(*
    Title: 	Fields - Implementation
    LastEdit:	"Fri Jan 29 10:48:44 1988"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE Fields;
(* $T-,$R- *)
FROM SYSTEM IMPORT BITSPERWORD, BYTESPERWORD, ADR, WORD;
FROM SystemTypes IMPORT Byte, Short;
IMPORT Endian;

VAR
    P2Mod: ARRAY [0 .. BITSPERWORD-1] OF CARDINAL;
    P2Div: ARRAY [0 .. BITSPERWORD-1] OF CARDINAL;

PROCEDURE Extract(v: WORD; low, high: WordBitRange): CARDINAL;
    VAR
        oh:WordBitRange;
    BEGIN
        IF Endian.Ours = Endian.Big THEN
            oh := high; high := 31 - low; low := 31 - oh
        END;
        IF low # 0 THEN
            v := WORD(CARDINAL(v) DIV P2Div[low]);
            DEC(high, low);
        END (* if *);
        IF high = BITSPERWORD-1 THEN
            RETURN CARDINAL(v)
        ELSE
            RETURN CARDINAL(v) MOD P2Mod[high];
        END (* if *);
    END Extract;

PROCEDURE Insert(VAR v: WORD; low, high: WordBitRange; field: CARDINAL);
    VAR
        i:  CARDINAL;
        oh: WordBitRange;
        bvp:POINTER TO BITSET;
    BEGIN
        IF Endian.Ours = Endian.Big THEN
            oh := high; high := 31 - low; low := 31 - oh
        END;
        IF low # 0 THEN
            field := field * P2Div[low];
        END (* if *);
        bvp := ADR(v);
        FOR i := low TO high DO
            EXCL(bvp^, i);
        END (* for *);
        v := WORD(BITSET(v) + BITSET(field));
    END Insert;

TYPE CHARPTR = POINTER TO ARRAY [0..BYTESPERWORD-1] OF CHAR;

PROCEDURE ExtractByte(v: WORD; byteno: ByteRange): Byte;
    VAR
        cp: CHARPTR;
    BEGIN
        cp := ADR(v);
        IF Endian.Ours = Endian.Little THEN
          RETURN Byte( cp^[byteno] )
        ELSE
          RETURN Byte( cp^[BYTESPERWORD -1 - byteno] )
        END
    END ExtractByte;

PROCEDURE InsertByte(VAR v: WORD; byteno: ByteRange; value: Byte);
    VAR
        cp: CHARPTR;
    BEGIN
        cp := ADR(v);
        IF Endian.Ours = Endian.Little THEN
          cp^[byteno] := CHAR( value )
        ELSE
          cp^[BYTESPERWORD -1 - byteno] := CHAR( value )
        END        
    END InsertByte;

PROCEDURE ExtractShort(v: WORD; shortno: ShortRange): Short;
    BEGIN
        IF Endian.Ours = Endian.Big THEN
            shortno := 1 - shortno
        END;
        IF shortno=0 THEN
            RETURN CARDINAL(v) MOD 65536
        ELSE
            RETURN CARDINAL(v) DIV 65536
        END;
    END ExtractShort;

PROCEDURE InsertShort(VAR v: WORD; shortno: ShortRange; value: Short);
    BEGIN
        IF Endian.Ours = Endian.Big THEN
            shortno := 1 - shortno
        END;
        IF shortno=0 THEN
            v := WORD( BITSET(v) - BITSET(0FFFFH) + BITSET(value));
        ELSE
            v := WORD (BITSET(v) - BITSET(0FFFF0000H) +
			BITSET(value * 010000H));
        END (* if *);
    END InsertShort;

PROCEDURE Initialise;
    VAR ix: CARDINAL;
    BEGIN
        P2Div[0] := 1; P2Mod[0] := 2;
        FOR ix := 1 TO BITSPERWORD-1 DO
            P2Div[ix] := P2Div[ix-1]*2;
            P2Mod[ix] := P2Div[ix]*2;
        END (* for *);
    END Initialise;

BEGIN
    Initialise;
END Fields.
