(*
    Title: 	FileStream - Implementation
    LastEdit:	"Wed Jan 27 10:49:43 1988"
    Author: 	Mick Jordan
		Acorn Research Centre

    Unix implementation
*)

(* $T-,$R- *)

(* $Header: FileStream.mod,v 1.1 88/01/11 14:32:14 dm Exp $ *)
(* $Log:	FileStream.mod,v $
Revision 1.1  88/01/11  14:32:14  dm
Initial revision

Revision 1.1  87/03/02  11:23:05  can
Initial revision
 *)

IMPLEMENTATION MODULE FileStream;

FROM SYSTEM IMPORT WORD, ADR, ADDRESS; IMPORT SYSTEM;
FROM SystemTypes IMPORT LongINTEGER, LongCARDINAL;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM Streams IMPORT 
    Create, StreamProcs, ErrorCode, SeekMode, Stream, OptionalFunction;
FROM StreamRep IMPORT
	DefaultProcs, SetStreamError, Procs;
FROM UnixIO IMPORT Read, Write, Open, Close, Lseek, Creat,
  OpenForRead, OpenForWrite, OpenForReadWrite;
IMPORT Strings;
IMPORT BadStream;

CONST bufUpb = 1023;
CONST bufLength = bufUpb+1;

TYPE BUFFERLENGTH = [0 .. bufLength];
     MODE = (In, Out, InOut);
     BUFFER = ARRAY [0 .. bufLength] OF CHAR;
     BUFFERPTR = POINTER TO BUFFER;

TYPE BUFFERDATA = POINTER TO 
    RECORD
    fildes: CARDINAL;			(* Unix file descriptor	      *)
    size: BUFFERLENGTH;		        (* size of current input buffer*)
    streambufno: CARDINAL;              (* data pair specifying *)
    streambufoffset: BUFFERLENGTH;      (* position on stream *)
    filebufno: CARDINAL;                (* data pair specifying position *)
    filebufoffset: BUFFERLENGTH;        (* in file *)
    bufferismodified: BOOLEAN;
    mode: MODE;
    stream: Stream;
    eof: BOOLEAN;
    buf: BUFFER;
    END;

PROCEDURE CurPos(b: BUFFERDATA; VAR cp: LongCARDINAL);
(*  Combines the buffer count and byte count into a character pointer *)
    BEGIN
        WITH b^ DO
            cp := streambufno * bufLength + streambufoffset;
        END;
    END CurPos;

PROCEDURE SeekToBuffer(b: BUFFERDATA): BOOLEAN;
(*  makes sure that the file position matches that of the stream position.
*)
    VAR
        r: INTEGER;
        seeko: LongINTEGER;
    BEGIN
      WITH b^ DO
        IF (streambufno # filebufno) OR (filebufoffset # 0) THEN
            seeko := streambufno * bufLength;
            r := Lseek(fildes, seeko, 0);
            IF r<0 THEN
                SetStreamError(b^.stream, r);
            ELSE    
            (* note that we are positioned at the beginning of the buffer *)
                filebufno := streambufno;
                filebufoffset := 0;
                eof := FALSE;
                RETURN TRUE
            END (* if *);
        END (* if *);
      END (* with *);
      RETURN FALSE;
    END SeekToBuffer;
        
PROCEDURE SetupBuffer(f: CARDINAL; m: MODE): BUFFERDATA;
VAR b: BUFFERDATA;
BEGIN    
    NEW(b); (* allocate bufferdata record *)
    WITH b^  DO
        size := 0;
        streambufno := 0; streambufoffset := 0;
        filebufno := 0; filebufoffset := 0;
        fildes := f;
        mode := m;
	bufferismodified := FALSE; (* Bug fixed by, Cosmos Nicolaou 19/2/87. *)
        eof := FALSE;
    END;
    RETURN b;
END SetupBuffer;

PROCEDURE WriteBuffer(b: BUFFERDATA): INTEGER;
VAR r: INTEGER;
    dummy: BOOLEAN;
BEGIN
  WITH b^ DO
    IF bufferismodified THEN
        dummy := SeekToBuffer(b);
        r := Write(fildes, ADR(buf), streambufoffset);
	IF r <> INTEGER(streambufoffset) THEN
            SetStreamError(b^.stream, r);
	ELSE 
        (*  note that the file pointer has advanced cos of the write *)
            IF streambufoffset = bufLength THEN
                INC(filebufno);
		(*
		 *	PreRead incase the next read is also for this
		 *	buffer.
		 *	Bug fixed, Cosmos Nicolaou 25/2/87. 
                 *      Changed to not set error code if past end of file.
                 *      Mac, 27/1/87
		 *)
		ReadBufferNoError(b);
            ELSE
                filebufoffset := streambufoffset;
            END (* if *);
            bufferismodified := FALSE;
	END;
    END;
  END;
  RETURN r;
END WriteBuffer;

PROCEDURE ReadBufferNoError(b:BUFFERDATA);
  VAR
    r:INTEGER;
  BEGIN
    WITH b^ DO
      r := Read(fildes, buf, bufLength);
      IF r>0 THEN
        size := BUFFERLENGTH(r);
        IF size = bufLength THEN
            INC(filebufno);
        ELSE
            filebufoffset := size;
        END;
      END;
    END; (* WITH *)
  END ReadBufferNoError;
  
PROCEDURE ReadBuffer(b: BUFFERDATA): INTEGER;
VAR r: INTEGER;
BEGIN
  WITH b^ DO
    r := Read(fildes, buf, bufLength);
    IF r<0 THEN
        SetStreamError(b^.stream, r);
    ELSIF r=0 THEN
        b^.stream^.status := EndOfStream;
        eof := TRUE;
    ELSE
        IF size > 0 THEN    (* not first read (or after a SEEK) *)
            INC(streambufno);
            streambufoffset :=0;
        END;
        size := BUFFERLENGTH(r);
        IF size = bufLength THEN
            INC(filebufno);
        ELSE
            filebufoffset := size;
        END;
    END;
    RETURN r;
END; (* WITH *)
END ReadBuffer;


PROCEDURE Implements(s: Stream; f: OptionalFunction): BOOLEAN;
    BEGIN
        RETURN TRUE;
    END Implements;

PROCEDURE Get(s: Stream): WORD;
    VAR
        b: BUFFERDATA;
        r: INTEGER;
        ch: CHAR;
    BEGIN
        b := BUFFERDATA(s^.instanceData);
        WITH b^ DO
            IF (streambufoffset >= size) THEN 
                r := ReadBuffer(b);
                IF r<=0 THEN RETURN WORD(0); END;
            END (* if *);
            ch := buf[streambufoffset]; INC(streambufoffset);
            RETURN WORD(ch);
        END (* with *);
    END Get;

PROCEDURE GetN(s: Stream; a: ADDRESS; n: CARDINAL): CARDINAL;
    VAR
        b: BUFFERDATA;
        i, bytesLeft, toGet, totalGet: CARDINAL;
        getNBuf: BUFFERPTR;
    BEGIN
        b := BUFFERDATA(s^.instanceData);
        WITH b^ DO
            getNBuf := BUFFERPTR(a);
            totalGet := 0;
            WHILE n > 0 DO
                IF streambufoffset >= size THEN
                    IF ReadBuffer(b) <= 0 THEN
                        RETURN totalGet;
                    END (* if *);
                END (* if *);
                bytesLeft := size - streambufoffset;
                IF bytesLeft < n THEN
                    toGet := bytesLeft
                ELSE
                    toGet := n;
                END (* if *);
                FOR i := 0 TO toGet-1 DO
                    getNBuf^[totalGet+i] := buf[streambufoffset+i];
                END (* for *);
                INC(streambufoffset, toGet);
                DEC(n, toGet); INC(totalGet, toGet);
            END (* while *);
        END (* with *);
        RETURN totalGet;
    END GetN;

PROCEDURE Backspace(s: Stream);
    BEGIN
        Seek(s, Current, -1);
    END Backspace;

PROCEDURE Put(s: Stream; w: WORD);
    VAR
        b: BUFFERDATA;
        r: INTEGER;
    BEGIN
        b := BUFFERDATA(s^.instanceData);
        WITH b^ DO
            bufferismodified := TRUE;
            buf[streambufoffset] := CHAR(w);
            INC(streambufoffset);
            IF streambufoffset > bufUpb THEN
                r := WriteBuffer(b);
                IF NOT (r<0) THEN   (* no error occurred *)
                    INC(streambufno);
                    streambufoffset := 0;
                END (* if *);
            END (* if *);
        END (* with *);
    END Put;

PROCEDURE PutN(s: Stream; a: ADDRESS; n: CARDINAL): CARDINAL;
    VAR
        b: BUFFERDATA;
        r: INTEGER;
        i, freeBytes, toPut, totalPut: CARDINAL;
        putNBuf: BUFFERPTR;
    BEGIN
        b := BUFFERDATA(s^.instanceData);
        WITH b^ DO
            putNBuf := BUFFERPTR(a);
            totalPut := 0; r := 0;
            WHILE (n > 0) AND (r >= 0) DO
                freeBytes := bufLength - streambufoffset;
                IF freeBytes < n THEN
                    toPut := freeBytes
                ELSE
                    toPut := n;
                END (* if *);
                IF toPut > 0 THEN
                    FOR i := 0 TO toPut-1 DO
                        buf[streambufoffset+i] := putNBuf^[totalPut+i]
                    END (* for *);
                    INC(streambufoffset, toPut);
                    bufferismodified := TRUE;
                END (* if *);
                IF streambufoffset > bufUpb THEN
                    r := WriteBuffer(b);
                    IF NOT (r<0) THEN   (* no error occurred *)
                        INC(streambufno);
                        streambufoffset := 0;
                    END (* if *);
                END (* if *);
                DEC(n, toPut);
                INC(totalPut, toPut);
            END (* while *);
        END (* with *);
        RETURN totalPut;
    END PutN;

PROCEDURE Flush(s: Stream);
    VAR
        b: BUFFERDATA;
        r: INTEGER;
    BEGIN
        b := BUFFERDATA(s^.instanceData);
        IF b^.mode # In THEN
            r := WriteBuffer(b);
        END (* if *);
    END Flush;

PROCEDURE Delete(s: Stream);
    VAR
        b: BUFFERDATA;
        r: INTEGER;
    BEGIN
        b := BUFFERDATA(s^.instanceData);
        IF b^.mode # In THEN
            r := WriteBuffer(b);
        END (* if *);
	r := Close(b^.fildes);
        IF (s^.status=Success) & (r<0) THEN
            SetStreamError(b^.stream, r);
        END (* if *);
	DISPOSE(b);
    END Delete;

PROCEDURE Tell(s: Stream; VAR p: LongCARDINAL);
    VAR
        b: BUFFERDATA;
    BEGIN
        b := BUFFERDATA(s^.instanceData);
        CurPos(b, p);
    END Tell;

PROCEDURE Seek(s: Stream; sm: SeekMode; offset: LongINTEGER); 
(*  Computes the true file offset from the sm, offset pair, 
    where the interpretation of offset depends on sm. (see Streams.def)
*)
VAR 
    b: BUFFERDATA;
    r: INTEGER;
    trueoffset: LongINTEGER;
    curpos: LongCARDINAL;
    newbuf, newstreambufoffset: CARDINAL;
    void: BOOLEAN;
BEGIN
    b := BUFFERDATA(s^.instanceData);
    CASE sm OF
      Beginning:        
        trueoffset := offset;
    | Current:
        IF b^.eof THEN s^.status := EndOfStream; RETURN END;
        CurPos(b, curpos);
        trueoffset := LongINTEGER(curpos) + offset;
    | End:
        trueoffset := 0;
        IF b^.bufferismodified THEN r := WriteBuffer(b) END (* if *);
        r := Lseek(b^.fildes, trueoffset, CARDINAL(sm));
        (*  ok, so now we know where we are *)
        b^.filebufno := trueoffset DIV bufLength;
        b^.filebufoffset := trueoffset MOD bufLength;
        (*  so this is where we want to go to *)
        trueoffset := trueoffset + offset;
    END;

    newbuf := trueoffset DIV bufLength;
    newstreambufoffset := trueoffset MOD bufLength;

    IF b^.bufferismodified THEN r := WriteBuffer(b); END (* if *);
    (* always flush, cos dont keep a high water mark *)
    IF (newbuf # b^.streambufno) OR (b^.size=0) (* new file *) THEN
    (*  a change of buffer is needed  *)
        b^.streambufno := newbuf;
        void := SeekToBuffer(b);
        b^.size := 0;
        r := ReadBuffer(b)
    END (* if *);
    b^.streambufoffset := newstreambufoffset;
END Seek;

TYPE OPENBLOCK = RECORD
        CASE mode: MODE OF
          InOut, Out:  fileMode: FileMode
        | In:   (* no field *)
        END (* case *);
        procs: Procs;
    END (* record *);

PROCEDURE GenericOpen(m: OPENBLOCK; name: ARRAY OF CHAR): Stream;
    VAR
        unixName: ARRAY [0 .. 255] OF CHAR;
        b: BUFFERDATA;
        r: INTEGER;
        s: Stream;
    BEGIN
    (*  pad the name with a null, as required by Unix.
        This is automatic in the Strings.CopyCC procedure.
    *)
        Strings.CopyCC(name, unixName);

        IF m.mode = In THEN
	    r := Open(unixName, OpenForRead, {});
        ELSIF m.mode = Out THEN
            r := Creat(unixName, BITSET(m.fileMode));
        ELSE
            r := Open(unixName, OpenForReadWrite, {});
            IF (r<0) AND (r=-2) THEN   (* file not found *)
                r := Creat(unixName, BITSET(m.fileMode));
                IF r>0 THEN
                    r := Close(r);
                    r := Open(unixName, OpenForReadWrite, {});
                END (* if *);
            END (* if *);
        END (* if *);
        
        IF r < 0 THEN
            s := BadStream.CreateInOutput();
            SetStreamError(s, -r);
        ELSE
	    b := SetupBuffer(r, m.mode);
            s := Create(StreamProcs(m.procs), b);
            b^.stream := s;
            s^.instanceData := WORD(b);
        END (* if *);
        RETURN s;
    END GenericOpen;

PROCEDURE CreateInput(name: ARRAY OF CHAR): Stream;
(*  Create an input stream to the standard input *)    
    BEGIN
        RETURN GenericOpen(InBlock, name);
    END CreateInput;

PROCEDURE CreateOutput(name: ARRAY OF CHAR): Stream;
(*  Create an output stream to the file called 'name'  *)
    BEGIN
        RETURN GenericOpen(OutBlock, name);
    END CreateOutput;

PROCEDURE CreateInOutput(name: ARRAY OF CHAR): Stream;
    BEGIN
        RETURN GenericOpen(InOutBlock, name);
    END CreateInOutput;

PROCEDURE CreateOutputWithMode(name: ARRAY OF CHAR; mode: FileMode): Stream;
    VAR
        openBlock: OPENBLOCK;
    BEGIN
        openBlock := OutBlock;
        openBlock.fileMode := mode;
        RETURN GenericOpen(openBlock, name);
    END CreateOutputWithMode;

PROCEDURE StreamErrorCode(s: Stream): INTEGER;
    BEGIN

	(* Another bug fixed, Cosmos Nicolaou 23/2/87. *)
	(*
	 *	Streams.create does NOT initialise streamError, only
	 *	status, therefore must check status before reporting 
	 *	an error.
	 *)
	IF s^.status # Success THEN
       	  RETURN INTEGER(s^.streamError);
	ELSE
          RETURN( 0 );
	END;

    END StreamErrorCode;

PROCEDURE InitProcs(mode: MODE): Procs;
    VAR
        pt: Procs;
    BEGIN
        pt := DefaultProcs(8);
        IF mode # In THEN
            pt^.Put := Put; pt^.PutN := PutN;
        END (* if *);
        IF mode # Out THEN 
            pt^.Get := Get; pt^.GetN := GetN;
            pt^.Backspace := Backspace;
        END (* if *);
        pt^.Flush := Flush; 
        pt^.Delete := Delete;
        pt^.Implements := Implements;
        pt^.Seek := Seek;
        pt^.Tell := Tell;
        RETURN pt;
    END InitProcs;

VAR     
    InBlock, OutBlock, InOutBlock: OPENBLOCK;
CONST
    DefaultMode = ReadByOwner+WriteByOwner
                  +ReadByGroup+WriteByGroup
                  +ReadByWorld+WriteByWorld;

BEGIN
    InBlock.mode := In; InBlock.procs := InitProcs(In);
    OutBlock.mode := Out; OutBlock.procs := InitProcs(Out);
    OutBlock.fileMode := DefaultMode;
    InOutBlock.fileMode := DefaultMode;
    InOutBlock.mode := InOut; InOutBlock.procs := InitProcs(InOut);

END FileStream .
