(*
    Title: 	Opening Modula-2 files - Implementation
    LastEdit:	"Thu Jan 28 18:38:17 1988"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory
*)

(*  This version provides the capability to hash files
    with a given extension but also supports the long-winded
    lookup.

    23-May-83   Support for reading from archive files
    12-Mar-84   Support for 4.1/4.2 directories
    22-Mar-84   Implement SetM2PathFromFile
    05-Sep-84   Update to new library/drop 4.1 support/
                /long module names/lose '.ext' in hidden files
    30-Apr-86   Vax byte order changed to 68000 byte order
*)

IMPLEMENTATION MODULE M2Files;

FROM SYSTEM IMPORT ADR, CALLEXTFUNC, CALLEXTPROC, WORD;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
IMPORT UnixEnv; FROM UnixEnv IMPORT GetEnvVar, SetEnvVar; 
FROM Streams IMPORT 
    Stream, ErrorCode, StreamProcs, Get, GetN, Create, Delete;
IMPORT BadStream;
FROM SysStreams IMPORT sysOut;
FROM TextIO IMPORT WriteChars, NewLine, ReadChars, WhiteSpace, EndOfLine;
FROM WriteF IMPORT WriteF1, H;
IMPORT HashStrings; 
FROM HashStrings IMPORT HashTable, HashId,
    EnterString, EnterChars, LookupChars, Assoc, Retrieve;
FROM UnixArchive IMPORT 
    ArchiveId, MemberId, OpenArchive, FindNextMember, CreateFastInput;
IMPORT Strings; FROM Strings IMPORT String;
IMPORT FileStream; 
FROM UnixVersion IMPORT UnixVersion, GetUnixVersion;
FROM Fields IMPORT ExtractShort;
IMPORT UnixLibCall;

CONST
	MAXNAMLEN=255;

TYPE FPROC = PROCEDURE(ARRAY OF CHAR): Stream;

TYPE FILEDATA = POINTER TO RECORD
        CASE inarchive: BOOLEAN OF
          TRUE:
              archiveid: ArchiveId;
              memberid: MemberId;
              archivename: String;
        | FALSE:
              directory: String;
        END (* case *);
    END (* record *);

TYPE
	DirHandle = POINTER TO Dir;
	Dir = RECORD
		ddFd:	INTEGER;
		ddLoc:	INTEGER;
		ddSize:	INTEGER;
		ddBbase:	INTEGER;
		ddEntno:	INTEGER;
		ddBsize:	INTEGER;
		ddBuf:	String;
	END;

	PtrDirect = POINTER TO Direct;
	Direct = RECORD
		dFileno: INTEGER;		(* file number of entry *)
		kludge: CARDINAL;
(* Kludge includes these two fields. *)
(* u_short *d_reclen;		(* length of this record *)*)
(* u_short *d_namlen;		(* length of string in d_name *)*)

		(* name (up to MAXNAMLEN + 1) *)
		dName: ARRAY [0..MAXNAMLEN + 1] OF CHAR;
	END;

	PtrSun43Direct = POINTER TO Sun43Direct;
	Sun43Direct = RECORD
		doff: INTEGER;
		dFileno: INTEGER;		(* file number of entry *)
		kludge: CARDINAL;
(* Kludge includes these two fields. *)
(* u_short *d_reclen;		(* length of this record *)*)
(* u_short *d_namlen;		(* length of string in d_name *)*)

		(* name (up to MAXNAMLEN + 1) *)
		dName: ARRAY [0..MAXNAMLEN + 1] OF CHAR;
	END;

VAR 
    dot, slash, separator: ARRAY [0 .. 0] OF CHAR;
    slashdot: ARRAY [0 .. 1] OF CHAR;
    archive: ARRAY [0 .. 10] OF CHAR;
    notify: BOOLEAN;
    extht, ht: HashTable;

CONST MaxFileLength = 255;  (* set to 10 for BSD4.1 *)

TYPE MODE = (In, Out, InOut);

PROCEDURE Open(m: MODE; name: ARRAY OF CHAR; ext: Extension): Stream;
VAR l: CARDINAL;
    s: Stream;
    dext: ARRAY [0 .. 5] OF CHAR;
    directoryPrefix: ARRAY [0..255] OF CHAR;
    fproc: FPROC;
    hashid: HashId;
    filedata: FILEDATA;
BEGIN
  (* have we hashed this extension type? *)
  IF (m=Out) OR LookupChars(extht, ext, hashid) THEN	
    Strings.CopyCC(name, Filename);
    Filename[MaxFileLength] := 0C;	(*  truncate to allowed limit *)
    ext[HIGH(ext)] := 0C;               (*  just in case *)
    dext[0] := '.'; dext[1] := 0C;
    Strings.ConcatCC(dext, ext, dext);		(*  'xxx' -> '.xxx' *)
    

    CASE m OF
      In:   fproc := FileStream.CreateInput;
    | Out:  fproc := FileStream.CreateOutput;
    | InOut: fproc := FileStream.CreateInOutput;
    END (* case *);

        IF HiddenFile(ext) THEN
            Strings.ConcatCC(dext, slash, dext);
            Strings.ConcatCC(dext, Filename, Filename);
        ELSE
            Strings.ConcatCC(Filename, dext, Filename);	
            (*  'name' -> 'name.xxx' *)
        END;
    
    IF m=Out THEN
        s := Try(Filename, fproc);
    ELSE
        IF LookupChars(ht, Filename, hashid) THEN
            filedata := FILEDATA( Retrieve(ht, hashid) );
            WITH filedata^ DO
                IF inarchive THEN
                    s := CreateFastInput(archiveid, memberid);
                    Strings.CopySC(archivename, Filename);
                ELSE
                    Strings.CopySC(directory, directoryPrefix);
                    Strings.ConcatCC(directoryPrefix, slash, directoryPrefix);
                    Strings.ConcatCC(directoryPrefix, Filename, Filename);
                    s := Try(Filename, fproc);
                END (* if *);
            END (* with *);
        ELSE
            s:= BadStream.CreateInOutput();
        END (* if *);
    END (* if *);
        
    RETURN s;				(*  will be 'invalid' if failed *)
  ELSE
    RETURN SlowOpen(m, name, ext);
  END (* if *)
END Open;

PROCEDURE Try(name: ARRAY OF CHAR; fproc: FPROC): Stream;
BEGIN
    IF notify THEN
        WriteChars(sysOut, "Trying ");
        WriteChars(sysOut, name); NewLine(sysOut);
    END (* if *);
    RETURN fproc(name);
END Try;

PROCEDURE CreateInput(name: ARRAY OF CHAR; ext: Extension): Stream;
BEGIN
    RETURN Open(In, name, ext);
END CreateInput;

PROCEDURE CreateOutput(name: ARRAY OF CHAR; ext: Extension): Stream;
BEGIN
    RETURN Open(Out, name, ext);
END CreateOutput;

PROCEDURE HashExt(ext: Extension);
    VAR
        m2path: ARRAY [0 .. 2047] OF CHAR;
        m2dir: ARRAY [0 .. 79] OF CHAR;
        dext: ARRAY [0..5] OF CHAR;
        endvar: BOOLEAN;
        firstindex, lastindex: CARDINAL;
        rc: INTEGER; envrc: UnixEnv.EnvRc;
        hiddenfile: BOOLEAN;
        hashid: HashId;
        unixVersion: UnixVersion;
        dirL: CARDINAL;
    BEGIN
      unixVersion := GetUnixVersion();
      IF unixVersion = Unknown THEN
      (*  this is interpreted as an instruction to suppress all hashing *)
        RETURN
      END (* if *);

      ext[HIGH(ext)] := 0C;                   (*  just in case *)
      IF EnterChars(extht, ext, hashid) THEN	      (*  new type *)
        hiddenfile := HiddenFile(ext);
        endvar := TRUE; firstindex := 0; lastindex := 0;
        m2path[0] := 0C;
        envrc := GetEnvVar("M2PATH", m2path);
        LOOP
            endvar := Strings.FindC(m2path, ':', lastindex);
            IF lastindex=firstindex THEN EXIT END;
            (* extract first directory to search *)
            Strings.ExtractCC(m2path, firstindex, lastindex, m2dir);
            dirL := lastindex - firstindex;
            INC(lastindex);
            firstindex :=  lastindex;

            IF hiddenfile THEN
                Strings.ConcatCC(m2dir, slashdot, m2dir);
                Strings.ConcatCC(m2dir, ext, m2dir);
                dext[0] := '.'; dext[1] := 0C;
                Strings.ConcatCC(dext, ext, dext); (*  'xxx' -> '.xxx' *)
                Strings.ConcatCC(dext, slash, dext); (* -> '.xxx/' *)
            ELSE
                dirL := 0;
                dext[0] := 0C;
            END;
            IF ReadArchive(m2dir, dext) THEN
            (*  use that instead of directory data *)
            ELSE
                ReadDirectory(m2dir, dirL, dext, unixVersion);
            END (* if *);
        END (* loop *);
      END (* if *)        
    END HashExt;

PROCEDURE ReadDirectory(m2dir: ARRAY OF CHAR; dirL: CARDINAL;
        dext: ARRAY OF CHAR; unixVersion: UnixVersion);
    VAR
        direntry: ARRAY [0 .. 255] OF CHAR;
        m2dirstring: String;
        s: Stream;
        filedata: FILEDATA;
        hashid: HashId;
        direct: PtrDirect;	
	Sun43direct: PtrSun43Direct;
        d : DirHandle;

    PROCEDURE GetNextEntry(): BOOLEAN;
        VAR
            stuff: ARRAY [0..1] OF CHAR;
            n: CARDINAL;
            ino: CARDINAL;
            discard: CHAR;
            reclen, namlen: CARDINAL;

        PROCEDURE RdShort(): CARDINAL;
            VAR
                msb: CARDINAL;
            BEGIN
                msb := CARDINAL(Get(s));
                RETURN CARDINAL(Get(s)) + msb * 256;
            END RdShort;

        PROCEDURE RdCard(): CARDINAL;
        VAR mss : CARDINAL;
	BEGIN
            mss := RdShort();
	    RETURN RdShort() + (mss * 65536);
        END RdCard;
        
        BEGIN
          IF unixVersion = BSD41 THEN
            LOOP
                stuff[0] := CHAR( Get(s) ); 
                IF s^.status = EndOfStream THEN RETURN FALSE END;
                stuff[1] := CHAR( Get(s) );
                n := GetN(s, ADR(direntry), 14);
                (* ensure 14 char names properly terminated *)
                direntry[14] := 0C; 
                IF NOT ((stuff[0] = 0C) AND (stuff[1] = 0C)) THEN
                    EXIT
                END (* if *);
            END (* loop *);
          ELSE  (* BSD4.2 OR ULTRIX1.2 *)
            LOOP
                ino := RdCard();
                IF s^.status = EndOfStream THEN RETURN FALSE END;
                reclen := RdShort();
                namlen := RdShort();
                n := 1;
                WHILE (n <= reclen-8) AND (s^.status # EndOfStream) DO
                    IF n>HIGH(direntry) THEN
                        discard := CHAR( Get(s) );
                    ELSE
                        direntry[n-1] := CHAR( Get(s) );
                    END (* if *);
                    INC(n);
                END (* for *);
                IF namlen <= HIGH(direntry) THEN
                    direntry[namlen] := 0C;
                END (* if *);
                IF ino # 0 THEN
                (* genuine entry *)
                    EXIT
                END (* if *);
            END (* loop *);
          END (* if *);
          RETURN TRUE;
        END GetNextEntry;
    BEGIN
    IF unixVersion = BSD41 THEN
        s := FileStream.CreateInput(m2dir);
        IF s^.status = Success THEN
            m2dirstring := Strings.CopyCS(m2dir);
            IF dirL # 0 THEN
                (* truncate back to just the directory *)
                m2dirstring^[dirL] := 0C;
            END (* if *);
            LOOP
                IF GetNextEntry() THEN
                    Strings.ConcatCC(dext, direntry, direntry);
                    IF EnterChars(ht, direntry, hashid) THEN
                        NEW(filedata);
                        WITH filedata^ DO
                            inarchive := FALSE;
                            directory := m2dirstring;
                        END (* with *);
                        Assoc(ht, hashid, filedata);
                    END (* if *);   
                ELSE
                    EXIT
                END (* if *);
            END (* loop *);
            Delete(s); 
            DISPOSE(s)
        END (* if *);
    ELSIF unixVersion = SUN43 THEN
        m2dirstring := Strings.CopyCS(m2dir);
	d := DirHandle (CALLEXTFUNC (UnixLibCall.opendir, m2dirstring));
	IF CARDINAL (d) # 0 THEN 
            IF dirL # 0 THEN
                (* truncate back to just the directory *)
                m2dirstring^[dirL] := 0C;
            END (* if *);
            LOOP
		Sun43direct := PtrSun43Direct (CALLEXTFUNC (
			UnixLibCall.readdir, d));
                IF CARDINAL(Sun43direct) # 0 THEN
		    Strings.ExtractCC (Sun43direct^.dName, 0,
				ExtractShort(Sun43direct^.kludge, 1) + 1,
			        direntry);
                    Strings.ConcatCC(dext, direntry, direntry);
                    IF EnterChars(ht, direntry, hashid) THEN
                        NEW(filedata);
                        WITH filedata^ DO
                            inarchive := FALSE;
                            directory := m2dirstring;
                        END (* with *);
                        Assoc(ht, hashid, filedata);
                    END (* if *);   
                ELSE
                    EXIT
                END (* if *);
            END (* loop *);
            CALLEXTPROC(UnixLibCall.closedir, d)
	END; (* IF CARDINAL (d) # 0 *)
    ELSE (*  BSD42 OR ULTRIX1.2 *)
        m2dirstring := Strings.CopyCS(m2dir);
	d := DirHandle (CALLEXTFUNC (UnixLibCall.opendir, m2dirstring));
	IF CARDINAL (d) # 0 THEN 
            IF dirL # 0 THEN
                (* truncate back to just the directory *)
                m2dirstring^[dirL] := 0C;
            END (* if *);
            LOOP
		direct := PtrDirect (CALLEXTFUNC (UnixLibCall.readdir, d));
                IF CARDINAL(direct) # 0 THEN
		    Strings.ExtractCC (direct^.dName, 0,
				ExtractShort(direct^.kludge, 1) + 1,
			        direntry);
                    Strings.ConcatCC(dext, direntry, direntry);
                    IF EnterChars(ht, direntry, hashid) THEN
                        NEW(filedata);
                        WITH filedata^ DO
                            inarchive := FALSE;
                            directory := m2dirstring;
                        END (* with *);
                        Assoc(ht, hashid, filedata);
                    END (* if *);   
                ELSE
                    EXIT
                END (* if *);
            END (* loop *);
            CALLEXTPROC(UnixLibCall.closedir, d)
	END; (* IF CARDINAL (d) # 0 *)
    END;
    END ReadDirectory;

PROCEDURE ReadArchive(m2dir: ARRAY OF CHAR; dext: ARRAY OF CHAR): BOOLEAN;
(*  returns TRUE if the archive exists *)
    VAR
        m2archive: ARRAY [0 .. 255] OF CHAR;
        m2archivestring: String;
        arcid: ArchiveId;
        memid: MemberId;
        memname: ARRAY [0 .. 19] OF CHAR;
        hashid: HashId;
        filedata: FILEDATA;
    BEGIN        
        Strings.ConcatCC(m2dir, archive, m2archive);
        IF OpenArchive(m2archive, arcid) THEN
            m2archivestring := Strings.CopyCS(m2archive);
            WHILE FindNextMember(arcid, memname, memid) DO
                Strings.ConcatCC(dext, memname, memname);
                IF EnterChars(ht, memname, hashid) THEN
                    NEW(filedata);
                    WITH filedata^ DO
                        inarchive := TRUE;
                        archiveid := arcid;
                        memberid := memid;
                        archivename := m2archivestring;
                    END (* with *);
                    Assoc(ht, hashid, filedata);
                END (* if *);
            END (* while *);
            RETURN TRUE;
        ELSE
            RETURN FALSE;
        END (* if *);
    END ReadArchive;


PROCEDURE HiddenFile(ext: Extension): BOOLEAN;
    BEGIN
       RETURN Strings.EqualCC(ext, "sym") OR Strings.EqualCC(ext, "ref") OR
              Strings.EqualCC(ext, "lnk") OR Strings.EqualCC(ext, "key");
    END HiddenFile;

PROCEDURE CreateInOutput(name: ARRAY OF CHAR; ext: Extension): Stream;
BEGIN
    RETURN Open(InOut, name, ext);
END CreateInOutput;

PROCEDURE SlowOpen(m: MODE; name: ARRAY OF CHAR; ext: Extension): Stream;
VAR l, firstindex, lastindex: CARDINAL;
    s: Stream;
    tFilename: ARRAY [0 .. 255] OF CHAR;
    m2path: ARRAY [0 .. 2047] OF CHAR;
    dext: ARRAY [0 ..  4] OF CHAR;
    m2dir: ARRAY [0 .. 255] OF CHAR;
    endvar: BOOLEAN;
    hiddenfile: BOOLEAN;
    fproc: FPROC;
    rc: UnixEnv.EnvRc;
BEGIN
    s := BadStream.CreateInOutput();    (* failure *)
    endvar := TRUE; firstindex := 0; lastindex := 0;
(*  append extension *)
    Strings.CopyCC(name, Filename);
    Filename[MaxFileLength] := 0C;	(*  truncate to allowed limit *)
    ext[HIGH(ext)] := 0C;               (*  just in case *)
    dext[0] := '.'; dext[1] := 0C;
    Strings.ConcatCC(dext, ext, dext);		(*  'xxx' -> '.xxx' *)
    
    rc := GetEnvVar("M2PATH", m2path);
    IF rc=UnixEnv.NotFound THEN RETURN s END;
    
    
    IF HiddenFile(ext) THEN
        hiddenfile := TRUE;
            Strings.ConcatCC(dext, slash, dext);
            Strings.ConcatCC(dext, Filename, Filename);
    ELSE
        hiddenfile := FALSE;
        Strings.ConcatCC(Filename, dext, Filename);	
        (*  'name' -> 'name.xxx' *)
    END;
    Strings.CopyCC(Filename, tFilename);	(*  save *)

    CASE m OF
      In:   fproc := FileStream.CreateInput;
    | Out:  fproc := FileStream.CreateOutput;
    | InOut: fproc := FileStream.CreateInOutput;
    END (* case *);
    
    LOOP
        endvar := Strings.FindC(m2path, ':', lastindex);
        IF lastindex=firstindex THEN EXIT END;
        (* extract first directory to search *)
        Strings.ExtractCC(m2path, firstindex, lastindex, m2dir);
        INC(lastindex);
        firstindex :=  lastindex;
        
        IF (m2dir[0]='/') AND (m=Out) THEN
        (*  don't try here *)
        ELSE
            Strings.ConcatCC(m2dir, slash, m2dir);
            Strings.CopyCC(tFilename, Filename);	(*  recover *)
            Strings.ConcatCC(m2dir, Filename, Filename);
            s := Try(Filename, fproc);
            IF s^.status = Success THEN EXIT END;
        END;
    END;
    RETURN s;				(*  will be 'invalid' if failed *)
END SlowOpen;

PROCEDURE SetM2PathFromFile;
(*  the file 'm2path' should be of the form:
    setenv M2PATH directory-string
*)
    VAR
        s: Stream;
        line: ARRAY [0..2047] OF CHAR;
        n, x: CARDINAL;
        envrc: UnixEnv.EnvRc;
    BEGIN
        s := FileStream.CreateInput("m2path");
        IF s^.status = Success THEN
            ReadChars(s, line, EndOfLine);
            Delete(s); 
            DISPOSE(s);
            x := 0; n := Strings.LengthC(line);
            (* move to second space *)
            IF Strings.FindC(line, 'H', x) THEN
                INC(x);
                (*  skip  spaces  *)
                WHILE (x<n) AND (line[x] = ' ') AND (line[x] # 0C) DO
                    INC(x);
                END (* while *);
                Strings.ExtractCC(line, x, n, line);
                envrc := SetEnvVar("M2PATH", line);
            END (* if *);
        END (* if *);
    END SetM2PathFromFile;

BEGIN
(* construct single character strings ! *)
    dot[0] := '.'; 
    slash[0] := '/';
    slashdot := "/.";
    archive := "/archive.ar";
    notify := (*TRUE*)FALSE;
    ht := HashStrings.New(128); extht := HashStrings.New(9);
END M2Files.
