IMPLEMENTATION MODULE MCLib;


IMPORT
  Strings, UnixLibCall;

FROM SYSTEM IMPORT ADDRESS, BYTESPERWORD, CALLEXTPROC, WORD;
FROM SystemTypes IMPORT String;


PROCEDURE MoveBytes(from, to:BytePtr; n:CARDINAL);
  BEGIN
    CALLEXTPROC(UnixLibCall.bcopy, from, to, n)
  END MoveBytes;

PROCEDURE MoveWords(from, to:ADDRESS; nOfWords:CARDINAL);
  BEGIN
    CALLEXTPROC(UnixLibCall.bcopy, from, to, nOfWords * BYTESPERWORD)
  END MoveWords;

PROCEDURE MoveOverlapBytes(from, to:BytePtr; n:CARDINAL);
  BEGIN
    CALLEXTPROC(UnixLibCall.bcopy, from, to, n)
  END MoveOverlapBytes;

PROCEDURE ClearBytes(at:BytePtr; n:CARDINAL);
  BEGIN
    CALLEXTPROC(UnixLibCall.bzero, at, n) 
  END ClearBytes;

PROCEDURE SetBytes(at:BytePtr; n:CARDINAL; to:CHAR);
  BEGIN
    WHILE n # 0 DO
      at^[0] := to;
      at := BytePtr(ADDRESS(at) + 1);
      INC(n);
    END;
  END SetBytes;

PROCEDURE SetWords(at:ADDRESS; nOfWords:CARDINAL; to:CARDINAL);
  BEGIN
    WHILE nOfWords # 0 DO
      at^ := WORD(to);
      at := at + BYTESPERWORD;
      DEC(nOfWords);
    END;
  END SetWords;


CONST
  Magic = - 1664117991;


PROCEDURE HashBytes(at:BytePtr; n:INTEGER; mod:CARDINAL):CARDINAL;
  VAR
    i:INTEGER;
  BEGIN
    IF n < 0 THEN n := Strings.LengthS(at) END;
    IF n > 3 THEN RETURN HashWords(at, n DIV 4, mod) END;
    FOR i := 0 TO n-1 DO
      n := n * Magic + INTEGER(ORD(at^[i]));
    END;
    RETURN CARDINAL(n) MOD mod
  END HashBytes;


PROCEDURE HashWords(at:ADDRESS; n, mod:CARDINAL):CARDINAL;
  VAR
    by,
    hash:INTEGER;
    lim: ADDRESS;
  BEGIN
    lim := at + (n * BYTESPERWORD);
    hash := n;
    by := BYTESPERWORD * (n DIV 5); 
    IF by = 0 THEN by := 4 END;
    WHILE at < lim DO
      hash := hash * Magic + INTEGER(at^);
      INC(at, by)
    END;
    RETURN CARDINAL(hash) MOD mod
  END HashWords;


END MCLib.
