IMPLEMENTATION MODULE OSStorage;
(*
    Title: 	UNIX Heap Allocator for bytes addressed machines
*)

FROM SYSTEM IMPORT ADDRESS, CALLEXTFUNC, WORD, TSIZE, BYTESPERWORD;
IMPORT UnixLibCall;

CONST MinOSIncrement = 32 * 1024;
    UnusedBytes = 1024;


(* Because clever allocation strategies may cause heap arrays to have
   less storage than their type declaration would suggest, and because
   passing such an object by value to a procedure will cause implicit
   copying to the high bound, it is prudent to leave a small buffer of
   valid but unused locations at the end of the heap. We do this in the
   first ever call to HeapAllocate
*)

    
PROCEDURE HeapAllocate (of: CARDINAL; VAR base: ADDRESS; 
    VAR size: CARDINAL): BOOLEAN;
    VAR rem: CARDINAL;
    VAR ofb: CARDINAL;
BEGIN
    ofb := of * AddressesPerUnit;
    IF ofb <= MinOSIncrement THEN ofb := MinOSIncrement END;
    base := ADDRESS (CALLEXTFUNC (UnixLibCall.malloc, ofb + UnusedBytes));
    IF base > 0 THEN
        size := ofb DIV AddressesPerUnit;
        RETURN TRUE
    ELSE RETURN FALSE
    END;
END HeapAllocate;

BEGIN
  AddressesPerUnit := BYTESPERWORD DIV TSIZE(WORD)
END OSStorage.
