(*
    Title: 	general terminal handler for Unix
    LastEdit:	"Wed Jan 27 17:18:50 1988"
    Author: 	acm/mjj
		Acorn Research Centre

*)

(* $T-,$R- *)

IMPLEMENTATION MODULE Terminal;

IMPORT 
  UnixCall, UnixErrNo, UnixMCI, UnixSelect;

FROM SYSTEM IMPORT WORD, ADDRESS, CALLEXTFUNC, CALLEXTPROC, ADR;
FROM CharCodes IMPORT NewLineCh, CReturnCh;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM StreamRep IMPORT 
	DefaultProcs, SetStreamError, PutProc, Procs, ImplementsProc;
FROM Streams IMPORT
        Create, Stream, StreamProcs, ErrorCode, OptionalFunction;
FROM Exceptions IMPORT RAISE;

CONST 
    bufReadLength = 80;
    bufLength = bufReadLength + 1;  (* to accomodate backspace *)

TYPE OutBuffer = POINTER TO 
	RECORD
            outFd: CARDINAL;
	    outindex: CARDINAL;
	    outbuf: ARRAY [0 .. bufLength-1] OF CHAR;
        END;
        
    StreamBuf = POINTER TO
        RECORD
            eof: BOOLEAN;
            mode: ScMode;
            translate: BOOLEAN;
            inindex, size: CARDINAL;
	    inbuf: ARRAY [0 .. bufLength-1] OF CHAR;
            buffer: OutBuffer;
            backspaceOk: BOOLEAN;
            delproc: DelProc;
	END;

CONST
    STDIN = 0; STDOUT = 1; STDERR = 2;

VAR
    ErrorStream:        Stream;         (* avoid multiple error streams  *)
    CurrentStream:      Stream;         (* most recent used stream *)
    StdBuffer:          OutBuffer;      (* common buffer for normal streams *)
    ErrBuffer:          OutBuffer;      (* common buffer for error streams *)
    StdMode:            ScMode;         (* terminal mode on entry *)

PROCEDURE ReadBuffer(b: StreamBuf): INTEGER;
  (*  Read one line from standard input *)
  CONST
              FIONREAD = 4004667fH;
  (* #define FIONREAD _IOR(f, 127, int)  /* get # bytes to read */ *)
  VAR
    r, n: INTEGER;
  BEGIN
    WITH b^ DO
      inbuf[bufLength-1] := inbuf[size-1];  (* save for Backspace *)
      UnixSelect.WaitForRead(STDIN);
      REPEAT
        r := INTEGER(CALLEXTFUNC(UnixCall.read, 
                                 STDIN, ADR(inbuf), bufReadLength))
      UNTIL r # UnixErrNo.EINTR;
      IF r<0 THEN
        r := -UnixMCI.errnoPtr^;
      ELSIF r>0 THEN
        inindex := 0;
        size := r;
        backspaceOk := TRUE;
      END
    END; (* WITH *)
    RETURN r
(*
      r := INTEGER(CALLEXTFUNC(UnixCall.ioctl, FIONREAD, ADR(n)));
      IF r<0 THEN
        r := -UnixMCI.errnoPtr^
      ELSE
        IF bufReadLength < n THEN n := bufReadLength END;
        IF TRUE (* n # 0 *) THEN
          UnixSelect.WaitForRead(STDIN);
          REPEAT
            r := INTEGER(CALLEXTFUNC(UnixCall.read, STDIN, ADR(inbuf), n))
          UNTIL r # UnixErrNo.EINTR;
          IF r<0 THEN
            r := -UnixMCI.errnoPtr^;
          ELSIF r>0 THEN
            inindex := 0;
            size := r;
            backspaceOk := TRUE;
          END
        ELSE
          size := 0
        END
      END
    END; (* WITH *)
*)
  END ReadBuffer;

PROCEDURE WriteBuffer(s: Stream; b: OutBuffer);
    VAR
        r: INTEGER;
    BEGIN
        WITH b^ DO
            UnixSelect.WaitForWrite(outFd);
            r := INTEGER(CALLEXTFUNC(UnixCall.write, 
                                     outFd, ADR(outbuf), outindex));
            IF r<0 THEN
                SetStreamError(s, UnixMCI.errnoPtr^);
            ELSE
                outindex := 0;
            END (* if *);
        END (* with *);
    END WriteBuffer;

PROCEDURE Implements(s: Stream; f: OptionalFunction): BOOLEAN;
    BEGIN
        RETURN FALSE;
    END Implements;

PROCEDURE Backspace(s: Stream);
    VAR
        b: StreamBuf;
        i: CARDINAL;
        lastCharInPrevBuffer: CHAR;
    BEGIN
        b := StreamBuf(s^.instanceData);
        WITH b^ DO
            IF eof THEN
                s^.status := EndOfStream;
            ELSIF inindex=0 THEN
              IF backspaceOk THEN
                lastCharInPrevBuffer := inbuf[bufLength-1];
                i := size;
                WHILE i>0 DO    (* shuffle up *)
                    inbuf[i] := inbuf[i-1];
                    DEC(i);
                END (* while *);
                inbuf[0] := lastCharInPrevBuffer;
                INC(size);
                backspaceOk := FALSE;
              END (* if *);
            ELSE
                DEC(inindex);
            END (* if *);
        END (* with *);
    END Backspace;

PROCEDURE Get(s: Stream): WORD;
    VAR
        sd: StreamBuf;
        ch: CHAR;
        r: INTEGER;
    BEGIN
        CheckOK (s);
        sd := StreamBuf (s^.instanceData);
        WITH sd^ DO
        (*  flush any buffered output *)
            IF buffer^.outindex>0 THEN
                WriteBuffer(s, buffer);
            END (* if *);
            IF inindex >= size THEN 
                r := ReadBuffer(sd);
                IF r = 0 THEN
                    s^.status := EndOfStream;
                    eof := TRUE;
                    RETURN WORD(0C);
                ELSIF r < 0 THEN 
                    SetStreamError(s, -r);
                    RETURN WORD(0C);
                END (* if *);
            END (* if *);
            ch := inbuf[inindex]; INC(inindex);
            IF sd^.translate AND (ch = CReturnCh) THEN
                ch := NewLineCh
            END;
        END (* with *);
	RETURN WORD(ch);
    END Get;

PROCEDURE Put(s: Stream; ch: WORD);
    VAR
        sd: StreamBuf;
    BEGIN
        CheckOK (s);
        sd := StreamBuf (s^.instanceData);
        WITH sd^ DO
            IF translate AND (CHAR (ch) = NewLineCh) THEN
                Put (s, CReturnCh)
            END;
            WITH buffer^ DO
                outbuf[outindex] := CHAR(ch);
                INC(outindex);
                IF (outindex > HIGH(outbuf)) OR (CHAR(ch) = NewLineCh) THEN
                    WriteBuffer(s, buffer);
                END (* if *);
            END;
        END (* with *);
    END Put;

PROCEDURE Flush(s: Stream);
    VAR
        sd: StreamBuf;
        b:  OutBuffer;
    BEGIN
        sd := StreamBuf(s^.instanceData);
        b := sd^.buffer;
        IF b^.outindex > 0 THEN
            WriteBuffer(s, b);
        END (* if *);
    END Flush;

PROCEDURE Delete(s: Stream);
    VAR
        sd: StreamBuf;
        b:  OutBuffer;
    BEGIN
        sd := StreamBuf(s^.instanceData);
        sd^.delproc (s);
        b := sd^.buffer;
        IF b^.outindex > 0 THEN
            WriteBuffer(s, b);
        END (* if *);
        DISPOSE(sd);
        IF ErrorStream = s THEN ErrorStream := NIL END;
        IF CurrentStream = s THEN 
            CurrentStream := NIL;
            SetFlags (s, StdMode);
        END;
    END Delete;

PROCEDURE PutE(s: Stream; ch: WORD);
    VAR
        b:  OutBuffer;
        sd: StreamBuf;
    BEGIN
        CheckOK (s);
        sd := StreamBuf (s^.instanceData);
        b := sd^.buffer;
        b^.outbuf[0] := CHAR(ch);
        b^.outindex := 1;
        WriteBuffer(s, b);
    END PutE;

PROCEDURE NewBuffer (fd: CARDINAL): OutBuffer; 
    VAR b: OutBuffer;
    BEGIN
        NEW(b);
	WITH b^ DO 
            outindex := 0;
            outFd := fd;
	END (* with *);
        RETURN b;
    END NewBuffer;

PROCEDURE NewStream 
    (b: OutBuffer; m: ScMode; t: BOOLEAN; d: DelProc): StreamBuf;
VAR sd: StreamBuf;
    i: CARDINAL;
BEGIN
    NEW (sd);
    WITH sd^ DO
        FOR i := 0 TO bufLength-1 DO
           inbuf[i] := 0C;
        END (* for *);
        mode := m;
        translate := t;
        inindex := bufReadLength; 
        size := bufReadLength;
        mode := m;
        translate := t;
        buffer := b;
        eof := FALSE;
        backspaceOk := FALSE;
        delproc := d;
    END;
    RETURN sd
END NewStream;

PROCEDURE NullDelete (s: Stream);
BEGIN
END NullDelete;

PROCEDURE CreateInOutput
    (mode: ScMode; translate: BOOLEAN; delete: DelProc): Stream;
    VAR s: Stream;
    BEGIN
        s := Create (InitProcs (Put, TRUE), 
            NewStream (StdBuffer, mode, translate, delete));
        RETURN s;
    END CreateInOutput;

PROCEDURE CreateStdInOutput (): Stream;
BEGIN
    RETURN CreateInOutput (StdMode, FALSE, NullDelete)
END CreateStdInOutput;

PROCEDURE CreateOutputForErrors(): Stream;
    BEGIN
      IF ErrorStream = NIL THEN
            ErrorStream := Create (InitProcs (PutE, FALSE), 
                    NewStream (ErrBuffer, StdMode, FALSE, NullDelete));
      END (* if *);
      RETURN ErrorStream
    END CreateOutputForErrors;

PROCEDURE StreamErrorCode(s: Stream): INTEGER;
    VAR
        sProcs: Procs;
    BEGIN
        sProcs := Procs(s^.procs);
        IF TerminalStream (s) THEN
            RETURN INTEGER(s^.streamError);
        ELSE
            RAISE("WrongClass");
            RETURN 0;
        END (* if *);
    END StreamErrorCode;

PROCEDURE InitProcs(proc: PutProc; input: BOOLEAN):StreamProcs;
    VAR
        pt: Procs;
    BEGIN
        pt := DefaultProcs(8);
        pt^.Put := proc; 
        IF input THEN 
            pt^.Get := Get; pt^.Backspace := Backspace;
        END (* if *);
        pt^.Flush := Flush; pt^.Delete := Delete;
        pt^.Implements := Implements;
        RETURN StreamProcs (pt);
    END InitProcs;

PROCEDURE TranslateCR (stream: Stream; yes: BOOLEAN);
VAR status: StreamBuf;
BEGIN
    IF NormalStream (stream) THEN
        status := StreamBuf (stream^.instanceData);
        status^.translate := yes;
    ELSE RAISE ("WrongClass");
    END;
END TranslateCR;

PROCEDURE NormalStream (stream: Stream): BOOLEAN;
VAR sb: StreamBuf;
BEGIN
    sb := StreamBuf (stream^.instanceData);
    RETURN sb^.buffer = StdBuffer;
END NormalStream;

PROCEDURE TerminalStream (stream: Stream): BOOLEAN;
VAR sb: StreamBuf;
BEGIN
    sb := StreamBuf (stream^.instanceData);
    RETURN (sb^.buffer = StdBuffer) OR (sb^.buffer = ErrBuffer)
END TerminalStream;

(* This code depends on the fact that an ScMode set is aligned on a word in
   the same way as a BITSET, covering all WORDSIZE bits of the C structure
*)


TYPE
    SGTTYB = RECORD
        chars:          ARRAY [0..3] OF CHAR;
        (* respectively sgispeed, sgospeed, sgerase, sgkill *)
        sgflags:        ScMode;
    END;
    
    CONST Mask = ScMode {TANDEM, CBREAK, LCASE, ECHO, CRMOD, RAW};


(**************************************************************************)

(*
   To make this happen, an "stty" call to Unix is made to cause a
   change in the terminal driver behaviour. This will cause other
   terminal input streams to behave oddly, and you may not open
   another stream with different characteristics.			  
*)

PROCEDURE CurrentMode (): ScMode;
    VAR sgttyb: SGTTYB;
    VAR rc: INTEGER;
BEGIN
    CALLEXTPROC(UnixCall.gtty, 0, ADR(sgttyb)); (* 0 = terminal input *)
    RETURN sgttyb.sgflags * Mask;
END CurrentMode;

PROCEDURE SetFlags (s: Stream; value: ScMode);
    VAR sgttyb: SGTTYB;
    VAR rc: INTEGER;
BEGIN
    IF INTEGER(CALLEXTFUNC(UnixCall.gtty, STDIN, ADR(sgttyb))) < 0 THEN
      RETURN
    END;
    WITH sgttyb DO
      sgflags := (sgflags - Mask) + value;
    END;
    IF INTEGER(CALLEXTFUNC(UnixCall.stty, STDIN, ADR(sgttyb))) # 0 THEN
      SetStreamError(s, -UnixMCI.errnoPtr^);
    END;
END SetFlags;

PROCEDURE CheckOK (s: Stream);
VAR nb: StreamBuf;
BEGIN
    IF s # CurrentStream THEN
        (* get terminal up to date *)
        IF CurrentStream # NIL THEN
            Flush (CurrentStream);   (* flush output on old terminal type *)
        END;
        nb := StreamBuf (s^.instanceData);
        SetFlags (s, nb^.mode);  (* change mode *)
        CurrentStream := s;
    END;
END CheckOK;

BEGIN
    CurrentStream := NIL;
    ErrorStream := NIL;
    StdMode := CurrentMode ();
    StdBuffer := NewBuffer (STDOUT);
    ErrBuffer := NewBuffer (STDERR);
END Terminal .
