(*
    Title: 	TimeDate - Implementation
    LastEdit:	"Wed Sep  5 16:11:50 1984"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory

    Unix Version
*)

IMPLEMENTATION MODULE TimeDate;
IMPORT SYSTEM, UnixCalls;

MODULE SystemTime;
    FROM SYSTEM IMPORT ADR, UNIXCALL;
    FROM UnixCalls IMPORT time;

    EXPORT GetTime, TIME;
    
    TYPE TIME = RECORD
            time, date: CARDINAL;
        END (* record *);

    (*  "Constants". *)
    VAR secondsperminute, minutesperday: CARDINAL;
    VAR dmonth: ARRAY [1 .. 12] OF CARDINAL;

    PROCEDURE SysTime(VAR t: CARDINAL);
    (*  Interface to the UNIX system clock. *)
        VAR rc: BOOLEAN;
        BEGIN
            t:= UNIXCALL(time, 0);
        END SysTime;

    PROCEDURE GetTime(VAR time : TIME);
    (*  Get Greenwich mean time and date. *)
        VAR s, m, d: CARDINAL;
        VAR Unixtime: CARDINAL;
        VAR yr, mth, dy: CARDINAL;
        BEGIN
            SysTime(Unixtime);
            s := Unixtime MOD secondsperminute; 
            m := Unixtime DIV secondsperminute;   (* minutes *)
            d := m DIV minutesperday;             (* days *)
            m := m MOD minutesperday;
            dy := d+1;
            yr := 70;        (* 1970. *)
        (*  Groups of 4 years - 3 ordinary + 1 leap.*)
            yr := yr + 4 * ( dy DIV (365 * 4 + 1));
            dy := dy MOD (365 * 4 + 1);

        (*  Single years. *)
            LOOP
                IF (yr MOD 4) = 0 THEN
                    IF dy > 366 THEN
                        dy := dy - 366
                    ELSE
                        EXIT
                    END
                ELSE
                    IF dy > 365 THEN
                        dy := dy - 365
                    ELSE
                        EXIT
                    END
                END;
                yr := yr + 1
            END;

        (*  Months. *)
            mth := 1;
            LOOP
                IF mth = 2 THEN
                    IF (yr MOD 4) = 0 THEN
                    (*  Leap February. *)
                        IF dy > 29 THEN
                            dy := dy - 29
                        ELSE
                            EXIT
                        END
                    ELSE
                        IF dy > 28 THEN
                            dy := dy - 28
                        ELSE
                            EXIT
                        END
                    END
                ELSE
                    IF dy > dmonth[mth] THEN
                        dy := dy - dmonth[mth]
                    ELSE
                        EXIT
                    END
                END;
                mth := mth + 1
            END;

            WITH time DO
                time := ((m DIV 60) * 64 + (m MOD 60)) * 64 + s;
                date := ((yr * 16) + mth) * 32 + dy;
            END
        END GetTime;

    BEGIN
        secondsperminute := 60;
        minutesperday := 1440;
        dmonth[1] := 31;    dmonth[2] := 28;    dmonth[3] := 31;
        dmonth[4] := 30;    dmonth[5] := 31;    dmonth[6] := 30;
        dmonth[7] := 31;    dmonth[8] := 31;    dmonth[9] := 30;
        dmonth[10] := 31;   dmonth[11] := 30;   dmonth[12] := 31
    END SystemTime;

(* -----------------portable exported procedures of TimeDate -------------- *)

VAR monthstrings: ARRAY [0 .. 12*3] OF CHAR;

PROCEDURE Time(): CARDINAL;
    VAR
        t: TIME;
    BEGIN
        GetTime(t);
        RETURN t.time
    END Time;

PROCEDURE Date(): CARDINAL;
    VAR
        t: TIME;
    BEGIN
        GetTime(t);
        RETURN t.date;
    END Date;

PROCEDURE Plant2(VAR v: ARRAY OF CHAR; i: CARDINAL; val: CARDINAL);
    VAR
        f: CARDINAL;
    BEGIN
        IF val>9 THEN
            f := val DIV 10;
            val := val MOD 10;
        ELSE
            f := 0;
        END (* if *);
        
        v[i] := CHAR(ORD('0') + f);
        v[i+1] := CHAR(ORD('0') + val);
    END Plant2;

PROCEDURE TimeString(time: CARDINAL; VAR ts: ARRAY OF CHAR);
    BEGIN
        ts[2] := ':'; ts[5] := ':'; ts[8] := 0C;
        Plant2(ts, 0, time DIV (64*64));
        Plant2(ts, 3, (time DIV 64) MOD 64);
        Plant2(ts, 6, time MOD 64);
    END TimeString;

PROCEDURE DateString(date: CARDINAL; VAR ds: ARRAY OF CHAR);
    VAR
        mx, i: CARDINAL;
    BEGIN
        mx :=(((date DIV 32) MOD 16) - 1)*3;
        ds[2] := '-'; ds[6] := '-'; ds[9] := 0C;
        Plant2(ds, 0, date MOD 32);
        FOR i := 0 TO 2 DO
            ds[3+i] := monthstrings[mx+i];
        END (* for *);
        Plant2(ds, 7, (date DIV (32*16)) MOD 128);
    END DateString;

BEGIN
    monthstrings := "JanFebMarAprMayJunJulAugSepOctNovDec";
END TimeDate.
