(*
    Title: 	TypeToChars - Implementation
    LastEdit:	"Tue May 12 14:10:28 1987"
    Author: 	Mick Jordan
		Acorn Research Centre
*)

IMPLEMENTATION MODULE TypeToChars;
FROM SYSTEM IMPORT MAXINT, BYTESPERWORD, MOSTSIGBIT, BITSPERWORD;
FROM Exceptions IMPORT RAISE;

VAR MaxTRUNC: REAL;

TYPE
    DIGITS = ARRAY [0..BITSPERWORD - 1] OF CHAR;
    REALDIGITS = ARRAY [0..79] OF CHAR;

(* $T-,$R- *)

PROCEDURE INTEGERToChars(int : INTEGER; VAR chars: ARRAY OF CHAR): CARDINAL;
VAR  i, j: CARDINAL;  neg: BOOLEAN;  ch: CHAR;  buf: DIGITS;
BEGIN
    IF int < 0 THEN neg:= TRUE;int:= -int ELSE neg:= FALSE END;
    IF neg AND (int<0) THEN (* MININT *)
        IF BYTESPERWORD=4 THEN
            buf:= "8463847412-"; i:= 11;
        ELSE
            buf:= "86723-"; i:= 6
        END (* if *);
    ELSE
        i:= 0;
        REPEAT
            buf[i]:= CHAR(CARDINAL(int) MOD 10 + ORD('0'));
            int:= int DIV 10;  INC(i)
        UNTIL int = 0;
        IF neg THEN buf[i]:= '-'; INC(i) END;
    END (* if *);
    RETURN CopyBuf(buf, i, chars);
END INTEGERToChars;

PROCEDURE BasedCARDINALToChars(card: CARDINAL; base: NumberBase;
   VAR chars: ARRAY OF CHAR): CARDINAL;
VAR  buf: DIGITS;  digit, i, j: CARDINAL;
BEGIN
    i:= 0;
    REPEAT
        digit:= card MOD base;
        IF digit < 10 THEN  buf[i]:= CHAR(digit + ORD('0'))
        ELSE  buf[i]:= CHAR(digit - 10 + ORD('A')) END;
        card:= card DIV base; INC(i)
    UNTIL (card=0);
    RETURN CopyBuf(buf, i, chars);
END BasedCARDINALToChars;

PROCEDURE BOOLEANToChars(bool: BOOLEAN; VAR chars: ARRAY OF CHAR): CARDINAL;
BEGIN
    IF bool THEN
        RETURN CopyBuf("EURT", 4, chars);
    ELSE
        RETURN CopyBuf("ESLAF", 5, chars) ;
    END;
END BOOLEANToChars;

PROCEDURE REALToChars(real: REAL;  decplaces: CARDINAL; 
  VAR chars: ARRAY OF CHAR): CARDINAL;

    PROCEDURE power10 (r: REAL; power: CARDINAL): REAL;
    (*  RETURNS r * 10Epower  *)
    VAR  i: CARDINAL;
    BEGIN
        FOR i := 1 TO power DO r:= r * 10.0 END;
        RETURN r;
    END power10;

VAR  integer, int, prcount: INTEGER;
    digit, ccount, decpos: CARDINAL;
    count, rcount: CARDINAL;
    remainder: REAL;
    neg: BOOLEAN;
    buf: REALDIGITS;
BEGIN
    neg:= real < 0.0;
    IF neg THEN real:= -real END;
    count:= 0;
    IF decplaces <> 0 THEN 
      buf[decplaces]:= '.';
      decpos := decplaces + 1;
    ELSE
      decpos := 0;
    END;
    REPEAT
        (*  Divide it down until it can safely be TRUNCed      *)
        rcount:= 0;
        WHILE real >= MaxTRUNC DO real:= real / 10.0; INC (rcount) END;
        (*  Now store the non-fractional digits      *)
        integer:= TRUNC(real);
        int:= integer;  
        ccount:= decpos;
        IF count = 0 THEN
          REPEAT
              buf [ccount+rcount]:= CHR((CARDINAL(int MOD 10)) + ORD('0'));
              int:= int DIV 10;  INC(ccount);
          UNTIL int = 0;
          count:= rcount + ccount;
        ELSE
          REPEAT 
              buf[ccount+rcount]:= CHR((CARDINAL(int MOD 10)) + ORD('0'));
              int:= int DIV 10; INC(ccount); DEC(prcount);
          UNTIL prcount = 0
        END;
        prcount := rcount;
        (*  recover fractional part and restore to true significance      *)
        remainder:= real - FLOAT(integer);
        real:= power10(remainder, rcount);
    UNTIL rcount = 0;
    IF neg THEN buf [count]:= '-'; INC (count) END;
    IF decplaces > 0 THEN
        FOR ccount:= 1 TO decplaces DO
            remainder:= remainder * 10.0;
            digit:= TRUNC(remainder);
            remainder:= remainder - FLOAT(digit);
            buf[decplaces-ccount]:= CHR(ORD('0') + digit);
        END;
    END;
    RETURN CopyBuf(buf, count, chars);
END REALToChars;

PROCEDURE CARDINALToChars(cardinal: CARDINAL; 
  VAR chars: ARRAY OF CHAR): CARDINAL;
BEGIN
    RETURN BasedCARDINALToChars(cardinal, 10, chars);
END CARDINALToChars;

(* $T+ *)   (* from here on we need the range checking *)

PROCEDURE BITSETToChars(bitset: BITSET; setSize: CARDINAL;
  VAR chars: ARRAY OF CHAR): CARDINAL;
VAR
    i, j, ccount: CARDINAL;
    first, increasing: BOOLEAN;

    PROCEDURE Put(ch: CHAR);
    BEGIN
        chars[ccount]:= ch;
        INC(ccount);
    END Put;

BEGIN
    increasing:= MOSTSIGBIT # 0;
    IF increasing THEN
        i:= 0
    ELSE
        i:= BITSPERWORD
    END (* if *);
    j:= 0; ccount:= 0;
    Put('{'); first:= TRUE;
    WHILE (j < setSize) AND (j < BITSPERWORD) DO
        IF i IN bitset THEN
            IF NOT first THEN
                Put(',');
            END (* if *);
            first:= FALSE;
            IF i>9 THEN
                Put(CHAR( (i DIV 10) + ORD('0') ));
            END (* if *);
            Put(CHAR( (i MOD 10) + ORD('0') ));
        END (* if *);
        INC(j);
        IF increasing THEN
            INC(i)
        ELSE
            DEC(i);
        END (* if *);
    END (* while *);
    Put('}');
    IF ccount <= HIGH(chars) THEN
        chars[ccount]:= 0C;
    END (* if *);
    RETURN ccount;
END BITSETToChars;

PROCEDURE CopyBuf(buf: ARRAY OF CHAR; maxIndex: CARDINAL;
VAR chars: ARRAY OF CHAR): CARDINAL;
    (*  the characters are stored in 'buf' in reverse order,
       from buf[maxIndex-1] -> buf[0], where maxIndex>0
       *)
VAR
    j: CARDINAL;
BEGIN
    j:= 0;
    WHILE maxIndex > 0 DO
        DEC(maxIndex);
        chars[j]:= buf[maxIndex];  INC(j)
    END (* while *);
    IF j <= HIGH(chars) THEN
        chars[j]:= 0C;
    END (* if *);
    RETURN j;
END CopyBuf;

BEGIN
  MaxTRUNC    := FLOAT (MAXINT);
END TypeToChars.
