(*
    Title: 	UnixIOXtra - Implementation
    LastEdit:	"Fri Jan 29 18:51:13 1988"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory

*)

IMPLEMENTATION MODULE UnixIOXtra; 
(*$T-,$R-*)

IMPORT UnixCall, UnixMCI;

FROM SYSTEM IMPORT  ADR, CALLEXTFUNC, CALLEXTPROC, WORD;
FROM SystemTypes IMPORT ShortINTEGER, ShortCARDINAL, LongINTEGER;
FROM UnixVersion IMPORT UnixVersion, GetUnixVersion;
FROM Fields IMPORT InsertShort, ExtractShort;

PROCEDURE Chmod(name : ARRAY OF CHAR; mode : BITSET) : INTEGER;
VAR r0: INTEGER;
BEGIN
    r0 := INTEGER(CALLEXTFUNC(UnixCall.chmod, ADR(name), mode));
    
    IF r0<0 THEN
	RETURN -UnixMCI.errnoPtr^;
    ELSE
	RETURN 0
    END
END Chmod;

PROCEDURE Dup(fildes : INTEGER) : INTEGER;
VAR r0: INTEGER;
BEGIN
    r0 := INTEGER(CALLEXTFUNC(UnixCall.dup, fildes));
    
    IF r0<0 THEN
	RETURN -UnixMCI.errnoPtr^
    ELSE
	RETURN r0
    END
END Dup;

PROCEDURE Dup2(fildes, fildes2 : INTEGER) : INTEGER;
CONST DUP2FLAG = 100B;
VAR r0: INTEGER;
BEGIN
    r0 := INTEGER(CALLEXTFUNC(UnixCall.dup, fildes+DUP2FLAG, fildes2));
    
    IF r0<0 THEN
	RETURN -UnixMCI.errnoPtr^
    ELSE
	RETURN r0
    END
END Dup2;

PROCEDURE Link(name1, name2 : ARRAY OF CHAR) : INTEGER;
VAR r0: INTEGER;
BEGIN
    r0 := INTEGER(CALLEXTFUNC(UnixCall.link, ADR(name1), ADR(name2)));
    
    IF r0<0 THEN
	RETURN -UnixMCI.errnoPtr^
    ELSE
	RETURN 0
    END
END Link;

TYPE
    BSD41StatBuf = RECORD    (* Unix uses packed structures *)
            devXino: CARDINAL;
            modeXnlink: CARDINAL;
            uidXgid: CARDINAL;
            Xrdev: CARDINAL;
            Xsize, Xatime, Xmtime, Xctime: CARDINAL;
        END (* record *);

    BSD42StatBuf = RECORD    (* Unix uses packed structures *)
            Xdev: CARDINAL;  (* C short, 'ino' aligned to longword *)
            Xino: CARDINAL;
            modeXnlink: CARDINAL;
            uidXgid: CARDINAL;
            Xrdev: CARDINAL; (* C short, 'size' aligned to longword *)
            Xsize: CARDINAL;
            Xatime, spare1: CARDINAL;
            Xmtime, spare2: CARDINAL;
            Xctime, spare3: CARDINAL;
            Xblksize, Xblocks: CARDINAL;
            spare4: ARRAY [0..2] OF CARDINAL;
	    hpuxfill: ARRAY [0..10] OF CARDINAL;
        END (* record *);

    SUNBSD42StatBuf = RECORD    (* Unix uses packed structures *)
            devXinoH : CARDINAL;
            inoLXmode: CARDINAL;
            nlinkXuid: CARDINAL;
            gidXrdev: CARDINAL;
            Xsize: CARDINAL;
            Xatime, spare1: CARDINAL;
            Xmtime, spare2: CARDINAL;
            Xctime, spare3: CARDINAL;
            Xblksize, Xblocks: CARDINAL;
            spare4: ARRAY [0..2] OF CARDINAL;
        END (* record *);

PROCEDURE CopySUN42StatBuf(
      fd:  INTEGER; 
  VAR name:ARRAY OF CHAR; 
  VAR buf: StatBuf
 ):INTEGER;
    VAR
      r0:  INTEGER;
      Ubuf:SUNBSD42StatBuf;
    BEGIN
      IF fd >= 0 THEN
        r0 := INTEGER(CALLEXTFUNC(UnixCall.fstat, fd, ADR(Ubuf)))
      ELSE      
        r0 := INTEGER(CALLEXTFUNC(UnixCall.stat, ADR(name), ADR(Ubuf)))
      END;
      IF r0<0 THEN
          RETURN -UnixMCI.errnoPtr^
      END (* if *);
      WITH buf DO
       WITH Ubuf DO
        dev := ExtractShort(devXinoH, 0);
        ino := 0;
        InsertShort(ino, 0, ExtractShort(devXinoH, 1));
        InsertShort(ino, 1, ExtractShort(inoLXmode, 0));
        mode := BITSET( ExtractShort(inoLXmode, 1) );
        nlink := ExtractShort(nlinkXuid, 0);
        uid := ExtractShort(nlinkXuid, 1);
        gid := ExtractShort(gidXrdev, 0);
        rdev := ExtractShort(gidXrdev, 1);
        size := Xsize;
        atime := Xatime; mtime := Xmtime; ctime := Xctime; 
        blocks := Xblocks; blksize := Xblksize;
       END (* with *)
      END (* with *);
      RETURN 0
    END CopySUN42StatBuf;

PROCEDURE Copy42StatBuf(
      fd:INTEGER;
  VAR name:ARRAY OF CHAR; 
  VAR buf: StatBuf
 ):INTEGER;
    VAR
      r0:  INTEGER;
      Ubuf:BSD42StatBuf;
    BEGIN
      IF fd >= 0 THEN
        r0 := INTEGER(CALLEXTFUNC(UnixCall.fstat, fd, ADR(Ubuf)))
      ELSE      
        r0 := INTEGER(CALLEXTFUNC(UnixCall.stat, ADR(name), ADR(Ubuf)))
      END;
      IF r0<0 THEN
          RETURN -UnixMCI.errnoPtr^
      END (* if *);
      WITH buf DO
       WITH Ubuf DO
        dev := ExtractShort(Xdev, 0);
        ino := Xino;
        mode := BITSET( ExtractShort(modeXnlink, 0) );
        nlink := ExtractShort(modeXnlink, 1);
        uid := ExtractShort(uidXgid, 0);
        gid := ExtractShort(uidXgid, 1);
        rdev := ExtractShort(Xrdev, 0);
        size := Xsize;
        atime := Xatime; mtime := Xmtime; ctime := Xctime; 
        blocks := Xblocks; blksize := Xblksize;
       END (* with *)
      END (* with *);
      RETURN 0
    END Copy42StatBuf;

PROCEDURE Copy41StatBuf(
      fd:  INTEGER;
  VAR name:ARRAY OF CHAR; 
  VAR buf: StatBuf
 ):INTEGER;
    VAR
      r0:  INTEGER;
      Ubuf:BSD41StatBuf;
    BEGIN
      IF fd >= 0 THEN
        r0 := INTEGER(CALLEXTFUNC(UnixCall.fstat, fd, ADR(Ubuf)))
      ELSE      
        r0 := INTEGER(CALLEXTFUNC(UnixCall.stat, ADR(name), ADR(Ubuf)))
      END;
      IF r0<0 THEN
          RETURN -UnixMCI.errnoPtr^
      END (* if *);
      WITH buf DO
       WITH Ubuf DO
        dev := ExtractShort(devXino, 0);
        ino := ExtractShort(devXino, 1);
        mode := BITSET( ExtractShort(modeXnlink, 0) );
        nlink := ExtractShort(modeXnlink, 1);
        uid := ExtractShort(modeXnlink, 0);
        gid := ExtractShort(uidXgid, 1);
        rdev := ExtractShort(Xrdev, 0);
        size := Xsize;
        atime := Xatime; mtime := Xmtime; ctime := Xctime; 
       END (* with *)
      END (* with *);
      RETURN 0
    END Copy41StatBuf;


PROCEDURE Stat(name : ARRAY OF CHAR; VAR buf : StatBuf) : INTEGER;
BEGIN
    CASE GetUnixVersion() OF
      SUN42,
      SUN43:
        RETURN CopySUN42StatBuf(-1, name, buf)
    | BSD42,
      ULTRIX11,
      ULTRIX12:
        RETURN Copy42StatBuf(-1, name, buf)
    | BSD41,
      Unknown:
        RETURN Copy41StatBuf(-1, name, buf)
    END
END Stat;

PROCEDURE Fstat(fildes : INTEGER; VAR buf : StatBuf) : INTEGER;
VAR
  dumDum:ARRAY [0..1] OF CHAR;
BEGIN
    CASE GetUnixVersion() OF
      SUN42,
      SUN43:
        RETURN CopySUN42StatBuf(fildes, dumDum, buf)
    | BSD42,
      ULTRIX11,
      ULTRIX12:
        RETURN Copy42StatBuf(fildes, dumDum, buf)
    | BSD41,
      Unknown:
        RETURN Copy41StatBuf(fildes, dumDum, buf)
    END
END Fstat;

PROCEDURE Unlink(name : ARRAY OF CHAR) : INTEGER;
VAR r0: INTEGER;
BEGIN
    r0 := INTEGER(CALLEXTFUNC(UnixCall.unlink, ADR(name)));
    
    IF r0<0 THEN
	RETURN -UnixMCI.errnoPtr^
    ELSE
	RETURN 0
    END
END Unlink;

PROCEDURE Umask(VAR oldmask: BITSET; newmask: BITSET);
BEGIN
    oldmask := BITSET(CALLEXTFUNC(UnixCall.umask, newmask));
END Umask;

END UnixIOXtra .
