(*
    Title: 	Unix Socket Interface
    LastEdit:	"Fri Jan 29 19:28:11 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE UnixSocket;


IMPORT
  Fields, SystemTypes, UnixCall, UnixErrNo, UnixMCI;

FROM SYSTEM IMPORT ADR, ADDRESS, CALLEXTFUNC;


TYPE
  UnixAddr = ARRAY [0..3] OF CARDINAL;


CONST
  AddrBytes = 16; (* TSIZE(UnixAddr); *)


PROCEDURE Socket(
      af:  Family;
      type:Type; 
      prot:Protocol
 ):T;
  VAR
    n:INTEGER;
  BEGIN
    n := INTEGER(CALLEXTFUNC(UnixCall.socket, af, type, prot));
    IF n < 0 THEN RETURN -UnixMCI.errnoPtr^ ELSE RETURN n END;
  END Socket;


PROCEDURE Accept(
      sock:   T;
  VAR addr:   Addr
 ):T;
  VAR
    n,
    unixLen:INTEGER;
    unixAddr:UnixAddr;
  BEGIN
    ConvertToUnix(addr, unixAddr);
    unixLen := AddrBytes;
    n := INTEGER(CALLEXTFUNC(UnixCall.accept, sock, ADR(unixAddr), ADR(unixLen)));
    IF n < 0 THEN RETURN -UnixMCI.errnoPtr^ ELSE RETURN n END;
  END Accept;


PROCEDURE Bind(
      sock:   T;
  VAR name:   Addr
 ):INTEGER;
  VAR
    n:INTEGER;
    unixAddr:UnixAddr;
  BEGIN
    ConvertToUnix(name, unixAddr);
    n := INTEGER(CALLEXTFUNC(UnixCall.bind, sock, ADR(unixAddr), AddrBytes));
    IF n < 0 THEN RETURN -UnixMCI.errnoPtr^ ELSE RETURN 0 END;
  END Bind;


PROCEDURE GetSockName(
      sock:   T; 
  VAR name:   Addr
 ):INTEGER;
  VAR
    n,
    unixLen:INTEGER;
    unixAddr:UnixAddr;
  BEGIN
    unixLen := AddrBytes;
    n := INTEGER(CALLEXTFUNC(UnixCall.getsockname, 
                             sock, ADR(unixAddr), ADR(unixLen)));
    IF n < 0 THEN
      RETURN -UnixMCI.errnoPtr^ 
    ELSE
      ConvertFromUnix(unixAddr, name);
      RETURN 0
    END
  END GetSockName;


PROCEDURE Listen(
      sock:T;
      backlog:INTEGER
 ):INTEGER;
  VAR
    n:INTEGER;
  BEGIN
    n := INTEGER(CALLEXTFUNC(UnixCall.listen, sock, backlog));
    IF n < 0 THEN RETURN -UnixMCI.errnoPtr^ ELSE RETURN 0 END;
  END Listen;


PROCEDURE RecV(
      sock: T; 
      buf:  ADDRESS;
      len:  INTEGER;
      flags:Flags
 ):INTEGER;
  VAR
    n:INTEGER;
  BEGIN
    REPEAT
      n := INTEGER(CALLEXTFUNC(UnixCall.recv, sock, buf, len, flags))
    UNTIL n # UnixErrNo.EINTR;
    IF n < 0 THEN RETURN -UnixMCI.errnoPtr^ ELSE RETURN n END;
  END RecV;


PROCEDURE RecVFrom(
      sock:   T; 
      buf:    ADDRESS;
      len:    INTEGER;
      flags:  Flags;
  VAR from:   Addr
 ):INTEGER;
  VAR
    n,
    unixLen:INTEGER;
    unixAddr:UnixAddr;
  BEGIN
    unixLen := AddrBytes;
    REPEAT
      n := INTEGER(CALLEXTFUNC(UnixCall.recvfrom, sock,
                               buf, len, flags, ADR(unixAddr), ADR(unixLen)))
    UNTIL n # UnixErrNo.EINTR;
    IF n < 0 THEN
      RETURN -UnixMCI.errnoPtr^ 
    ELSE
      ConvertFromUnix(unixAddr, from);
      RETURN n
    END
  END RecVFrom;


PROCEDURE Send(
      sock: T;
      msg:  ADDRESS;
      len:  INTEGER;
      flags:Flags
 ):INTEGER;
  VAR
    n:INTEGER;
  BEGIN
    n := INTEGER(CALLEXTFUNC(UnixCall.send, sock, msg, len, flags));
    IF n < 0 THEN RETURN -UnixMCI.errnoPtr^ ELSE RETURN n END;
  END Send;


PROCEDURE SendTo(
      sock: T;
      msg:  ADDRESS;
      len:  INTEGER;
      flags:Flags;
  VAR to:   Addr
 ):INTEGER;
  VAR
    n:INTEGER;
    unixAddr:UnixAddr;
  BEGIN
    ConvertToUnix(to, unixAddr);
    n := INTEGER(CALLEXTFUNC(UnixCall.sendto, sock, 
                             msg, len, flags, ADR(unixAddr), AddrBytes));
    IF n < 0 THEN RETURN -UnixMCI.errnoPtr^ ELSE RETURN n END;
  END SendTo;


PROCEDURE Shutdown(
      sock:T; 
      how: HowShut
 ):INTEGER;
  VAR
    n:INTEGER;
  BEGIN
    n := INTEGER(CALLEXTFUNC(UnixCall.shutdown, sock, how));
    IF n < 0 THEN RETURN -UnixMCI.errnoPtr^ ELSE RETURN 0 END;
  END Shutdown;


PROCEDURE SocketPair(
      af:  Family;
      type:Type;
      prot:Protocol;
  VAR sv:  PairT
 ):INTEGER;
  VAR
    n:INTEGER;
  BEGIN
    n := INTEGER(CALLEXTFUNC(UnixCall.socketpair, af, type, prot, ADR(sv)));
    IF n < 0 THEN RETURN -UnixMCI.errnoPtr^ ELSE RETURN 0 END;
  END SocketPair;


PROCEDURE GetSockOpt(
      sock:T; 
      level, optname:INTEGER; 
      optval:ADDRESS;
  VAR optlen:INTEGER
 ):INTEGER;
  VAR
    n:INTEGER;
  BEGIN
    n := INTEGER(CALLEXTFUNC(UnixCall.getsockopt, 
                             sock, level, optname, optval, ADR(optlen)));
    IF n < 0 THEN RETURN -UnixMCI.errnoPtr^ ELSE RETURN 0 END;
  END GetSockOpt;


PROCEDURE SetSockOpt(
      sock:T; 
      level, optname:INTEGER; 
      optval:ADDRESS;
  VAR optlen:INTEGER
 ):INTEGER;
  VAR
    n:INTEGER;
  BEGIN
    n := INTEGER(CALLEXTFUNC(UnixCall.setsockopt, 
                     sock, level, optname, optval, ADR(optlen)));
    IF n < 0 THEN RETURN -UnixMCI.errnoPtr^ ELSE RETURN 0 END;
  END SetSockOpt;


PROCEDURE ConvertToUnix(VAR addr:Addr; VAR unixAddr:UnixAddr);
  VAR
    i:INTEGER;
  BEGIN
    FOR i := 0 TO 3 DO unixAddr[i] := 0 END;
    Fields.InsertShort(unixAddr[0], 0, ORD(addr.family));
    CASE addr.family OF
      afUnix: 
        Fields.InsertShort(unixAddr[0], 1, 
                           Fields.ExtractShort(addr.name, 0));
        Fields.InsertShort(unixAddr[1], 0, 
                           Fields.ExtractShort(addr.name, 1));
    | afInet: 
        Fields.InsertShort(unixAddr[0], 1, addr.port);
        unixAddr[1] := addr.ipAddr
    | afUnspecified,
      afIMPLink,
      afPUP,
      afChaos,
      afNS,
      afNBS,
      afECMA,
      afDatakit,
      afCCITT,
      afSNA:
    END
  END ConvertToUnix;


PROCEDURE ConvertFromUnix(VAR unixAddr:UnixAddr; VAR addr:Addr);
  VAR
    n:CARDINAL;
  BEGIN
    addr.family := Family(Fields.ExtractShort(unixAddr[0], 0));
    CASE addr.family OF
      afUnix: 
        Fields.InsertShort(addr.name, 0, Fields.ExtractShort(unixAddr[0], 1));
        Fields.InsertShort(addr.name, 1, Fields.ExtractShort(unixAddr[0], 0));
    | afInet: 
        addr.port := Fields.ExtractShort(unixAddr[0], 1);
        addr.ipAddr := unixAddr[1]
    | afUnspecified,
      afIMPLink,
      afPUP,
      afChaos,
      afNS,
      afNBS,
      afECMA,
      afDatakit,
      afCCITT,
      afSNA:
    END
  END ConvertFromUnix;


END UnixSocket.
