(*
    Title: 	UnixTermcap - Implementation
    LastEdit:	"Wed Feb 27 18:41:27 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)
IMPLEMENTATION MODULE UnixTermcap;

FROM UnixEnv IMPORT
    GetEnvVar, EnvRc;
FROM CharCodes IMPORT EscapeCh, CapitalCh;


PROCEDURE FindKey (key: ARRAY OF CHAR; VAR entry, value: ARRAY OF CHAR): CARDINAL;
    VAR i: CARDINAL;
    
    PROCEDURE Compare (): BOOLEAN;
        VAR j: CARDINAL;
    BEGIN
        FOR j := 0 TO HIGH (key) DO
            IF (i > HIGH (entry)) OR (entry [i] = 0C) 
            OR (entry [i] # key [j]) THEN
                RETURN FALSE
            ELSE INC (i);
            END;
        END;
        IF entry [i] = "=" THEN INC (i); RETURN TRUE ELSE RETURN FALSE END;
    END Compare;
    
    PROCEDURE Skip ();
    BEGIN
        LOOP 
            IF (i > HIGH (entry)) OR (entry [i] = 0C) THEN RETURN END;
            INC (i);
            IF (entry[i-1] = ":") THEN RETURN END;
        END;
    END Skip;
    
    PROCEDURE Copy (): CARDINAL;
        VAR j: CARDINAL;
    BEGIN
        FOR j := 0 TO HIGH (value) DO
            IF (i > HIGH (entry)) OR (entry[i] = 0C) OR (entry [i] = ":") THEN
                value [j] := 0C;
                RETURN j;
            ELSE value [j] := entry [i];
                INC (i);
            END;
        END;
        RETURN HIGH (value)+1
    END Copy;

BEGIN
    i := 0;
    LOOP
        IF (i > HIGH (entry)) OR (entry [i] = 0C) THEN RETURN 0 END;
        IF Compare () THEN RETURN Copy () ELSE Skip () END;
    END;
END FindKey;


PROCEDURE UnEscape (VAR s: ARRAY OF CHAR; l: CARDINAL): CARDINAL;

    VAR i: CARDINAL;

    PROCEDURE Shift ();
        VAR j: CARDINAL;
    BEGIN
        FOR j := i TO l-2 DO s [j] := s [j+1] END;
        DEC (l);
    END Shift;
    
    PROCEDURE Octal (i: CARDINAL): CHAR;
        VAR j, octal: CARDINAL;
    BEGIN
        octal := 0;
        FOR j := 1 TO 3 DO
            IF (s[i] >= "0") AND (s[i] <= "7") THEN
                octal := octal * 8 + ORD (s[i]) - ORD ("0");
                Shift;
            END;
        END;
        RETURN VAL (CHAR, octal);
    END Octal;

    PROCEDURE CtrlCh (c: CHAR): CHAR;
    VAR cap: CHAR;
    VAR control: INTEGER;
    CONST MaxControlChOrd = 31;
    BEGIN
        control := ORD (CapitalCh (c)) - ORD ("@");
        IF (control >= 0) AND (control <= MaxControlChOrd) THEN
            RETURN VAL (CHAR, control);
        ELSE RETURN c
        END;
    END CtrlCh;

BEGIN
    IF l > 0 THEN
        i := 0;
        LOOP
            IF i >= l THEN
                EXIT
            ELSIF s [i] = "^" THEN
                Shift;
                s [i] := CtrlCh (s[i]);
            ELSIF s[i] = "\" THEN
                Shift;
                IF s[i] = "E" THEN
                    s[i] := EscapeCh;
                ELSIF (s[i] # "\") AND (s[i] # "^") THEN
                    s[i] := Octal (i);
                END;
            END;
            INC (i);
        END;
        RETURN l;
    ELSE RETURN 0
    END;
END UnEscape;


PROCEDURE TermCap (key: ARRAY OF CHAR; VAR value: ARRAY OF CHAR): CARDINAL;
    VAR termcapentry: ARRAY [0..1023] OF CHAR;
    VAR entry:        ARRAY [0..127] OF CHAR;
    VAR i, l: CARDINAL;
BEGIN
    IF GetEnvVar ("TERMCAP", termcapentry) = Found THEN
        l := FindKey (key, termcapentry, entry);
        l :=  UnEscape (entry, l);
        IF (l > 0) AND (l <= HIGH (value)+1) THEN
            FOR i := 0 TO l-1 DO value [i] := entry [i] END;
            IF l <= HIGH (value) THEN value [l] := 0C END;
            RETURN l;
        END;
    END;
    RETURN 0
END TermCap;



PROCEDURE CursorUp (VAR value: ARRAY OF CHAR): CARDINAL;
BEGIN
    RETURN TermCap ("ku", value);
END CursorUp;


PROCEDURE CursorDown  (VAR value: ARRAY OF CHAR): CARDINAL;
BEGIN
    RETURN TermCap ("kd", value);
END CursorDown;


PROCEDURE CursorLeft  (VAR value: ARRAY OF CHAR): CARDINAL;
BEGIN
    RETURN TermCap ("kl", value);
END CursorLeft;


PROCEDURE CursorRight (VAR value: ARRAY OF CHAR): CARDINAL;
BEGIN
    RETURN TermCap ("kr", value);
END CursorRight;

END UnixTermcap.
