IMPLEMENTATION MODULE Arguments;

(*
    Title: 	A General Command Line Parser
    LastEdit:	"Thu Apr 25 17:03:37 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

FROM SystemTypes IMPORT LongINTEGER, LongCARDINAL;
FROM Streams IMPORT
    Stream, SeekMode, Put, Get, Backspace, Seek, Tell;
FROM CharCodes IMPORT SpaceCh, NewLineCh, CapitalCh;
FROM TextIO IMPORT ReadCARDINAL, ReadINTEGER, ReadREAL;
FROM WriteF IMPORT WriteF0;
FROM Strings IMPORT EqualCC, CopyCC;
FROM Arithmetic IMPORT CMin, CMax;



CONST
    MaxArg = 20; (* No call of Arguments may specify more than this *)
    MaxChar= 260;(* Max no of chars which arguments may occupy      *)
    EndStreamCh = 0C;

TYPE
    ATTR  =      (Switch, Mandatory, Keyword, KeyMatched, Matched);
    ATTRSET =    SET OF ATTR;
    ITEM =       (Empty, EqualSign, Quoted, Unquoted);

    ARG = RECORD
        Attr:      ATTRSET;
        SeekPos:   LongCARDINAL; (* seek position of arg *)
        Length:    CARDINAL;
    END;

VAR Format:      ARRAY [0..MaxChar] OF CHAR; (* last arg to Arguments *)
    
VAR Arg:         ARRAY [1..MaxArg] OF ARG;
VAR ArgHigh:     CARDINAL;
VAR In, Out:     Stream;



PROCEDURE Item (VAR s: ARRAY OF CHAR; VAR pos, length: LongCARDINAL): ITEM;
    VAR quoted: BOOLEAN;
    VAR ch: CHAR;
BEGIN
    LOOP
        ch := CHAR (Get (In));
        IF (ch = NewLineCh) OR (ch = EndStreamCh) THEN
            RETURN Empty
        ELSIF ch # SpaceCh THEN
            EXIT
        END;
    END;
    Tell (In, pos);
    pos := pos - 1;
    IF ch = "=" THEN
        RETURN EqualSign
    ELSIF ch = '"' THEN
        quoted := TRUE;
        Tell (In, pos);
        ch := CHAR (Get (In));
    ELSE quoted := FALSE;
    END;
    s [0] := ch;
    length := 1;
    LOOP
        ch := CHAR (Get (In));
        IF (ch = NewLineCh) OR (ch = EndStreamCh) THEN
            EXIT
        ELSIF quoted THEN
            IF ch = '"' THEN 
                EXIT
            END;
        ELSE IF ch = SpaceCh THEN
                EXIT
            ELSIF ch = "=" THEN
                Backspace (In);
                EXIT
            END;
        END;
        s [length] := ch;
        INC (length);
        IF length >= HIGH (s) THEN EXIT END;
    END;
    s [length] := 0C;
    IF quoted THEN RETURN Quoted ELSE RETURN Unquoted END;
END Item;


PROCEDURE FindKey (VAR key: ARRAY OF CHAR): CARDINAL;
    VAR arg, i, j: CARDINAL;
    VAR keystart: BOOLEAN;
BEGIN
    keystart := TRUE;
    arg := 1;
    FOR i := 0 TO MaxChar DO
        IF keystart THEN
            j := 0;
            LOOP
                IF ((j > HIGH (key)) OR (key [j] = EndStreamCh))
                    AND ((i+j > MaxChar)
                    OR (Format [i+j] = ",") OR (Format[i+j] = "/")) THEN
                    RETURN arg
                ELSIF Format [i+j] = "," THEN
                    INC (arg);
                    keystart := TRUE;
                    EXIT
                ELSIF (Format [i+j] = "/") OR (key [j] # Format [i+j]) THEN
                    keystart := FALSE;
                    EXIT;
                ELSE INC (j);
                END;
            END;
        ELSIF Format [i] = "," THEN
            INC (arg);
            keystart := TRUE;
        END;
    END;
    RETURN 0;
 END FindKey;

    
PROCEDURE MatchKeyArg (VAR key: ARRAY OF CHAR): BOOLEAN;
    VAR arg: CARDINAL;
    VAR value: ARRAY [0..MaxChar] OF CHAR;
    VAR item: ITEM;
BEGIN
    arg := FindKey (key);
    IF (arg = 0) OR (Matched IN Arg[arg].Attr) THEN
        RETURN FALSE
    ELSE WITH Arg [arg] DO
            IF NOT (Switch IN Attr) THEN 
                item := Item (value, SeekPos, Length);
                IF item = EqualSign THEN
                    item := Item (value, SeekPos, Length);
                END;
            END;
            INCL (Attr, KeyMatched);
            INCL (Attr, Matched);
            RETURN TRUE
        END;
    END;
END MatchKeyArg;


PROCEDURE MatchPosArg (pos, length: LongCARDINAL): BOOLEAN;
    VAR arg: CARDINAL;
BEGIN
    FOR arg := 1 TO ArgHigh DO
        WITH Arg [arg] DO
            IF Attr * ATTRSET {Matched, Keyword, Switch} = ATTRSET {} THEN
                SeekPos := pos;
                Length := length;
                INCL (Attr, Matched);
                RETURN TRUE;
            END;
        END;
    END;
    RETURN FALSE;
END MatchPosArg;


PROCEDURE MatchArgs (): BOOLEAN;
    VAR s: ARRAY [0..MaxChar] OF CHAR;
    VAR pos, length: LongCARDINAL;
    VAR first: BOOLEAN;
BEGIN
    first := TRUE;
    LOOP
        CASE Item (s, pos, length) OF
            Empty:      RETURN TRUE;
        |   EqualSign:  RETURN FALSE;
        |   Quoted:     IF NOT MatchPosArg (pos, length) THEN
                            RETURN FALSE
                        END;
        |   Unquoted:   IF first AND (length = 1) AND (s[0] = "?") 
                        THEN Put (Out, '"');
                            WriteF0 (Out, Format);
                            WriteF0 (Out, '"\N');
                            RETURN FALSE;
                        ELSE IF NOT MatchKeyArg (s) 
                            AND NOT MatchPosArg (pos, length) THEN
                                RETURN FALSE
                            END;
                        END;
        ELSE RETURN FALSE;
        END;
    first := FALSE;
    END;
END MatchArgs;


PROCEDURE DefineArgs (VAR format: ARRAY OF CHAR): BOOLEAN;
    VAR c: CHAR;
    VAR pos: CARDINAL;
BEGIN
    pos := 0;
    WHILE (pos <= HIGH (format)) 
    AND (format[pos] # 0C)
    AND (ArgHigh < MaxArg) DO
        INC (ArgHigh);
        WITH Arg [ArgHigh] DO
            IF (pos > 0) AND (format [pos] = ",") THEN INC (pos) END;
            Attr := ATTRSET {};
            WHILE (pos <= HIGH (format)) 
            AND (format [pos] # "/")
            AND (format [pos] # ",") DO
                INC (pos)
            END;
            WHILE (pos < HIGH (format)) AND (format [pos] = "/") DO
                c := CapitalCh (format [pos+1]);
                INC (pos, 2);
                IF c = "S" THEN
                    INCL (Attr, Switch)
                ELSIF c = "M" THEN
                    INCL (Attr, Mandatory)
                ELSIF c = "K" THEN
                    INCL (Attr, Keyword)
                END;
            END;
        END;
        IF (pos <= HIGH (format)) AND (format [pos] # ",") THEN
            RETURN FALSE;
        END;
    END;
    RETURN TRUE;
END DefineArgs;


PROCEDURE Arguments (in, out: Stream; format: ARRAY OF CHAR): BOOLEAN;
    VAR arg: CARDINAL;
BEGIN
    In := in;
    Out := out;
    CopyCC (format, Format);
    ArgHigh := 0;
    IF NOT DefineArgs (format) OR NOT MatchArgs () THEN
        ArgHigh := 0;
        RETURN FALSE
    END;
    FOR arg := 1 TO ArgHigh DO
        WITH Arg [arg] DO
            IF Attr * ATTRSET {Mandatory, Matched} = 
                ATTRSET {Mandatory} THEN
                ArgHigh := 0;
                RETURN FALSE
            END;
        END;
    END;
    RETURN TRUE;
END Arguments;


PROCEDURE Present (arg: CARDINAL): BOOLEAN;
BEGIN
    IF (1 <= arg) AND (arg <= ArgHigh) THEN
        RETURN Matched IN Arg [arg].Attr
    ELSE RETURN FALSE
    END;
END Present;


PROCEDURE Readable (arg: CARDINAL): BOOLEAN;
BEGIN
    IF Present (arg) AND NOT (Switch IN Arg[arg].Attr) THEN
        Seek (In, Beginning, LongINTEGER (Arg [arg].SeekPos));
        RETURN TRUE;
    ELSE RETURN FALSE
    END;
END Readable;



PROCEDURE CardArg (arg: CARDINAL; VAR c: CARDINAL);
BEGIN
    IF Readable (arg) THEN c := ReadCARDINAL (In); END;
END CardArg;


PROCEDURE IntArg (arg: CARDINAL; VAR i: INTEGER);
BEGIN
    IF Readable (arg) THEN i := ReadINTEGER (In); END;
END IntArg;


PROCEDURE RealArg (arg: CARDINAL; VAR r: REAL);
BEGIN
    IF Readable (arg) THEN r := ReadREAL (In); END;
END RealArg;


PROCEDURE StringArg (arg: CARDINAL; VAR s: ARRAY OF CHAR): CARDINAL;
    VAR length, i: CARDINAL;
BEGIN
    IF Readable (arg) THEN
        WITH Arg [arg] DO
            length := CMin (HIGH (s) + 1, Length);
            IF length > 0 THEN
                Seek (In, Beginning, LongINTEGER (SeekPos));
                FOR i := 0 TO length - 1 DO
                    s [i] := CHAR (Get (In));
                END;
                FOR i := length TO HIGH (s) DO 
                    s [i] := 0C;
                END;
            END;
            RETURN length
        END;
    ELSE RETURN 0;
    END;
END StringArg;

END Arguments.
