IMPLEMENTATION MODULE Binomial;

(*
    Title: 	Computations for a binomial distriution.
    LastEdit:	"Tue Aug 20 13:41:19 1985"
    Author: 	Alan Jones,T72,232,322723
		Cambridge University Computer Laboratory
*)

FROM Arithmetic IMPORT IMax, IMin, Round ;

FROM WriteF IMPORT R ;

FROM Numerical IMPORT Power ;

FROM UtilError IMPORT ReportPError ;

PROCEDURE nCr (n, r : INTEGER) : REAL ;
VAR LargeFact, SmallFact, RIndex, Result : REAL ;
    Index : INTEGER ;
    BEGIN
        LargeFact := FLOAT (n) ;
        SmallFact := FLOAT (IMin (r, n-r)) ;
        IF SmallFact < 0.5 THEN
            RETURN 1.0
        ELSIF SmallFact < 1.5 THEN
            RETURN FLOAT (n) ;
        END (* if *);
        Result := FLOAT (n + 1) - SmallFact ;
        FOR Index := 0 TO Round (SmallFact) - 2 DO
            RIndex := FLOAT (Index) ;
            Result := Result * (LargeFact - RIndex) /
                        (SmallFact - RIndex)
        END (* for *);
        RETURN Result ;
    END nCr;

PROCEDURE Ordinate (Psuccess : REAL ;
                    n, r : INTEGER) : REAL ;
    BEGIN
        IF (r > n) OR (r < 0) THEN
            ReportPError ("Binomial.Ordinate", TRUE,
                          "n (%I0) and r (%I0) must satisfy n >= r >= 0\N",
                          n, r) ;
        END (* if *);
        IF (Psuccess < 0.0) OR (Psuccess > 1.0) THEN
            ReportPError ("Binomial.Ordinate", TRUE,
                          "Psuccess (%F4) must lie between 0 and 1 inclusive\N",
                          R (Psuccess), 0) ;
        END (* if *);
        RETURN nCr (n, r) * Power (Psuccess, r) * Power (1.0 - Psuccess, n-r)
    END Ordinate;

PROCEDURE Cumulative (Psuccess : REAL ;
                      n, rmax : INTEGER) : REAL;
VAR sum : REAL ;
    rindex : INTEGER ;
    BEGIN
        sum := 0.0 ;
        FOR rindex := 0 TO rmax DO
            sum := sum + Ordinate (Psuccess, n, rindex) ;
        END (* for *);
        RETURN sum
    END Cumulative;

END Binomial.
