IMPLEMENTATION MODULE FileUtil;

(*
    Title: 	Useful file functions
    LastEdit:	"Tue Jul 16 13:29:49 1985"
    Author: 	ahj
		Cambridge University Computer Laboratory
*)

FROM CharCodes IMPORT SpaceCh, NewLineCh,
                      CReturnCh, TabCh, FormFeedCh ;

FROM FileStream IMPORT CreateInput, CreateOutput, StreamErrorCode ;

FROM WriteF IMPORT H ;

FROM Streams IMPORT Get, Backspace, 
                     ErrorCode,
                     Stream ;

FROM UnixIOXtra IMPORT Link, Unlink ;
                     
FROM UtilError IMPORT ReportPError ;                     

PROCEDURE OpenStreamFromFile (FileName : ARRAY OF CHAR ;
                              VAR OpenStream : Stream) : BOOLEAN ;
    BEGIN
        OpenStream := CreateInput (FileName) ;
        IF OpenStream^.status # Success THEN
            ReportPError ("FileUtil", FALSE,
                          "Could not open file %S for input, error code %I0",
                          H (FileName), OpenStream^.status) ;
            RETURN FALSE
        END (* if *);
        RETURN TRUE
    END OpenStreamFromFile;

PROCEDURE OpenStreamToFile (FileName : ARRAY OF CHAR ;
                            VAR OpenStream : Stream) : BOOLEAN ;
    BEGIN
        OpenStream := CreateOutput (FileName) ;
        IF OpenStream^.status # Success THEN
            ReportPError ("FileUtil", FALSE,
                          "Could not open file %S for output, error code %I0",
                          H (FileName), OpenStream^.status) ;
            RETURN FALSE
        END (* if *);
        RETURN TRUE
    END OpenStreamToFile;

PROCEDURE ForceLink (FileName, LinkName : ARRAY OF CHAR) : INTEGER ;
VAR rc : INTEGER ;
    BEGIN
        IF Unlink (LinkName) # 0 THEN (* who cares? *) END (* if *);
        rc := Link (FileName, LinkName) ;
        IF rc # 0 THEN
            ReportPError ("FileUtil", FALSE,
                          "Link to %S failed",
                          H (FileName),
                          0) ;
        END (* if *);
        RETURN rc
    END ForceLink;

PROCEDURE ForceUnlink (LinkName : ARRAY OF CHAR) : INTEGER ;
VAR rc : INTEGER ;
    BEGIN
        RETURN Unlink (LinkName) ;
    END ForceUnlink;

PROCEDURE ForceRename (FileName, LinkName : ARRAY OF CHAR) : INTEGER ;
VAR rc : INTEGER ;
    BEGIN
        IF Unlink (LinkName) # 0 THEN (* who cares? *) END (* if *);
        rc := Link (FileName, LinkName) ;
        IF rc # 0 THEN
            ReportPError ("FileUtil", FALSE,
                          "Link to %S failed",
                          H (FileName),
                          0) ;
            RETURN rc
        END (* if *);
        rc := Unlink (FileName) ;
        IF rc # 0 THEN
            ReportPError ("FileUtil", FALSE,
                          "Unlink of %S failed",
                          H (FileName),
                          0) ;
            RETURN rc
        END (* if *);
        RETURN rc
    END ForceRename;

PROCEDURE Readln (s : Stream) ;
VAR ch : CHAR ;
    BEGIN 
        ch := CHAR (Get (s)) ;
        WHILE (ch # NewLineCh)
        DO ch := CHAR (Get (s)) END 
    END Readln ;

    
PROCEDURE SkipLayout (s : Stream) : CHAR ;
VAR c : CHAR ;
    BEGIN
        c := CHAR (Get (s)) ;
        WHILE (c = NewLineCh)
           OR (c = SpaceCh) 
           OR (c = TabCh)
           OR (c = FormFeedCh)
           OR (c = CReturnCh)
        DO c := CHAR (Get (s)) 
        END (* while *) ;
        RETURN c
    END SkipLayout ;

    
PROCEDURE SkipLeadingSpaces (s : Stream) ;
VAR c : CHAR ;
    BEGIN
        c := CHAR (Get (s)) ;
        WHILE c = SpaceCh DO c := CHAR (Get (s)) END ;
        Backspace (s)
    END SkipLeadingSpaces ;
    

PROCEDURE SkipSpacesToEndOfLine (s : Stream) ;
VAR c : CHAR ;
    BEGIN
        c := CHAR (Get (s)) ;
        WHILE c # NewLineCh
        DO IF c # SpaceCh
           THEN ReportPError ("FileUtil", FALSE,
                              "non space character read while skipping line",
                              0, 0)
           END (* if *) ;
           c := CHAR (Get (s))
        END (* while *)
    END SkipSpacesToEndOfLine ;
    
PROCEDURE ReadWord (VAR s : ARRAY OF CHAR ;
                    st : Stream) ;
VAR index : CARDINAL ;
    high  : CARDINAL ;
    ch : CHAR ;
    BEGIN
        index := 0 ;
        high := HIGH (s) ;
        ch := SkipLayout (st) ;
        WHILE (ch # SpaceCh) 
          AND (ch # NewLineCh)
          AND (ch # TabCh)
          AND (ch # FormFeedCh)
          AND (ch # CReturnCh)
          AND (index <= high)
        DO s [index] := ch ;
           INC (index) ;
           ch := CHAR (Get (st))
        END (* while *) ;
        IF index <= high THEN s [index] := 0C END (* if *)
    END ReadWord ;

END FileUtil.
