IMPLEMENTATION MODULE Filter;
(*
    Title: 	General Stream Filter
    LastEdit:	"Wed Feb 27 11:10:52 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)


FROM SYSTEM IMPORT WORD, ADR;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM StreamRep IMPORT
    Procs, GetProc, PutProc, BackspaceProc, DeleteProc, FlushProc,
    DefaultProcs, SetStreamError;
FROM Streams IMPORT
    Stream, StreamProcs, ErrorCode, StreamErrorCodeType, Get, Create;
FROM Context IMPORT Context, Cursor, NewCursor, Advance, Bound, ContextProc;

CONST EndStreamCh = 0C;

TYPE 
    FILTERDATA = RECORD
        stream:     Stream;  (* the unfiltered stream *)
        context:    Context;
        cursor:     Cursor;
        b:          ARRAY [0..HighChar] OF CHAR;
        next,                       (* next char in buffer *)
        first,                      (*  first char in buffer    *)
        free:       [0..HighChar];  (* first free postition in buffer *)
        lastch:     WORD;           (* last value read - for Backspace *)
        pending:    BOOLEAN;        (* Backspace pending *)
    END;
    
    FILTER = POINTER TO FILTERDATA;

PROCEDURE NextCh (f: FILTER): CHAR;
    VAR input: Stream;
BEGIN
    WITH f^ DO
        IF next = free THEN
            IF free = HighChar THEN free := 0 ELSE INC (free) END;
            b[next] := CHAR (Get (stream));
        END;
        IF next = HighChar THEN
            next := 0;
            RETURN b [HighChar]
        ELSE INC (next);
            RETURN b [next - 1]
        END;
    END;
END NextCh;


PROCEDURE Insert (f: FILTER; ch: CHAR);
    BEGIN
        WITH f^ DO
            IF ((first = next) AND (next = free)) (* completely empty *)
            OR (free # first) (* not full *) THEN
                IF free = HighChar THEN
                    b [HighChar] := ch;
                    free := 0;
                ELSE
                    b [free] := ch;
                    INC (free);
                END (* if *);
            END (* if *);
        END (* with *);
    END Insert;
        

PROCEDURE FilterPutProc (s: Stream; a: WORD);
VAR f: FILTER;
BEGIN
    f := FILTER (s^.instanceData);
    Insert (f, CHAR (a));
END FilterPutProc;

PROCEDURE FilterGetProc (s: Stream): WORD;
VAR f: FILTER;
    match: BOOLEAN;
    p: ContextProc;
    ch: CHAR;
BEGIN
    match := FALSE;
    f := FILTER (s^.instanceData);
    WITH f^ DO
        IF pending THEN pending := FALSE; RETURN lastch END;
        LOOP
            ch := NextCh (f);
            IF Advance (cursor, ch) THEN
                match := TRUE;
                IF Bound (cursor, p) THEN
                    NewCursor (cursor, context);
                    first := next;
                    p (f^.stream, s);  (*  pass unfiltered stream for input *)
                    IF ch = EndStreamCh THEN 
                        lastch := WORD (ch);
                        RETURN lastch
                    END;
                END;
            ELSE IF match THEN NewCursor (cursor, context) END;
                next := first;
                IF first = HighChar THEN 
                    first := 0;
                ELSE INC (first);
                END;
                lastch := WORD (NextCh (f));
                RETURN lastch;
            END;
        END;
    END;
    RETURN WORD (0);
END FilterGetProc;

PROCEDURE FilterBackspaceProc (s: Stream);
VAR f: FILTER;
BEGIN
    f := FILTER (s^.instanceData);
    f^.pending := TRUE;
END FilterBackspaceProc;

PROCEDURE FilterDeleteProc (s: Stream);
VAR f: FILTER;
BEGIN
    f := FILTER (s^.instanceData);
    Insert (f, EndStreamCh);
    DISPOSE (f);
END FilterDeleteProc;

PROCEDURE CreateFilterInput(unfiltered: Stream; c: Context): Stream;
    VAR s: Stream;
    VAR f: FILTER;
    VAR procs: Procs;
BEGIN
    NEW (f);
    WITH f^ DO
        first := 0;
        next := 0;
        free := 0;
        stream := unfiltered;
        context := c;
        lastch := WORD (EndStreamCh);
        pending := FALSE;
        NewCursor (cursor, c);
    END;
    procs := DefaultProcs (8);
    WITH procs^ DO
        Put := FilterPutProc;
        Get := FilterGetProc;
        Backspace := FilterBackspaceProc;
        Delete := FilterDeleteProc;
    END;
    s := Create (StreamProcs (procs), f);
    RETURN s;
END CreateFilterInput;




END Filter .
