IMPLEMENTATION MODULE GKS;

(*
    Title: 	Graphical Kernel System User Interface Implementation
    LastEdit:	"Tue Feb 26 14:24:33 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

FROM SYSTEM IMPORT WORD;

FROM Vector IMPORT RVEC, RBOX;

FROM GKSTypes IMPORT
    WS, FirstWS, CurrentWS, ERRORPROC, GKSError,
    RBOXPOINTER, TRANSFORM, SEGMENT,
    ATTRTYPE, ATTRIBUTE, GDPCODE, ESCAPECODE,
    PATH, PRECISION, LINESTYLE, MARKERSTYLE, FILLSTYLE;

IMPORT GKSDraw;
IMPORT GKSWork;
IMPORT GKSNorm;
IMPORT GKSAttr;


PROCEDURE OpenGKS (p: ERRORPROC);
BEGIN
    GKSError := p;
END OpenGKS;


PROCEDURE CloseGKS;
    VAR ws, next: WS;
BEGIN
    ws := FirstWS;
    WHILE ws # NIL DO
        next := ws^.Next;
        GKSWork.CloseWS (ws);
        ws := next;
    END;
END CloseGKS;


PROCEDURE OpenWS (type, instance: ARRAY OF CHAR): WS;
BEGIN
    RETURN GKSWork.OpenWS (type, instance);
END OpenWS;

PROCEDURE CloseWS (ws: WS);
BEGIN
    GKSWork.CloseWS (ws);
END CloseWS;


PROCEDURE ActivateWS (ws: WS);
BEGIN
    GKSWork.ActivateWS (ws);
END ActivateWS;

PROCEDURE DeactivateWS (ws: WS);
BEGIN
    GKSWork.DeactivateWS (ws);
END DeactivateWS;

PROCEDURE ClearWS (ws: WS; obligatory: BOOLEAN);
BEGIN
    GKSWork.ClearWS (ws, obligatory);
END ClearWS;

PROCEDURE UpdateWS (ws: WS);
BEGIN
    GKSWork.UpdateWS (ws);
END UpdateWS;

PROCEDURE SetDeferralState (ws: WS;
    AllowUpdateDeferral, AllowImplicitRedraw: BOOLEAN);
BEGIN
    GKSWork.SetDeferralState (ws, AllowUpdateDeferral, AllowImplicitRedraw);
END SetDeferralState;

PROCEDURE Message (ws: WS; s: ARRAY OF CHAR);
BEGIN
    GKSWork.Message (ws, s);
END Message;

PROCEDURE Escape (ws: WS; function: ESCAPECODE; VAR w: WORD; 
        VAR a: ARRAY OF WORD);
BEGIN
    GKSWork.Escape (ws, function, w, a);
END Escape;

PROCEDURE SetWSWindow (ws: WS; r: RBOX);
BEGIN
    GKSWork.SetWSWindow (ws, r);
END SetWSWindow;

PROCEDURE SetWSViewport (ws: WS; r: RBOX);
BEGIN
    GKSWork.SetWSViewport (ws, r);
END SetWSViewport;

PROCEDURE GetWSTransformation (ws: WS; VAR w, v: RBOX);
BEGIN
    WITH ws^ DO
        w := Window;
        v := Viewport;
    END;
END GetWSTransformation;


PROCEDURE GetWSDisplaySurface (ws: WS; VAR r: RBOX);
BEGIN
    r := ws^.DisplaySurface;
END GetWSDisplaySurface;


PROCEDURE PolyLine (count: INTEGER; points: ARRAY OF RVEC);
BEGIN
    GKSDraw.PolyLine (count, points);
END PolyLine;


PROCEDURE Line (from, to: RVEC);
BEGIN
    GKSDraw.Line (from, to);
END Line;


PROCEDURE PolyMarker (count: INTEGER; points: ARRAY OF RVEC);
BEGIN
    GKSDraw.PolyMarker (count, points);
END PolyMarker;

PROCEDURE Marker (at: RVEC);
BEGIN
    GKSDraw.Marker (at);
END Marker;

PROCEDURE Text (at: RVEC; text: ARRAY OF CHAR);
BEGIN
    GKSDraw.Text (at, text);
END Text;

PROCEDURE Fill (count: INTEGER; points: ARRAY OF RVEC);
BEGIN
    GKSDraw.Fill (count, points);
END Fill;

PROCEDURE Pixels (in: RBOX; rows, columns: INTEGER; 
    pixels: ARRAY OF CARDINAL);
BEGIN
    GKSDraw.Pixels (in, rows, columns, pixels);
END Pixels;

PROCEDURE Gdp (function: GDPCODE; data: WORD; count: INTEGER;
    points: ARRAY OF RVEC);
BEGIN
    GKSDraw.Gdp (function, data, count, points);
END Gdp;

PROCEDURE SelectNormalization (index: CARDINAL);
BEGIN
    GKSNorm.SelectNormalization (index);
END SelectNormalization;

PROCEDURE SetWindow (index: CARDINAL; r: RBOX);
BEGIN
    GKSNorm.SetWindow (index, r);
END SetWindow;

PROCEDURE SetViewport (index: CARDINAL; r: RBOX);
BEGIN
    GKSNorm.SetViewport (index, r);
END SetViewport;

PROCEDURE SetClipping (on: BOOLEAN);
BEGIN
    GKSNorm.SetClipping (on);
END SetClipping;

PROCEDURE GetNormalizationIndex (): CARDINAL;
BEGIN
    RETURN GKSNorm.GetNormalizationIndex ();
END GetNormalizationIndex;

PROCEDURE GetNormalization (index: CARDINAL; VAR w, v: RBOX);
BEGIN
    GKSNorm.GetNormalization (index, w, v);
END GetNormalization;

PROCEDURE SetPolyLineIndex (index: CARDINAL);
BEGIN
    GKSAttr.SetPolyLineIndex (index);
END SetPolyLineIndex;

PROCEDURE SetPolyMarkerIndex (index: CARDINAL);
BEGIN
    GKSAttr.SetPolyMarkerIndex (index);
END SetPolyMarkerIndex;

PROCEDURE SetTextIndex (index: CARDINAL);
BEGIN
    GKSAttr.SetTextIndex (index);
END SetTextIndex;

PROCEDURE SetFillAreaIndex (index: CARDINAL);
BEGIN
    GKSAttr.SetFillAreaIndex (index);
END SetFillAreaIndex;

PROCEDURE SetPolyLine (ws: WS; index, style, colour: CARDINAL; width: REAL);
BEGIN
    GKSAttr.SetPolyLine (ws, index, style, colour, width);
END SetPolyLine;

PROCEDURE SetPolyMarker (ws: WS; index, style, colour: CARDINAL; size: REAL);
BEGIN
    GKSAttr.SetPolyMarker (ws, index, style, colour, size);
END SetPolyMarker;

PROCEDURE SetText (ws: WS; index, font, colour: CARDINAL;
    precision: PRECISION);
BEGIN
    GKSAttr.SetText (ws, index, font, colour, precision);
END SetText;

PROCEDURE SetFillArea (ws: WS; index, style, pattern, colour: CARDINAL);
BEGIN
    GKSAttr.SetFillArea (ws, index, style, pattern, colour);
END SetFillArea;

PROCEDURE SetPattern (ws: WS; index: CARDINAL; rowhigh, columnhigh: INTEGER; 
    pattern: ARRAY OF CARDINAL);
BEGIN
    GKSAttr.SetPattern (ws, index, rowhigh, columnhigh, pattern);
END SetPattern;

PROCEDURE SetColour (ws: WS; index: CARDINAL; red, green, blue: REAL);
BEGIN
    GKSAttr.SetColour (ws, index, red, green, blue);
END SetColour;

PROCEDURE SetCharHeight (h: REAL);
BEGIN
    GKSAttr.SetCharHeight (h);
END SetCharHeight;

PROCEDURE SetCharUpVector (v: RVEC);
BEGIN
    GKSAttr.SetCharUpVector (v);
END SetCharUpVector;

PROCEDURE SetCharExpand (e: REAL);
BEGIN
    GKSAttr.SetCharExpand (e);
END SetCharExpand;

PROCEDURE SetCharPath (d: PATH);
BEGIN
    GKSAttr.SetCharPath (d);
END SetCharPath;

PROCEDURE SetCharSpacing (s: REAL);
BEGIN
    GKSAttr.SetCharSpacing (s);
END SetCharSpacing;

PROCEDURE SetPatternSize (sx, sy: REAL);
BEGIN
    GKSAttr.SetPatternSize (sx, sy);
END SetPatternSize;

PROCEDURE SetPatternReferencePoint (p: RVEC);
BEGIN
    GKSAttr.SetPatternReferencePoint (p);
END SetPatternReferencePoint;


PROCEDURE GetActiveWS (VAR a: ARRAY OF WS; VAR c: INTEGER);
    VAR ws: WS;
BEGIN
    c := 0;
    ws := FirstWS;
    WHILE ws # NIL DO
        a [c] := ws;
        ws := ws^.Next;
        INC (c);
    END;
END GetActiveWS;


END GKS.
