IMPLEMENTATION MODULE GKSAttr;
(*
    Title: 	GKS Attributes Module
    LastEdit:	"Tue Mar  3 15:53:38 1987" by Glenn Nicholls
    Author: 	jd
		Cambridge University Computer Laboratory
*)

FROM Vector IMPORT
    RVEC, AXIS, REqual, RAdd, RSubtract, RLength, RVec, RSetLength;

FROM GKSTypes IMPORT
    WS, FirstWS, SelectWS, TRANSFORM, PATH, PRECISION,
    LINESTYLE, MARKERSTYLE, FILLSTYLE,
    ATTRTYPE, ATTRIBUTE, DropAttributes, DropTransform, SetFontAspect,
    StorePixels;

FROM GKSDev IMPORT
    SetDevLine, SetDevMarker, SetDevText,
    SetDevColour, SetDevArea, SetDevPattern,
    SetDevPatternRef, SetDevChar;

FROM GKSTrans IMPORT
    TransformPoint, TransformVector, RotationTransform, IdentityTransform;

FROM GKSM IMPORT 
    RECORDTYPE, WrRecord, WrInt, WrReal, WrPath, WrVector, WrPixels;

FROM MathLib IMPORT Atan2, Degrees;

FROM Storage IMPORT ALLOCATE;


CONST DefaultColour = 0;
    DefaultRed = 0.0;
    DefaultGreen = 0.0;
    DefaultBlue = 0.0;
    DefaultText = 1;
    DefaultLineWidth = 1.0;
    NullStyle = 0;
    NullColour = 0;
    DefaultPattern = 0;

VAR DefaultIndex: ARRAY ATTRTYPE OF CARDINAL;

VAR normalization: TRANSFORM;


PROCEDURE SetPolyLineIndex (index: CARDINAL);
BEGIN
    CurrentLine := index;
END SetPolyLineIndex;


PROCEDURE SetPolyMarkerIndex (index: CARDINAL);
BEGIN
    CurrentMarker := index;
END SetPolyMarkerIndex;


PROCEDURE SetTextIndex (index: CARDINAL);
BEGIN
    CurrentText := index;
END SetTextIndex;


PROCEDURE SetFillAreaIndex (index: CARDINAL);
BEGIN
    CurrentFillArea := index;
END SetFillAreaIndex;


MODULE CharAttributes;
(********************)
    IMPORT
        RVEC, AXIS, PATH, ATTRIBUTE, ATTRTYPE, RECORDTYPE,
        NullStyle, NullColour,
        CurrentChar, ALLOCATE, TRANSFORM, 
        RVec, RSetLength, REqual,
        normalization, TransformPoint, TransformVector,
        RotationTransform;

    EXPORT
        SetCharHeight, SetCharUpVector, SetCharExpand,
        SetCharPath, SetCharSpacing, SetChar, SetMetaChar;

    VAR FirstChar: ATTRIBUTE;
    VAR Rotate270: TRANSFORM;


    VAR Height, Expand, Space: REAL;
    VAR Path: PATH;

    VAR UpVector, Vertical, Base: RVEC;

    PROCEDURE SetChar;
        VAR attr: ATTRIBUTE;
        VAR pathrotate: TRANSFORM;
    BEGIN
        attr := FirstChar;
        WHILE attr # NIL DO
            WITH attr^ DO
                IF REqual (Vertical, CharUp, 0.0)
                    AND REqual (Base, CharBase, 0.0)
		    AND (Expand = CharExpand)
                    AND (Space = CharSpace)
		    AND (Path = CharPath) THEN
                    CurrentChar := attr;
                    RETURN
                END;
            END;
            attr := attr^.Next
        END;
        NEW (attr, CharAttr);
        WITH attr^ DO
            Metafiled := FALSE;
            Type := CharAttr;
            CharUp := Vertical;
            CharBase := Base;
            CharPath := Path;
            CharSpace := Space;
	    CharExpand := Expand;
            Index := 0; Style := NullStyle; Colour:= NullColour;
            Next := FirstChar;
        END;
        FirstChar := attr;
        CurrentChar := attr;
    END SetChar;


    PROCEDURE SetMetaChar (vertical, base: RVEC);
    BEGIN
        Vertical := vertical;
        Base := base;
        SetChar;
    END SetMetaChar;
    
    PROCEDURE ComputeBaseAndVertical;
    BEGIN
        Vertical := UpVector;
        RSetLength (Vertical, Height);
        Base := Vertical;
        TransformVector (Rotate270, Base);
        TransformVector (normalization, Vertical);
        TransformVector (normalization, Base);
    END ComputeBaseAndVertical;
        

    PROCEDURE SetCharHeight (h: REAL);
    BEGIN
        IF h >= 0.0 THEN
            Height := h;
            ComputeBaseAndVertical;
            SetChar;
        END;
    END SetCharHeight;


    PROCEDURE SetCharUpVector (v: RVEC);
    BEGIN
        IF (v[X] # 0.0) OR (v[Y] # 0.0) THEN
            UpVector := v;
            ComputeBaseAndVertical;
            SetChar;
        END;
    END SetCharUpVector;


    PROCEDURE SetCharExpand (e: REAL);
    BEGIN
        IF e > 0.0 THEN
            Expand := e;
            SetChar;
        END;
    END SetCharExpand;


    PROCEDURE SetCharPath (d: PATH);
    BEGIN
        Path := d;
        SetChar;
    END SetCharPath;


    PROCEDURE SetCharSpacing (s: REAL);
    BEGIN
        Space := s;
        SetChar;
    END SetCharSpacing;

    BEGIN
        FirstChar := NIL;
        RotationTransform (270.0, Rotate270);
        Height := 0.01;
        Path := Right;
        Space := 0.0;
        Expand := 1.0;
        RVec (UpVector, 0.0, 0.01);
        RVec (Vertical, 0.0, 0.01);
        RVec (Base, 0.01, 0.0);
        SetChar;                        (* need default text attr  *)

END CharAttributes;


MODULE PatternRefAttributes;
(**************************)

    IMPORT
        RVEC, AXIS, ATTRIBUTE, ATTRTYPE, ALLOCATE, RECORDTYPE,
        RVec, REqual, CurrentPatternRef,
        normalization, TransformPoint, TransformVector;

    EXPORT
        SetPatternSize, SetPatternReferencePoint, 
        SetPatternRef, SetMetaPattern;

    VAR FirstPatternRef: ATTRIBUTE;
    
    VAR Vertical, Base, At: RVEC;
        

    PROCEDURE SetPatternRef;
        VAR attr: ATTRIBUTE;
    BEGIN
        attr := FirstPatternRef;
        WHILE attr # NIL DO
            WITH attr^ DO
                IF REqual (Vertical, PatternUp, 0.0) AND
                    REqual (Base, PatternBase, 0.0) AND
                    REqual (At, PatternAt, 0.0) THEN
                    CurrentPatternRef := attr;
                    RETURN
                END;
            END;
            attr := attr^.Next
        END;
        NEW (attr, PatternRefAttr);
        WITH attr^ DO
            Metafiled := FALSE;
            Type := PatternRefAttr;
            PatternUp := Vertical;
            PatternBase := Base;
            PatternAt := At;
            Next := FirstPatternRef;
        END;
        FirstPatternRef := attr;
        CurrentPatternRef := attr;
    END SetPatternRef;



    PROCEDURE SetMetaPattern (vertical, base: RVEC);
    BEGIN
        Vertical := vertical;
        Base := base;
        SetPatternRef;
    END SetMetaPattern;
        


    PROCEDURE SetPatternSize (sx, sy: REAL);
    BEGIN
        RVec (Vertical, 0.0, sy);
        RVec (Base, sx, 0.0);
        TransformVector (normalization, Vertical);
        TransformVector (normalization, Base);
        SetPatternRef;
    END SetPatternSize;


    PROCEDURE SetPatternReferencePoint (p: RVEC);
    BEGIN
        At := p;
        TransformPoint (normalization, At);
        SetPatternRef;
    END SetPatternReferencePoint;

BEGIN
    FirstPatternRef := NIL;
    RVec (Base, 0.0, 0.0);
    RVec (Vertical, 0.0, 0.0);
    RVec (At, 0.0, 0.0);

END PatternRefAttributes;


PROCEDURE SetAttribute (ws: WS; type: ATTRTYPE;
    index, style, colour: CARDINAL): ATTRIBUTE;
    VAR a: ATTRIBUTE;
    VAR found: BOOLEAN;
BEGIN
    a := ws^.Attributes;
    found := FALSE;
    WHILE (a # NIL) AND NOT found DO
        IF (a^.Type = type) AND (a^.Index = index) THEN
            found := TRUE;
        ELSE a := a^.Next
        END;
    END;
    IF NOT found THEN
        NEW (a);
        WITH a^ DO
            Type := type;
            Index := index;
            Next := ws^.Attributes;
        END;
        ws^.Attributes := a;
    END;
    WITH a^ DO
        Style := style;
        Colour := colour;
        Metafiled := FALSE;
    END;
    IF ws^.CurrentAttributes [type] = a THEN
        ws^.CurrentAttributes [type] := NIL;
    END;
    RETURN a;
END SetAttribute;
    

PROCEDURE SetPolyLine (ws: WS; index, style, colour: CARDINAL; width: REAL);
    VAR a: ATTRIBUTE;
BEGIN
    a := SetAttribute (ws, LineAttr, index, style, colour);
    a^.Width := width;
END SetPolyLine;


PROCEDURE SetPolyMarker (ws: WS; index, style, colour: CARDINAL; size: REAL);
    VAR a: ATTRIBUTE;
BEGIN
    a := SetAttribute (ws, MarkerAttr, index, style, colour);
    a^.Size := size;
END SetPolyMarker;


PROCEDURE SetText (ws: WS; index, font, colour: CARDINAL;
    precision: PRECISION);
    VAR a: ATTRIBUTE;
BEGIN
    a := SetAttribute (ws, TextAttr, index, font, colour);
    a^.Precision := precision;
END SetText;


PROCEDURE SetFillArea (ws: WS; index, style, pattern, colour: CARDINAL);
    VAR a: ATTRIBUTE;
BEGIN
    a := SetAttribute (ws, AreaAttr, index, style, colour);
    a^.Pattern := pattern;
END SetFillArea;


PROCEDURE SetPattern (ws: WS; index: CARDINAL; rowhigh, columnhigh: INTEGER;
        pattern: ARRAY OF CARDINAL);
    VAR a: ATTRIBUTE;
BEGIN
    a := SetAttribute (ws, PatternAttr, index, NullStyle, NullColour);
    StorePixels (a^.Pixels, rowhigh, columnhigh, pattern);
END SetPattern;


PROCEDURE SetColour (ws: WS; index: CARDINAL; red, green, blue: REAL);
    VAR a: ATTRIBUTE;
BEGIN
    a  := SetAttribute (ws, ColourAttr, index, NullStyle, NullColour);
    a^.Red := red;
    a^.Green := green;
    a^.Blue := blue;
END SetColour;



PROCEDURE DefaultAttributes (ws: WS);
    VAR i: ATTRTYPE;
    VAR pattern: ARRAY [0..0] OF CARDINAL;
BEGIN 
    WITH ws^ DO
        Attributes := NIL;
        FOR i := LineAttr TO PatternRefAttr DO
            CurrentAttributes [i] := NIL;
        END;
        FontAspect := 1.0;
        SetColour (ws, DefaultColour, DefaultRed, DefaultGreen, DefaultBlue);
        SetPolyLine (ws, ORD (SolidLine), ORD (SolidLine), DefaultColour,
            DefaultLineWidth);
        SetPolyLine (ws, ORD (DashedLine), ORD (DashedLine), DefaultColour,
            DefaultLineWidth);
        SetPolyLine (ws, ORD (DottedLine), ORD (DottedLine), DefaultColour,
            DefaultLineWidth);
        SetPolyLine (ws, ORD (DashedDottedLine), ORD (DashedDottedLine),
            DefaultColour, DefaultLineWidth);
        SetPolyMarker (ws, ORD (DotMark), ORD (DotMark), DefaultColour, 1.0);
        SetPolyMarker (ws, ORD (PlusMark), ORD (PlusMark), DefaultColour, 1.0);
        SetPolyMarker (ws, ORD (StarMark), ORD (StarMark), DefaultColour, 1.0);
        SetPolyMarker (ws, ORD (ZeroMark), ORD (ZeroMark), DefaultColour, 1.0);
        SetPolyMarker (ws, ORD (XMark), ORD (XMark), DefaultColour, 1.0);
        SetText (ws, DefaultText, DefaultText, 
            DefaultColour, StringPrecision);
        SetFillArea (ws, ORD (HollowArea), ORD (HollowArea), 0,  DefaultColour);
        SetFillArea (ws, ORD (SolidArea), ORD (SolidArea), 0, DefaultColour);
        pattern [0] := DefaultColour;
        SetPattern (ws, DefaultPattern, 0, 0, pattern);
    END;
END DefaultAttributes;


PROCEDURE FindAttribute (ws: WS; type: ATTRTYPE; index: CARDINAL): ATTRIBUTE;
    VAR def, prev: ATTRIBUTE;
BEGIN
    prev := NIL;
    def := ws^.Attributes;
    WHILE def # NIL DO
        IF (def^.Type = type) AND (def^.Index = index) THEN
            IF prev # NIL THEN
                (* put found attribute at head of list for next search *)
                prev^.Next := def^.Next;
                def^.Next := ws^.Attributes;
                ws^.Attributes := def;
            END;
            RETURN def
        ELSE prev := def;
            def := def^.Next
        END;
    END;
    RETURN NIL;
END FindAttribute;



PROCEDURE MetaAttribute (a: ATTRIBUTE);
BEGIN
    WITH a^ DO
        CASE Type OF
          LineAttr:
                IF NOT Metafiled THEN
                    WrRecord (LineDefRec);
                    WrInt (Index); WrInt (Style); 
                    WrReal (Width); WrInt (Colour);
                END;
                WrRecord (LineIndexRec); WrInt (Index);
        | MarkerAttr:
                IF NOT Metafiled THEN
                    WrRecord (MarkerDefRec);
                    WrInt (Index); WrInt (Style);
                    WrReal (Size); WrInt (Colour);
                END;
                WrRecord (MarkIndexRec); WrInt (Index);
        | TextAttr:
                IF NOT Metafiled THEN
                    WrRecord (TextDefRec);
                    WrInt (Index); WrInt (Style); 
                    WrInt (ORD (Precision)); WrInt (Colour);
                END;
                WrRecord (TextIndexRec); WrInt (Index);
        | AreaAttr:
                IF NOT Metafiled THEN
                    WrRecord (FillDefRec);
                    WrInt (Index); WrInt (Style);
                    WrInt (Pattern); WrInt (Colour);
                END;
                WrRecord (FillIndexRec); WrInt (Index);
        | PatternAttr:
                IF NOT Metafiled THEN
                    WrRecord (PatternDefRec);
                    WrInt (Index);
                    WrPixels (Pixels^.RowHigh, Pixels^.ColumnHigh, 
                        Pixels^.Pixels);
                END;
        | ColourAttr:
                IF NOT Metafiled THEN
                    WrRecord (ColourDefRec);
                    WrInt (Index);
                    WrReal (Red); WrReal (Green); WrReal (Blue);
                END;
        | CharAttr:
                WrRecord (CharHeightRec);
                    WrReal (RLength (CharUp));
                WrRecord (CharUpRec); 
                    WrVector (CharUp); WrVector (CharBase);
                WrRecord (CharExpandRec); WrReal (CharExpand);
                WrRecord (CharPathRec);  WrPath (CharPath);
                WrRecord (CharSpaceRec); WrReal (CharSpace);
        | PatternRefAttr:
                WrRecord (PatternSizeRec);
                    WrVector (PatternBase); WrVector (PatternUp);
                WrRecord (PatternRefRec);
                    WrVector (PatternAt);
        END;
        Metafiled := TRUE;
    END;
END MetaAttribute;

TYPE ATTRSET = SET OF ATTRTYPE;
VAR ContainsColour: ATTRSET;

PROCEDURE SelectPen (ws: WS; type: ATTRTYPE; index: CARDINAL);
    VAR def: ATTRIBUTE;
BEGIN
    IF (ws^.CurrentAttributes [type] = NIL) OR
       (ws^.CurrentAttributes [type]^.Index # index) THEN
        def := FindAttribute (ws, type, index);
        IF def # NIL THEN
            IF ws^.Metafile THEN
                MetaAttribute (def);
            ELSE CASE type OF
                    LineAttr:
                        SetDevLine (def^.Style, def^.Width);
                  | MarkerAttr:
                        SetDevMarker (def^.Style, def^.Size);
                  | TextAttr:
                        SetFontAspect (SetDevText (def^.Style, def^.Precision));
                  | AreaAttr:
                        SetDevArea (def^.Style, def^.Pattern);
                        SelectPen (ws, PatternAttr, def^.Style);
                  | PatternAttr:
                        SetDevPattern (def^.Pixels);
                  | ColourAttr:
                        SetDevColour (def^.Red, def^.Green, def^.Blue);
                ELSE RETURN END;
            END;
            ws^.CurrentAttributes [type] := def;
        ELSE SelectPen (ws, type, DefaultIndex [type]);
        END;
    END;
    IF type IN ContainsColour THEN
        SelectPen (ws, ColourAttr, ws^.CurrentAttributes [type]^.Colour);
    END;
END SelectPen;


PROCEDURE SpaceVector (up, base: RVEC; space: REAL;
    path: PATH; VAR v: RVEC): BOOLEAN;
    VAR p: RVEC;
BEGIN
    IF (space = 0.0) AND (path = Right) THEN 
        v := base;
        RETURN TRUE
    ELSE IF (path = Left) OR (path = Right) THEN p := base ELSE p := up END;
        RSetLength (p, space * RLength (p));
        v[X] := 0.0; v[Y] := 0.0;
        CASE path OF
            Right: RAdd (p, v);
                   RAdd (base, v);
          | Left:  RSubtract (p, v);
                   RSubtract (base, v);
          | Up:    RAdd (up, v);
                   RAdd (p, v);
          | Down:  RSubtract (up, v);
                   RSubtract (p, v);
        ELSE END;
        RETURN FALSE;
    END;
END SpaceVector;


PROCEDURE SelectAttribute (ws: WS; attr: ATTRIBUTE; display: TRANSFORM);
    VAR at, up, base, space: RVEC;
    VAR height, width, rotation, slant: REAL;
    VAR abut: BOOLEAN;
BEGIN
    IF (ws^.CurrentAttributes [attr^.Type]) # attr THEN
        IF ws^.Metafile THEN
            MetaAttribute (attr)
        ELSE WITH attr^ DO
                CASE Type OF
                    CharAttr:
                        up := CharUp;
                        base := CharBase;
                        TransformVector (display, up);
                        height := RLength (up);
                        TransformVector (display, base);
                        width := RLength(base) * CharExpand * ws^.FontAspect;
                        RSetLength (base, width);
                        rotation := Degrees (Atan2 (base[Y], base[X]));
                        slant := Degrees (Atan2 (up[Y], up[X])) - rotation;
                        abut := SpaceVector (up, base, CharSpace,
                            CharPath, space);
                        SetDevChar (slant, rotation, height, width, space, abut);
                  | PatternRefAttr:
                        at := PatternAt;
                        TransformPoint (display, at);
                        up := PatternUp;
                        TransformVector (display, up);
                        base := PatternBase;
                        TransformVector (display, base);
                        SetDevPatternRef (at, up, base);
                ELSE RETURN END;
            END;
        END;
        ws^.CurrentAttributes [attr^.Type] := attr;
    END;
END SelectAttribute;


PROCEDURE ChangeNormalization (t: TRANSFORM);
    VAR ws: WS;
BEGIN
    normalization := t;
    SetChar;
    SetPatternRef;
    ws := FirstWS;
    WHILE ws # NIL DO 
        DropAttributes (ws);
        DropTransform (ws);
        ws := ws^.Next;
    END;
END ChangeNormalization;


BEGIN
    ContainsColour := ATTRSET {LineAttr, MarkerAttr, TextAttr, AreaAttr};
    DefaultIndex [LineAttr] := ORD (SolidLine);
    CurrentLine := ORD (SolidLine);
    DefaultIndex [MarkerAttr] := ORD (DotMark);
    CurrentMarker := ORD (DotMark);
    DefaultIndex [TextAttr] := DefaultText;
    CurrentText := DefaultText;
    DefaultIndex [AreaAttr] := ORD (SolidArea);
    CurrentFillArea := ORD (SolidArea);
    DefaultIndex [PatternAttr] := DefaultPattern;
    DefaultIndex [ColourAttr] := DefaultColour;
    ChangeNormalization (IdentityTransform);
END GKSAttr.
