IMPLEMENTATION MODULE GKSDefs;

(*
    Title: 	GKS standard colours package
    LastEdit:	"Mon Sep 13 09:13:41 1982"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

FROM Palette IMPORT
    COLOUR, PALETTE, INTENSITY, Intensity;
FROM GKSTypes IMPORT
    WS, LINESTYLE, MARKERSTYLE, FILLSTYLE, PRECISION;
FROM GKS IMPORT
    SetColour, SetPolyLine, SetPolyLineIndex,
    SetPolyMarker, SetPolyMarkerIndex,
    SetText, SetTextIndex,
    SetFillArea, SetFillAreaIndex;

PROCEDURE Colour (ws: WS; c: COLOUR);
    VAR i: INTENSITY;
BEGIN
    Intensity (c, i);
    WITH i DO
        SetColour (ws, ORD (c), R, G, B);
    END;
END Colour;

PROCEDURE LineNumber (c: COLOUR; style: LINESTYLE): CARDINAL;
BEGIN
    RETURN 100 + ORD (style) * (ORD (White) + 1) + ORD (c);
END LineNumber;

PROCEDURE DefineLines (ws: WS; palette: PALETTE;
    style: LINESTYLE; width: REAL);
    VAR c: COLOUR;
BEGIN
    FOR c := Black TO White DO
        IF c IN palette THEN
            Colour (ws, c);
            SetPolyLine (ws, LineNumber (c, style),
                ORD (style), ORD (c), width);
        END;
    END;
END DefineLines;

PROCEDURE SelectLine (c: COLOUR; style: LINESTYLE);
BEGIN
    SetPolyLineIndex (LineNumber (c, style));
END SelectLine;

PROCEDURE MarkerNumber (c: COLOUR; style: MARKERSTYLE): CARDINAL;
BEGIN
    RETURN 100 + ORD (style) * (ORD (White) + 1) + ORD (c);
END MarkerNumber;


PROCEDURE DefineMarkers (ws: WS; palette: PALETTE;
    style: MARKERSTYLE; size: REAL);
    VAR c: COLOUR;
BEGIN
    FOR c := Black TO White DO
        IF c IN palette THEN
            Colour (ws, c);
            SetPolyMarker (ws, MarkerNumber (c, style),
                ORD (style), ORD (c), size)
        END;
    END;
END DefineMarkers;

PROCEDURE SelectMarker (c: COLOUR; style: MARKERSTYLE);
BEGIN
    SetPolyMarkerIndex (MarkerNumber (c, style));
END SelectMarker;

PROCEDURE TextNumber (c: COLOUR; font: CARDINAL; p: PRECISION): CARDINAL;
BEGIN
    RETURN 100 + font * (ORD (StrokePrecision) + 1) * (ORD (White) + 1)
        + ORD (p) * (ORD (White) + 1) + ORD (c);
END TextNumber;


PROCEDURE DefineTexts (ws: WS; palette: PALETTE; 
    font: CARDINAL; p: PRECISION);
    VAR c: COLOUR;
BEGIN
    FOR c := Black TO White DO
        IF c IN palette THEN
            Colour (ws, c);
            SetText (ws, TextNumber (c, font, p), font, ORD (c), p);
        END;
    END;
END DefineTexts;

PROCEDURE SelectText (c: COLOUR; font: CARDINAL; p: PRECISION);
BEGIN
    SetTextIndex (TextNumber (c, font, p));
END SelectText;

PROCEDURE FillAreaNumber (c: COLOUR; style: FILLSTYLE): CARDINAL;
BEGIN
    RETURN 100 + ORD (style) * (ORD (White) + 1) + ORD (c);
END FillAreaNumber;

PROCEDURE DefineFillAreas (ws: WS; palette: PALETTE; 
    style: FILLSTYLE; pattern: CARDINAL);
    VAR c: COLOUR;
BEGIN
    FOR c := Black TO White DO
        IF c IN palette THEN
            Colour (ws, c);
            SetFillArea (ws, FillAreaNumber (c, style), 
                ORD (style), pattern, ORD (c));
        END;
    END;
END DefineFillAreas;

PROCEDURE SelectFillArea (c: COLOUR; style: FILLSTYLE);
BEGIN
    SetFillAreaIndex (FillAreaNumber (c, style));
END SelectFillArea;

END GKSDefs.
