IMPLEMENTATION MODULE GKSDev;

(*
    Title: 	BBC Microcomputer Terminal Device Driver;
    LastEdit:	"Fri Mar  8 15:46:58 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)
FROM Vector IMPORT
    RVEC, AXIS, RBOX, BOUND, RBox, RLength, RInBox;
FROM GKSBuf IMPORT Wrch, Flush;
FROM GKSTypes IMPORT
    SEGMENT, PRIMITIVE, TRANSFORM, PRECISION, PIXELS,
    LINESTYLE, LineStyle, MARKERSTYLE, MarkerStyle, FILLSTYLE, FillStyle,
    SetFontAspect, ESCAPECODE, GDPCODE;
FROM Palette IMPORT COLOUR, INTENSITY, PALETTE, Colour, Rainbow;
FROM Arithmetic IMPORT Round, RMin, RMax, RSign;
FROM SYSTEM IMPORT  MAXINT, WORD;
FROM Streams IMPORT Get, Stream;
FROM SysStreams IMPORT sysOut;
FROM WriteF IMPORT WriteF0;

CONST
    epsilon    = 0.001;
    PixelsWide = 1280.0;
    PixelsHigh = 1024.0;
    VerticalOffset = 128.0; (* used for text at bottom of screen *)
    
CONST
    SP = 11C;
    LF = 12C;
    CR = 15C;
    ESC= 33C;
    FF = 36C;


MODULE BBCIO;

    (* In which we deal with the more sordid aspects of the interface *)
    IMPORT MAXINT, Wrch, LF, SetFontAspect, RMin, RMax, epsilon,
        PixelsWide, PixelsHigh, VerticalOffset, Round,
        COLOUR, INTENSITY, PALETTE, Colour, Rainbow;
    
    EXPORT
        Mode, PlainMode, SelectMode, CurrentMode, FontHeight, FontWidth,
        AbsMove, RelMove, TextTo, AbsDraw, RelDraw, Dot,
        LineDotted, LineSolid, LineType, ClearScreen, 
        JoinCursors, SeparateCursors,
        ENABLECURSOR, EnableCursor, DisableCursor, GraphicsColour, Fill;
    
    TYPE PLANE = (Foreground, Background);
    
    (* The useful control codes *)
    CONST
        ENABLECURSOR = 2C;
        DISABLECURSOR = 3C;
        JOINCURSORS = 5C;
        SEPARATECURSORS = 6C;
        CLEARTEXT = 14C;
        CLEARGRAPHICS = 20C;
        TEXTCOLOUR = 21C;
        GRAPHICSCOLOUR = 22C;
        DEFINECOLOUR = 23C;
        MODE = 26C;
        GRAPHICSWINDOW = 30C;
        TEXTWINDOW = 34C;
        HOME = 36C;
        PLOT = 31C;
        TEXTCURSOR = 37C;
    
    VAR LineType: INTEGER;
        Palette: PALETTE; (* the ones available in current mode *)    
        CurrentColour: ARRAY PLANE OF COLOUR;
        Intensity: INTENSITY; (* of the currently selected graphics colour *)
        MaxChar: INTEGER; (* maximum char X pos in current mode *)
        FontHeight, FontWidth: REAL; (* set according to mode *)
        GraphicsX, GraphicsY: INTEGER;
    
    PROCEDURE Byte (c: CARDINAL);
        CONST msnibble = 7 * 16; lsnibble = 6 * 16;
    BEGIN
        (* avoid sending some control chars and large values in clear *)
        IF c <= 15 THEN
            Wrch (VAL (CHAR, lsnibble + c));
        ELSIF c <= 95 THEN
            Wrch (VAL (CHAR, c))
        ELSE Wrch (VAL (CHAR, msnibble + (c DIV 16)));
            Wrch (VAL (CHAR, lsnibble + (c MOD 16)));
        END;
    END Byte;
    
    PROCEDURE Word (c: INTEGER);
    BEGIN
        Byte (CARDINAL (c) MOD 256);
        Byte ((CARDINAL (c) DIV 256) MOD 256);
    END Word;
    
    PROCEDURE MapColour (from, to: COLOUR);
        (* Map logical to physical colours *)
    BEGIN
        Wrch (DEFINECOLOUR);
        Byte (ORD (from)); Byte (ORD (to)); Byte (0); Byte (0); Byte (0);
    END MapColour;
    
    
    PROCEDURE Plot (style: INTEGER; x, y: REAL; relative: BOOLEAN);
        VAR cx, cy: INTEGER;
    BEGIN
        cx := TRUNC (x);
        cy := TRUNC (y);
        Wrch (PLOT);
        Byte (style);
        IF relative THEN
            Word (cx - GraphicsX); 
            Word (cy - GraphicsY);
        ELSE Word (cx); 
            Word (cy);
        END;
        GraphicsX := cx;
        GraphicsY := cy;
    END Plot;

    
    PROCEDURE AbsMove (x, y: REAL);
    BEGIN
        Plot (4, x, y, FALSE);
    END AbsMove;

    
    PROCEDURE RelMove (x, y: REAL);
    BEGIN
        Plot (0, x, y, TRUE);
    END RelMove;

    
    PROCEDURE TextTo (x, y: REAL);
    BEGIN
        AbsMove (x, y + FontHeight); (* BBC machine positions on TOP left! *)
    END TextTo;


    PROCEDURE AbsDraw (x, y: REAL);
    BEGIN
        Plot (5 + LineType, x, y, FALSE);
    END AbsDraw;


    PROCEDURE RelDraw (x, y: REAL);
    BEGIN
        Plot (1 + LineType, x, y, TRUE);
    END RelDraw;


    PROCEDURE Dot (x, y: REAL);
    BEGIN
        Plot (5 + 64, x, y, FALSE);
    END Dot;

    
    PROCEDURE LineDotted;
    BEGIN
        LineType := 16;
    END LineDotted;

    
    PROCEDURE LineSolid;
    BEGIN
        LineType := 0;
    END LineSolid;


    PROCEDURE GCol (plane: PLANE; colour: COLOUR);
        VAR c: INTEGER;
    BEGIN
        CurrentColour [plane] := colour;
        c := ORD (colour);
        IF plane = Background THEN INC (c, 128) END;
        Wrch (GRAPHICSCOLOUR); Byte (0); Byte (c);
    END GCol;

    PROCEDURE GraphicsColour (r, g, b: REAL);
    BEGIN
        WITH Intensity DO R := r; G := g; B := b END;
        GCol (Foreground, Colour (Intensity, Palette));
    END GraphicsColour;
    
    PROCEDURE TextColour (plane: PLANE; colour: COLOUR);
        VAR c: INTEGER;
    BEGIN
        c := ORD (colour);
        IF plane = Background THEN INC (c, 128) END;
        Wrch (TEXTCOLOUR); Byte (c);
    END TextColour;
    
    PROCEDURE JoinCursors ();
    BEGIN
        Wrch (JOINCURSORS);
    END JoinCursors;

    PROCEDURE SeparateCursors ();
    BEGIN
        Wrch (SEPARATECURSORS);
    END SeparateCursors;
    
    
    PROCEDURE Fill (x, y: REAL);
    BEGIN
        Plot (81, x, y, TRUE);
    END Fill;
    
    
    PROCEDURE EnableCursor (x, y: REAL);
    BEGIN
        Wrch (ENABLECURSOR); 
        Word (INTEGER(Round (x))); Word (INTEGER(Round (y)));
    END EnableCursor;

    PROCEDURE DisableCursor;
    BEGIN
        Wrch (DISABLECURSOR);
    END DisableCursor;

    PROCEDURE ClearScreen;
    BEGIN
        Wrch (CLEARGRAPHICS);
    END ClearScreen;
    

    VAR CurrentMode: INTEGER;

    PROCEDURE Mode (mode: INTEGER); 
    BEGIN
        Wrch (MODE); Byte (mode); 
        IF mode <= 2 THEN (* interesting for graphics *)
            SetFontAspect (FontWidth / FontHeight);
            IF mode = 0 THEN
                (* for best definition, we draw white on a black background *)
                MapColour (White, Black);
                MapColour (Black, White);
            ELSE (* White backgrounds flicker, so map White to Blue *)
                MapColour (White, Blue);
                IF Green IN Palette THEN MapColour (Green, Green) END;
                IF Blue IN Palette THEN MapColour (Blue, White) END;
            END;
            GraphicsColour (Intensity.R, Intensity.G, Intensity.B);
            GCol (Background, White);
            TextColour (Foreground, Black);
            TextColour (Background, White);
            ClearScreen;
            Wrch (HOME);
            (* Now divide screen:  bottom 1/8 for text, top 7/8 for graphics *)
            Wrch (TEXTWINDOW); Byte (0); Byte (31); Byte (MaxChar); Byte (28);
            Wrch (GRAPHICSWINDOW);
                Word (0); Word (TRUNC (VerticalOffset)); 
                Word (TRUNC (PixelsWide-1.0)); Word (TRUNC (PixelsHigh-1.0));
        END;
    END Mode;

    PROCEDURE PlainMode (mode: INTEGER); 
    BEGIN
        Wrch (MODE); 
        Byte (mode); 
    END PlainMode;

    
    PROCEDURE SelectMode (mode: INTEGER);
    BEGIN
        CASE mode OF
            0: FontHeight := 32.0; FontWidth := 16.0;MaxChar := 79;
               Palette := PALETTE {Black,White};
        |   1: FontHeight := 32.0; FontWidth := 32.0;MaxChar := 39;
               Palette := PALETTE {Black,Red,Green,White};
        |   2: FontHeight := 32.0; FontWidth := 64.0;MaxChar := 19;
               Palette := Rainbow;
        ELSE Palette := PALETTE {};  (* not interesting for graphics *)
        END;
        CurrentMode := mode;
    END SelectMode;

    
    BEGIN
       LineSolid;
       CurrentColour [Foreground] := Black;
END BBCIO;

        
VAR Active: BOOLEAN;

        (********** General **********)
	


PROCEDURE OpenDev (device: ARRAY OF CHAR; 
        VAR DisplaySurface, PixelSurface: RBOX;
        VAR Defer, ImplicitRedraw: BOOLEAN);
        CONST ScreenWidth = 0.28; (* for a 14" monitor *)
BEGIN
    (* The BBC machine draws on a graphics area of PixelsWide by PixelsHigh
       but we reserve VerticalOffset pixels for text only
    *)
    RBox (DisplaySurface, 0.0, 0.0, 
        ScreenWidth, ScreenWidth * (PixelsHigh-VerticalOffset) / PixelsWide);
    RBox (PixelSurface, 0.0, VerticalOffset, PixelsWide-1.0, PixelsHigh-1.0);
    Defer := FALSE;
    ImplicitRedraw := TRUE;
    SelectMode (1);
    Active := FALSE;
END OpenDev;


PROCEDURE CloseDev; BEGIN Flush (TRUE); END CloseDev;


PROCEDURE ActivateDev; 
BEGIN
    Active := TRUE; 
    Mode (CurrentMode); 
END ActivateDev;

PROCEDURE DeactivateDev;
BEGIN
    Active := FALSE;
    DisableCursor;
    PlainMode (0);
END DeactivateDev;


PROCEDURE DevMessage (high: INTEGER; VAR s: ARRAY OF CHAR);
BEGIN
    WriteF0 (sysOut, s);
END DevMessage;


PROCEDURE ClearDev (obligatory: BOOLEAN);
BEGIN
    ClearScreen;
    Flush (TRUE);
END ClearDev;


PROCEDURE UpdateDev ();
BEGIN
    Flush (TRUE);
END UpdateDev;


        (********** Selection of Current Primitive Attributes **********)
        

PROCEDURE SetDevLine (style: CARDINAL; width: REAL);
BEGIN
    CASE LineStyle (style) OF
        SolidLine:
	    LineSolid;
      | DashedLine:
            LineDotted;
      | DottedLine:
            LineDotted;
      | DashedDottedLine:
            LineDotted;
    ELSE END;
END SetDevLine;


VAR MarkSize: REAL;
VAR MarkProc: PROCEDURE (REAL, REAL);

PROCEDURE DrawMark (x, y: REAL);
    VAR style: CARDINAL;
BEGIN
    IF (x >= MarkSize) AND (x < PixelsWide - MarkSize)
    AND (y >= MarkSize) AND (y < PixelsHigh - MarkSize) THEN
        style := LineType;
        LineSolid;
        MarkProc (x, y);
        LineType := style;
    END;
END DrawMark;


PROCEDURE DrawPlus (x, y: REAL);
BEGIN
    AbsMove (x - MarkSize, y);
    RelDraw (x + MarkSize, y);
    RelMove (x, y + MarkSize);
    RelDraw (x, y - MarkSize);
END DrawPlus;

PROCEDURE DrawX (x, y: REAL);
BEGIN
    AbsMove (x - MarkSize, y - MarkSize);
    RelDraw (x + MarkSize, y + MarkSize);
    RelMove (x + MarkSize, y - MarkSize);
    RelDraw (x - MarkSize, y + MarkSize);
END DrawX;

PROCEDURE DrawStar (x, y: REAL);
    VAR half: REAL;
BEGIN
    half := MarkSize / 2.0;
    AbsMove (x - MarkSize, y - half);
    RelMove (x + MarkSize, y - half);
    Fill (x, y + MarkSize);
    RelMove (x - MarkSize, y + half);
    RelMove (x + MarkSize, y + half);
    Fill (x, y - MarkSize);
END DrawStar;

PROCEDURE DrawZero (x, y: REAL);
BEGIN
    AbsMove (x - MarkSize, y - MarkSize);
    RelMove (x + MarkSize, y - MarkSize);
    Fill (x + MarkSize, y + MarkSize);
    RelMove (x - MarkSize, y + MarkSize);
    Fill (x - MarkSize, y - MarkSize);
END DrawZero;


PROCEDURE SetDevMarker (style: CARDINAL; size: REAL);
    VAR ord: CARDINAL;
BEGIN
    CASE MarkerStyle (style) OF
        DotMark:  MarkProc := Dot;      MarkSize := 1.0;
      | PlusMark: MarkProc := DrawPlus; MarkSize := 4.0 * size;
      | StarMark: MarkProc := DrawStar; MarkSize := 6.0 * size;
      | ZeroMark: MarkProc := DrawZero; MarkSize := 4.0 * size;
      | XMark:    MarkProc := DrawX;    MarkSize := 4.0 * size;
    ELSE END;
END SetDevMarker;

    
VAR CharSpace: RVEC;
VAR SpaceText: BOOLEAN;
VAR Precision: PRECISION;


PROCEDURE SetDevText (font: CARDINAL; precision: PRECISION): REAL;
BEGIN
    Precision := precision;
    RETURN FontWidth / FontHeight;
END SetDevText;

PROCEDURE SetDevChar (slant, rotation, height, width: REAL;
    VAR space: RVEC; abut: BOOLEAN);
BEGIN
    CharSpace := space;
    SpaceText := NOT abut OR (ABS (CharSpace[X] - FontWidth) > epsilon)
        OR (ABS (CharSpace[Y]) > epsilon);
    SpaceText := SpaceText AND (Precision # StringPrecision);
END SetDevChar;


VAR AreaFill: BOOLEAN; (* selects between solid and outline filling *)

PROCEDURE SetDevArea (style, pattern: CARDINAL);
BEGIN
    AreaFill := (FillStyle (style) = SolidArea);
END SetDevArea;


PROCEDURE SetDevPattern (pixels: PIXELS);
BEGIN
    (* Not interested in area fill patterns *)
END SetDevPattern;


PROCEDURE SetDevColour (red, green, blue: REAL);
BEGIN
    GraphicsColour (red, green, blue);
END SetDevColour;



PROCEDURE SetDevPatternRef (VAR at, up, base: RVEC);
BEGIN
    (* Can't fill areas, so can't do much about this *)
END SetDevPatternRef;


        (********** Output Functions **********)

PROCEDURE Line (high: INTEGER; VAR at: ARRAY OF RVEC; close: BOOLEAN);
    VAR i: INTEGER;
BEGIN
    AbsMove (at[0, X], at[0, Y]);
    FOR i := 1 TO high DO
        RelDraw (at[i, X], at[i, Y]);
    END;
    IF close THEN RelDraw (at[0, X], at[0, Y]) END;
    Flush (FALSE);
END Line;


PROCEDURE DevLine (high: INTEGER; VAR at: ARRAY OF RVEC);
BEGIN
    Line (high, at, FALSE);
END DevLine;


PROCEDURE DevMarker (high: INTEGER; VAR at: ARRAY OF RVEC);
    VAR i: INTEGER;
BEGIN
    FOR i := 0 TO high DO
        DrawMark (at[i, X], at[i, Y]);
    END;
    Flush (FALSE);
END DevMarker;


PROCEDURE DevFill (high: INTEGER; VAR boundary: ARRAY OF RVEC);
    VAR i: INTEGER;
BEGIN
    IF high >= 2 THEN
        IF AreaFill THEN
            FOR i := 1 TO high-1 DO
                AbsMove (boundary[0, X], boundary[0, Y]);
                RelMove (boundary[i, X], boundary[i, Y]);
                Fill (boundary[i+1, X], boundary[i+1, Y]);
            END;
        ELSE Line (high, boundary, TRUE);
        END;
    END;
    Flush (FALSE);
END DevFill;

(* NB We interpret escapes in the manner of BFormatIO *)

PROCEDURE DevText (VAR at: RVEC; VAR in: RBOX; VAR s: ARRAY OF CHAR);
    VAR i: CARDINAL;
    VAR visible: BOOLEAN;
BEGIN
    JoinCursors;
    visible := RInBox (at, in);
    IF visible THEN TextTo (at[X], at[Y]) END;
    i := 0;
    WHILE (i <= HIGH (s)) AND (s[i] # 0C) DO
        IF (s [i] = "\") AND (i < HIGH (s)) THEN
            INC (i);
            IF visible THEN
                IF s [i] = "C" THEN Wrch (CR);
                ELSIF s[i] = "P" THEN Wrch (FF);
                ELSIF s[i] = "N" THEN Wrch (CR); Wrch (LF);
                ELSIF s[i] = "T" THEN Wrch (SP);
                ELSIF s[i] = "S" THEN Wrch (SP);
                ELSIF s[i] = "\" THEN Wrch ("\");
                ELSE Wrch ("\"); Wrch (s[i]);
                END;
            END;
        ELSE IF visible THEN Wrch (s[i]) END;
        END;
        INC (i);
        IF SpaceText THEN
            at[X] := at[X] + CharSpace[X];
            at[Y] := at[Y] + CharSpace[Y];
        ELSE at[X] := at[X] + FontWidth;
        END;
        IF RInBox (at, in) THEN
            IF NOT visible OR SpaceText THEN
                TextTo (at[X], at[Y]); 
            END;
            visible := TRUE;
        ELSE visible := FALSE
        END;
    END;
    SeparateCursors;
    Flush (FALSE);
END DevText;


PROCEDURE DevPixels (VAR boundary: ARRAY OF RVEC; VAR ClipTo: RBOX;
    rowhigh, columnhigh: INTEGER; VAR pixels: ARRAY OF CARDINAL);
BEGIN
    Line (HIGH (boundary), boundary, TRUE); (* minimum simulation *)
END DevPixels;


PROCEDURE DevGdp (fn: GDPCODE; arg: WORD;
    high: INTEGER; VAR points: ARRAY OF RVEC; ClipTo: RBOX);
END DevGdp;


PROCEDURE DevEscape (fn: ESCAPECODE; VAR w: WORD; VAR a: ARRAY OF WORD);
VAR coord, byte, mult, value: INTEGER;
    in: Stream;
BEGIN
    CASE fn OF
        SetMode:
                SelectMode (INTEGER (w));
                IF Active THEN
                    Mode (INTEGER (w));
                    Flush (TRUE);
                END;
    |   CursorOn:
                EnableCursor (REAL (a[0]), REAL (a[1]));
    |   CursorOff:
                DisableCursor;
    |   CursorPrefix:			(*  ESC-E  *)
                w := WORD (2);
                a [0] := WORD (ESC);
                a [1] := WORD ('E');
    |   ReadCursor:
                in := Stream (w);
                w := WORD (2);
                FOR coord := 0 TO 1 DO
                    mult := 1;
                    value := 0;
                    FOR byte := 1 TO 2 DO
                        INC (value, (INTEGER (Get (in)) MOD 16) * 16 * mult);
                        INC (value, (INTEGER (Get (in)) MOD 16) * mult);
                        mult := mult * 256
                    END;
                    a [coord] := WORD (FLOAT (value));
                END;
    END;
    Flush (FALSE);
END DevEscape;


        (********** Segment Functions **********)


PROCEDURE ReDrawDev (): BOOLEAN;
BEGIN
END ReDrawDev;

PROCEDURE SegAdd (s: SEGMENT): BOOLEAN;
BEGIN
END SegAdd;

PROCEDURE SegDelete (s: SEGMENT): BOOLEAN;
BEGIN
END SegDelete;

PROCEDURE SegPrimChange (s: SEGMENT; p: PRIMITIVE; new: BOOLEAN): BOOLEAN;
BEGIN
END SegPrimChange;

PROCEDURE SegTransform (s: SEGMENT; T: TRANSFORM): BOOLEAN;
BEGIN
END SegTransform;

PROCEDURE SegVisible (s: SEGMENT; on: BOOLEAN): BOOLEAN;
BEGIN
END SegVisible;

PROCEDURE SegLit (s: SEGMENT; on: BOOLEAN): BOOLEAN;
BEGIN
END SegLit;

PROCEDURE SegPriority (s: SEGMENT; priority: REAL): BOOLEAN;
BEGIN
END SegPriority;

END GKSDev.
