IMPLEMENTATION MODULE GKSDraw;

(*
    Title: 	GKS Client Drawing Interface
    LastEdit:	"Tue Feb 26 14:25:01 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

FROM Vector IMPORT RVEC, AXIS, RBOX, BOUND;

FROM GKSTypes IMPORT TRANSFORM, WS, FirstWS, PRIMITIVE, GDPCODE;

FROM GKSWork IMPORT
    SelectDisplay,
    DisplayLine, DisplayMarker, DisplayText, DisplayArea,
    DisplayPixels, DisplayGdp;

FROM GKSNorm IMPORT
    ClippingBox, NormalizePoint, NormalizeLine;

FROM GKSSeg IMPORT
    StoreLine, StoreMarker, StoreText, StoreFillArea, StorePixelArray,
    StoreGdp, OpenSegment;

FROM GKSAttr IMPORT
    CurrentLine, CurrentMarker, CurrentText, CurrentFillArea,
    CurrentChar, CurrentPatternRef;

FROM GKSDev IMPORT SegPrimChange;

FROM Storage IMPORT ALLOCATE;

FROM SYSTEM IMPORT WORD;

PROCEDURE PolyLine (count: INTEGER; VAR points: ARRAY OF RVEC);
BEGIN
    IF count > 0 THEN
        NormalizeLine (count - 1, points);
        InsertLine (count - 1, points);
    END;
END PolyLine;


PROCEDURE Line (VAR from, to: RVEC);
    VAR p: ARRAY [0..1] OF RVEC;
BEGIN
    p[0] := from;
    p[1] := to;
    PolyLine (2, p);
END Line;


PROCEDURE PolyMarker (count: INTEGER; VAR points: ARRAY OF RVEC);
BEGIN
    IF count > 0 THEN
        NormalizeLine (count - 1, points);
        InsertMarker (count - 1, points);
    END;
END PolyMarker;


PROCEDURE Marker (VAR at: RVEC);
    VAR p: ARRAY [0..0] OF RVEC;
BEGIN
    p[0] := at;
    PolyMarker (1, p);
END Marker;


PROCEDURE Text (VAR at: RVEC; VAR text: ARRAY OF CHAR);
BEGIN
    NormalizePoint (at);
    InsertText (at, HIGH (text), text);
END Text;


PROCEDURE Fill (count: INTEGER; VAR points: ARRAY OF RVEC);
BEGIN
    IF count > 2 THEN
        NormalizeLine (count - 1, points);
        InsertFill (count - 1, points);
    END;
END Fill;


PROCEDURE Pixels (in: RBOX; rows, columns: INTEGER;
        VAR pixels: ARRAY OF CARDINAL);
        VAR b: BOUND;
BEGIN
    IF (rows > 0) AND (columns > 0) THEN
        FOR b := Low TO High DO NormalizePoint (in [b]) END;
        InsertPixels (in, rows - 1, columns - 1, pixels);
    END;
END Pixels;


PROCEDURE Gdp (function: GDPCODE; data: WORD; count: INTEGER;
    points: ARRAY OF RVEC);
BEGIN
    IF count > 0 THEN
        NormalizeLine (count-1, points);
        InsertGdp (function, data, count-1, points);
    END;
END Gdp;


PROCEDURE DisplayPrimitive (ws: WS; p: PRIMITIVE): BOOLEAN;
BEGIN
    IF ws^.Active THEN
        IF (p # NIL) AND SegPrimChange (OpenSegment, p, TRUE) THEN
            RETURN FALSE
        ELSE SelectDisplay (ws, OpenSegment, ClippingBox (FALSE));
            RETURN TRUE
        END;
    ELSE RETURN FALSE
    END;
END DisplayPrimitive;



PROCEDURE InsertLine (high: INTEGER; VAR points: ARRAY OF RVEC);
    VAR ws: WS;
    VAR p: PRIMITIVE;
BEGIN
    p := StoreLine (CurrentLine, high, points);
    ws := FirstWS;
    WHILE ws # NIL DO
        IF DisplayPrimitive (ws, p) THEN
            DisplayLine (CurrentLine, high, points);
        END;
        ws := ws^.Next
    END;
END InsertLine;


PROCEDURE InsertMarker (high: INTEGER; VAR points: ARRAY OF RVEC);
    VAR ws: WS;
    VAR p: PRIMITIVE;
BEGIN
    p := StoreMarker (CurrentMarker, high, points);
    ws := FirstWS;
    WHILE ws # NIL DO
        IF DisplayPrimitive (ws, p) THEN
            DisplayMarker (CurrentMarker, high, points);
        END;
        ws := ws^.Next
    END;
END InsertMarker;


PROCEDURE InsertText (at: RVEC; high: INTEGER; VAR text: ARRAY OF CHAR);
    VAR ws: WS;
    VAR p: PRIMITIVE;
BEGIN
    p := StoreText (CurrentText, at, CurrentChar, text);
    ws := FirstWS;
    WHILE ws # NIL DO
        IF DisplayPrimitive (ws, p) THEN
            DisplayText (at, CurrentText, CurrentChar, text);
        END;
        ws := ws^.Next
    END;
END InsertText;


PROCEDURE InsertFill (high: INTEGER; VAR points: ARRAY OF RVEC);
    VAR ws: WS;
    VAR p: PRIMITIVE;
BEGIN
    p := StoreFillArea (CurrentPatternRef, CurrentFillArea, high, points);
    ws := FirstWS;
    WHILE ws # NIL DO
        IF DisplayPrimitive (ws, p) THEN
            DisplayArea (CurrentFillArea, high, CurrentPatternRef, points);
        END;
        ws := ws^.Next
    END;
END InsertFill;


PROCEDURE InsertPixels (in: RBOX; rowhigh, columnhigh: INTEGER;
        VAR pixels: ARRAY OF CARDINAL);
    VAR ws: WS;
    VAR p: PRIMITIVE;
BEGIN
    p := StorePixelArray (in, rowhigh, columnhigh, pixels);
    ws := FirstWS;
    WHILE ws # NIL DO
        IF DisplayPrimitive (ws, p) THEN
            DisplayPixels (in, rowhigh, columnhigh, pixels);
        END;
        ws := ws^.Next
    END;
END InsertPixels;


PROCEDURE InsertGdp (function: GDPCODE; arg: WORD; high: INTEGER;
        VAR points: ARRAY OF RVEC);
    VAR ws: WS;
    VAR p: PRIMITIVE;
BEGIN
    p := StoreGdp (function, arg, high, points);
    ws := FirstWS;
    WHILE ws # NIL DO
        IF DisplayPrimitive (ws, p) THEN
            DisplayGdp (function, arg, high, points);
        END;
        ws := ws^.Next
    END;
END InsertGdp;

BEGIN
END GKSDraw.
