IMPLEMENTATION MODULE GKSFount;

(*
    Title: 	standard handling of fount numbers for GKS
    LastEdit:	"Tue Sep 24 15:50:23 1985"
    Author: 	pr
		Cambridge University Computer Laboratory
*)

FROM WriteF IMPORT H ;

FROM StringF IMPORT StringF1, StringF2 ;

FROM Strings IMPORT CopyCC, EqualCC, ConcatCC ;

CONST MaxFountNumber = 9 ;

TYPE FOUNTNAME = ARRAY [0..11] OF CHAR ;

VAR FountNames : ARRAY [0..MaxFountNumber] OF FOUNTNAME ;

PROCEDURE FountNumber
    (family: ARRAY OF CHAR; style: STYLE; VAR number: CARDINAL): BOOLEAN;
VAR StylePart, FountNameIndex : CARDINAL ;
    BEGIN
        CASE style OF
          Roman  : StylePart := 1 |
          Italic : StylePart := 2 |
          Bold   : StylePart := 3
        ELSE (* STYLE must have changed *)
            RETURN FALSE
        END (* case *) ;

        FountNameIndex := 0 ;
        
        WHILE (FountNameIndex <= MaxFountNumber)
          AND NOT EqualCC (FountNames [FountNameIndex], family) DO
            INC (FountNameIndex)
        END (* while *);
        
        IF FountNameIndex > MaxFountNumber THEN
            RETURN FALSE
        ELSE number := 10 * FountNameIndex + 100 + StylePart
        END (* if *);
        
        RETURN TRUE
    END FountNumber;


PROCEDURE FountName (number: CARDINAL; VAR name: ARRAY OF CHAR): BOOLEAN;
VAR FamilyPart : CARDINAL ;
    BEGIN
        FamilyPart := (number - 100) DIV 10 ;

        IF (FamilyPart < 0) OR (FamilyPart > MaxFountNumber) THEN
            CopyCC ("unknown", name) ;
            RETURN FALSE
        ELSE
            CopyCC (FountNames [FamilyPart], name)
        END (* if *);

        number := number MOD 10 ;
        
        CASE number OF
          1 : ConcatCC (name, " roman", name) | 
          2 : ConcatCC (name, " italic", name) |
          3 : ConcatCC (name, " bold", name)
        ELSE ConcatCC (name, " roman", name)
        END (* case *) ;
        
        RETURN TRUE
    END FountName;


PROCEDURE FileName
    (number, pointsize: CARDINAL; VAR filename: ARRAY OF CHAR): BOOLEAN;
VAR FamilyPart : CARDINAL ;
    style : CHAR ;
    ScratchString : ARRAY [0..39] OF CHAR ;
    BEGIN
        FamilyPart := (number - 100) DIV 10 ;

        IF (FamilyPart < 0) OR (FamilyPart > MaxFountNumber) THEN
            FamilyPart := 0
        END (* if *);
        
        number := number MOD 10 ;
        
        CASE number OF
          1 : style := "r" |
          2 : style := "i" |
          3 : style := "b"
        ELSE style := "r"
        END (* case *) ;
        
        IF FamilyPart = 0
        THEN 
            StringF1 (filename, "%C", CAP (style))
        ELSE
            StringF2 (filename,
                      "%S.%C",
                      H (FountNames [FamilyPart]),
                      style)
        END ;

        StringF2 (ScratchString, "/usr/lib/vfont/%S.%I", H (filename), pointsize) ;
        CopyCC (ScratchString, filename) ;
        RETURN TRUE
        
    END FileName;


BEGIN
    PixelsPerInch := 200.0;
    
    FountNames [0] := "hershey";        
    FountNames [1] := "basker";
    FountNames [2] := "bodoni";
    FountNames [3] := "cm";
    FountNames [4] := "delegate";
    FountNames [5] := "gacham";
    FountNames [6] := "meteor";
    FountNames [7] := "nonie";
    FountNames [8] := "stare";
    FountNames [9] := "times";

END GKSFount.
