IMPLEMENTATION MODULE GKSInput;

(*
    Title: 	handling graphical input for GKS
    LastEdit:	"Tue Feb 26 15:21:38 1985"
    Author: 	pr
		Cambridge University Computer Laboratory
*)


FROM SYSTEM IMPORT WORD;

FROM Context IMPORT Context, Bind, Unbind;

FROM Dispatch IMPORT FindRealInput;

FROM Vector IMPORT AXIS, RVEC, RVec;

FROM UnixTermcap IMPORT 
    CursorRight, CursorUp, CursorLeft, CursorDown;

FROM Streams IMPORT Stream;

FROM GKS IMPORT Escape, UpdateWS;

FROM GKSNorm IMPORT NormalizePoint, UnNormalizePoint;

FROM GKSTrans IMPORT TransformPoint, UnTransformPoint;

FROM GKSTypes IMPORT WS, ESCAPECODE, SelectWS;


VAR
(*  these variables should be stored separately for each workstation,
    and the appropriate ones accessed in the routines...		  *)
    Ws:         WS;
    Position,				(*  in device coordinates	  *)
    Grid:       RVEC;			(*  in world coordinates	  *)
    Stupid,				(*  relates to a workstation that can
                                            not volunteer cursor coords	  *)
    Visible:    BOOLEAN;


PROCEDURE BindCursorInput (ws: WS; immediate: Context; enabled: BOOLEAN): 
        BOOLEAN;
    VAR
        l,
        length:     INTEGER;
        array:      ARRAY [0 .. 10] OF WORD;
        prefix:     ARRAY [0 .. 10] OF CHAR;
    BEGIN
        SelectWS (ws);
        Escape (ws, CursorPrefix, length, array);
        IF length > 0 THEN
            FOR l := 0 TO length - 1 DO
                prefix [l] := CHAR (array [l]);
            END (* for *);
            prefix [length] := 0C;
            IF enabled THEN
            (*  should really bind prefix to ws in a private context for
                use in recognition of workstaions below	but for the time 
                being we just remember it globally			  *)
                Ws := ws;
                Stupid := prefix [0] = 0C;
            (*  no binding necessary - this is a stupid workstation that
                will obtain cursor information in GetCursor		  *)
                IF NOT Stupid THEN
                    Bind (immediate, prefix, DoReadCursor);
                    RVec (Position, 0.0, 0.0);
                END (* if *);
            ELSE
            (*  and unbind it here					  *)
                Ws := NIL;
                IF NOT Stupid THEN
                    Unbind (immediate, prefix);
                END (* if *);
            END (* if *);
            RETURN TRUE;
        ELSE
            RETURN FALSE;
        END (* if *);
    END BindCursorInput;

PROCEDURE DoReadCursor (in, out: Stream);
    BEGIN
    (*  should really find out the command that caused us to be invoked
        and then chain down all the WSDATA records to find one with the
        same CursorPrefix, and select that workstation as current	  *)
        IF Ws # NIL THEN
            Escape (Ws, ReadCursor, in, Position);
        END (* if *);
    END DoReadCursor;


PROCEDURE BindCursorKeys (ws: WS; immediate: Context; enabled: BOOLEAN): BOOLEAN;
    VAR
        up, down,
        left, right:    ARRAY [0 .. 10] OF CHAR;
        unit:           RVEC;
    BEGIN
        IF (CursorUp (up) > 0) AND (CursorDown (down) > 0) 
        AND (CursorLeft (left) > 0) AND (CursorRight (right) > 0) THEN
            IF enabled THEN
            (*  same as Cursor input again here...			  *)
                Ws := ws;
                Bind (immediate, up, GoUp);
                Bind (immediate, down, GoDown);
                Bind (immediate, left, GoLeft);
                Bind (immediate, right, GoRight);
                RVec (unit, 1.0, 1.0);
                CursorGrid (ws, unit);
            ELSE
                Ws := NIL;
                CursorOutput (ws, FALSE);
                Unbind (immediate, up);
                Unbind (immediate, down);
                Unbind (immediate, left);
                Unbind (immediate, right);
            END (* if *);
            RETURN TRUE;
        END (* if *);
        RETURN FALSE;
    END BindCursorKeys;

PROCEDURE GoUp (in, out: Stream);
    BEGIN
        Step (Y, +1.0);
    END GoUp;

PROCEDURE GoDown (in, out: Stream);
    BEGIN
        Step (Y, -1.0);
    END GoDown;

PROCEDURE GoLeft (in, out: Stream);
    BEGIN
        Step (X, -1.0);
    END GoLeft;

PROCEDURE GoRight (in, out: Stream);
    BEGIN
        Step (X, +1.0);
    END GoRight;

PROCEDURE Step (z: AXIS; sign: REAL);
    VAR
        position:       RVEC;
    BEGIN
    (*  should really find out the command that caused us to be invoked
        and then chain down all the WSDATA records to find one with the
        same cursor key string, and select that workstation as current	  *)
        IF (Ws # NIL) AND (NOT Stupid) THEN
        (*  there's no point in stepping the cursor of a stupid WS	  *)
            GetCursor (Ws, position);
            position [z] := position [z] + sign * Grid [z];
            SetCursor (Ws, position);
        END (* if *);
    END Step;

PROCEDURE CursorGrid (ws: WS; grid: RVEC);
    BEGIN
        IF Ws = ws THEN
            Grid := grid;
        END (* if *);
    END CursorGrid;


PROCEDURE CursorInput (ws: WS; enabled: BOOLEAN);
    VAR
        length:     INTEGER;
        array:      ARRAY [0 .. 10] OF WORD;
    BEGIN
        SelectWS (ws);
        Escape (ws, CursorPrefix, length, array);
        IF enabled THEN
            Ws := ws;
            Stupid := CHAR (array [0]) = 0C;
        ELSE
            Ws := NIL;
        END (* if *);
    END CursorInput;


PROCEDURE CursorOutput (ws: WS; enabled: BOOLEAN);
    BEGIN
        Visible := enabled;
        DrawCursor (ws);
    END CursorOutput;


PROCEDURE DrawCursor (ws: WS);
    VAR
        junk:       WORD;
        array:      ARRAY [0 .. 0] OF WORD;
    BEGIN
        IF Visible THEN
            Escape (ws, CursorOn, junk, Position);
        ELSE
            Escape (ws, CursorOff, junk, array);
        END (* if *);
        UpdateWS (ws);
    END DrawCursor;


PROCEDURE SetCursor (ws: WS; at: RVEC);
    BEGIN
        NormalizePoint (at);
        TransformPoint (ws^.Transform, at);
        Position := at;
        DrawCursor (ws);
    END SetCursor;


PROCEDURE GetCursor (ws: WS; VAR position: RVEC);
    VAR
        input:      Stream;
    BEGIN
        IF Stupid THEN
            input := FindRealInput ();
            Escape (ws, ReadCursor, input, Position);
        END (* if *);
        position := Position;
        IF UnTransformPoint (ws^.Transform, position) 
        AND UnNormalizePoint (position) THEN
        (*  goody  *)
        END (* if *);
    END GetCursor;


BEGIN
    Ws := NIL;
    RVec (Grid, 1.0, 1.0);

END GKSInput.
