IMPLEMENTATION MODULE GKSM;

(*
    Title: 	GKS Meta File Procedures
    LastEdit:	"Fri Feb 26 16:00:05 1988" by Tony Mann
    Previous:	"Tue Sep 17 19:15:22 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

FROM Vector IMPORT RVEC, RBOX, AXIS, BOUND;

FROM GKSTypes IMPORT TRANSFORM, PATH, GKSError;

FROM Streams IMPORT
    Stream, SeekMode, Get, Put, Delete, ErrorCode, Create, StreamProcs, Flush;
    
FROM StreamRep IMPORT Procs, PutProc, DeleteProc, DefaultProcs, NullFlush;

FROM SysStreams IMPORT sysErr;

FROM WriteF IMPORT WriteF1, WriteF2, WriteF0, H;

FROM CharCodes IMPORT NewLineCh;

FROM TextIO IMPORT 
    WriteREAL, WriteINTEGER, ReadChars, TermProc, 
    ReadCARDINAL, ReadINTEGER, ReadREAL, EndOfLine, WhiteSpace;

FROM StringStream IMPORT CreateInOutChars;

FROM StringF IMPORT StringF1;
    
FROM Storage IMPORT ALLOCATE, DEALLOCATE;

FROM SYSTEM IMPORT WORD;

FROM Strings IMPORT LengthC, EqualCC;


CONST EndCh = 0C;



VAR Sink,                                  (* the destination metafile   *)
    Source,                                (* the input metafile         *)
    Record: Stream;                        (* output buffer stream       *)
    Buffer: ARRAY [0..RecordHigh] OF CHAR; (* the record buffer          *)
    Pos:    INTEGER;                       (* next offset to read/write  *)
    Begun:  BOOLEAN;                       (* TRUE when a record open    *)

    InputLine,
    InputRecord,
    InputLength: INTEGER;                 (* of the current record      *)
    

PROCEDURE TransferDeleteProc (s: Stream);
BEGIN
END TransferDeleteProc;

PROCEDURE TransferPutProc (s: Stream; a: WORD);
BEGIN
    IF (Pos > RecordHigh) OR (Pos < 0) THEN
        s^.status := StreamError;
    ELSE Buffer [Pos] := CHAR (a);
        INC (Pos);
    END;
END TransferPutProc;

PROCEDURE NewStream (): Stream;
VAR procs: Procs;
BEGIN
    Pos := 0;
    procs := DefaultProcs (8);
    WITH procs^ DO
        Put := TransferPutProc;
        Delete := TransferDeleteProc;
    END;
    RETURN Create (StreamProcs (procs), 0)
END NewStream;

PROCEDURE InValid (s: Stream): BOOLEAN;
BEGIN
    IF s^.status # Success THEN
        Delete (s); DISPOSE (s);
        RETURN TRUE
    ELSE RETURN FALSE
    END;
END InValid;

PROCEDURE SelectMetaOutput (to: Stream): BOOLEAN;
BEGIN
    Sink := to;
    IF NOT InValid (to) THEN
        Record :=  NewStream ();
        Begun := FALSE;
        WriteF0 (to, "GKSM");
        WriteF0 (to, " CUCL Modula-2 metafile: author J. Dion ");
        WriteF0 (to, "20/09/82"); (* no dates *)
        WriteF0 (to, " 1"); (* version 1 of the metafile *)
        WriteF0 (to, " 4"); (* 4 bytes of "GKSM" in each record *)
        WriteF0 (to, " 3"); (* 3 bytes for record type *)
        WriteF0 (to, " 4"); (* 4 bytes for record length *)
        WriteF0 (to, " 0"); (* free format integers *)
        WriteF0 (to, " 0"); (* free format reals *)
        WriteF0 (to, " 1"); (* character form of number representation *)
        WriteF0 (to, " 1"); (* reals are reals *);
        WriteF0 (to, "           "); (* not significant *)
        WriteF0 (to, "           "); (* not significant *)
        WriteF0 (to, "\N");
        RETURN TRUE;
    ELSE Record := sysErr;
        RETURN FALSE;
    END;
END SelectMetaOutput;

PROCEDURE Substring (VAR s: ARRAY OF CHAR; at: CARDINAL; 
    sub: ARRAY OF CHAR): BOOLEAN;
    VAR i: CARDINAL;
BEGIN
    FOR i := 0 TO HIGH (sub) DO
        IF (at+i > HIGH (s)) OR (s[at+i] # sub[i]) THEN RETURN FALSE END;
    END;
    RETURN TRUE;
END Substring;


PROCEDURE InputError;
    VAR error: ARRAY [0..50] OF CHAR;
BEGIN
    StringF1 (error, "Metafile format error in line %I\N", InputLine);
    GKSError (error);
END InputError;


PROCEDURE SelectMetaInput (s: Stream): BOOLEAN;
    VAR header: ARRAY [0..90] OF CHAR;
    VAR ok: BOOLEAN;
BEGIN
    InputLine := 1;
    IF (ReadCharLength (s, header, EndOfLine) = 90)
    AND Substring (header, 0, "GKSM")
    AND Substring (header, 52, " 1") THEN
        Source := s;
        ok := TRUE;
    ELSE Source := sysErr;
        ok := FALSE;
        InputError;
    END;
    RETURN ok;
END SelectMetaInput;


PROCEDURE EndMetaOutput;
BEGIN
    WrRecord (EndRec);
    EndOutputRecord;
    Delete (Record); DISPOSE (Record);
    Delete (Sink); DISPOSE (Sink);
END EndMetaOutput;


PROCEDURE EndMetaInput;
BEGIN
    Delete (Source); DISPOSE (Source);
    Source := sysErr;
END EndMetaInput;


PROCEDURE BeginOutputRecord (type: INTEGER);
BEGIN
    WriteF1 (Sink, "GKSM%I3", type);
    Pos := 0;
    Begun := TRUE;
END BeginOutputRecord;


PROCEDURE EndOutputRecord;
    VAR i: INTEGER;
BEGIN
    IF Begun THEN
        Begun := FALSE;
        WriteF1 (Sink, "%I4", Pos);
        FOR i := 0 TO Pos-1 DO Put (Sink, Buffer[i]) END;
        WriteF0 (Sink, "\N");
    END;
END EndOutputRecord;


PROCEDURE WrRecord (type: RECORDTYPE);
BEGIN
    EndOutputRecord;
    BeginOutputRecord (ORD (type));
END WrRecord;

PROCEDURE RdRecord (VAR type: CARDINAL; VAR length: INTEGER);
    VAR tag: ARRAY [0..3] OF CHAR;
    VAR s: Stream;
BEGIN
    IF (CHAR (Get(Source)) = NewLineCh)
    AND (ReadCharLength (Source, tag, WhiteSpace) = 4)
    AND EqualCC (tag, "GKSM") THEN
        InputRecord := ReadCARDINAL (Source);
        InputLength := ReadCARDINAL (Source);
        type := InputRecord;
        length := InputLength;
        INC (InputLine);
    ELSE type := ORD (EndRec);
        length := 0;
        InputError;
    END;
END RdRecord;


PROCEDURE ReadCharLength
    (s: Stream; VAR chars: ARRAY OF CHAR; T: TermProc): CARDINAL;
BEGIN
    ReadChars (s, chars, T);
    RETURN LengthC (chars);
END ReadCharLength;


PROCEDURE WrUserRecord (type: INTEGER);
BEGIN
    EndOutputRecord;
    BeginOutputRecord (type);
END WrUserRecord;

PROCEDURE RdUserRecord (): Stream;
    VAR buffer: ARRAY [0..RecordHigh] OF CHAR;
    VAR i: INTEGER;
BEGIN
    FOR i := 0 TO InputLength-1 DO buffer[i] := CHAR (Get (Source)) END;
    IF InputLength <= RecordHigh THEN buffer [InputLength] := 0C END;
    RETURN CreateInOutChars (buffer);
END RdUserRecord;
        

PROCEDURE WrInt (i: INTEGER);
BEGIN
    WriteF1 (Record, " %I", i);
END WrInt;


PROCEDURE RdInt (VAR i: INTEGER);
    VAR s: Stream;
BEGIN
    i := ReadINTEGER (Source);
END RdInt;

PROCEDURE WrReal (r: REAL);
    VAR e: INTEGER;
    VAR real: REAL;
BEGIN
    e := 0;
    real := ABS (r);
    IF real > 10.0 THEN
        WHILE real >=10.0 DO INC (e); real := real / 10.0 END;
    ELSIF real # 0.0 THEN
        WHILE real < 0.1 DO DEC (e); real := real * 10.0 END;
    END;
    Put (Record, " ");
    IF r < 0.0 THEN Put (Record, "-") END;
    WriteREAL (Record, real, 6);
    IF e # 0 THEN 
        Put (Record, "E");
        WriteINTEGER (Record, e);
    END;
END WrReal;


PROCEDURE RdReal (VAR r: REAL);
BEGIN
    r := ReadREAL (Source);
END RdReal;

PROCEDURE WrPath (p: PATH);
    VAR d: INTEGER;
BEGIN
    CASE p OF 
        Right: d:=0; | Left: d:=1; | Up: d:=2; | Down: d:=3;
    END;
    WrInt (d);
END WrPath;


PROCEDURE RdPath (VAR p: PATH);
    VAR i: INTEGER;
BEGIN
    RdInt (i);
    CASE i OF 
        1: p := Left; | 2: p := Up; | 3: p := Down;
    ELSE p := Right;
    END;
END RdPath;


PROCEDURE WrVector (v: RVEC);
BEGIN
    WrReal (v[X]);
    WrReal (v[Y]);
END WrVector;


PROCEDURE RdVector (VAR v: RVEC);
BEGIN
    v[X] := ReadREAL (Source);
    v[Y] := ReadREAL (Source);
END RdVector;


PROCEDURE WrBox (r: RBOX);
BEGIN
    WrReal (r[Low, X]);
    WrReal (r[High, X]);
    WrReal (r[Low, Y]);
    WrReal (r[High, Y]);
END WrBox;


PROCEDURE RdBox (VAR r: RBOX);
BEGIN
    r[Low, X] := ReadREAL (Source);
    r[High, X] := ReadREAL (Source);
    r[Low, Y] := ReadREAL (Source);
    r[High, Y] := ReadREAL (Source);
END RdBox;


PROCEDURE WrPoints (high: INTEGER; VAR p: ARRAY OF RVEC);
    VAR i: INTEGER;
BEGIN
    WriteF1 (Record, " %I", high+1);
    FOR i := 0 TO high DO WrReal (p[i, X]); WrReal (p[i, Y]) END;
END WrPoints;


PROCEDURE RdPoints (VAR count: INTEGER; VAR p: ARRAY OF RVEC);
    VAR i: INTEGER;
BEGIN
    count := ReadCARDINAL (Source);
    FOR i := 0 TO INTEGER(count-1) DO
        p[i, X] := ReadREAL (Source);
        p[i, Y] := ReadREAL (Source);
    END;
END RdPoints;


PROCEDURE WrString (VAR string: ARRAY OF CHAR);
    VAR l, i: INTEGER;
BEGIN
    l := LengthC (string);
    WriteF1 (Record, " %I ", l);
    FOR i := 0 TO l-1 DO 
        Put (Record, string[i]) (* WriteF0 interprets escapes! *)
    END;
END WrString;


PROCEDURE RdString (VAR string: ARRAY OF CHAR);
    VAR l: CARDINAL;
    VAR i: INTEGER;
    VAR ch: CHAR;
BEGIN
    l := ReadCARDINAL (Source);
    ch := CHAR (Get (Source)); (* the terminator! *)
    FOR i := 0 TO INTEGER(l-1) DO
        ch := CHAR (Get (Source));
        IF i <= INTEGER (HIGH (string)) THEN string[i] := ch END;
    END;
    IF l <= HIGH (string) THEN string[l] := 0C END;
END RdString;


PROCEDURE WrPixels (rowhigh, colhigh: INTEGER;
    VAR pixels: ARRAY OF CARDINAL);
    VAR row, col, rowoffset: INTEGER;
BEGIN
    WriteF2 (Record, " %I %I", rowhigh + 1, colhigh + 1);
    FOR row := 0 TO rowhigh DO
        rowoffset := row * (colhigh+1);
        FOR col := 0 TO colhigh DO
            WriteF1 (Record, " %I", pixels [rowoffset + col]);
        END;
    END;
END WrPixels;


PROCEDURE RdPixels (VAR rows, cols: INTEGER; VAR pixels: ARRAY OF CARDINAL);
    VAR row, col, rowoffset: INTEGER;
BEGIN
    rows := ReadCARDINAL (Source);
    cols := ReadCARDINAL (Source);
    FOR row := 0 TO rows - 1 DO
        rowoffset := row * cols;
        FOR col := 0 TO cols - 1 DO
            pixels [rowoffset + col] := ReadCARDINAL (Source);
        END;
    END;
END RdPixels;


PROCEDURE WrTransform (t: TRANSFORM);
    VAR i, j: INTEGER;
BEGIN
    WITH t DO
        FOR i := 1 TO 3 DO
            FOR j := 1 TO 3 DO WrReal (Matrix [i,j]); END;
        END;
    END;
END WrTransform;


PROCEDURE RdTransform (VAR t: TRANSFORM);
    VAR i, j: INTEGER;
BEGIN
    WITH t DO
        FOR i := 1 TO 3 DO
            FOR j := 1 TO 3 DO Matrix [i,j] := ReadREAL (Source); END;
        END;
    END;
END RdTransform;

END GKSM.


