IMPLEMENTATION MODULE GKSMInput;

(*
    Title: 	GKS Meta File Procedures
    LastEdit:	"Wed Sep 18 15:29:39 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

FROM SYSTEM IMPORT WORD;

FROM Vector IMPORT RVEC, RBOX, AXIS, BOUND, RBox;

FROM GKSTypes IMPORT WS, TRANSFORM, PRECISION, PATH, GDPCODE,
    FirstWS;

FROM GKSM IMPORT
    RECORDTYPE, RecordHigh, RdInt, RdReal, RdVector, RdPoints, RdPixels,
    RdString, RdPath, RdBox, RdRecord, RdUserRecord, SelectMetaInput,
    EndMetaInput;

FROM GKS IMPORT
    Line, PolyLine, Marker, PolyMarker, Text, Fill, Pixels, Gdp, 

    SetPolyLineIndex, SetPolyLine,
    SetPolyMarkerIndex, SetPolyMarker,
    SetTextIndex, SetText,
    SetFillAreaIndex, SetFillArea, SetPattern,
    SetCharHeight, SetCharUpVector, SetCharExpand,
    SetCharPath, SetCharSpacing,
    SetPatternSize, SetPatternReferencePoint,
    SetColour,

    SelectNormalization, SetWindow, SetViewport,
    GetNormalizationIndex, GetNormalization, SetClipping,

    ClearWS, UpdateWS, SetDeferralState, Message, 
    SetWSWindow, SetWSViewport;

FROM GKSAttr IMPORT SetMetaChar, SetMetaPattern;

FROM Streams IMPORT Stream, Delete, ErrorCode;

FROM CharCodes IMPORT NewLineCh;

FROM SysStreams IMPORT sysErr;

FROM FileStream IMPORT CreateInput;

FROM Storage IMPORT DEALLOCATE;
    
FROM WriteF IMPORT
    WriteF1, WriteF2, WriteF0, H;


PROCEDURE InValid (s: Stream): BOOLEAN;
BEGIN
    IF s^.status # Success THEN
        DISPOSE (s);
        RETURN TRUE
    ELSE RETURN FALSE
    END;
END InValid;


PROCEDURE DiscardRecord (type, length: INTEGER): BOOLEAN;
BEGIN
    Delete (RdUserRecord ());
    RETURN TRUE;
END DiscardRecord;


PROCEDURE InterpretRecord (type: CARDINAL; length: INTEGER): BOOLEAN;
    VAR ws: WS;
    VAR i, j, index, style, colour: INTEGER;
    VAR rows, cols: INTEGER;
    VAR r, g, b, size: REAL;
    VAR vx, vy: RVEC;
    VAR p: PATH;
    VAR box: RBOX;
    VAR s: ARRAY [0..400] OF CHAR;
    VAR pixels: ARRAY [0..100] OF CARDINAL;
    VAR points: ARRAY [0..50] OF RVEC;
BEGIN
    ws := FirstWS;
    IF type <= ORD (SegPriority) THEN
        CASE VAL (RECORDTYPE, type) OF
          EndRec:
                EndMetaInput ();
        | ClearRec:
                RdInt (i);
                WHILE ws # NIL DO
                    IF ws^.Active THEN ClearWS (ws, i=1); END;
                    ws := ws^.Next
                END;
        | RedrawRec:
                RETURN FALSE; (* segments not implemented *)
        | UpdateRec:
                WHILE ws # NIL DO
                    IF ws^.Active THEN UpdateWS (ws); END;
                    ws := ws^.Next
                END;
        | MessageRec:
                RdString (s);
                WHILE ws # NIL DO
                    IF ws^.Active THEN Message (ws, s); END;
                    ws := ws^.Next
                END;
        | BoxRec:
                RdBox (box);
                SetWindow (GetNormalizationIndex (), box);
                SetClipping (TRUE);
        | LineIndexRec:
                RdInt (i);
                SetPolyLineIndex (i);
        | MarkIndexRec:
                RdInt (i);
                SetPolyMarkerIndex (i);
        | TextIndexRec:
                RdInt (i);
                SetTextIndex (i);
        | CharHeightRec:
                RdReal (r);
                SetCharHeight (r);
        | CharUpRec:
                RdVector (vy);
                RdVector (vx);
                SetMetaChar (vy, vx);
        | CharExpandRec:
                RdReal (r);
                SetCharExpand (r);
        | CharPathRec:
                RdPath (p);
                SetCharPath (p);
        | CharSpaceRec:
                RdReal (r);
                SetCharSpacing (r);
        | FillIndexRec:
                RdInt (i);
                SetFillAreaIndex (i);
        | PatternSizeRec:
                RdVector (vx); RdVector (vy);
                SetMetaPattern (vx, vy);
        | PatternRefRec:
                RdVector (vx);
                SetPatternReferencePoint (vx);
        | LineDefRec:
                RdInt (index); RdInt (style); RdReal (size); RdInt (colour);
                WHILE ws # NIL DO
                    IF ws^.Active THEN
                        SetPolyLine (ws, index, style, colour, size);
                    END;
                    ws := ws^.Next
                END;
        | MarkerDefRec:
                RdInt (index); RdInt (style); RdReal (size); RdInt (colour);
                WHILE ws # NIL DO
                    IF ws^.Active THEN
                        SetPolyMarker (ws, index, style, colour, size);
                    END;
                    ws := ws^.Next
                END;
        | TextDefRec:
                RdInt (index); RdInt (style); RdInt (i); RdInt (colour);
                WHILE ws # NIL DO
                    IF ws^.Active THEN
                        SetText (ws, index, style, colour, 
                            VAL (PRECISION, i));
                    END;
                    ws := ws^.Next;
                END;
        | FillDefRec:
                RdInt (index); RdInt (style); RdInt (i); RdInt (colour);
                WHILE ws # NIL DO
                    IF ws^.Active THEN
                        SetFillArea (ws, index, style, i, colour);
                    END;
                    ws := ws^.Next;
                END;
        | PatternDefRec:
                RdInt (index); RdPixels (rows, cols, pixels);
                WHILE ws # NIL DO
                    IF ws^.Active THEN
                        SetPattern (ws, index, rows, cols, pixels);
                    END;
                    ws := ws^.Next;
                END;
        | ColourDefRec:
                RdInt (index); RdReal (r); RdReal (g); RdReal (b);
                WHILE ws # NIL DO
                    IF ws^.Active THEN
                        SetColour (ws, index, r, g, b);
                    END;
                    ws := ws^.Next;
                END;
        | DeferralRec:
                RdInt (i); RdInt (j);
                WHILE ws # NIL DO
                    IF ws^.Active THEN
                        SetDeferralState (ws, i#0, j=0);
                    END;
                    ws := ws^.Next;
                END;
        | WindowRec:
                RdBox (box);
                WHILE ws # NIL DO
                    IF ws^.Active THEN
                        SetWSWindow (ws, box);
                    END;
                    ws := ws^.Next;
                END;
        | ViewportRec:
                RdBox (box);
                WHILE ws # NIL DO
                    IF ws^.Active THEN
                        SetWSViewport (ws, box);
                    END;
                    ws := ws^.Next;
                END;
        | LineRec:
            RdPoints (rows, points);
            PolyLine (rows, points);
        | MarkerRec:
            RdPoints (rows, points);
            PolyMarker (rows, points);
        | TextRec:
            RdVector (vx);
            RdString (s);
            Text (vx, s);
        | FillRec:
            RdPoints (rows, points);
            Fill (rows, points);
        | PixelRec:
            RdBox (box);
            RdPixels (rows, cols, pixels);
            Pixels (box, rows, cols, pixels);
        | GdpRec:
            RdPoints (rows, points);
            RdInt (style);
            RdInt (style);
            RdInt (j);
            RdInt (index);
            Gdp (VAL (GDPCODE, style), WORD (j), rows, points);
        | CreateSegRec, CloseSegRec, RenameSegRec, DeleteSetRec,
          SegTransRec,  SegVisibleRec, SegLitRec,  SegPriority:
            RETURN FALSE;
        ELSE RETURN FALSE;
        END;
        RETURN TRUE;
    ELSE RETURN FALSE;
    END;
END InterpretRecord;


PROCEDURE InterpretStream (s: Stream; p: RECORDPROC);
    VAR type, index: CARDINAL;
        length: INTEGER;
        w, v, id: RBOX;
BEGIN
    IF SelectMetaInput (s) THEN
        index := GetNormalizationIndex ();
        GetNormalization (index, w, v);
        RBox (id, 0.0, 0.0, 1.0, 1.0);
        SetWindow (index, id);
        SetViewport (index, id);
        REPEAT
            RdRecord (type, length);
            IF InterpretRecord (type, length)
            OR p (type, length)
            OR DiscardRecord (type, length) THEN
            END;
        UNTIL type = ORD (EndRec);
        SetWindow (index, w);
        SetViewport (index, v);
    END;
END InterpretStream;


PROCEDURE InterpretFile (name: ARRAY OF CHAR; p: RECORDPROC);
    VAR s: Stream;
BEGIN
    s := CreateInput (name);
    IF InValid (s) THEN
        WriteF0 (sysErr, "Can't read `");
        WriteF0 (sysErr, name);
        WriteF0 (sysErr, "'\N");
    ELSE InterpretStream (s, p)
    END;
END InterpretFile;


END GKSMInput.
