IMPLEMENTATION MODULE GKSMenu;

(*
    Title: 	menu manipulation under GKS
    LastEdit:	"Wed Oct  2 18:53:22 1985"
    Author: 	pr
		Cambridge University Computer Laboratory
*)


FROM Arguments IMPORT Arguments;
FROM Context IMPORT Context, Bind, Unbind, EnumerateContext, BoundName;
FROM Dispatch IMPORT FindContexts;
FROM Vector IMPORT AXIS, BOUND, RVEC, RVec, RBOX, RBox;

FROM WriteF IMPORT WriteF1, H;
FROM Streams IMPORT Stream;

FROM GKS IMPORT 
    GetNormalizationIndex, SetWindow, SelectNormalization, Text, 
    ActivateWS, DeactivateWS, GetActiveWS, UpdateWS;
FROM GKSTypes IMPORT WS, PRECISION;
FROM GKSDefs IMPORT DefineTexts, SelectText;
FROM GKSInput IMPORT CursorOutput, SetCursor, GetCursor;
FROM Palette IMPORT COLOUR, PALETTE;


CONST
    CommandLength = 20;
    CommandCount = 30;


TYPE
    COMMANDRANGE = [0 .. CommandCount];
    COMMANDLENGTH = [0 .. CommandLength];


VAR 
    WorkStation:    WS;
    OldNormalization,
    Normalization:  INTEGER;
    Redraw:         BOOLEAN;
    Contexts:       Context;
    Command:        ARRAY COMMANDRANGE, COMMANDLENGTH OF CHAR;
    Limit:          INTEGER;
    RememberedCursor: BOOLEAN;
    RememberedNormalization: BOOLEAN;
    OldCursor,
    Cursor:         RVEC;


PROCEDURE BindMenu (immediate: Context; display, select: ARRAY OF CHAR;
        enabled: BOOLEAN);
    BEGIN
        IF enabled THEN
            Bind (immediate, display, DisplayMenu);
            Bind (immediate, select, MenuSelect);
        ELSE
            Unbind (immediate, display);
            Unbind (immediate, select);
        END (* if *);
    END BindMenu;


PROCEDURE MenuInput (ws: WS; normalization: INTEGER; enabled: BOOLEAN);
    VAR
        window:         RBOX;
    BEGIN
        IF enabled THEN
            WorkStation := ws;
            Normalization := normalization;
            RBox (window, 0.0, 0.0,
                FLOAT (CommandLength), FLOAT (CommandCount));
            SetWindow (Normalization, window);
            CursorOutput (WorkStation, TRUE);
            DefineTexts (ws, PALETTE {Green, White}, 1, StringPrecision);
        ELSE
            WorkStation := NIL;
        END (* if *);
    END MenuInput;


PROCEDURE CancelMenu;
    VAR
        oldnormal:      INTEGER;
    BEGIN
        IF WorkStation # NIL THEN
            oldnormal := GetNormalizationIndex ();
            SelectNormalization (Normalization);
            Display (White);
            UpdateWS (WorkStation);
            Limit := -1;
            SelectNormalization (oldnormal);
        END (* if *);
        RedrawMenu;
    END CancelMenu;


PROCEDURE RedrawMenu;
    BEGIN
        Redraw := TRUE;
    END RedrawMenu;


PROCEDURE DisplayMenu (in, out: Stream);
    VAR
        newcursor:      RVEC;
        immediate,
        delayed:        Context;
        changedcontext: BOOLEAN;
    BEGIN
        IF (WorkStation # NIL) AND NOT RememberedNormalization THEN
            GetCursor (WorkStation, OldCursor);
            OldNormalization := GetNormalizationIndex ();
            SelectNormalization (Normalization);
            RememberedNormalization := TRUE;
            FindContexts (immediate, delayed);
            changedcontext := INTEGER (delayed) # INTEGER (Contexts);
            IF Redraw OR changedcontext THEN
                IF changedcontext THEN
                    IF NOT Redraw THEN
                        Display (White);
                    END (* if *);
                    Contexts := delayed;
                END (* if *);
                Limit := -1;
                EnumerateContext (in, out, Contexts, Record);
                IF changedcontext THEN
                    RVec (Cursor, FLOAT (CommandLength) / 2.0,
                        FLOAT (Limit + 1) / 2.0);
                END (* if *);
                Display (Green);
                Redraw := FALSE;
            END (* if *);
            GetCursor (WorkStation, newcursor);
            RememberedCursor := newcursor [X] < 0.0;
            IF RememberedCursor THEN
                SetCursor (WorkStation, Cursor);
            ELSE
                Cursor := newcursor;
            END (* if *);
            UpdateWS (WorkStation);
        END (* if *);
    END DisplayMenu;


PROCEDURE Record (in, out: Stream; context: Context);
    BEGIN
        INC (Limit);
        BoundName (context, Command [Limit]);
    END Record;


PROCEDURE Display (colour: COLOUR);
    VAR
        wss:            ARRAY [1 .. 10] OF WS;
        c, count:       INTEGER;
        command:        INTEGER;
        location:       RVEC;
    BEGIN
        GetActiveWS (wss, count);
        FOR c := 1 TO count DO
            IF wss [c] # WorkStation THEN
                DeactivateWS (wss [c]);
            END (* if *);
        END (* for *);
        SelectText (colour, 1, StringPrecision);
        FOR command := 0 TO Limit DO
            RVec (location, 0.0, FLOAT (command));
            Text (location, Command [command]);
        END (* for *);
        FOR c := 1 TO count DO
            IF wss [c] # WorkStation THEN
                ActivateWS (wss [c]);
            END (* if *);
        END (* for *);
    END Display;


PROCEDURE MenuSelect (in, out: Stream);
    CONST args = 0.9;
    VAR
        cursor:         RVEC;
        current:        INTEGER;
    BEGIN
        IF WorkStation # NIL THEN
            GetCursor (WorkStation, cursor);
            IF (cursor [X] >= 0.0) AND (cursor [Y] >= 0.0) THEN
                current := TRUNC (cursor [Y]);
                IF current <= Limit THEN
                    IF TRUNC (cursor [X] / args) < CommandLength THEN
                        WriteF1 (out, "%S\N", H (Command [current]));
                    ELSE WriteF1 (out, "%S ", H (Command [current]));
                    END;
                    Cursor [X] := cursor [X];
                    Cursor [Y] := FLOAT (current) + 0.5;
                END (* if *);
            END (* if *);
            SelectNormalization (OldNormalization);
            RememberedNormalization := FALSE;
            IF RememberedCursor THEN
                SetCursor (WorkStation, OldCursor);
                UpdateWS (WorkStation);
            END (* if *);
        END (* if *);
    END MenuSelect;


PROCEDURE ArgumentsMenu (in, out: Stream; format: ARRAY OF CHAR): BOOLEAN;
    BEGIN
        RETURN Arguments (in, out, format);
    END ArgumentsMenu;


BEGIN
    WorkStation := NIL;
    Redraw := TRUE;
    RememberedNormalization := FALSE;
    Contexts := Context (NIL);
    Limit := -1;

END GKSMenu.
