IMPLEMENTATION MODULE GKSNorm;

(*
    Title: 	GKS Normalization Transformations
    LastEdit:	"Tue Feb 26 14:25:56 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

FROM Vector IMPORT RVEC, AXIS, RBOX, BOUND;

FROM GKSTypes IMPORT
    RBOXPOINTER, TRANSFORM;


FROM GKSTrans IMPORT
    IdentityBox, IdentityTransform,
    ValidBox, ConstructTransform,
    TransformPoint, UnTransformPoint, TransformVector, TransformPolyLine;

FROM GKSAttr IMPORT
    ChangeNormalization;

FROM Storage IMPORT ALLOCATE;


TYPE
    NORMALIZE = POINTER TO NORMALIZEDATA;

    NORMALIZEDATA = RECORD
        Number:                INTEGER;
        Next:                  NORMALIZE;
        Transform:             TRANSFORM;
        Window:                RBOX;
        Viewport:              RBOXPOINTER;
        ViewportReference:     BOOLEAN;
                               (* if TRUE make a new Viewport on a change *)
    END;


VAR FirstNormalize, CurrentNormalize: NORMALIZE;

VAR IdentityClip: RBOXPOINTER;

CONST IdentityNormalize = 0;

VAR Clipping: BOOLEAN;


PROCEDURE GetNormalizationIndex (): INTEGER;
BEGIN
    RETURN CurrentNormalize^.Number;
END GetNormalizationIndex;

PROCEDURE GetNormalization (index: INTEGER; VAR w, v: RBOX);
    VAR n: NORMALIZE;
BEGIN
    n := FindNormalize (index);
    WITH n^ DO
        w := Window;
        v := Viewport^
    END;
END GetNormalization;


PROCEDURE ClippingBox (StoredRef: BOOLEAN): RBOXPOINTER;
BEGIN
    IF Clipping THEN
        IF StoredRef THEN CurrentNormalize^.ViewportReference := TRUE; END;
        RETURN CurrentNormalize^.Viewport;
    ELSE RETURN IdentityClip
    END;
END ClippingBox;


PROCEDURE FindNormalize (index: INTEGER): NORMALIZE;
    VAR normal: NORMALIZE;
BEGIN
    normal := FirstNormalize;
    WHILE normal # NIL DO
        IF normal^.Number = index THEN
                RETURN normal
        ELSE normal := normal^.Next
        END;
    END;
    NEW (normal);
    WITH normal^ DO
        Number := index;
        Next := FirstNormalize;
        ViewportReference := FALSE;
        NEW (Viewport);
        Window := IdentityBox;
        Viewport^ := IdentityBox;
        Transform := IdentityTransform;
    END;
    FirstNormalize := normal;
    RETURN normal;
END FindNormalize;


PROCEDURE SelectNormalization (index: INTEGER);
    VAR normal: NORMALIZE;
BEGIN
    CurrentNormalize := FindNormalize (index);
    ChangeNormalization (CurrentNormalize^.Transform);
END SelectNormalization;


PROCEDURE SetWindow (index: INTEGER; r: RBOX);
    VAR normal: NORMALIZE;
BEGIN
    IF (index # IdentityNormalize) AND ValidBox (r) THEN
        normal := FindNormalize (index);
        WITH normal^ DO
            Window := r;
            ConstructTransform (Window, Viewport^, FALSE, Transform);
        END;
        IF normal = CurrentNormalize THEN
            ChangeNormalization (normal^.Transform)
        END;
    END;
END SetWindow;

    
PROCEDURE SetViewport (index: INTEGER; r: RBOX);
    VAR normal: NORMALIZE;
BEGIN
    IF (index # IdentityNormalize) AND ValidBox (r) THEN
        normal := FindNormalize (index);
        WITH normal^ DO
            IF ViewportReference THEN
                NEW (Viewport);
                ViewportReference := FALSE;
            END;
            Viewport^ := r;
            ConstructTransform (Window, Viewport^, FALSE, Transform);
        END;
        IF normal = CurrentNormalize THEN
            ChangeNormalization (normal^.Transform)
        END;
    END;
END SetViewport;


PROCEDURE SetClipping (on: BOOLEAN);
BEGIN
    Clipping := on;
END SetClipping;


PROCEDURE NormalizePoint (VAR p: RVEC);
BEGIN
    TransformPoint (CurrentNormalize^.Transform, p);
END NormalizePoint;


PROCEDURE UnNormalizePoint (VAR p: RVEC): BOOLEAN;
BEGIN
    RETURN UnTransformPoint (CurrentNormalize^.Transform, p);
END UnNormalizePoint;


PROCEDURE NormalizeVector (VAR v: RVEC);
BEGIN
    TransformVector (CurrentNormalize^.Transform, v);
END NormalizeVector;


PROCEDURE NormalizeLine (high: INTEGER; VAR p: ARRAY OF RVEC);
BEGIN
    TransformPolyLine (CurrentNormalize^.Transform, high, p)
END NormalizeLine;


BEGIN
    FirstNormalize := NIL;
    CurrentNormalize := NIL;
    SelectNormalization (IdentityNormalize);
    IdentityClip := CurrentNormalize^.Viewport;
    SetClipping (FALSE);
END GKSNorm.
