IMPLEMENTATION MODULE GKSSeg;

(*
    Title: 	GKS Segment Functions
    LastEdit:	"Tue Feb 26 14:25:36 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

FROM Vector IMPORT RVEC, AXIS, RBOX, BOUND;
FROM Strings IMPORT String, CopySC, CopyCS;

FROM GKSTypes IMPORT
    WS, FirstWS, TRANSFORM,
    RVECS, StorePoints, PIXELS, StorePixels, DropSegment,
    ATTRIBUTE, PRIMITIVE, PRIMTYPE, SEGMENT, SEGLINK, GDPCODE;

FROM GKSWork IMPORT
    ClearWS, UpdateWS, SelectDisplay,
    DisplayLine, DisplayMarker, DisplayText, DisplayArea,
    DisplayPixels, DisplayGdp;

FROM GKSNorm IMPORT
    ClippingBox;

FROM GKSTrans IMPORT
    CheckIdentity, IdentityTransform;

FROM Storage IMPORT ALLOCATE;

FROM SYSTEM IMPORT WORD;


PROCEDURE DrawPrimitive (ws: WS; s: SEGMENT; p: PRIMITIVE);
VAR text: ARRAY [0..255] OF CHAR;
BEGIN
    SelectDisplay (ws, s, p^.Clip);
    WITH p^ DO
        CASE Type OF
            LinePrim:
                DisplayLine (LineIndex, LineAt^.High, LineAt^.Points);
          | MarkerPrim:
                DisplayMarker (MarkerIndex, MarkAt^.High, MarkAt^.Points);
          | TextPrim:
                CopySC (String, text);
                DisplayText (TextAt, TextIndex, Char, text);
          | AreaPrim:
                DisplayArea (AreaIndex, Boundary^.High, PatternRef,
                        Boundary^.Points);
          | PixelPrim:
                DisplayPixels (Box,
                        Pixels^.RowHigh, Pixels^.ColumnHigh, Pixels^.Pixels);
          | GdpPrim:
                DisplayGdp (Function, Arg, GdpPoints^.High, GdpPoints^.Points);
        ELSE END;
    END;
END DrawPrimitive;


PROCEDURE DrawSeg (ws: WS; seg: SEGMENT);
    VAR p: PRIMITIVE;
    VAR T: TRANSFORM;
    VAR R: RBOX;
BEGIN
    p := seg^.Primitive;
    WHILE p # NIL DO
        DrawPrimitive (ws, seg, p);
        p := p^.Next
    END;
END DrawSeg;


PROCEDURE RedrawWS (ws: WS);
    VAR seglink: SEGLINK;
BEGIN
    UpdateWS (ws);
    ClearWS (ws, FALSE);
    seglink := ws^.Segments;
    WHILE seglink # NIL DO
        DrawSeg (ws, seglink^.Segment);
        seglink := seglink^.Next
    END;
END RedrawWS;


VAR LastPrimitive: PRIMITIVE; (* the last primitive on the open segment *)


PROCEDURE CreateSegment (): SEGMENT;
    VAR ws: WS;
BEGIN
    IF OpenSegment # NIL THEN CloseSegment () END;
    NEW (OpenSegment);
    WITH OpenSegment^ DO
        Transform := IdentityTransform;
        Visible := TRUE;
        Lit := FALSE;
        Priority := 0.0;
        Primitive := NIL;
    END;
    LastPrimitive := NIL;
    ws := FirstWS;
    WHILE ws # NIL DO
        IF ws^.Active THEN AssociateWithWS (ws, OpenSegment) END;
        ws := ws^.Next;
        END;
    RETURN OpenSegment;
END CreateSegment;


PROCEDURE CloseSegment ();
BEGIN
    OpenSegment := NIL;
    LastPrimitive := NIL;
END CloseSegment;


PROCEDURE DeleteSegment (s: SEGMENT);
BEGIN
END DeleteSegment;


PROCEDURE DeleteFromWS (ws: WS; s: SEGMENT);
BEGIN
END DeleteFromWS;


PROCEDURE AssociateWithWS (ws: WS; s: SEGMENT);
BEGIN
END AssociateWithWS;


PROCEDURE CopyToWS (ws: WS; s: SEGMENT);
BEGIN
END CopyToWS;


PROCEDURE MakeTransform (t: TRANSFORMATION; VAR T: TRANSFORM);
    VAR i, j: INTEGER;
BEGIN
    FOR i := 1 TO 2 DO FOR j := 1 TO 3 DO T.Matrix [i,j] := t [i,j] END END;
    WITH T DO
        Matrix [3,1] := 0.0;
        Matrix [3,2] := 0.0;
        Matrix [3,3] := 1.0;
    END;
    CheckIdentity (T)
END MakeTransform;


PROCEDURE InsertSegment (s: SEGMENT; t: TRANSFORMATION);
BEGIN
END InsertSegment;


PROCEDURE SetTransformation (seg: SEGMENT; t: TRANSFORMATION);
BEGIN
    MakeTransform (t, seg^.Transform);
    DropSegment (seg);
END SetTransformation;


PROCEDURE SetVisibility (s: SEGMENT; visible: BOOLEAN);
BEGIN
END SetVisibility;


PROCEDURE SetHighlighting (s: SEGMENT; highlit: BOOLEAN);
BEGIN
END SetHighlighting;


PROCEDURE SetPriority (s: SEGMENT; priority: REAL);
BEGIN
END SetPriority;


PROCEDURE NewPrimitive (VAR p: PRIMITIVE; type: PRIMTYPE): BOOLEAN;
BEGIN
    IF OpenSegment # NIL THEN
        NEW (p);
        p^.Type := type;
        p^.Next := NIL;
        p^.Clip := ClippingBox (TRUE);;
        IF LastPrimitive = NIL THEN
            OpenSegment^.Primitive := p
        ELSE LastPrimitive^.Next := p
        END;
        RETURN TRUE
    ELSE RETURN FALSE
    END;
END NewPrimitive;


PROCEDURE StoreLine (index, high: INTEGER;
    VAR points: ARRAY OF RVEC): PRIMITIVE;
    VAR p: PRIMITIVE;
BEGIN
    IF NewPrimitive (p, LinePrim) THEN
        WITH p^ DO
            LineIndex := index;
            StorePoints (LineAt, high, points);
        END;
        RETURN p
    ELSE RETURN NIL;
    END;
END StoreLine;


PROCEDURE StoreMarker (index, high: INTEGER;
    VAR points: ARRAY OF RVEC): PRIMITIVE;
    VAR p: PRIMITIVE;
BEGIN
    IF NewPrimitive (p, MarkerPrim) THEN
        WITH p^ DO
            MarkerIndex := index;
            StorePoints (MarkAt, high, points);
        END;
        RETURN p
    ELSE RETURN NIL
    END;
END StoreMarker;


PROCEDURE StoreText (index: INTEGER; at: RVEC;
    char: ATTRIBUTE; VAR text: ARRAY OF CHAR): PRIMITIVE;
    VAR p:  PRIMITIVE;
BEGIN
    IF NewPrimitive (p, TextPrim) THEN
        WITH p^ DO
            TextIndex := index;
            Char := char;
            String := CopyCS (text);
            TextAt := at;
        END;
        RETURN p
    ELSE RETURN NIL
    END;
END StoreText;


PROCEDURE StoreFillArea (ref: ATTRIBUTE;
    index, high: INTEGER; VAR points: ARRAY OF RVEC): PRIMITIVE;
    VAR p: PRIMITIVE;
BEGIN
    IF NewPrimitive (p, AreaPrim) THEN
        WITH p^ DO
            AreaIndex := index;
            PatternRef := ref;
            StorePoints (Boundary, high, points);
        END;
        RETURN p
    ELSE RETURN NIL
    END;
END StoreFillArea;

PROCEDURE StorePixelArray (in: RBOX;
    rowhigh, columnhigh: INTEGER; pixels: ARRAY OF CARDINAL): PRIMITIVE;
    VAR p: PRIMITIVE;
BEGIN
    IF NewPrimitive (p, PixelPrim) THEN
        WITH p^ DO
            Box := in;
            StorePixels (Pixels, rowhigh, columnhigh, pixels);
        END;
        RETURN p
    ELSE RETURN NIL
    END;
END StorePixelArray;


PROCEDURE StoreGdp (function: GDPCODE; arg: WORD; high: INTEGER; 
    VAR points: ARRAY OF RVEC): PRIMITIVE;
    VAR p: PRIMITIVE;
BEGIN
    IF NewPrimitive (p, GdpPrim) THEN
        WITH p^ DO
            Function := function;
            Arg := arg;
            StorePoints (GdpPoints, high, points);
        END;
        RETURN p
    ELSE RETURN NIL
    END;
END StoreGdp;


BEGIN
    OpenSegment := NIL;

END GKSSeg.

