IMPLEMENTATION MODULE GKSTypes;

(*
    Title: 	A Few Generally Useful Operations on GKS Internal Types
    LastEdit:	"Mon Feb 25 17:22:51 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)
FROM Streams IMPORT Stream;
FROM SysStreams IMPORT sysErr;
FROM WriteF IMPORT WriteF0;
FROM Arithmetic IMPORT SquareRoot;
FROM Vector IMPORT RVEC, BOUND, AXIS, RBOX, RBox;
FROM Storage IMPORT ALLOCATE;

FROM SYSTEM IMPORT TSIZE;


PROCEDURE SelectWS (ws: WS);
BEGIN
    (* SelectModule (ws^.GLOBALFRAME); *)
    CurrentWS := ws;
END SelectWS;




PROCEDURE DropAttributes (ws: WS);
BEGIN
    WITH ws^ DO
        CurrentAttributes [CharAttr] := NIL;
        CurrentAttributes [PatternRefAttr] := NIL;
    END;
END DropAttributes;

PROCEDURE DropTransform (ws: WS);
BEGIN
    WITH ws^ DO
        DisplayTransform := Transform;
        DisplayClip := NIL;
        RBox (DisplayBox, 0.0, 0.0, 1.0, 1.0);
    END;
END DropTransform;


PROCEDURE DropSegment (seg: SEGMENT);
    VAR ws: WS;
BEGIN
    ws := FirstWS;
    WHILE ws # NIL DO
        IF ws^.DisplaySeg = seg THEN
            DropTransform (ws);
            ws^.DisplaySeg := NIL;
        END;
        ws := ws^.Next;
    END;
END DropSegment;

PROCEDURE SetFontAspect (aspect: REAL);
BEGIN
    WITH CurrentWS^ DO FontAspect := aspect END;
    DropAttributes (CurrentWS);
END SetFontAspect;

PROCEDURE StandardError (e: ARRAY OF CHAR);
BEGIN
    WriteF0 (sysErr, e);
END StandardError;


PROCEDURE LineStyle (selector: CARDINAL): LINESTYLE;
BEGIN
    IF (selector >= ORD (SolidLine))
    AND (selector <= ORD (DashedDottedLine)) THEN
        RETURN VAL (LINESTYLE, selector)
    ELSE RETURN UnspecifiedLine
    END;
END LineStyle;


PROCEDURE MarkerStyle (selector: CARDINAL): MARKERSTYLE;
BEGIN
    IF (selector >= ORD (DotMark))
    AND (selector <= ORD (XMark)) THEN
        RETURN VAL (MARKERSTYLE, selector)
    ELSE RETURN UnspecifiedMark
    END
END MarkerStyle;


PROCEDURE FillStyle (selector: CARDINAL): FILLSTYLE;
BEGIN
    IF selector < ORD (HatchArea) THEN
        RETURN VAL (FILLSTYLE, selector);
    ELSE RETURN SolidArea;
    END;
END FillStyle;


PROCEDURE StorePoints (VAR points: RVECS; high: INTEGER;
    VAR p: ARRAY OF RVEC);
    VAR i: INTEGER;
BEGIN
    ALLOCATE (points, TSIZE (INTEGER) + (high+1) * TSIZE (RVEC));
    WITH points^ DO
        High := high;
        FOR i := 0 TO high DO Points [i] := p [i] END;
    END;
END StorePoints;


PROCEDURE StorePixels (VAR pixels: PIXELS; rowhigh, columnhigh: INTEGER;
    VAR p: ARRAY OF CARDINAL);
    VAR i, entries: INTEGER;
BEGIN
    entries := (rowhigh+1) * (columnhigh+1);
    ALLOCATE (pixels, (2 + entries) * TSIZE (INTEGER));
    WITH pixels^ DO
        RowHigh := rowhigh;
        ColumnHigh := columnhigh;
        FOR i := 0 TO entries - 1 DO Pixels [i] := p [i] END;
    END;
END StorePixels;


BEGIN
    FirstWS := NIL;
    CurrentWS := NIL;
    GKSError := StandardError;
END GKSTypes.
