IMPLEMENTATION MODULE GKSWork;

(*
    Title: 	Graphical Kernel System Workstation Interface
    LastEdit:	"Wed Feb 26 10:30:59 1986"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

FROM Vector IMPORT RVEC, AXIS, RBOX, BOUND;

FROM GKSTypes IMPORT
    WS, TRANSFORM, RBOXPOINTER, SEGMENT, SEGLINK, PIXELARRAY,
    ATTRTYPE, ATTRIBUTE, GKSError, FirstWS, CurrentWS, SelectWS,
    DropAttributes, DropTransform, ESCAPECODE, GDPCODE;

FROM GKSAttr IMPORT
    SelectPen, SelectAttribute,
    DefaultAttributes;

FROM GKSDev IMPORT
    OpenDev, CloseDev, ActivateDev, DeactivateDev,
    ClearDev, UpdateDev, DevEscape, DevMessage,
    DevLine, DevMarker, DevText, DevFill, DevPixels, DevGdp;

FROM GKSTrans IMPORT
    IdentityBox, IdentityTransform,
    ComposeTransforms, ConstructTransform,
    TransformPoint, TransformPolyLine, TransformBox;

FROM GKSClip IMPORT
    ClipBox, ClipPoints, ClipPolyLine, ClipPolygon;

FROM GKSM IMPORT
    RECORDTYPE, SelectMetaOutput, EndMetaOutput, WrRecord, 
    WrInt, WrReal, WrVector, WrBox, WrPoints, WrString, WrPixels;

FROM Arithmetic IMPORT RMin, CMin;

FROM Strings IMPORT EqualCC;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;

FROM M2Files IMPORT CreateOutput;

FROM WriteF IMPORT H;

FROM Streams IMPORT Stream, ErrorCode, Delete;

FROM StringF IMPORT StringF1;

FROM SYSTEM IMPORT WORD;


(* In this module, as in other places in GKS, metafile output is treated as
   a special case. It has a WS allocated to it, but no procedures in GKSDev
   are called to do output. This should be changed if dynamic selection of
   different implentations of an interface (GKSDev in this case) ever becomes
   possible, and is only done as a stopgap measure to allow metafile creation
   while some other workstation is drawing
*)


CONST HighPoint = 50;
      Ext       = "mf";

TYPE RVECBUFFER = ARRAY [0..HighPoint] OF RVEC;


PROCEDURE InValid (s: Stream): BOOLEAN;
BEGIN
    IF s^.status # Success THEN
        Delete (s); DISPOSE (s);
        RETURN TRUE
    ELSE RETURN FALSE
    END;
END InValid;


PROCEDURE MakeWSTransform (ws: WS);
    VAR T: TRANSFORM;
BEGIN
    WITH ws^ DO
        ConstructTransform (Window, Viewport, TRUE, T);
        ComposeTransforms (T, PixelTransform, Transform);
        DropTransform (ws);
        DropAttributes (ws);
    END;
END MakeWSTransform;


PROCEDURE OpenWS (type, instance: ARRAY OF CHAR): WS;
    VAR ws: WS;
    (* VAR frame: GLOBALFRAME; *)
    VAR PixelSurface: RBOX;
    VAR s: Stream;
    VAR error: ARRAY [0..100] OF CHAR;
BEGIN
    NEW (ws);
    WITH ws^ DO
        (* frame := FindModule (type); *)
        DeferUpdates := TRUE;
        RedrawImplicitly := FALSE;
        IF EqualCC (type, "GKSM") THEN
            s := CreateOutput (instance, Ext);
            IF InValid (s) OR NOT SelectMetaOutput (s) THEN
                StringF1 (error, "Unable to open %s\N", H(instance));
                GKSError (error);
                DISPOSE (ws);
                RETURN NIL;
            ELSE Metafile := TRUE;
                PixelSurface := IdentityBox;
                DisplaySurface := IdentityBox;
            END;
        ELSE SelectWS (ws);
            OpenDev (instance, DisplaySurface, PixelSurface,
                DeferUpdates, RedrawImplicitly);
            Metafile := FALSE;
        END;
        Active := FALSE;
        Segments := NIL;
        DisplayClip := NIL;
        DisplayBox := IdentityBox;
        DisplaySeg := NIL;
        PendingRedraw := FALSE;
        DefaultAttributes (ws);
        Window := IdentityBox;
        ConstructTransform (DisplaySurface, PixelSurface, 
            FALSE, PixelTransform);
        Viewport := DisplaySurface;
        MakeWSTransform (ws);
        DisplayTransform := Transform;
    END;
    ws^.Next := FirstWS;
    FirstWS := ws;
    RETURN ws;
END OpenWS;


PROCEDURE CloseWS (ws: WS);
    VAR w: WS;
    VAR a: ATTRIBUTE;
    VAR s: SEGLINK;
BEGIN
    DeactivateWS (ws);
    IF ws^.Metafile THEN
        EndMetaOutput;
    ELSE CloseDev ();
    END;
    IF ws = FirstWS THEN
        FirstWS := ws^.Next
    ELSE w := FirstWS;
        WHILE w^.Next # ws DO w := w^.Next END;
        w^.Next := ws^.Next;
    END;
    WHILE ws^.Attributes # NIL DO
        a := ws^.Attributes;
        ws^.Attributes := a^.Next;
        DISPOSE (a)
    END;
    WHILE ws^.Segments # NIL DO
        s := ws^.Segments;
        ws^.Segments := s^.Next;
        DISPOSE (s);
    END;
    DISPOSE (ws);
END CloseWS;


PROCEDURE ActivateWS (ws: WS);
BEGIN
    IF NOT ws^.Active THEN
        IF NOT ws^.Metafile THEN
            SelectWS (ws);
            ActivateDev;
            UpdateDev;
        END;
        ws^.Active := TRUE;
    END;
END ActivateWS;


PROCEDURE DeactivateWS (ws: WS);
BEGIN
    IF ws^.Active THEN
        IF NOT ws^.Metafile THEN
            SelectWS (ws);
            DeactivateDev;
            UpdateDev;
        END;
        ws^.Active := FALSE;
    END;
END DeactivateWS;


PROCEDURE ClearWS (ws: WS; obligatory: BOOLEAN);
BEGIN
    IF ws^.Metafile THEN
        WrRecord (ClearRec);
        WrInt (ORD (obligatory));
    ELSE SelectWS (ws);
        ClearDev (obligatory);
    END;
END ClearWS;


PROCEDURE UpdateWS (ws: WS);
BEGIN
    IF ws^.Metafile THEN
        WrRecord (UpdateRec);
    ELSE SelectWS (ws);
        UpdateDev;
    END;
END UpdateWS;


PROCEDURE SetDeferralState (ws: WS; AllowUpdateDeferral,
    AllowImplicitRedraw: BOOLEAN);
BEGIN
    WITH ws^ DO
        DeferUpdates := AllowUpdateDeferral;
        RedrawImplicitly := AllowImplicitRedraw;
    END;
    IF ws^.Metafile THEN
        WrRecord (DeferralRec);
        WrInt (3 * ORD (AllowUpdateDeferral));
        WrInt (ORD (NOT AllowImplicitRedraw));
    END;
END SetDeferralState;


PROCEDURE Message (ws: WS; VAR s: ARRAY OF CHAR);
BEGIN
    IF ws^.Metafile THEN
        WrRecord (MessageRec);
        WrString (s);
    ELSE SelectWS (ws);
        DevMessage (HIGH (s), s);
    END;
END Message;
 

PROCEDURE Escape (ws: WS; function: ESCAPECODE; VAR w: WORD; 
        VAR a: ARRAY OF WORD);
BEGIN
    IF NOT ws^.Metafile THEN
        SelectWS (ws);
        DevEscape (function, w, a);
    END;
END Escape;


PROCEDURE SetWSWindow (ws: WS; r: RBOX);
BEGIN
    IF ClipBox (r, IdentityBox) THEN
        WITH ws^ DO
            Window := r;
            MakeWSTransform (ws);
        END;
        IF ws^.Metafile THEN
            WrRecord (WindowRec);
            WrBox (r);
        END;
    END;
END SetWSWindow;

    
PROCEDURE SetWSViewport (ws: WS; r: RBOX);
BEGIN
    SelectWS (ws);
    IF ClipBox (r, ws^.DisplaySurface) THEN
        WITH ws^ DO
            Viewport := r;
            MakeWSTransform (ws);
        END;
        IF ws^.Metafile THEN
            WrRecord (ViewportRec);
            WrBox (r);
        END;
    END;
END SetWSViewport;



MODULE Display;
(*
    This module maintains encached information for each workstation which
    determines the transformation and clipping to be done at the time of
    displaying it. This information is kept in the WS, and depends both
    on the segment from which we are drawing and the workstation on which
    we are drawing. The module exports the clipping box and
    transformation for the currently selected module.
    
    The encached transformation must be recomputed whenever 
        1) the workstation transformation changes
        2) the current segment for this workstation changes
        3) the transformation of the segment current on this workstation
           changes.
*)

    IMPORT
        TRANSFORM, AXIS, RVEC, RBOX, BOUND, RBOXPOINTER, 
        WS, SEGMENT, RECORDTYPE,
        FirstWS, IdentityTransform, IdentityBox,
        SelectWS, DropAttributes,
        ComposeTransforms, ClipBox, TransformPoint,
        WrRecord, WrBox;
 
    EXPORT SelectDisplay,
        TryDisplay, DisplayBox, DisplayTransform, 
        VisibleBox;

    VAR DisplayTransform: TRANSFORM;
    VAR DisplayBox, VisibleBox: RBOX;
    VAR TryDisplay: BOOLEAN;   (* IF FALSE, DisplayBox has area  0 *)


    PROCEDURE SelectDisplay (ws: WS; s: SEGMENT; r: RBOXPOINTER);
        VAR b: BOUND;
    BEGIN
        SelectWS (ws); 
        IF ws^.DisplaySeg # s THEN
            IF ((s # NIL) AND NOT s^.Transform.Identity) 
            OR ((ws^.DisplaySeg # NIL) 
                AND NOT ws^.DisplaySeg^.Transform.Identity) THEN
                ComposeTransforms (s^.Transform, ws^.Transform,
                    ws^.DisplayTransform);
                DropAttributes (ws)
            END;
            ws^.DisplaySeg := s;
        END;
        DisplayTransform := ws^.DisplayTransform;

        IF ws^.DisplayClip # r THEN
            ws^.DisplayClip := r;
            ws^.DisplayBox := r^;
            IF ws^.Metafile THEN
                WrRecord (BoxRec); WrBox (r^);
            END;
            TryDisplay := ClipBox (ws^.DisplayBox, ws^.Window);
            ws^.VisibleBox := ws^.DisplayBox;
            FOR b := Low TO High DO
                TransformPoint (ws^.Transform, ws^.VisibleBox [b]);
            END;
        END;
        DisplayBox := ws^.DisplayBox;
        VisibleBox := ws^.VisibleBox;
    END SelectDisplay;

END Display;


PROCEDURE DisplayLine (index, high: INTEGER; VAR points: ARRAY OF RVEC);
    VAR low, tohigh: INTEGER;
    VAR to: RVECBUFFER;
BEGIN
    SelectPen (CurrentWS, LineAttr, index);
    IF CurrentWS^.Metafile THEN
        WrRecord (LineRec);
        WrPoints (high, points);
    ELSIF TryDisplay THEN
        low := 0;
        tohigh := HighPoint;
        WHILE ClipPolyLine (low, high, points, tohigh, to,
            DisplayBox) DO
            TransformPolyLine (DisplayTransform, tohigh, to);
            DevLine (tohigh, to);
            tohigh := HighPoint;
        END;
    END;
END DisplayLine;


PROCEDURE DisplayMarker (index, high: INTEGER; VAR points: ARRAY OF RVEC);
    VAR low, tohigh: INTEGER;
    VAR to: RVECBUFFER;
BEGIN
    SelectPen (CurrentWS, MarkerAttr, index);
    IF CurrentWS^.Metafile THEN
        WrRecord (MarkerRec);
        WrPoints (high, points);
    ELSIF TryDisplay THEN
        low := 0;
        tohigh := HighPoint;
        IF ClipPoints (high, points, tohigh, to, DisplayBox) THEN
            TransformPolyLine (DisplayTransform, tohigh, to);
            DevMarker (tohigh, to);
        END;
    END;
END DisplayMarker;


PROCEDURE DisplayText (at: RVEC; index: INTEGER; char: ATTRIBUTE; 
    VAR text: ARRAY OF CHAR);
BEGIN
    (* Text MUST be selected before char to set font height & width *)
    SelectPen (CurrentWS, TextAttr, index);
    SelectAttribute (CurrentWS, char, DisplayTransform);
    IF CurrentWS^.Metafile THEN
        WrRecord (TextRec);
        WrVector (at);
        WrString (text);
    ELSIF TryDisplay THEN
        TransformPoint (DisplayTransform, at);
        DevText (at, VisibleBox, text);
    END;
END DisplayText;


PROCEDURE DisplayArea (index, high: INTEGER;
    patternref: ATTRIBUTE; VAR points: ARRAY OF RVEC);
    VAR tohigh: INTEGER;
    VAR to: RVECBUFFER;
BEGIN
    SelectPen (CurrentWS, AreaAttr, index);
    SelectAttribute (CurrentWS, patternref, DisplayTransform);
    IF CurrentWS^.Metafile THEN
        WrRecord (FillRec);
        WrPoints (high, points);
    ELSIF TryDisplay THEN
        tohigh := HighPoint;
        IF ClipPolygon (high, points, tohigh, to, DisplayBox) THEN
            TransformPolyLine (DisplayTransform, tohigh, to);
            DevFill (tohigh, to);
        END;
    END;
END DisplayArea;


PROCEDURE DisplayPixels (in: RBOX;
    RowHigh, ColumnHigh: INTEGER; VAR pixels: ARRAY OF CARDINAL);
    VAR parallelogram : ARRAY [0..3] OF RVEC;
BEGIN
    IF CurrentWS^.Metafile THEN
        WrRecord (PixelRec);
        WrBox (in);
        WrPixels (RowHigh, ColumnHigh, pixels);
    ELSIF TryDisplay THEN
        TransformBox (DisplayTransform, in, parallelogram);
        DevPixels (parallelogram, VisibleBox,
            RowHigh, ColumnHigh, pixels);
    END;
END DisplayPixels;


PROCEDURE DisplayGdp (function: GDPCODE; arg: WORD;
    high: INTEGER; VAR points: ARRAY OF RVEC);
    VAR i, tohigh: INTEGER;
    VAR to: RVECBUFFER;
BEGIN
    IF CurrentWS^.Metafile THEN
        WrRecord (GdpRec);
        WrPoints (high, points);
        WrInt (2);
        WrInt (INTEGER (ORD (function)));
        WrInt (INTEGER (arg));
        WrInt (0);
    ELSIF TryDisplay THEN
        tohigh := CMin (high, HighPoint);
        FOR i := 0 TO tohigh DO to [i] := points [i] END;
        TransformPolyLine (DisplayTransform, tohigh, to);
        DevGdp (function, arg, tohigh, to, VisibleBox);
    END;
END DisplayGdp;


END GKSWork.
