IMPLEMENTATION MODULE Numerical;

(*
    Title: 	Proper mathematical functions etc.
    LastEdit:	"Tue Aug 20 14:26:54 1985"
    Author: 	ahj
		Cambridge University Computer Laboratory
*)

FROM SYSTEM IMPORT MAXCARD ;

FROM UtilError IMPORT ReportPError ;

VAR Index : CARDINAL ;

PROCEDURE Floor (r : REAL) : INTEGER ;
    BEGIN
        IF r >= 0.0 THEN
            RETURN INTEGER (TRUNC (r))
        ELSE
            RETURN - INTEGER (TRUNC (- r * justLessThanOne) + 1)
        END (* if *);
    END Floor;

PROCEDURE Ceiling (r : REAL) : INTEGER ;
    BEGIN
        RETURN - Floor (- r) ;
    END Ceiling;

PROCEDURE Power (n : REAL ;
                 i : INTEGER) : REAL ;
    BEGIN
        IF i > 1 THEN
            IF ODD (i) THEN
                RETURN n * Power (n*n, i DIV 2)
            ELSE
                RETURN Power (n*n, i DIV 2)
            END (* if *);
        ELSIF i = 1 THEN
            RETURN n
        ELSIF i < 0 THEN
            IF n = 0.0 THEN
                ReportPError ("Numerical", TRUE,
                              "Trying to raise zero to the power %I0",
                              i, 0)
            ELSE
                RETURN 1.0 / Power (n, -i)
            END (* if *);
        ELSIF i = 0 THEN
            IF n = 0.0 THEN
                RETURN 1.0 (* Make binomial statistics work correctly *)
            ELSE
                RETURN 1.0
            END (* if *)
        END (* if *);
    END Power;

PROCEDURE Round (Value : REAL) : REAL ;
    BEGIN
        RETURN FLOAT (RoundToInt (Value))
    END Round;

PROCEDURE RoundToInt (Value : REAL) : INTEGER ;
    BEGIN
        IF Value >= 0.0 THEN
            RETURN TRUNC (Value + 0.5)
        ELSE
            RETURN - INTEGER (TRUNC (0.5 - Value))
        END (* if *);
        
    END RoundToInt;

PROCEDURE Frac (Value : REAL) : REAL ;
    BEGIN
        RETURN Value - FLOAT (TRUNC (Value))
    END Frac;

PROCEDURE TrueMOD (a : INTEGER ; b : CARDINAL) : CARDINAL ;
    BEGIN
        IF a >= 0 THEN
            RETURN CARDINAL (a) MOD b
        ELSE
            RETURN b - ((CARDINAL (- a - 1) MOD b) + 1)
        END (* if *);
    END TrueMOD;

PROCEDURE TrueDIV (a : INTEGER ; b : CARDINAL) : INTEGER ;
    BEGIN
        IF a >= 0 THEN
            RETURN CARDINAL (a) DIV b
        ELSE
            RETURN - INTEGER ((b + CARDINAL (-a) - 1) DIV b)
        END (* if *);
    END TrueDIV;

PROCEDURE Even (a: INTEGER) : BOOLEAN ;
    BEGIN
        RETURN NOT ODD (a) ;
    END Even ;

BEGIN
    TwoPow [0] := 1 ;
    FOR Index := 1 TO 31 DO
        TwoPow [Index] := 2 * TwoPow [Index - 1]
    END (* for *);
    
    ThreePow [0] := 1 ;
    FOR Index := 1 TO 20 DO
        ThreePow [Index] := 3 * ThreePow [Index - 1]
    END (* for *);
    
    realRelativePrecision := 1.0 ;
    WHILE (1.0 - realRelativePrecision) # 1.0 DO
        realRelativePrecision := realRelativePrecision / 2.0 ;
    END (* while *);
    realRelativePrecision := realRelativePrecision * 2.0 ;
    WHILE (1.0 + realRelativePrecision) = 1.0 DO
        realRelativePrecision := realRelativePrecision * 2.0 ;
    END (* while *);
    justLessThanOne := 1.0 - realRelativePrecision ;
    justGreaterThanOne := 1.0 + realRelativePrecision ;
END Numerical.
