IMPLEMENTATION MODULE Palette;

(*
    Title: 	Standard Colours for GKS
    LastEdit:	"Fri Feb 13 12:33:55 1987" by Tony Mann
    Author: 	jd / jacs
		Cambridge University Computer Laboratory
*)

(*  Updated 30/4/86 to provide 16 colour interface.
*)

(* This module provides a standard colour palette of the more generally
   useful colours. It is not part of GKS, and is an addition to it purely
   for convenience.
*)

VAR Map: ARRAY COLOUR OF INTENSITY;

PROCEDURE Intensity (c: COLOUR; VAR i: INTENSITY);
BEGIN
    i := Map [c];
END Intensity;

PROCEDURE Distance (i1, i2: INTENSITY): REAL;
BEGIN
    RETURN ABS (i1.R-i2.R) + ABS (i1.G-i2.G) + ABS (i1.B-i2.B);
END Distance;

PROCEDURE Colour (i: INTENSITY; p: PALETTE): COLOUR;
    VAR d, distance: REAL;
    VAR c, colour: COLOUR;
BEGIN
    colour := Black;
    distance := Distance (Map[Black], i);
    FOR c := Red TO LightGrey DO
        IF c IN p THEN
            d := Distance (Map[c], i);
            IF d < distance THEN 
                distance := d;
                colour := c
            END;
        END;
    END;
    RETURN colour;
END Colour;

VAR c: COLOUR;
VAR i: INTENSITY;

BEGIN
    WITH i DO
        R := 0.0; G := 0.0; B := 0.0; 
        FOR c := Black TO White DO
            Map [c] := i;
            IF R = 1.0 THEN
                R := 0.0;
                IF G = 1.0 THEN G := 0.0; B := 1.0; ELSE G := 1.0 END;
            ELSE R := 1.0;
            END;
        END;
    END;
    i.R := 1.0; i.G := 0.5; i.B :=0.0; Map [Orange] := i;
    i.R := 0.5; i.G := 1.0; i.B :=0.0; Map [GreenYellow] := i;
    i.R := 0.0; i.G := 1.0; i.B :=0.5; Map [GreenCyan] := i;
    i.R := 0.0; i.G := 0.5; i.B :=1.0; Map [BlueCyan] := i;
    i.R := 0.5; i.G := 0.0; i.B :=1.00; Map [BlueMagenta] := i;
    i.R := 1.0; i.G := 0.0; i.B :=0.5; Map [RedMagenta] := i;
    i.R := 0.5; i.G := 0.5; i.B :=0.5; Map [DarkGrey] := i;
    i.R := 0.75; i.G := 0.75; i.B :=0.75; Map [LightGrey] := i;
END Palette.
