IMPLEMENTATION MODULE RandomExtras;

FROM Arithmetic IMPORT SquareRoot;

FROM CLibrary IMPORT Ln;

FROM Numerical IMPORT Frac;

FROM Random IMPORT RandomCard;

FROM SYSTEM IMPORT MAXCARD;

VAR (* Useful values *)
    RealMaxCard, RealMaxCardDiv4 : REAL;
    (* Globals for Random01 *)
    needNewNormals : BOOLEAN;
    x2 : REAL;
    (* Globals for Uniform01 *)
    x, y, z : INTEGER;

PROCEDURE Random01 () : REAL ;
    BEGIN
        RETURN FLOAT (RandomCard () DIV 4) / RealMaxCardDiv4
    END Random01;

PROCEDURE Uniform01 () : REAL;
(*
   From Applied Statistics, v.31, no.2, 1982, page 188.
   Algorithm AS 183
   By Wichmann B.A. and Hill I.D.
   This version requires integer arithmetic up to 5212632,
   an alternative version is available for 16-bit arithmetic.
*)
    BEGIN
        x := (171 * x) MOD 30269;
        y := (172 * y) MOD 30307;
        z := (170 * z) MOD 30323;
        
        RETURN Frac (FLOAT (x) / 30269.0
                   + FLOAT (y) / 30307.0
                   + FLOAT (z) / 30323.0)
    END Uniform01;

PROCEDURE Normal1 () : REAL ;
(* Yield a variable normally distributed with mean zero, s.d. one *)
(* Uses the Polar Method described in Knuth, 
   Seminumerical algorithms, page 104 *)
VAR v1, v2, s, nasty, x1 : REAL ;
(* Note that x2 is a global *)
    BEGIN
        IF needNewNormals THEN
	    REPEAT
		v1 := 1.0 - 2.0 * Uniform01 ();
		v2 := 1.0 - 2.0 * Uniform01 ();
		s := v1 * v1 + v2 * v2;
	    UNTIL s < 1.0;
            nasty := SquareRoot (- 2.0 * Ln (s) / s);
            x1 := v1 * nasty;
            x2 := v2 * nasty;
            needNewNormals := FALSE;
            RETURN x1
        ELSE
            needNewNormals := TRUE;
            RETURN x2
        END (* if *);
    END Normal1;

BEGIN
    (* The following are necessary because FLOAT (MAXCARD) doesn't work *)
    RealMaxCardDiv4 := FLOAT (MAXCARD DIV 4);
    RealMaxCard := RealMaxCardDiv4 * 4.0 + FLOAT (MAXCARD MOD 4);
    needNewNormals := TRUE;
    x := 10;
    y := 200;
    z := 3000;
END RandomExtras.
