IMPLEMENTATION MODULE GKSDev;

(*
    Title: 	generic GKS driver for raster devices
    LastEdit:	"Wed Sep 18 11:16:25 1985"
    Author: 	pr
		Cambridge University Computer Laboratory
*)


IMPORT Arithmetic;
FROM Arithmetic IMPORT Round, IMax, RMax, RMin, SquareRoot;
FROM CharCodes IMPORT CReturnCh, NewLineCh, SpaceCh;
FROM GKSBuf IMPORT Wrch, Flush;
IMPORT GKSTypes;
FROM GKSTypes IMPORT
    SEGMENT, PRIMITIVE, TRANSFORM,
    PATH, PRECISION, PIXELS, ESCAPECODE, GDPCODE,
    MARKERSTYLE, LINESTYLE,
    GKSError, LineStyle, MarkerStyle;
IMPORT MathLib;
FROM Palette IMPORT INTENSITY;
IMPORT RasterIO;
FROM RasterIO IMPORT 
    Width, Height, PixelsWide, PixelsHigh, SelectColour, PlotBox, PrintPage;
IMPORT RasterOp;
IMPORT RasterCh;
FROM SYSTEM IMPORT WORD, MAXINT;
IMPORT Vector;
FROM Vector IMPORT 
    RVEC, AXIS, IVEC, RVec, REqual, RAdd, RBOX, BOUND, RBox, RInBox;


(*****  General                                                       *****)

PROCEDURE OpenDev (device: ARRAY OF CHAR;
        VAR DisplaySurface, PixelSurface: RBOX;
        VAR Defer, ImplicitRedraw: BOOLEAN);
BEGIN
    RBox (DisplaySurface, 0.0, 0.0, Width, Height);
    RBox (PixelSurface, 0.0, 0.0, FLOAT (PixelsWide), FLOAT (PixelsHigh));
    Defer := FALSE;
    ImplicitRedraw := TRUE;
END OpenDev;


PROCEDURE CloseDev; 
BEGIN
    PrintPage;
END CloseDev;


PROCEDURE ActivateDev;
BEGIN
END ActivateDev;


PROCEDURE DeactivateDev; 
BEGIN
END DeactivateDev;


PROCEDURE ClearDev (obligatory: BOOLEAN);
BEGIN
    PrintPage;
END ClearDev;


PROCEDURE UpdateDev;
BEGIN
END UpdateDev;


(*  a few modules to hide some state variables				  *)

MODULE rasterops;

FROM GKSTypes IMPORT LINESTYLE, LineStyle, MarkerStyle;
FROM RasterIO IMPORT NominalLineWidth;
FROM RasterOp IMPORT
    LINETYPE, CHANGEINDEX, 
    LineWidth, LineType, NoOfChanges, Cycle, CycleLength, 
    CurrentMarkerStyle, MarkerSize,
    LineSegment, DrawCyclicLine, DrawMarker;
FROM Vector IMPORT RVEC;

EXPORT SetDevLine, SetDevMarker, DevLine, DevMarker;

PROCEDURE SetDevLine (style: CARDINAL; width: REAL);
BEGIN
    LineWidth := NominalLineWidth * width;
    CASE LineStyle (style) OF
      UnspecifiedLine, SolidLine : LineType := Solid;
    | DottedLine : Cycle [0] := 0.1;
                   Cycle [1] := 0.9;
                   NoOfChanges := 2;
                   CycleLength := LineWidth * 10.0;
                   LineType := Cyclic 
    | DashedLine : Cycle [0] := 0.25;
                   Cycle [1] := 0.75;
                   NoOfChanges := 2;
                   CycleLength := LineWidth * 20.0;
                   LineType := Cyclic
    | DashedDottedLine : Cycle [0] := 0.3;
                         Cycle [1] := 0.45;
                         Cycle [2] := 0.55;
                         Cycle [3] := 0.7;
                         NoOfChanges := 4;
                         CycleLength := 40.0 * LineWidth;
                         LineType := Cyclic
    END (* case *)                      
END SetDevLine;

PROCEDURE SetDevMarker (style: CARDINAL; size: REAL);
BEGIN
    CurrentMarkerStyle := MarkerStyle (style);
    MarkerSize := size;
END SetDevMarker;

PROCEDURE DevLine (high: INTEGER; VAR at: ARRAY OF RVEC);
VAR i : INTEGER;
BEGIN
    IF LineWidth > 0.01
    THEN
         IF LineType = Solid
         THEN
              FOR i := 0 TO high-1
              DO LineSegment (at [i], at [i+1], LineWidth)
              END (* for *)
         ELSE    
              DrawCyclicLine (high, at)
         END (* if *)
    END (* if *)    
END DevLine;

PROCEDURE DevMarker (high: INTEGER; VAR at: ARRAY OF RVEC);
    VAR index : INTEGER;
BEGIN
    FOR index := 0 TO high
    DO DrawMarker (at [index])
    END (* for *)
END DevMarker;

BEGIN
    LineType := Solid;

END rasterops;


MODULE writing;

FROM Arithmetic IMPORT Round;
FROM GKSTypes IMPORT PRECISION;
FROM MathLib IMPORT Sin, Cos;
FROM RasterIO IMPORT PixelsWide;
FROM RasterCh IMPORT SelectFount, SelectStyle, PlotCh;
FROM Vector IMPORT AXIS, RVEC, RVec, RAdd, RBOX, RInBox;

EXPORT SetDevText, SetDevChar, DevText;

CONST TabPositions = 10.0;

VAR LineSpace: RVEC;

PROCEDURE SetDevText (fount: CARDINAL; precision: PRECISION): REAL;
BEGIN
    RETURN SelectFount (fount);
END SetDevText;

PROCEDURE SetDevChar (slant, rotation, height, width: REAL;
        VAR space: RVEC; abut: BOOLEAN);
BEGIN
    SelectStyle (slant, rotation, height, width, space, abut);
    RVec (LineSpace, height * Sin (rotation), - height * Cos (rotation));
END SetDevChar;

PROCEDURE DevText (VAR at: RVEC; VAR in: RBOX; VAR string: ARRAY OF CHAR);
VAR
    position:   RVEC;
    forcedmove: BOOLEAN;
    s:          CARDINAL;
BEGIN
    position := at;
    forcedmove := TRUE;
    s := 0;
    LOOP
        IF (s > HIGH (string)) OR (string [s] = 0C) THEN
            EXIT;
        END;
        IF forcedmove THEN
            IF NOT RInBox (position, in) THEN
                EXIT;
            END (* if *);
            forcedmove := FALSE;
        END (* if *);
        IF string [s] = '\' THEN
            INC (s);
            IF (s > HIGH (string)) OR (string [s] = 0C) THEN
                EXIT;
            END;
            CASE string [s] OF
              'c', 'C':
                position := at;
                forcedmove := TRUE;
            | 'n', 'N':
                position := at;
                RAdd (LineSpace, position);
                at := position;
                forcedmove := TRUE;
            | 's', 'S':
                PlotCh (' ', position);
            | 't', 'T':
                position [X] := FLOAT (
                    (Round (TabPositions * position [X]) DIV PixelsWide + 1)
                        * PixelsWide) / TabPositions;
                forcedmove := TRUE;
            ELSE
                PlotCh (string [s], position);
            END (* case *);
        ELSE
            PlotCh (string [s], position);
        END (* if *);
        INC (s);
    END (* loop *);
END DevText;

END writing;


(*****  Selection of Current Primitive Attributes		      *****)

PROCEDURE SetDevArea (style, pattern: CARDINAL);
BEGIN END SetDevArea;


PROCEDURE SetDevPattern (pixels: PIXELS);
BEGIN END SetDevPattern;


PROCEDURE SetDevColour (red, green, blue: REAL);
VAR
    intensity:  INTENSITY;
BEGIN
    WITH intensity DO
        R := red;
        G := green;
        B := blue;
    END (* with *);
    SelectColour (intensity);
END SetDevColour;


PROCEDURE SetDevPatternRef (VAR at, up, base: RVEC);
BEGIN END SetDevPatternRef;


(*****  Output Functions 					       *****)

PROCEDURE DevFill (high: INTEGER; VAR boundary: ARRAY OF RVEC);

CONST epsilon = 0.001;

    PROCEDURE IsBox (): BOOLEAN;
        VAR
            b: INTEGER;
        BEGIN
            IF (high = 4) 
            AND REqual (boundary [0], boundary [4], epsilon) THEN
                FOR b := 0 TO 3 DO
                    IF ABS ((boundary [b+1, X] - boundary [b, X]) *
                        (boundary [b+1, Y] - boundary [b, Y])) > epsilon THEN
                            RETURN FALSE;
                    END (* if *);
                END (* for *);
                RETURN TRUE;
            END (* if *);
            RETURN FALSE;
        END IsBox;

VAR r: RBOX;
BEGIN
    IF IsBox () THEN
        r [Low, X] := boundary [0, X];
        r [Low, Y] := boundary [0, Y];
        r [High, X] := boundary [2, X];
        r [High, Y] := boundary [2, Y];
        PlotBox (r);
    ELSE
        DevLine (high, boundary);
    END (* if *);
END DevFill;


PROCEDURE DevPixels (VAR boundary: ARRAY OF RVEC; VAR ClipTo: RBOX;
    rowhigh, columnhigh: INTEGER; VAR pixels: ARRAY OF CARDINAL);
BEGIN
END DevPixels;


PROCEDURE DevGdp (fn: GDPCODE; arg: WORD;
    high: INTEGER; VAR points: ARRAY OF RVEC; ClipTo: RBOX);
END DevGdp;


PROCEDURE DevEscape (fn: ESCAPECODE; VAR w: WORD; VAR a: ARRAY OF WORD);
BEGIN
    w := WORD (0);
END DevEscape;


PROCEDURE DevMessage (high: INTEGER; VAR s: ARRAY OF CHAR);
VAR
    c: CARDINAL;
BEGIN
    c := 0;
    LOOP
        IF (c > HIGH (s)) OR (s [c] = 0C) THEN
            EXIT
        END (* if *);
        IF s [c] = '\' THEN
            INC (c);
            IF (c > HIGH (s)) OR (s [c] = 0C) THEN
                EXIT
            END (* if *);
            CASE s [c] OF
              'c', 'C':
                Wrch (CReturnCh);
            | 'n', 'N', 'p', 'P':
                Wrch (CReturnCh);
                Wrch (NewLineCh);
            | 's', 'S', 't', 'T':
                Wrch (SpaceCh);
            ELSE
                Wrch (s [c]);
            END (* case *);
        ELSE
            Wrch (s [c]);
        END (* if *);
        INC (c);
    END (* loop *);
    Flush (TRUE);
END DevMessage;


(*****  Segment Functions 				              *****)

(*  this bit is rather easy to do					  *)

PROCEDURE ReDrawDev (): BOOLEAN;
BEGIN
    RETURN FALSE;
END ReDrawDev;


PROCEDURE SegAdd (s: SEGMENT): BOOLEAN;
BEGIN
    RETURN FALSE;
END SegAdd;


PROCEDURE SegDelete (s: SEGMENT): BOOLEAN;
BEGIN
    RETURN FALSE;
END SegDelete;


PROCEDURE SegPrimChange (s: SEGMENT; p: PRIMITIVE; new: BOOLEAN): BOOLEAN;
BEGIN
    RETURN FALSE;
END SegPrimChange;


PROCEDURE SegTransform (s: SEGMENT; T: TRANSFORM): BOOLEAN;
BEGIN
    RETURN FALSE;
END SegTransform;


PROCEDURE SegVisible (s: SEGMENT; on: BOOLEAN): BOOLEAN;
BEGIN
    RETURN FALSE;
END SegVisible;


PROCEDURE SegLit (s: SEGMENT; on: BOOLEAN): BOOLEAN;
BEGIN
    RETURN FALSE;
END SegLit;


PROCEDURE SegPriority (s: SEGMENT; priority: REAL): BOOLEAN;
BEGIN
    RETURN FALSE;
END SegPriority;

(*  or not as the case may be						  *)


END GKSDev.
