IMPLEMENTATION MODULE RasterCh;

(*
    Title: 	raster fount handling
    LastEdit:	"Wed Sep 18 11:25:02 1985"
    Author: 	pr
		Cambridge University Computer Laboratory
*)


FROM Arithmetic IMPORT SquareRoot, Round;

FROM FileStream IMPORT CreateInput;

FROM GKSFount IMPORT STYLE, PixelsPerInch, PointsPerInch,
                     FountNumber, FountName, FileName;

FROM MathLib IMPORT Sin, Cos, Radians;

FROM RasterIO IMPORT PixelsWide, PixelsHigh, PlotBox;

FROM RasterOp IMPORT SortByXThenY, SortByYThenX;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;

FROM Streams IMPORT Stream, Get, Put, Seek, SeekMode, Tell, Delete;

FROM Strings IMPORT CopyCC;

FROM SYSTEM IMPORT WORD, BITSPERWORD;

FROM SystemTypes IMPORT LongCARDINAL;

FROM Vector IMPORT AXIS, RVEC, RVec, RAdd, BOUND, RBOX, RSetLength, RLength;


CONST MaxLeft  = 30;
      MaxRight = 80;
      MaxUp    = 100;
      MaxDown  = 60;
      MaxCharacterNumber = 255; (* The characters in VFONT run from 0 to 255 *)
      MaxFontNumber = 300;
      MaxFontCount = 10; (* Number of simultaneously loaded fonts *)
      
      MaxFileNameLength = 40;
      
      ShortSize = 2;
      CharSize  = 1;


TYPE
     BYTE = [0..255];
     BYTEELEM = [0..7];
     BYTESET = SET OF BYTEELEM; 
            
     FILENAME = ARRAY [1..MaxFileNameLength] OF CHAR;

     CHARBITS = ARRAY 
                    [0..((MaxLeft+MaxRight-1)*(MaxDown+MaxUp))DIV BITSPERWORD]
                    OF BITSET;
     
     CHARDESC = RECORD
                   Width: REAL; (* Offset to the next character position *)
                   
                   Left : INTEGER; (* The valid bounds of the Bitmap,     *)
                   Right: INTEGER; (* with respect to the [0,0] character *)
                   Up   : INTEGER; (* reference point                     *)
                   Down : INTEGER;
                   
                   BitMap: CHARBITS
                END;
                
           FONT = RECORD
               BitsHeight: REAL; (* The height of the font for scaling *)
               BitsWidth : REAL; (* The width of the font for scaling *)
               
               Characters: ARRAY [0..MaxCharacterNumber] OF CHARDESC 
            END;

     FONTPOINTER = POINTER TO FONT;
     
            
VAR
    DummyReal: REAL;
    i : INTEGER;
    Fonts: ARRAY [1..MaxFontNumber] OF FONTPOINTER;
    FontCount: INTEGER;
    CurrentFont: FONTPOINTER;
    CharSlant, CharRotation, CharHeight, CharWidth: REAL;
    CharAbut: BOOLEAN;
    CharSpace: RVEC;
    (* transformation coefficients between Vfont and Canon bitmaps: *)
    VxCx, VxCy, VyCx, VyCy, CxVx, CxVy, CyVx, CyVy: REAL;
    CanonMoveRight, CanonMoveDown, VfontMoveRight,
    VfontRightVec, VfontDownVec, CanonRightVec: RVEC;
    RowLength: INTEGER;
    

PROCEDURE ReadUnsignedShort (in: Stream): INTEGER;
    VAR i: INTEGER;
    BEGIN
        i:= INTEGER (Get (in));
        i:= i + 256 * INTEGER (Get (in));
        RETURN i
    END ReadUnsignedShort;
    
PROCEDURE ReadShort (in: Stream): INTEGER;
    VAR i: INTEGER;
    BEGIN
        i:= ReadUnsignedShort (in);
        RETURN i (* I am not reading any -ve no's yet *)
    END ReadShort;    
    
PROCEDURE BitSet (VAR CharDesc: CHARDESC; Row, Column: INTEGER;
                  Value: BOOLEAN);
    VAR Index: INTEGER;
    BEGIN
        WITH CharDesc DO
           Index:= (Row + MaxDown) * RowLength + (Column + MaxLeft);
           IF Value 
           THEN INCL (BitMap [Index DIV BITSPERWORD], Index MOD BITSPERWORD)
           ELSE EXCL (BitMap [Index DIV BITSPERWORD], Index MOD BITSPERWORD)
           END (* if *)
        END (* with *)   
    END BitSet;

PROCEDURE BitRead (VAR CharDesc: CHARDESC; Row, Column: INTEGER): BOOLEAN;
    VAR Index: CARDINAL;
    BEGIN
        WITH CharDesc DO
           IF (Row > Up) OR (Row < -Down)
              OR (Column < -Left) OR (Column > Right)
           THEN
                RETURN FALSE
           ELSE     
                Index:= (Row+MaxDown) * RowLength + (Column + MaxLeft);
                RETURN (Index MOD BITSPERWORD)
                          IN (BitMap [Index DIV BITSPERWORD])
           END (* if *)       
        END (* with *)   
    END BitRead;

PROCEDURE ReadVfontChar (in: Stream): INTEGER;
    VAR i: INTEGER;
    BEGIN
        i:= INTEGER (Get (in));
        IF i > 128 (* Who knows where they draw the line ! *)
        THEN RETURN i - 256
        ELSE RETURN i
        END (* if *)
    END ReadVfontChar;


PROCEDURE SetTransformCoefficients;
    VAR determinant: REAL;
    BEGIN
        WITH CurrentFont^ DO 
            VxCx:=   (CharWidth / BitsWidth) * Cos (CharRotation);
            VxCy:= - (CharWidth / BitsWidth) * Sin (CharRotation);
            VyCx:=   (CharHeight / BitsHeight) 
                      * Cos (CharRotation + CharSlant);
            VyCy:= - (CharHeight / BitsHeight)
                      * Sin (CharRotation + CharSlant)
        END (* with *);    
           
        determinant:= VxCx * VyCy - VxCy * VyCx;
        
        CxVx:=   VyCy / determinant;
        CxVy:= - VxCy / determinant;
        CyVx:= - VyCx / determinant;
        CyVy:=   VxCx / determinant;

        RVec (CanonMoveRight, 1.0, 0.0);
        RVec (CanonMoveDown,  0.0, 1.0);
        VfontMoveRight:= CanonMoveRight;
        
        TransformToVfont (CanonMoveRight, VfontRightVec);
        TransformToVfont (CanonMoveDown, VfontDownVec);
        TransformToCanon (VfontMoveRight, CanonRightVec) 
        
END SetTransformCoefficients;

PROCEDURE TransformToVfont (canon: RVEC; VAR vfont: RVEC);
VAR cx, cy: REAL;  
BEGIN
    cx:= canon [X];
    cy:= canon [Y];
    vfont [X]:= CxVx * cx + CyVx * cy;
    vfont [Y]:= CxVy * cx + CyVy * cy
END TransformToVfont;

PROCEDURE TransformToCanon (vfont: RVEC; VAR canon: RVEC);
VAR vx, vy: REAL;
BEGIN
    vx:= vfont [X];
    vy:= vfont [Y];
    canon [X]:= VxCx * vx + VyCx * vy;
    canon [Y]:= VxCy * vx + VyCy * vy
END TransformToCanon;


PROCEDURE TransformArrayToCanon (high: INTEGER; Vectors: ARRAY OF RVEC);
VAR i: INTEGER;
BEGIN
    FOR i:= 0 TO high
    DO TransformToCanon (Vectors [i], Vectors [i])
    END
END TransformArrayToCanon;
   

PROCEDURE RemoveAllFonts;
VAR FontNumber: INTEGER;
    BEGIN
        FOR FontNumber:= 1 TO MaxFontNumber DO
            IF Fonts [FontNumber] <> NIL THEN
                DISPOSE (Fonts [FontNumber]);
                Fonts [FontNumber]:= NIL
            END (* if *);
        END (* for *);
    END RemoveAllFonts;
    

PROCEDURE ReadFont (FontNumber: INTEGER);
VAR fontin               : Stream;
    FontFile             : FILENAME;
    CharNumber           : INTEGER;
    NextDispatch         : LongCARDINAL;
    VfontHeaderSize      : INTEGER;
    VfontDispatchSize    : INTEGER;
    MapsBegin            : INTEGER;
    DummyInt             : INTEGER;
    MapOffset            : INTEGER;
    BitMapSize           : INTEGER;
    Maxx, Maxy           : INTEGER;
    NBytes               : INTEGER;
    Row, Column          : INTEGER;
    CurrentBit           : INTEGER;
    CurrentByte          : BYTESET;
    FontPointer          : FONTPOINTER;    
    TooBig               : BOOLEAN;

BEGIN
    VfontHeaderSize:= 5 * ShortSize;
    VfontDispatchSize:= 3 * ShortSize + 4 * CharSize;
    MapsBegin:= VfontHeaderSize + 256 * VfontDispatchSize;

    TooBig:= FALSE;

    IF (FontNumber < 0) OR (FontNumber > MaxFontNumber)
    THEN
         FontNumber:= 1
    END (* if *);
    
    IF Fonts [FontNumber] = NIL
    THEN INC (FontCount);
         IF FontCount > MaxFontCount THEN
            RemoveAllFonts;
            FontCount:= 1
         END (* if *);
         NEW (Fonts [FontNumber]);
         WITH Fonts [FontNumber]^
         DO 
            IF NOT FileName (FontNumber, 12, FontFile)
            THEN CopyCC ("/usr/lib/vfont/R.12", FontFile)
            END (* if *);
            
            fontin := CreateInput (FontFile);
            (* Read vfont header *)
                
            DummyInt:= ReadShort (fontin);
            
            IF DummyInt # 286 (* magic number 0436 octal *) THEN
            END (* if *);
            
            BitMapSize:= ReadUnsignedShort (fontin);
            Maxx:= ReadShort (fontin);
            Maxy:= ReadShort (fontin);
            
            IF (Maxx <= 0) OR (Maxy <= 0)
            THEN
               Maxx:= TRUNC (SquareRoot (FLOAT (BitMapSize) / 9.0)) + 1;
               Maxy:= Maxx
            END (* if *);
            
            BitsWidth:= FLOAT (Maxx) / 2.0;
            BitsHeight:= FLOAT (Maxy) / 2.0;
            
            DummyInt:= ReadShort (fontin);

                (* Read the character descriptions *)
                
            FOR CharNumber:= 0 TO MaxCharacterNumber
            DO WITH Characters [CharNumber]
               DO 
                  (* Read Vfont Dispatch *)
                
                  MapOffset:= ReadUnsignedShort (fontin);
                  NBytes:= ReadShort (fontin);
                  Up:= ReadVfontChar (fontin);
                  Down:= ReadVfontChar (fontin);
                  Left:= ReadVfontChar (fontin);
                  Right:= ReadVfontChar (fontin);
                  Width:= FLOAT (ReadShort (fontin));

                  IF NBytes > 0 THEN (* Read Bit Map *)
                     Tell (fontin, NextDispatch);
                     Seek (fontin, Beginning, MapsBegin + MapOffset);
                     FOR Row:= Up-1 TO -Down BY -1
                     DO (*
                        Read a row of bytes into BitMap [row]
                        *)
                        CurrentBit:= -1;
                        FOR Column:= -Left TO Right-1
                        DO IF CurrentBit < 0
                           THEN CurrentBit:= 7;
                                CurrentByte:= BYTESET (Get (fontin))
                           END (* if *);     
                           
                           IF     (Row < MaxUp-1) 
                              AND (Row > -MaxDown) 
                              AND (Column > -MaxLeft)  
                              AND (Column < MaxRight-1)
                           THEN
                              BitSet (Characters [CharNumber],
                                      Row,
                                      Column,
                                      BYTEELEM (CurrentBit) IN CurrentByte
                                     ) 
                           ELSE TooBig:= TRUE
                           END;
                           DEC (CurrentBit) 
                        END (* for *)  
                     END (* for *);
                     Seek (fontin, Beginning, NextDispatch)
                  END (* if *)                   
               END (* with *)
            END; (* for *)
            Delete (fontin); (* Throw away Stream *)
            DISPOSE (fontin);
        END; (* with *)
    END (* if *)
END ReadFont;


PROCEDURE SelectFount (FontNumber: INTEGER): REAL;
    BEGIN
        IF (FontNumber < 0) OR (FontNumber > MaxFontNumber)
        THEN
            FontNumber:= 1
        END (* if *);    

        IF Fonts [FontNumber] = NIL
        THEN
            ReadFont (FontNumber)
        END (* if *);
        
        CurrentFont:= Fonts [FontNumber];
        
        SetTransformCoefficients;
        
        RETURN CurrentFont^.BitsWidth / CurrentFont^.BitsHeight
        
    END SelectFount;
    

PROCEDURE SelectStyle (slant, rotation, height, width: REAL;
                        space: RVEC; abut: BOOLEAN);
    BEGIN
        CharSlant:= Radians (slant);
        CharRotation:= Radians (rotation);
        CharHeight:= height;  (* used to /2 to look better *)
        CharWidth:= width;
        CharSpace:= space;
        RSetLength (CharSpace, RLength (space) / 3.0);
        CharSpace [Y]:= - CharSpace [Y];
        CharAbut:= abut;
        SetTransformCoefficients
    END SelectStyle;


PROCEDURE PlotCh (c: CHAR; VAR at: RVEC);
VAR 
    x, y: INTEGER;
    VfontChar: CHARDESC;
    Ytop, Ybottom, Xleft, Xright: INTEGER;
    UpX, UpY, DownX, DownY, LeftX, LeftY, RightX, RightY: REAL;
    corners: ARRAY [0..3] OF RVEC;
    CanonTopLeft, CanonBottomRight,
    VfontYRef, VfontVec, CharMove: RVEC;
    CanonTLX, CanonTLY, CanonBRX, CanonBRY: INTEGER;
    splat: RBOX;
    running: BOOLEAN;
BEGIN      
    IF c = ' ' THEN
        RAdd (CharSpace, at);
        RETURN;
    END (* if *);
    VfontChar:= CurrentFont^.Characters [INTEGER (c)];    
    WITH VfontChar DO
        UpX:= FLOAT (Up) * VyCx;
        UpY:= FLOAT (Up) * VyCy;
        DownX:= - FLOAT (Down) * VyCx;
        DownY:= - FLOAT (Down) * VyCy;
        
        LeftX:= - FLOAT (Left) * VxCx;
        LeftY:= - FLOAT (Left) * VxCy;
        RightX:= FLOAT (Right) * VxCx;
        RightY:= FLOAT (Right) * VxCy;
        
        corners [0] [X]:= UpX + LeftX;
        corners [0] [Y]:= UpY + LeftY;
        corners [1] [X]:= UpX + RightX;
        corners [1] [Y]:= UpY + RightY;
        corners [2] [X]:= DownX + LeftX;
        corners [2] [Y]:= DownY + LeftY;
        corners [3] [X]:= DownX + RightX;
        corners [3] [Y]:= DownY + RightY;

        SortByYThenX (3, corners);  (* Scope for improvement here ! *)
        Ytop:= Round (corners [0] [Y]);
        Ybottom:= Round (corners [3] [Y]);
        
        SortByXThenY (3, corners);
        Xleft:= Round (corners [0] [X]);
        Xright:= Round (corners [3] [X]);

        RVec (CanonTopLeft, FLOAT (Xleft), FLOAT (Ytop));
        RVec (CanonBottomRight, FLOAT (Xright), FLOAT (Ybottom));
        
        CanonTLX:= Round (CanonTopLeft [X] + at [X]);
        CanonTLY:= Round (CanonTopLeft [Y] + at [Y]);
        CanonBRX:= Round (CanonBottomRight [X] + at [X]);
        CanonBRY:= Round (CanonBottomRight [Y] + at [Y]);

        IF (CanonTLX >= 0) AND (CanonTLY >= 0)
            AND (CanonBRX <= PixelsWide) AND (CanonBRY <= PixelsHigh)
        THEN   
             TransformToVfont (CanonTopLeft, VfontYRef);
             VfontVec:= VfontYRef;
             
             FOR y := Ytop TO Ybottom DO
                splat [Low, Y] := at [Y] - FLOAT (y);
                splat [High, Y] := splat [Low, Y];
                running := FALSE;
                FOR x := Xleft TO Xright DO
                    IF BitRead (VfontChar, 
                        Round (VfontVec [Y]), Round (VfontVec [X])) # running
                    THEN
                        IF running THEN
                            splat [High, X] := at [X] + FLOAT (x-1);
                            PlotBox (splat);
                            running := FALSE;
                        ELSE
                            splat [Low, X] := at [X] + FLOAT (x);
                            running := TRUE;
                        END (* if *);
                    END (* if *);
                    RAdd (VfontRightVec, VfontVec)                   
                END (* for *);
                IF running THEN
                    splat [High, X] := at [X] + FLOAT (Xright);
                    PlotBox (splat);
                END (* if *);
                RAdd (VfontDownVec, VfontYRef);
                VfontVec:= VfontYRef 
             END (* for *)
        END (* if *);
      
        IF CharAbut
        THEN
             CharMove:= VfontMoveRight;
             RSetLength (CharMove, Width);
             TransformToCanon (CharMove, CharMove);
             RAdd (CharMove, at)
        ELSE 
             RAdd (CharSpace, at)
        END (* if *)
        
    END (* with *)              
END PlotCh;


BEGIN
    RowLength:= (MaxLeft + MaxRight - 1);

    FOR i:= 1 TO MaxFontNumber
    DO Fonts [i]:= NIL
    END;
    FontCount:= 0;
    
    CharSlant:= Radians (90.0); (* so that SetTransformCoordinates *)
    CharRotation:= 0.0;         (* does not have to bother about   *)
    CharHeight:= 15.0;          (* initialisation problems         *)
    CharWidth:= 10.0;
    CharAbut:= TRUE;
    RVec (CharSpace, 10.0, 0.0)

END RasterCh.
