IMPLEMENTATION MODULE RasterIO;

(*
    Title: 	raster driver for Diablo C-150 ink jet printer
    LastEdit:	"Tue Sep 17 11:57:55 1985"
    Author: 	pr
		Cambridge University Computer Laboratory
*)


FROM Arithmetic IMPORT Round, IMin, IMax;
FROM GKSBuf IMPORT Wrch, Flush;
FROM Palette IMPORT INTENSITY, COLOUR, PALETTE, Rainbow, Colour;
FROM SYSTEM IMPORT WORD;
FROM Vector IMPORT AXIS, BOUND, ISEG, ISeg, RBOX;


CONST
    PixelsPerInch = 120.0;
    MetresPerInch = 0.00254;

    PixelsPerChar = 2;
    PageWidth = 899;			(*  7 1/2"  *)
    BufferWidth = (PageWidth + 1) DIV PixelsPerChar - 1;
    PageHeight = 1199;			(*  by 10"  *)

    LineFeedCh = 12C;
    EscapeCh = 33C;
    SpaceCh = 40C;


VAR
    PixelsPerMetre: REAL;

    Buffer: ARRAY [0 .. PageHeight], [0 .. BufferWidth] OF CHAR;
    CurrentColour: INTEGER;
    VerticalLimit: ISEG;

    Control: ARRAY [0 .. 3] OF PALETTE;


PROCEDURE ClearBuffer;
    VAR
        x, y: INTEGER;
        z: CHAR;
    BEGIN
        z := CHAR (INTEGER (White) * 11H);
        FOR y := VerticalLimit [Low] TO VerticalLimit [High] DO
            FOR x := 0 TO BufferWidth DO
                Buffer [y, x] := z;
            END (* for *);
        END (* for *);
        VerticalLimit [Low] := PageHeight;
        VerticalLimit [High] := 0;
    END ClearBuffer;


PROCEDURE SelectColour (intensity: INTENSITY);
    BEGIN
        CurrentColour := INTEGER (Colour (intensity, Rainbow));
    END SelectColour;


PROCEDURE PlotBox (box: RBOX);
    VAR
        u, v: ISEG;
        xx, y, z: INTEGER;
    BEGIN
        IF (box [High, X] < 0.0) OR (box [High, Y] < 0.0) THEN RETURN END;
        ISeg (IMax (Round (box [Low, X]), 0), 
            IMin (Round (box [High, X]), PageWidth), u);
        ISeg (IMax (Round (box [Low, Y]), 0), 
            IMin (Round (box [High, Y]), PageHeight), v);
        IF v [Low] < VerticalLimit [Low] THEN 
            VerticalLimit [Low] := v [Low];
        END (* if *);
        IF v [High] > VerticalLimit [High] THEN 
            VerticalLimit [High] := v [High];
        END (* if *);
        FOR y := v [Low] TO v [High] DO
            IF ODD (u [Low]) THEN
                xx := u [Low] DIV 2;
                z := INTEGER (Buffer [y, xx]);
                Buffer [y, xx] := CHAR ((z DIV 16) * 16 + CurrentColour);
            END (* if *);
            FOR xx := (u [Low] + 1) DIV 2 TO (u [High] - 1) DIV 2 DO
                Buffer [y, xx] := CHAR (CurrentColour * 11H);
            END (* for *);
            IF NOT ODD (u [High]) THEN
                xx := u [High] DIV 2;
                z := INTEGER (Buffer [y, xx]);
                Buffer [y, xx] := CHAR (16 * CurrentColour + z MOD 16);
            END (* if *);
        END (* for *);
    END PlotBox;


PROCEDURE Diablo (byte: WORD);
    VAR c: CARDINAL;
    CONST msnibble = 7 * 16; lsnibble = 6 * 16;
    BEGIN
        c := CARDINAL (byte);
        Wrch (EscapeCh);
        (* avoid sending some control chars and large values in clear *)
        IF c <= 15 THEN
            Wrch (VAL (CHAR, lsnibble + c));
        ELSIF c <= 95 THEN
            Wrch (VAL (CHAR, c))
        ELSE Wrch (VAL (CHAR, msnibble + (c DIV 16)));
            Wrch (VAL (CHAR, lsnibble + (c MOD 16)));
        END;
        Flush (FALSE);
    END Diablo;


PROCEDURE PrintPage;
    VAR
        m, p, y, c, x, xx, z: INTEGER;
        colour: COLOUR;
        high, byte, bit: INTEGER;
        picoline: ARRAY [0 .. 128] OF BITSET;
    BEGIN
        IF VerticalLimit [Low] > VerticalLimit [High] THEN RETURN END;
        FOR byte := 1 TO 128 DO picoline [byte] := {} END;
        FOR m := VerticalLimit [High] TO VerticalLimit [Low] BY -4 DO
            FOR p := 0 TO 3 DO
                y := m - p;
                IF y >= 0 THEN
                    FOR c := 0 TO 3 DO
                        byte := 1;
                        bit := 7;
                        FOR xx := 0 TO PageWidth DIV 2 DO
                            z := INTEGER (Buffer [y, xx]);
                            colour := COLOUR (z DIV 16);
                            IF colour IN Control [c] THEN
                                INCL (picoline [byte], bit);
                            END (* if *);
                            DEC (bit);
                            colour := COLOUR (z MOD 16);
                            IF colour IN Control [c] THEN
                                INCL (picoline [byte], bit);
                            END (* if *);
                            IF bit > 0 THEN
                                DEC (bit);
                            ELSE
                                bit := 7;
                                INC (byte);
                            END (* if *);
                        END (* for *);
                        high := IMin (byte, 128);
                        WHILE (high > 0) AND (picoline [high] = {}) DO
                            DEC (high);
                        END (* while *);
                        IF high > 0 THEN
                            Diablo (EscapeCh);
                            Diablo ('g');
                            Diablo (INTEGER ('0') + 4 * c + p);
                            IF high >= 100 THEN Diablo ('1') END;
                            IF high >= 10 THEN 
                                Diablo (INTEGER ('0') + (high MOD 100) DIV 10)
                            END (* if *);
                            Diablo (INTEGER ('0') + high MOD 10);
                            Diablo (SpaceCh);
                            FOR byte := 1 TO high DO
                                Diablo (picoline [byte]);
                                picoline [byte] := {};
                            END (* for *);
                        END (* if *);
                    END (* for *);
                END (* if *);
            END (* for *);
            Diablo (EscapeCh);
            Diablo ('k');
            Diablo ('1');
        END (* for *);
        FOR p := 1 TO 6 DO
            Diablo (LineFeedCh);
        END (* for *);
        Flush (TRUE);
        ClearBuffer;
    END PrintPage;


BEGIN
    PixelsPerMetre := PixelsPerInch / MetresPerInch;

    PixelsWide := PageWidth;
    PixelsHigh := PageHeight;
    Width := FLOAT (PixelsWide + 1) / PixelsPerMetre;
    Height := FLOAT (PixelsHigh + 1) / PixelsPerMetre;

    NominalLineWidth := 1.0;

    Control [0] := PALETTE {Black};
    Control [1] := PALETTE {Red, Blue, Magenta};    (* Magenta ink *)
    Control [2] := PALETTE {Red, Green, Yellow};    (* Yellow ink *)
    Control [3] := PALETTE {Green, Blue, Cyan};     (* Cyan ink *)

    CurrentColour := INTEGER (Black);

    ISeg (0, PageHeight, VerticalLimit);
    ClearBuffer;

END RasterIO.
