IMPLEMENTATION MODULE RasterIO;

(*
    Title: 	raster driver for HP 2686A LaserJet Printer
    LastEdit:	"Thu Sep 19 09:41:54 1985"
    Author: 	pr /dm
		Cambridge University Computer Laboratory
*)


FROM Arithmetic IMPORT Round, IMin, IMax;
FROM CharCodes IMPORT CReturnCh, NewLineCh, EscapeCh, FormFeedCh;
FROM GKSBuf IMPORT Wrch, Flush;
FROM Palette IMPORT INTENSITY, COLOUR, PALETTE, Rainbow, Colour, Intensity;
FROM SYSTEM IMPORT WORD;
FROM Vector IMPORT AXIS, BOUND, ISEG, ISeg, RBOX;


CONST
    PixelsPerInch = 75.0;       (* Note this to be a REAL form of *)
    IntPixelsPerInch = 75;      (* this INTEGER ! *)
    MetresPerInch = 0.00254;

    PixelsPerChar = 2;
    PageWidth = 749;                       (* PixelsPerInch * 7.5 *)
    BufferWidth = ((PageWidth + 1) DIV PixelsPerChar) - 1;
    PageHeight = 749;                     (* PixelsPerInch * 7.5 *)
    PixelBytesPerLine = PageWidth DIV 8 + 1;


VAR
    PixelsPerMetre: REAL;

    Buffer: ARRAY [0 .. PageHeight], [0 .. BufferWidth] OF CHAR;
    CurrentColour: INTEGER;
    VerticalLimit: ISEG;


PROCEDURE ClearBuffer;
    VAR
        x, y: INTEGER;
        z: CHAR;
    BEGIN
        z := CHAR (INTEGER (White) * 11H);
        FOR y := VerticalLimit [Low] TO VerticalLimit [High] DO
            FOR x := 0 TO BufferWidth DO
                Buffer [y, x] := z;
            END (* for *);
        END (* for *);
        VerticalLimit [Low] := PageHeight;
        VerticalLimit [High] := 0;
    END ClearBuffer;


PROCEDURE SelectColour (intensity: INTENSITY);
    BEGIN
        CurrentColour := INTEGER (Colour (intensity, Rainbow));
    END SelectColour;


PROCEDURE PlotBox (box: RBOX);
    VAR
        u, v: ISEG;
        xx, y, z: INTEGER;
    BEGIN
        IF CurrentColour # INTEGER (Black) THEN RETURN; END;
        IF (box [High, X] < 0.0) OR (box [High, Y] < 0.0) THEN RETURN END;
        ISeg (IMax (Round (box [Low, X]), 0), 
            IMin (Round (box [High, X]), PageWidth), u);
        ISeg (IMax (Round (box [Low, Y]), 0), 
            IMin (Round (box [High, Y]), PageHeight), v);
        IF v [Low] < VerticalLimit [Low] THEN 
            VerticalLimit [Low] := v [Low];
        END (* if *);
        IF v [High] > VerticalLimit [High] THEN 
            VerticalLimit [High] := v [High];
        END (* if *);
        FOR y := v [Low] TO v [High] DO
            IF ODD (u [Low]) THEN
                xx := u [Low] DIV 2;
                z := INTEGER (Buffer [y, xx]);
                Buffer [y, xx] := CHAR ((z DIV 16) * 16 + CurrentColour);
            END (* if *);
            FOR xx := (u [Low] + 1) DIV 2 TO (u [High] - 1) DIV 2 DO
                Buffer [y, xx] := CHAR (CurrentColour * 11H);
            END (* for *);
            IF NOT ODD (u [High]) THEN
                xx := u [High] DIV 2;
                z := INTEGER (Buffer [y, xx]);
                Buffer [y, xx] := CHAR (16 * CurrentColour + z MOD 16);
            END (* if *);
        END (* for *);
    END PlotBox;


PROCEDURE Laser (byte: CHAR);
    BEGIN
        Wrch (byte);
        Flush (FALSE);
    END Laser;


PROCEDURE GrLaser(arg: BOOLEAN; value: INTEGER; start, end: CHAR);
    
    PROCEDURE Digit(value, size: INTEGER);
        BEGIN
            IF (value >= size) THEN
                Laser (CHAR (INTEGER ("0") + (value DIV size) MOD 10));
            END;
        END Digit;
        
    BEGIN
        Laser (EscapeCh);
        Laser ("*");
        Laser (start);
        IF arg THEN
            IF value > 1000 THEN HALT; END;
            IF value = 0 THEN Laser("0");
            ELSE Digit(value, 100);
                Digit(value, 10);
                Digit(value, 1);
            END;
        END;
        Laser (end);
    END GrLaser;


PROCEDURE SetAndStartRasta ();
    BEGIN
        Laser (CReturnCh);
        Laser (NewLineCh);
        (* Set IntPixelsPerInch on printer *)
        GrLaser (TRUE, IntPixelsPerInch, "t", "R");
        (* Start graphics mode *)
        GrLaser (TRUE, 0, "r", "A");
    END SetAndStartRasta;


PROCEDURE SetLine(bytes: INTEGER);
    BEGIN
        (* Set up for 'high' bytes of graphics *)
        GrLaser (TRUE, bytes, "b", "W");
    END SetLine;


PROCEDURE Pixel (VAR byte: BITSET; bit: INTEGER; colour: COLOUR);
    VAR i: INTENSITY;
    BEGIN
        Intensity (colour, i);
        IF (i.R * 0.3) + (i.G * 0.59) + (i.B * 0.11) <= 0.90 THEN
            INCL (byte, bit);
        END;
    END Pixel;


PROCEDURE EndAndFeed ();
    BEGIN
        GrLaser (FALSE, 0, "r", "B");
        Laser (FormFeedCh);
    END EndAndFeed;
    
    
PROCEDURE PrintPage;
    VAR m, p, y, c, x, xx, z: INTEGER;
        colour: COLOUR;
        high, byte, bit: INTEGER;
        picoline: ARRAY [1..PixelBytesPerLine] OF BITSET;
    BEGIN
        IF VerticalLimit [Low] > VerticalLimit [High] THEN RETURN END;
        SetAndStartRasta ();
        FOR byte := 1 TO PixelBytesPerLine DO picoline [byte] := {} END;
        FOR y := VerticalLimit [High] TO VerticalLimit [Low] BY -1 DO
                byte := 1; bit := 7;
                FOR xx := 0 TO PageWidth DIV 2 DO
                    z := INTEGER (Buffer [y, xx]);
                    Pixel (picoline [byte], bit, COLOUR (z DIV 16));
                    DEC (bit);
                    Pixel (picoline [byte], bit, COLOUR (z MOD 16));
                    IF bit > 0 THEN DEC (bit);
                    ELSE bit := 7; INC (byte);
                    END (* if *);
                END (* for *);
                high := IMin (byte, PixelBytesPerLine);
                WHILE (high > 0) AND (picoline [high] = {}) DO
                    DEC (high);
                END;
                SetLine (high);
                FOR byte := 1 TO high DO
                    Laser (CHAR (picoline [byte]));
                    picoline [byte] := {};
                END (* for *);
        END (* for y *);
        EndAndFeed ();
        Flush (TRUE);
        ClearBuffer;
    END PrintPage;


BEGIN
    PixelsPerMetre := PixelsPerInch / MetresPerInch;

    PixelsWide := PageWidth;
    PixelsHigh := PageHeight;
    Width := FLOAT (PixelsWide + 1) / PixelsPerMetre;
    Height := FLOAT (PixelsHigh + 1) / PixelsPerMetre;

    NominalLineWidth := 1.0;

    CurrentColour := INTEGER (Black);

    ISeg (0, PageHeight, VerticalLimit);
    ClearBuffer;

END RasterIO.
