IMPLEMENTATION MODULE RasterOp;

(*
    Title: 	plotting operations for RasterDev
    LastEdit:	"Tue Sep 17 12:06:47 1985"
    Author: 	pr (well ahj's CanonDev, really)
		Cambridge University Computer Laboratory
*)


FROM Arithmetic IMPORT SquareRoot;
FROM GKSTypes IMPORT MARKERSTYLE;
FROM RasterIO IMPORT PixelsWide, PixelsHigh, NominalLineWidth, PlotBox;
FROM Vector IMPORT AXIS, RVEC, RVec, RAdd, RSubtract, REqual, RBOX, RBox;


TYPE FOURRVECS = ARRAY [1 .. 4] OF RVEC;

CONST 
    epsilon = 0.001;
    ComparisonTolerance = 0.01;


VAR
    HorizontalResolution, VerticalResolution,
    UnitX, UnitY,
    UnitXSquared, UnitYSquared: REAL;

	
PROCEDURE FRAC (x : REAL) : REAL;
    BEGIN
        IF x >= 0.0
        THEN RETURN x - FLOAT (TRUNC (x))
        ELSE RETURN 1.0 + x + FLOAT (TRUNC (-x))
        END 
    END FRAC;

PROCEDURE FindPerpendicularVector (in : RVEC; VAR out : RVEC);
    BEGIN
        RVec (out, - (in [Y]), in [X] )
    END FindPerpendicularVector;

PROCEDURE CanonFindPerpendicularVector (in : RVEC; VAR out : RVEC);
    BEGIN
        RVec (out,
                - (in [Y] * HorizontalResolution / VerticalResolution),
                   in [X] * VerticalResolution / HorizontalResolution)
    END CanonFindPerpendicularVector;

PROCEDURE CanonSetRLength (VAR v : RVEC; len : REAL);
    VAR factor : REAL;
    BEGIN
        factor := len / SquareRoot ( v[X] * v[X] / UnitXSquared
                                   + v[Y] * v[Y] / UnitYSquared);
        v[X] := v[X] * factor;
        v[Y] := v[Y] * factor
    END CanonSetRLength;
    
PROCEDURE CanonRDistance (VAR v1, v2 : RVEC) : REAL;
VAR dx, dy : REAL;
    BEGIN
        dx := v1 [X] - v2 [X];
        dy := v1 [Y] - v2 [Y];
        RETURN SquareRoot (dx * dx / UnitXSquared + dy * dy / UnitYSquared)
    END CanonRDistance;
    
PROCEDURE SetFourVector (v1, v2, v3, v4 : RVEC; VAR FourVector :FOURRVECS);
    BEGIN
        FourVector [1] := v1;
        FourVector [2] := v2;
        FourVector [3] := v3;
        FourVector [4] := v4
    END SetFourVector;

PROCEDURE Frac (x : REAL) : REAL;
    BEGIN
        RETURN (x - FLOAT (TRUNC (x)))
    END Frac;


PROCEDURE SortByYThenX (high : INTEGER; VAR CoOrds : ARRAY OF RVEC);

    VAR top, current : INTEGER;

    PROCEDURE arrange (VAR v1, v2 : RVEC);
        VAR t : RVEC;
        BEGIN
            IF (((v1[Y] - v2[Y]) > ComparisonTolerance)
                 OR ((ABS (v1[Y] - v2[Y]) < ComparisonTolerance)
                      AND (v1[X] > v2[X])))
            THEN
                t := v1; v1 := v2; v2 := t
            END (* if *);
        END arrange;
    
BEGIN        
    FOR top := high TO 1 BY -1
    DO FOR current := 1 TO top
       DO
          arrange (CoOrds [current-1], CoOrds [current])
       END   
    END
END SortByYThenX;

PROCEDURE SortByXThenY (high : INTEGER; VAR CoOrds : ARRAY OF RVEC);

    VAR top, current : INTEGER;

    PROCEDURE arrange (VAR v1, v2 : RVEC);
        VAR t : RVEC;
        BEGIN
            IF (((v1[X] - v2[X]) > ComparisonTolerance)
                 OR ((ABS (v1[X] - v2[X]) < ComparisonTolerance)
                      AND (v1[Y] > v2[Y])))
            THEN
                t := v1; v1 := v2; v2 := t
            END (* if *);
        END arrange;
    
BEGIN        
    FOR top := high TO 1 BY -1
    DO FOR current := 1 TO top
       DO
          arrange (CoOrds [current-1], CoOrds [current])
       END   
    END
END SortByXThenY;


PROCEDURE FillBox (xleft, xright, ytop, ybottom : INTEGER);
    VAR
        box: RBOX;
    BEGIN
        RBox (box, 
            FLOAT (xleft), FLOAT (ybottom), FLOAT (xright), FLOAT (ytop));
        PlotBox (box);
    END FillBox;          

PROCEDURE DrawHorizontal (y, xleft, xright : INTEGER);
    BEGIN
        FillBox (xleft, xright, y, y)        
    END DrawHorizontal;

PROCEDURE DrawVertical (x, ytop, ybottom : INTEGER);
    BEGIN
        FillBox (x, x, ytop, ybottom)
    END DrawVertical;
    

PROCEDURE DrawHorizontals (ystart, yend,
                           incleft, incright : REAL;
                           xleft, xright : REAL);

    VAR y : INTEGER;
        IntDiff, IntOffset : INTEGER;
        IntLeft, IntRight : INTEGER;
    BEGIN
        xleft  := xleft  + incleft  * (FLOAT (TRUNC (ystart) + 1) - ystart);
        xright := xright + incright * (FLOAT (TRUNC (ystart) + 1) - ystart);

        IF ABS (incleft - incright) < 0.00001
        THEN
             IntDiff := TRUNC (xright - xleft + 0.5);
             xright := (xright + xleft + FLOAT (IntDiff)) / 2.0;
             IntOffset := IntDiff - 1;
        
             FOR y := TRUNC (ystart) + 1 TO TRUNC (yend)
             DO
                IntRight := TRUNC (xright);
                IntLeft := IntRight - IntOffset;
                DrawHorizontal (y, IntLeft, IntRight);
                xright := xright + incright 
             END 
        ELSE
             FOR y := TRUNC (ystart) + 1 TO TRUNC (yend)
             DO
                IntDiff := TRUNC (xright - xleft + 0.5);
                IntRight := TRUNC ((xright + xleft + FLOAT (IntDiff)) / 2.0);
                IntLeft := IntRight - IntDiff + 1;
                DrawHorizontal (y, IntLeft, IntRight);
                xright := xright + incright;
                xleft := xleft + incleft
             END (* for *)   
        END (* if *)
    END DrawHorizontals;


PROCEDURE DrawVerticals (xstart, xend,
                           inctop, incbottom : REAL;
                           ytop, ybottom : REAL);

    VAR x : INTEGER;
        IntDiff, IntOffset : INTEGER;
        IntTop, IntBottom : INTEGER;
    BEGIN
        ytop  := ytop  + inctop  * (FLOAT (TRUNC (xstart) + 1) - xstart);
        ybottom := ybottom + incbottom * (FLOAT (TRUNC(xstart) + 1) - xstart);

        IF ABS (inctop - incbottom) < 0.00001
        THEN
             IntDiff := TRUNC (ybottom - ytop + 0.5);
             ybottom := (ybottom + ytop + FLOAT (IntDiff)) / 2.0;
             IntOffset := IntDiff - 1;
        
             FOR x := TRUNC (xstart) + 1 TO TRUNC (xend)
             DO
                IntBottom := TRUNC (ybottom);
                IntTop := IntBottom - IntOffset;
                DrawVertical (x, IntTop, IntBottom);
                ybottom := ybottom + incbottom 
             END 
        ELSE
             FOR x := TRUNC (xstart) + 1 TO TRUNC (xend)
             DO
                IntDiff := TRUNC (ybottom - ytop + 0.5);
                IntBottom := TRUNC ((ybottom + ytop + FLOAT (IntDiff)) / 2.0);
                IntTop := IntBottom - IntDiff + 1;
                DrawVertical (x, IntTop, IntBottom);
                ybottom := ybottom + incbottom;
                ytop := ytop + inctop
             END (* for *)   
        END (* if *)
    END DrawVerticals;


PROCEDURE HorizontalFillBlock (CoOrds : FOURRVECS);
VAR xa, xb, y1, y2, y3, y4, x1, x2, x3, x4 : REAL;
BEGIN
     SortByYThenX (3, CoOrds);
     x1 := CoOrds [1][X]; y1 := CoOrds [1][Y];
     x2 := CoOrds [2][X]; y2 := CoOrds [2][Y];
     x3 := CoOrds [3][X]; y3 := CoOrds [3][Y];
     x4 := CoOrds [4][X]; y4 := CoOrds [4][Y];

     IF TRUNC (y2) # TRUNC (y1)
     THEN
          xa := x1; xb := x1;
          IF x2 < x3
          THEN
               DrawHorizontals (y1, y2,
                                (x2-x1)/(y2-y1),
                                (x3-x1)/(y3-y1),
                                xa, xb)
          ELSE 
               DrawHorizontals (y1, y2,
                                (x3-x1)/(y3-y1),
                                (x2-x1)/(y2-y1),
                                xa, xb)
          END
     END;
  
     IF x1 < x2 
     THEN xa := x1 + (x3 - x1) * (y2 - y1) / (y3 - y1); 
          xb := x2
     ELSE xa := x2;
          xb := x1 + (x3 - x1) * (y2 - y1) / (y3 - y1)
     END;

     IF TRUNC (y3) # TRUNC (y2)
     THEN
           DrawHorizontals (y2, y3,
                         (x3-x1)/(y3-y1),
                         (x3-x1)/(y3-y1),
                         xa, xb) 
     END (* if *);

     IF x2 < x3
     THEN xa := x2 + (x4 - x2) * (y3 - y2) / (y4 - y2);
          xb := x3
     ELSE xa := x3;
          xb := x2 + (x4 - x2) * (y3 - y2) / (y4 - y2)
     END (* if *);
     
     IF TRUNC (y3) # TRUNC (y4)
     THEN 
          IF x2 < x3
          THEN
               DrawHorizontals (y3, y4,
                                (x4-x2)/(y4-y2),
                                (x4-x3)/(y4-y3),
                                xa, xb)
          ELSE
               DrawHorizontals (y3, y4,
                                (x4-x3)/(y4-y3),
                                (x4-x2)/(y4-y2),
                                xa, xb)
          END
     END 
     
END HorizontalFillBlock;

PROCEDURE VerticalFillBlock (CoOrds : FOURRVECS);
VAR xa, xb, ya, yb, y1, y2, y3, y4, x1, x2, x3, x4 : REAL;
BEGIN
     SortByXThenY (3, CoOrds);
     x1 := CoOrds [1][X]; y1 := CoOrds [1][Y];
     x2 := CoOrds [2][X]; y2 := CoOrds [2][Y];
     x3 := CoOrds [3][X]; y3 := CoOrds [3][Y];
     x4 := CoOrds [4][X]; y4 := CoOrds [4][Y];

     IF TRUNC (x2) # TRUNC (x1)
     THEN
          ya := y1; yb := y1;
          IF y2 < y3
          THEN
               DrawVerticals (x1, x2,
                                (y2-y1)/(x2-x1),
                                (y3-y1)/(x3-x1),
                                ya, yb)
          ELSE 
               DrawVerticals (x1, x2,
                                (y3-y1)/(x3-x1),
                                (y2-y1)/(x2-x1),
                                ya, yb)
          END
     END;
  
     IF y1 < y2 
     THEN ya := y1 + (y3 - y1) * (x2 - x1) / (x3 - x1); 
          yb := y2
     ELSE ya := y2;
          yb := y1 + (y3 - y1) * (x2 - x1) / (x3 - x1)
     END;

     IF TRUNC (x3) # TRUNC (x2)
     THEN
           DrawVerticals (x2, x3,
                         (y3-y1)/(x3-x1),
                         (y3-y1)/(x3-x1),
                         ya, yb) 
     END (* if *);

     IF y2 < y3
     THEN ya := y2 + (y4 - y2) * (x3 - x2) / (x4 - x2);
          yb := y3
     ELSE ya := y3;
          yb := y2 + (y4 - y2) * (x3 - x2) / (x4 - x2)
     END (* if *);
     
     IF TRUNC (x3) # TRUNC (x4)
     THEN 
          IF y2 < y3
          THEN
               DrawVerticals (x3, x4,
                                (y4-y2)/(x4-x2),
                                (y4-y3)/(x4-x3),
                                ya, yb)
          ELSE
               DrawVerticals (x3, x4,
                                (y4-y3)/(x4-x3),
                                (y4-y2)/(x4-x2),
                                ya, yb)
          END
     END 
     
END VerticalFillBlock;


PROCEDURE HorizontalFillTriangle (p1, p2, p3 : RVEC) ;
VAR parray : ARRAY [0..2] OF RVEC ;
    x1, x2, x3, y1, y2, y3, xa, xb : REAL ;
    BEGIN
(*
        ReverseForCanon (p1) ;
        ReverseForCanon (p2) ;
        ReverseForCanon (p3) ;
*)
        parray [0] := p1 ;
        parray [1] := p2 ;
        parray [2] := p3 ;
        SortByYThenX (2, parray) ;
        
        x1 := parray [0,X]; y1 := parray [0,Y];
        x2 := parray [1,X]; y2 := parray [1,Y];
        x3 := parray [2,X]; y3 := parray [2,Y];

        IF TRUNC (y2) # TRUNC (y1)
        THEN
            xa := x1 ;
            xb := x1 ;
             IF (x2 < x3) OR ((x2 = x3) AND ((x2-x1)/(y2-y1) < (x3-x1)/(y3-y1)))
             THEN
                  DrawHorizontals (y1, y2,
                                   (x2-x1)/(y2-y1),
                                   (x3-x1)/(y3-y1),
                                   xa, xb)
             ELSE 
                  DrawHorizontals (y1, y2,
                                   (x3-x1)/(y3-y1),
                                   (x2-x1)/(y2-y1),
                                   xa, xb)
             END
        END;
     
        IF TRUNC (y2) # TRUNC (y3)
        THEN 
            xa := x1 + (x3 - x1) * (y2 - y1) / (y3 - y1) ;
            xb := x2 ;
             IF (x1 < x2) OR ((x1 = x2) AND ((x3-x1)/(y3-y1) > (x3-x2)/(y3-y2)))
             THEN
                  DrawHorizontals (y2, y3,
                                   (x3-x1)/(y3-y1),
                                   (x3-x2)/(y3-y2),
                                   xa,
                                   xb)
             ELSE
                  DrawHorizontals (y2, y3,
                                   (x3-x2)/(y3-y2),
                                   (x3-x1)/(y3-y1),
                                   xb,
                                   xa)
             END
        END 
        
    END HorizontalFillTriangle;


PROCEDURE LineSegment (vfrom, vto : RVEC; width : REAL);
    VAR EndPush, WidthPush, Push1, Push2,
        v1, v2, v3, v4 : RVEC;
        FourVector : FOURRVECS;
        i : INTEGER;
BEGIN
      IF NOT REqual (vfrom, vto, epsilon) THEN
          EndPush := vto;
          RSubtract (vfrom, EndPush);
          CanonSetRLength (EndPush, width / 2.0);
          CanonFindPerpendicularVector (EndPush, WidthPush);
          Push1 := EndPush;
          Push2 := EndPush;
          RAdd (WidthPush, Push1);
          RSubtract (WidthPush, Push2);
          v1 := vto; v2 := vto;
          RAdd (Push1, v1); RAdd (Push2, v2);
          v3 := vfrom; v4 := vfrom;
          RSubtract (Push1, v3); RSubtract (Push2, v4);
(*
          ReverseForCanon (v1);
          ReverseForCanon (v2);
          ReverseForCanon (v3);
          ReverseForCanon (v4);
*)
          SetFourVector (v1, v2, v3, v4, FourVector);
          IF ABS (EndPush [X]) > ABS (EndPush [Y])
          THEN HorizontalFillBlock (FourVector)
          ELSE VerticalFillBlock (FourVector)
          END (* if *)
     END (* if *);
END LineSegment;


PROCEDURE DrawCyclicLine (high : INTEGER; VAR at : ARRAY OF RVEC);
VAR i : INTEGER;
    FullVector : RVEC;  (* Current line direction *)
    PushVector : RVEC;  (* Current pattern segment vector *)
    Position   : RVEC;  (* Where the line has reached *)
    CyclesSoFar : REAL;
    NextCornerCycle : REAL;
    NextChangeIndex : CHANGEINDEX;
    CyclesToNextChange : REAL;
    DistanceToNextChange : REAL;
    Drawing : BOOLEAN;  (* TRUE => pen down *) 
BEGIN
    NextCornerCycle := 0.0;    
    NextChangeIndex := 0;
    CyclesSoFar := 0.0;
    Drawing := TRUE;    (* start cycle with pen down *)
    
    FOR i := 1 TO high
    DO IF NOT REqual (at [i-1], at [i], epsilon) THEN
            NextCornerCycle := NextCornerCycle
                      + CanonRDistance (at [i-1], at [i]) / CycleLength;
            CyclesToNextChange := FRAC (Cycle[NextChangeIndex] - CyclesSoFar);
            Position := at [i-1];
            FullVector := at [i];
            RSubtract (Position, FullVector);
            
            WHILE CyclesSoFar + CyclesToNextChange < NextCornerCycle 
            DO
               DistanceToNextChange := CyclesToNextChange * CycleLength;
               PushVector := FullVector;
               CanonSetRLength (PushVector, DistanceToNextChange);
               RAdd (Position, PushVector);
               IF Drawing THEN
                 LineSegment (Position, PushVector, LineWidth);
               END;
               Drawing := NOT Drawing;
               Position := PushVector;
               CyclesSoFar := CyclesSoFar + CyclesToNextChange;
               NextChangeIndex := (NextChangeIndex + 1) MOD NoOfChanges;
               CyclesToNextChange := FRAC (Cycle [NextChangeIndex] - CyclesSoFar)
            END (* while *);
            
            (* Draw up to the corner *)
            
            DistanceToNextChange := 
                (NextCornerCycle - CyclesSoFar) * CycleLength;
            PushVector := FullVector;
            CanonSetRLength (PushVector, DistanceToNextChange);
            RAdd (Position, PushVector);  (* PushVector should now be
                                             approximately equal to at [i] *)
            IF Drawing THEN LineSegment (Position, PushVector, LineWidth) END;
            CyclesSoFar := NextCornerCycle;
            CyclesToNextChange := Cycle [NextChangeIndex] - FRAC (CyclesSoFar)

       END (* if *)
    END (* for *)
END DrawCyclicLine;


PROCEDURE DrawMarker (at : RVEC);
VAR Size : REAL;
    Vxref, Vyref, Vref1, Vend1, Vend2 : RVEC;
    i : INTEGER;
BEGIN
    Size := NominalLineWidth * MarkerSize;
    
    CASE CurrentMarkerStyle OF
    
      DotMark    : RVec (Vend1, epsilon * 2.0, 0.0);
                   Vend2 := at;
                   RAdd (Vend1, Vend2);
                   LineSegment (at, Vend2, Size * 2.0) |
                   
      PlusMark   : RVec (Vxref, 1.0, 0.0);
                      (* The constant 3.0 may be altered at will *)
                   CanonSetRLength (Vxref, Size * 6.0);
                   CanonFindPerpendicularVector (Vxref, Vyref);
                   Vend1 := at; Vend2 := at;
                   RAdd (Vxref, Vend1);
                   RSubtract (Vxref, Vend2);
                   LineSegment (Vend1, Vend2, Size);
                   Vend1 := at; Vend2 := at;
                   RAdd (Vyref, Vend1);
                   RSubtract (Vyref, Vend2);
                   LineSegment (Vend1, Vend2, Size) |
                   
      StarMark   : RVec (Vxref, 1.0, 0.0);
                   CanonSetRLength (Vxref, Size * 6.0);
                   CanonFindPerpendicularVector (Vxref, Vyref);
                   Vend1 := at; Vend2 := at;
                   RAdd (Vxref, Vend1);
                   RSubtract (Vxref, Vend2);
                   LineSegment (Vend1, Vend2, Size);
                   Vend1 := at; Vend2 := at;
                   RAdd (Vyref, Vend1);
                   RSubtract (Vyref, Vend2);
                   LineSegment (Vend1, Vend2, Size);
                   
                   RAdd (Vxref, Vyref);
                   CanonSetRLength (Vyref, Size * 6.0);
                   CanonFindPerpendicularVector (Vyref, Vxref);
                   Vend1 := at; Vend2 := at;
                   RAdd (Vxref, Vend1);
                   RSubtract (Vxref, Vend2);
                   LineSegment (Vend1, Vend2, Size);
                   Vend1 := at; Vend2 := at;
                   RAdd (Vyref, Vend1);
                   RSubtract (Vyref, Vend2);
                   LineSegment (Vend1, Vend2, Size) |
                   
      ZeroMark   : RVec (Vxref, 1.0, 0.0);
                   RVec (Vyref, 0.0, 1.0);
                   CanonSetRLength (Vxref, Size * 4.0);
                   CanonSetRLength (Vyref, Size);
                   Vref1 := at;
                   RAdd (Vxref, Vref1);
                   Vend1 := Vref1;
                   Vend2 := Vref1;
                   RAdd (Vyref, Vend1);
                   RSubtract (Vyref, Vend2);
                   LineSegment (Vend1, Vend2, Size);
                   CanonSetRLength (Vxref, Size * 8.0);
                   RSubtract (Vxref, Vend1);
                   RSubtract (Vxref, Vend2);
                   LineSegment (Vend1, Vend2, Size);
                   RSubtract (at, Vend1);
                   RSubtract (at, Vend2);
                   CanonFindPerpendicularVector (Vend1, Vend1);
                   CanonFindPerpendicularVector (Vend2, Vend2);
                   RAdd (at, Vend1);
                   RAdd (at, Vend2);
                   LineSegment (Vend1, Vend2, Size);
                   CanonSetRLength (Vyref, Size * 8.0);
                   RAdd (Vyref, Vend1);
                   RAdd (Vyref, Vend2);
                   LineSegment (Vend1, Vend2, Size);
                   Vref1 := Vxref;
                   CanonSetRLength (Vref1, -Size / 2.0);
                   CanonSetRLength (Vyref,  Size / 2.0);
                   RAdd (Vyref, Vref1);
                   RAdd (Vref1, Vend1);
                   Vref1 := Vyref;
                   CanonSetRLength (Vref1, -Size * 0.7071);
                   RAdd (Vref1, Vend1);
                   Vend2 := Vend1;
                   RSubtract (at, Vend2);
                   CanonFindPerpendicularVector (Vend2, Vend2);
                   Vend2 [Y] := - Vend2 [Y];
                   RAdd (at, Vend2);
                   LineSegment (Vend1, Vend2, Size);
                   FOR i := 1 TO 3
                   DO RSubtract (at, Vend1);
                      RSubtract (at, Vend2);
                      CanonFindPerpendicularVector (Vend1, Vend1);
                      CanonFindPerpendicularVector (Vend2, Vend2);
                      RAdd (at, Vend1);
                      RAdd (at, Vend2);
                      LineSegment (Vend1, Vend2, Size)
                   END (* for *) |
      
      XMark      : 
                   RVec (Vxref, 1.0, 0.0);
                   CanonFindPerpendicularVector (Vxref, Vyref);
                   RAdd (Vxref, Vyref);
                   CanonSetRLength (Vyref, Size * 6.0);
                   CanonFindPerpendicularVector (Vyref, Vxref);
                   Vend1 := at; Vend2 := at;
                   RAdd (Vxref, Vend1);
                   RSubtract (Vxref, Vend2);
                   LineSegment (Vend1, Vend2, Size);
                   Vend1 := at; Vend2 := at;
                   RAdd (Vyref, Vend1);
                   RSubtract (Vyref, Vend2);
                   LineSegment (Vend1, Vend2, Size)
    ELSE
              (* Use a small dot mark *)
                   RVec (Vend1, epsilon * 2.0, 0.0);
                   Vend2 := at;
                   RAdd (Vend1, Vend2);
                   LineSegment (at, Vend2, Size)
    END (* case *);
END DrawMarker;


BEGIN
    HorizontalResolution := FLOAT (PixelsWide);
    VerticalResolution := FLOAT (PixelsHigh);
    
    IF VerticalResolution >= HorizontalResolution 
    THEN 
         UnitX := 1.0;
         UnitY := VerticalResolution / HorizontalResolution
    ELSE
         UnitX := HorizontalResolution / VerticalResolution;
         UnitY := 1.0
    END (* if *);
    
    UnitXSquared := UnitX * UnitX;
    UnitYSquared := UnitY * UnitY

END RasterOp.
