IMPLEMENTATION MODULE Scientific;

(*
    Title: 	Write reals in scientific format
    LastEdit:	"Wed Mar  6 14:16:08 1985"
    Author: 	Alan Jones,T72,232,322723
		Cambridge University Computer Laboratory
*)

FROM Ascii IMPORT DigitsCh ;

FROM Numerical IMPORT Round, Frac ;

FROM Streams IMPORT Stream, Put ;

FROM TextIO IMPORT WriteCARDINAL ;

FROM UtilError IMPORT ReportPError ;

FROM WriteF IMPORT R ;


PROCEDURE WriteFrac (stream : Stream ;
                     Value : REAL ;
                     DigitsToGo : CARDINAL) ;
    BEGIN
        IF Value >= 1.0 THEN
            ReportPError ("Scientific", FALSE,
                          "WriteFrac will ignore integer part of argument %E12.3\N",
                          R (Value), 0) ;
            Value := Value - FLOAT (TRUNC (Value)) ;
        ELSIF Value <= -1.0 THEN
            ReportPError ("Scientific", FALSE,
                          "WriteFrac will ignore integer part of argument %E12.3\N",
                          R (Value), 0) ;
            Value := Value + FLOAT (TRUNC (ABS (Value))) ;
        END (* if *);
        
        IF DigitsToGo > 0 THEN
            Value := 10.0 * Value ;
            Put (stream, DigitsCh [TRUNC (Value)]) ;
            WriteFrac (stream, Frac (Value), DigitsToGo - 1)
        END (* if *);
    END WriteFrac;


PROCEDURE WriteScientific (stream : Stream ;
                           Value : REAL ;
                           SignificantDigits : CARDINAL);
(* Write out the Value in scientific notation, that is
   as a real times a power of 1000 *)
VAR Exponent : INTEGER ;
    IntegerDigits, FractionalDigits : INTEGER ;
    Index : INTEGER ;
    Rounding : REAL ;
    BEGIN
        IF SignificantDigits = 0 THEN
            SignificantDigits := 8
        END (* if *);
        Exponent := 0 ;
        IF Value < 0.0 THEN
           Value := - Value ;
           Put (stream, "-") ;
        END (* if *);
        
        IF Value > 0.0 THEN
	    WHILE Value < 1.0 DO
		Value := Value * 1000.0 ;
		DEC (Exponent, 3)
	    END (* while *);
        END (* if *);
        
        WHILE Value >= 1000.0 DO
            Value := Value / 1000.0 ;
            INC (Exponent, 3)
        END (* while *);
        
        IF Value >= 100.0 THEN
            IntegerDigits := 3
        ELSIF Value >= 10.0 THEN
            IntegerDigits := 2
        ELSE
            IntegerDigits := 1
        END (* if *);
        
        FractionalDigits := INTEGER (SignificantDigits) - IntegerDigits ;

	Rounding := 0.5 ;
	FOR Index := 1 TO FractionalDigits DO
	    Rounding := Rounding * 0.1 ;
	END (* for *) ;
        Value := Value + Rounding ;
        
        WriteCARDINAL (stream, TRUNC (Value)) ;
        Put (stream, ".") ;
        
        IF FractionalDigits <= 0 THEN
            Put (stream, "0")
        ELSE
            WriteFrac (stream, Frac (Value), FractionalDigits) 
        END (* if *);
        
        IF Exponent # 0 THEN
            Put (stream, "E") ;
            IF Exponent < 0 THEN
                Put (stream, "-")
            ELSE
                Put (stream, "+")
            END (* if *);
            WriteCARDINAL (stream, ABS (Exponent))
        END (* if *);
    END WriteScientific;

END Scientific.
