IMPLEMENTATION MODULE Timing;

(*
    Title: 	Time various parts of execution
    LastEdit:	"Fri May  3 16:28:59 1985"
    Author: 	ahj
		Cambridge University Computer Laboratory
*)

FROM Streams IMPORT Stream ;

FROM SYSTEM IMPORT UNIXCALL, ADR, ADDRESS ;

FROM UnixCalls IMPORT times ;

FROM WriteF IMPORT WriteF3, R ;

CONST TicksPerSecond = 60.0 ;

TYPE TIMES = RECORD
            UserTime,
            SystemTime,
            DescendantUserTime,
            DescendantSystemTime:   CARDINAL;
        END (* record *);

VAR initialTimes, lastBreakTimes, lastButOneBreakTimes : TIMES ;

PROCEDURE TimeBetween (t1, t2 : TIMES ;
                       VAR t3 : TIMES) ;
(* t3 := t1 - t2 *)
    BEGIN
        t3.UserTime := t2.UserTime - t1.UserTime ;
        t3.SystemTime := t2.SystemTime - t1.SystemTime ;
        t3.DescendantUserTime := t2.DescendantUserTime - t1.DescendantUserTime ;
        t3.DescendantSystemTime := t2.DescendantSystemTime - t1.DescendantSystemTime ;
    END TimeBetween;

PROCEDURE ZeroTimes;
VAR rc : CARDINAL ;
    BEGIN
        rc := UNIXCALL (times, ADR (initialTimes)) ;
        rc := UNIXCALL (times, ADR (lastBreakTimes)) ;
        rc := UNIXCALL (times, ADR (lastButOneBreakTimes)) ;
    END ZeroTimes;

PROCEDURE BreakTimes;
VAR rc : CARDINAL ;
    BEGIN
        lastButOneBreakTimes := lastBreakTimes ;
        rc := UNIXCALL (times, ADR (lastBreakTimes)) ;
    END BreakTimes;

PROCEDURE PrintTotalTimes (stream : Stream) ;
VAR totalTimes : TIMES ;
    BEGIN
        TimeBetween (initialTimes, lastBreakTimes, totalTimes) ;
        WITH totalTimes DO
        WriteF3 (stream, "Total %F4, User %F4, System %F4",
                 R (FLOAT (UserTime + SystemTime) / TicksPerSecond),
                 R (FLOAT (UserTime) / TicksPerSecond),
                 R (FLOAT (SystemTime) / TicksPerSecond)) ;
        END (* with *);
    END PrintTotalTimes;

PROCEDURE PrintPartialTimes (stream : Stream) ;
VAR partialTimes : TIMES ;
    BEGIN
        TimeBetween (lastButOneBreakTimes, lastBreakTimes, partialTimes) ;
        WITH partialTimes DO
        WriteF3 (stream, "Partial %F4, User %F4, System %F4",
                 R (FLOAT (UserTime + SystemTime) / TicksPerSecond),
                 R (FLOAT (UserTime) / TicksPerSecond),
                 R (FLOAT (SystemTime) / TicksPerSecond)) ;
        END (* with *);
    END PrintPartialTimes;

END Timing.
