IMPLEMENTATION MODULE Vector;

(*
    Title: 	interval, vector and rectangle operations
    LastEdit:	"Fri May 17 18:05:03 1985"
    Author: 	jd
		Cambridge University Computer Laboratory
*)

(* $T-,$R- *)

FROM Arithmetic IMPORT SquareRoot, RMax, RMin, IMin, IMax;

CONST Large = 1.0E20; (* a large real number *)


PROCEDURE Orthogonal (z: AXIS): AXIS;
BEGIN
    IF z = X THEN RETURN Y ELSE RETURN X END;
END Orthogonal;


PROCEDURE Opposite (l: BOUND): BOUND;
BEGIN
    IF l = Low THEN RETURN High ELSE RETURN Low END;
END Opposite;


PROCEDURE Left (VAR z: AXIS; VAR l: BOUND);
BEGIN
    IF z = Y THEN l := Opposite (l) END;
    z := Orthogonal (z);
END Left;


PROCEDURE Right (VAR z: AXIS; VAR l: BOUND);
BEGIN
    IF z = X THEN l := Opposite (l) END;
    z := Orthogonal (z);
END Right;


PROCEDURE RSeg (a, b: REAL; VAR s: RSEG);
BEGIN
    s [Low] := RMin (a, b);
    s [High] := RMax (a, b);
END RSeg;


PROCEDURE ISeg (a, b: INTEGER; VAR s: ISEG);
BEGIN
    s [Low] := IMin (a, b);
    s [High] := IMax (a, b);
END ISeg;


PROCEDURE RSize (VAR s: RSEG): REAL;
BEGIN
    RETURN s [High] - s [Low]
END RSize;


PROCEDURE ISize (VAR s: ISEG): INTEGER;
BEGIN
    RETURN s [High] - s [Low]
END ISize;


PROCEDURE RInClosed (p: REAL; VAR s: RSEG): BOOLEAN;
BEGIN
    RETURN (p >= s [Low]) AND (p <= s [High]);
END RInClosed;


PROCEDURE IInClosed (p: INTEGER; VAR s: ISEG): BOOLEAN;
BEGIN
    RETURN (p >= s [Low]) AND (p <= s [High]);
END IInClosed;


PROCEDURE RInOpen (p: REAL; VAR s: RSEG): BOOLEAN;
BEGIN
    RETURN (p > s [Low]) AND (p < s [High]);
END RInOpen;


PROCEDURE IInOpen (p: INTEGER; VAR s: ISEG): BOOLEAN;
BEGIN
    RETURN (p > s [Low]) AND (p < s [High]);
END IInOpen;


PROCEDURE RIntersect (VAR s1, s2, intersection: RSEG): BOOLEAN;
BEGIN
    intersection [Low] := RMax (s1 [Low], s2 [Low]);
    intersection [High] := RMin (s1 [High], s2 [High]);
    RETURN intersection [High] >= intersection [Low]
END RIntersect;


PROCEDURE IIntersect (VAR s1, s2, intersection: ISEG): BOOLEAN;
BEGIN
    intersection [Low] := IMax (s1 [Low], s2 [Low]);
    intersection [High] := IMin (s1 [High], s2 [High]);
    RETURN intersection [High] >= intersection [Low]
END IIntersect;


PROCEDURE RVec (VAR v: RVEC; x, y: REAL);
BEGIN
    v[X] := x; 
    v[Y] := y
END RVec;


PROCEDURE IVec (VAR v: IVEC; x, y: INTEGER);
BEGIN
    v[X] := x; 
    v[Y] := y
END IVec;


PROCEDURE REqual (VAR v1, v2: RVEC; epsilon: REAL): BOOLEAN;
BEGIN
    IF (ABS (v1[X]-v2[X]) <= epsilon) AND (ABS (v1[Y]-v2[Y]) <= epsilon) THEN
        RETURN TRUE
    ELSE RETURN FALSE
    END;
END REqual;


PROCEDURE IEqual (VAR v1, v2: IVEC): BOOLEAN;
BEGIN
    RETURN (v1[X] = v2[X]) AND (v1[Y] = v2[Y]);
END IEqual;


PROCEDURE RAdd (VAR v, to: RVEC);
BEGIN
    to [X] := to [X] + v [X];
    to [Y] := to [Y] + v [Y];
END RAdd;


PROCEDURE IAdd (VAR v, to: IVEC);
BEGIN
    to [X] := to [X] + v [X];
    to [Y] := to [Y] + v [Y];
END IAdd;


PROCEDURE RSubtract (VAR v, from: RVEC);
BEGIN
    from [X] := from [X] - v [X];
    from [Y] := from [Y] - v [Y];
END RSubtract;


PROCEDURE ISubtract (VAR v, from: IVEC);
BEGIN
    from [X] := from [X] - v [X];
    from [Y] := from [Y] - v [Y];
END ISubtract;


PROCEDURE RDistance (VAR v1, v2: RVEC): REAL;
    VAR v: RVEC;
BEGIN
    v[X] := v1[X] - v2[X];
    v[Y] := v1[Y] - v2[Y];
    RETURN RLength (v);
END RDistance;


PROCEDURE IDistance (VAR v1, v2: IVEC): CARDINAL;
BEGIN
    RETURN ABS (v1[X] - v2[X]) + ABS (v1[Y] - v2[Y]);
END IDistance;


PROCEDURE RLength (VAR v: RVEC): REAL;
BEGIN
    RETURN SquareRoot (v[X] * v[X] + v[Y] * v[Y]);
END RLength;


PROCEDURE RSetLength (VAR v: RVEC; l: REAL);
    VAR r: REAL;
BEGIN
    r := RLength (v);
    IF r > 0.0 THEN
        l := l / r;
        v[X] := v[X] * l;
        v[Y] := v[Y] * l;
    END;
END RSetLength;


PROCEDURE RBox (VAR b: RBOX; lx, ly, hx, hy: REAL);
BEGIN
    RVec (b [Low], lx, ly);
    RVec (b [High], hx, hy);
END RBox;


PROCEDURE IBox (VAR b: IBOX; lx, ly, hx, hy: INTEGER);
BEGIN
    IVec (b [Low], lx, ly);
    IVec (b [High], hx, hy);
END IBox;


PROCEDURE RBoxSeg (VAR r: RBOX; z: AXIS; VAR s: RSEG);
BEGIN
    s [Low] := r [Low, z];
    s [High] := r [High, z];
END RBoxSeg;


PROCEDURE IBoxSeg (VAR r: IBOX; z: AXIS; VAR s: ISEG);
BEGIN
    s [Low] := r [Low, z];
    s [High] := r [High, z];
END IBoxSeg;


PROCEDURE RSetBoxSeg (VAR r: RBOX; z: AXIS; VAR s: RSEG);
BEGIN
    r [Low, z] := s [Low];
    r [High, z] := s [High];
END RSetBoxSeg;


PROCEDURE ISetBoxSeg (VAR r: IBOX; z: AXIS; VAR s: ISEG);
BEGIN
    r [Low, z] := s [Low];
    r [High, z] := s [High];
END ISetBoxSeg;


PROCEDURE RInBox (VAR v: RVEC; VAR b: RBOX): BOOLEAN;
    VAR z: AXIS;
BEGIN
    FOR z := X TO Y DO
        IF (v [z] < b [Low, z]) OR (v [z] > b [High, z]) THEN RETURN FALSE END
    END;
    RETURN TRUE
END RInBox;


PROCEDURE IInBox (VAR v: IVEC; VAR b: IBOX): BOOLEAN;
    VAR z: AXIS;
BEGIN
    FOR z := X TO Y DO
        IF (v [z] < b [Low, z]) OR (v [z] > b [High, z]) THEN RETURN FALSE END
    END;
    RETURN TRUE
END IInBox;

END Vector.
