IMPLEMENTATION MODULE VectorExtras;

(*
    Title: 	More vector and box operations
    LastEdit:	"Tue Aug 20 13:37:30 1985"
    Author: 	ahj
		Cambridge University Computer Laboratory
*)

FROM Vector IMPORT AXIS, BOUND, IVEC, IBOX, RBOX, RBox, RVEC ;

PROCEDURE FloatBox (ibox : IBOX ; VAR rbox : RBOX) ;
VAR axis : AXIS ;
    bound : BOUND ;
    BEGIN
        FOR axis := X TO Y DO
            FOR bound := Low TO High DO
                rbox [bound, axis] := FLOAT (ibox [bound, axis])
            END (* for *);
        END (* for *);
    END FloatBox;

PROCEDURE ExpandBox (VAR Box : RBOX ; f: REAL) ;
VAR lowx, highx, lowy, highy, sumx, sumy, diffx, diffy : REAL ;
    BEGIN
        lowx := Box [Low,X] ;
        highx := Box [High,X] ;
        lowy := Box [Low,Y] ;
        highy := Box [High,Y] ;
        
        sumx := (lowx + highx) ;
        sumy := (lowy + highy) ;
        diffx := (highx - lowx) * f ;
        diffy := (highy - lowy) * f ;
        
        Box [Low,X] := (sumx - diffx) / 2.0 ;
        Box [Low,Y] := (sumy - diffy) / 2.0 ;
        Box [High,X] := (sumx + diffx) / 2.0 ;
        Box [High,Y] := (sumy + diffy) / 2.0 ;
    END ExpandBox ;
    

PROCEDURE ExpandRBoxAxis (VAR Box : RBOX ;
                          Factor : REAL ;
                          Axis : AXIS) ;
VAR low, high, sum, diff : REAL ;
    BEGIN
        low := Box [Low, Axis] ;
        high := Box [High, Axis] ;
        sum := low + high ;
        diff := (high - low) * Factor ;
        Box [Low, Axis] := (sum - diff) / 2.0 ;
        Box [High, Axis] := (sum + diff) / 2.0 ;
    END ExpandRBoxAxis;


PROCEDURE ShiftBox (VAR Box : RBOX ;
                    f : REAL ;
                    a : AXIS) ;
VAR shove : REAL ;
    BEGIN
        shove := (Box [High,a] - Box [Low,a]) * f ;  
        Box [High,a] := Box [High,a] + shove ;
        Box [Low,a] := Box [Low,a] + shove ;
    END ShiftBox ;

PROCEDURE MakeRBox (VAR Box : RBOX ;
                    Vec1, Vec2 : RVEC) ;
VAR x1, x2, y1, y2 : REAL ;
    BEGIN
        x1 := Vec1 [X] ;
        x2 := Vec2 [X] ;
        y1 := Vec1 [Y] ;
        y2 := Vec2 [Y] ;
        IF x1 < x2 THEN
            Box [Low,X] := x1 ;
            Box [High,X] := x2
        ELSE
            Box [Low,X] := x2 ;
            Box [High,X] := x1 
        END (* if *);
        IF y1 < y2 THEN
            Box [Low,Y] := y1 ;
            Box [High,Y] := y2
        ELSE
            Box [Low,Y] := y2 ;
            Box [High,Y] := y1 
        END (* if *);
    END MakeRBox;


PROCEDURE ExpandRBoxToSquare (VAR Box : RBOX) ;
VAR Ratio : REAL ;
    BEGIN
        Ratio := (Box [High,X] - Box [Low,X]) / (Box [High,Y] - Box [Low,Y]) ;
        IF Ratio > 1.0 THEN
            ExpandRBoxAxis (Box, Ratio, Y) ;
        ELSE
            ExpandRBoxAxis (Box, 1.0 / Ratio, X) ;
        END (* if *);
    END ExpandRBoxToSquare;

PROCEDURE SquashRBoxToSquare (VAR Box : RBOX) ;
VAR Ratio : REAL ;
    BEGIN
        Ratio := (Box [High,X] - Box [Low,X]) / (Box [High,Y] - Box [Low,Y]) ;
        IF Ratio > 1.0 THEN
            ExpandRBoxAxis (Box, 1.0 / Ratio, X) ;
        ELSE
            ExpandRBoxAxis (Box, Ratio, Y) ;
        END (* if *);
    END SquashRBoxToSquare;


END VectorExtras.
