(*
    Title: 	Real implementation
    LastEdit:	"Thu Feb 25 15:00:26 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE NetMember;


IMPORT
  Marshal, MCLib, RPC, RPCNets, Strings, SysStreams, SYSTEM, Threads, 
  TimeDate, Timers, WriteF;

FROM Storage     IMPORT ALLOCATE, DEALLOCATE;
FROM SystemTypes IMPORT String;


CONST
  TableSize = 128;
  TableTop  = TableSize - 1;


TYPE
  Entry  = POINTER TO EntryR;
  EntryR = 
    RECORD
      nextName: Entry;
      name:     String;
      dist:     RPC.Distance;
      place:    RPC.Place
    END;
  Table = ARRAY [0..TableTop] OF Entry;


VAR
  nameTable: Table;
  tableLock: Threads.Mutex;


PROCEDURE Find(
      name: String;
      dist: RPC.Distance;
  VAR place:RPC.Place    (* Out *)
 ):RPC.MemberRC;
  VAR
    mrc:   RPC.MemberRC;
    entry: Entry;
    pentry:Entry;
  BEGIN
    Threads.Acquire(tableLock);
      mrc := LookupByName(name, entry, pentry);
      IF mrc = RPC.BindOK THEN place := entry^.place END;
    Threads.Release(tableLock);
    Log("Find", mrc, name, place);
    RETURN mrc
  END Find;


PROCEDURE Offer(
      name: String;
      dist: RPC.Distance;
      place:RPC.Place
 ):RPC.MemberRC;
  VAR
    mrc:    RPC.MemberRC;
    hash:   INTEGER;
    entry,
    nameEnt,
    namePre:Entry;
  BEGIN
    Threads.Acquire(tableLock);
      mrc := LookupByName(name, nameEnt, namePre);
      IF mrc = RPC.BindOK THEN
        entry := nameEnt
      ELSIF mrc = RPC.NotFound THEN
        NEW(entry);
        namePre^.nextName := entry;
        entry^.nextName := NIL;
        entry^.name := Strings.CopySS(name);
        mrc := RPC.BindOK
      END;
      IF mrc = RPC.BindOK THEN
        entry^.dist := dist;
        entry^.place := place
      END;      
    Threads.Release(tableLock);
    Log("Offer", mrc, name, place);
    RETURN mrc
  END Offer;


PROCEDURE Retract(
      name: String;
      place:RPC.Place;
      dist: RPC.Distance
 );
  VAR
    mrc:   RPC.MemberRC;
    entry: Entry;
    pentry:Entry;
  BEGIN
    Threads.Acquire(tableLock);
      IF  (LookupByName(name, entry, pentry) = RPC.BindOK)
      AND (RPC.SamePlace(place, entry^.place)) THEN
        pentry^.nextName := entry^.nextName;
        Strings.Dispose(entry^.name);
        DISPOSE(entry)
      END;
    Threads.Release(tableLock);
    Log("Retract", RPC.BindOK, name, place);
  END Retract;


PROCEDURE FindAt(
      b:    RPC.Binder;
      name: String;
  VAR place:RPC.Place    (* Out *)
 ):RPC.MemberRC;
  BEGIN
    RETURN Find(name, RPC.WithinDomain, place)
  END FindAt;


PROCEDURE LookupByName(
      name:  String;
  VAR entry: Entry;
  VAR pentry:Entry
 ):RPC.MemberRC;
  VAR
    mrc:RPC.MemberRC;
  BEGIN
    pentry := SYSTEM.ADR(nameTable[MCLib.HashBytes(name, -1, TableSize)]);
    LOOP
      entry := pentry^.nextName;
      IF entry = NIL THEN mrc := RPC.NotFound; EXIT END;
      IF Strings.EqualSS(name, entry^.name) THEN
        mrc := RPC.BindOK; EXIT
      END;
      pentry := entry
    END;
    RETURN mrc
  END LookupByName;


PROCEDURE Dumper(arg:Threads.ForkeeArg):Threads.ForkeeReturn;
  BEGIN
    LOOP
      Time();
      WriteF.WriteF0(SysStreams.sysOut, " Oghma running\n");
      Timers.WaitSecs(60, Timers.FromNow)
    END;
    RETURN Threads.ForkeeReturn(NIL)
  END Dumper;


PROCEDURE Log(
      type: ARRAY OF CHAR; 
      mrc:  RPC.MemberRC; 
      name: String; 
      place:RPC.Place
 );
  BEGIN
    Time();
    WriteF.WriteF2(SysStreams.sysOut, 
                   " %s7 %s8", WriteF.H(type), WriteF.S(name));
    WriteF.WriteF4(SysStreams.sysOut, 
                   " %x2 %x8 %x8 %x8", 
                   place.location.word[0], 
                   place.location.word[1], 
                   place.location.word[2], 
                   place.service);
    WriteF.WriteF1(SysStreams.sysOut, " [%x]\n", ORD(mrc))
  END Log;


PROCEDURE Time();
  VAR
    ds,
    ts:ARRAY [0..15] OF CHAR;
  BEGIN
    TimeDate.DateString(TimeDate.Date(), ds);
    TimeDate.TimeString(TimeDate.Time(), ts);
    WriteF.WriteF3(SysStreams.sysOut, 
                   "%s %s", WriteF.H(ds), WriteF.H(ts));
  END Time;


PROCEDURE Init();
  VAR
    i:     INTEGER;
    thread:Threads.Thread;
  BEGIN
    FOR i := 0 TO TableTop DO
      nameTable[i]  := NIL
    END;
    Threads.InitMutex(tableLock);
    thread := Threads.Fork(Dumper, Threads.ForkeeArg(NIL))
  END Init;


END NetMember.
