(*
    Title: 	Server Stub for NetMember
    LastEdit:	"Fri Jan  8 15:01:06 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE SSNetMember;


IMPORT
  Binder, Exceptions, Marshal, NetMember, RPC, Strings, SYSTEM, Threads;   

FROM SystemTypes IMPORT String;


PROCEDURE FindStub(m:Marshal.T);
  VAR
    name: String;
    dist: RPC.Distance;
    place:RPC.Place;
    mrc:  RPC.MemberRC;
  BEGIN
    Marshal.AString(name, m);
    Marshal.AWord(dist, m);
    IF RPC.EndOfArgs(m) = RPC.OK THEN
      mrc := NetMember.Find(name, dist, place);
      Marshal.AWord(mrc, m);
      IF mrc = RPC.BindOK THEN RPC.MarshalPlace(place, m) END
    ELSE
      RPC.SetReturnCode(RPC.ArgError, m)
    END;
    Strings.Dispose(name)
  END FindStub;


PROCEDURE OfferStub(m:Marshal.T);
  VAR
    mrc:  RPC.MemberRC;
    name: String;
    dist: RPC.Distance;
    place:RPC.Place;
  BEGIN
    Marshal.AString(name, m);
    Marshal.AWord(dist, m);
    RPC.MarshalPlace(place, m);
    IF RPC.EndOfArgs(m) = RPC.OK THEN
      Marshal.ValueWord(NetMember.Offer(name, dist, place), m);
    ELSE
      RPC.SetReturnCode(RPC.ArgError, m)
    END;
    Strings.Dispose(name)
  END OfferStub;


PROCEDURE RetractStub(m:Marshal.T);
  VAR
    name: String;
    place:RPC.Place;
    dist: RPC.Distance;
  BEGIN
    Marshal.AString(name, m);
    RPC.MarshalPlace(place, m);
    Marshal.AWord(dist, m);
    IF RPC.EndOfArgs(m) = RPC.OK THEN
      NetMember.Retract(name, place, dist)
    ELSE
      RPC.SetReturnCode(RPC.ArgError, m)
    END;
    Strings.Dispose(name)
  END RetractStub;  


PROCEDURE FindAtStub(m:Marshal.T);
  VAR
    mrc:  RPC.MemberRC;
    name: String;
    place:RPC.Place;
  BEGIN
    Marshal.AString(name, m);
    IF RPC.EndOfArgs(m) = RPC.OK THEN
      mrc := NetMember.FindAt(place, name, place);
      Marshal.AWord(mrc, m);
      IF mrc = RPC.BindOK THEN RPC.MarshalPlace(place, m) END
    ELSE
      RPC.SetReturnCode(RPC.ArgError, m)
    END;
    Strings.Dispose(name)
  END FindAtStub;


PROCEDURE Receiver(p:RPC.PktBuf):RPC.PktBuf;
  VAR
    m:   Marshal.T;
    rc:  RPC.RC;
    proc:CARDINAL;
  BEGIN
    rc := RPC.ServerProcedureVCheck(p, m, proc, NetMember.Version);
    IF rc = RPC.OK THEN
      CASE NetMember.Procedures(proc) OF
        NetMember.FindProc:    FindStub(m)
      | NetMember.OfferProc:   OfferStub(m)
      | NetMember.RetractProc: RetractStub(m)
      | NetMember.FindAtProc:  FindAtStub(m)
      ELSE                     RPC.SetReturnCode(RPC.ProcError, m)
      END
    ELSE
      RPC.SetReturnCode(rc, m)
    END;
    RETURN RPC.EndOfCall(m)
  END Receiver;


PROCEDURE Offer();
  VAR
    name:String;
    svId:RPC.ServiceId;
  BEGIN
    name := Strings.CopyCS(Binder.ServiceName);
    IF (RPC.Offer(name, RPC.SameNet, Receiver, 0, svId) # RPC.BindOK)
    OR (NOT RPC.OfferId(Binder.ServiceId, Receiver, 0)) THEN
      Exceptions.RAISE("Unable to offer Oghma")
    END;
    Strings.Dispose(name)
  END Offer;


END SSNetMember.
