(*
    Title: 	Marshalling code
    LastEdit:	"Wed Jan 27 14:15:06 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE Marshal;

IMPORT
  Arithmetic, Endian, Exceptions, MCLib, MCNetLib, PktBufs, 
  RPCBasics, Strings;

FROM Storage     IMPORT ALLOCATE, DEALLOCATE;
FROM SYSTEM      IMPORT ADR, WORD, ADDRESS, BITSPERWORD, BYTESPERWORD, SIZE;
FROM SystemTypes IMPORT String;


PROCEDURE New(p:PktBufs.T):T;
  VAR
    m:T;
  BEGIN
    NEW(m);
    WITH m^ DO
      iPktBuf := p;
      cPktBuf := p;
      ptr := 0;
      end := m^.cPktBuf^.size;
      swapEndian := (p^.protocolHdr.endian # Endian.Ours);
    END; (* if *)
    RETURN m;
  END New;


PROCEDURE Start(m:T);
  BEGIN
    WITH m^ DO
      ptr := 0;
      end := m^.iPktBuf^.size
    END
  END Start;


PROCEDURE EndOfArgs(m:T):BOOLEAN;
  (* Only relevant if unmarshal *)
  BEGIN
    WITH m^ DO
      RETURN (marshal) OR
             (cPktBuf^.top = ptr * BYTESPERWORD)
    END
  END EndOfArgs;


PROCEDURE Discard(m:T);
  BEGIN
    DISPOSE(m);
  END Discard;


PROCEDURE AWord(VAR x:WORD; m:T);
  VAR
    y:WORD;
  BEGIN
    WITH m^ DO
      IF ptr = end THEN NewPktBuf(m) END;
      WITH cPktBuf^ DO
        IF marshal THEN
          IF swapEndian THEN 
            data[ptr] := MCNetLib.ChangeWordSex(x)
          ELSE
            data[ptr] := x
          END;
          top := top + BYTESPERWORD
        ELSE
          x := data[ptr];
          IF swapEndian THEN x := MCNetLib.ChangeWordSex(x) END
        END
      END;
      INC(ptr)
    END
  END AWord;


PROCEDURE AChar(VAR c:CHAR; m:T);
  VAR
    card:CARDINAL;
  BEGIN
    IF m^.marshal THEN
      card := ORD(c);
      AWord(card, m)
    ELSE
      AWord(card, m);
      c := CHR(card)
    END
  END AChar;


PROCEDURE Words(start:ADDRESS; n:CARDINAL; m:T);
  BEGIN
    Bytes(start, n, m, m^.swapEndian)
  END Words;


PROCEDURE Bytes(start:ADDRESS; n:CARDINAL; m:T; swap:BOOLEAN);
  VAR
    i,j:CARDINAL;
    to,
    from:ADDRESS;
  BEGIN
    WITH m^ DO
      WHILE n # 0 DO
        i := Arithmetic.CMin(n, end - ptr);
        IF i = 0 THEN 
          NewPktBuf(m);
          i := Arithmetic.CMin(n, end - ptr)            
        END;
        IF marshal THEN
          from := start; to := ADR(cPktBuf^.data[ptr]);
          INC(cPktBuf^.top, i * BYTESPERWORD)
        ELSE
          from := ADR(cPktBuf^.data[ptr]); to := start
        END;
        IF swap THEN 
          MCNetLib.ChangeBufferSex(from, to, i);
        ELSE  
          MCLib.MoveBytes(from, to, i * BYTESPERWORD)
        END;
        n := n - i;
        start := start + i * BYTESPERWORD;
        ptr := ptr + i
      END
    END
  END Bytes;


PROCEDURE Chars(start:ADDRESS; n:CARDINAL; m:T);
(*This is really for marshalling arrays and strings*)
(*Should this be dealing with non-aligned strings? Yes*)
  BEGIN
    Bytes(start, (n+3) DIV 4, m, FALSE);
  END Chars;


PROCEDURE AString(VAR x:String; m:T);
  VAR
    l:CARDINAL;
  BEGIN
    IF m^.marshal THEN
      l := Strings.LengthS(x);
      AWord(l, m);
      Chars(x, l, m)
    ELSE
      AWord(l, m);
      x := Strings.New(l);
      Chars(x, l, m);
      x^[l] := 0C
    END
  END AString;


PROCEDURE AnArrayOfWord(
  VAR x:ARRAY OF WORD;
  VAR n:CARDINAL;
      m:T
 );
  VAR
    a:CARDINAL;
  BEGIN
    a := HIGH(x) + 1;
    AWord(n, m);
    IF n > a THEN
      IF m^.marshal THEN
        n := a
      ELSE
        Exceptions.RAISE("UnMarshalError")
      END
    END;
    Words(ADR(x), n, m)
  END AnArrayOfWord;


PROCEDURE AnArrayOfChar(VAR x:ARRAY OF CHAR;
                             VAR n:CARDINAL;
                                 m:T);
  VAR
    a:CARDINAL;
  BEGIN
    a := HIGH(x) + 1;
    AWord(n, m);
    IF n > a THEN
      IF m^.marshal THEN
        n := a
      ELSE
        Exceptions.RAISE("UnMarshalError")
      END
    END;
    Chars(ADR(x), n, m)
  END AnArrayOfChar;


PROCEDURE NewPktBuf(m:T);
  BEGIN
    WITH m^ DO
      IF marshal THEN
        cPktBuf^.next := PktBufs.New();
        cPktBuf := cPktBuf^.next
      ELSE
        cPktBuf := cPktBuf^.next;
        IF cPktBuf = NIL THEN Exceptions.RAISE("UnMarshalError") END
      END;
      ptr := 0;
      end := cPktBuf^.size
    END
  END NewPktBuf;


PROCEDURE ValueWord(x:WORD; m:T);
  VAR
    xx:WORD;
  BEGIN
    IF (NOT m^.marshal) THEN Exceptions.RAISE("UnMarshalError") END;
    xx := x;
    AWord(xx, m)
  END ValueWord;


PROCEDURE UnWord(m:T):WORD;
  VAR
    xx:WORD;
  BEGIN
    IF m^.marshal THEN Exceptions.RAISE("MarshalError") END;
    AWord(xx, m);
    RETURN xx
  END UnWord;


END Marshal.
