(*
    Title: 	Things to chain together
    LastEdit:	"Fri Jan 29 19:27:11 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE PktBufs;

IMPORT
  Endian, MCLib, MCNetLib, RPCBasics, SYSTEM, Threads;

FROM Storage     IMPORT ALLOCATE, DEALLOCATE;

FROM SYSTEM      IMPORT ADR, BYTESPERWORD, TSIZE, WORD;


VAR
  dummyLoc:RPCBasics.Location;

  pktLock:Threads.Mutex; 
  pktList:T;


PROCEDURE New():T;
  VAR
    p:    T;
    allfs:CARDINAL;
  BEGIN
    p := Newish(Size);
    WITH p^ DO
      top := 0;
      data[0] := WORD(RPCBasics.OK);
      WITH protocolHdr DO
        endian := Endian.Ours;
        pktrate := 0;
        seqNumber := 1;
        maxNumber := 1
      END;
      location[RPCBasics.ClientEnd] := dummyLoc;
      location[RPCBasics.ServerEnd] := dummyLoc;
      serviceId := 0FFFFFFFFH;
      thread := 0FFFFFFFFH;
      callId := 0;
      converseId := 0
    END;
    RETURN p
  END New;


PROCEDURE Newish(wordsofdata:CARDINAL):T;
  VAR
    p:T;    
  BEGIN
    p := NIL;
    IF wordsofdata = Size THEN
      Threads.Acquire(pktLock);
        p := pktList;
        IF p # NIL THEN pktList := p^.next END;
      Threads.Release(pktLock)
    ELSIF wordsofdata = 0 THEN
      wordsofdata := 1 (* Space for Return Code *)
    END;
    IF p = NIL THEN ALLOCATE(p, TSIZE(WORD) * (wordsofdata + recSize)) END;
    WITH p^ DO
      next := NIL;
      size := wordsofdata
    END;
    RETURN p
  END Newish;


PROCEDURE Extra(p: T); (*RAISES NotEndOfChain*)
(*Chains another on the end*)
  BEGIN
    IF p^.next # NIL THEN
      (*RAISE NotEndOfChain*)
      RETURN;
    END; (* if *)
    p^.next := New();
  END Extra;


PROCEDURE Discard(VAR p: T (* InOut *));
(*Discards, with all followers*)
  VAR
    q:T;
  BEGIN
    q := p;
    WHILE p # NIL DO
      q := p^.next;
      IF p^.size = ByteSize THEN
        Threads.Acquire(pktLock);
          p^.next := pktList;
          pktList := p;
        Threads.Release(pktLock)
      ELSE
        DEALLOCATE(p, TSIZE(WORD) * (p^.size + recSize))
      END;
      p := q
    END
  END Discard;


PROCEDURE Copy(p: T): T;
(*Copies an entire chain*)
  VAR
    q,r:T;
  BEGIN
    IF p = NIL THEN
      RETURN NIL
    ELSE
      q := Newish(p^.top);
      MCLib.MoveBytes(ADR(p^.location), ADR(q^.location), headBSize+p^.top);
      q^.protocolHdr := p^.protocolHdr;
      r := q;
      WHILE p^.next # NIL DO
        p := p^.next;
        q^.next := Newish(p^.top);
        q := q^.next;
        MCLib.MoveBytes(ADR(p^.location), ADR(q^.location), headBSize+p^.top);
        q^.protocolHdr := p^.protocolHdr
      END;
      RETURN r
    END
  END Copy;


PROCEDURE CopyHeader(from, to:T);
  BEGIN
    MCLib.MoveBytes(ADR(from^.location), ADR(to^.location), headBSize);
    to^.protocolHdr := from^.protocolHdr
  END CopyHeader;


PROCEDURE FetchDataWORD(p:T; offset:CARDINAL):WORD;
  BEGIN
    IF p^.protocolHdr.endian # Endian.Ours THEN
      RETURN WORD(MCNetLib.ChangeWordSex(p^.data[offset]))
    ELSE
      RETURN p^.data[offset]
    END
  END FetchDataWORD;


(*
TYPE
  Packed = 
    RECORD
      pack:BITS 4 FOR PacketType;
      prot:BITS 3 FOR ProtocolType;
      endi:BITS 1 FOR Endian.Sex;
      rate:BITS 8 FOR PacketRate;
      seqN:BITS 8 FOR SequenceNumber;
      maxN:BITS 8 FOR SequenceNumber
    END;
*)    
TYPE
  FourChars = POINTER TO ARRAY [0..3] OF CHAR;
  
PROCEDURE PackHeader(p:T);
  VAR
    cp:FourChars;
  BEGIN
    cp := SYSTEM.ADR(p^.packedHdr);
    WITH p^.protocolHdr DO
(*
      WITH w DO
        pack := packet;
        prot := protocol;
        endi := endian;
        rate := pktrate;
        seqN := seqNumber;
        maxN := maxNumber
      END
*)
      cp^[0] := CHAR(CARDINAL(packet) + 
                     16 * (CARDINAL(protocol) + 
                           8 * CARDINAL(endian)));
      cp^[1] := CHAR(pktrate);
      cp^[2] := CHAR(seqNumber);
      cp^[3] := CHAR(maxNumber)
    END
  END PackHeader;


PROCEDURE UnPackHeader(p:T);
  VAR
    w: CARDINAL;
    cp:FourChars;
  BEGIN
    cp := SYSTEM.ADR(p^.packedHdr);
    WITH p^.protocolHdr DO
(*
      WITH w DO
        packet := pack;
        protocol := prot;
        endian := endi;
        pktrate := rate;
        seqNumber := seqN;
        maxNumber := maxN
      END;
*)
      w := ORD(cp^[0]);
      packet := RPCBasics.PacketType(w MOD 16);
      protocol := RPCBasics.ProtocolType((w DIV 16) MOD 8);
      endian := Endian.Sex(w DIV 128);
      pktrate := RPCBasics.PacketRate(cp^[1]);
      seqNumber := RPCBasics.SequenceNumber(cp^[2]);
      maxNumber := RPCBasics.SequenceNumber(cp^[3]);
     END
  END UnPackHeader;


PROCEDURE Init();
  VAR
    pkt:T;
  BEGIN
    WITH dummyLoc DO
      word[0] := 0FFFFFFFFh;
      word[1] := 0FFFFFFFFh;
      word[2] := 0FFFFFFFFh;
    END;
    Threads.InitMutex(pktLock);
    pktList := NIL;
    pkt := T(0);
    headBSize := ADR(pkt^.data) - ADR(pkt^.location);
    headSize := headBSize DIV BYTESPERWORD;
    recBSize := ADR(pkt^.data) - ADR(pkt^.next);
    recSize := recBSize DIV BYTESPERWORD
  END Init;


BEGIN
  Init()
END PktBufs.
