(*
    Title: 	Server end of things
    LastEdit:	"Thu Feb 25 18:04:36 1988"
    Author: 	Derek McAuley,Oms1,4652
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE RPCCalls;


IMPORT
  NetMember, PktBufs, RPCNaff, RPCNets, RPCReceive, RPCStates,
  SYSTEM, Threads, ThreadsExtra, Timers, RPCTable;

FROM Threads   IMPORT Thread;
FROM RPCBasics IMPORT
  CallEnd, CallId, RC, Instance, PacketType, ProtocolType;


CONST
  MinimumReceivers =  2; (* Leave at least two receivers *)
  KillingLimit     =  6; (* If there are this many redundant, kill some *)
  RPCThreads       =  4; (* Default receiver threads *)
  MaximumReceivers = 16; (* Never fork more than this *)
  OverseerTimeout  = 10; (* Seconds *)

  ReplyRetryLimit  = 16;      (* Number of times to try to reply *)
  ReplyTimer       = 3;       (* Seconds between replies *)


TYPE
  RecStats =
    RECORD
      total,
      normal,
      except,
      broken:CARDINAL
    END;


VAR
  minReceivers:  CARDINAL;      (* Minimum freeReceivers in last cycle *)
  freeReceivers: CARDINAL;      (* The number free (i.e. not in a call) *)
  totalReceivers:CARDINAL;      (* The total number of receivers *)
  lastTime:      CARDINAL;
  receiversLock: Threads.Mutex; (* Lock on all above *)


  recStats:      RecStats;


PROCEDURE Receiver(arg:Threads.ForkeeArg):Threads.ForkeeReturn;
  VAR
    p,q:    PktBufs.T;
    type:   CallEnd;
    die:    BOOLEAN;
    time:   Timers.Value;
    state,
    pstate: RPCStates.T;
    congest:BOOLEAN;
  BEGIN
    ThreadsExtra.SetMyPriority(ThreadsExtra.DefPriority - 1);
    state := RPCStates.New();
    die := FALSE;
    WITH state^ DO
      REPEAT
        p := RPCReceive.NewCall(pstate); (* NewCallLock is held *)
        INC(recStats.total);
        p^.next := NIL;
        RPCStates.CopyOutOfPkt(p, state);

        IF (pstate # NIL) AND (p^.protocolHdr.protocol = ExactlyOnce) THEN
          q := PktBufs.Newish(0);
          PktBufs.CopyHeader(p, q);
          q^.callId := pstate^.sCallId;
          q^.protocolHdr.packet := TermPkt; (* Prods other thread *)
          ReceivedPkt(q, pstate)
        END;

        Threads.Acquire(receiversLock);
          congest := (freeReceivers <= MinimumReceivers);
          IF congest THEN
            IF totalReceivers < MaximumReceivers THEN
              ForkAnother();
              INC(totalReceivers);
              congest := FALSE
            END
          ELSE
            DEC(freeReceivers)
          END;
        Threads.Release(receiversLock);

        IF congest THEN
          RPCStates.SendADeadOne(state, Congested);
          PktBufs.Discard(p);
          RPCReceive.NewCallRelease()
        ELSE
          DoCall(state, p);

          RPCStates.DisposeAndNew(state);

          Threads.Acquire(receiversLock);
         (* minReceivers := Arithmetic.CMin(minReceivers, freeReceivers);
            Timers.Read(time);
            IF time.secs > lastTime + OverseerTimeout THEN
              die := (minReceivers > (KillingLimit + MinimumReceivers));
              minReceivers := totalReceivers;
              lastTime := time.secs
            END;
            IF die THEN DEC(totalReceivers) ELSE INC(freeReceivers) END; *)
            INC(freeReceivers);
          Threads.Release(receiversLock);
        END;
      UNTIL die
    END;
    RPCStates.Dispose(state);
    RETURN Threads.ForkeeReturn(NIL)
  END Receiver;


PROCEDURE DoCall(
  state:RPCStates.T;
  p:    PktBufs.T
 );
  TYPE
    PROCP = PROCEDURE(PktBufs.T);
  VAR
    r:    PktBufs.T;
    at:   PktBufs.ReceiveProc;
    proc: PROC;
    procp:PROCP;
  BEGIN
    WITH state^ DO
      Threads.Acquire(sLock);
        sArguments := p;
        sStatus := RPCStates.ServerExecute;

        RPCReceive.Add(ServerEnd, state);
      Threads.Release(sLock);
  
      IF NOT RPCTable.FindId(sArguments^.serviceId, at, sArguments^.i) THEN
        RPCStates.SendADeadOne(state, UnknownService);
        RPCReceive.Remove(state);
        RETURN
      END;
      sResults := NIL;
      ThreadsExtra.SetMyPriority(ThreadsExtra.DefPriority);
        sResults := at(sArguments);
      ThreadsExtra.SetMyPriority(ThreadsExtra.DefPriority - 2);
      PktBufs.Discard(sArguments);

      Threads.Acquire(sLock);
        IF sResults = NIL THEN
          INC(recStats.broken);
          RPCStates.SendADeadOne(state, DomainUnknown)
        ELSE
          IF (sResults^.next # NIL) THEN
            (* "Hrrrmm.", "I don't recall playing any 'Hrrrmm'" *)
            INC(recStats.broken);
            RPCStates.SendADeadOne(state, ReplyTooBig)
          ELSE
            RPCStates.CopyInToPkt(state, state^.sResults);
            state^.sResults^.protocolHdr.packet := ReplyPkt;
            IF RPCNets.Send(ClientEnd, state^.sResults) = OK THEN
              IF sProtocol = ExactlyOnce THEN TerminateEO(state) END;
              IF ORD(UserException) 
               = CARDINAL(PktBufs.FetchDataWORD(sResults, 0)) THEN
                INC(recStats.except)
              ELSE
                INC(recStats.normal)
              END
            ELSE
              INC(recStats.broken)
            END
          END;
          PktBufs.Discard(sResults)
        END;
        RPCReceive.Remove(state);
      Threads.Release(sLock);
      ThreadsExtra.SetMyPriority(ThreadsExtra.DefPriority - 1)
    END
  END DoCall;


PROCEDURE TerminateEO(state:RPCStates.T);
  VAR
    retry:CARDINAL;
  BEGIN
    WITH state^ DO
      retry := ReplyRetryLimit;
      REPEAT
        DEC(retry);
        state^.sStatus := RPCStates.ServerTerminate;
        RPCNaff.WaitTimeOut(sLock, sCondition, ReplyTimer);
        IF sUnknown # NIL THEN
          CASE sUnknown^.protocolHdr.packet OF
            CallPkt, AlivePkt:
              PktBufs.Discard(sUnknown) 
          ELSE
              PktBufs.Discard(sUnknown);
              RETURN (* Either a TermPkt or protocol error, finish anyway *)
          END;
        END;
      UNTIL (RPCNets.Send(ClientEnd, state^.sResults) # OK) OR (retry = 0)
    END
  END TerminateEO;


PROCEDURE ReceivedPkt(
  p:    PktBufs.T;
  state:RPCStates.T
 );
  BEGIN
    Threads.Acquire(state^.sLock);
      CASE state^.sStatus OF
        RPCStates.ServerExecute:
          CASE p^.protocolHdr.packet OF
            CallPkt, AlivePkt:
              PktBufs.Discard(p);
              RPCStates.SendAliveOne(ClientEnd, state)
          ELSE
              RPCStates.SignalPacket(state, p)
          END
      | RPCStates.ServerTerminate:
          RPCStates.SignalPacket(state, p)
      ELSE
          PktBufs.Discard(p)
      END;
    Threads.Release(state^.sLock)
  END ReceivedPkt;


PROCEDURE Calls(
  VAR total: CARDINAL; (* Out *)
  VAR normal:CARDINAL; (* Out *)
  VAR except:CARDINAL; (* Out *)
  VAR broken:CARDINAL; (* Out *)
  VAR active:CARDINAL; (* Out *)
      reset: BOOLEAN
 );
  BEGIN
    total := recStats.total;
    normal := recStats.normal;
    except := recStats.except;
    broken := recStats.broken;
    active := totalReceivers - freeReceivers;
    IF reset THEN ResetCalls() END;
  END Calls;


PROCEDURE ResetCalls();
  BEGIN
    recStats.total := 0;
    recStats.normal := 0;
    recStats.except := 0;
    recStats.broken := 0
  END ResetCalls;


PROCEDURE ForkAnother();
  VAR
    thread:Thread;
  BEGIN
    thread := Threads.Fork(Receiver, Threads.ForkeeArg(NIL));
    (* ThreadsExtra.NoJoinNeeded(thread) *)
  END ForkAnother;


PROCEDURE Init();
  VAR
    i:  CARDINAL;
    now:Timers.Value;
  BEGIN
    ResetCalls();
    Threads.InitMutex(receiversLock);
    Timers.Read(now);
    lastTime := now.secs;
    freeReceivers := RPCThreads;
    totalReceivers := RPCThreads;
    minReceivers := RPCThreads;
    FOR i := 1 TO RPCThreads DO ForkAnother() END
  END Init;


END RPCCalls.
