(*
    Title: 	Network interface stuff
    LastEdit:	"Thu Feb 25 18:12:15 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE RPCNets;


IMPORT
  Cfr, Endian, Exceptions, Fields, MCNetLib, PktBufs, RPCBasics, SYSTEM, 
  Threads, Timers, UDP, UnixMCI;
IMPORT
  SysStreams, WriteF;


VAR
  txSocket: UDP.Socket;
  rxSocket: UDP.Socket;
  udpHandle:UDP.Handle;
  
  udpRecVCount,
  cfrRecVCount:INTEGER;


PROCEDURE Send(
      theirEnd:RPCBasics.CallEnd; 
      pkt:     PktBufs.T
 ):RPCBasics.RC;
  (* Must not destroy PktBufs.T *)
  VAR
    q:PktBufs.T;
    sock:UDP.Socket;
    rc:  UDP.SendRC;
    top: CARDINAL;
    call:RPCBasics.CallId;
  BEGIN
    IF pkt = NIL THEN RETURN RPCBasics.OK END;
    CASE RPCBasics.NetworkType(ORD(pkt^.location[theirEnd].byte[0])) OF
      RPCBasics.UdpIpNet: (* In line *)
    | RPCBasics.UdlNet:
        RETURN Cfr.Send(theirEnd, pkt) 
    ELSE
        RETURN RPCBasics.MachineAdUnknown
    END;
 (* WriteOutPkt(pkt, "Send"); *)

    WITH pkt^.location[theirEnd] DO
      sock.addr := machineId;
      sock.port := Fields.ExtractShort(udpPort, 0)
    END;
 (* WriteF.WriteF1(SysStreams.sysOut, "Send to port &%x\n", sock.port); *)

    PktBufs.PackHeader(pkt);
    IF Endian.Ours = Endian.Big THEN
      call := pkt^.callId;
      top := pkt^.top;
      pkt^.callId := RPCBasics.CallId(MCNetLib.ChangeWordSex(call));
      pkt^.top := CARDINAL(MCNetLib.ChangeWordSex(top));
      rc := UDP.Send(sock, udpHandle, SYSTEM.ADR(pkt^.location), 
                     top + PktBufs.headBSize);
      pkt^.callId := call;
      pkt^.top := top
    ELSE
      rc := UDP.Send(sock, udpHandle, SYSTEM.ADR(pkt^.location), 
                     pkt^.top + PktBufs.headBSize);
    END;
    IF rc < 0 THEN
      WriteF.WriteF1(SysStreams.sysOut, "Bad send rc = &%x\n", -rc);
      RETURN RPCBasics.MachineAdUnknown
    END;
    RETURN RPCBasics.OK
  END Send;
           

PROCEDURE RecV():PktBufs.T;
  VAR
    p:   PktBufs.T;
    sock:UDP.Socket;
    len: CARDINAL;
    rc:  UDP.RecVRC;
  BEGIN
    IF (Cfr.isActive) AND (cfrRecVCount < udpRecVCount) THEN
      INC(cfrRecVCount);
      p := Cfr.RecV();
      DEC(cfrRecVCount);
      RETURN p
    END;
    INC(udpRecVCount);
    p := PktBufs.New();
    REPEAT
      sock := rxSocket;
      len := p^.size + PktBufs.headBSize;
      rc := UDP.RecV(udpHandle, sock, SYSTEM.ADR(p^.location), len);
      IF rc >=0  THEN
        IF Endian.Ours = Endian.Big THEN
          p^.callId := RPCBasics.CallId(MCNetLib.ChangeWordSex(p^.callId));
          p^.top := CARDINAL(MCNetLib.ChangeWordSex(p^.top))
        END;
        IF (len < p^.top + PktBufs.headBSize) THEN 
          WriteF.WriteF2(SysStreams.sysOut, "Len %d, top %d\n", len, p^.top);
          rc := -1
        END
   (* ELSE
        WriteF.WriteF1(SysStreams.sysOut, "RecV failed, rc = &%x\n", rc) *)
      END
    UNTIL rc >= 0;
       
    PktBufs.UnPackHeader(p);

 (* WriteOutPkt(p, "RecV"); *)
 (* WriteF.WriteF0(SysStreams.sysOut, "RecV from port %x\n", sock.port); *)

    DEC(udpRecVCount);
    RETURN p
  END RecV;
                       

PROCEDURE SetErrorRate(dropeveryonein:CARDINAL);
  BEGIN
  END SetErrorRate;


PROCEDURE WriteOutPkt(p:PktBufs.T; who:ARRAY OF CHAR);
  BEGIN
    WriteF.WriteF1(SysStreams.sysOut, "to   : %x8\n", p^.serviceId);
    WriteF.WriteF1(SysStreams.sysOut, "id   : %x8\n", p^.callId);
    WriteF.WriteF1(SysStreams.sysOut, "pkt  : %x8\n", INTEGER(p^.protocolHdr.packet));
    WriteF.WriteF1(SysStreams.sysOut, "from : %x8\n", p^.thread)
  END WriteOutPkt;


PROCEDURE SetAddress(
  VAR location:RPCBasics.Location;
  VAR socket:  UDP.Socket;
      ipaddr:  UDP.Address
 );
  BEGIN
    WITH location DO
      Fields.InsertByte(udpType, 0, ORD(RPCBasics.UdpIpNet));
      Fields.InsertShort(udpPort, 0, rpcPort);
      machineId := ipaddr
    END;
    WITH socket DO
      addr := ipaddr;
      port := rpcPort
    END
  END SetAddress;


PROCEDURE Init(port:CARDINAL);
  VAR
    ouraddr:UDP.Address;
    p:      PktBufs.T;
    foo:    BOOLEAN;
    time:   Timers.Value;
  BEGIN
    ouraddr := UDP.ReadIPAddress();
    IF port # 0 THEN
      rpcPort := MCNetLib.HToNS(port);
      SetAddress(thisLocation, txSocket, ouraddr);
      udpHandle := UDP.ReserveSocket(txSocket)
    ELSE
      udpHandle := UDP.ReserveAnySocket(txSocket);
      rpcPort := txSocket.port;
      SetAddress(thisLocation, txSocket, ouraddr)
    END;
    IF udpHandle < 0 THEN
      WriteF.WriteF2(SysStreams.sysErr, 
                     "No Socket, rc %i, err %i\n", 
                     udpHandle, UnixMCI.errnoPtr^);
      Exceptions.RAISE("No Socket")
    END;
    SetAddress(wildLocation, rxSocket, UDP.ReadWildIPAddress());
    Cfr.Init();
    cfrRecVCount := 0;
    udpRecVCount := 0
  END Init;

END RPCNets.
