(*
    Title: 	De-multiplex packets
    LastEdit:	"Thu Feb 25 18:06:24 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE RPCReceive;

                 
IMPORT
  Endian, MCNetLib, PktBufs, RPC, RPCCalls, RPCNets, 
  RPCRemote, RPCStates, SYSTEM, Threads;

FROM Storage   IMPORT ALLOCATE, DEALLOCATE;
FROM Threads   IMPORT Thread;

FROM RPCBasics IMPORT
  CallEnd, CallId, RC, Instance, Location, PacketType, ServiceId, ThreadHand;


CONST
  TableSize     = 32;


TYPE
  ActiveCall = POINTER TO ActiveCallR;
  ActiveCallR =
    RECORD
      next:    ActiveCall; (* First field *)
      callEnd: CallEnd;
      rpcState:RPCStates.T
    END;
  ActiveTable = ARRAY [0..TableSize-1] OF ActiveCall;
  
  LookResult  = (LRFound, LRFoundPrev, LRNotFound);


VAR
  activeTable:ActiveTable;
  tableLock:  Threads.Mutex;
  activeList: ActiveCall;
  newCallLock:Threads.Mutex;


PROCEDURE FoundActiveCall(
      call:  CallId;
      thread:ThreadHand;
      client:Location;
  VAR act:   ActiveCall (* Out *)
 ):LookResult;
  VAR
    lact: ActiveCall;
    ocall:CallId;
  BEGIN
    lact := activeTable[Hash(thread)];
    act := NIL;
    WHILE lact # NIL DO
      WITH lact^.rpcState^ DO
        IF (sThreadId = thread)
        AND (RPC.SameLocation(client, sLocation[ClientEnd])) THEN
          act := lact;
          IF sCallId = call THEN RETURN LRFound END
        END
      END;
      lact := lact^.next
    END;
    IF act # NIL THEN 
      ocall := act^.rpcState^.sCallId;
      IF  ((ocall DIV 1000000H) = (call DIV 1000000H))
      AND ((call - ocall + 1000000H) MOD 1000000H < 10000H) THEN
        RETURN LRFoundPrev 
      END
    END;
    RETURN LRNotFound
  END FoundActiveCall;


PROCEDURE InsertCall(type:CallEnd; state:RPCStates.T);
  VAR
    nact:ActiveCall;
    hash:CARDINAL;
  BEGIN
    hash := Hash(state^.sThreadId);
    nact := activeList;
    IF nact = NIL THEN NEW(nact) ELSE activeList := nact^.next END;
    WITH nact^ DO
      next := activeTable[hash];
      callEnd := type;
      rpcState := state
    END;
    activeTable[hash] := nact
  END InsertCall;


PROCEDURE Add(type:CallEnd; state:RPCStates.T);
  BEGIN
    Threads.Acquire(tableLock);
      InsertCall(type, state);
    Threads.Release(tableLock);
    IF type = ServerEnd THEN Threads.Release(newCallLock) END
  END Add;
                         

PROCEDURE Lookup(
      call:  CallId;
      thread:ThreadHand;
      client:Location;
  VAR type:  CallEnd;    (* Out *)
  VAR state: RPCStates.T (* Out *)
 ):LookResult;
  VAR
    act:  ActiveCall;
    found:LookResult;
  BEGIN
    Threads.Acquire(tableLock);
      found := FoundActiveCall(call, thread, client, act);
      CASE found OF
        LRFound, LRFoundPrev:
          WITH act^ DO
            type := callEnd;
            state := rpcState
          END
      ELSE
      END;
    Threads.Release(tableLock);
    RETURN found
  END Lookup;


PROCEDURE Remove(state:RPCStates.T);
  VAR
    act,
    pact:ActiveCall;
  BEGIN
    Threads.Acquire(tableLock);
      pact := SYSTEM.ADR(activeTable[Hash(state^.sThreadId)]);
      LOOP
        act := pact^.next;
        IF (act = NIL) THEN EXIT END;
        WITH act^ DO
          IF state = rpcState THEN
            pact^.next := act^.next;
            act^.next := activeList; 
            activeList := act;
            rpcState := NIL;
            EXIT
          END
        END;
        pact := act
      END;
    Threads.Release(tableLock)
  END Remove;


PROCEDURE NewCallRelease();
  BEGIN
    Threads.Release(newCallLock)
  END NewCallRelease;


PROCEDURE NewCall(VAR state:RPCStates.T (* Out *)):PktBufs.T;
  (* This calls an interested person but if none is found various actions
     may be taken depending on packet type, etc., etc., etc.. *)
  VAR
    p:     PktBufs.T;
    type:  CallEnd;
    callid:CallId;
  BEGIN
    LOOP
      state := NIL;
      p := RPCNets.RecV();
      Threads.Acquire(newCallLock);
      IF Lookup(p^.callId, p^.thread, 
                p^.location[ClientEnd], type, state) = LRFound THEN
        Threads.Release(newCallLock);
        relevantProcedure[type](p, state)
      ELSIF p^.protocolHdr.packet = CallPkt THEN
        RETURN(p) (* N.B. Lock still held - InsertCall will release *)
      ELSE
        Threads.Release(newCallLock);
        CASE p^.protocolHdr.packet OF
          AlivePkt, AliveAckPkt: (* There be some poor orphan out there *)
            RPCStates.ReplyWithADeadOne(p, NoSuchCall)
        | ReplyPkt: (* I've finished that call                     *)
            RPCStates.ReplyWithTerminator(p)
        ELSE
            PktBufs.Discard(p)
        END
      END
    END
  END NewCall;


PROCEDURE Hash(thread:ThreadHand):CARDINAL;
  BEGIN
    RETURN ((thread DIV 16) MOD TableSize)
  END Hash;


PROCEDURE Init();
  VAR
    i:CARDINAL;
  BEGIN
    FOR i := 0 TO TableSize-1 DO
      activeTable[i] := NIL
    END;
    activeList := NIL;
    Threads.InitMutex(tableLock);
    Threads.InitMutex(newCallLock);
    relevantProcedure[ClientEnd] := RPCRemote.ReceivedPkt;
    relevantProcedure[ServerEnd] := RPCCalls.ReceivedPkt
  END Init;


END RPCReceive.
