(*
    Title: 	Various random utility procedures
    LastEdit:	"Thu Feb 25 18:07:30 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE RPCStates;


IMPORT 
  PktBufs, RPCNets, SYSTEM, Threads;

IMPORT Storage; FROM Storage IMPORT ALLOCATE, DEALLOCATE;

FROM RPCBasics IMPORT
  CallEnd, CallId, RC, PacketType, ProtocolType, Place;


VAR
  stateChain:T;
  chainLock: Threads.Mutex;


PROCEDURE CopyInToPkt(state:T; p:PktBufs.T);
  BEGIN
    WITH p^ DO
      WITH state^ DO
        protocolHdr.protocol := sProtocol;
        location[ClientEnd] := sLocation[ClientEnd];
        location[ServerEnd] := sLocation[ServerEnd];
        serviceId := sServiceId;
        thread := sThreadId;
        callId := sCallId
      END
    END
  END CopyInToPkt;

  
PROCEDURE CopyOutOfPkt(p:PktBufs.T; state:T);
  BEGIN
    WITH p^ DO
      WITH state^ DO
        sProtocol := protocolHdr.protocol;
        sLocation[ClientEnd] := location[ClientEnd];
        sLocation[ServerEnd] := location[ServerEnd];
        sServiceId := serviceId;
        sThreadId := thread;
        sCallId := callId
      END
    END
  END CopyOutOfPkt;


(* Cache these things, as otherwise it's a lot of Threads.Init .... *)
PROCEDURE New():T;
  VAR
    state:T;
  BEGIN
    Threads.Acquire(chainLock);
      IF stateChain = NIL THEN
        NEW(state);
        WITH state^ DO
          Threads.InitMutex(sLock);
          Threads.InitCondition(sCondition)
        END
      ELSE
        state := stateChain;
        stateChain := stateChain^.sNext
      END;
    Threads.Release(chainLock);
    WITH state^ DO
      sNext := NIL;
      sArguments := NIL;
      sResults := NIL;
      sUnknown := NIL
    END;
    RETURN state
  END New;


PROCEDURE Dispose(state:T);
  BEGIN
    DisposeAndNew(state);
    Threads.Acquire(chainLock);
      state^.sNext := stateChain;
      stateChain := state;
    Threads.Release(chainLock)
  END Dispose;


PROCEDURE DisposeAndNew(state:T);
  BEGIN
    WITH state^ DO
      IF sUnknown # NIL THEN PktBufs.Discard(sUnknown) END;
    END
  END DisposeAndNew;


PROCEDURE SetupPDU(
  state:T;
  type: PacketType
 ):PktBufs.T;
  VAR
    p: PktBufs.T;
  BEGIN
    p := PktBufs.New();
    CopyInToPkt(state, p);
    WITH p^.protocolHdr DO
      packet := type;
      protocol := state^.sProtocol
    END;
    p^.converseId := 0;
    RETURN p
  END SetupPDU;


PROCEDURE ReplySetupPDU(
  p:   PktBufs.T;
  type:PacketType
 );
  VAR
    me:Place;
  BEGIN
    WITH p^ DO
      protocolHdr.packet := type;
      top := 0
    END
  END ReplySetupPDU;


PROCEDURE SendPDU(end:CallEnd; p:PktBufs.T);
  VAR
    cr:RC;
  BEGIN
    cr := RPCNets.Send(end, p);
    PktBufs.Discard(p)
  END SendPDU;


PROCEDURE SendTerminator(state:T);
  BEGIN
    SendPDU(ClientEnd, SetupPDU(state, TermPkt))
  END SendTerminator;


PROCEDURE ReplyWithTerminator(p:PktBufs.T);
  BEGIN
    ReplySetupPDU(p, TermPkt);
    SendPDU(ClientEnd, p)
  END ReplyWithTerminator;


PROCEDURE SendAliveOne(end:CallEnd; state:T);
  BEGIN
    IF end = ClientEnd THEN
      SendPDU(end, SetupPDU(state, AliveAckPkt))
    ELSE
      SendPDU(end, SetupPDU(state, AlivePkt))
    END
  END SendAliveOne;


PROCEDURE SendADeadOne(
  state:T;
  err:  RC
 );
  VAR
    p:PktBufs.T;
  BEGIN
    p := SetupPDU(state, DeadPkt);
    InsertDeadInfo(p, err);
    SendPDU(ClientEnd, p)
  END SendADeadOne;


PROCEDURE ReplyWithADeadOne(
  p:  PktBufs.T;
  err:RC
 );
  BEGIN
    ReplySetupPDU(p, DeadPkt);
    InsertDeadInfo(p, err);
    SendPDU(ClientEnd, p)
  END ReplyWithADeadOne;


PROCEDURE InsertDeadInfo(
  p:  PktBufs.T;
  err:RC
 );
  BEGIN
    p^.data[0] := SYSTEM.WORD(err);
    p^.top := 4
  END InsertDeadInfo;


PROCEDURE SignalPacket(
  state:T;
  pkt:  PktBufs.T
 );
  BEGIN
    WITH state^ DO
      IF sUnknown # NIL THEN PktBufs.Discard(sUnknown) END;
      sUnknown := pkt;
      Threads.Signal(sCondition)
    END
  END SignalPacket;


PROCEDURE Init();
  BEGIN
    stateChain := NIL;
    Threads.InitMutex(chainLock)
  END Init;


END RPCStates.
