IMPLEMENTATION MODULE UDP;


IMPORT
  Exceptions, Hosts, SYSTEM, Strings, Timers, UnixErrNo, 
  UnixIO, UnixSelect, UnixSocket;
  
FROM SystemTypes IMPORT String;

IMPORT
  SysStreams, WriteF;


CONST
  HostLength = 256;
  SocketRetryLimit = 100;
  
  
VAR
  noFlags: UnixSocket.Flags;
  hostId:  Address;
  ourSocks:BITSET;


PROCEDURE Send(
  to:    Socket;
  hand:  Handle;
  data:  SYSTEM.ADDRESS;
  length:CARDINAL
 ):SendRC;
  VAR
    sockAddr:UnixSocket.Addr;
  BEGIN
    ConvertToUnixSocket(to, sockAddr);
    IF NOT (CARDINAL(hand) IN ourSocks) THEN RETURN -1 END;
    UnixSelect.WaitForWrite(hand);
    RETURN UnixSocket.SendTo(hand, data, INTEGER(length), noFlags, sockAddr)
  END Send;


PROCEDURE ReserveSocket(sock:Socket):Handle;
  VAR
    fd:      UnixSocket.T;
    sockAddr:UnixSocket.Addr;
    i, n:    INTEGER;
  BEGIN
    fd := UnixSocket.Socket(UnixSocket.afInet, 
                            UnixSocket.sockDatagram, 
                            UnixSocket.afUnspecified);
    IF fd < 0 THEN RETURN fd END;
    ConvertToUnixSocket(sock, sockAddr);
    n := UnixSocket.Bind(fd, sockAddr);
    IF n < 0 THEN 
      IF UnixIO.Close(fd) < 0 THEN END;
      RETURN n 
    END;
    INCL(ourSocks, fd);
    RETURN Handle(fd)
  END ReserveSocket;


PROCEDURE ReserveAnySocket(VAR sock:Socket):Handle;
  VAR
    hand:   Handle;
    retry:  INTEGER;
    timeVal:Timers.Value;
  BEGIN
    Timers.Read(timeVal);
    sock.addr := hostId;
    sock.port := timeVal.usecs MOD 08000H;
    retry := 0;
    LOOP
      sock.port := 1024 + ((sock.port + 17) MOD 08000H);
      hand := ReserveSocket(sock);
      IF hand > 0 THEN RETURN hand END;
      INC(retry);
      IF retry = SocketRetryLimit THEN RETURN hand END;
    END
  END ReserveAnySocket;


PROCEDURE ReadWildIPAddress():Address;
  BEGIN
    RETURN Address(0)
  END ReadWildIPAddress;


PROCEDURE RecV(
      hand: Handle;
  VAR from: Socket;         (* Out *)
      data: SYSTEM.ADDRESS;
  VAR len:  CARDINAL        (* InOut *)
 ):RecVRC;
  VAR
    ilen:    INTEGER;
    fromAddr:UnixSocket.Addr;
  BEGIN
    IF NOT (CARDINAL(hand) IN ourSocks) THEN RETURN -1 END;
    UnixSelect.WaitForRead(hand);
    REPEAT
      ilen := UnixSocket.RecVFrom(hand, data, len, noFlags, fromAddr)
    UNTIL ilen # -UnixErrNo.EINTR;
    IF ilen < 0 THEN RETURN ilen END;
    from.port := fromAddr.port;
    from.addr := fromAddr.ipAddr;
    RETURN 0
  END RecV;
    

PROCEDURE ReadIPAddress():Address;
  BEGIN
    RETURN hostId
  END ReadIPAddress;


PROCEDURE ConvertToUnixSocket(VAR to:Socket; VAR sockaddr:UnixSocket.Addr);
  BEGIN
    WITH sockaddr DO
      family := UnixSocket.afInet;
      port := to.port;
      ipAddr := to.addr
    END
  END ConvertToUnixSocket;


BEGIN
  IF UnixSelect.ValidFD > 32 THEN
    Exceptions.RAISE("UDP implemented with 32 FD")
  END;
  noFlags := UnixSocket.Flags{};
  hostId := Hosts.OwnId();
  WriteF.WriteF1(SysStreams.sysOut, "hostId: %x\n", hostId)
END UDP.
