(*
    Title: 	Stub generator guts
    LastEdit:	"Tue Mar  7 15:25:09 1989"
    Author: 	Mac,Oms101,4652,
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE Generate;

IMPORT
  CharCodes, FileStream, M2Files, Streams, Strings, SysStreams, SYSTEM,
  TextIO, TimeDate, UtilError, Version, WriteF;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;

CONST
  MaxKeyWord = 32;
  NumArgs = 32;
  NumProcs = 256;
  MaxImports = 32;


TYPE
  Token = (EOFTk, SymbolTk, ProcedureTk, CardTk, IntTk, CharTk, BoolTk, 
           StringTk, VersionTk, AloTk, AmoTk, EoTk, EqualTk, ArrayTk,
           InTk, InOutTk, OutTk, LocalTk, EndTk, ProtocolTk,
           LengthTk, MaxLengthTk);

  ArgEntry = POINTER TO ArgEntryR;
  ArgEntryR =
    RECORD
      link:ArgEntry;
      name:Strings.String;
      type:Token;
      module,
      usertype:Strings.String;
      dir: Token;
      length,
      maxlength:Strings.String;
    END;

  ProcEntry = POINTER TO ProcEntryR;
  ProcEntryR =
    RECORD
      link:ProcEntry;
      name:Strings.String;
      protocol:Token;
      args:ARRAY [0..NumArgs-1] OF ArgEntry;
      curArg:INTEGER;
    END;

  TokenType = (WordType, ExpressionType, VarType, DottedType);

VAR
  stInput,
  stDef,
  stReal,
  stMod,
  stSSDef,
  stSSMod:Streams.Stream;

  csName,
  ssName,
  realName:ARRAY [0..127] OF CHAR;
  
  maxKeyWord:CARDINAL;
  keyWord:ARRAY [1..MaxKeyWord] OF Strings.String;
  keyWordToken:ARRAY [1..MaxKeyWord] OF Token;
  
  lastToken:Token;
  lastSymbol:Strings.String;
  
  curProc:INTEGER;
  procTable: ARRAY [0..NumProcs-1] OF ProcEntry;

  cardStr, intStr, chrStr, boolStr, strStr:Strings.String;

  amoStr, aloStr, eoStr:Strings.String;

  inStr, ioStr, ouStr, localStr:Strings.String;
  
  comStr:Strings.String;
  
  lineNumber:INTEGER;
  
  theVersion:CARDINAL;
  
  curImport:INTEGER;
  theImports:ARRAY [0..MaxImports-1] OF Strings.String;

  
PROCEDURE Error(str:ARRAY OF CHAR; arg1, arg2:SYSTEM.WORD);
  BEGIN
    WriteF.WriteF4(SysStreams.sysErr,  
      "%c%s.stb%c, line %i:  ", '"', WriteF.H(realName), '"', lineNumber);
    UtilError.ReportError(TRUE, str, arg1, arg2)
  END Error;
  

PROCEDURE NewProc(nme:Strings.String):ProcEntry;
  VAR
    proc:ProcEntry;
  BEGIN
    NEW(proc);
    WITH proc^ DO
      link := NIL;
      name := nme;
      curArg := 0
    END;
    procTable[curProc] := proc;
    INC(curProc);
    IF curProc = NumProcs THEN
      Error("Too many procedures", 0, 0)
    END;
    RETURN proc
  END NewProc;


PROCEDURE NewArg(proc:ProcEntry):ArgEntry;
  VAR
    arg:ArgEntry;
  BEGIN
    NEW(arg);
    WITH arg^ DO
      name := NIL;
      length := NIL;
      maxlength := NIL
    END;    
    WITH proc^ DO
      args[curArg] := arg;
      INC(curArg);
      IF curArg = NumArgs THEN
        Error("Too many arguments", 0, 0)
      END
    END;
    RETURN arg
  END NewArg;


PROCEDURE SkipToChar():BOOLEAN;
  VAR
    ch:CHAR;
    inComment:INTEGER;
  BEGIN
    inComment := 0;
    LOOP
      ch := TextIO.ReadCHAR(stInput);
      IF stInput^.status # Streams.Success THEN RETURN FALSE END;
      IF ch = CharCodes.NewLineCh THEN INC(lineNumber) END;
      IF inComment > 0 THEN
        CASE ch OF
          '(': INC(inComment);
        | ')': DEC(inComment);
        ELSE
        END;
      ELSE
        IF NOT TextIO.WhiteSpace(ch) THEN
          CASE ch OF
            '(':
              INC(inComment);
              Verbose(comStr, TRUE)
          | ')':
              Error("Unclosed comment", 0, 0); 
          ELSE
            Streams.Backspace(stInput);
            RETURN TRUE
          END
        END
      END
    END
  END SkipToChar;


PROCEDURE AlphaWord(ch:CHAR):BOOLEAN;
  BEGIN
    CASE ch OF
      'A'..'Z',
      'a'..'z',
      '0'..'9': RETURN FALSE
    ELSE
    END;
    RETURN TRUE
  END AlphaWord;


PROCEDURE DottedWord(ch:CHAR):BOOLEAN;
  BEGIN
    CASE ch OF
      'A'..'Z',
      'a'..'z',
      '0'..'9',
      '.': RETURN FALSE
    ELSE
    END;
    RETURN TRUE
  END DottedWord;


PROCEDURE Verbose(st:Strings.String; istoken:BOOLEAN);
  BEGIN
    IF NOT beVerbose THEN RETURN END;
    WriteF.WriteF4(SysStreams.sysOut,  
      "%c%s.stb%c, line %i:  ", '"', WriteF.H(realName), '"', lineNumber);
    IF istoken THEN
      WriteF.WriteF1(SysStreams.sysOut, "TOKEN:  %s\n", WriteF.S(st))
    ELSE
      WriteF.WriteF1(SysStreams.sysOut, "SYMBOL: %s\n", WriteF.S(st))
    END
  END Verbose;


PROCEDURE GetToken(exp:TokenType):Token;
  VAR
    i:CARDINAL;
    word:Strings.String;
  BEGIN
    IF NOT SkipToChar() THEN 
      Verbose(Strings.CopyCS("EOF"), TRUE);
      RETURN EOFTk
    END;
    CASE exp OF
      WordType:
        word := TextIO.ReadString(stInput, AlphaWord)
    | ExpressionType:
        word := TextIO.ReadString(stInput, TextIO.WhiteSpace)
    | DottedType, VarType:
        word := TextIO.ReadString(stInput, DottedWord)
    END;
    FOR i := 1 TO maxKeyWord DO
      IF Strings.EqualSS(keyWord[i], word) THEN
        lastSymbol := NIL;
        lastToken := keyWordToken[i];
        Verbose(word, TRUE);
        RETURN keyWordToken[i]
      END
    END;
    lastSymbol := Capitalise(exp, word);
    lastToken := SymbolTk;
    Verbose(lastSymbol, FALSE);
    RETURN SymbolTk
  END GetToken;


PROCEDURE Expression():Strings.String;
  BEGIN
    IF (Past('=')) AND
       (GetToken(ExpressionType) = SymbolTk) THEN
      RETURN lastSymbol
    END;
    Error("Bad expression for a length", 0, 0)
  END Expression;


PROCEDURE DoArray(arg:ArgEntry):BOOLEAN;
  BEGIN
    LOOP
      CASE GetToken(WordType) OF
        LengthTk:
          arg^.length := Expression()
      | MaxLengthTk:
          arg^.maxlength := Expression()
      | InTk, InOutTk, OutTk, LocalTk:
          IF arg^.length = NIL THEN Error("No length given", 0, 0) END;
          arg^.dir := lastToken;
          RETURN TRUE
      ELSE
          Error("Bad length for arrays", 0, 0)
      END
    END
  END DoArray;


PROCEDURE Capitalise(exp:TokenType; str:Strings.String):Strings.String;
  VAR
    off,
    index:CARDINAL;
  BEGIN
    CASE exp OF
      DottedType:
        str^[0] := CharCodes.CapitalCh(str^[0])
    | ExpressionType:
        index := 0;
        IF Strings.FindS(str, '.', index) THEN
          str^[0] := CharCodes.CapitalCh(str^[0])
        END
    ELSE
    END;
    off := 0; index := 0;
    LOOP
      IF str^[index] = '.' THEN
        IF exp = ExpressionType THEN
          str^[index+1] := CharCodes.CapitalCh(str^[index+1]);
          AddImport(Strings.ExtractSS(str, 0, index));
          exp := DottedType
        ELSE
          INC(off);
          str^[index] := CharCodes.CapitalCh(str^[index+off])
        END
      ELSIF off # 0 THEN 
        str^[index] := str^[index+off]
      END;
      IF str^[index] = 0C THEN EXIT END;
      INC(index)
    END;
    RETURN str
  END Capitalise;


PROCEDURE DoArgs(proc:ProcEntry):BOOLEAN;
  VAR
    arg:ArgEntry;
  BEGIN
    WITH proc^ DO
      LOOP
        CASE GetToken(WordType) OF
          EndTk: RETURN TRUE
        | CharTk, CardTk, IntTk, BoolTk, StringTk, SymbolTk:
            arg := NewArg(proc);
            arg^.type := lastToken;
            IF arg^.type = SymbolTk THEN
              arg^.module := lastSymbol;
              IF (NOT Past('.')) OR
                 (GetToken(WordType) # SymbolTk) THEN
                Error("Bad Module.Type", 0, 0)
              END;
              AddImport(arg^.module);
              arg^.usertype := lastSymbol;
            END;
            IF GetToken(VarType) # SymbolTk THEN RETURN FALSE END;
            arg^.name := lastSymbol;
            IF arg^.type = StringTk THEN
              IF GetToken(WordType) # MaxLengthTk THEN
                Error("No MAXLENGTH for STRING", 0, 0)
              END;
              arg^.maxlength := Expression()
            END;
            CASE GetToken(WordType) OF
              InTk, InOutTk, OutTk, LocalTk:
                arg^.dir := lastToken;
            | ArrayTk:
                IF NOT DoArray(arg) THEN RETURN FALSE END;
            ELSE RETURN FALSE
            END;
        ELSE RETURN FALSE
        END
      END
    END
  END DoArgs;


PROCEDURE DoProcedure():RC;
  VAR
    proc:ProcEntry;
  BEGIN
    CASE GetToken(WordType) OF
      ProcedureTk:
        IF (GetToken(DottedType) = SymbolTk) THEN
          proc := NewProc(lastSymbol);
          IF (GetToken(WordType) = ProtocolTk) THEN
            CASE GetToken(WordType) OF
              AloTk, AmoTk, EoTk:
                proc^.protocol := lastToken;
                IF (DoArgs(proc)) AND
                   (GetToken(DottedType) = SymbolTk) AND
                   (Strings.EqualSS(proc^.name, lastSymbol)) THEN
                  RETURN OK
                END
            ELSE
            END
          END;
          Error("Bad PROCEDURE syntax in %s\n", WriteF.S(proc^.name), 0);
          RETURN ParseError
        END
    | EOFTk:
          RETURN Done
    ELSE
    END;
    Error("Bad PROCEDURE name\n", 0, 0);
    RETURN ParseError
  END DoProcedure;


PROCEDURE DoAllProcs():RC;
  VAR
    rc:RC;
  BEGIN
    LOOP
      rc := DoProcedure();
      IF rc = Done THEN RETURN OK END;
      IF rc # OK THEN RETURN rc END
    END
  END DoAllProcs;


PROCEDURE Past(chin:CHAR):BOOLEAN;
  VAR
    ch:CHAR;
  BEGIN
    LOOP
      ch := TextIO.ReadCHAR(stInput);
      IF stInput^.status # Streams.Success THEN RETURN FALSE END;
      IF NOT TextIO.WhiteSpace(ch) THEN RETURN (ch = chin) END;
    END;
  END Past;

  
PROCEDURE DoVersion ():RC;
  VAR
    vers:CARDINAL;
  BEGIN
    IF GetToken(WordType) = VersionTk THEN
      IF Past("=") THEN
        theVersion := TextIO.ReadCARDINAL(stInput);
        IF theVersion # 0 THEN RETURN OK END
      END;
      Error("Bad VERSION syntax", 0, 0);
    END;
    Error("No VERSION", 0, 0);
    RETURN NoVersion
  END DoVersion;


PROCEDURE AddImport(imp:Strings.String);
  VAR
    i:INTEGER;
  BEGIN
    FOR i := 0 TO curImport-1 DO
      IF Strings.EqualSS(imp, theImports[i]) THEN RETURN END
    END;
    theImports[curImport] := imp;
    INC(curImport);
  END AddImport;


PROCEDURE Imports(st:Streams.Stream);
  VAR
    i:INTEGER;
  BEGIN
    FOR i := 0 TO curImport-2 DO
      WriteF.WriteF1(st, "%s, ", WriteF.S(theImports[i]));
    END;
    WriteF.WriteF1(st, "%s;\n\n", WriteF.S(theImports[curImport-1]));
  END Imports;


PROCEDURE DateEtc(st:Streams.Stream);
  VAR
    ds,
    ts:ARRAY [0..15] OF CHAR;
  BEGIN
    TimeDate.DateString(TimeDate.Date(), ds);
    TimeDate.TimeString(TimeDate.Time(), ts);
    WriteF.WriteF3(st, "(* m2stub V%s: at %s %s *)\n\n", 
                   WriteF.H(Version.idString),
                   WriteF.H(ds), WriteF.H(ts))
  END DateEtc;
  

PROCEDURE Headers();
  BEGIN
    DateEtc(stDef);
    WriteF.WriteF1(stDef, "DEFINITION MODULE %s;\n\n", WriteF.H(csName));
    WriteF.WriteF0(stDef, "IMPORT\n");
    Imports(stDef);
    WriteF.WriteF0(stDef, 
      "EXPORT QUALIFIED\n  Version, Procedures,\n  Find,\n");
    
    DateEtc(stReal);
    WriteF.WriteF1(stReal, "DEFINITION MODULE %s;\n\n", WriteF.H(realName));
    WriteF.WriteF0(stReal, "IMPORT\n");
    Imports(stReal);
    WriteF.WriteF0(stReal, 
      "EXPORT QUALIFIED\n  Version, Procedures,\n");
    
    DateEtc(stSSDef);
    WriteF.WriteF1(stSSDef, "DEFINITION MODULE %s;\n\n", WriteF.H(ssName));
    WriteF.WriteF0(stSSDef, "IMPORT\n");
    Imports(stSSDef);
    WriteF.WriteF0(stSSDef, "EXPORT QUALIFIED Offer;\n\n");
    WriteF.WriteF0(stSSDef,
 "PROCEDURE Offer(name:Strings.String; dist:RPC.Distance):RPC.MemberRC;\n\n");
      
    AddImport(Strings.CopyCS("SYSTEM"));
    DateEtc(stMod);
    WriteF.WriteF1(stMod, "IMPLEMENTATION MODULE %s;\n\n",WriteF.H(csName));
    WriteF.WriteF0(stMod, "IMPORT\n  Marshal, ");
    Imports(stMod);
    
    AddImport(Strings.CopyCS(realName));
    DateEtc(stSSMod);
    WriteF.WriteF1(stSSMod, "IMPLEMENTATION MODULE %s;\n\n",WriteF.H(ssName));
    WriteF.WriteF0(stSSMod, "IMPORT\n  Marshal,");
    Imports(stSSMod);
  END Headers;


PROCEDURE InOutType(arg:ArgEntry):Strings.String;
  BEGIN
    CASE arg^.dir OF
      InTk: RETURN inStr;
    | InOutTk: RETURN ioStr;
    | OutTk: RETURN ouStr;
    | LocalTk: RETURN localStr;
    ELSE
      Error("Bad IO Type", 0, 0)
    END
  END InOutType;


PROCEDURE TypeOf(arg:ArgEntry; st:Streams.Stream; colon, aLocal:BOOLEAN);
  VAR
    str:Strings.String;
  BEGIN
    CASE arg^.type OF
      CardTk:
        str := cardStr
    | IntTk:
        str := intStr
    | CharTk:
        str := chrStr
    | BoolTk:
        str := boolStr
    | StringTk:
        str :=  strStr
    | SymbolTk:
        WriteF.WriteF2(st, 
          "%s.%s", WriteF.S(arg^.module), WriteF.S(arg^.usertype));
        IF colon THEN WriteF.WriteF0(st, "\;") END;
        WriteF.WriteF1(st, " (* %s *)\n", WriteF.S(InOutType(arg)));
        RETURN
    ELSE
        Error("Bad Typeof %i", arg^.type, 0)
    END;
    
    IF arg^.length # NIL THEN
      WriteF.WriteF0(st, "ARRAY ");
      IF aLocal THEN
        IF (arg^.maxlength = NIL) OR (arg^.type = OutTk) THEN
          WriteF.WriteF1(st, "[0..%s-1] ", WriteF.S(arg^.length))
        ELSE
          WriteF.WriteF1(st, "[0..%s-1] ", WriteF.S(arg^.maxlength))
        END
      END;
      WriteF. WriteF0(st, "OF ")
    END;
    WriteF.WriteF1(st, "%s", WriteF.S(str));
        
    IF colon THEN WriteF.WriteF0(st, "\;") END;
    WriteF.WriteF1(st, " (* %s *)\n", WriteF.S(InOutType(arg)))
  END TypeOf;


PROCEDURE Def(proc:ProcEntry; st:Streams.Stream);
  VAR
    i:INTEGER;
  BEGIN
    WriteF.WriteF1(st, "PROCEDURE %s(\n", WriteF.S(proc^.name));
    WITH proc^ DO
      FOR i := 0 TO curArg-1 DO
        WriteF.WriteF1(st, "VAR %s:\t", WriteF.S(args[i]^.name));
        TypeOf(args[i], st, (i # curArg-1), FALSE);
      END
    END;
    WriteF.WriteF0(st, "):RPC.RC;\n");
  END Def;


PROCEDURE Call(proc:ProcEntry; st:Streams.Stream);
  VAR
    i:INTEGER;
  BEGIN
    WITH proc^ DO
      WriteF.WriteF2(st, 
        "rpcRc := %s.%s(", WriteF.H(realName), WriteF.S(name));
      IF (curArg > 0) THEN
        FOR i := 0 TO curArg-2 DO
          WriteF.WriteF1(st, "%s, ", WriteF.S(args[i]^.name));
        END;
        WriteF.WriteF1(st, "%s);\n", WriteF.S(args[curArg-1]^.name));
      ELSE
        WriteF.WriteF0(st, ");\n");
      END;
    END;
  END Call;


PROCEDURE Locals(proc:ProcEntry);
  VAR
    i:INTEGER;
  BEGIN
    WriteF.WriteF0(stSSMod, "rpcRc:RPC.RC;\n");
    WITH proc^ DO
      FOR i := 0 TO curArg-1 DO
        WriteF.WriteF1(stSSMod, "%s:", WriteF.S(args[i]^.name));
        TypeOf(args[i], stSSMod, TRUE, TRUE)
      END
    END
  END Locals;


PROCEDURE FixArray(arg:ArgEntry; st:Streams.Stream);
  BEGIN
    WITH arg^ DO
      CASE type OF
        CardTk, IntTk, BoolTk: 
          WriteF.WriteF2(st, 
            "Marshal.Words(SYSTEM.ADR(%s), %s, m);\n", 
            WriteF.S(name), WriteF.S(length));
      | CharTk:
          WriteF.WriteF2(st, 
            "Marshal.Chars(SYSTEM.ADR(%s), %s, m);\n", 
            WriteF.S(name), WriteF.S(length));
      ELSE
          Error("Only array of INT, CHAR and BOOL supported", 0, 0)
      END
    END
  END FixArray;


PROCEDURE VarArray(arg:ArgEntry; st:Streams.Stream);
  BEGIN
    WITH arg^ DO
      CASE type OF
        CardTk, IntTk, BoolTk: 
          WriteF.WriteF2(st, 
            "Marshal.AnArrayOfWord(%s, %s, m);\n", 
            WriteF.S(name), WriteF.S(length));
      | CharTk:
          WriteF.WriteF2(st, 
            "Marshal.AnArrayOfChar(%s, %s, m);\n", 
            WriteF.S(name), WriteF.S(length));
      ELSE
          Error("Only array of INT, CHAR and BOOL supported", 0, 0)
      END
    END
  END VarArray;


PROCEDURE Single(unmarshal:BOOLEAN; arg:ArgEntry; st:Streams.Stream);
  BEGIN
    WITH arg^ DO
      CASE type OF
        CardTk, IntTk, BoolTk: 
          WriteF.WriteF1(st, 
            "Marshal.AWord(%s, m);\n", WriteF.S(name));
      | CharTk:
          WriteF.WriteF1(st, 
            "Marshal.AChar(%s, m);\n", WriteF.S(name));
      | StringTk: 
          WriteF.WriteF1(st, 
            "Marshal.AString(%s, m);\n", WriteF.S(name));
          IF unmarshal THEN
            WriteF.WriteF1(st,
              "Strings.Delete(%s);\n", WriteF.S(name))
          END
      | SymbolTk:
          WriteF.WriteF3(st, 
            "%s.Marshal%s(%s, m);\n", WriteF.S(module), 
            WriteF.S(usertype), WriteF.S(name));
      ELSE
          Error("Bad Marshal", 0, 0)
      END
    END
  END Single;


PROCEDURE Both(proc:ProcEntry; st:Streams.Stream; tipe:Token);
  VAR
    i:INTEGER;
  BEGIN
    WITH proc^ DO
      FOR i := 0 TO curArg-1 DO
        WITH args[i]^ DO
          IF (dir = InOutTk) OR (dir = tipe) THEN
            IF length # NIL THEN
              IF maxlength # NIL THEN
                VarArray(args[i], st)
              ELSE
                FixArray(args[i], st)
              END;
            ELSE
              Single(tipe = OutTk, args[i], st);
            END
          END
        END
      END
    END
  END Both;


PROCEDURE Marshal(proc:ProcEntry; st:Streams.Stream);
  BEGIN
    Both(proc, st, InTk)
  END Marshal;


PROCEDURE UnMarshal(proc:ProcEntry; st:Streams.Stream);
  BEGIN
    Both(proc, st, OutTk)
  END UnMarshal;


PROCEDURE ProtocolOf(proc:ProcEntry):Strings.String;
  BEGIN
    CASE proc^.protocol OF
      AmoTk: RETURN amoStr;
    | AloTk: RETURN aloStr;
    | EoTk:  RETURN eoStr;
    ELSE 
      Error("Bad PROTOCOL %i", proc^.protocol, 0)
    END
  END ProtocolOf;


PROCEDURE Mods(proc:ProcEntry);
  BEGIN
    WriteF.WriteF0(stMod, "VAR\nm:Marshal.T;\nrpcRc:RPC.RC;\nBEGIN\n");
    WriteF.WriteF2(stMod, 
      "m := RPC.ClientArguments(boundTo, ORD(%sProc), Version, %s);\n", 
      WriteF.S(proc^.name), WriteF.S(ProtocolOf(proc)));
    Marshal(proc, stMod);
    WriteF.WriteF0(stMod, "rpcRc := RPC.Call(m);\n");
    WriteF.WriteF0(stMod, "IF rpcRc # RPC.OK THEN RETURN rpcRc END;\n");
    UnMarshal(proc, stMod);
    WriteF.WriteF1(stMod, 
      "RETURN RPC.EndOfResults(m);\nEND %s;\n\n", WriteF.S(proc^.name));

    WriteF.WriteF1(stSSMod, 
      "PROCEDURE %sStub(m:Marshal.T);\nVAR\n", WriteF.S(proc^.name));
    Locals(proc);
    WriteF.WriteF0(stSSMod, "BEGIN\n");
    Marshal(proc, stSSMod);
    WriteF.WriteF0(stSSMod,
      "rpcRc := RPC.EndOfArgs(m);\nIF rpcRc = RPC.OK THEN\n");
    Call(proc, stSSMod);
    WriteF.WriteF0(stSSMod, "IF rpcRc = RPC.OK THEN\n");
    UnMarshal(proc, stSSMod);
    WriteF.WriteF1(stSSMod, 
      "RETURN\nEND\nEND;\nRPC.SetReturnCode(rpcRc, m)\nEND %sStub;\n\n", 
      WriteF.S(proc^.name))
  END Mods;
  

PROCEDURE Case();
  VAR
    i:INTEGER;
    proc:ProcEntry;
  BEGIN
    FOR i := 0 TO curProc-1 DO
      proc := procTable[i];
      IF i # 0 THEN
        WriteF.WriteF0(stSSMod, "| ")
      ELSE
        WriteF.WriteF0(stSSMod, "  ")
      END;
      WriteF.WriteF3(stSSMod, 
        "%s.%sProc: %sStub(m)\n", 
        WriteF.H(realName), WriteF.S(proc^.name),WriteF.S(proc^.name))        
    END
  END Case;


PROCEDURE Rxer();
  BEGIN
    WriteF.WriteF0(stSSMod, "PROCEDURE Receiver(p:RPC.PktBuf):RPC.PktBuf;\n");
    WriteF.WriteF0(stSSMod,"VAR\nm:Marshal.T;\nrpcRc:RPC.RC;\nproc:CARDINAL;\n");
    WriteF.WriteF1(stSSMod,
      "BEGIN\nrpcRc := RPC.ServerProcedureVCheck(p, m, proc, %s.Version);\n", 
      WriteF.H(realName));
    WriteF.WriteF1(stSSMod,
      "IF rpcRc = RPC.OK THEN\n CASE %s.Procedures(proc) OF\n", 
      WriteF.H(realName));
    Case();
    WriteF.WriteF0(stSSMod,
      "ELSE RPC.SetReturnCode(RPC.ProcError, m)\nEND\nELSE\n");
    WriteF.WriteF0(stSSMod,
      "RPC.SetReturnCode(rpcRc, m)\nEND;\nRETURN RPC.EndOfCall(m)\n");
    WriteF.WriteF0(stSSMod, "END Receiver;\n\n");
    WriteF.WriteF0(stSSMod, 
   "PROCEDURE Offer(name:Strings.String; dist:RPC.Distance):RPC.MemberRC;\n");
    WriteF.WriteF0(stSSMod, 
      "VAR\nid:RPC.ServiceId;\nBEGIN\n");
    WriteF.WriteF0(stSSMod, 
      "RETURN RPC.Offer(name, dist, Receiver, 0, id);\nEND Offer;\n\n");
  END Rxer;
  

PROCEDURE Find();
  BEGIN
    WriteF.WriteF0(stDef,
  "PROCEDURE Find(name:Strings.String; dist:RPC.Distance):RPC.MemberRC;\n\n");
      
    WriteF.WriteF0(stMod, 
  "PROCEDURE Find(name:Strings.String; dist:RPC.Distance):RPC.MemberRC;\n");
    WriteF.WriteF0(stMod,
      "BEGIN\nRETURN RPC.Find(name, dist, boundTo);\nEND Find;\n\n");
  END Find;

  
PROCEDURE Internals();
  VAR
    i:INTEGER;
  BEGIN
    FOR i := 0 TO curProc-2 DO
      WriteF.WriteF1(stDef, "%s, ", WriteF.S(procTable[i]^.name));
      WriteF.WriteF1(stReal, "%s, ", WriteF.S(procTable[i]^.name))
    END;
    WriteF.WriteF1(stReal, "%s;\n\n", WriteF.S(procTable[curProc-1]^.name));
    WriteF.WriteF1(stDef, "%s;\n\n", WriteF.S(procTable[curProc-1]^.name));
    WriteF.WriteF1(stReal, "CONST\n  Version = %i;\n\n", theVersion);
    WriteF.WriteF0(stReal, "TYPE\n  Procedures = (");
    WriteF.WriteF1(stDef, "CONST\n  Version = %i;\n\n", theVersion);
    WriteF.WriteF0(stDef, "TYPE\n  Procedures = (");
    FOR i := 0 TO curProc-2 DO
      WriteF.WriteF1(stDef, "%sProc, ", WriteF.S(procTable[i]^.name));
      WriteF.WriteF1(stReal, "%sProc, ", WriteF.S(procTable[i]^.name));
    END;
    WriteF.WriteF1(stDef,
      "%sProc);\n\n", WriteF.S(procTable[curProc-1]^.name));
    WriteF.WriteF1(stReal,
      "%sProc);\n\n", WriteF.S(procTable[curProc-1]^.name));
    WriteF.WriteF0(stMod, "VAR\nboundTo:RPC.Place;\n\n");
    FOR i := 0 TO curProc-1 DO
      Def(procTable[i], stDef);
      Def(procTable[i], stReal);
      TextIO.NewLine(stDef);
      TextIO.NewLine(stReal);
      Def(procTable[i], stMod);
      Mods(procTable[i]);
    END;
    Rxer();
    Find()
  END Internals;
  

PROCEDURE Tails();
  BEGIN
    WriteF.WriteF1(stDef, "END %s.\n", WriteF.H(csName));
    WriteF.WriteF1(stReal, "END %s.\n", WriteF.H(realName));
    WriteF.WriteF1(stMod, "END %s.\n", WriteF.H(csName));
    WriteF.WriteF1(stSSDef, "END %s.\n", WriteF.H(ssName));
    WriteF.WriteF1(stSSMod, "END %s.\n", WriteF.H(ssName))
  END Tails;


PROCEDURE DoIt ():RC;
  VAR
    rc:RC;
  BEGIN
    rc := DoVersion();
    IF rc # OK THEN RETURN rc END;
    rc := DoAllProcs();
    IF rc # OK THEN RETURN rc END;
    Headers();
    Internals();
    Tails();
    RETURN OK;
  END DoIt;


PROCEDURE Files(name:ARRAY OF CHAR):RC;
  VAR
    rc:RC;
  BEGIN
    Strings.CopyCC(name, realName);
    stInput := M2Files.CreateInput(name, "stb");
    IF stInput^.status # Streams.Success THEN RETURN BadInFile END;
    stReal := M2Files.CreateOutput(realName, "def");
    IF stReal^.status # Streams.Success THEN RETURN BadOutFile1 END;
    Strings.ConcatCC("CS", name, csName);
    stDef := M2Files.CreateOutput(csName, "def");
    IF stDef^.status # Streams.Success THEN RETURN BadOutFile2 END;
    stMod := M2Files.CreateOutput(csName, "mod");
    IF stMod^.status # Streams.Success THEN RETURN BadOutFile3 END;
    Strings.ConcatCC("SS", name, ssName);
    stSSDef := M2Files.CreateOutput(ssName, "def");
    IF stSSDef^.status # Streams.Success THEN RETURN BadOutFile4 END;
    stSSMod := M2Files.CreateOutput(ssName, "mod");
    IF stSSMod^.status # Streams.Success THEN RETURN BadOutFile5 END;
    rc := DoIt();
    Streams.Delete(stInput);
    Streams.Delete(stDef);
    Streams.Delete(stReal);
    Streams.Delete(stMod);
    Streams.Delete(stSSDef);
    Streams.Delete(stSSMod);
    RETURN rc
  END Files;


PROCEDURE AddKeyWord(word:ARRAY OF CHAR; tk:Token);
  BEGIN
    IF maxKeyWord = MaxKeyWord THEN
      Error("Too many keywords for table", 0, 0)
    END;
    INC(maxKeyWord);
    keyWord[maxKeyWord] := Strings.CopyCS(word);
    keyWordToken[maxKeyWord] := tk;
  END AddKeyWord;


BEGIN
  curProc := 0;
  curImport := 0;
  maxKeyWord := 0;
  lineNumber := 1;

  cardStr := Strings.CopyCS("CARDINAL");
  intStr := Strings.CopyCS("CARDINAL");
  chrStr := Strings.CopyCS("CHAR");
  boolStr := Strings.CopyCS("BOOLEAN");
  strStr := Strings.CopyCS("Strings.String");

  amoStr := Strings.CopyCS("RPC.AtMostOnce");
  aloStr := Strings.CopyCS("RPC.AtLeastOnce");
  eoStr := Strings.CopyCS("RPC.ExactlyOnce");

  inStr := Strings.CopyCS("In");
  ioStr := Strings.CopyCS("InOut");
  ouStr := Strings.CopyCS("Out");
  localStr := Strings.CopyCS("Local");

  comStr := Strings.CopyCS("Comment");
  
  AddImport(Strings.CopyCS("RPC"));
  AddImport(Strings.CopyCS("Strings"));

  AddKeyWord("PROCEDURE", ProcedureTk);
  AddKeyWord("CARD", CardTk);
  AddKeyWord("INT", IntTk);
  AddKeyWord("CHAR", CharTk);
  AddKeyWord("BOOL", BoolTk);
  AddKeyWord("STRING", StringTk);
  AddKeyWord("VERSION", VersionTk);
  AddKeyWord("atleastonce", AloTk);
  AddKeyWord("atmostonce", AmoTk);
  AddKeyWord("exactlyonce", EoTk);
  AddKeyWord("ARRAY", ArrayTk);
  AddKeyWord("IN", InTk);
  AddKeyWord("INOUT", InOutTk);
  AddKeyWord("OUT", OutTk);
  AddKeyWord("LOCAL", LocalTk);
  AddKeyWord("END", EndTk);
  AddKeyWord("PROTOCOL", ProtocolTk);
  AddKeyWord("LENGTH", LengthTk);
  AddKeyWord("MAXLENGTH", MaxLengthTk);
  
  beVerbose := FALSE
END Generate.
