MODULE Pooh;


IMPORT
  Streams, StringStream, SysStreams, TextIO, Threads, Timers, UDP, WriteF;


FROM Storage IMPORT ALLOCATE, DEALLOCATE;

  
TYPE
  Buffer = POINTER TO ARRAY [0..1023] OF CHAR;
  ForkType = (Prod, Wait, Send, Run1, Run2);


VAR
  buffLock:Threads.Mutex;
  ourHand1,
  ourHand2:UDP.Handle;
  ourSock1,
  ourSock2:UDP.Socket;
  theString:ARRAY [0..1023] OF CHAR;
  stringOut:Streams.Stream;


PROCEDURE Sender();
  VAR
    buffer:Buffer;
    rc:    UDP.SendRC;
  BEGIN
    NEW(buffer);
    LOOP
      Timers.WaitSecs(3, Timers.FromNow);
      rc := UDP.Send(ourSock1, ourHand2, buffer, 1);
      Threads.Acquire(buffLock);
        WriteF.WriteF1(SysStreams.sysOut, "Send returned %u\n", rc);
      Threads.Release(buffLock)
    END
  END Sender;
  

PROCEDURE Waiter();
  VAR
    buffer:  Buffer;
    length:  CARDINAL;
    rc:      UDP.RecVRC;
    fromSock:UDP.Socket;
  BEGIN
    NEW(buffer);
    LOOP
      length := HIGH(buffer^);
      rc := UDP.RecV(ourHand1, fromSock, buffer, length);
      Threads.Acquire(buffLock);
        WriteF.WriteF2(SysStreams.sysOut,
                       "RecV returned %u, %u bytes\n", rc, length);
      Threads.Release(buffLock);
    END
  END Waiter;


PROCEDURE Consumer();
  VAR
    ch:CHAR;
  BEGIN
    LOOP
      ch := TextIO.ReadCHAR(SysStreams.sysIn);
      Threads.Acquire(buffLock);
        WriteF.WriteF0(SysStreams.sysOut, "Consumer: Hello world\n");
      Threads.Release(buffLock);
    END;
  END Consumer;


PROCEDURE Producer();
  VAR
    ch:CHAR;
  BEGIN
    LOOP
      Timers.WaitSecs(2, Timers.FromNow);
      Threads.Acquire(buffLock);
        WriteF.WriteF0(SysStreams.sysOut, "Producer: Hello world\n");
      Threads.Release(buffLock);
    END;
  END Producer;


PROCEDURE Runner(no:INTEGER);
  VAR
    i:INTEGER;
  BEGIN
    i := 0;
    LOOP 
      INC(i);
      IF i MOD 60 = 0 THEN 
        Threads.Acquire(buffLock);
        WriteF.WriteF1(stringOut, "%x", no);
        IF i MOD 3600 = 0 THEN
          i := 0;
          WriteF.WriteF1(stringOut, "%c", 0C);
          WriteF.WriteF1(SysStreams.sysOut, "%s\n", WriteF.H(theString));
          Streams.Seek(stringOut, Streams.Beginning, 0)
        END;
        Threads.Release(buffLock)
      END;
    END;
  END Runner;


PROCEDURE Forker(arg:Threads.ForkeeArg):Threads.ForkeeReturn;
  BEGIN
    CASE ForkType(arg) OF
      Prod: Producer();
    | Wait: Waiter();
    | Send: Sender();
    | Run1: Runner(1)
    | Run2: Runner(2)
    END;
    RETURN Threads.ForkeeReturn(NIL)
  END Forker;


PROCEDURE GetSocket(VAR sock:UDP.Socket; VAR hand:UDP.Handle):BOOLEAN;
  BEGIN
    hand := UDP.ReserveAnySocket(sock);
    IF hand < 0 THEN
      WriteF.WriteF1(SysStreams.sysOut, "Reserve returned %u\n", hand);
      RETURN FALSE
    ELSE
      WriteF.WriteF1(SysStreams.sysOut, "Port is %u\n", sock.port);
      RETURN TRUE
    END
  END GetSocket;


PROCEDURE Main();
  VAR
    thread:Threads.Thread;
  BEGIN
    stringOut := StringStream.CreateInOutChars(theString);
    IF  GetSocket(ourSock1, ourHand1)
    AND GetSocket(ourSock2, ourHand2) THEN
      Threads.InitMutex(buffLock);
(*    thread := Threads.Fork(Forker, Threads.ForkeeArg(Prod));
      thread := Threads.Fork(Forker, Threads.ForkeeArg(Wait));
      thread := Threads.Fork(Forker, Threads.ForkeeArg(Send)); *)
      thread := Threads.Fork(Forker, Threads.ForkeeArg(Run1));
      thread := Threads.Fork(Forker, Threads.ForkeeArg(Run2));
      Consumer()
    END;
  END Main;

  
BEGIN
  Main()
END Pooh.
