(*
    Title: 	RPC test
    LastEdit:	"Fri Jan  8 16:01:05 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

MODULE RPC1;


IMPORT
  Marshal, RPC, RPCInit, Strings, SysStreams, SYSTEM, Threads, Timers, WriteF;

FROM SystemTypes IMPORT String;
FROM Threads     IMPORT Thread;


CONST
  NumThreads = 2;
  RPCFailure = "RPCFailure";


VAR
  serverPlace:RPC.Place;


PROCEDURE Receiver(p:RPC.PktBuf):RPC.PktBuf;
  VAR
    m:   Marshal.T;
    q:   RPC.PktBuf;
    i, j,
    proc,
    vers:CARDINAL;
    rc:  RPC.RC;
  BEGIN
    proc := RPC.ServerProcedure(p, m, vers);
    Marshal.AWord(i, m);
    Marshal.AWord(j, m);
    rc := RPC.EndOfArgs(m);
    IF rc = RPC.OK THEN
      DoService(i, j)
    ELSE
      RPC.SetReturnCode(rc, m)
    END;
    RETURN RPC.EndOfCall(m)
  END Receiver;


PROCEDURE DoService(i, j:CARDINAL);
  BEGIN
  (*  WriteF.WriteF2(SysStreams.sysOut, "%x7 %x3 ", i, j) *)
  END DoService;
     

PROCEDURE Caller(arg:Threads.ForkeeArg):Threads.ForkeeReturn;
  CONST
    Num = 128;
  VAR
    i,j:CARDINAL;
  BEGIN
    i := CARDINAL(arg);
    j := 0;
    WHILE (j < Num) AND CallRPC(i, j) DO INC(j) END;
    IF j < Num THEN
      WriteF.WriteF1(SysStreams.sysOut, "RPC Failed %x\n", Num - j)
    END;
    WriteF.WriteF1(SysStreams.sysOut, "Finished Thread %x\n", i);
    RETURN Threads.ForkeeReturn(NIL)
  END Caller;


PROCEDURE CallRPC(i, j:CARDINAL):BOOLEAN;
  VAR
    m: Marshal.T;
    n: CARDINAL;
    rc:RPC.RC;
  BEGIN
    m := RPC.ClientArguments(serverPlace, 0, 0, RPC.ExactlyOnce);
    Marshal.AWord(j, m);
    Marshal.AWord(i, m);
    rc := RPC.Call(m);
    IF rc # RPC.OK THEN Error(ORD(rc), "RPC call"); RETURN FALSE END;
    rc := RPC.EndOfResults(m);
    IF rc # RPC.OK THEN Error(ORD(rc), "EndOfResults"); RETURN FALSE END;
    RETURN TRUE
  END CallRPC;


PROCEDURE Error(rc:CARDINAL; what:ARRAY OF CHAR);
  BEGIN
    WriteF.WriteF2(SysStreams.sysOut, 
                   "%s error, rc = %x\n", WriteF.H(what), rc)
  END Error;


PROCEDURE Main();
  VAR
    i:          INTEGER;
    mrc:        RPC.MemberRC;
    wait:       Threads.Condition;
    string:     String;
    forkReturn: Threads.ForkeeReturn;
    service:    RPC.ServiceId;
    callThreads:ARRAY [1..NumThreads] OF Thread;
  BEGIN
    string := Strings.CopyCS("Mac1");
    WriteF.WriteF0(SysStreams.sysOut, "Mac1 starting ... ");
    RPCInit.Init(0);
  
    mrc := RPC.Offer(string, RPC.SameMachine, Receiver, 0, service);
    IF mrc # RPC.BindOK THEN Error(ORD(mrc), "Offer") END;
  
    WriteF.WriteF1(SysStreams.sysOut, "ServiceId %x8\n", service);
    WriteF.WriteF0(SysStreams.sysOut, "Lookup .. ");
  
    mrc := RPC.Find(string, RPC.SameNet, serverPlace);
    IF mrc # RPC.BindOK THEN
      Error(ORD(mrc), "Not Found");
      WriteF.WriteF0(SysStreams.sysOut, "Server Dying!\n");
      HALT
    END;
      
    WriteF.WriteF1(SysStreams.sysOut, "ServiceId %x8\n", serverPlace.service);
    FOR i := 1 TO NumThreads DO
      callThreads[i] := Threads.Fork(Caller, Threads.ForkeeArg(i));
      WriteF.WriteF1(SysStreams.sysOut, "Fork Thread %x\n", i)
    END;
    FOR i := 1 TO NumThreads DO
      forkReturn := Threads.Join(callThreads[i]);
      WriteF.WriteF1(SysStreams.sysOut, "Join Thread %x\n", i)
    END;
    WriteF.WriteF0(SysStreams.sysOut, "The End\n\n")
  END Main;


BEGIN
  Main()
END RPC1.
