(*
    Title: 	RPC test
    LastEdit:	"Mon Feb  1 15:01:10 1988"
    Author: 	Derek McAuley,Oms1,4663,329629
		Cambridge University Computer Laboratory
*)

MODULE RPC2;


IMPORT
  Exceptions, Marshal, RPC, RPCInit, Strings, SysStreams, SYSTEM, 
  Threads, TimeDate, Timers, UnixArgs, UnixMCI, UnixMonitor, WriteF;

FROM SystemTypes IMPORT String;
FROM Threads     IMPORT Thread;


CONST
  NumThreads = 10;
  RPCFailure = "RPCFailure";


VAR
  dieNow:BOOLEAN;
  serverPlace:RPC.Place;


PROCEDURE Receiver(p:RPC.PktBuf):RPC.PktBuf;
  VAR
    m:   Marshal.T;
    q:   RPC.PktBuf;
    die: BOOLEAN;
    i, j,
    proc,
    vers:CARDINAL;
    rc:  RPC.RC;
  BEGIN
    proc := RPC.ServerProcedure(p, m, vers);
    Marshal.AWord(j, m);
    Marshal.AWord(i, m);
    Marshal.AWord(die, m);
    rc := RPC.EndOfArgs(m);
    IF rc = RPC.OK THEN
      DoService(die, i, j)
    ELSE
      RPC.SetReturnCode(rc, m)
    END;
    RETURN RPC.EndOfCall(m)
  END Receiver;


PROCEDURE DoService(die:BOOLEAN; i, j:CARDINAL);
  BEGIN
    dieNow := die;
    IF j MOD 64 = 0 THEN
      WriteF.WriteF2(SysStreams.sysOut, "%x7 %x3\n", i, j)
    END
  END DoService;
     

PROCEDURE Caller(arg:Threads.ForkeeArg):Threads.ForkeeReturn;
  CONST
    Num = 1028;
  VAR
    i,j:  CARDINAL;
    ts1,
    ts2:  ARRAY [0..15] OF CHAR;
    time1,
    time2:CARDINAL;
  BEGIN
    i := CARDINAL(arg);
    j := 0;
    time1 := TimeDate.Time();
    WHILE (j < Num) AND CallRPC(FALSE, i, j) DO INC(j) END;
    time2 := TimeDate.Time();
    TimeDate.TimeString(time1, ts1);
    TimeDate.TimeString(time2, ts2);
    WriteF.WriteF3(SysStreams.sysOut, 
                   "%i calls %s - %s\n", Num, WriteF.H(ts1), WriteF.H(ts2));
    IF j < Num THEN
      WriteF.WriteF1(SysStreams.sysOut, "RPC Failed %x\n", Num - j)
    END;
    WriteF.WriteF1(SysStreams.sysOut, "Finished Thread %x\n", i);
    RETURN Threads.ForkeeReturn(NIL)
  END Caller;


PROCEDURE CallRPC(die:BOOLEAN; i, j:CARDINAL):BOOLEAN;
  VAR
    m: Marshal.T;
    n: CARDINAL;
    rc:RPC.RC;
  BEGIN
    IF j MOD 64 = 0 THEN
      WriteF.WriteF2(SysStreams.sysOut, "%x7 %x3\n", i, j)
    END;
    m := RPC.ClientArguments(serverPlace, 0, 0, RPC.ExactlyOnce);
    Marshal.AWord(j, m);
    Marshal.AWord(i, m);
    Marshal.AWord(die, m);
    rc := RPC.Call(m);
    IF rc # RPC.OK THEN Error(ORD(rc), "RPC call"); RETURN FALSE END;
    rc := RPC.EndOfResults(m);
    IF rc # RPC.OK THEN Error(ORD(rc), "EndOfResults"); RETURN FALSE END;
    RETURN TRUE
  END CallRPC;


PROCEDURE Error(rc:CARDINAL; what:ARRAY OF CHAR);
  BEGIN
    WriteF.WriteF2(SysStreams.sysOut, 
                   "%s error, rc = %x\n", WriteF.H(what), rc)
  END Error;


PROCEDURE Main();
  VAR
    i,
    num:        INTEGER;
    arg:        ARRAY [0..79] OF CHAR;
    mrc:        RPC.MemberRC;
    wait:       Threads.Condition;
    string:     String;
    forkReturn: Threads.ForkeeReturn;
    service:    RPC.ServiceId;
    callThreads:ARRAY [1..NumThreads] OF Thread;
  BEGIN
    UnixArgs.GetArg(1, arg);
    string := Strings.CopyCS("Mac2");
    WriteF.WriteF0(SysStreams.sysOut, "Mac2 starting ... ");
    RPCInit.Init(0);
  
    
    CASE arg[0] OF
      's', 'S':
        dieNow := FALSE;
        mrc := RPC.Offer(string, RPC.SameMachine, Receiver, 0, service);
        IF mrc # RPC.BindOK THEN Error(ORD(mrc), "Offer") END;
    
        WriteF.WriteF1(SysStreams.sysOut, "ServiceId %x8\n", service);
        REPEAT Timers.WaitSecs(5, Timers.FromNow) UNTIL dieNow
    | '0'..'9':
        num := ORD(arg[0]) - ORD('0');
        UnixMonitor.MonStart(0, UnixMCI.endText, 1);
        mrc := RPC.Find(string, RPC.SameNet, serverPlace);
        IF mrc # RPC.BindOK THEN
          Error(ORD(mrc), "Not Found");
          Exceptions.RAISE("Not Found!")
        END;      

        WriteF.WriteF1(SysStreams.sysOut, 
                       "Lookup .. id %x8\n", serverPlace.service);
        FOR i := 1 TO num DO
          callThreads[i] := Threads.Fork(Caller, Threads.ForkeeArg(i));
          WriteF.WriteF1(SysStreams.sysOut, "Fork Thread %x\n", i)
        END;
        FOR i := 1 TO num DO
          forkReturn := Threads.Join(callThreads[i]);
          WriteF.WriteF1(SysStreams.sysOut, "Join Thread %x\n", i)
        END;
        WriteF.WriteF0(SysStreams.sysOut, "Kill Server\n");
     (* IF CallRPC(TRUE, 0, 0) THEN END; *)
        UnixMonitor.MonFinish()
    ELSE
        WriteF.WriteF0(SysStreams.sysOut, "Bad Args\n");
    END;
    WriteF.WriteF0(SysStreams.sysOut, "The End\n\n")
  END Main;


BEGIN
  Main()
END RPC2.
