IMPLEMENTATION MODULE AuxGen3;

FROM SYSTEM IMPORT WORD;
FROM items IMPORT ITEM, MODE, REGISTER, Operator, GlobalLabel, MonOperator,
                  InDataReg, equal, NullRegItem, OnRealStack, PackItem,
		  PackRealStackItem, PackReg, immediate,
		  ConditionCode, DataReg0;
FROM RegSlv IMPORT Slave, protect, FreeReg, StopSlaving,
                   DeleteFromSlave, PutIntoSlave, UseBitSet,
		   SetRegFree, IsFree, MarkOnStack;
FROM PsStk IMPORT PseudStk, base, top, PseudoStackEmpty, MoveIntoScratchSpace;
FROM GenOp IMPORT SetResultItem;
FROM labels IMPORT NewLabel, PushLabelStack;
FROM GenProcs IMPORT GenJsrToGlobalLabel, GenMove, GenExg, GenMoveToReg, GenNeg,
                     GenShift, GenRelBranch, GenMoveImm, GenTst;
FROM AuxGen1 IMPORT AddAndOr;
FROM AuxGen2 IMPORT shift;
FROM MonGen IMPORT SimpleMonadicOps;
FROM UtilError IMPORT ReportError;

VAR DataReg4, DataReg5, DataReg7 : ITEM;


(* The 32 bit multiplication and division routines are
   filched from Ian Wilson                                          *)


(* Multiplication is done by a standard subroutine which expects the
   two operands in D4 and D5 and returns the result in D4. Note that
   the contents of D4 to D7 are overwritten

   For the assembler routine see the Modula 2 assembler arithmetic
   library - m2:assembler.arithmetic
*)

PROCEDURE Multiply (Op : Operator; VAR Opd1, Opd2, Dest : ITEM);
    VAR power : CARDINAL;
    BEGIN
        IF immediate (Opd1) 
            AND PowerOfTwo (Opd1.DispOrData, power, (Op=Mul))
        THEN
            ShiftMulDiv (Op, power, Opd1, Opd2, Dest);
        ELSIF immediate (Opd2) 
            AND PowerOfTwo (Opd2.DispOrData, power, (Op=Mul))
        THEN
            ShiftMulDiv (Op, power, Opd2, Opd1, Dest);
        ELSE
            NormalMultiply (Op, Opd1, Opd2, Dest);
        END (* if *);
    END Multiply;

PROCEDURE ShiftMulDiv (Op : Operator; power : CARDINAL;
                            VAR ImmedOpd, OtherOpd, Dest : ITEM);
    VAR 
        PowerItem, SlvOtherOpd : ITEM;
        ShiftOp : Operator;
        signed, NegateResult : BOOLEAN;
    BEGIN
        CASE Op OF
            Mul :   ShiftOp := Ashl; signed := TRUE;
        |   Umul:   ShiftOp := Shl;  signed := FALSE;
        |   Udiv:   ShiftOp := Shr;  signed := FALSE;
        END (* case *);
        NegateResult := signed AND (ImmedOpd.DispOrData < 0);
        IF power = 0 THEN (* multipy or divide by 1 or -1 *)
            MulDivBy1 (OtherOpd, Dest, NegateResult);
        ELSE
            PackItem (PowerItem, Immediate, NoReg, power, NoReg);
            shift (ShiftOp, PowerItem, OtherOpd, Dest, NegateResult);
        END (* if *);
    END ShiftMulDiv;

PROCEDURE MulDivBy1 (VAR opd, dest: ITEM; neg: BOOLEAN);
    VAR slaved: ITEM;
BEGIN
    IF neg THEN
        SimpleMonadicOps (Neg, opd, dest);
    ELSIF equal (dest, NullRegItem) THEN
        SetResultItem (opd);
    ELSE
        Slave (opd, slaved);
        GenMove (slaved, dest);
        PutIntoSlave (slaved, dest, FALSE);
    END (* if *);
END MulDivBy1;

PROCEDURE NormalMultiply (Op : Operator; VAR Opd1, Opd2, Dest : ITEM);
    VAR SlvOpd1, SlvOpd2 : ITEM;
    BEGIN
        (* Note slaving done before preparing for standard subroutine
           because this procedure will free registers D4 to D7      *)
        Slave (Opd1, SlvOpd1);
        Slave (Opd2, SlvOpd2);
        PrepareForJsr (2, D4);
        (* 2 arguments on pseudo stack, registers D4 to D7 needed
           by multiply subroutine                                   *)
        IF equal (SlvOpd1, DataReg5) THEN
            GenMoveToReg (SlvOpd2, D4);
        ELSIF equal (SlvOpd2, DataReg4) THEN
            GenMoveToReg (SlvOpd1, D5);
        ELSE
            GenMoveToReg (SlvOpd1, D4);
            GenMoveToReg (SlvOpd2, D5);
        END (* if *);
        GenJsrToGlobalLabel (multiply);
        RecoverFromJsr ();
        IF equal (Dest, NullRegItem) THEN
            SetResultItem (DataReg4);
        ELSE
            GenMove (DataReg4, Dest);
            PutIntoSlave (DataReg4, Dest, FALSE);
        END (* if *);
    END NormalMultiply;

(* Unsigned Div & Mod are handled by calling a standard assembler 
   subroutine UDIV. This subroutine requires the operands
   to be in D4 and D5 and returns D4 DIV D5 in D7 and D4 MOD D5
   in D4.

   The routine for signed division is very similar and takes arguments
   and returns results in the same way. It does, however, use another
   register (D3) to store the sign of the result.

   For assembler subroutines see Modula 2 assembler arithmetic library
   in m2:assembler.arithmetic
*)

PROCEDURE DivMod (Op : Operator; VAR Opd1, Opd2, Dest : ITEM);
    VAR power : CARDINAL;
    BEGIN
        (* Op could be any one of Mod, Umod, Div or Udiv *)
        IF (Op # Mod) AND immediate (Opd1) 
            AND PowerOfTwo (Opd1.DispOrData, power, (Op=Div))
        THEN
            CASE Op OF
                Umod:   AndUmod (power, Opd2, Dest);
            |   Udiv:   ShiftMulDiv (Op, power, Opd1, Opd2, Dest);
            |   Div:    ShiftSignedDiv (power, Opd1, Opd2, Dest);
            END (* case *);
        ELSE
            NormalDivMod (Op, Opd1, Opd2, Dest);
        END (* if *);
    END DivMod;


PROCEDURE ShiftSignedDiv (power : CARDINAL; 
                            VAR ImmedOpd, OtherOpd, Dest : ITEM);
    VAR
        ShiftedReg, Shifter : ITEM;
        NegateResult : BOOLEAN;
        label1, label2 : CARDINAL;
    BEGIN
        NegateResult := ImmedOpd.DispOrData < 0;
        IF power = 0 (* DIV by 1 or -1 *) THEN
            MulDivBy1 (OtherOpd, Dest, NegateResult);
        ELSE
            Slave (OtherOpd, ShiftedReg);
            IF NOT InDataReg (ShiftedReg) THEN
                IF equal (Dest, DataReg0) THEN
                    ShiftedReg := DataReg0;
                    protect (DataReg0);
                ELSE
                    FreeReg (ShiftedReg, UseBitSet);
                END (* if *);
                GenMove (OtherOpd, ShiftedReg);
            END (* if *);
            IF power <= 8 THEN
                PackItem (Shifter, Immediate, NoReg, power, NoReg);
            ELSE
                FreeReg (Shifter, NOT UseBitSet);
                GenMoveImm (power, Shifter);
            END (* if *);
            GenTst (ShiftedReg);
            label1 := NewLabel ();
            GenRelBranch (PL, label1);
            GenNeg (ShiftedReg);
            GenShift (Shr, Shifter, ShiftedReg);
            IF NOT NegateResult THEN
                GenNeg (ShiftedReg);
            END (* if *);
            label2 := NewLabel ();
            GenRelBranch (T, label2);
            PushLabelStack (label1);
            GenShift (Shr, Shifter, ShiftedReg);
            IF NegateResult THEN
                GenNeg (ShiftedReg);
            END (* if *);
            PushLabelStack (label2);
            IF equal (Dest, NullRegItem) THEN
                SetResultItem (ShiftedReg);
            ELSE
                (* if Dest = DataReg0 the next 2 calls do no harm *)
                GenMove (ShiftedReg, Dest);
                PutIntoSlave (ShiftedReg, Dest, TRUE);
            END (* if *);
        END (* if *);
    END ShiftSignedDiv;

PROCEDURE AndUmod (power : CARDINAL; VAR OtherOpd, Dest : ITEM);
    VAR
        mask, i : CARDINAL;
        MaskItem : ITEM;
    BEGIN
        mask := 0;
        FOR i := 1 TO power DO
            mask := mask * 2;
            INC (mask);
        END (* for *);
        PackItem (MaskItem, Immediate, NoReg, mask, NoReg);
        AddAndOr (And, MaskItem, OtherOpd, Dest);
    END AndUmod;

PROCEDURE NormalDivMod (Op : Operator; VAR Opd1, Opd2, Dest : ITEM);
    VAR SlvOpd1, SlvOpd2, result : ITEM;
        subr                     : GlobalLabel;
        MinUsedReg               : REGISTER;
    BEGIN
        Slave (Opd1, SlvOpd1); Slave (Opd2, SlvOpd2);
        IF (Op = Div) OR (Op = Mod) THEN
            subr := signeddivide;
            MinUsedReg := D3;
        ELSE (* unsigned div or mod *)
            subr := divide;
            MinUsedReg := D4;
        END (* if *);
        PrepareForJsr (2, MinUsedReg);
        IF equal (SlvOpd2, DataReg5) AND equal (SlvOpd1, DataReg4) THEN
            GenExg (SlvOpd1, SlvOpd2);
        ELSIF equal (SlvOpd1, DataReg4) THEN
            GenMove (SlvOpd1, DataReg5);
            GenMove (SlvOpd2, DataReg4);
        ELSE
            GenMove (SlvOpd2, DataReg4);
            GenMove (SlvOpd1, DataReg5);
        END (* if *);
        GenJsrToGlobalLabel (subr);
        RecoverFromJsr ();
        IF (Op = Div) OR (Op = Udiv) THEN
            result := DataReg7
        ELSE (* Mod or Umod *)
            result := DataReg4
        END (* if *);
        IF equal (Dest, NullRegItem) THEN
            SetResultItem (result);
        ELSE
            GenMove (result, Dest);
            PutIntoSlave (result, Dest, FALSE);
        END (* if *);
   END NormalDivMod;

PROCEDURE PowerOfTwo (val : WORD; VAR power : CARDINAL; signed : BOOLEAN) 
                                                                    : BOOLEAN;
    VAR c : CARDINAL;
    BEGIN
        IF signed THEN
            c := ABS (INTEGER (val));
        ELSE
            c := CARDINAL (val);
        END (* if *);
        power := 0;
        IF c # 0 THEN
            WHILE c MOD 2 = 0 DO
                c := c DIV 2;
                INC (power);
            END (* if *);
        END (* if *);
        RETURN (c = 1);
    END PowerOfTwo;

VAR BottomOfPsStkOnRealStack : BOOLEAN;
    LowerBoundRegister : REGISTER;
    BottomItem, SafePlace : ITEM;
    
(*  This procedure prepares the registers and stack for a call to one of the
    standard subroutines. This preparation must be done because such a call
    may :
        1)  Write over valuable data held just above the stackpointer.
        2)  Write over valuable data held in the data registers.
    To stop this happening this procedure stores any valuable data in the
    location above the stackpointer in a safe location (further above the top
    of the stack) and frees enough registers for the standard subroutine to 
    work in. Note that in the case of, say, division one of the registers to
    be freed may contain an operand to the division subroutine. In this case
    the value of the register is NOT saved in store - it is assumed that it
    will be taken care of elsewhere.                                        *)

PROCEDURE PrepareForJsr (ArgsOnPsStk : CARDINAL; LWBreg : REGISTER);

    PROCEDURE MustBeSaved (VAR item : ITEM) : BOOLEAN;
        VAR reg : REGISTER;
        BEGIN
            IF InDataReg (item) THEN
                reg := item.Register;
            ELSE
                RETURN FALSE;
            END (* if *);
            RETURN (LWBreg <= reg) AND (reg <= D7)
        END MustBeSaved;
    VAR count : CARDINAL;
        reg   : REGISTER;
        found : BOOLEAN;
    
    BEGIN
        LowerBoundRegister := LWBreg;
        IF PseudoStackEmpty () OR ((top - ArgsOnPsStk) <= base) THEN
            RETURN
        END (* if *);
        BottomItem := PseudStk [base+1];
        IF OnRealStack (BottomItem) OR MustBeSaved (BottomItem) THEN
            reg := D0; found := FALSE;
            WHILE (reg < LWBreg) AND (NOT found) DO
                IF IsFree (reg) THEN
                    found := TRUE;
                    MarkOnStack (reg);
                    PackReg (SafePlace, reg);
                ELSE
                    INC (reg)
                END (* if *);
            END (* while *);
            IF found THEN
                GenMove (BottomItem, SafePlace);
                PseudStk [base+1] := SafePlace;
            ELSE
ReportError (FALSE, "AuxGen3 moving onto real stack", 0, 0);
		MoveIntoScratchSpace (base + 1);
(*                PackRealStackItem (SafePlace, top + 2);
                BottomOfPsStkOnRealStack := TRUE;
                GenMove (BottomItem, SafePlace);   *)
            END (* if *);
        END (* if *);
        FOR count := base + 2 TO top - ArgsOnPsStk DO
            IF MustBeSaved (PseudStk [count]) THEN
                MoveIntoScratchSpace (count);
            END (* if *);
        END (* for *);
    END PrepareForJsr;

PROCEDURE RecoverFromJsr ();
    VAR reg : REGISTER;
    BEGIN
        FOR reg := LowerBoundRegister TO D7 DO
            SetRegFree (reg);
        END (* for *);
        IF BottomOfPsStkOnRealStack THEN
ReportError (FALSE, "AuxGen3 moving to real stack in recover from JSR", 0, 0);
            PackRealStackItem (BottomItem, base + 1);
            GenMove (SafePlace, BottomItem);
            PseudStk [base + 1] := BottomItem;
            BottomOfPsStkOnRealStack := FALSE;
        END (* if *);
    END RecoverFromJsr;

BEGIN
    BottomOfPsStkOnRealStack := FALSE;
    PackReg (DataReg4, D4);
    PackReg (DataReg5, D5);
    PackReg (DataReg7, D7);
END AuxGen3.
