IMPLEMENTATION MODULE Float;

FROM items IMPORT
    ITEM, MODE, REGISTER, InDataReg, PackRealStackItem, BytesPerWord,
    OnRealStack, GlobalLabel;
FROM CodeProcs IMPORT next;
FROM PsStk IMPORT
    DecTopOfPsStk, DoAnyPendingOp, Push, PseudStk, top, base,
    MoveIntoScratchSpace;
FROM RegSlv IMPORT
    slave, SetRegFree;
FROM GenProcs IMPORT
    GenMove, GenMoveToReg, GenExg, GenAddaImm, GenSubaImm, GenJsrToGlobalLabel;

FROM UtilError IMPORT ReportError;
FROM WriteF IMPORT H;
FROM MCMnemonics IMPORT FADD, FSUB, FMUL, FDIV, FCMP, FABS, FNEG, FFCT;

PROCEDURE float (instr : CARDINAL);
    VAR
        i : CARDINAL;
        s : ARRAY [0..5] OF CHAR;
    BEGIN
        DoAnyPendingOp ();
        CASE instr OF

            FADD:
            CallFPRoutine (TRUE, FALSE, FALSE, D0, D1, D7, NoReg, fadd);

        |   FSUB:
            CallFPRoutine (TRUE, FALSE, TRUE, D1, D0, D7, NoReg, fsub);

        |   FMUL:
            CallFPRoutine (TRUE, FALSE, FALSE, D0, D1, D7, NoReg, fmul);

        |   FDIV:
            CallFPRoutine (TRUE, FALSE, TRUE, D1, D0, D7, NoReg, fdiv);

        |   FCMP:
            CallFPRoutine (TRUE, TRUE, TRUE, D0, D1, D0, D1, fcmp);

        |   FABS:
            CallFPRoutine (FALSE, FALSE, FALSE, D0, NoReg, D0, NoReg, fabs);

        |   FNEG:
            CallFPRoutine (FALSE, FALSE, FALSE, D0, NoReg, D0, NoReg, fneg);

        |   FFCT: (* FLOAT and TRUNC *)
            i := next ();
            IF (i=1) OR (i=3) THEN
                IF i = 1 THEN s := "FLOAT" ELSE s := "TRUNC" END;
                ReportError
                    (TRUE, "Double length %S not implemented*N", H(s), 0);
            ELSE
                IF i = 0 THEN
                    CallFPRoutine (FALSE, FALSE, FALSE, D0, NoReg, D0, NoReg, ffloat);
                ELSE
                    CallFPRoutine (FALSE, FALSE, FALSE, D0, NoReg, D0, NoReg, ftrunc);
                END (* if *);
            END (* if *);
        END (* case *);
    END float;

PROCEDURE CallFPRoutine (TwoArgs, TwoResults, ArgOrderMatters : BOOLEAN;
                         arg1, arg2, res1, res2 : REGISTER;
                         func : GlobalLabel);
    VAR
        Save, Temp, Arg1, Arg2 : ITEM;
        i, NoOfArgs : CARDINAL;
        reg : REGISTER;
        ItemsSavedOnRealStack : BOOLEAN;
    BEGIN
        IF TwoArgs THEN NoOfArgs := 2 ELSE NoOfArgs := 1 END;
        ItemsSavedOnRealStack := FALSE;
        IF top > NoOfArgs THEN
            FOR i := base+1 TO top - NoOfArgs DO
                Temp := PseudStk [i];
                IF InDataReg (Temp) THEN
		    MoveIntoScratchSpace (i);
(*                    PackRealStackItem (Save, i);
                    GenMove (Temp, Save);
                    PseudStk [i] := Save;             *)
                    ItemsSavedOnRealStack := TRUE;
                ELSIF OnRealStack (Temp) THEN
                    ItemsSavedOnRealStack := TRUE;
                END (* if *);
            END (* for *);
        END (* if *);
        Arg1 := PseudStk [top];
        slave (Arg1);
        IF TwoArgs THEN
            Arg2 := PseudStk [top-1];
            slave (Arg2);
            Get2ArgsReady (Arg1, Arg2, arg1, arg2, ArgOrderMatters);
        ELSE
            GenMoveToReg (Arg1, arg1);
        END (* if *);
        DecTopOfPsStk (NoOfArgs);
        IF ItemsSavedOnRealStack THEN
            GenSubaImm (top * BytesPerWord, S);
            GenJsrToGlobalLabel (func);
            GenAddaImm (top * BytesPerWord, S);
        ELSE
            GenJsrToGlobalLabel (func);
        END (* if *);
        FOR reg := D0 TO D7 DO
            SetRegFree (reg);
        END (* for *);
        Push (DataReg, res1, 0);
        IF TwoResults THEN
            Push (DataReg, res2, 0);
        END (* if *);
    END CallFPRoutine;

PROCEDURE Get2ArgsReady (VAR Arg1, Arg2 : ITEM; arg1, arg2 : REGISTER;
                            ArgOrderMatters : BOOLEAN);
    BEGIN
        IF ArgOrderMatters THEN
            IF InDataReg (Arg2) AND (Arg2.Register = arg1) THEN
                IF InDataReg (Arg1) AND (Arg1.Register = arg2) THEN
                    GenExg (Arg1, Arg2);
                ELSE
                    GenMoveToReg (Arg2, arg2);
                    GenMoveToReg (Arg1, arg1);
                END (* if *);
            ELSE
                GenMoveToReg (Arg1, arg1);
                GenMoveToReg (Arg2, arg2);
            END (* if *);
        ELSE
            IF InDataReg (Arg2) AND (Arg2.Register = arg1) THEN
                GenMoveToReg (Arg1, arg2);
            ELSE
                GenMoveToReg (Arg1, arg1);
                GenMoveToReg (Arg2, arg2);
            END (* if *)
        END (* if *);
    END Get2ArgsReady;

END Float.
    


