IMPLEMENTATION MODULE Misc;

FROM items IMPORT ITEM, REGISTER, MODE, ConditionCode, Operator, InDataReg, 
                  OnRealStack, UsesSlavedAddrReg, GlobalLabel,
                  StackPreDec, PackRealStackItem,
                  BytesPerWord, DataReg0, NILvalue, address, immediate,
                  PackItem, equal;
FROM Exceptions IMPORT StdException;
FROM PsStk IMPORT DoAnyPendingOp, PseudStk, top, base, DecTopOfPsStk, 
                  DyadicOpPending, MonadicOpPending, pop, push, Push,
                  AssertEmptyPsStk, NotOnPsStk, PsStkState, STATE, STATEBIT,
		  ScratchHwm;
FROM RegSlv IMPORT FreeReg, UseBitSet, slave, protect,
                   PutIntoSlave, DeleteFromSlave, ResetUseBits;
FROM GenProcs IMPORT GenMove, GenAddaImm, GenCmp, GenRelBranch, GenMoveQ,
                     GenJsrToGlobalLabel, GenTst, GenMoveToReg, GenSubaImm,
                     GenSubImm, GenDBra, GenPea, GenLea, GenMovea, GenLabel,
                     GenShiftImm, GenBsrToRangeCheck, SetFrameSize;
FROM AddrReg IMPORT IncUseCount, GetAddrRegsForBlockMove, InAddrRegSlave,
                    ReleaseAddrReg, FreeAddrReg, AddressMove;
FROM CodeProcs IMPORT next;
FROM Relations IMPORT compare;
FROM SimpleStks IMPORT Stack, PushOnto, PopFrom, initialise;
FROM MainPass IMPORT ExtraBytesProcessed;
FROM labels IMPORT PushLabelStack, NewLabel;
FROM UtilError IMPORT ReportError;
FROM MCMnemonics IMPORT
    UCHK, CHK, TRAP, CHKZ, CHKS, RDS, LODFW, STORE, STOFV, STOT, COPT,
    DECS, LIN, MOV, LSWn, LSW, LSD0, LSA, ADD, UADD;
FROM Procs IMPORT ParmSpace, InitSize;
CONST stofv = TRUE;

VAR TopStack : Stack;

PROCEDURE traps (instruction : CARDINAL);
BEGIN
    CASE instruction OF
      TRAP : Trap ()
    | CHK  : SignedBoundsCheck ()
    | CHKZ : CheckZ ()
    | CHKS : CheckSign ()
    ELSE UnsignedBoundsCheck ()
    END;
END traps;
    
PROCEDURE Trap ();
    VAR TopItem : ITEM;
    BEGIN
        DoAnyPendingOp ();
        TopItem := PseudStk [top];
        DecTopOfPsStk (1);
        IF immediate (TopItem) AND
	   (TopItem.DispOrData = INTEGER (ORD (ArrayIndexOutOfRange))) THEN
            (* range trap *)
            GenBsrToRangeCheck;
        ELSE
            slave (TopItem);
            GenMoveToReg (TopItem, D7);
            GenJsrToGlobalLabel (trap);
        END (* if *);
    END Trap;

PROCEDURE CheckZ ();
    VAR Upb, Index : ITEM;
        cc : ConditionCode;
        label : CARDINAL;
    BEGIN
        DoAnyPendingOp ();
        Upb := PseudStk [top];
        Index := PseudStk [top - 1];
        DecTopOfPsStk (1);
        IF immediate (Upb) AND immediate (Index) THEN
            IF Index.DispOrData > Upb.DispOrData THEN
                ReportError (TRUE,
                       "Constant array index or set element out of range*N",
                             0, 0);
            END (* if *);
        ELSE
            cc := compare (Uleq, Upb, Index);
            label := NewLabel ();
            GenRelBranch (cc, label);
            GenBsrToRangeCheck;
            PushLabelStack (label);
        END (* if *);
    END CheckZ;

PROCEDURE CheckSign ();
    VAR TopItem : ITEM;
        label : CARDINAL;
    BEGIN
        DoAnyPendingOp ();
        TopItem := PseudStk [top];
        slave (TopItem);
        GenTst (TopItem);
        label := NewLabel ();
        GenRelBranch (PL, label);
        GenBsrToRangeCheck;
        PushLabelStack (label);
    END CheckSign;

PROCEDURE UnsignedBoundsCheck ();
    BEGIN
        BoundsCheck (HI, CC);
    END UnsignedBoundsCheck;

PROCEDURE SignedBoundsCheck ();
    BEGIN
        BoundsCheck (GT, GE);
    END SignedBoundsCheck;

PROCEDURE BoundsCheck (GreaterThan, GreaterThanOrEqual : ConditionCode);
    VAR Upb, Lwb, Index, RegItem : ITEM;
        label1, label2 : CARDINAL;
    BEGIN
        DoAnyPendingOp ();
        Upb := PseudStk [top];
        Lwb := PseudStk [top - 1];
        Index := PseudStk [top - 2];
        DecTopOfPsStk (2);
        slave (Upb); slave (Lwb); slave (Index);
        IF NOT InDataReg (Index) THEN
            protect (Upb); protect (Lwb);
            FreeReg (RegItem, NOT UseBitSet);
            GenMove (Index, RegItem);
            PutIntoSlave (RegItem, Index, FALSE);
            Index := RegItem;
        END (* if *);
        label1 := NewLabel (); label2 := NewLabel ();
        GenCmp (Upb, Index);
        GenRelBranch (GreaterThan, label1);
        GenCmp (Lwb, Index);
        GenRelBranch (GreaterThanOrEqual, label2);
        PushLabelStack (label1);
        GenBsrToRangeCheck;
        PushLabelStack (label2);
    END BoundsCheck;

PROCEDURE misc (instruction : CARDINAL);
    BEGIN
        IF instruction = LIN THEN
            GenLabel (); Push (Immediate, NoReg, NILvalue);    
        ELSIF instruction = MOV THEN
            BlockMove ()
        ELSE
            CASE instruction OF
              LODFW : lodfw ()
            | STORE : SavePseudStk (NOT stofv)
            | STOFV : SavePseudStk (stofv)
            | STOT  : CopyToRealStack ()
            | COPT : CopyTopOfStack ()
            END (* case *);
        END;
    END misc;

PROCEDURE ResetStaticStackHeight;
    BEGIN
        StaticStackHeight := 0
    END ResetStaticStackHeight;

PROCEDURE CopyToRealStack ();
    VAR TopItem, RegItem, RealStackPosn, OldPos : ITEM;
        count : CARDINAL;
    BEGIN
        DoAnyPendingOp ();
        IF next () = DECS THEN
            (* take no action at all *)
            DecTopOfPsStk (1);
            ExtraBytesProcessed := 1;
            RETURN;
        ELSE INCL (PsStkState[top], STOTed);
        END (* if *);
    END CopyToRealStack;

PROCEDURE CopyTopOfStack ();
    VAR PsStkTop, Reg : ITEM;
        AddrReg : REGISTER;
        nextinst : CARDINAL;
    BEGIN
        DoAnyPendingOp ();
        PsStkTop := PseudStk [top];
        IF InDataReg (PsStkTop) OR OnRealStack (PsStkTop) THEN
            nextinst := next ();
            IF ((LSWn+0 <= nextinst) AND (nextinst <= LSWn+15)) OR
               ((LSW <= nextinst) AND (nextinst <= LSD0)) OR
               (nextinst = LSA) (* variation on LSW, LSA, LSD *) THEN
                DecTopOfPsStk (1);
                AddrReg := FreeAddrReg (NotOnPsStk);
                GenMovea (PsStkTop, AddrReg);
                Push (ADDRaddrRegDisp, AddrReg, 0);
                IncUseCount (AddrReg, top);
                Push (ADDRaddrRegDisp, AddrReg, 0);
                IncUseCount (AddrReg, top);
            ELSIF (nextinst = UADD) OR (nextinst = ADD) THEN
                ExtraBytesProcessed := 1;
                IF OnRealStack (PsStkTop) THEN
                    FreeReg (Reg, NOT UseBitSet);
                    GenMove (PsStkTop, Reg);
                    DecTopOfPsStk (1);
                    push (Reg);
                END (* if *);
                GenShiftImm (Shl, 1, PseudStk [top]);
            ELSE
                FreeReg (Reg, NOT UseBitSet);
                GenMove (PsStkTop, Reg);
                push (Reg);
            END (* if *);
        ELSE 
            push (PsStkTop);
            IF UsesSlavedAddrReg (PsStkTop) THEN
                IncUseCount (PsStkTop.Register, top);
            END;
        END (* if *);
    END CopyTopOfStack;

PROCEDURE lodfw ();
    BEGIN
        RestorePseudStk ();
        push (DataReg0);
    END lodfw;

PROCEDURE SavePseudStk (IsStofv : BOOLEAN);

(*    PROCEDURE AdjustStackPointer ();
        BEGIN
            IF StackPointerOffset <> 0 THEN
                GenSubaImm (StackPointerOffset, S);
                StackPointerOffset := 0;
            END (* if *);
        END AdjustStackPointer;
*)
    VAR point, STOTcount, GBcount, StackPointerOffset : CARDINAL;
        CurrentItem : ITEM;
        save, dest : ITEM;  
    BEGIN
        DoAnyPendingOp ();
	IF IsStofv THEN
	    save := PseudStk [top];
	    DecTopOfPsStk (1);
	END;
(*        StackPointerOffset := 0;   *)
        FOR point := base + 1 TO top DO
	    PackItem (dest, AddrRegDisp, L, -(INTEGER (point - base)
		   + InitSize - ParmSpace + StaticStackHeight), NoReg);
            CurrentItem := PseudStk [point];
         IF address (CurrentItem) THEN
	    AddressMove (CurrentItem, dest)
         ELSIF NOT equal (CurrentItem, dest) THEN
            GenMove (CurrentItem, dest)
	 END;
(*            IF OnRealStack (CurrentItem) THEN
                INC (StackPointerOffset, BytesPerWord);
              ELSE
                AdjustStackPointer ();
                IF address (CurrentItem) THEN
                    GenPea (CurrentItem);
                ELSE
                    GenMove (PseudStk [point], StackPreDec);
                END (* if *);
            END (* if *);
*)
        END (* for *);
	StaticStackHeight := StaticStackHeight + INTEGER (top);
	IF CARDINAL (StaticStackHeight) > ScratchHwm
	THEN ScratchHwm := CARDINAL (StaticStackHeight);
	    SetFrameSize (ScratchHwm)
        END;
(*        AdjustStackPointer ();  *)
        GBcount := 0;
        FOR point := base + 1 TO top DO
            IF GB IN PsStkState [point] THEN
                PushOnto (TopStack, point);
                INC (GBcount);
            END (* if *);
        END (* for *);
        PushOnto (TopStack, GBcount);

	STOTcount := 0;
	FOR point := base + 1 TO top DO
	    IF STOTed IN PsStkState [point] THEN
		PushOnto (TopStack, point);
		INC (STOTcount);
	    END (* if *);
	    PsStkState [point] := STATE {};
	END (* for *);
	PushOnto (TopStack, STOTcount);
(*        IF IsStofv THEN
            PushOnto (TopStack, top - 1);
        ELSE
            PushOnto (TopStack, top);
        END (* if *);			*)



	PushOnto (TopStack, top);


        DecTopOfPsStk (top);
	IF IsStofv THEN
	    push (save);
	    INCL (PsStkState [top], STOTed)
	END;
	    
    END SavePseudStk;

PROCEDURE RestorePseudStk ();
    VAR GBcount, point, STOTcount : CARDINAL;
        StackPosition, source : ITEM;
    BEGIN
        top := PopFrom (TopStack);
(*        GenAddaImm (top * BytesPerWord, S);    *)
	StaticStackHeight := StaticStackHeight - INTEGER (top);
        FOR point := base + 1 TO top DO
	    PackItem (source, AddrRegDisp, L, -(INTEGER (point - base)
		+ InitSize - ParmSpace + StaticStackHeight), NoReg);
	    PseudStk [point] := source;
(*            PackRealStackItem (StackPosition, point);
            PseudStk [point] := StackPosition;		*)
            PsStkState [point] := STATE {};
        END (* for *);

	STOTcount := PopFrom (TopStack);
	FOR point := 1 TO STOTcount DO
	    INCL (PsStkState [PopFrom (TopStack)], STOTed);
	END (* FOR *);
        GBcount := PopFrom (TopStack);
        FOR point := 1 TO GBcount DO
            INCL (PsStkState [PopFrom (TopStack)], GB);
        END (* for *);
    END RestorePseudStk;

PROCEDURE BlockMove ();
    VAR Size, Source, Dest, Reg : ITEM;
        label : CARDINAL;
    BEGIN
        DoAnyPendingOp ();
        Size := PseudStk [top];
        Source := PseudStk [top - 1];
        Dest := PseudStk [top - 2];
        DecTopOfPsStk (3);
        protect (Source); protect (Dest);
        IF immediate (Size) THEN
            FreeReg (Reg, NOT UseBitSet);
            DEC (Size.DispOrData);
            GenMove (Size, Reg);
            Size := Reg;
        ELSE ReportError (TRUE, "unexpected MOV", 0, 0);
            (*IF NOT InDataReg (Size) THEN
                slave (Size);
                IF InDataReg (Size) THEN
                    DeleteFromSlave (Size.Register);
                ELSE
                    FreeReg (Reg, NOT UseBitSet);
                    GenMove (Size, Reg);
                    Size := Reg;
                END (* if *);
            END (* if *);
            GenSubImm (1, Size);*)
        END;
        GetAddrRegsForBlockMove (Source, Dest);
        label := NewLabel ();
        PushLabelStack (label);
        GenMove (Source, Dest);
        GenDBra (Reg.Register, label);
        ResetUseBits ();
    END BlockMove;

BEGIN
    initialise (TopStack);
END Misc.
