IMPLEMENTATION MODULE MonGen;

FROM items IMPORT ITEM, MonOperator, equal, NullRegItem, InDataReg,
                  Operator, REGISTER, ConditionCode, DataReg0;
FROM GenOp IMPORT SetResultItem;
FROM RegSlv IMPORT Slave, FreeReg, DeleteFromSlave, UseBitSet, StopSlaving,
                   PutIntoSlave;
FROM GenProcs IMPORT GenMove, GenXorImm, GenCom, GenNeg, GenExg, GenMoveImm,
                     GenMoveQ, GenShift, GenRelBranch, GenTst;
FROM labels IMPORT NewLabel, PushLabelStack;

PROCEDURE GenAbsSequence (VAR opd : ITEM);
    VAR label : CARDINAL;
    BEGIN
        GenTst (opd);
        label := NewLabel ();
        GenRelBranch (PL, label);
        GenNeg (opd);
        PushLabelStack (label);
    END GenAbsSequence;

PROCEDURE AbsoluteValue (VAR Opd, Dest : ITEM);
    VAR SlvOpd, Reg : ITEM;
    BEGIN
        Slave (Opd, SlvOpd);
        IF InDataReg (SlvOpd) THEN
            DeleteFromSlave (SlvOpd.Register);
            GenAbsSequence (SlvOpd);
            IF equal (NullRegItem, Dest) THEN
                SetResultItem (SlvOpd);
            ELSE
                GenMove (SlvOpd, Dest);
                PutIntoSlave (SlvOpd, Dest, TRUE);
            END (* if *);
        ELSIF equal (NullRegItem, Dest) THEN
            FreeReg (Reg, NOT UseBitSet);
            GenMove (Opd, Reg);
            GenAbsSequence (Reg);
            SetResultItem (Reg);
        ELSE
            GenMove (Opd, Dest);
            GenAbsSequence (Dest);
            StopSlaving (Dest);
        END (* if *);
    END AbsoluteValue;

PROCEDURE SimpleMonadicOps (Op : MonOperator; VAR Opd, Dest : ITEM);
    VAR SlvOpd, Reg : ITEM;
    BEGIN
        Slave (Opd, SlvOpd);
        IF equal (Dest, NullRegItem) THEN
            IF InDataReg (SlvOpd) THEN
                DeleteFromSlave (SlvOpd.Register);
                SetResultItem (SlvOpd);
                SimpleMonGen (Op, SlvOpd);
            ELSE
                FreeReg (Reg, NOT UseBitSet);
                SetResultItem (Reg);
                GenMove (Opd, Reg);
                SimpleMonGen (Op, Reg);
            END (* if *);
        ELSIF equal (Dest, DataReg0) THEN
            GenMove (SlvOpd, DataReg0);
            SimpleMonGen (Op, DataReg0);
        ELSE
            IF equal (Opd, Dest) THEN
                StopSlaving (Dest);
                SimpleMonGen (Op, Dest);
            ELSIF InDataReg (SlvOpd) THEN
                SimpleMonGen (Op, SlvOpd);
                GenMove (SlvOpd, Dest);
                PutIntoSlave (SlvOpd, Dest, TRUE);
            ELSE
                GenMove (SlvOpd, Dest);
                SimpleMonGen (Op, Dest);
                StopSlaving (Dest);
            END (* if *);
        END (* if *);
    END SimpleMonadicOps;

PROCEDURE SimpleMonGen (Op : MonOperator; VAR opd : ITEM);
    BEGIN
        CASE Op OF
          
          Not   :   GenXorImm (1, opd)      |
          Com   :   GenCom (opd)            |
          Neg   :   GenNeg (opd)
          
        END (* case *);
    END SimpleMonGen;

PROCEDURE ShiftMonadicOps (Op : MonOperator; VAR Opd, Dest : ITEM);
    VAR SlvOpd, ShiftByReg, ShiftedReg : ITEM;
    BEGIN
        Slave (Opd, SlvOpd);
        IF InDataReg (SlvOpd) THEN
            ShiftByReg := SlvOpd;
        ELSE
            FreeReg (ShiftByReg, UseBitSet);
            GenMove (Opd, ShiftByReg);
            PutIntoSlave (ShiftByReg, Opd, FALSE);
        END (* if *);
        IF equal (Dest, DataReg0) THEN
            IF equal (ShiftByReg, DataReg0) THEN
                FreeReg (ShiftedReg, NOT UseBitSet);
                GenExg (ShiftedReg, ShiftByReg);
                ShiftByReg := ShiftedReg;
                ShiftedReg := DataReg0;
            END (* if *);
        ELSE
            FreeReg (ShiftedReg, NOT UseBitSet);
        END (* if *);
        IF Op = Bit THEN
            GenMoveQ (1, ShiftedReg.Register);
            GenShift (Shl, ShiftByReg, ShiftedReg);
        ELSE (* Op = Msk *)
            GenMoveImm (-1, ShiftedReg);
            GenShift (Shl, ShiftByReg, ShiftedReg);
            GenCom (ShiftedReg);
        END (* if *);
        IF equal (Dest, NullRegItem) OR equal (Dest, DataReg0) THEN
            SetResultItem (ShiftedReg);
        ELSE
            GenMove (ShiftedReg, Dest);
            PutIntoSlave (ShiftedReg, Dest, TRUE);
        END (* if *)
    END ShiftMonadicOps;
    
END MonGen.
