IMPLEMENTATION MODULE Object;
(* $T-,$R- *)
FROM SYSTEM IMPORT
    BYTESPERWORD, TSIZE, WORD, ADDRESS, ADR;
FROM Storage IMPORT 
    ALLOCATE;
FROM Jumps IMPORT
    RegJustLoadedOnStack;
FROM SimpleStks IMPORT
    Stack, PushOnto, PopFrom, initialise, empty;
FROM labels IMPORT
    PushLabelStack, PopLabelStack, none, NewLabel, GetPermanentLabelTable;
FROM UtilError IMPORT
    ReportError;
FROM WriteF IMPORT R;
FROM CharCodes IMPORT SpaceCh;
(* FROM BBulkIO IMPORT BulkWrite; *)
FROM ReadMcd IMPORT 
    MCDERROR, IMPORTFRAMEPTR, IMPORTFRAME, MODULENAME,
    KEY, KEYSIZE, MODULENAMESIZE,
    CodeSize, ProcadrSize, CodeByte, McdError, Importlist, ImportlistLength,
    Headerframe, StringSize;
FROM items IMPORT GlobalLabel;
FROM UnixCalls IMPORT
  exit, fork, read, write, open, close, wait,
  creat, link, unlink, exec, chdir, time, 
  mknod, chmod, chown, break, stat, lseek, 
  getpid, mount, umount, setuid, getuid, stime,
  ptrace, alarm, fstat, pause, utime, smdate,
  stty, gtty, access, nice, ftime, sync, kill,
  switch, setpgrp, tell, dup, pipe, times, profil,
  setgid, getgid, signal, acct, setphys, syslock,
  ioctl, reboot, mpx, exece, umask, chroot, vfork, 
  vread, vwrite, brk, sbrk;

IMPORT Strings;
(*                          BITSET Utilities                                 *)

CONST
    shl1    = 2;        shl2    = 4;        shl3    = 8;
    shl4    = 16;       shl5    = 32;       shl6    = 64;
    shl7    = 128;      shl8    = 256;      shl9    = 512;
    shl10   = 1024;     shl11   = 2048;     shl12   = 4096;
    shl13   = 8192;     shl14   = 16384;    shl15   = 32768;

    Max16BitCard    = 65535;
    Max16BitInt     = 32767;
    MaxNeg16BitInt  = Max16BitInt + 1;
    Max8BitCard     = 255;
    Max8BitInt      = 127;
    MaxNeg8BitInt   = Max8BitInt + 1;

    SkipGSetUp      = 6;	(* used to bypass module data base setup
				   when calling from inside module      *)

PROCEDURE ObjCodeAddress () : ADDRESS;
    BEGIN
	RETURN ADDRESS (ObjCode);
    END ObjCodeAddress;

PROCEDURE ObjCodeBytes () : CARDINAL;
    BEGIN
  	RETURN ObjPtr - ProcAddrTableAddress;
    END ObjCodeBytes;
    
PROCEDURE ShiftLeft (b : BITSET; by : CARDINAL) : BITSET;
    BEGIN
        CASE by OF
            1:  RETURN BITSET (CARDINAL (b) * shl1);
        |   2:  RETURN BITSET (CARDINAL (b) * shl2);
        |   3:  RETURN BITSET (CARDINAL (b) * shl3);
        |   4:  RETURN BITSET (CARDINAL (b) * shl4);
        |   5:  RETURN BITSET (CARDINAL (b) * shl5);
        |   6:  RETURN BITSET (CARDINAL (b) * shl6);
        |   7:  RETURN BITSET (CARDINAL (b) * shl7);
        |   8:  RETURN BITSET (CARDINAL (b) * shl8);
        |   9:  RETURN BITSET (CARDINAL (b) * shl9);
        |   10: RETURN BITSET (CARDINAL (b) * shl10);
        |   11: RETURN BITSET (CARDINAL (b) * shl11);
        |   12: RETURN BITSET (CARDINAL (b) * shl12);
        |   13: RETURN BITSET (CARDINAL (b) * shl13);
        |   14: RETURN BITSET (CARDINAL (b) * shl14);
        |   15: RETURN BITSET (CARDINAL (b) * shl15);
        ELSE
            ReportError (TRUE, "Cant shift by %R0 using ShiftLeft*N", by, 0);
        END (* case *);
    END ShiftLeft;

PROCEDURE TooBig (i : INTEGER; bits : CARDINAL);
    BEGIN
        ReportError 
            (TRUE, "Too large an argument (%N) to Signed%R0Bit", i, bits);
    END TooBig;

PROCEDURE Signed16Bit (i : INTEGER) : BITSET;
    VAR absi : CARDINAL;
    BEGIN
        IF i < 0 THEN
            absi := -i;
            IF absi > MaxNeg16BitInt THEN
                TooBig (i, 16);
            ELSE
                RETURN BITSET (Max16BitCard - absi + 1);
            END;
        ELSE
            IF i > Max16BitInt THEN
                TooBig (i, 16);
            ELSE
                RETURN BITSET (i);
            END (* if *);
        END (* if *);
    END Signed16Bit;

PROCEDURE Signed8Bit (i : INTEGER) : BITSET;
    VAR absi : CARDINAL;
    BEGIN
        IF i < 0 THEN
            absi := -i;
            IF absi > MaxNeg8BitInt THEN
                TooBig (i, 8);
            ELSE
                RETURN BITSET (Max8BitCard - absi + 1);
            END;
        ELSE
            IF i > Max8BitInt THEN
                TooBig (i, 8);
            ELSE
                RETURN BITSET (i);
            END (* if *);
        END (* if *);
    END Signed8Bit;


(*                              LABELLING                                    *)

TYPE
    LabListPtr =
        POINTER TO LabList;
    LabList =
        RECORD
            Label:      CARDINAL;
            Position:   CARDINAL;
            Type:       LabelUseType;
            Next:       LabListPtr;
            Prev:       LabListPtr;
        END (* record *);

VAR
    LabListPool : LabListPtr;

PROCEDURE InitLL (VAR LL : LabListPtr);
    BEGIN
        LL := NIL;
    END InitLL;

PROCEDURE AddToLL (VAR LL : LabListPtr; label, posn : CARDINAL;
                                                    type : LabelUseType);
    VAR newLL : LabListPtr;
    BEGIN
        IF LabListPool = NIL THEN
            NEW (newLL);
        ELSE
            newLL := LabListPool;
            LabListPool := LabListPool^.Next;
        END (* if *);
        WITH newLL^ DO
            Label := label;
            Position := posn;
            Type := type;
            Prev := NIL;
            Next := LL;
        END (* with *);
        IF LL # NIL THEN
            LL^.Prev := newLL;
        END (* if *);
        LL := newLL;
    END AddToLL;
    
PROCEDURE DeleteFromLL (VAR LL, ptr : LabListPtr);
    VAR OldNext : LabListPtr;
    BEGIN
        WITH ptr^ DO
            OldNext := Next;
            IF Next # NIL THEN
                Next^.Prev := Prev;
            END (* if *);
            IF Prev # NIL THEN
                Prev^.Next := Next;
            ELSE
                LL := Next;
            END (* if *);
            Next := LabListPool;
        END (* with *);
        LabListPool := ptr;
        ptr := OldNext;
    END DeleteFromLL;

CONST
    LotsOfLabels = 200;

VAR
    PermanentLabels : POINTER TO ARRAY [0..LotsOfLabels] OF CARDINAL;

    TransitionLabel, LastLabel : CARDINAL;

    UndefinedList, TempLabelList : LabListPtr;

PROCEDURE label (lab : CARDINAL);
    BEGIN
        IF (lab # LastLabel) THEN
            PushLabelStack (lab);
            LastLabel := lab;
        END (* if *);
    END label;

PROCEDURE UseOfLabel (label, posn : CARDINAL; TypeOfUse : LabelUseType);
    VAR value : CARDINAL;
    BEGIN
        value := 0;
        IF label > TransitionLabel THEN
            value := SearchTempLabelList (label);
        ELSE
            value := PermanentLabels^[label];
        END (* if *);
        IF value = 0 THEN
            PutOnUndefinedList (label, posn, TypeOfUse);
        ELSE
            InsertLabelValue (posn, value, TypeOfUse);
        END (* if *);
    END UseOfLabel;

PROCEDURE BackwardBranch (label : CARDINAL; VAR short : BOOLEAN) : BOOLEAN;
    VAR
        value : CARDINAL;
    BEGIN
        value := PermanentLabels^[label];
        IF value = 0 THEN
            RETURN FALSE;
        ELSE
            short := ShortBackwardBranch (value);
            RETURN TRUE;
        END (* if *);
    END BackwardBranch;

PROCEDURE ShortBackwardBranch (value : CARDINAL) : BOOLEAN;
    VAR disp : INTEGER;
    BEGIN
        disp := INTEGER (value) - INTEGER (ObjPtr);
        IF ABS (disp) > MaxNeg8BitInt THEN
            RETURN FALSE;
        ELSE
            ObjCode^ [ObjPtr - ProcAddrTableAddress - 1] :=
                        BYTE (Signed8Bit (disp));
            RETURN TRUE;
        END (* if *);
    END ShortBackwardBranch;

PROCEDURE DefnOfLabel (label : CARDINAL);
    VAR AlreadyUsed : BOOLEAN;
    BEGIN
        AlreadyUsed := CheckUndefinedList (label, ObjPtr);
        IF label > TransitionLabel THEN
            IF NOT AlreadyUsed THEN
                PutOnTempLabelList (label, ObjPtr);
            END (* if *);
        ELSE
            PermanentLabels^[label] := ObjPtr;
        END (* if *);
    END DefnOfLabel;

PROCEDURE CheckUndefinedList (label, DefinedPosn : CARDINAL) : BOOLEAN;
    VAR 
        SomethingFound : BOOLEAN;
        ptr : LabListPtr;
    BEGIN
        SomethingFound := FALSE;
        ptr := UndefinedList;
        WHILE ptr # NIL DO
            WITH ptr^ DO
                IF Label = label THEN
                    SomethingFound := TRUE;
                    InsertLabelValue (Position, DefinedPosn, Type);
                    DeleteFromLL (UndefinedList, ptr);
                ELSE
                    ptr := Next;
                END (* if *);
            END (* with *);
        END (* while *);
        RETURN SomethingFound;
    END CheckUndefinedList;

PROCEDURE PutOnUndefinedList (label, PosnOfUse : CARDINAL;
                                        TypeOfUse : LabelUseType);
    BEGIN
        AddToLL (UndefinedList, label, PosnOfUse, TypeOfUse);
    END PutOnUndefinedList;

PROCEDURE SearchTempLabelList (label : CARDINAL) : CARDINAL;
    VAR
        value : CARDINAL;
        ptr : LabListPtr;
    BEGIN
        value := 0;
        ptr := TempLabelList;
        WHILE ptr # NIL DO
            WITH ptr^ DO
                IF Label = label THEN
                    value := Position;
                    DeleteFromLL (TempLabelList, ptr);
                    RETURN value;
                ELSE
                    ptr := Next;
                END (* if *);
            END (* with *);
        END (* while *);
        RETURN value;
    END SearchTempLabelList;

PROCEDURE PutOnTempLabelList (label, DefnPosn : CARDINAL);
    BEGIN
        AddToLL (TempLabelList, label, DefnPosn, LabelUseType (0));
    END PutOnTempLabelList;

PROCEDURE InsertLabelValue (PosnOfUse, LabelValue : CARDINAL;
                                            TypeOfUse : LabelUseType);
    VAR offset : CARDINAL;
    BEGIN 
        offset := PosnOfUse - ProcAddrTableAddress;
        CASE TypeOfUse OF
        
            Abs32Bit:
		Place4 (offset, LabelValue);
                Relocate (PosnOfUse);

	|   QuickBSR:  (* Just like Disp16BitInBranch but skip first
			instruction in proc                          *)
                Place2 (offset, CARDINAL (
                   Signed16Bit (SkipGSetUp + INTEGER (LabelValue)
						- INTEGER (PosnOfUse))));

        |   Disp16BitInBranch:
                Place2 (offset, CARDINAL (
                   Signed16Bit (INTEGER (LabelValue) - INTEGER (PosnOfUse))));

        |   DispInBranch:
                Place2 (offset, CARDINAL (
                   Signed16Bit (INTEGER (LabelValue) - INTEGER (PosnOfUse))));
                Statistics (INTEGER (LabelValue) - INTEGER (PosnOfUse));

        |   Disp8BitInBranch:
                ObjCode^ [offset-1] := 
                    BYTE (Signed8Bit 
                        (INTEGER (LabelValue) - INTEGER (PosnOfUse)));

        |   Disp8BitInPCWithIndex:
                ObjCode^ [offset+1] :=
                    BYTE (Signed8Bit 
                        (INTEGER (LabelValue) - INTEGER (PosnOfUse)));
        
        END (* case *);
    END InsertLabelValue;

VAR ShortBranches, TotalBranches : CARDINAL;
PROCEDURE Statistics (i :INTEGER);
    BEGIN
        INC (TotalBranches);
        IF (-MaxNeg8BitInt < i) AND (i < Max8BitInt) THEN
            INC (ShortBranches);
        END (* if *);
    END Statistics;


(*                              RANGE CHECK                                  *)

VAR RangeCheckLabelValue : CARDINAL;

PROCEDURE DefineRangeCheckLabel;
    BEGIN
        RangeCheckLabelValue := ObjPtr;
    END DefineRangeCheckLabel;

PROCEDURE RangeCheckBranchDisp;
    BEGIN
        Emit (BITSET (-INTEGER (ObjPtr-RangeCheckLabelValue)));
    END RangeCheckBranchDisp;


(*                              RELOCATION                                   *)


PROCEDURE Relocate (pos : CARDINAL);
    BEGIN
        PushOnto (RelocStack, pos);
        INC (RelocCount);
    END Relocate;


(*                              EXTRNS                                       *)


TYPE String8 = ARRAY [0 .. 8] OF CHAR;

VAR NamesIndex: CARDINAL;
(*  use the OSProcNames array for the external calls,
    filling in the strings dynamically.
    Note that this limits one module to HIGH(OSProcNames) different calls.
*)

PROCEDURE NewCall(name: ARRAY OF CHAR; VAR index: CARDINAL);
(*  look in OSProcNames array for this call.
    if found.  Otherwise bump NamesIndex.
*)
    BEGIN
        index := 1;
        WHILE index <= NamesIndex DO
            IF Strings.EqualCC(name, OSProcNames[index]^) THEN
                RETURN;
            ELSE
                INC(index);
            END (* if *);
        END (* while *);
        FixedName(name, index);
    END NewCall;

PROCEDURE FixedName(name: ARRAY OF CHAR; VAR index: CARDINAL);
    BEGIN
        INC(NamesIndex); index := NamesIndex;
        IF index > HIGH(OSProcNames) THEN
            ReportError(TRUE, "too many C or UNIX calls",0,0);
        END (* if *);
        NEW(OSProcNames[index]);
        Strings.CopyCC(name, OSProcNames[index]^);
    END FixedName;


PROCEDURE UnixCallName(id: CARDINAL; VAR name: String8);
BEGIN
    CASE id OF
    	exit: name := "__exit";
    |   fork: name := "_fork";
    |   read: name := "_read";
    |   write: name := "_write";
    |   open: name := "_open";
    |   close: name := "_close";
    |   wait: name := "_wait";
    |   creat: name := "_creat";
    |   link: name := "_link";
    |   unlink: name := "_unlink";
    |   exec: name := "_execv";
    |   chdir: name := "_chdir";
    |   time: name := "_time";
    |   mknod: name := "_mknod";
    |   chmod: name := "_chmod";
    |   chown: name := "_chown";
    |   break: name := "_break";
    |   stat: name := "_stat";
    |   lseek: name := "_lseek";
    |   getpid: name := "_getpid";
    |   mount: name := "_mount";
    |   umount: name := "_umount";
    |   setuid: name := "_setuid";
    |   getuid: name := "_getuid";
    |   stime: name := "_stime";
    |   ptrace: name := "_ptrace";
    |   alarm: name := "_alarm";
    |   fstat: name := "_fstat";
    |   pause: name := "_pause";
    |   utime: name := "_utime";
    |   smdate: name := "_smdate";
    |   stty: name := "_stty";
    |   gtty: name := "_gtty";
    |   access: name := "_access";
    |   nice: name := "_nice";
    |   ftime: name := "_ftime";
    |   sync: name := "_sync";
    |   kill: name := "_kill";
    |   dup: name := "_dup";
    |   pipe: name := "_pipe";
    |   times: name := "_times";
    |   profil: name := "_profil";
    |   setgid: name := "_setgid";
    |   getgid: name := "_getgid";
    |   signal: name := "_signal";
    |   acct: name := "_acct";
    |   setphys: name := "_setphys";
    |   syslock: name := "_syslock";
    |   ioctl: name := "_ioctl";
    |   mpx: name := "_mpx";
    |   exece: name := "_exece";
    |   umask: name := "_umask";
    |   chroot: name := "_chroot";
    |   vfork: name := "_vfork";
    |   vread: name := "_vread";
    |   vwrite: name := "_vwrite";
    |   brk: name := "_brk";
    |   sbrk: name := "_sbrk";
    END;
END UnixCallName;

PROCEDURE ExtrnUNIX (Call : CARDINAL);
    VAR
        unixcallname : String8;
        index : CARDINAL;
    BEGIN
        UnixCallName (Call, unixcallname);
	NewCall (unixcallname, index);
        PushOnto (ExtrnOSStacks [index], ObjPtr);
        INC (ExtrnOSCounts [index]);	
    END ExtrnUNIX;
PROCEDURE GlobalName (gl : GlobalLabel; VAR name : String8);
    BEGIN
        CASE gl OF
	    divide: name := "_div";
	|   signeddivide: name := "_sdiv";
	|   multiply: name := "_mul";
	|   fadd: name := "_fadd";
	|   fsub: name := "_fsub";
	|   fmul: name := "_fmul";
	|   fdiv: name := "_fdiv";
	|   fcmp: name := "_fcmp";
	|   fabs: name := "_fabs";
	|   fneg: name := "_fneg";
	|   ftrunc: name := "_ftrunc";
	|   ffloat: name := "_ffloat";
	|   newprocess: name := "_newproc";
	|   transfer: name := "_transfer";
	|   gft: name := "_gft";
	|   trap: name := "_trap";
        END;
    END GlobalName;

PROCEDURE ExtrnGlCall (sub : GlobalLabel);
    VAR glcallname : String8;
        index : CARDINAL;
    BEGIN
        GlobalName (sub, glcallname);
        NewCall (glcallname, index);
	PushOnto (ExtrnOSStacks [index], ObjPtr);
	INC (ExtrnOSCounts [index]);
    END ExtrnGlCall;

PROCEDURE ExtrnNamedCall (CallName : ARRAY OF CHAR);
    VAR index : CARDINAL;
    BEGIN
	NewCall (CallName, index);
        PushOnto (ExtrnOSStacks [index], ObjPtr);
        INC (ExtrnOSCounts [index]);	
    END ExtrnNamedCall;

PROCEDURE Extrn (type : ExtrnType; datum1, datum2 : WORD);
    BEGIN
        CASE type OF

            OtherModuleBase:
                PushOnto (ExtrnMBStacks^[CARDINAL (datum1)], ObjPtr);
                INC (ExtrnMBCounts^[CARDINAL (datum1)]);

        |   OtherModuleOffset:
                PushOnto (ExtrnMOStacks^[CARDINAL (datum1)], ObjPtr);
                INC (ExtrnMOCounts^[CARDINAL (datum1)]);

        |   OtherModuleProc:
                PushOnto (ExtrnMPStacks^[CARDINAL (datum1)], CARDINAL (datum2));
                PushOnto (ExtrnMPStacks^[CARDINAL (datum1)], ObjPtr);
                INC (ExtrnMPCounts^[CARDINAL (datum1)]);

        END (* case *);
    END Extrn;


(*                                  EMITS                                    *)

CONST
    MaxObjCodeSize = 32767; (* max signed 16 bit number *)

TYPE 
    BYTE = CHAR;

VAR
    ObjCodeSize : CARDINAL; (* actual code size *)
    BytePtr     : POINTER TO ARRAY [0..(BYTESPERWORD-1)] OF BYTE;
    BitsetPtr   : POINTER TO BITSET;
    CardPtr     : POINTER TO CARDINAL;

    ObjCode     : POINTER TO ARRAY [0..(MaxObjCodeSize-1)] OF BYTE;
    WordObjCode : POINTER TO 
                    ARRAY [0..(MaxObjCodeSize DIV BYTESPERWORD)-1] OF CARDINAL;

PROCEDURE AllocateCodeBuffer (bytes: CARDINAL);
    CONST DefaultSize = 24 * 1024; (* 24k bytes *)
BEGIN
    IF bytes = 0 THEN bytes := DefaultSize END;
    IF bytes <= MaxObjCodeSize THEN
        ALLOCATE (ObjCode, TSIZE(WORD) * ((bytes DIV BYTESPERWORD) + 1));
        WordObjCode := ADDRESS (ObjCode);
        ObjCodeSize := bytes;
    ELSE ReportError (TRUE, "Code buffer too big (%N bytes)", bytes, 0)
    END;
END AllocateCodeBuffer;

PROCEDURE InstructionEmit (b : BITSET);
    VAR l : CARDINAL;
    BEGIN
        l := PopLabelStack ();
        WHILE l # none DO
            DefnOfLabel (l);
            l := PopLabelStack ();
        END (* while *);
        RegJustLoadedOnStack := FALSE;
        Emit (b);
    END InstructionEmit;

PROCEDURE Emit (b : BITSET);
    VAR offset : CARDINAL;
	w : CARDINAL;
    BEGIN
        offset := ObjPtr - ProcAddrTableAddress;
        IF offset + 1 >= ObjCodeSize THEN
            CodeOverflow;
        END (* if *);
	Place2 (offset, CARDINAL (b));
        INC (ObjPtr, 2);
    END Emit;

PROCEDURE Patch (offset : CARDINAL; b : BITSET);
    BEGIN
        Place2 (offset - ProcAddrTableAddress, CARDINAL (b));
    END Patch;

PROCEDURE LongEmit (b : BITSET);
    VAR offset : CARDINAL;
    BEGIN
        offset := ObjPtr - ProcAddrTableAddress;
        IF offset + 3 >= ObjCodeSize THEN
            CodeOverflow;
        END (* if *);
	Place4 (offset, CARDINAL (b));
        INC (ObjPtr, 4);
    END LongEmit;

PROCEDURE Place4 (offset, c : CARDINAL);
    VAR high, low : CARDINAL;
	b : CARDINAL;
    BEGIN
	high := c DIV 65536; low := c MOD 65536;
	b := high DIV 256;
	ObjCode^[offset] := BYTE (b);
	b := high MOD 256;
	ObjCode^[offset + 1] := BYTE (b);
	b := low DIV 256;
	ObjCode^[offset + 2] := BYTE (b);
	b := low MOD 256;
	ObjCode^[offset + 3] := BYTE (b);
    END Place4;

PROCEDURE Place2 (offset, c : CARDINAL);
    VAR low : CARDINAL;
        b : CARDINAL;
    BEGIN
	low := c MOD 65536;
	b := low DIV 256;
	ObjCode^[offset] := BYTE (b);
	b := low MOD 256;
	ObjCode^[offset + 1] := BYTE (b);
    END Place2;

PROCEDURE CodeOverflow;
BEGIN
    ReportError (TRUE, "Object code overflow (buffer = %N bytes)", ObjCodeSize, 0);
END CodeOverflow;


(*                              Start and End                                *)

CONST
    TriposHeaderSize    = 2; (* these values are in long words *)
    VersionFrameSize    = 3;

    ExtrnAndEntrySize   = 16;

    ModuleTitleSize     = 8;
    WordModuleTitleSize = ModuleTitleSize DIV BYTESPERWORD;

    UnixMagic  = 407B;
    Thunk   =   1000;   Treloc  =   1001;
    Tend    =   1002;   Text    =   1005;
    
    EXTdefrel   =   1;
    EXTref      =   129;
    EXTproc     =   131;
    

TYPE
    ExtrnOrEntry    = ARRAY [0..ExtrnAndEntrySize-1] OF CHAR;
    Entrys          = (atD, atV);

VAR
    DataFrameSize : CARDINAL;

    ProcAddrTableAddress : CARDINAL;
    EndOfModula2Code : CARDINAL;

    BCPLStyleModuleTitle    : ARRAY [0..(ModuleTitleSize-1)] OF CHAR;
    EntryStyleModuleName    : ExtrnOrEntry;

    ModuleEntrys    : ARRAY [atD..atV] OF CARDINAL;

PROCEDURE CapCh (ch : CHAR) : CHAR;
    BEGIN
        IF ('a' <= ch) AND (ch <= 'z') THEN
            RETURN CHR (ORD ('A') + ORD (ch) - ORD ('a'));
        ELSE
            RETURN ch;
        END (* if *);
    END CapCh;
(*
PROCEDURE CurrentModuleName;
    VAR 
        i : CARDINAL;
        BCPLch, Entrych : CHAR;
        NameEnded : BOOLEAN;
    BEGIN
        NameEnded := FALSE;
        WITH Headerframe^ DO
            FOR i := 0 TO MODULENAMESIZE-1 DO
                IF (name [i] = 0C) AND NOT NameEnded THEN
                    NameEnded := TRUE;
                END (* if *);
                IF NameEnded THEN
                    BCPLch := SpaceCh;
                ELSE
                    BCPLch := name [i];
                END (* if *);
                Entrych := CapCh (BCPLch);
                IF i + 1 < ModuleTitleSize THEN
                    BCPLStyleModuleTitle [i+1] := BCPLch;
                END (* if *);
                IF i+3 < ExtrnAndEntrySize THEN
                    EntryStyleModuleName [i+3] := Entrych
                END (* if *);
            END (* for *);
        END (* with *);
        BCPLStyleModuleTitle [0] := CHAR (ModuleTitleSize-1);
        EntryStyleModuleName [0] := CHAR (EXTdefrel);
        EntryStyleModuleName [1] := '@';
    END CurrentModuleName;
*)
PROCEDURE StartObjectModule;
    VAR i : CARDINAL;
    BEGIN
        TransitionLabel := GetPermanentLabelTable (PermanentLabels);
        ALLOCATE (ExtrnMBStacks, (ImportlistLength+1) * TSIZE(Stack));
        ALLOCATE (ExtrnMBCounts, (ImportlistLength+1) * TSIZE(CARDINAL));
        ALLOCATE (ExtrnMOStacks, (ImportlistLength+1) * TSIZE(Stack));
        ALLOCATE (ExtrnMOCounts, (ImportlistLength+1) * TSIZE(CARDINAL));
        ALLOCATE (ExtrnMPStacks, (ImportlistLength+1) * TSIZE(Stack));
        ALLOCATE (ExtrnMPCounts, (ImportlistLength+1) * TSIZE(CARDINAL));

        FOR i := 0 TO ImportlistLength DO
            initialise (ExtrnMBStacks^[i]);
            ExtrnMBCounts^[i] := 0;
            initialise (ExtrnMOStacks^[i]);
            ExtrnMOCounts^[i] := 0;
            initialise (ExtrnMPStacks^[i]);
            ExtrnMPCounts^[i] := 0;
        END (* for *);
(*
        CurrentModuleName;
*)
        DataFrameSize := Headerframe^.datasize;
        ProcAddrTableAddress := 0;
(*	IF ODD (ProcAddrTableAddress) THEN INC (ProcAddrTableAddress) END;
        (* set up atD label ENTRY to point at first word of data frame *)
        ModuleEntrys [atD] := 
            (TriposHeaderSize + WordModuleTitleSize) * BYTESPERWORD;

        (* first word in data frame is label - must be relocated *)
        Relocate (ModuleEntrys [atD]);

        (* set up other ENTRY *)
        ModuleEntrys [atV] := ProcAddrTableAddress - 
                                (VersionFrameSize * BYTESPERWORD);
*)
        (* initialise ObjPtr *)
        ObjPtr := ProcAddrTableAddress;
    END StartObjectModule;

PROCEDURE EndObjectModule;
    VAR PerCent : REAL;
    BEGIN
        EndOfModula2Code := ObjPtr;
        (* end the segment with two aligned long word zeros *)
        IF ObjPtr MOD BYTESPERWORD # 0 THEN
            Emit (BITSET {});
        END (* if *);
        LongEmit (BITSET {});
        LongEmit (BITSET {});
        IF BranchStatisticsWanted THEN
            IF TotalBranches # 0 THEN
                PerCent := (FLOAT (ShortBranches) / FLOAT (TotalBranches)) * 100.0;
            ELSE
                PerCent := 0.0;
            END (* if *);
            ReportError (FALSE,"%R0 Branches %E4.1%% of them short",TotalBranches,R(PerCent));
        END (* if *);
    END EndObjectModule;

PROCEDURE OutputObjectCode;
    CONST 
        BCPLSecWord = 12345;
    VAR
        i       : CARDINAL;
        textsize, datasize : CARDINAL;
    BEGIN
(*
  	(* Write out the Header *)

        Wrword (UnixMagic);
        Wrword (textsize);
        Wrword (datasize);
        Wrword (DataFrameSize);		(* Uninitialised Data Size *)
        Wrword (0);			(* Symbol Table Size *)
        Wrword (0);			(* Entry *)
        Wrword (0);			(* Text Relocate *)
        Wrword (0);			(* Data Relocate *)

        FOR i := 0 TO ModuleTitleSize - 1 DO
            Wrch (BCPLStyleModuleTitle [i]);    (* Module name as *)
        END (* for *);                          (* BCPL string    *)

        (* now for the code proper *)
(*
        BulkWrite (WordObjCode, (ObjPtr - ProcAddrTableAddress) DIV BYTESPERWORD);
*)        
        (* next the STRING FRAME into the data area *)

        (* then the VERSION FRAME *)
        Wrword (EndOfModula2Code - ProcAddrTableAddress);
        Wrword (ProcadrSize DIV 2);
        Wrword (DataFrameSize);

        FlushOutBuff;
        
        
        LinkInfo;


        FlushOutBuff;
*)
    END OutputObjectCode;
(*
PROCEDURE LinkInfo;
    BEGIN
        WriteRelocInfo;
        WriteEntryExtrn;
    END LinkInfo;

PROCEDURE WriteRelocInfo;  
    BEGIN
        (* always at least one relocation (see StartObjectModule) *)
        Wrword (Treloc);
        Wrword (RelocCount);
        WHILE RelocCount # 0 DO
            Wrword (PopFrom (RelocStack));
            DEC (RelocCount);
        END (* while *);
    END WriteRelocInfo;

PROCEDURE WriteEntryExtrn;
    BEGIN
        Wrword (Text);
        WriteEntrys;
        WriteExtrns;
        Wrword (0);
    END WriteEntryExtrn;

PROCEDURE WriteEntrys;
    VAR 
        e : Entrys;
        i : CARDINAL;
    BEGIN
        FOR e := atD TO atV DO
            CASE e OF
                atD:    EntryStyleModuleName [2] := 'D';
            |   atV:    EntryStyleModuleName [2] := 'V';
            END (* case *);
            FOR i := 0 TO ExtrnAndEntrySize-1 DO
                Wrch (EntryStyleModuleName [i]);
            END (* for *);
            Wrword (ModuleEntrys [e]);
        END (* for *);
    END WriteEntrys;

PROCEDURE WriteExtrns;
    BEGIN
        WriteModuleExtrns;
        WriteSubExtrns;
    END WriteExtrns;

PROCEDURE WriteModuleExtrns;
    VAR
        Current     : ExtrnOrEntry;
        i, j        : CARDINAL;
        MBCount,
        MPCount,
        MOCount     : CARDINAL;
        NameEnded   : BOOLEAN;
        char        : CHAR;
        frameptr    : IMPORTFRAMEPTR;
        frame       : IMPORTFRAME;
    BEGIN
        Current [1] := '@';
        frameptr := Importlist;
        FOR i := 1 TO ImportlistLength DO
            frame := frameptr^;
            frameptr := frame.link;
            MBCount := ExtrnMBCounts^[i];
            MPCount := ExtrnMPCounts^[i];
            MOCount := ExtrnMOCounts^[i];
            IF (MBCount # 0) OR (MPCount # 0) OR (MOCount # 0) THEN
                NameEnded := FALSE;
                WITH frame DO
                    FOR j := 0 TO MODULENAMESIZE-4 DO
                        NameEnded := NameEnded OR (name [j] = 0C);
                        IF NameEnded THEN
                            char := SpaceCh;
                        ELSE
                            char := CapCh (name [j]);
                        END (* if *);
                        Current [j+3] := char;
                    END (* for *);
                END (* with *);
                Current [2] := 'D';
                WriteSingleExtrn (Current, MBCount, ExtrnMBStacks^[i]);
                WriteSingleProcExtrn (Current, MPCount, ExtrnMPStacks^[i]);
                Current [2] := 'O';
                WriteSingleExtrn (Current, MOCount, ExtrnMOStacks^[i]);
            END (* if *);
        END (* for *);
    END WriteModuleExtrns;

PROCEDURE WriteSubExtrns;
    VAR
        Current     : ExtrnOrEntry;
        i           : CARDINAL;
        NameEnded   : BOOLEAN;
    BEGIN
(*        FOR gl := divide TO trap DO
            CASE gl OF
                divide:
                    Current := " @_UDIV";
            |   signeddivide:
                    Current := " @_DIV";
            |   multiply:
                    Current := " @_MUL";
            |   fadd:
                    Current := " @_FADD";
            |   fsub:
                    Current := " @_FSUB";
            |   fmul:
                    Current := " @_FMUL";
            |   fdiv:
                    Current := " @_FDIV";
            |   fcmp:
                    Current := " @_FCMP";
            |   fabs:
                    Current := " @_FABS";
            |   fneg:
                    Current := " @_FNEG";
            |   ftrunc:
                    Current := " @_FTRUNC";
            |   ffloat:
                    Current := " @_FFLOAT";
            |   newprocess:
                    Current := " @_NEWPR";
            |   transfer:
                    Current := " @_TRANS";
            |   gft:
                    Current := " @_GFT";
            |   trap:
                    Current := " @_TRAP";
            END (* case *);
            IF ExtrnGLCounts [gl] # 0 THEN
                NameEnded := FALSE;
                FOR i := 1 TO ExtrnAndEntrySize-1 DO
                    NameEnded := NameEnded OR (Current [i] = 0C);
                    IF NameEnded THEN
                        Current [i] := SpaceCh;
                    END (* if *);
                END (* for *);
                WriteSingleExtrn 
                    (Current, ExtrnGLCounts [gl], ExtrnGLStacks [gl]);
            END (* if *);
        END (* for *);
*)    END WriteSubExtrns;

PROCEDURE WriteSingleExtrn (VAR x : ExtrnOrEntry; VAR c : CARDINAL;
                                VAR s : Stack);
    BEGIN
        IF c # 0 THEN
            x [0] := CHAR (EXTref);
            WriteExtrnOrEntry (x);
            Wrword (c);
            WriteStack (s, c);
        END (* if *);
    END WriteSingleExtrn;

PROCEDURE WriteSingleProcExtrn (VAR x : ExtrnOrEntry; VAR c : CARDINAL;
                                    VAR s : Stack);
    BEGIN
        IF c # 0 THEN
            x [0] := CHAR (EXTproc);
            WriteExtrnOrEntry (x);
            Wrword (c);
            c := c * 2;
            WriteStack (s, c);
        END (* if *);
    END WriteSingleProcExtrn;

PROCEDURE WriteExtrnOrEntry (VAR x : ExtrnOrEntry);
    VAR i : CARDINAL;
    BEGIN
        FOR i := 0 TO ExtrnAndEntrySize-1 DO
            Wrch (x [i]);
        END (* for *);
    END WriteExtrnOrEntry;

PROCEDURE WriteStack (VAR s : Stack; VAR c : CARDINAL);
    BEGIN
        WHILE c # 0 DO
            Wrword (PopFrom (s));
            DEC (c);
        END (* while *);
    END WriteStack;


(*                          OUTPUT PROCEDURES                                *)

CONST
    OutBuffSize     = 1024;
    OutBuffWordUpb  = OutBuffSize - 1;
    OutBuffCharUpb  = (OutBuffSize * BYTESPERWORD) - 1;

VAR
    OutBuffPtr  :   CARDINAL;
    OutBuffWord :   POINTER TO ARRAY [0..OutBuffWordUpb] OF WORD;
    OutBuffChar :   POINTER TO ARRAY [0..OutBuffCharUpb] OF CHAR;

PROCEDURE Wrch (ch : CHAR);
    BEGIN
        IF OutBuffPtr > OutBuffCharUpb THEN
            FlushOutBuff;
        END (* if *);
        OutBuffChar^ [OutBuffPtr] := ch;
        INC (OutBuffPtr);
    END Wrch;

PROCEDURE Wrword (w : WORD);
    BEGIN
        IF OutBuffPtr MOD BYTESPERWORD # 0 THEN
            ReportError (TRUE, "Object-mod - illegal Wrword*N", 0, 0);
        ELSIF OutBuffPtr > OutBuffCharUpb THEN
            FlushOutBuff;
        END (* if *);
        OutBuffWord^ [OutBuffPtr DIV BYTESPERWORD] := w;
        INC (OutBuffPtr, BYTESPERWORD);
    END Wrword;

PROCEDURE FlushOutBuff;
    BEGIN
        IF OutBuffPtr # 0 THEN
(*            BulkWrite (OutBuffWord, OutBuffPtr DIV BYTESPERWORD); *)
            OutBuffPtr := 0;
        END (* if *);
    END FlushOutBuff;

PROCEDURE Reverse (w : WORD) : CARDINAL;
   VAR p1, p2 : POINTER TO ARRAY [0..3] OF CHAR;
   VAR res : WORD;
   BEGIN
   p1 := ADR (w);
   p2 := ADR (res);
   p2^[0] := p1^[3];
   p2^[1] := p1^[2];
   p2^[2] := p1^[1];
   p2^[3] := p1^[0];
   RETURN CARDINAL (res);
END Reverse;
*)
(*                          MODULE INITIALISATION                            *)

VAR i : CARDINAL;

BEGIN
    LastLabel := none;

    InitLL (TempLabelList);
    InitLL (UndefinedList);
    InitLL (LabListPool);

    initialise (RelocStack);
    RelocCount := 0;

    FOR i := 0 TO MaxNStProcs DO
        initialise (ExtrnOSStacks [i]);
        ExtrnOSCounts [i] := 0;
    END (* for *);

    NEW (CardPtr);
    BytePtr := ADDRESS (CardPtr);
    BitsetPtr := ADDRESS (CardPtr);
(*
    NEW (OutBuffWord);
    OutBuffChar := ADDRESS (OutBuffWord);
    OutBuffPtr := 0;
*)
    BranchStatisticsWanted := FALSE;
    ShortBranches := 0;
    TotalBranches := 0;
    NamesIndex := 0;
END Object.
