
IMPLEMENTATION MODULE ReadRef;


FROM SYSTEM IMPORT ADR, WORD;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM SysStreams IMPORT sysOut;
FROM SystemTypes IMPORT String;
FROM Streams IMPORT Stream, ErrorCode, Get, Delete;
FROM UtilError IMPORT ReportError;
IMPORT Strings; 
FROM WriteF IMPORT H, S, WriteF0, WriteF1, WriteF2, WriteF3;


FROM M2RefTypes IMPORT RefSymbol;
FROM FileStream IMPORT 
    CreateOutput, StreamErrorCode;
FROM M2Files IMPORT CreateInput;
FROM UnixErrors IMPORT SysErrorString, sysErrList;
    
IMPORT Slist;

TYPE NAME = String;
     NUMBER = CARDINAL;
     

CONST
    maxsymlength = 30;
VAR 
    ModuleName : ARRAY [0..30] OF CHAR;
    refStream: Stream;
    modstack : ARRAY [0..20] OF NAME;
    modstacktop : CARDINAL;
    RefVersion : CARDINAL;
    underline : ARRAY [0..1] OF CHAR;
    
PROCEDURE Error(s: ARRAY OF CHAR; n1, n2: WORD);
VAR ss: ARRAY [0 .. 50] OF CHAR;
BEGIN
    Strings.ConcatCC("(ReadRef): ", s, ss);
    ReportError(TRUE, ss, n1, n2);
END Error;
    

PROCEDURE SetProcTable (mn : ARRAY OF CHAR);
    VAR i : CARDINAL;
    BEGIN
        Strings.CopyCC (mn, ModuleName);
        refStream := CreateInput (mn, "ref");
        IF refStream^.status # Success THEN
            ReportError(TRUE, "unable to open reference file for %S",
                                H(mn),0);
        END;
        FOR i := 0 TO maxprocs - 1 DO  
            IF ProcTable[i] # NIL THEN Strings.Dispose (ProcTable[i]) END
        END;
        modstacktop := 0;
        modstack [modstacktop] := Strings.CopyCS ("");
        RefFileVersion;
        SkipSymb();
        Module();
        Delete (refStream);
    END SetProcTable;
    
PROCEDURE RefFileVersion;
    VAR
        symb: RefSymbol;
    BEGIN
        symb := NextSymb(); (* should be reffileRS *)
        IF symb # reffileRS THEN
            Error("bad reference file",0,0);
        END (* if *);
        symb := NextSymb(); (* version number *)
        RefVersion := CARDINAL(symb);
        IF RefVersion < 32 THEN
            Error("wrong version of reference file", 0, 0);
        END;
    END RefFileVersion;

PROCEDURE Module();
VAR objnum : CARDINAL;
    name : NAME;
BEGIN
    SkipNumber();  
    objnum := Number();
    INC (modstacktop);
    Name (name);
    name := Strings.ConcatCS (underline, name);
    modstack [modstacktop] := Strings.ConcatSS (modstack[modstacktop - 1],
                                    name);
    modstack [modstacktop] := Strings.ConcatSC (modstack[modstacktop],
                                underline);
    ProcTable [objnum] := Strings.ConcatSS (modstack[modstacktop], name);
    IF Strings.LengthS (ProcTable [objnum]) > maxsymlength THEN
        Strings.Dispose (ProcTable [objnum]);
        ProcTable [objnum] := ModuleObjectNumber (modstack[1], objnum);
    END;
    Block();
    Strings.Dispose (name);
    Strings.Dispose (modstack[modstacktop]);
    DEC (modstacktop);
END Module;

PROCEDURE Block();
VAR
    symb: RefSymbol;
BEGIN
  LOOP
  symb := NextSymb();
  IF (symb = endRS) THEN EXIT
  ELSE
    CASE symb OF
      moduleRS:    
	    Module();
    | procRS:
	    Procedure();
    | varRS:
	    Variable();
    | typeRS:
            Type();
	
    ELSE Error("Block type expected, %O6 read", symb, 0);
    
    END; (* case *);
  END; (* if *)
  
  END (* loop *);
END Block;

PROCEDURE Procedure();
VAR procnum : NUMBER;
    procname : NAME;
BEGIN
    SkipNumber();  
    procnum := Number();
    Name(procname);
    ProcTable [procnum] := Strings.ConcatSS (modstack[modstacktop], procname);
    IF Strings.LengthS (ProcTable [procnum]) > maxsymlength THEN
        Strings.Dispose (ProcTable [procnum]);
        ProcTable [procnum] := ModuleObjectNumber (modstack[1], procnum);
    END;
    Strings.Dispose (procname);
    Block();
END Procedure;

PROCEDURE ModuleObjectNumber (mod : String; objnum : CARDINAL): String;
VAR numchar : ARRAY [0..3] OF CHAR;
BEGIN
    numchar [0] := CHAR (CARDINAL('0') + objnum DIV 100);
    numchar [1] := CHAR (CARDINAL('0') + (objnum MOD 100) DIV 10);
    numchar [2] := CHAR (CARDINAL('0') + objnum MOD 10);
    numchar [3] := 0C;
    RETURN Strings.ConcatSC (mod, numchar);
END ModuleObjectNumber;

PROCEDURE Variable();
BEGIN
    SkipNumber();
    SkipSymb();
    SkipNumber();
    SkipName();
    TypeRef();
END Variable;

PROCEDURE TypeRef();
    BEGIN
	IF NextSymb() = typerefRS THEN SkipNumber() END;
    END TypeRef;
    

PROCEDURE Type();
    VAR
        symb: RefSymbol;
        typedesc: RefSymbol;
        owner: NAME;
        importedtype: BOOLEAN;
        i: CARDINAL;
    BEGIN
	    SkipNumber();
            Name(owner); importedtype := owner^[0] # 0C;
            SkipName();
            typedesc := NextSymb();
            SkipNumber();
            Strings.Dispose(owner);
            IF importedtype AND (RefVersion > 32) THEN
                RETURN  (* nothing more to process *)
            END (* if *);

            CASE typedesc OF
              hiddenRS:
            | setRS, pointerRS, arrdynRS, openRS:
                TypeRef();
            | subrRS:
                SkipNumber(); SkipNumber();
                TypeRef();
            | arrayRS:
                TypeRef(); TypeRef();
            | enumRS:
                WHILE NextSymb() = constRS DO
                    SkipNumber();
                    SkipName ();
                END (* while *);
                
            | recordRS:
                WHILE NextSymb() = fieldRS DO
                    SkipNumber();
                    SkipName();
                    IF RefVersion >= 34 THEN
                        IF Number() # 0 THEN   (* subject of variant *)
                            FOR i := 1 TO Number() DO
                                    SkipNumber();
                            END (* for *);
                        END (* if *);
                    END (* if *);
                    TypeRef();
                END (* while *);
                
            ELSE Error("Unknown typedesc %U\n", typedesc, 0);
            END (* case *);
    END Type;

PROCEDURE Name(VAR np: NAME);
VAR
    i: CARDINAL;
    n: ARRAY [0 .. 40] OF CHAR;
BEGIN
    i := 0;
    REPEAT
        n[i] := Character();
	INC(i)
    UNTIL (n[i-1]  = 0C) OR (i > HIGH(n));
    IF (i > HIGH(n)) THEN
        n[i-1] := 0C;
        Error("name too long (%S)", H(n), 0);
    ELSE
    (*  Suppress types with 'invented' (numeric) names *)
        IF ('0'<=n[0]) AND (n[0]<='9') THEN n[0] := 0C END;
        np := Strings.CopyCS(n);
    END;
END Name;

PROCEDURE SkipName();
VAR
    i: CARDINAL;
    n: ARRAY [0 .. 40] OF CHAR;
BEGIN
    i := 0;
    REPEAT
        n[i] := Character();
	INC(i)
    UNTIL (n[i-1]  = 0C) OR (i > HIGH(n));
    IF (i > HIGH(n)) THEN
        n[i-1] := 0C;
        Error("name too long (%S)", H(n), 0);
    ELSE
    END;
END SkipName;


PROCEDURE NextSymb(): RefSymbol;
VAR c: RefSymbol;
BEGIN
    c := RefSymbol( Get(refStream) );
    IF refStream^.status # Success THEN
        RETURN endRS
    ELSE
        RETURN c;
    END (* if *);
END NextSymb;

PROCEDURE SkipSymb();
VAR c: WORD;
BEGIN
    c := Get(refStream);
END SkipSymb;


PROCEDURE Number(): NUMBER;
VAR n: NUMBER;
BEGIN
    n := GetBytes();
    RETURN n;
END Number;

PROCEDURE SkipNumber();
VAR n: NUMBER;
BEGIN
    n := GetBytes();
END SkipNumber;

PROCEDURE Character(): CHAR;
BEGIN
    RETURN CHAR( Get(refStream) );
END Character;

PROCEDURE GetBytes(): CARDINAL;
VAR b: CARDINAL;
BEGIN
    b := CARDINAL( Get(refStream) );
    IF b < 128 THEN
        RETURN b;
    ELSE
        RETURN (GetBytes() * 128) + b - 128;
    END;
END GetBytes;

VAR i : CARDINAL;
BEGIN
    modstacktop := 0;
    FOR i := 0 TO maxprocs - 1 DO  
        ProcTable[i] := NIL
    END;
    underline [0] := "_";
    underline [1] := 0C;
END ReadRef.
