IMPLEMENTATION MODULE Relations;

FROM items IMPORT ITEM, REGISTER, MODE, Operator, ConditionCode, equal,
                        NullRegItem, PackItem, NILvalue,
                        immediate, InDataReg, DataReg0, MonOperator;
FROM RegSlv IMPORT Slave, FreeReg, UseBitSet, protect,
                   PutIntoSlave, StopSlaving, DeleteFromSlave;
FROM GenProcs IMPORT GenCmp, GenMove, GenMoveImm, GenTst,
                     GenClr, GenRelBranch, GenMoveQ, CheckedGenCmp,
		     GenShift, GenCmpImm, GenSimpleDyadicOp,
                     GenShiftImm, GenAddImm, GenCmpAddress;
FROM labels IMPORT NewLabel, PushLabelStack;
FROM GenOp IMPORT SetResultItem;


FROM SysStreams IMPORT sysOut;
FROM WriteF IMPORT WriteF0;

CONST StoreResult = TRUE; DontStoreResult = FALSE;


PROCEDURE SelectCondCode (op : Operator) : ConditionCode;
   VAR cc : ConditionCode;
   BEGIN
      CASE op OF
       Eql  : cc := EQ |  Neq  : cc := NE |
       Ulss : cc := CS |  Lss  : cc := LT |
       Uleq : cc := LS |  Leq  : cc := LE |
       Ugtr : cc := HI |  Gtr  : cc := GT |
       Ugeq : cc := CC |  Geq  : cc := GE
      END;
      RETURN cc
   END SelectCondCode;


PROCEDURE opposite (CondCode : ConditionCode) : ConditionCode;
   VAR cc : ConditionCode;
   BEGIN
      CASE CondCode OF
         T  : cc := F  | F  : cc := T  |
         EQ : cc := NE | NE : cc := EQ |
	 LE : cc := GT | GT : cc := LE |
	 LT : cc := GE | GE : cc := LT |
	 LS : cc := HI | HI : cc := LS |
	 CS : cc := CC | CC : cc := CS
      END; RETURN cc
   END opposite;


PROCEDURE reverse (CondCode : ConditionCode) : ConditionCode;
   VAR cc : ConditionCode;
   BEGIN
      CASE CondCode OF
       LT : cc := GT |  LE : cc := GE |
       GT : cc := LT |  GE : cc := LE |
       CS : cc := HI |  LS : cc := CC |
       HI : cc := CS |  CC : cc := LS
      ELSE cc := CondCode
      END; RETURN cc
   END reverse;

PROCEDURE SameOperandsYieldsTRUE (cc : ConditionCode) : BOOLEAN;
    BEGIN
        CASE cc OF
            EQ, LE, GE, LS, CC  : RETURN TRUE;
        |   NE, LT, GT, CS, HI  : RETURN FALSE;
        END (* case *);
    END SameOperandsYieldsTRUE;

PROCEDURE compare (op : Operator; VAR opd1, opd2 : ITEM)
                                         : ConditionCode;
    VAR cc : ConditionCode;
        SlvOpd1, SlvOpd2, RegItem : ITEM;
    BEGIN
        IF op = In THEN
            FindIfIn (opd1, opd2, NullRegItem, DontStoreResult);
            cc := NE;
        ELSE
            Slave (opd1, SlvOpd1); Slave (opd2, SlvOpd2);
            cc := SelectCondCode (op);
            IF equal (opd1, opd2) THEN
                IF SameOperandsYieldsTRUE (cc) THEN
                    cc := T;
                ELSE
                    cc := F;
                END (* if *);
            ELSIF InDataReg (SlvOpd2) OR immediate (opd1) THEN
                cc := CheckedGenCmp (cc, SlvOpd1, SlvOpd2)
            ELSIF InDataReg (SlvOpd1) OR immediate (opd2) THEN
                cc := CheckedGenCmp (reverse (cc), SlvOpd2, SlvOpd1)
            ELSE
                FreeReg (RegItem, NOT UseBitSet);
                GenMove (opd2, RegItem);
                PutIntoSlave (RegItem, opd2, FALSE);
                GenCmp (opd1, RegItem);
            END (* if *);
        END (* if *);
        RETURN cc
    END compare;


PROCEDURE CompareAndStoreResult (op : Operator;
                          VAR opd1, opd2, dest : ITEM);

    PROCEDURE Option1 (cc : ConditionCode;
                          VAR opd1, opd2, dest : ITEM);
        VAR label : CARDINAL;
        BEGIN
            StopSlaving (dest);
            GenClr (dest);
            cc := CheckedGenCmp (cc, opd1, opd2);
            label := NewLabel ();
            GenRelBranch (opposite (cc), label);
            GenAddImm (1, dest);
            PushLabelStack (label);
        END Option1;

    PROCEDURE Option2 (cc : ConditionCode; VAR opd1, opd2, dest : ITEM);
        VAR label1, label2 : CARDINAL;
        BEGIN
            StopSlaving (dest);
            cc := CheckedGenCmp (cc, opd1, opd2);
            label1 := NewLabel (); label2 := NewLabel ();
            GenRelBranch (cc, label1);
            GenClr (dest);
            GenRelBranch (T, label2);
            PushLabelStack (label1);
            GenMoveImm (1, dest);
            PushLabelStack (label2);
        END Option2;

    PROCEDURE GenCmpAndStore (cc : ConditionCode; VAR opd1, opd2, dest : ITEM);
        BEGIN
            IF equal (opd1, dest) OR equal (opd2, dest) THEN
                Option2 (cc, opd1, opd2, dest);
            ELSE
                Option1 (cc, opd1, opd2, dest);
            END (* if *);
        END GenCmpAndStore;

    VAR cc : ConditionCode;
        reg1, reg2, SlvOpd1, SlvOpd2,
        Dest, SlvDest : ITEM;

    BEGIN
        IF op = In THEN
            FindIfIn (opd1, opd2, dest, StoreResult);
        ELSE
            Slave (opd1, SlvOpd1); Slave (opd2, SlvOpd2);
            cc := SelectCondCode (op);
            IF equal (dest, NullRegItem) THEN
                FreeReg (reg1, UseBitSet);
                Dest := reg1;
                SetResultItem (reg1);
            ELSE
                Dest := dest
            END (* if *);
            Slave (dest, SlvDest);
            protect (SlvDest);
            IF equal (opd1, opd2) THEN
                IF SameOperandsYieldsTRUE (cc) THEN
                    GenMoveImm (1, Dest)
                ELSE
                    GenClr (Dest)
                END (* if *);
            ELSE
                IF InDataReg (SlvOpd2) OR immediate (opd1) THEN
                    GenCmpAndStore (cc, SlvOpd1, SlvOpd2, Dest)
                ELSIF InDataReg (SlvOpd1) OR immediate (opd2) THEN
                    GenCmpAndStore (reverse (cc), SlvOpd2, SlvOpd1, Dest)
                ELSE
                    IF InDataReg (SlvDest)
                    AND NOT (equal(dest, NullRegItem) OR equal(dest, DataReg0))
                    THEN
                        reg2 := SlvDest;
                    ELSE
                        FreeReg (reg2, NOT UseBitSet);
                    END (* if *);
                    GenMove (opd2, reg2);
                    PutIntoSlave (reg2, opd2, TRUE);
                    GenCmpAndStore (cc, opd1, reg2, Dest)
                END (* if *);
            END (* if *);
        END (* if *);
    END CompareAndStoreResult;


PROCEDURE FindIfIn (VAR opd1, opd2, dest : ITEM;
                     ResultToBeStored : BOOLEAN);
   VAR SlvOpd1, SlvOpd2, SlvDest, Reg1, Reg2, bitval : ITEM;
       SlvDestAvailable : BOOLEAN;
       NullSet : BITSET;
       shiftval : CARDINAL;
    BEGIN
        Slave (opd1, SlvOpd1); Slave (opd2, SlvOpd2);
        Slave (dest, SlvDest);
        protect (SlvDest); StopSlaving (dest);
        (* slaved operands already protected (see PsStk.pop) *)
        SlvDestAvailable := InDataReg (SlvDest) AND
                                NOT equal (SlvDest, dest);
        IF immediate (opd2) THEN
            IF InDataReg (SlvOpd1) THEN
                Reg1 := SlvOpd1;
            ELSE
                IF SlvDestAvailable THEN
                    Reg1 := SlvDest;
                ELSE
                    FreeReg (Reg1, NOT UseBitSet);
                END (* if *);
                GenMove (opd1, Reg1);
            END (* if *);
            NullSet := {};
            shiftval := opd2.DispOrData;
            INCL (NullSet, shiftval);
            PackItem (bitval, Immediate, NoReg, INTEGER (NullSet), NoReg);
            GenSimpleDyadicOp (And, bitval, Reg1);
            IF ResultToBeStored THEN
                IF shiftval <> 0 THEN
		    IF shiftval > 8 THEN
			REPEAT
			    GenShiftImm ( Shr, 8, Reg1 );
	  		    shiftval := shiftval - 8;
			UNTIL shiftval <= 8;
		    END;
		    GenShiftImm (Shr, shiftval, Reg1);
                END (* if *);
                IF equal (dest, NullRegItem) THEN
                    SetResultItem (Reg1);
                ELSE
                    GenMove (Reg1, dest);
                    PutIntoSlave (Reg1, dest, TRUE);
                END (* if *);
            ELSE
                DeleteFromSlave (Reg1.Register);
            END (* if *);
        ELSE
            IF InDataReg (SlvOpd2) THEN
                Reg1 := SlvOpd2
            ELSE
                IF SlvDestAvailable THEN
                    GenMove (opd2, SlvDest);
                    Reg1 := SlvDest;
                    SlvDestAvailable := FALSE
                ELSE
                    FreeReg (Reg1, UseBitSet);
                    GenMove (opd2, Reg1)
                END (* if *);
                PutIntoSlave (Reg1, opd2, TRUE)
            END (* if *);
            IF SlvDestAvailable THEN
                Reg2 := SlvDest
            ELSE
                FreeReg (Reg2, NOT UseBitSet)
            END (* if *);
            GenMoveQ (1, Reg2.Register);
            GenShift (Shl, Reg1, Reg2);
            GenSimpleDyadicOp (And, SlvOpd1, Reg2);
            IF ResultToBeStored THEN
                GenShift (Shr, Reg1, Reg2);
                IF equal (dest, NullRegItem) THEN
                    SetResultItem (Reg2);
                ELSE
                    GenMove (Reg2, dest);
                    PutIntoSlave (Reg2, dest, TRUE);
                END (* if *);
            END (* if *);
        END (* if *);
   END FindIfIn;

PROCEDURE NotTest (VAR opd : ITEM) : ConditionCode;
    VAR SlvOpd : ITEM;
    BEGIN
        Slave (opd, SlvOpd);
        GenTst (SlvOpd);
        RETURN EQ;
    END NotTest;


END Relations.
