IMPLEMENTATION MODULE SYS;

FROM items IMPORT
    MODE, ITEM, REGISTER, PackRealStackItem, BytesPerWord, DataReg0, PackItem,
    StackPreDec;
FROM PsStk IMPORT PseudStk, top, PseudoStackEmpty, DoAnyPendingOp, pop, push,
                  PutInScratchSpace, Push;
FROM RegSlv IMPORT SetRegFree, FreeReg, UseBitSet, IsFree, PurgeRegSlaves;
FROM CodeProcs IMPORT next;
FROM GenProcs IMPORT
    GenSubaImm, GenAddaImm, GenUnixCall, GenMoveQ, GenMoveReg, GenMovea,
    GenNamedCall;
FROM UtilError IMPORT ReportError;
FROM Object IMPORT OSNAME;
FROM MainPass IMPORT ExtraBytesProcessed;
FROM MCMnemonics IMPORT FUNC;
VAR NumberOfArgumentBytes : CARDINAL;
    StackPostInc : ITEM;

PROCEDURE sys ();
    VAR
        SystemCall : CARDINAL;
    CONST
        NormalSystemCall    = 0;
        Register            = 1;
	CCall		    = 2;
    BEGIN
        NumberOfArgumentBytes := next ();
        SystemCall := next ();
        CASE SystemCall OF
          
          NormalSystemCall  :   DoSystemCall;
        
        | Register          :   GetRegisterValue;
	| CCall		    :   DoCCall;
        ELSE
            ReportError (TRUE,
                         "Unknown system call (%O3) after SYS*N",
                         SystemCall, 0);
        END (* case *);
    END sys;

PROCEDURE DoCCall;
    VAR NextByte, displacement, NoOfArgs, count : CARDINAL;
        name : ARRAY [0 .. 79] OF CHAR;
    BEGIN
	NoOfArgs := next ();
	IF (NumberOfArgumentBytes - 2) > (HIGH (name) - 1) THEN
	    ReportError (TRUE, "External name to long", 0, 0)
	END;
	FOR count := 0 TO HIGH (name) DO name [count] := 0C END;
	FOR count := 1 TO NumberOfArgumentBytes - 2 DO
	    name [count - 1] := CHAR (next ());
	END;
        DoAnyPendingOp ();
        (* We must save the dynnamic and static links *)
	GenMoveReg (G, StackPreDec);
	GenMoveReg (L, StackPreDec);
        displacement := NoOfArgs * BytesPerWord;
        FOR count := 1 TO NoOfArgs DO
            pop (StackPreDec);
        END (* for *);
        PutInScratchSpace (D0); (* if D0 on real stack then move contents
                                to safety *)
        SetRegFree (D0);
        GenNamedCall (name);
        IF displacement # 0 THEN GenAddaImm (displacement, S) END;
	GenMovea (StackPostInc, L);
	GenMovea (StackPostInc, G);
        NextByte := next ();
        ExtraBytesProcessed := 0;
        IF NextByte = FUNC (* is a function call *) THEN
            INC (ExtraBytesProcessed);
            push (DataReg0);
            NextByte := next ()
        END (* if *);
	PurgeRegSlaves (TRUE);
    END DoCCall;

PROCEDURE DoSystemCall;
    VAR 
        displacement, NoOfArgs, count,
	TypeOfSystemCall : CARDINAL;
    BEGIN
        DoAnyPendingOp ();
        TypeOfSystemCall := next ();
        NoOfArgs := next ();
        (* We must save the dynnamic and static links *)
	GenMoveReg (G, StackPreDec);
	GenMoveReg (L, StackPreDec);
        displacement := NoOfArgs * BytesPerWord;
        FOR count := 1 TO NoOfArgs DO
            pop (StackPreDec);
        END (* for *);
        PutInScratchSpace (D0); (* if D0 on real stack then move contents
                                to safety *)
        SetRegFree (D0);
        GenUnixCall (TypeOfSystemCall);
        IF displacement # 0 THEN GenAddaImm (displacement, S) END;
	GenMovea (StackPostInc, L);
	GenMovea (StackPostInc, G);
        push (DataReg0);
	PurgeRegSlaves (TRUE);
    END DoSystemCall;

PROCEDURE GetRegisterValue;
    VAR RegItem : ITEM;
        reg : REGISTER;
    BEGIN
        IF NumberOfArgumentBytes # 2 THEN
            ReportError (TRUE, "Illegal use of REGISTER*N", 0, 0);
        END (* if *);
        reg := VAL (REGISTER, next ());
        IF (D0 <= reg) AND (reg <= D7) AND IsFree (reg) THEN
            Push (DataReg, reg, 0);
        ELSE
            FreeReg (RegItem, NOT UseBitSet);
            GenMoveReg (reg, RegItem);
            push (RegItem);
        END (* if *);
    END GetRegisterValue;

BEGIN
    PackItem (StackPostInc, PostInc, S, 0, NoReg);
END SYS.
