(*
    Title: 	UnixObj - HpUx Implementation
    LastEdit:	Feb 1990
    Author: 	Ian Leslie
		Cambridge University Computer Laboratory
*)




IMPLEMENTATION MODULE UnixObj;

FROM SYSTEM IMPORT ADDRESS;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM Streams IMPORT Put, PutN, Seek, SeekMode, Stream;
IMPORT BinaryIO;
FROM UtilError IMPORT ReportError;
IMPORT HashStrings, Strings, Slist, SystemTypes, WriteF;

FROM SysStreams IMPORT sysErr;
FROM WriteF IMPORT WriteF3, S, H;

CONST bigseg = 100000;
        externbit = {27};
        long = {26};
      Rdata = {30};
      Rtext = {};
      Rextern = {30,31};
      Rlong = {29};

      HpUxundf = 0;
      HpUxtext = 2;
      HpUxdata = 3;

      HpUxext  = 020H;
      HpUxLong = 2;

      HpUxSegtext = 0;
      HpUxSegdata = 1; 
      HpUxSegext  = 3;

TYPE SYMBOLENTRYPTR = POINTER TO SYMBOLENTRY;
TYPE SYMBOLENTRY = RECORD
        link: SYMBOLENTRYPTR;
        hashid: HashStrings.HashId;
        ntype, value: CARDINAL;
        id: CARDINAL;
        namelength: CARDINAL;
    END (* record *);

VAR size, relocsize: CARDINAL;
    textsize: CARDINAL;
    textrelocsize: CARDINAL;
    binStream : Stream;
    relocatingtext : BOOLEAN;
    
PROCEDURE SetObjStream (s : Stream);
    BEGIN
        binStream := s;
    END SetObjStream;

PROCEDURE WriteCARDINAL (s : Stream; c: CARDINAL);
    BEGIN
        PutLong (c);
    END WriteCARDINAL;

PROCEDURE PutHeader(magic: CARDINAL);
    VAR
        i: CARDINAL;
    BEGIN
        WriteCARDINAL(binStream, magic);
        FOR i := 1 TO 15 DO
            WriteCARDINAL(binStream, 0);
        END (* for *);
	size := 0;
    END PutHeader;

PROCEDURE PutByte(b: BYTE);
    BEGIN
        Put(binStream, b);
        INC(size, 1);
    END PutByte;

PROCEDURE PutShort(s: CARDINAL);
    BEGIN
        Put(binStream, s DIV 256);
	Put(binStream, s MOD 256);
        INC(size, 2);
    END PutShort;

PROCEDURE PutLong(w: CARDINAL);
(*  no guarantee of word alignment, so output byte at a time if necessary *)
    VAR
        high, low: CARDINAL;
    BEGIN
        high := w DIV 65536; low := w MOD 65536;
        Put(binStream, high DIV 256);
        Put(binStream, high MOD 256);
        Put(binStream, low DIV 256);
        Put(binStream, low MOD 256);
        INC(size, 4);
    END PutLong;

PROCEDURE PutSeg (add : ADDRESS; bytes : CARDINAL);
    VAR i : CARDINAL;
    BEGIN
        i := PutN (binStream, add, bytes);
        INC (size, bytes);  
    END PutSeg;

PROCEDURE Flush(udatasize: CARDINAL);
    VAR datasize : CARDINAL;
	symtablesize: CARDINAL;
    BEGIN
	datasize := size;
	size := 0;
        PutSymbols;
	symtablesize := size;
        PutRelocs (TextRelocList);
        PutRelocs (DataRelocList);
        Seek(binStream, Beginning, 12);
        WriteCARDINAL(binStream, textsize);
        WriteCARDINAL(binStream, datasize);
        WriteCARDINAL(binStream, udatasize);
        WriteCARDINAL(binStream, textrelocsize);
        WriteCARDINAL(binStream, relocsize);
        WriteCARDINAL(binStream, 0);
        WriteCARDINAL(binStream, symtablesize);
        WriteCARDINAL(binStream, 0);
        WriteCARDINAL(binStream, 0);
        WriteCARDINAL(binStream, 0);
        WriteCARDINAL(binStream, 0);
        WriteCARDINAL(binStream, 0);
        WriteCARDINAL(binStream, 0);
    END Flush;

PROCEDURE SizeNow(): CARDINAL;
    BEGIN
        RETURN size;
    END SizeNow;

PROCEDURE EndText;
    BEGIN
        textsize := size;
        size := 0;
    END EndText;

PROCEDURE EndTextReloc;
    BEGIN
        relocatingtext := FALSE;
        textrelocsize := relocsize;
        relocsize := 0;
    END EndTextReloc;

TYPE RELOCPTR = POINTER TO RELOC;
    RELOC = RECORD
            link: RELOCPTR;
            address, symb, segment: CARDINAL;
        END (* record *);

VAR TextRelocList, DataRelocList: RELOCPTR;

	   

PROCEDURE AddRelocation(address, symb: CARDINAL; extern: BOOLEAN);
    VAR
        RelocList, reloc, preloc, xreloc: RELOCPTR;
        searching: BOOLEAN;
    BEGIN
        IF relocatingtext THEN
	    RelocList := TextRelocList;
	ELSE
	    RelocList := DataRelocList;
	END;
        NEW(reloc);
        reloc^.address := address;
	IF extern THEN
	    reloc^.symb := symb;
	    reloc^.segment := HpUxSegext;
	ELSE reloc^.symb := 0;
	   CASE symb OF
	     Ntext: reloc^.segment := HpUxSegtext;
	   | Ndata: reloc^.segment := HpUxSegdata;
	   ELSE ReportError (TRUE, "Bad symb for reloc %u", symb, 0);
	   END;
	END;
        xreloc := RelocList; preloc := RelocList; searching := TRUE;
        WHILE searching AND (xreloc # NIL) DO
            IF address < xreloc^.address THEN
                searching := FALSE;
            ELSE
                preloc := xreloc;
                xreloc := xreloc^.link;
            END (* if *);
        END (* while *);
        reloc^.link := xreloc;
        IF xreloc=RelocList THEN
	    IF relocatingtext THEN
            	TextRelocList := reloc
	    ELSE
	    	DataRelocList := reloc
	    END
        ELSE
            preloc^.link := reloc;
        END (* if *);
        INC(relocsize, 8);
    END AddRelocation;

PROCEDURE PutRelocs (VAR RelocList : RELOCPTR);
    VAR
        xr: RELOCPTR;
    BEGIN
        WHILE RelocList # NIL DO
	    WriteCARDINAL (binStream, RelocList^.address);
	    PutShort (RelocList^.symb);
	    PutByte (RelocList^.segment);
	    PutByte (HpUxLong);
            xr := RelocList;
            RelocList := RelocList^.link;
            DISPOSE(xr);
        END (* while *);
    END PutRelocs;


VAR
    symbolcount: CARDINAL;
    symbols: HashStrings.HashTable;
    symbolentries: SYMBOLENTRYPTR;
    stringtableindex: CARDINAL;

PROCEDURE AddSymbol(VAR name: ARRAY OF CHAR; xntype, xvalue: CARDINAL): CARDINAL;
    VAR
        xhashid: HashStrings.HashId;
        symbolentry: SYMBOLENTRYPTR;
        HpUxtype: CARDINAL;
    BEGIN
        CASE xntype OF
	   Ntext: HpUxtype := HpUxtext;
	|  Ndata: HpUxtype := HpUxdata;
	|  Nundf + Next: HpUxtype := HpUxundf + HpUxext;
	|  Ntext + Next: HpUxtype := HpUxtext + HpUxext;
	|  Ndata + Next: HpUxtype := HpUxdata + HpUxext;
	ELSE ReportError (TRUE, "Bad symbol ntype: %u", xntype, 0);
	END;   	
        IF HashStrings.EnterChars(symbols, name, xhashid) THEN
            NEW(symbolentry);
            WITH symbolentry^ DO
                ntype := HpUxtype;
                hashid := xhashid;
                value := xvalue;
                id := symbolcount;
                namelength := Strings.LengthC(name)+1;
            END (* with *);
            INC(symbolcount);
            Slist.AddE(symbolentries, symbolentry);
            HashStrings.Assoc(symbols, xhashid, symbolentry);
        ELSE
            symbolentry := SYMBOLENTRYPTR( HashStrings.Retrieve(symbols, xhashid) );
        END (* if *);
        RETURN symbolentry^.id;
    END AddSymbol;

PROCEDURE PutSymbols;
    VAR
        xs, s: SYMBOLENTRYPTR;
        name: SystemTypes.String;
        i: CARDINAL;
    BEGIN
        s := symbolentries;
        WHILE s # NIL DO
            WITH s^ DO
		WriteCARDINAL(binStream, value);
		PutByte (ntype);
		PutByte (namelength-1);
		PutShort (0);
		PutShort (0);			
		name := HashStrings.StringAt (symbols, hashid);

		(* Namelengths include null byte *)
		FOR i := 0 TO namelength - 2 DO
		    PutByte (ORD (name^[i]));
		    END;
                s := link;
            END (* with *);
        END (* while *);

    END PutSymbols;

BEGIN
    size := 0;
    relocsize := 0; TextRelocList := NIL; DataRelocList := NIL;
    symbolcount := 0; symbolentries := NIL;
    stringtableindex := 4;
    symbols := HashStrings.New(50);
    relocatingtext := TRUE;
END UnixObj.
