IMPLEMENTATION MODULE labels;

FROM SYSTEM IMPORT ADDRESS, MAXCARD, TSIZE;
FROM SetUpCode IMPORT MaxCodeSize;
FROM Storage IMPORT ALLOCATE;
FROM SimpleStks IMPORT Stack, initialise, PushOnto, PopFrom, empty;

CONST sz   = 8; (* This value gives the number of
                   MCODE bytes each entry in the
                   label table represents
                *)


TYPE LinkPtr = POINTER TO Link;
     Link = RECORD
              CodeByteNo,
              Label,
              LabelUseCount,
              ExitLabel  : CARDINAL;
	      next : LinkPtr
	    END;


VAR LabelTable : ARRAY [0..MaxCodeSize DIV sz] OF LinkPtr;

    NewLink, Current, Last : LinkPtr;
    LabelNumber, count : CARDINAL;

    LabelStack : Stack;


PROCEDURE NewLabel () : CARDINAL;
    BEGIN
        INC (LabelNumber); RETURN LabelNumber
    END NewLabel;

PROCEDURE PushLabelStack (lab : CARDINAL);
    BEGIN
        PushOnto (LabelStack, lab);
    END PushLabelStack;

PROCEDURE PopLabelStack () : CARDINAL;
    VAR lab : CARDINAL;
    BEGIN
        IF empty (LabelStack) THEN
            RETURN none;
        ELSE
            lab := PopFrom (LabelStack);
	    RETURN lab;
        END (* if *);
    END PopLabelStack;

PROCEDURE SetNewLink (ByteNo, label, labelUseCount, JumpLabel : CARDINAL;
			                   Next : LinkPtr);
   BEGIN
      NEW (NewLink);
      WITH NewLink^ DO CodeByteNo := ByteNo;
		       Label := label;
		       LabelUseCount := labelUseCount;
		       ExitLabel := JumpLabel;
		       next := Next
		    END
   END SetNewLink;

PROCEDURE InsertNewLabel (ByteNo : CARDINAL) : CARDINAL;

   VAR current, last : LinkPtr;

   BEGIN
      current := LabelTable [ByteNo DIV sz];
      last := NIL;
      LOOP
	 IF (current = NIL)
	 OR (current^.CodeByteNo > ByteNo)
	 THEN 
              SetNewLink (ByteNo, NewLabel (), 1, none, current);
	      IF last = NIL
	      THEN LabelTable [ByteNo DIV sz] := NewLink
	      ELSE last^.next := NewLink
	      END;
	      IF (Current = current)
	      AND (ByteNo DIV sz = CurrentInstrByteNo DIV sz)
	      THEN IF ByteNo >= CurrentInstrByteNo
	           THEN Current := NewLink
		   ELSE Last := NewLink
		   END
	      END;
	      RETURN NewLink^.Label
	 ELSIF current^.CodeByteNo = ByteNo
	 THEN WITH current^ DO
		 IF Label = none THEN
		     Label := NewLabel ();
                     LabelUseCount := 1;
                 ELSE
                     INC (LabelUseCount);
		 END;
		 RETURN Label
	      END
	 END;
	 last := current; current := current^.next
      END
   END InsertNewLabel;



PROCEDURE StepLabelPointers (NoOfBytes : CARDINAL);

   VAR NewByteNo : CARDINAL;

   BEGIN
      NewByteNo := CurrentInstrByteNo + NoOfBytes;
      IF (CurrentInstrByteNo MOD sz + NoOfBytes) >= sz
      THEN Current := LabelTable [NewByteNo DIV sz];
	   Last := NIL
      END;
      LOOP
	 IF (Current = NIL) 
	 OR (Current^.CodeByteNo >= NewByteNo)
	 THEN EXIT
	 END;
	 Last := Current; Current := Current^.next
      END
   END StepLabelPointers;


PROCEDURE MarkCurrentByte (label : CARDINAL);
   BEGIN
      IF (Current <> NIL)
      AND (Current^.CodeByteNo = CurrentInstrByteNo)
      THEN Current^.ExitLabel := label
      ELSE SetNewLink (CurrentInstrByteNo, none, 0,
		       label, Current);
	   IF Last = NIL
	   THEN LabelTable [CurrentInstrByteNo DIV sz]
			      := NewLink
	   ELSE Last^.next := NewLink
	   END;
	   Last := NewLink
      END
   END MarkCurrentByte;


PROCEDURE CurrentLabel () : CARDINAL;
   BEGIN RETURN FindCurrent (TRUE) 
   END CurrentLabel;


PROCEDURE CurrentExitLabel () : CARDINAL;
   BEGIN RETURN FindCurrent (FALSE)
   END CurrentExitLabel;


PROCEDURE FindCurrent (LabelWanted : BOOLEAN) : CARDINAL;
   BEGIN
      IF (Current <> NIL) 
      AND (Current^.CodeByteNo = CurrentInstrByteNo)
      THEN WITH Current^ DO
	      IF LabelWanted
	      THEN RETURN Label
	      ELSE RETURN ExitLabel
	      END
	   END
      ELSE RETURN none
      END
   END FindCurrent;


(* Must only be used if CurrentLabel has returned value not equal to none *)
PROCEDURE DecCurrentLabelUseCount (amount : CARDINAL) : CARDINAL;
    BEGIN
        WITH Current^ DO
            DEC (LabelUseCount, amount);
            IF LabelUseCount = 0 THEN
                Label := none;
            END (* if *);
            RETURN Label;
        END (* with *);
    END DecCurrentLabelUseCount;


PROCEDURE FindExitLabel (ByteNo : CARDINAL) : CARDINAL;

   VAR current, last : LinkPtr;

   BEGIN
      current := LabelTable [ByteNo DIV sz];
      last := NIL;
      LOOP
	 WITH current^ DO
	    IF (current = NIL)
	    OR (CodeByteNo > ByteNo)
	    THEN RETURN none
	    ELSIF CodeByteNo = ByteNo
	    THEN RETURN ExitLabel
	    END;
	    last := current; current := next
	 END
      END
   END FindExitLabel;


VAR TransitionLabel : CARDINAL;

CONST LotsOfLabels = 512;
VAR PermanentLabels : POINTER TO ARRAY [0..LotsOfLabels] OF CARDINAL;


(* $T-,$R- *)
PROCEDURE SetUpPermanentLabelTable;
    BEGIN
        TransitionLabel := NewLabel ();
        ALLOCATE (PermanentLabels, (TransitionLabel+1)*TSIZE(ADDRESS));
    END SetUpPermanentLabelTable;


PROCEDURE GetPermanentLabelTable (VAR a : ADDRESS) : CARDINAL;
    VAR i : CARDINAL;
    BEGIN
        a := ADDRESS (PermanentLabels);
        FOR i := 0 TO TransitionLabel DO
            PermanentLabels^[i] := 0;
        END (* for *);
        RETURN TransitionLabel;
    END GetPermanentLabelTable;
(* $T+,$R+ *)


PROCEDURE InitialisePointers ();
   BEGIN 
      CurrentInstrByteNo := 0;
      CurrentByteNo := 0;
      Current := LabelTable [0]; Last := NIL
   END InitialisePointers;


(* Note that in this initialisation sequence the very
   first MCODE byte, i.e. the start of the table of
   addresses which precedes the code proper, is labelled
*)
BEGIN
    LabelNumber := none;
    SetNewLink (0, NewLabel (), 1, none, NIL);
    LabelTable [0] := NewLink;
    FOR count := 1 TO (MaxCodeSize DIV sz) DO
        LabelTable [count] := NIL
    END (* for *);
    initialise (LabelStack);
END labels.
