IMPLEMENTATION MODULE AddrReg; (* $T- *)

FROM SYSTEM IMPORT MAXCARD;
FROM items IMPORT ITEM, MODE, REGISTER, equal, PackItem,
                  slaveable, address, Max8BitInt,
		  UsesSlavedAddrReg, InDataReg,
		  PackRealStackItem, BytesPerWord;
FROM RegSlv IMPORT AdjustRegSlave, Slave, slave, MarkOnStack;
FROM PsStk IMPORT
    PseudStk, top, base, NotOnPsStk, FreeAnyRegUsedBy, PsStkState, STATE,
    STATEBIT, ScratchHwm;
FROM CaseFor IMPORT ForNesting;
FROM GenProcs IMPORT GenMovea, GenLea, GenMove, GenLeaExternal,
                     GenAddImm, GenMoveReg, GenAddaImm, SetFrameSize;
FROM UtilError IMPORT ReportError;
FROM Procs IMPORT InitSize, ParmSpace;

TYPE AddrRegSlvItem = RECORD
                         UseCount     : CARDINAL;
			 For          : CARDINAL;
			 External     : CARDINAL;
			 Slaving      : BOOLEAN;
			 Item         : ITEM;
                         Protected    : BOOLEAN;
                      END;


VAR AddrRegSlave : ARRAY [A0..A2] OF AddrRegSlvItem;
    empty        : AddrRegSlvItem;


PROCEDURE ProtectAddress (VAR Address : ITEM);
    BEGIN
        IF UsesSlavedAddrReg (Address) THEN
            AddrRegSlave [Address.Register].Protected := TRUE;
        END (* if *);
    END ProtectAddress;

PROCEDURE UnprotectAddress (VAR Address : ITEM);
    BEGIN
        IF UsesSlavedAddrReg (Address) THEN
            AddrRegSlave [Address.Register].Protected := FALSE;
        END (* if *);
    END UnprotectAddress;

(* Removes an address from the slave *)
PROCEDURE ClearFromAddrRegSlave (VAR item : ITEM);
    VAR Reg : REGISTER;
    BEGIN
        IF slaveable (item) OR address (item) THEN
            FOR Reg := A0 TO A2 DO
                WITH AddrRegSlave [Reg] DO
                    IF Slaving AND equal (item, Item) THEN
                        Slaving := FALSE
                    END (* if *);
                END (* with *);
            END (* for *);
        END (* if *);
    END ClearFromAddrRegSlave;

(* Frees register so it can be used for some other purpose *)
PROCEDURE ReleaseAddrReg (reg : REGISTER);
    BEGIN
        IF AddrRegSlave [reg].Protected THEN
            ReportError (TRUE,
                         "Attempt to release protected Address Register*N",
                         0, 0);
        END (* if *);
        AdjustAddrRegSlave (reg);
        AdjustPsStk (reg, NotOnPsStk);
        AddrRegSlave [reg] := empty;
    END ReleaseAddrReg;

PROCEDURE PutIntoAddrRegSlave (Reg : REGISTER; VAR item : ITEM);
    PROCEDURE valid (VAR i : ITEM; r : REGISTER) : BOOLEAN;
        BEGIN
            IF UsesSlavedAddrReg (i) THEN
                IF i.Register = r THEN
                    RETURN FALSE
                ELSE
                    WITH AddrRegSlave [i.Register] DO
                        RETURN Slaving OR (External # 0);
                    END (* with *);
                END (* if *);
            ELSE
                RETURN TRUE;
            END (* if *);
        END valid;
    BEGIN
        IF (address (item) OR slaveable (item))
           AND valid (item, Reg) THEN
            WITH AddrRegSlave [Reg] DO
                Slaving := TRUE; Item := item
            END (* with *);
        END (* if *);
    END PutIntoAddrRegSlave;

(* When a pseudo stack item which used an address
   register is taken off the pseudo stack the
   use count of that register must be altered
*)
PROCEDURE DecUseCount (Reg : REGISTER; PsStkPosn : CARDINAL);
    BEGIN
        DEC (AddrRegSlave [Reg].UseCount);
        EXCL (PsStkState [PsStkPosn], SlaveAddrReg);
    END DecUseCount;

PROCEDURE IncUseCount (Reg : REGISTER; PsStkPosn : CARDINAL);
    BEGIN
        INC (AddrRegSlave [Reg].UseCount);
        INCL (PsStkState [PsStkPosn], SlaveAddrReg);
    END IncUseCount;

(* Tests if the given address is slaved.
   If it is it returns TRUE and sets the
   variable 'Reg' to the address
   register slaving the address
*)
PROCEDURE InAddrRegSlave (VAR item : ITEM;
                          VAR Reg : REGISTER) : BOOLEAN;
    VAR reg : REGISTER;
    BEGIN
        IF slaveable (item) OR address (item) THEN
            FOR reg := A0 TO A2 DO
                WITH AddrRegSlave [reg] DO
                    IF Slaving AND equal (Item, item) THEN
                        Reg := reg; RETURN TRUE
                    END (* if *);
                END (* with *);
            END (* for *);
        END (* if *);
        RETURN FALSE
    END InAddrRegSlave;

PROCEDURE PrepareResultReg (ResultReg : REGISTER;
                            PsStkPosn : CARDINAL);
    VAR Reg : REGISTER;
    BEGIN
        AddrRegSlave [ResultReg] := empty;
        IF PsStkPosn <> NotOnPsStk THEN
            INC (AddrRegSlave [ResultReg].UseCount);
            INCL (PsStkState [PsStkPosn], SlaveAddrReg);
        END (* if *);
        AdjustAddrRegSlave (ResultReg);
    END PrepareResultReg;

PROCEDURE AdjustAddrRegSlave (reg : REGISTER);
    VAR Reg : REGISTER;
    VAR RegChanged : ARRAY [A0..A2] OF BOOLEAN;
    VAR NoChange : BOOLEAN;
    BEGIN
        FOR Reg := A0 TO A2 DO
            RegChanged [Reg] := FALSE;
        END (* for *);
        RegChanged [reg] := TRUE;
        REPEAT
            NoChange := TRUE;
            FOR Reg := A0 TO A2 DO
                WITH AddrRegSlave [Reg] DO
                    IF (NOT RegChanged [Reg]) AND Slaving
                    AND UsesSlavedAddrReg (Item)
                    AND RegChanged [Item.Register]
                    THEN
                        RegChanged [Reg] := TRUE;
                        NoChange := FALSE;
                    END (* if *)
                END (* with *);
            END (* for *);
        UNTIL NoChange;
        FOR Reg := A0 TO A2 DO
            IF RegChanged [Reg] THEN
                AddrRegSlave [Reg].Slaving := FALSE;
                AdjustRegSlave (Reg);
            END (* if *);
        END (* for *);
    END AdjustAddrRegSlave;

PROCEDURE FreeAddrReg (PsStkPosn : CARDINAL) : REGISTER;
    VAR Reg, MinForReg, LeastUsedReg : REGISTER;
        MinFor, MinUseCount, stage : CARDINAL;
    BEGIN
        FOR stage := 1 TO 3 DO
            FOR Reg := A0 TO A2 DO
                WITH AddrRegSlave [Reg] DO
                    CASE stage OF

             1 : IF (UseCount = 0) AND (For = 0)
                 AND (External = 0) AND NOT Slaving AND NOT Protected THEN
                     PrepareResultReg (Reg, PsStkPosn);
		     RETURN Reg;
                 END (* if *) |
             2 : IF (UseCount = 0) AND (For = 0)
                 AND (External = 0) AND NOT Protected THEN
                     PrepareResultReg (Reg, PsStkPosn);
		     RETURN Reg;
                 END (* if *) |
             3 : IF (UseCount = 0) AND (For = 0) AND NOT Protected THEN
                     PrepareResultReg (Reg, PsStkPosn);
		     RETURN Reg;
                 END (* if *)

                    END (* case *);
                END (* with *);
            END; (* Reg loop *)
        END; (* stage loop *)
        MinFor := MAXCARD; MinUseCount := 16;
        FOR Reg := A0 TO A2 DO
            WITH AddrRegSlave [Reg] DO
                IF (UseCount = 0) AND (For < MinFor) AND NOT Protected THEN
	            MinFor := For; MinForReg := Reg
	        END (* if *);
                IF (UseCount < MinUseCount) AND NOT Protected THEN
                    MinUseCount := UseCount;
                    LeastUsedReg := Reg
                END (* if *);
            END (* with *);
        END (* for *);
        IF MinFor <> MAXCARD THEN
            PrepareResultReg (MinForReg, PsStkPosn);
            RETURN MinForReg;
        ELSE
            IF MinUseCount <> MAXCARD THEN
                AdjustPsStk (LeastUsedReg, PsStkPosn);
	        PrepareResultReg (LeastUsedReg, PsStkPosn);
	        RETURN LeastUsedReg;
            ELSE
                ReportError (TRUE,
                             "Address Reg Slave clogged *N",
                             0, 0);
            END (* if *);
        END (* if *);
    END FreeAddrReg;

(* This procedure is used when an address register is about
   to have its value changed. It searches through the pseudo
   stack and if it finds an entry that uses 'AddrReg' as
   its base register then the value of that entry is moved
   to a safe place (the real stack or a data register);
   It then looks through the register slave and marks as free
   any items in that which are slaving addresses which use
   'AddrReg' as their base.
*)
PROCEDURE AdjustPsStk (reg : REGISTER; IgnorePosn : CARDINAL);
    VAR Point : CARDINAL;
        StkItem, RealStkPosn : ITEM;
	dest : ITEM;
    BEGIN
        FOR Point := base+1 TO top DO
            IF (SlaveAddrReg IN PsStkState [Point]) AND (Point # IgnorePosn) THEN
                StkItem := PseudStk [Point];
                IF StkItem.Register = reg THEN
                    slave (StkItem);
                    IF InDataReg (StkItem) THEN
	                MarkOnStack (StkItem.Register);
	                PseudStk [Point] := StkItem
                    ELSE
        PackItem (dest, AddrRegDisp, L, -(INTEGER (Point - base)
		 + InitSize - ParmSpace), NoReg);

                        IF address (StkItem) THEN
                            AddressMove (StkItem, dest)
                        ELSE
                            GenMove (StkItem, dest)
                        END (* if *);

                        PseudStk [Point] := dest
                    END (* if *);
                END (* if *);
            END (* if *);
        END (* for *);
	IF top > ScratchHwm
	THEN ScratchHwm := top;
            SetFrameSize (ScratchHwm)
   	END;

    END AdjustPsStk;

PROCEDURE AddressMove (VAR addr, StkPosn : ITEM);
    VAR TrueDisp : INTEGER;
    BEGIN
        GenMoveReg (addr.Register, StkPosn);
        TrueDisp := 4*addr.DispOrData;
        IF TrueDisp <> 0 THEN
             GenAddImm (TrueDisp, StkPosn)
        END (* if *);
    END AddressMove;

PROCEDURE MarkForFor (Reg : REGISTER);
    BEGIN
        AddrRegSlave [Reg].For := ForNesting
    END MarkForFor;

PROCEDURE FreeForFor () : REGISTER;
    VAR Reg : REGISTER;
    BEGIN
        Reg := FreeAddrReg (NotOnPsStk);
        AddrRegSlave [Reg].For := ForNesting;
        RETURN Reg;
    END FreeForFor;

PROCEDURE AddrRegStillAvailable (VAR Reg : REGISTER) : BOOLEAN;
    VAR reg : REGISTER;
    BEGIN
        FOR reg := A0 TO A2 DO
            IF AddrRegSlave [reg].For = ForNesting THEN
	        Reg := reg; RETURN TRUE
	    END (* if *);
        END (* for *);
        RETURN FALSE;
    END AddrRegStillAvailable;

PROCEDURE FreeAfterFor (Reg : REGISTER);
    BEGIN
        AddrRegSlave [Reg].For := 0;
    END FreeAfterFor;

PROCEDURE PurgeAddrRegSlave (AfterProcCall : BOOLEAN);
    VAR Reg : REGISTER;
    BEGIN
        FOR Reg := A0 TO A2 DO
            WITH AddrRegSlave [Reg] DO
                IF (For = 0) OR AfterProcCall THEN
                    AddrRegSlave [Reg] := empty
	        ELSE
                    Slaving := FALSE; External := 0;
                    (* For not affected *)
                END (* if *);
	    END (* with *);
        END (* for *);
    END PurgeAddrRegSlave;

(* When a stack store is made it cannot
   affect values on the data frame table
   or the address of a FOR control variable
   so a limited purge is done.
*)
PROCEDURE PurgeVariablesInAddrRegSlave ();
    VAR Reg : REGISTER;
    BEGIN
        FOR Reg := A0 TO A2 DO
            WITH AddrRegSlave [Reg] DO
                IF Slaving AND (External = 0) AND (For = 0) THEN
	            Slaving := FALSE;
	        END (* if *);
            END (* with *);
        END (* for *);
    END PurgeVariablesInAddrRegSlave;

PROCEDURE SlavedExternal (offset : CARDINAL; VAR reg : REGISTER) : BOOLEAN;
    VAR Reg : REGISTER;
    BEGIN
        FOR Reg := A0 TO A2 DO
            WITH AddrRegSlave [Reg] DO
                IF External = offset THEN
                    reg := Reg;
                    RETURN TRUE;
                END (* if *);
            END (* with *);
        END (* for *);
        RETURN FALSE;
    END SlavedExternal;

PROCEDURE SlaveExternal (offset : CARDINAL; reg : REGISTER);
    BEGIN
        AddrRegSlave [reg].External := offset;
    END SlaveExternal;

PROCEDURE GetExternalBaseAddr (disp : INTEGER) : REGISTER;
    VAR AddrReg : REGISTER;
    BEGIN
        IF NOT SlavedExternal (disp, AddrReg) THEN
            AddrReg := FreeAddrReg (NotOnPsStk);
            GenLeaExternal (disp, AddrReg);
            SlaveExternal (disp, AddrReg);
        END;
        RETURN AddrReg
    END GetExternalBaseAddr;

PROCEDURE GetAddrOnPsStkIntoUseableForm
                  (PsStkPosn : CARDINAL; ForIndex : BOOLEAN);
    VAR AddrReg : REGISTER;
        item, SlvItem : ITEM;
    BEGIN
        item := PseudStk [PsStkPosn];
        IF address (item)
        AND (NOT ForIndex OR (item.DispOrData * 4 <= Max8BitInt)) THEN
            RETURN
        ELSE
            FreeAnyRegUsedBy (PsStkPosn);
            IF address (item) THEN
                AddrReg := FreeAddrReg (PsStkPosn);
                GenLea (item, AddrReg);
                PutIntoAddrRegSlave (AddrReg, item);
            ELSIF InAddrRegSlave (item, AddrReg) THEN
                IncUseCount (AddrReg, PsStkPosn);
            ELSE
                AddrReg := FreeAddrReg (PsStkPosn);
                Slave (item, SlvItem);
                GenMovea (SlvItem, AddrReg);
                PutIntoAddrRegSlave (AddrReg, item)
            END (* if *);
        END (* if *);
        PackItem (PseudStk [PsStkPosn],
                  ADDRaddrRegDisp, AddrReg, 0, NoReg)
    END GetAddrOnPsStkIntoUseableForm;

PROCEDURE GetAddrRegsForBlockMove (VAR source, dest : ITEM);

    PROCEDURE GetItemIntoAddrReg (VAR item : ITEM; reg : REGISTER);
        BEGIN
            IF address (item) THEN
                GenLea (item, reg);
            ELSE
                slave (item);
                GenMovea (item, reg);
            END (* if *);
        END GetItemIntoAddrReg;

    PROCEDURE PostIncrement (VAR item : ITEM; reg : REGISTER);
        BEGIN
            PackItem (item, PostInc, reg, 0, NoReg);
        END PostIncrement;

    VAR SourceAddrReg, DestAddrReg, SaveRegister : REGISTER;
        SaveUseCount : CARDINAL;
        SourceSlaved, DestSlaved : BOOLEAN;
    BEGIN
        SourceSlaved := FALSE; DestSlaved := FALSE;
        IF InAddrRegSlave (source, SourceAddrReg)
           AND (AddrRegSlave [SourceAddrReg].UseCount = 0) THEN
            ReleaseAddrReg (SourceAddrReg);
            SourceSlaved := TRUE;
        ELSE
            SourceAddrReg := FreeAddrReg (NotOnPsStk);
        END (* if *);
        IF InAddrRegSlave (dest, DestAddrReg)
           AND (AddrRegSlave [DestAddrReg].UseCount = 0) THEN
            ReleaseAddrReg (DestAddrReg);
            DestSlaved := TRUE;
        ELSE
            WITH AddrRegSlave [SourceAddrReg] DO
                SaveUseCount := UseCount;
                UseCount := 16;
                DestAddrReg := FreeAddrReg (NotOnPsStk);
                UseCount := SaveUseCount;
            END (* with *);
        END (* if *);
        IF NOT (SourceSlaved AND DestSlaved) THEN
            IF SourceSlaved THEN
                GetItemIntoAddrReg (dest, DestAddrReg);
            ELSIF DestSlaved THEN
                GetItemIntoAddrReg (source, SourceAddrReg);
            ELSE
                IF UsesSlavedAddrReg(source)
                   AND (source.Register=DestAddrReg) THEN
                    SaveRegister := SourceAddrReg;
                    SourceAddrReg := DestAddrReg;
                    DestAddrReg := SaveRegister;
                END (* if *);
                GetItemIntoAddrReg (dest, DestAddrReg);
                GetItemIntoAddrReg (source, SourceAddrReg);
            END (* if *);
        END (* if *);
        PostIncrement (source, SourceAddrReg);
        PostIncrement (dest, DestAddrReg);
    END GetAddrRegsForBlockMove;

BEGIN
   WITH empty DO
      UseCount := 0; For := 0; External := 0; 
      Slaving := FALSE; Protected := FALSE;
   END;
END AddrReg.
