IMPLEMENTATION MODULE AuxGen1;

FROM items IMPORT ITEM, REGISTER, Operator, equal, 
                  InDataReg, immediate, NullRegItem,
		  DataReg0;
FROM RegSlv IMPORT Slave, protect, FreeReg, StopSlaving,
                   DeleteFromSlave, PutIntoSlave, UseBitSet;
FROM GenProcs IMPORT GenSimpleDyadicOp, GenMove, GenClr, GenExg, GenNeg;
FROM GenOp IMPORT SetResultItem;

PROCEDURE Option1 (op : Operator; VAR opd, dest : ITEM);
   BEGIN
      GenSimpleDyadicOp (op, opd, dest);
      StopSlaving (dest);
   END Option1;


PROCEDURE Option2 (op : Operator; VAR opd1, opd2, dest : ITEM);
   BEGIN
      GenSimpleDyadicOp (op, opd1, opd2);
      GenMove (opd2, dest);
      PutIntoSlave (opd2, dest, TRUE)
   END Option2;


PROCEDURE Option3 (op : Operator; VAR opd, reg, dest : ITEM);
   BEGIN
      GenMove (opd, reg);
      GenSimpleDyadicOp (op, reg, dest);
      StopSlaving (dest);
      PutIntoSlave (reg, opd, TRUE);
   END Option3;


PROCEDURE Option4 (op : Operator; VAR opd1, opd2, dest : ITEM);
   BEGIN
      GenMove (opd2, dest);
      GenSimpleDyadicOp (op, opd1, dest);
      StopSlaving (dest);
   END Option4;


PROCEDURE Option5 (op : Operator; VAR opd1, opd2,
                                     reg, dest : ITEM);
   BEGIN
      GenMove (opd2, reg);
      GenSimpleDyadicOp (op, opd1, reg);
      GenMove (reg, dest);
      PutIntoSlave (reg, dest, TRUE)
   END Option5;

PROCEDURE Option6 (op : Operator; VAR opd1, opd2, reg, dest : ITEM);
    BEGIN
        GenMove (opd1, reg);
        PutIntoSlave (reg, opd1, TRUE);
        GenMove (opd2, dest);
        GenSimpleDyadicOp (op, reg, dest);
        StopSlaving (dest);
    END Option6;

   (* Add, And & Or are dealt with together because they are
      all commutative operations (e.g. A AND B = B AND A) &
      because the addressing modes available on the 68000
      are the same for each operation (of the form :
         <ea> + data reg -> data reg       or
         <ea> + data reg -> <ea>           or   
         <ea> + immediate -> <ea>             )
   *)
PROCEDURE AddAndOr (Op : Operator; VAR Opd1, Opd2, Dest : ITEM);
     VAR SlvOpd1, SlvOpd2, SlvDest, Reg : ITEM;
     BEGIN
        Slave (Opd1, SlvOpd1);
        Slave (Opd2, SlvOpd2);
        Slave (Dest, SlvDest);
        IF equal (Dest, DataReg0) THEN
        (* This should only be true if are about to
           return from a function *)
            IF equal (SlvOpd1, DataReg0) THEN
                GenSimpleDyadicOp (Op, Opd2, DataReg0)
            ELSE
	        GenMove (SlvOpd2, DataReg0);
	        GenSimpleDyadicOp (Op, Opd1, DataReg0)
            END (* if *);
        ELSIF equal (Dest, NullRegItem) THEN
            IF InDataReg (SlvOpd1) THEN
                Option1 (Op, SlvOpd2, SlvOpd1);
                SetResultItem (SlvOpd1);
            ELSIF InDataReg (SlvOpd2) THEN
                Option1 (Op, SlvOpd1, SlvOpd2);
                SetResultItem (SlvOpd2);
            ELSE
                FreeReg (Reg, NOT UseBitSet);
                Option4 (Op, Opd1, Opd2, Reg);
                SetResultItem (Reg)
            END (* if *);
        ELSIF equal (Opd1, Dest) 
        AND (InDataReg (SlvOpd2) OR immediate (SlvOpd2)) THEN
            Option1 (Op, SlvOpd2, Dest)
        ELSIF equal (Opd2, Dest)
        AND (InDataReg (SlvOpd1) OR immediate (SlvOpd1)) THEN
            Option1 (Op, SlvOpd1, Dest)
        ELSIF InDataReg (SlvOpd1) THEN
            Option2 (Op, SlvOpd2, SlvOpd1, Dest)
        ELSIF InDataReg (SlvOpd2) THEN
            Option2 (Op, Opd1, SlvOpd2, Dest)
        ELSIF equal (Opd1, Dest) THEN
            FreeReg (Reg, NOT UseBitSet);
            Option3 (Op, Opd2, Reg, Dest)
        ELSIF equal (Opd2, Dest) THEN
            FreeReg (Reg, NOT UseBitSet);
            Option3 (Op, Opd1, Reg, Dest)
        ELSIF immediate (Opd1) THEN
            Option4 (Op, Opd1, Opd2, Dest)
        ELSIF immediate (Opd2) THEN
            Option4 (Op, Opd2, Opd1, Dest)
        ELSIF InDataReg (SlvDest) THEN
            Option5 (Op, Opd1, Opd2, SlvDest, Dest)
        ELSE 
            FreeReg (Reg, NOT UseBitSet);
            Option5 (Op, Opd1, Opd2, Reg, Dest)
        END
    END AddAndOr;
      

   (* The effect required is to subtract opd1 from opd2 and to
      put the result in Dest. The available addressing modes on
      the 68000 are :
	  <ea> - data reg -> <ea>
	  data reg - <ea> -> data reg
	  <ea> - immediate -> <ea>
   *)
PROCEDURE Subtract (Op : Operator; VAR Opd1, Opd2, Dest : ITEM);
    VAR SlvOpd1, SlvOpd2, SlvDest, Reg : ITEM;
    BEGIN
        Slave (Opd1, SlvOpd1);
        Slave (Opd2, SlvOpd2);
        Slave (Dest, SlvDest);
        IF equal (Opd1, Opd2) THEN
            IF equal (Dest, NullRegItem) THEN
	        FreeReg (Reg, NOT UseBitSet);
		GenClr (Reg); 
		SetResultItem (Reg)
            ELSE 
                GenClr (Dest); StopSlaving (Dest)
            END (* if *);
        ELSIF equal (Dest, DataReg0) THEN
            IF equal (SlvOpd2, DataReg0) THEN
                GenSimpleDyadicOp (Op, Opd1, DataReg0)
	    ELSIF equal (SlvOpd1, DataReg0) 
	          AND InDataReg (SlvOpd2) THEN
	        GenExg (SlvOpd2, SlvOpd1);
	        GenSimpleDyadicOp (Op, SlvOpd1, DataReg0)
            ELSIF equal (SlvOpd1, DataReg0) THEN
	        GenNeg (SlvOpd1);
	        IF Op = Sub THEN
		    GenSimpleDyadicOp (Add, Opd2, DataReg0)
		ELSE (*Usub*)
		    GenSimpleDyadicOp (Uadd, Opd2, DataReg0)
                END (* if *);
            ELSE 
                Option4 (Op, SlvOpd1, SlvOpd2, DataReg0);
            END (* if *);
        ELSIF equal (Dest, NullRegItem) THEN
            IF InDataReg (SlvOpd2) THEN
                Option1 (Op, SlvOpd1, SlvOpd2);
                SetResultItem (SlvOpd2)
            ELSE
                FreeReg (Reg, NOT UseBitSet);
                Option4 (Op, SlvOpd1, Opd2, Reg);
                SetResultItem (Reg)
            END (* if *);
        ELSIF equal (Opd2, Dest)
              AND (InDataReg (SlvOpd1) OR immediate (Opd1)) THEN
            Option1 (Op, SlvOpd1, Dest)
        ELSIF InDataReg (SlvOpd2) THEN
            Option2 (Op, SlvOpd1, SlvOpd2, Dest)
        ELSIF equal (Opd2, Dest) THEN
            FreeReg (Reg, NOT UseBitSet);
            Option3 (Op, Opd1, Reg, Dest)
        ELSIF immediate (Opd1) OR InDataReg (SlvOpd1) THEN
            Option4 (Op, SlvOpd1, Opd2, Dest)
        ELSIF InDataReg (SlvDest) THEN
            Option5 (Op, Opd1, Opd2, SlvDest, Dest)
        ELSE 
            FreeReg (Reg, NOT UseBitSet);
            Option5 (Op, Opd1, Opd2, Reg, Dest)
        END (* if *);
    END Subtract;

   (* The effect required is to XOR Opd1 and Opd2 together
      and put the result in dest. The addressing modes
      available are :
	  data reg XOR <ea> -> <ea>
	  immediate XOR <ea> -> <ea>
   *)
PROCEDURE ExclusiveOr (VAR Opd1, Opd2, Dest : ITEM);
    VAR SlvOpd1, SlvOpd2, SlvDest, Reg1, Reg2 : ITEM;
    BEGIN
        Slave (Opd1, SlvOpd1);
        Slave (Opd2, SlvOpd2);
        Slave (Dest, SlvDest);
        (* turns out that for this operation
           there is no need to put in any special
           working to handle the case when dest is
           data reg 0 as the normal handling of the
           operation handles it as well as possible *)
        IF equal (Opd1, Opd2) THEN
            IF equal (Dest, NullRegItem) THEN
	        FreeReg (Reg1, NOT UseBitSet);
		GenClr (Reg1); 
		SetResultItem (Reg1)
            ELSE 
                GenClr (Dest); StopSlaving (Dest)
            END (* if *);
        ELSIF equal (Dest, NullRegItem) THEN
            IF InDataReg (SlvOpd1) 
               AND (immediate (Opd2) OR InDataReg (SlvOpd2)) THEN
                Option1 (Xor, SlvOpd2, SlvOpd1);
                SetResultItem (SlvOpd1)
            ELSIF InDataReg (SlvOpd2) AND immediate (Opd1) THEN
                Option1 (Xor, Opd1, SlvOpd2);
                SetResultItem (SlvOpd2);
            ELSIF (immediate (Opd1) OR InDataReg (SlvOpd1)) THEN
                FreeReg (Reg1, NOT UseBitSet);
                Option4 (Xor, SlvOpd1, Opd2, Reg1);
                SetResultItem (Reg1)
            ELSIF  (immediate (Opd2) OR InDataReg (SlvOpd2)) THEN
                FreeReg (Reg1, NOT UseBitSet);
                Option4 (Xor, SlvOpd2, Opd1, Reg1);
                SetResultItem (Reg1)
            ELSE
                FreeReg (Reg1, UseBitSet);
                FreeReg (Reg2, NOT UseBitSet);
                Option6 (Xor, Opd1, Opd2, Reg1, Reg2);
                SetResultItem (Reg2) 
            END (* if *);
        ELSIF equal (Opd1, Dest)
              AND (InDataReg (SlvOpd2) OR immediate (Opd2)) THEN
            Option1 (Xor, SlvOpd2, Dest);
        ELSIF equal (Opd2, Dest)
              AND (InDataReg (SlvOpd1) OR immediate (Opd1)) THEN
            Option1 (Xor, SlvOpd1, Dest);
        ELSIF InDataReg (SlvOpd1) 
              AND (immediate (Opd2) OR InDataReg (SlvOpd2)) THEN
            Option2 (Xor, SlvOpd2, SlvOpd1, Dest);
        ELSIF InDataReg (SlvOpd2) AND immediate (Opd1) THEN
            Option2 (Xor, Opd1, SlvOpd2, Dest);
        ELSIF equal (Opd1, Dest) THEN
            FreeReg (Reg1, NOT UseBitSet);
            Option3 (Xor, Opd2, Reg1, Dest)
        ELSIF equal (Opd2, Dest) THEN
            FreeReg (Reg1, NOT UseBitSet);
            Option3 (Xor, Opd1, Reg1, Dest);
        ELSIF immediate (Opd1) OR InDataReg (SlvOpd1) THEN
            Option4 (Xor, SlvOpd1, Opd2, Dest);
        ELSIF immediate (Opd2) OR InDataReg (SlvOpd2) THEN
            Option4 (Xor, SlvOpd2, Opd1, Dest);
        ELSE 
            FreeReg (Reg1, NOT UseBitSet);
            Option6 (Xor, Opd1, Opd2, Reg1, Dest);
        END (* if *);
    END ExclusiveOr;

END AuxGen1.
