(*
    Title: 	BinOutput - Implementation
    LastEdit:	"Mon Sep 17 11:42:42 1984"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE BinOutput;

FROM SYSTEM IMPORT BYTESPERWORD, ADDRESS;

IMPORT Object; FROM Object IMPORT ExtrnMPStacks, ExtrnMPCounts,
				  ExtrnMBStacks, ExtrnMBCounts,
				  ExtrnOSStacks, ExtrnOSCounts,
				  OSProcNames, OSNAME, MaxNStProcs,
				  RelocCount, RelocStack;
FROM SimpleStks IMPORT PopFrom;
IMPORT Strings; FROM Strings IMPORT String;
FROM Streams IMPORT Stream;
FROM UnixObj IMPORT 
    PutHeader, PutByte, PutLong, SizeNow, EndText, EndTextReloc, SetObjStream,
    AddSymbol, AddRelocation, Ntext, Ndata, Nbss, Next, Nundf, Flush,
    PutSeg;
FROM ReadMcd IMPORT IMPORTFRAME, IMPORTFRAMEPTR, MODULENAME, MODULENAMESIZE,
    Stringframe, StringSize, Headerframe, Importlist, ImportlistLength;
FROM MainPass IMPORT ProcedureCount;
FROM UtilError IMPORT ReportError;
VAR
    ModuleName:  MODULENAME;
    CodeModuleName, DataModuleName, 
    VModuleName, TModuleName: ARRAY [0 .. 30] OF CHAR;
    CodeId, DataId: ARRAY [0 .. 2] OF CHAR;
    IDataSize, UDataSize, CodeTable: CARDINAL;
    underline: ARRAY [0 .. 1] OF CHAR;
    proctabindex: CARDINAL;
    ProcName: ARRAY [0 .. 79] OF CHAR;
    Dsymid, Tsymid : CARDINAL;
    Psymid : ARRAY [0..1001] OF CARDINAL;

PROCEDURE OutputObjectCode(s: Stream);
    CONST lotsofprocs = 100;
    VAR
        dummy, symbid, i : CARDINAL;
	proctab : POINTER TO ARRAY [0..lotsofprocs] OF CARDINAL;
    BEGIN
        SetObjStream (s);
        ModuleName := Headerframe^.name;
        Strings.CopyCC(ModuleName, CodeModuleName);
        Strings.CopyCC(ModuleName, DataModuleName);
        Strings.CopyCC(ModuleName, VModuleName);
        Strings.CopyCC(ModuleName, TModuleName);
        Strings.ConcatCC(CodeModuleName, CodeId, CodeModuleName);
        Strings.ConcatCC(DataModuleName, DataId, DataModuleName);
        Strings.ConcatCC(VModuleName, "__v", VModuleName);
        Strings.ConcatCC(TModuleName, "__t", TModuleName);

        PutHeader(407B);        (* updated later *)
        IF ODD(StringSize) THEN
            INC(StringSize);
        END (* if *);
        (* we must know this before outputting the text *)
        TextSize := Align(2, Object.ObjCodeBytes() +
                        12, FALSE);

        dummy := Align(2, SizeNow(), TRUE);  (* force word alignment *)
        CodeTable := SizeNow();
        CodeStart := CodeTable;
        symbid := AddSymbol(CodeModuleName, Ntext+Next, CodeStart);
        Tsymid := AddSymbol(TModuleName, Ntext, CodeTable);
	proctab := ADDRESS (Object.ObjCodeAddress());
        PutSeg (Object.ObjCodeAddress(), Object.ObjCodeBytes());

	FOR i := 0 TO ProcedureCount -1 DO
	    ProcSymbolName (ModuleName, i, ProcName);
	    Psymid [i] := AddSymbol (ProcName, Ntext + Next,
			(* reverse *) (proctab^[i])); (* SORRY *)
(*	    AddRelocation (CodeTable + i * 4, Tsymid, FALSE);  *)
 	    END;

        (* output debugger info *)
        Dsymid := AddSymbol(DataModuleName, Ndata+Next, TextSize);
        symbid := AddSymbol(VModuleName, Ntext+Next, SizeNow());
        PutLong(SizeNow()-CodeStart);
        PutLong(ProcedureCount);
        PutLong(0);

        dummy := Align(2, SizeNow(), TRUE);

        EndText();

        OutputData();
	OutputTextRelocations();
        EndTextReloc();
	OutputDataRelocations();
        Flush(UDataSize);
    END OutputObjectCode;

(*PROCEDURE reverse (c: CARDINAL) : CARDINAL;
    VAR high, low, res : CARDINAL;
    BEGIN
	high := c DIV 65536; low := c MOD 65536;
    	res := low MOD 256;
    	res := res * 256 + low DIV 256;
    	res := res * 256 + high MOD 256;
    	res := res * 256 + high DIV 256;
        RETURN (res);
    END reverse;
*)
PROCEDURE Align(n: CARDINAL; sz: CARDINAL; output: BOOLEAN): CARDINAL;
    VAR
        i, mod: CARDINAL;
    BEGIN
        CASE n OF
          1: mod := 2;
        | 2: mod := 4;
        END (* case *);
        IF (sz MOD mod) # 0 THEN
            FOR i := 1 TO mod - (sz MOD mod) DO
                IF output THEN
                    PutByte(0);
                END (* if *);
                INC(sz);
            END (* for *);
        END (* if *);
        RETURN sz;
    END Align;

PROCEDURE OutputStrings();
    VAR i : CARDINAL;
    BEGIN
        IF StringSize # 0
    	THEN FOR i := 0 TO StringSize - 1 DO
	        PutByte (CARDINAL (Stringframe [i]));
            END;
	END;
    END OutputStrings;

PROCEDURE OutputData;
    VAR
        i: CARDINAL;
    BEGIN
	FOR i := 1 TO Headerframe^.datasize - (StringSize DIV BYTESPERWORD) DO
            PutLong(0);
	    END;
	OutputStrings();
    END OutputData;

PROCEDURE OutputDataRelocations;
    BEGIN
	AddRelocation (0, Tsymid, FALSE);
    END OutputDataRelocations;

PROCEDURE OutputTextRelocations;
    VAR i, j : CARDINAL;
	frame : IMPORTFRAME;
	frameptr: IMPORTFRAMEPTR;
	offset, procnum, symid : CARDINAL;
	symname : ARRAY [0..80] OF CHAR;

    BEGIN
	FOR i := 1 TO RelocCount DO
	    AddRelocation (PopFrom (RelocStack), Tsymid, FALSE);
	    END;

	FOR j := 1 TO ExtrnMBCounts^ [0] DO
	    offset := PopFrom (ExtrnMBStacks^ [0]);
	    AddRelocation (CodeTable + offset, Dsymid, TRUE);
 	    END;

        FOR j := 1 TO ExtrnMPCounts^ [0] DO
	    offset := PopFrom (ExtrnMPStacks^ [0]);
	    procnum := PopFrom (ExtrnMPStacks^ [0]);
	    AddRelocation (CodeTable + offset, Psymid[procnum], TRUE);
	    END;

    	frameptr := Importlist;
	FOR i := 1 TO ImportlistLength DO
	    frame := frameptr^;

	    IF ExtrnMPCounts^ [i] # 0 THEN
	        FOR j := 1 TO ExtrnMPCounts^ [i] DO
		    offset := PopFrom (ExtrnMPStacks^ [i]);
		    procnum := PopFrom (ExtrnMPStacks^ [i]);
		    ProcSymbolName (frame.name, procnum, symname);
		    symid := AddSymbol (symname, Nundf + Next, 0);
		    AddRelocation (CodeTable + offset, symid, TRUE);
		    END;
		END;

    	    IF ExtrnMBCounts^ [i] # 0 THEN
	        Strings.ConcatCC (frame.name, DataId, symname);
		symid := AddSymbol (symname, Nundf + Next, 0);
	        FOR j := 1 TO ExtrnMBCounts^ [i] DO
		    offset := PopFrom (ExtrnMBStacks^ [i]);
		    AddRelocation (CodeTable + offset, symid, TRUE);
		    END;
    		END;

	    frameptr := frame.link;
	    END;

	FOR i := 0 TO MaxNStProcs DO
	    IF ExtrnOSCounts [i] # 0 THEN
	        symid := AddSymbol (OSProcNames[i]^, Nundf + Next, 0);
		FOR j := 1 TO ExtrnOSCounts [i] DO
		    offset := PopFrom (ExtrnOSStacks [i]);
		    AddRelocation (CodeTable + offset, symid, TRUE);
		    END;
	        END;
	    END;
 	END OutputTextRelocations;

PROCEDURE ProcSymbolName (ModuleName : ARRAY OF CHAR; procnum : CARDINAL;
				VAR ProcName: ARRAY OF CHAR);
    VAR numchar : ARRAY [0..4] OF CHAR;
        huns, tens, ones : CARDINAL;
    BEGIN
	IF procnum = 0 THEN
	    numchar [0] := '_';
	    numchar [1] := '_';
	    numchar [2] := 0C;
	    Strings.ConcatCC (ModuleName, numchar, ProcName);
	    Strings.ConcatCC (ProcName, ModuleName, ProcName);
	ELSE
	    huns := procnum DIV 100;
	    tens := (procnum - huns * 100) DIV 10;
	    ones := procnum MOD 10;
            numchar [0] := '_';
	    numchar [1] := CHAR (ORD('0') + huns);
	    numchar [2] := CHAR (ORD('0') + tens);
	    numchar [3] := CHAR (ORD('0') + ones);
   	    numchar [4] := 0C;
	    Strings.ConcatCC (ModuleName, numchar, ProcName);
	    END;
    END ProcSymbolName;
BEGIN
    CodeId := "__c";
    DataId := "__d";
    UDataSize := 0;
    underline[0] := '_'; underline[1] := 0C;
END BinOutput.
