IMPLEMENTATION MODULE CaseFor;

FROM SYSTEM IMPORT ADDRESS;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM items IMPORT ITEM, MODE, REGISTER, ConditionCode,
                  Operator, PackItem, InDataReg, immediate,
		  address, PackRealStackItem, StackPreDec,
                  PackReg, Int16Bit, BytesPerWord;
FROM RegSlv IMPORT FreeReg, UseBitSet, slave;
FROM MainPass IMPORT ExtraBytesProcessed;
FROM AddrReg IMPORT InAddrRegSlave, FreeForFor, MarkForFor, FreeAfterFor, 
                    PutIntoAddrRegSlave, FreeAddrReg, AddrRegStillAvailable,
                    ProtectAddress, UnprotectAddress;
FROM SimpleStks IMPORT Stack, PushOnto, TopValueOf, PopFrom,
                       empty, initialise, RemoveTopItemFrom;
FROM CodeProcs IMPORT next2, next4, next, StepToNextInstruction;
FROM labels IMPORT CurrentLabel, CurrentExitLabel;
FROM PsStk IMPORT DoAnyPendingOp, PseudStk, top, FreeAnyRegUsedBy, NotOnPsStk,
                  AssertEmptyPsStk, DecTopOfPsStk;
FROM Relations IMPORT reverse;
FROM labels IMPORT PushLabelStack, NewLabel;
FROM GenProcs IMPORT GenMove, GenPushLabel, GenLabel,
                     CheckedGenCmp, GenCmpImm, GenSubImm, GenRelBranch,
		     GenClr, GenShiftImm, GenJump, GenLongBcc,
		     GenTableEntry, GenRtn, GenBra, GenPea,
		     GenCmp, GenBcc, GenLea, GenMovea,
		     GenAddImm, GenAddaImm, GenIndexedMovea;
FROM MCMnemonics IMPORT ENTC, FOR1, DECS;

TYPE
    VALONSTACK = POINTER TO VALONSTACKDATA;
    VALONSTACKDATA =
        RECORD
            Next:   VALONSTACK;
            Saved:  BOOLEAN;
        END;

CONST TwoBytes = 2; FourBytes = 4;

VAR CaseStack, ForStack : Stack;
VAR ValsOnStack: VALONSTACK;

PROCEDURE ValOnStack (saved: BOOLEAN);
    VAR v: VALONSTACK;
BEGIN
    NEW (v);
    v^.Next := ValsOnStack;
    v^.Saved := saved;
    ValsOnStack := v;
END ValOnStack;

PROCEDURE ValOffStack;
    VAR v: VALONSTACK;
BEGIN
    v := ValsOnStack^.Next;
    DISPOSE (ValsOnStack);
    ValsOnStack := v;
END ValOffStack;

PROCEDURE case (Instruction : CARDINAL);
    BEGIN
        IF Instruction = ENTC THEN
            EnterCase ()
        ELSE (* EXC *)
            GenRtn ()
        END (* if *);
    END case;


PROCEDURE EnterCase ();
   VAR Reg, PsStkTop : ITEM;
       DataRegister, AddressRegister : REGISTER;
   BEGIN
      DoAnyPendingOp ();
      PsStkTop := PseudStk [top];
      IF InDataReg (PsStkTop)
      THEN DataRegister := PsStkTop.Register
      ELSE FreeReg (Reg, NOT UseBitSet);
           GenMove (PsStkTop, Reg);
	   DataRegister := Reg.Register
      END;
      DecTopOfPsStk (1);
      AddressRegister := FreeAddrReg (NotOnPsStk);
      PushOnto (CaseStack, ORD (AddressRegister));
      PushOnto (CaseStack, ORD (DataRegister));
      ValOnStack (FALSE);
      GenLongBcc (T);
   END EnterCase;


PROCEDURE SelectCase ();
    VAR Reg, AddressRegItem : ITEM;
        AddressRegister : REGISTER;
        cc : ConditionCode;
        low, hi : INTEGER;
        entries, count, label1, label2 : CARDINAL;
    BEGIN
        GenLabel ();
        PackReg (Reg, VAL (REGISTER, PopFrom (CaseStack)));
        AddressRegister := VAL (REGISTER, PopFrom (CaseStack));
        PackItem (AddressRegItem, AddrRegDisp, AddressRegister, 0, NoReg);
        GenPushLabel ();
        low := INTEGER (next4 ());
        StepToNextInstruction (FourBytes);
        hi := INTEGER (next4 ());
        entries := CARDINAL (ABS (hi-low)) + 2;
        IF low > hi THEN cc := HI;
        ELSIF low = hi THEN cc := NE;
        ELSE cc := GT;
        END (* if *);
        GenCmpImm (hi, Reg);
        (* checking if Reg is less than or equal to the upper bound *)
        label1 := NewLabel (); label2 := NewLabel ();
        GenRelBranch (cc, label1);
        GenSubImm (low-1, Reg);
        GenRelBranch (cc, label2);
        PushLabelStack (label1);
        GenClr (Reg);
        PushLabelStack (label2);
        GenShiftImm (Shl, 2, Reg);
        GenIndexedMovea (Reg, AddressRegister);
        GenJump (AddressRegItem);

        (* now handle case label table *)

        StepToNextInstruction (FourBytes);
        GenLabel ();
        FOR count := 1 TO entries DO
            GenTableEntry ();
            StepToNextInstruction (TwoBytes)
        END (* for *);
        ValOffStack;
    END SelectCase;


PROCEDURE for (Instruction : CARDINAL);
    BEGIN 
        IF Instruction = FOR1 THEN
            For1 ()
        ELSE 
            For2 ()
        END (* if *);
    END for;

TYPE FORATTR = (ImmediateHi, AddressableControlVar);
TYPE FORTYPE = SET OF FORATTR;

PROCEDURE SaveItem (VAR i : ITEM);
    VAR t : POINTER TO ITEM;
    BEGIN
        NEW (t);
        t^ := i;
        PushOnto (ForStack, CARDINAL (t));
    END SaveItem;

PROCEDURE RestoreItem (VAR i : ITEM);
    VAR t : POINTER TO ITEM;
    BEGIN
        t := ADDRESS (PopFrom (ForStack));
        i := t^;
        DISPOSE (t);
    END RestoreItem;

PROCEDURE DirectlyAddressable (VAR i : ITEM) : BOOLEAN;
    BEGIN
        WITH i DO
            RETURN (Mode = ADDRaddrRegDisp)
                AND ((Register = L) OR (Register = G));
        END (* if *);
    END DirectlyAddressable;

PROCEDURE For1 ();
    VAR count : CARDINAL;
        hi, low, adr, Reg, ControlVar : ITEM;
        GoingUp, immhi, immlo : BOOLEAN;
        cc : ConditionCode;
        AddrReg : REGISTER;
        ForType : FORTYPE;
    BEGIN
        INC (ForNesting);
        ForType := FORTYPE {};
        DoAnyPendingOp ();
        hi := PseudStk [top]; low := PseudStk [top-1];
        adr := PseudStk [top-2];
        DecTopOfPsStk (3);
        slave (hi); slave (low);
        GoingUp := (next () = 0);
        immhi := immediate (hi); immlo := immediate (low);
        IF immhi AND immlo THEN
            IF (GoingUp AND (hi.DispOrData < low.DispOrData))
               OR (NOT GoingUp AND (hi.DispOrData > low.DispOrData)) THEN
	        GenLongBcc (T)
	    END (* if *);
        ELSE IF GoingUp THEN cc := GT ELSE cc := LT END;
            IF NOT immlo AND (immhi OR InDataReg (low)) THEN
	        cc := CheckedGenCmp (cc, hi, low);
	        GenLongBcc (cc)
            ELSIF immlo OR InDataReg (hi) THEN
	        cc := CheckedGenCmp (cc, low, hi);
	        GenLongBcc (reverse (cc))
            ELSE FreeReg (Reg, NOT UseBitSet);
                GenMove (low, Reg); low := Reg;
		GenCmp (hi, low);
		GenLongBcc (cc)
            END (* if *);
        END;
        IF DirectlyAddressable (adr) THEN
            INCL (ForType, AddressableControlVar);
            ControlVar := adr;
            ControlVar.Mode := AddrRegDisp;
            GenMove (low, ControlVar);
            SaveItem (ControlVar);
        ELSE IF InAddrRegSlave (adr, AddrReg) THEN
                MarkForFor (AddrReg)
            ELSE
                ProtectAddress (hi); (* hi may use a slaved address register *)
                AddrReg := FreeForFor (); (* so protect it while we get another *)
                UnprotectAddress (hi);
                IF address (adr) THEN
                    GenLea (adr, AddrReg)
                ELSE GenMovea (adr, AddrReg)
                END (* if *);
                (* not worth putting into slave because just about to enter loop *)
            END;
            PackItem (ControlVar, AddrRegDisp, AddrReg, 0, NoReg);
            GenMove (low, ControlVar);
            GenPea (ControlVar);
        END (* if *);
        IF immhi THEN
            SaveItem (hi);
            INCL (ForType, ImmediateHi);
        ELSE GenMove (hi, StackPreDec);
        END (* if *);
        PushOnto (ForStack, CARDINAL (ForType));
        ValOnStack (immhi);
        ValOnStack (AddressableControlVar IN ForType);
    END For1;
                           
      
PROCEDURE For2 ();
    VAR StepSize, StackDecAmount : INTEGER;
        cc : ConditionCode;
        SavedItem, ControlVar, hi : ITEM;
        AddrReg : REGISTER;
        ForType : FORTYPE;
        immhi : BOOLEAN;
    BEGIN
        StackDecAmount := 0;
        ForType := FORTYPE (PopFrom (ForStack));
        StepSize := Int16Bit (next2 ());
        IF StepSize > 0 THEN cc := GE ELSE cc := LE END;
        immhi := ImmediateHi IN ForType;
        IF immhi THEN
            RestoreItem (hi);
        ELSE
            PackRealStackItem (SavedItem, 0);
            INC (StackDecAmount);
            FreeReg (hi, NOT UseBitSet);
            GenMove (SavedItem, hi);
        END (* if *);
        IF AddressableControlVar IN ForType THEN
            RestoreItem (ControlVar);
        ELSE
            PackRealStackItem (SavedItem, -StackDecAmount);
            INC (StackDecAmount);
            IF AddrRegStillAvailable (AddrReg) THEN
                FreeAfterFor (AddrReg)
            ELSE 
                AddrReg := FreeAddrReg (NotOnPsStk);
                GenMovea (SavedItem, AddrReg);
            END (* if *);
            PackItem (ControlVar, AddrRegDisp, AddrReg, 0, NoReg);
        END (* if *);
        GenAddImm (StepSize, ControlVar);
        IF immhi THEN
            cc := CheckedGenCmp (cc, hi, ControlVar);
            GenLongBcc (reverse (cc));
        ELSE GenCmp (ControlVar, hi);
            GenLongBcc (cc);
        END (* if *);
        GenSubImm (StepSize, ControlVar);
        IF StackDecAmount > 0 THEN GenAddaImm (StackDecAmount*BytesPerWord, S) END;
        ValOffStack;
        ValOffStack;
        DEC (ForNesting);
    END For2;

PROCEDURE DoEXIT;
    VAR count, WordsToDec : CARDINAL;
        v : VALONSTACK;
        OldPos : ITEM;
BEGIN
    AssertEmptyPsStk;
    v := ValsOnStack;
    WordsToDec := 0;
    LOOP
        IF NOT v^.Saved THEN INC (WordsToDec) END;
        IF next () # DECS THEN EXIT END;
        v := v^.Next;
        INC (ExtraBytesProcessed);
    END (* while *);
    GenAddaImm (WordsToDec * BytesPerWord, S);
END DoEXIT;


BEGIN
    initialise (CaseStack);
    initialise (ForStack);
    ValsOnStack := NIL;
    ForNesting := 0;
END CaseFor.
