IMPLEMENTATION MODULE CodeProcs; (* $T- *)

FROM items IMPORT Int16Bit;
FROM Storage IMPORT ALLOCATE;
FROM SetUpCode IMPORT 
    code, FirstCodeByteNo, LastCodeByteNo, FixupByte, SetToNOP;
FROM SimpleStks IMPORT Stack, PushOnto, TopValueOf, empty,
		       RemoveTopItemFrom, initialise; 
FROM labels IMPORT InsertNewLabel, StepLabelPointers, 
		   MarkCurrentByte, CurrentLabel, 
		   CurrentExitLabel, FindExitLabel, 
		   InitialisePointers,
                   CurrentByteNo, CurrentInstrByteNo;
FROM UtilError IMPORT ReportError;
FROM MCMnemonics IMPORT
    LIn, LLWn, SLWn, LGWn, SGWn, SSWn, LSD0, LXFW, LXB, DSHR, SSD0, FNEG, READ,
    ESC, EXP, UGEQ, LODFW, DECS, UADD, SHR, EXC, PACK, GB1, ALOC, RTN, CF, CLn,
    LIB, LLA, LSA, JPFC, LLD, SLW, SLD, LGW, LGD, SGW, SGD, LSW, LSD, LSTA,
    SSW, SSD, FFCT, ENTP, TRA, PCOP, GB, ENTR, CI, CL, LIW, LEA, JP, LEW, LED,
    SEW, SED, ENTC, CX, FOR1, LID, FOR2, SYS, RDS, JPF, JPB;


TYPE CaseList = POINTER TO CaseItem;
TYPE CaseItem =
    RECORD
        ByteNumber: CARDINAL;
        Next:       CaseList;
    END (* record *);

VAR FirstCaseTable, CurrentCaseTable : CaseList;

PROCEDURE NewCI (bn : CARDINAL; cl : CaseList) : CaseList;
    VAR ci : CaseList;
    BEGIN
        NEW (ci); WITH ci^ DO ByteNumber := bn; Next := cl; END; RETURN ci;
    END NewCI;

PROCEDURE PutInCaseList (ByteNo : CARDINAL);
    VAR ci, chain : CaseList;
    BEGIN
        IF (FirstCaseTable = NIL) OR (ByteNo < FirstCaseTable^.ByteNumber) THEN
            ci := NewCI (ByteNo, FirstCaseTable);
            FirstCaseTable := ci;
        ELSE
            chain := FirstCaseTable;
            WHILE (chain^.Next # NIL) AND (chain^.Next^.ByteNumber < ByteNo) DO
                chain := chain^.Next;
            END (* while *);
            ci := NewCI (ByteNo, chain^.Next);
            chain^.Next := ci;
        END (* if *);
        CurrentCaseTable := ci;
    END PutInCaseList;


PROCEDURE MarkStartOfCaseLabelTable ();
    VAR label, ByteNo : CARDINAL;
    BEGIN
        ByteNo := CurrentInstrByteNo + next2 () + 1;
        label := InsertNewLabel (ByteNo);
        MarkCurrentByte (label);
        PutInCaseList (ByteNo);
    END MarkStartOfCaseLabelTable;

PROCEDURE LoadingFormalProcedure () : BOOLEAN;
    BEGIN
        IF FixupByte (CurrentInstrByteNo + 1) THEN
            RETURN TRUE;
        ELSE
            RETURN FALSE;
        END (* if *);
    END LoadingFormalProcedure;

PROCEDURE AtCaseLabelTable () : BOOLEAN; 
   BEGIN
      IF (CurrentCaseTable # NIL)
      AND (CurrentCaseTable^.ByteNumber = CurrentInstrByteNo)
      THEN CurrentCaseTable := CurrentCaseTable^.Next;
	   RETURN TRUE
      ELSE RETURN FALSE
      END
   END AtCaseLabelTable;


PROCEDURE WipeOut (NoOfBytes : CARDINAL);
    VAR byte : CARDINAL;
    BEGIN
        FOR byte := 0 TO NoOfBytes - 1 DO
            SetToNOP (CurrentInstrByteNo + byte);
        END (* for *);
    END WipeOut;


PROCEDURE StepToNextInstruction (NoOfBytes : CARDINAL);
   BEGIN
      StepLabelPointers (NoOfBytes);
      INC (CurrentInstrByteNo, NoOfBytes);
      CurrentByteNo := CurrentInstrByteNo;
      IF CurrentByteNo > LastCodeByteNo
      THEN NotAtEndOfCode := FALSE
      ELSIF CurrentByteNo >= FirstCodeByteNo
      THEN NotAtEndOfTable := FALSE
      END
   END StepToNextInstruction;


PROCEDURE next () : CARDINAL;
   BEGIN
      INC (CurrentByteNo);
      RETURN code (CurrentByteNo - 1)
   END next;


PROCEDURE next2 () : CARDINAL;
   BEGIN
      INC (CurrentByteNo, 2);
      RETURN code (CurrentByteNo - 2) * 400B +
		  code (CurrentByteNo - 1)
   END next2;

PROCEDURE next4 () : CARDINAL;
    VAR c: CARDINAL;
    BEGIN
        c := next2 ();
        RETURN (c * 200000B) + next2 ();
    END next4;

PROCEDURE ValueOfNextByte () : CARDINAL;
   BEGIN
      RETURN code (CurrentByteNo)
   END ValueOfNextByte;


PROCEDURE ValueOfNextTwoBytes () : CARDINAL;
   BEGIN
      RETURN code (CurrentByteNo) * 400B +
	     code (CurrentByteNo + 1)
   END ValueOfNextTwoBytes;


PROCEDURE NextInstruction (VAR length: CARDINAL): CARDINAL;
    VAR inst: CARDINAL;
BEGIN
    inst := next ();
    CASE inst OF
      LIn+0..LIn+15, LLWn+4..LLWn+15, SLWn+4..SLWn+15, LGWn+2..LGWn+15,
      SGWn+2..SSWn+15, LSD0..LXFW, LXB..DSHR, SSD0..FNEG, READ..ESC, EXP..UGEQ,
      LODFW..DECS, UADD..SHR, EXC..PACK, ALOC, RTN, CF, CLn+1..CLn+15:
        length := 1;
    | LIB, LLA..LSA, JPFC..LLD, SLW, SLD, LGW, LGD, SGW, SGD, LSW, LSD, LSTA,
      SSW, SSD, FFCT, ENTP, TRA, PCOP, GB1, CI, CL:
        length := 2;
    | LIW, LEA..JP, LEW, LED, SEW, SED, ENTC, CX, ENTR, GB:
        length := 3;
    | FOR1: length := 4;
    | LID, FOR2: length := 5;
    | SYS: length := ValueOfNextByte () + 2;
    | RDS: length := ValueOfNextByte () + 3;
    ELSE ReportError (TRUE, "Unknown code byte (%O3)", inst, 0);
    END;
    RETURN inst;
END NextInstruction;


PROCEDURE ignore (value : CARDINAL);
   BEGIN END ignore;


PROCEDURE ResetCodeAndLabelPtrs ();
   BEGIN
      InitialisePointers ();
      CurrentCaseTable := FirstCaseTable;
      NotAtEndOfTable := TRUE;
      NotAtEndOfCode := TRUE
   END ResetCodeAndLabelPtrs;


PROCEDURE LabelRelativeJump (offset : CARDINAL);
   VAR int : INTEGER; 
       ByteNo : CARDINAL;
   BEGIN
      int := Int16Bit (offset);
      IF int < 0
      THEN ByteNo := CurrentByteNo - CARDINAL (-int)
      ELSE ByteNo := CurrentByteNo + offset
      END;
      LabelAbsoluteJump (ByteNo);
   END LabelRelativeJump;


PROCEDURE LabelRelativeJumpBack (offset : CARDINAL);
   BEGIN
      LabelAbsoluteJump (CurrentByteNo - offset)
   END LabelRelativeJumpBack;


PROCEDURE LabelAbsoluteJump (ByteNo : CARDINAL);
    VAR label : CARDINAL;
        offset : INTEGER;

    BEGIN
        IF ByteNo # 0 THEN (* some address table entries are zero *)
            LOOP
                CASE code (ByteNo) OF
                    JP:
                        offset := Int16Bit (code (ByteNo+1) * 400B + 
                                                        code (ByteNo + 2)) + 1;
                        IF offset >= 0 THEN
                            INC (ByteNo, CARDINAL (offset));
                        ELSE
                            DEC (ByteNo, CARDINAL (-offset));
                        END (* if *);
                |   JPF:
                        INC (ByteNo, code (ByteNo + 1) + 1);
                |   JPB:
                        DEC (ByteNo, code (ByteNo + 1) - 1);
                ELSE
                    EXIT;
                END (* case *);
            END (* loop *);
        END (* if *);
        label := InsertNewLabel (ByteNo);
        MarkCurrentByte (label);
    END LabelAbsoluteJump;


PROCEDURE FindStartOfProc (TableNo : CARDINAL);
   BEGIN
      MarkCurrentByte (FindExitLabel (TableNo*2))
   END FindStartOfProc;


BEGIN
    FirstCaseTable := NIL;
    ResetCodeAndLabelPtrs ()
END CodeProcs.
