IMPLEMENTATION MODULE DeadWood;

FROM CodeProcs IMPORT
    NotAtEndOfTable, NotAtEndOfCode, StepToNextInstruction,
    NextInstruction, ValueOfNextByte, WipeOut,
    AtCaseLabelTable, next4;
FROM MCMnemonics IMPORT RTN, EXC, JP, JPF, JPB;

IMPORT SYSTEM;
IMPORT labels;

MODULE PermLabelTable;
(* $T-,$R- *)

FROM SYSTEM IMPORT MAXCARD;
FROM labels IMPORT
    CurrentLabel, CurrentExitLabel, GetPermanentLabelTable,
    DecCurrentLabelUseCount, none;

EXPORT GetTable, ProtectExitLab, DecExitLabUseCount, Labelled;

CONST
    LotsOfLabels = 1024;
    DontTouch = MAXCARD;

VAR Table : POINTER TO ARRAY [0..LotsOfLabels] OF CARDINAL;

PROCEDURE GetTable;
    VAR dummy : CARDINAL;
    BEGIN
        dummy := GetPermanentLabelTable (Table);
    END GetTable;

PROCEDURE ProtectExitLab;
    VAR lab : CARDINAL;
    BEGIN
        lab := CurrentExitLabel ();
        IF lab # none THEN
            Table^[lab] := DontTouch;
        END (* if *);
    END ProtectExitLab;

PROCEDURE DecExitLabUseCount;
    VAR lab : CARDINAL;
    BEGIN
        lab := CurrentExitLabel ();
        IF (lab # none) AND (Table^[lab] # DontTouch) THEN
            INC (Table^[lab]);
        END (* if *);
    END DecExitLabUseCount;

PROCEDURE Labelled () : BOOLEAN;
    VAR amount, lab : CARDINAL;
    BEGIN
        lab := CurrentLabel ();
        amount := Table^[lab]; (* note Table^[none] is always 0 *)
        IF (amount # 0) AND (amount # DontTouch) THEN
            lab := DecCurrentLabelUseCount (amount);
        END (* if *);
        RETURN (lab # none);
    END Labelled;

(* $T+,$R+ *)
END PermLabelTable;

TYPE ACTION = (Wipeout,LeaveAlone);

PROCEDURE EatCaseLabelTable (ToDo : ACTION);
    VAR low, hi : INTEGER;
    VAR entries, bytes : CARDINAL;
    BEGIN
        IF ToDo = LeaveAlone THEN
            ProtectExitLab;
        ELSE (* ToDo = Wipeout *)
            DecExitLabUseCount;
        END (* if *);
        low := INTEGER (next4 ());
        hi := INTEGER (next4 ());
        entries := CARDINAL (ABS (hi-low)) + 2; 
        (* +2 - one for 'telegraph pole' and one for ELSE exit *)
        bytes := 8 + (2 * entries); (* 8 for the two next4 ()s *)
        IF ToDo = Wipeout THEN WipeOut (bytes) END;
        StepToNextInstruction (bytes);
    END EatCaseLabelTable;


PROCEDURE UnconditionalTransferOfControl (inst : CARDINAL) : BOOLEAN;
    BEGIN
        RETURN (inst = RTN) OR (inst = EXC)
            OR (inst = JP) OR (inst = JPF)  OR (inst = JPB);
    END UnconditionalTransferOfControl;


PROCEDURE ProcessCode;
    VAR
        DeadWood : BOOLEAN;
        c : CARDINAL;
    BEGIN
        DeadWood := FALSE;
        WHILE NotAtEndOfCode DO
            IF NOT DeadWood THEN
                IF AtCaseLabelTable () THEN
                    EatCaseLabelTable (LeaveAlone);
                    DeadWood := TRUE;
                ELSE
                    IF UnconditionalTransferOfControl (ValueOfNextByte ()) THEN
                        DeadWood := TRUE;
                    END (* if *);
                    NextInst (LeaveAlone);
                END (* if *);
            ELSE
                IF Labelled () THEN
                    DeadWood := FALSE;
                ELSE
                    DecExitLabUseCount;
                    IF AtCaseLabelTable () THEN
                        EatCaseLabelTable (Wipeout);
                    ELSE
                        NextInst (Wipeout);
                    END (* if *);
                END (* if *);
            END (* if *);
        END (* while *);
    END ProcessCode;

PROCEDURE NextInst (ToDo : ACTION);
    VAR instr, bytes : CARDINAL;
    BEGIN
        instr := NextInstruction (bytes);
        IF ToDo = Wipeout THEN WipeOut (bytes) END;
        StepToNextInstruction (bytes);
    END NextInst;


PROCEDURE ProcessTable;
    BEGIN
        WHILE NotAtEndOfTable DO
            ProtectExitLab;
            StepToNextInstruction (2);
        END (* while *);
    END ProcessTable;

PROCEDURE DestroyDeadWood;
    BEGIN
        GetTable;
        ProcessTable;
        ProcessCode;
    END DestroyDeadWood;

END DeadWood.
