IMPLEMENTATION MODULE GenProcs;

FROM items IMPORT ITEM, REGISTER, MODE, Operator, GlobalLabel, PackReg,
		  ConditionCode, InDataReg, Max16BitInt, NILvalue,
                  PackItem, equal, immediate, address,
                  Max8BitInt, BytesPerWord;
FROM RegSlv IMPORT PurgeRegSlaves;
FROM Exceptions IMPORT StdException;
FROM Relations IMPORT reverse;
FROM labels IMPORT CurrentLabel, CurrentExitLabel;
FROM UtilError IMPORT
    ReportError;
FROM Object IMPORT
    LabelUseType, ExtrnType, ObjPtr, 
    BackwardBranch, UseOfLabel, label, Extrn,
    InstructionEmit, LongEmit, Emit,
    StartObjectModule, EndObjectModule,
    Signed16Bit, Signed8Bit, ShiftLeft,
    DefineRangeCheckLabel, RangeCheckBranchDisp,
    ExtrnNamedCall, ExtrnGlCall, ExtrnUNIX,
    Patch;

CONST
    ADD         = BITSET {15,14,12,7};
    ADDA        = BITSET {15,14,12,8,7,6};
    ADDAWORD    = BITSET {15,14,12,7,6};
    ADDI        = BITSET {10,9,7};
    ADDQ        = BITSET {14,12,7};
    ANDOP       = BITSET {15,14,7};
    ANDI        = BITSET {9,7};
    ASL         = BITSET {15,14,13,8,7};
    ASR         = BITSET {15,14,13,7};
    BCC         = BITSET {14,13};
    BSR         = BITSET {14,13,8};
    CLR         = BITSET {14,9,7};
    CMP         = BITSET {15,13,12,7};
    CMPA        = BITSET {15,13,12,8,7,6};
    CMPI        = BITSET {11,10,7};
    DBRA        = BITSET {14,12,8,7,6,3};
    EOR         = BITSET {15,13,12,8,7};
    EORI        = BITSET {11,9,7};
    EXG         = BITSET {15,14,8,6}; (* Data register EXG only *)
    JMP         = BITSET {14,11,10,9,7,6};
    JSR         = BITSET {14,11,10,9,7};
    LEA         = BITSET {14,8,7,6};
    LINK        = BITSET {14,11,10,9,6,4};
    LSL         = BITSET {15,14,13,8,7,3};
    LSR         = BITSET {15,14,13,7,3};
    MOVE        = BITSET {13};
    MOVEA       = BITSET {13,6};
    MOVEWORD    = BITSET {13,12};
    MOVEBYTE    = BITSET {12};
    MOVETOCCR   = BITSET {14,10,7,6};
    MOVEMWTOEA  = BITSET {14,11,7};
    MOVEMWTOREG = BITSET {14,11,10,7};
    MOVEQ       = BITSET {14,13,12};
    NEG         = BITSET {14,10,7};
    NOTOP       = BITSET {14,10,9,7};
    OROP        = BITSET {15,7};
    ORI         = BITSET {7};
    PEA         = BITSET {14,11,6};
    ROR         = BITSET {15,14,13,7,4,3};
    RTS         = BITSET {14,11,10,9,6,5,4,2,0};
    SUB         = BITSET {15,12,7};
    SUBA        = BITSET {15,12,8,7,6};
    SUBAWORD    = BITSET {15,12,7,6};
    SUBI        = BITSET {10,7};
    SUBQ        = BITSET {14,12,8,7};
    TRAP        = BITSET {14,11,10,9,6};
    TST         = BITSET {14,11,9,7};
    UNLK        = BITSET {14,11,10,9,6,4,3};

    VALImmediate        = BITSET {5,4,3,2};
    VALDataReg          = BITSET {};
    VALAddrRegDirect    = BITSET {3};
    VALAddrRegIndirect  = BITSET {4};
    VALAddrRegDisp      = BITSET {5,3};
    VALAddrRegIndex     = BITSET {5,4};
    VALPreDec           = BITSET {5};
    VALPostInc          = BITSET {4,3};
    VALAbsLongAddress   = BITSET {5,4,3,0};
    VALPCWithDisp       = BITSET {5,4,3,1};
    VALPCWithIndex      = BITSET {5,4,3,1,0};

    IndexExtWord        = BITSET {11};


TYPE
    SHIFTTYPE = (RegisterShift, ImmediateShift);

    ITEMTYPE = (LongWord, Word, Byte);

VAR
    Instruction, ExtWord, ExtraExtWord  : BITSET;
    ExtWordSet, ExtraExtWordSet, LongExtWord : BOOLEAN;
    LastLinkOffset : CARDINAL;
    LastLinkValue    : INTEGER;

PROCEDURE NormalEmitInstruction;
    BEGIN
        InstructionEmit (Instruction);
        IF ExtWordSet THEN
            ExtWordSet := FALSE;
            IF LongExtWord THEN
                LongExtWord := FALSE;
                LongEmit (ExtWord);
            ELSE
                Emit (ExtWord);
            END (* if *);
        END (* if *);
    END NormalEmitInstruction;

PROCEDURE DataRegVal (d : REGISTER) : BITSET;
    BEGIN
        RETURN BITSET (CARDINAL (d) - CARDINAL (D0));
    END DataRegVal;

PROCEDURE AddrRegVal (a : REGISTER) : BITSET;
    BEGIN
        RETURN BITSET (CARDINAL (a) - CARDINAL (A0));
    END AddrRegVal;

PROCEDURE RegVal (r : REGISTER) : BITSET;
    BEGIN
        IF r >= A0 THEN
            RETURN AddrRegVal (r);
        ELSE
            RETURN DataRegVal (r);
        END (* if *);
    END RegVal;

PROCEDURE NormalEA (VAR ea : ITEM; ItemType : ITEMTYPE);
    VAR Disp : INTEGER;
    BEGIN
        WITH ea DO
            IF (Mode=AddrRegDisp) OR (Mode=AddrRegIndex) THEN
                IF ItemType = LongWord THEN
                    Disp := DispOrData * BytesPerWord;
                ELSE
                    Disp := DispOrData;
                END (* if *);
            END (* if *);
            CASE Mode OF

                Immediate:
                    Instruction := Instruction + VALImmediate;
                    IF ItemType = LongWord THEN
                        LongExtWord := TRUE;
                        ExtWord := BITSET (DispOrData);
                    ELSIF DispOrData < 0 THEN
                        CASE ItemType OF
                            Word:
                                ExtWord := Signed16Bit (DispOrData);
                        |   Byte:
                                ExtWord := Signed8Bit (DispOrData);
                        END (* case *);
                    ELSE
                        ExtWord := BITSET (DispOrData);
                    END (* if *);
                    ExtWordSet := TRUE;

            |   DataReg:
                    Instruction := Instruction + VALDataReg + 
                                    DataRegVal (Register);

            |   AddrRegDisp:
                    IF DispOrData = 0 THEN
                        Instruction := Instruction + VALAddrRegIndirect +
                                        AddrRegVal (Register);
                    ELSE
                        Instruction := Instruction + VALAddrRegDisp +
                                        AddrRegVal (Register);
                        ExtWord := Signed16Bit (Disp);
                        ExtWordSet := TRUE;
                    END (* if *);

            |   AddrRegIndex:
                    (* assumes always use the form DISP(An,Dn.L) *)
                    Instruction := Instruction + VALAddrRegIndex +
                                    AddrRegVal (Register);
                    ExtWord := IndexExtWord + 
                                ShiftLeft (DataRegVal (Index), 12) +
                                Signed8Bit (Disp);
                    ExtWordSet := TRUE;

            |   PreDec:
                    Instruction := Instruction + VALPreDec + 
                                    AddrRegVal (Register);

            |   PostInc:
                    Instruction := Instruction + VALPostInc + 
                                    AddrRegVal (Register);

            ELSE
                ReportError 
         (TRUE, "Bad ITEM (mode %i, reg %i) given to NormalEA in GenProcs",
		    		ORD (Mode), ORD (Register));
            END (* case *);
        END (* with *);
    END NormalEA;

PROCEDURE NormalImmediateEA (code : BITSET; data : INTEGER; VAR ea : ITEM);
    BEGIN
        Instruction := code;
        NormalEA (ea, LongWord);
        InstructionEmit (Instruction);
        LongEmit (BITSET (data));
        IF ExtWordSet THEN
            ExtWordSet := FALSE;
            (* No chance of a LongEmit *)
            Emit (ExtWord);
        END (* if *);
    END NormalImmediateEA;

PROCEDURE NormalRegEA (code : BITSET; reg : REGISTER; VAR ea : ITEM);
    BEGIN
        Instruction := code + ShiftLeft (RegVal (reg), 9);
        NormalEA (ea, LongWord);
        NormalEmitInstruction;
    END NormalRegEA;

PROCEDURE NormalEAOnly (code : BITSET; VAR ea : ITEM);
    BEGIN
        Instruction := code;
        NormalEA (ea, LongWord);
        NormalEmitInstruction;
    END NormalEAOnly;

PROCEDURE EmitExtraExtWord;
    BEGIN
        IF ExtraExtWordSet THEN
            ExtraExtWordSet := FALSE;
            Emit (ExtraExtWord);
        END (* if *);
    END EmitExtraExtWord;

PROCEDURE DataRegInBits9To11 (d : REGISTER) : BITSET;
    BEGIN
        RETURN ShiftLeft (DataRegVal (d), 9);
    END DataRegInBits9To11;

PROCEDURE AddrRegInBits9To11 (a : REGISTER) : BITSET;
    BEGIN
        RETURN ShiftLeft (AddrRegVal (a), 9);
    END AddrRegInBits9To11;

PROCEDURE GenDataAndStringFrames ();
    BEGIN
        StartObjectModule;
    END GenDataAndStringFrames;

PROCEDURE GenEndOfCodeSequence ();
    BEGIN
        EndObjectModule;
    END GenEndOfCodeSequence;

PROCEDURE GenRangeCheckSequence;
    BEGIN
        DefineRangeCheckLabel;
	GenMoveQ (INTEGER (ORD (ArrayIndexOutOfRange)), D7);
        GenJsrToGlobalLabel (trap);
    END GenRangeCheckSequence;

PROCEDURE QuickSize (data : INTEGER) : BOOLEAN;
    BEGIN
        RETURN (data > 0) AND (data <9);
    END QuickSize;

PROCEDURE MoveQuickSize (data : INTEGER) : BOOLEAN;
    BEGIN
        RETURN (ABS (data) <= 127);
    END MoveQuickSize;


(*  **** MOVES                              ****  *)

PROCEDURE SetMoveDest (VAR dest : ITEM; ItemType : ITEMTYPE);
    VAR Disp : INTEGER;
    BEGIN
        WITH dest DO
            IF (Mode=AddrRegDisp) OR (Mode=AddrRegIndex) THEN
                IF ItemType = LongWord THEN
                    Disp := DispOrData * BytesPerWord;
                ELSE
                    Disp := DispOrData;
                END (* if *);
            END (* if *);
            CASE Mode OF

                DataReg:
                    Instruction := Instruction + ShiftLeft(VALDataReg, 3) + 
                                    DataRegInBits9To11 (Register);

            |   AddrRegDisp:
                    IF DispOrData = 0 THEN
                        Instruction := Instruction + 
                                        ShiftLeft (VALAddrRegIndirect, 3) +
                                        AddrRegInBits9To11 (Register);
                    ELSE
                        Instruction := Instruction + 
                                        ShiftLeft (VALAddrRegDisp, 3) +
                                        AddrRegInBits9To11 (Register);
                        ExtraExtWord := Signed16Bit (Disp);
                        ExtraExtWordSet := TRUE;
                    END (* if *);

            |   AddrRegIndex:
                    Instruction := Instruction + 
                                    ShiftLeft (VALAddrRegIndex, 3) +
                                    AddrRegInBits9To11 (Register);
                    ExtraExtWord := IndexExtWord + Signed8Bit (Disp) +
                                        ShiftLeft (DataRegVal (Index), 12);
                    ExtraExtWordSet := TRUE;

            |   PreDec:
                    Instruction := Instruction + ShiftLeft (VALPreDec, 3) +
                                    AddrRegInBits9To11 (Register);

            |   PostInc:
                    Instruction := Instruction + ShiftLeft (VALPostInc, 3) +
                                    AddrRegInBits9To11 (Register);

            ELSE
                ReportError 
                    (TRUE, "Bad ITEM given to SetMoveDest in GenProcs", 0, 0);
            END (* case *);
        END (* with *);
    END SetMoveDest;

PROCEDURE GenMove (VAR source, dest : ITEM);
    BEGIN
        IF NOT equal (source, dest) THEN
            IF source.Mode = FormalProcedure THEN
                StoreFormalProc (source, dest);
            ELSIF InDataReg (dest) THEN
                GenMoveToReg (source, dest.Register);
            ELSE
                IF immediate (source) AND (source.DispOrData = NILvalue) THEN
                    GenMoveReg (NILRegister, dest); 
                ELSIF immediate (source) AND (source.DispOrData = 0) THEN
                    GenMoveReg (ZeroRegister, dest);
                ELSE
                    Instruction := MOVE;
                    NormalEA (source, LongWord);
                    SetMoveDest (dest, LongWord);
                    NormalEmitInstruction;
                    EmitExtraExtWord;
                END (* if *);
            END; (* if *)
        END; (* if *)
    END GenMove;

PROCEDURE StoreFormalProc (VAR proc, dest : ITEM);
    VAR
        val, ModuleOffset, ProcOffset : CARDINAL;
        save : INTEGER;
    BEGIN
        val := proc.DispOrData;
(*        save := dest.DispOrData;
        dest.DispOrData := save * BytesPerWord;   *)
        ModuleOffset := val DIV 400B;
        ProcOffset := (val MOD 400B); (* * BytesPerWord; *)
        Instruction := MOVE + VALImmediate;
        SetMoveDest (dest, LongWord);    
        InstructionEmit (Instruction);
(*        Extrn (OtherModuleOffset, ModuleOffset, 0);  *)
(* FORMAL PROC CHANGE *)
	Extrn (OtherModuleProc, ModuleOffset, ProcOffset);
        LongEmit (BITSET {});
        EmitExtraExtWord;
(*        INC (dest.DispOrData, 2);
        Instruction := MOVEWORD + VALImmediate;
        SetMoveDest (dest, Word);
        InstructionEmit (Instruction);
        Emit (BITSET (ProcOffset));
        EmitExtraExtWord;
        dest.DispOrData := save;  *)
    END StoreFormalProc;

PROCEDURE GenMoveImm (data : INTEGER; VAR dest : ITEM);
    VAR ImmItem : ITEM;
    BEGIN
        PackItem (ImmItem, Immediate, NoReg, data, NoReg);
        GenMove (ImmItem, dest);
    END GenMoveImm;

PROCEDURE GenMoveReg (reg : REGISTER; VAR dest : ITEM);
    VAR RegItem : ITEM;
    BEGIN
        IF reg >= A0 THEN
            Instruction := MOVE + VALAddrRegDirect + AddrRegVal (reg);
        ELSE
            Instruction := MOVE + VALDataReg + DataRegVal (reg);
        END (* if *);
        SetMoveDest (dest, LongWord);
        InstructionEmit (Instruction);
        EmitExtraExtWord;        
    END GenMoveReg;

PROCEDURE GenMoveToReg (VAR item : ITEM; reg : REGISTER);
    BEGIN
        IF NOT (InDataReg (item) AND (item.Register = reg)) THEN
            IF immediate (item) AND MoveQuickSize (item.DispOrData) THEN
                GenMoveQ (item.DispOrData, reg);
            ELSE
                Instruction := MOVE + DataRegInBits9To11 (reg) + 
                                ShiftLeft (VALDataReg, 3);
                NormalEA (item, LongWord);
                NormalEmitInstruction;
            END (* if *);
        END (* if *);
    END GenMoveToReg;

PROCEDURE GenMovea (VAR item : ITEM; AddrReg : REGISTER);
    BEGIN
        NormalRegEA (MOVEA, AddrReg, item);
    END GenMovea;

PROCEDURE GenMoveaReg (reg, AddrReg : REGISTER);
    BEGIN
        IF reg >= A0 THEN
            InstructionEmit (MOVEA + AddrRegInBits9To11 (AddrReg) +
                                VALAddrRegDirect + AddrRegVal (reg));
        ELSE
            InstructionEmit (MOVEA + AddrRegInBits9To11 (AddrReg) +
                                VALDataReg + DataRegVal (reg));
        END (* if *);
    END GenMoveaReg;

PROCEDURE GenLeaExternal (offset : CARDINAL; AddrReg : REGISTER);
    BEGIN
        InstructionEmit 
            (LEA + AddrRegInBits9To11 (AddrReg) + VALAbsLongAddress);
        Extrn (OtherModuleBase, offset, 0);
        LongEmit (BITSET {});
    END GenLeaExternal;

PROCEDURE GenMoveByte (VAR opd1, opd2 : ITEM);
    BEGIN
        IF NOT equal (opd1, opd2) THEN
            Instruction := MOVEBYTE;
            NormalEA (opd1, Byte);
            SetMoveDest (opd2, Byte);
            NormalEmitInstruction;
            EmitExtraExtWord;
        END (* if *);
    END GenMoveByte;

PROCEDURE GenMoveQ (data : INTEGER; dest : REGISTER);
    BEGIN
        InstructionEmit (MOVEQ + DataRegInBits9To11 (dest) + 
                            BITSET (Signed8Bit (data)));
    END GenMoveQ;

PROCEDURE GenExg (VAR opd1, opd2 : ITEM);
    BEGIN
        InstructionEmit (EXG + DataRegInBits9To11 (opd1.Register) +
                            DataRegVal (opd2.Register));
    END GenExg;

PROCEDURE GenMoveToCCR (value : CARDINAL);
    BEGIN
        InstructionEmit (MOVETOCCR + VALImmediate);
        Emit (BITSET (value));
    END GenMoveToCCR;

(*   **** ADDRESS OPERATIONS                       ****   *)

PROCEDURE GenPea (VAR item : ITEM);
    VAR addr: BOOLEAN;
    BEGIN
        addr := address (item);
        IF addr THEN item.Mode := AddrRegDisp END;
        NormalEAOnly (PEA, item);
        IF addr THEN item.Mode := ADDRaddrRegDisp END;
    END GenPea;


PROCEDURE GenLea (VAR item : ITEM; AddrReg : REGISTER);
    VAR addr: BOOLEAN;
    BEGIN
        addr := address (item);
        IF addr THEN item.Mode := AddrRegDisp END;
        NormalRegEA (LEA, AddrReg, item);
        IF addr THEN item.Mode := ADDRaddrRegDisp END;
    END GenLea;

PROCEDURE GenLeaGlobalLabel (lab : GlobalLabel; AddrReg : REGISTER);
    BEGIN
ReportError (FALSE, "Would have generated Lea Global", 0, 0);
(*
        InstructionEmit (LEA + ShiftLeft (AddrRegVal (AddrReg), 9)
                            + VALAbsLongAddress);
        Extrn (GlobLab, lab, 0);
        LongEmit (BITSET {});
*)
    END GenLeaGlobalLabel;

(*   **** COMPARISONS                          ****   *)

PROCEDURE CheckedGenCmp (cc : ConditionCode; VAR opd1, opd2 : ITEM) : ConditionCode;
    BEGIN
        IF immediate (opd1) THEN
            IF (opd1.DispOrData = 0) THEN
                GenCmpAddress (opd2, ZeroRegister);
                RETURN reverse (cc);
            ELSIF (opd1.DispOrData = NILvalue) THEN
                GenCmpAddress (opd2, NILRegister);
                RETURN reverse (cc);
            ELSE
                GenCmp (opd1, opd2);
                RETURN cc;
            END (* if *);
        ELSE
            GenCmp (opd1, opd2);
            RETURN cc;
        END (* if *);
    END CheckedGenCmp;

PROCEDURE GenCmp (VAR opd1, opd2 : ITEM);
    BEGIN
        IF immediate (opd1) THEN
            GenCmpImm (opd1.DispOrData, opd2)
        ELSE
            NormalRegEA (CMP, opd2.Register, opd1);
        END; (* if *)
    END GenCmp;

PROCEDURE GenCmpAddress (VAR opd : ITEM; AddrReg : REGISTER);
    BEGIN
        NormalRegEA (CMPA, AddrReg, opd);
    END GenCmpAddress;

PROCEDURE GenCmpImm (data : INTEGER; VAR item : ITEM);
    BEGIN
        NormalImmediateEA (CMPI, data, item);
    END GenCmpImm;

(*   **** DYADIC OPERATORS                     ****   *)

PROCEDURE GenSimpleDyadicOp (op : Operator; VAR source, dest : ITEM);
    CONST ToEABit = BITSET {8};
    BEGIN
        IF immediate (source) THEN
            CASE op OF

              Uadd, Add     :   GenAddImm (source.DispOrData, dest)   |
              Usub, Sub     :   GenSubImm (source.DispOrData, dest)   |
              And           :   GenAndImm (source.DispOrData, dest)   |
              Or            :   GenOrImm  (source.DispOrData, dest)   |
              Xor           :   GenXorImm (source.DispOrData, dest)

            END (* case *);
        ELSE
            CASE op OF

                Uadd, Add   :   Instruction := ADD;
            |   Usub, Sub   :   Instruction := SUB;
            |   And         :   Instruction := ANDOP;
            |   Or          :   Instruction := OROP;
            |   Xor         :   Instruction := EOR;

            END (* case *);
	    (*  Fix because EOR ain't like the rest *)
            IF InDataReg (dest) AND (Instruction # EOR) THEN
                Instruction := Instruction + DataRegInBits9To11(dest.Register);
                NormalEA (source, LongWord);
            ELSE IF NOT InDataReg (source) THEN
	             ReportError (TRUE, "Error in GenProc", 0, 0);
		END;
                Instruction := Instruction + ToEABit +
                                DataRegInBits9To11 (source.Register);
                NormalEA (dest, LongWord);
            END (* if *);
            NormalEmitInstruction;
        END (* if *);
    END GenSimpleDyadicOp;

PROCEDURE GenAndImm (imm : INTEGER; VAR item : ITEM);
    BEGIN
        NormalImmediateEA (ANDI, imm, item);
    END GenAndImm;

PROCEDURE GenOrImm (data : INTEGER; VAR item : ITEM);
    BEGIN
        NormalImmediateEA (ORI, data, item);
    END GenOrImm;

PROCEDURE GenXorImm (data : INTEGER; VAR item : ITEM);
    BEGIN
        NormalImmediateEA (EORI, data, item);
    END GenXorImm;

PROCEDURE qOp (code : BITSET; data : CARDINAL; VAR ea : ITEM);
    VAR qdata : CARDINAL;
    BEGIN
        IF data = 8 THEN
            qdata := 0;
        ELSE
            qdata := data;
        END (* if *);
        Instruction := code + ShiftLeft (BITSET (qdata), 9);
        NormalEA (ea, LongWord);
        NormalEmitInstruction;
    END qOp;

CONST NastyNeg = 80000000H;

PROCEDURE GenSubImm (imm : INTEGER; VAR item : ITEM);
    BEGIN
        IF (imm < 0) AND (imm # INTEGER (NastyNeg)) THEN
            GenAddImm (-imm, item);
        ELSIF QuickSize (imm) THEN
            qOp (SUBQ, imm, item);
        ELSE
            NormalImmediateEA (SUBI, imm, item);
        END (* if *);
    END GenSubImm;

PROCEDURE GenAddImm (imm : INTEGER; VAR item : ITEM);
    BEGIN
        IF (imm < 0) AND (imm # INTEGER (NastyNeg)) THEN
            GenSubImm (-imm, item);
        ELSIF QuickSize (imm) THEN
            qOp (ADDQ, imm, item);
        ELSE
            NormalImmediateEA (ADDI, imm, item);
        END (* if *);
    END GenAddImm;

PROCEDURE GenShift (op : Operator; VAR opd1, opd2 : ITEM);
    BEGIN
        IF immediate (opd1) THEN
            GenShiftImm (op, opd1.DispOrData, opd2);
        ELSE
            (* opd1 must be data register *)
            Shifts (op, opd1.Register, 0, opd2, RegisterShift);
        END (* if *);
    END GenShift;


PROCEDURE GenShiftImm (op : Operator; imm : INTEGER;
                                 VAR item : ITEM);
    VAR data : CARDINAL;
    BEGIN
        IF imm = 8 THEN
            data := 0
        ELSE
            data := imm
        END (* if *);
        Shifts (op, NoReg, data, item, ImmediateShift);
    END GenShiftImm;

PROCEDURE Shifts (op : Operator; reg1 : REGISTER; imm : CARDINAL;
                    VAR reg2 : ITEM; TypeOfShift : SHIFTTYPE);
    CONST RegShiftBit = BITSET {5};
    BEGIN
        CASE op OF
            Shl  : Instruction := LSL;
        |   Shr  : Instruction := LSR;
        |   Ashl : Instruction := ASL;
        |   Ashr : Instruction := ASR;
        |   Ror  : Instruction := ROR;
        END (* case *);
        Instruction := Instruction + DataRegVal (reg2.Register);
        IF TypeOfShift = RegisterShift THEN
            Instruction := Instruction + DataRegInBits9To11 (reg1)
                            + RegShiftBit;
        ELSE
	    IF imm > 8 THEN
		ReportError (TRUE, "Tried to generate shift of %i", imm, 0);
	    END;
            Instruction := Instruction + ShiftLeft (BITSET (imm), 9);
        END (* if *);
        InstructionEmit (Instruction);
    END Shifts;

PROCEDURE SubaAddaImm (code : BITSET; data : INTEGER; AddrReg : REGISTER; 
                        long : BOOLEAN);
    BEGIN
        IF data # 0 THEN
            Instruction := code + AddrRegInBits9To11 (AddrReg) + VALImmediate;
            ExtWordSet := TRUE;
            IF long THEN
                LongExtWord := TRUE;
                ExtWord := BITSET (data);
            ELSE
                ExtWord := Signed16Bit (data);
            END (* if *);
            NormalEmitInstruction;
        END (* if *)
    END SubaAddaImm;

PROCEDURE GenSuba (VAR item : ITEM; AddrReg : REGISTER);
    BEGIN
        IF immediate (item) THEN
            GenSubaImm (item.DispOrData, AddrReg);
        ELSE
            NormalRegEA (SUBA, AddrReg, item);
        END (* if *);
    END GenSuba;

PROCEDURE GenSubaImm (imm : INTEGER; AddrReg : REGISTER);
    BEGIN
        IF ABS (imm) < Max16BitInt THEN
            SubaAddaImm (SUBAWORD, imm, AddrReg, FALSE);
        ELSE
            SubaAddaImm (SUBA, imm, AddrReg, TRUE);
        END;
    END GenSubaImm;

PROCEDURE GenAddaImm (imm : INTEGER; AddrReg : REGISTER);
    BEGIN
        IF ABS (imm) < Max16BitInt THEN
            SubaAddaImm (ADDAWORD, imm, AddrReg, FALSE);
        ELSE
            SubaAddaImm (ADDA, imm, AddrReg, TRUE);
        END (* if *);
    END GenAddaImm;

(*   **** JUMPS AND LABELS                     ****   *)

PROCEDURE GenLabel ();
    CONST AfterLabel = FALSE;
    VAR lab : CARDINAL;
    BEGIN
        lab := CurrentLabel ();
        IF lab <> 0 THEN
            PurgeRegSlaves (AfterLabel);
            label (lab);
        END; (* if *)
    END GenLabel;

PROCEDURE GenTableEntry ();
    BEGIN
        InstructionEmit (BITSET {});
        Emit (BITSET {});
        UseOfLabel (CurrentExitLabel (), ObjPtr-4, Abs32Bit);
    END GenTableEntry;

PROCEDURE branch (cc : ConditionCode) : BITSET;
    BEGIN
        CASE cc OF
            T  : RETURN BITSET {};
        |   EQ : RETURN BITSET {10,9,8};
        |   NE : RETURN BITSET {10,9};
        |   LE : RETURN BITSET {11,10,9,8};
        |   GT : RETURN BITSET {11,10,9};
        |   LT : RETURN BITSET {11,10,8};
        |   GE : RETURN BITSET {11,10};
        |   LS : RETURN BITSET {9,8};
        |   HI : RETURN BITSET {9};
        |   CS : RETURN BITSET {10,8};
        |   CC : RETURN BITSET {10};
        |   MI : RETURN BITSET {11,9,8};
        |   PL : RETURN BITSET {11,9};
        |   VC : RETURN BITSET {11};
        |   VS : RETURN BITSET {11,8};
        END (* case *);
    END branch;

PROCEDURE GenBcc (cc : ConditionCode);
    VAR lab : CARDINAL;
    VAR short : BOOLEAN;
    BEGIN
        IF cc # F THEN
            InstructionEmit (BCC + branch (cc));
            lab := CurrentExitLabel ();
            IF BackwardBranch (lab, short) THEN
                IF NOT short THEN
                    Emit (BITSET {});
                    UseOfLabel (lab, ObjPtr-2, Disp16BitInBranch);
                END (* if *);
            ELSE
                Emit (BITSET {});
                UseOfLabel (lab, ObjPtr-2, DispInBranch);
            END (* if *);
        END (* if *);
    END GenBcc;

PROCEDURE GenBra ();
   BEGIN GenBcc (T) END GenBra;

PROCEDURE GenLongBcc (cc : ConditionCode);
    VAR lab : CARDINAL;
    VAR short : BOOLEAN;
    BEGIN
        IF cc # F THEN
            InstructionEmit (BCC + branch (cc));
            lab := CurrentExitLabel ();
            IF BackwardBranch (lab, short) THEN
                IF NOT short THEN
                    Emit (BITSET {});
                    UseOfLabel (lab, ObjPtr-2, Disp16BitInBranch);
                END (* if *);
            ELSE
                Emit (BITSET {});
                UseOfLabel (lab, ObjPtr-2, Disp16BitInBranch);
            END (* if *);
        END (* if *);
    END GenLongBcc;

PROCEDURE GenBsrToRangeCheck;
    BEGIN
        InstructionEmit (BSR);
        RangeCheckBranchDisp;
    END GenBsrToRangeCheck;

PROCEDURE GenBsr ();
   BEGIN
        InstructionEmit (BSR);
        Emit (BITSET {});
        UseOfLabel (CurrentExitLabel (), ObjPtr-2, QuickBSR);
   END GenBsr;

PROCEDURE GenJsr (VAR item : ITEM);
   BEGIN
        NormalEAOnly (JSR, item);
   END GenJsr;

PROCEDURE GenJsrToExternalProc (ModuleDisp, ProcNumber : CARDINAL);
    BEGIN
        InstructionEmit (JSR + VALAbsLongAddress);
        Extrn (OtherModuleProc, ModuleDisp, ProcNumber);
        LongEmit (BITSET {});
    END GenJsrToExternalProc;

PROCEDURE GenRelBranch (cc : ConditionCode;
                         lab : CARDINAL);
    BEGIN
        IF cc # F THEN
            InstructionEmit (BCC + branch (cc));
            UseOfLabel (lab, ObjPtr, Disp8BitInBranch);
        END (* if *);
    END GenRelBranch;

PROCEDURE GenDBra (reg : REGISTER; lab : CARDINAL);
    BEGIN
        InstructionEmit (DBRA + DataRegVal (reg));
        Emit (BITSET {});
        UseOfLabel (lab, ObjPtr-2, Disp16BitInBranch);
    END GenDBra;

PROCEDURE GenIndexedMovea (VAR reg : ITEM; AddressReg : REGISTER);
    BEGIN
        InstructionEmit 
            (MOVEA + AddrRegInBits9To11 (AddressReg) + VALPCWithIndex);
        Emit (IndexExtWord + ShiftLeft (DataRegVal (reg.Register), 12));
        UseOfLabel (CurrentExitLabel (), ObjPtr-2, Disp8BitInPCWithIndex);
    END GenIndexedMovea;

PROCEDURE GenPushLabel ();
    BEGIN
        InstructionEmit (PEA + VALPCWithDisp);
        Emit (BITSET {});
        UseOfLabel (CurrentExitLabel (), ObjPtr-2, Disp16BitInBranch);
    END GenPushLabel;

PROCEDURE GenJump (VAR dest : ITEM);
    BEGIN
        NormalEAOnly (JMP, dest);
    END GenJump;

(*   **** SINGLE OPERAND OPERATIONS                ****   *)

PROCEDURE GenClr (VAR item : ITEM);
    BEGIN
        GenMoveReg (ZeroRegister, item); (* quicker than CLR.L *)
    END GenClr;

PROCEDURE GenTst (VAR item : ITEM);
    BEGIN
        NormalEAOnly (TST, item);
    END GenTst;

PROCEDURE GenNeg (VAR item : ITEM);
    BEGIN
        NormalEAOnly (NEG, item);
    END GenNeg;

PROCEDURE GenCom (VAR item : ITEM);
    BEGIN
        NormalEAOnly (NOTOP, item);
    END GenCom;

(*   **** OPERATIONS USED WITH PROCEDURES              ****   *)

PROCEDURE GenRtn ();
    BEGIN
        InstructionEmit (RTS);
    END GenRtn;

PROCEDURE SetFrameSize(Hwm : CARDINAL);
    BEGIN
	Patch (LastLinkOffset, Signed16Bit (-INTEGER (Hwm * BytesPerWord) +
					LastLinkValue));
    END SetFrameSize;

PROCEDURE GenLink (AddrReg : REGISTER; disp : INTEGER);
    BEGIN
        InstructionEmit (LINK + AddrRegVal (AddrReg));
	LastLinkOffset := ObjPtr;
	LastLinkValue := disp;
        Emit (Signed16Bit (disp));
    END GenLink;

PROCEDURE GenUnlink (AddrReg : REGISTER);
    BEGIN
        InstructionEmit (UNLK + AddrRegVal (AddrReg));
    END GenUnlink;

(*   **** STANDARD SUBROUTINE CALL AND SYSCALL            ****   *)

PROCEDURE GenJsrToGlobalLabel (sub : GlobalLabel);
    BEGIN
        InstructionEmit (JSR + VALAbsLongAddress);
	ExtrnGlCall (sub);
        LongEmit (BITSET {});
    END GenJsrToGlobalLabel;

PROCEDURE GenJmpToGlobalLabel (sub : GlobalLabel);
    BEGIN
        (*  Just like a call as far as the external info *)
        InstructionEmit (JMP + VALAbsLongAddress);
        ExtrnGlCall (sub);
        LongEmit (BITSET {});
    END GenJmpToGlobalLabel;

PROCEDURE GenNamedCall (CallName : ARRAY OF CHAR);
    BEGIN
        InstructionEmit (JSR + VALAbsLongAddress);
	ExtrnNamedCall (CallName);
        LongEmit (BITSET {});
    END GenNamedCall;

PROCEDURE GenUnixCall (CallNo : CARDINAL);
    BEGIN
        InstructionEmit (JSR + VALAbsLongAddress);
	ExtrnUNIX (CallNo);
        LongEmit (BITSET {});
    END GenUnixCall;

(*   **** COMMENT                          ****   *)

PROCEDURE GenComment (str : ARRAY OF CHAR);
    BEGIN
    END GenComment;

PROCEDURE GenMoveMultipleWords (VAR source : ITEM; LWBreg, UPBreg : REGISTER);
    VAR 
        RegBit : CARDINAL;
        PreDecMode : BOOLEAN;
        r : REGISTER;
    BEGIN
        Instruction := MOVEMWTOREG;
        NormalEA (source, LongWord);
        PreDecMode := (source.Mode = PreDec);
        ExtraExtWord := BITSET {};
        FOR r := LWBreg TO UPBreg DO
            RegBit := CARDINAL (r) - CARDINAL (D0);
            IF PreDecMode THEN
                RegBit := 15 - RegBit;
            END (* if *);
            INCL (ExtraExtWord, RegBit);
        END (* for *);
        InstructionEmit (Instruction);
        Emit (ExtraExtWord);
        IF ExtWordSet THEN
            ExtWordSet := FALSE;
            Emit (ExtWord);
        END (* if *);
    END GenMoveMultipleWords;

BEGIN
    ExtWordSet := FALSE;
    ExtraExtWordSet := FALSE;
    LongExtWord := FALSE;
END GenProcs.
