IMPLEMENTATION MODULE Jumps;

FROM items IMPORT ITEM, MODE, REGISTER, ConditionCode, PackRealStackItem,
                  PackItem, DataReg0, immediate;
FROM PsStk IMPORT top, PseudStk, DoBooleanOp, push,
                  BooleanOpPending, DoAnyPendingOp, DecTopOfPsStk,
                  PutPsStkOnRealStack, DyadicOpPending;
FROM GenProcs IMPORT GenTst, GenBcc, GenRelBranch, GenClr,
                     GenMoveImm, GenMoveQ, GenMoveToCCR,
                     GenLabel, GenLongBcc;
FROM SimpleStks IMPORT Stack, PushOnto, empty, TopValueOf,
                       RemoveTopItemFrom, initialise;
FROM CodeProcs IMPORT ValueOfNextByte, StepToNextInstruction,
                      next, ignore;
FROM RegSlv IMPORT FreeReg, UseBitSet, RegCanBeFreedSimply;
FROM Relations IMPORT opposite;
FROM GenOp IMPORT SetResultItem;
FROM LoadStore IMPORT LocalOffset;
FROM labels IMPORT NewLabel, PushLabelStack, CurrentLabel, CurrentExitLabel;
FROM MCMnemonics IMPORT
    NOTT, RTN, JPC, JP, JPFC, JPF, JPBC, JPB, ORJP, ANDJP, SLW, SGW, SLWn, SGWn;

CONST TrueValue = 0; FalseValue = 4;

VAR DestOfAndJpOrJp : Stack;
    PreparedCondCode : BOOLEAN;
    CondCode : ConditionCode;
    label1, label2 : CARDINAL;


(* This procedure is called when a conditional jump is to
   be made. In the LILITH machine, when a conditional jump
   is about to be made, the top of the expression stack is
   a boolean value and this can be tested to see if it is
   TRUE or FALSE. 
   On the pseudo stack the situation is not quite so simple
   because there may be a pending operation on the stack.
   Also the 68000 branches according to condition codes so
   that it is possible to do 'branch if x < y' by :
         CMP.L y,x
	 BLS label
   whereas the LILITH can only do 'branch if TRUE' or
   'branch if FALSE'. This procedure operates on the
   top one or two items of the pseudo stack so that a
   condition code is set in the status register of the
   68000. It then returns the condition, say 'XX' such
   that 'BXX label' has the same effect as 'branch if
   TRUE to label' on the LILITH. e.g. consider the
   instruction sequence:   
   Starting with x, y on the expression stack.
   The LILITH would do -
   pop x, pop y, 
   IF x < y THEN push TRUE ELSE push FALSE, 
   branch to label if value on top of stack is TRUE 
   What the procedure Cond does if it starts with x and
   y on the pseudo stack is to generate CMP.L  y,x and
   return the condition 'LS' so that
     BLS label
  has the same effect as the LILITH branch if TRUE.
  Note that in fact most jumps on the LILITH are done
  if the top stack value is FALSE so instead of 'BLS'
  'BGE' is the branch needed but I prefer Cond this
  way even if you dont.
*)
PROCEDURE Cond () : ConditionCode;
    VAR cc : ConditionCode;
        StackTop : ITEM;
    BEGIN
        IF BooleanOpPending () THEN
            cc := DoBooleanOp ();
        ELSIF PreparedCondCode THEN
            PreparedCondCode := FALSE;
            cc := CondCode;
        ELSIF RegJustLoadedOnStack THEN
            DecTopOfPsStk (1); cc := NE
        ELSE 
            DoAnyPendingOp ();
            StackTop := PseudStk [top];
            IF immediate (StackTop) THEN
                IF StackTop.DispOrData = 0 THEN
                    cc := F;
                ELSE
                    cc := T;
                END (* if *);
            ELSE
                GenTst (StackTop);
                cc := NE;
            END (* if *);
            DecTopOfPsStk (1);
        END (* if *);
        RETURN cc
    END Cond;


PROCEDURE AndJump ();
    VAR cc : ConditionCode;
    BEGIN
        IF DyadicOpPending THEN
            IF top > 2 THEN
                PutPsStkOnRealStack (top-2)
            END (* if *);
        ELSIF (top > 1) THEN
            PutPsStkOnRealStack (top-1)
        END (* if *);
        PushOntoLabelStack (CurrentExitLabel ());
        cc := Cond (); GenLabel ();
        (* note that a call of Cond may couse code
          to be written - the order of Cond and label
          is therefore important *)
        IF cc = NE THEN
            GenBcc (EQ)
        ELSE
            label1 := NewLabel ();
            GenRelBranch (cc, label1);
            (* 2 instructions, 16 bit immediate value + 16 bit displacement *)
            GenMoveToCCR (FalseValue);
            GenBcc (T);
            PushLabelStack (label1);
        END (* if *);
    END AndJump;


PROCEDURE OrJump ();
    VAR cc : ConditionCode;
    BEGIN
        IF DyadicOpPending THEN
            IF top > 2 THEN
                PutPsStkOnRealStack (top-2)
            END (* if *);
        ELSIF (top > 1) THEN
            PutPsStkOnRealStack (top-1)
        END (* if *);
        PushOntoLabelStack (CurrentExitLabel ());
        cc := Cond (); GenLabel ();
        IF cc = NE THEN
            GenBcc (NE)
        ELSE
            label1 := NewLabel ();
            GenRelBranch (opposite (cc), label1);
            GenMoveToCCR (TrueValue);
            GenBcc (T);
            PushLabelStack (label1);
        END (* if *);
   END OrJump;


PROCEDURE PushOntoLabelStack (lab : CARDINAL);
   BEGIN
      IF NOT empty (DestOfAndJpOrJp)
      AND (TopValueOf (DestOfAndJpOrJp) = lab) 
      THEN RETURN 
      ELSE PushOnto (DestOfAndJpOrJp, lab)
      END;
   END PushOntoLabelStack;


PROCEDURE CondJump (long : BOOLEAN);
   VAR cc : ConditionCode;
   BEGIN
      cc := opposite (Cond ()); (*note that this has side effects*)
      GenLabel ();
      IF long THEN
          GenLongBcc (cc);
      ELSE
          GenBcc (cc);
      END (* if *);
   END CondJump;


PROCEDURE AtDestOfOrJpOrAndJp () : BOOLEAN;
   BEGIN
      RETURN 
        NOT empty (DestOfAndJpOrJp)
        AND (TopValueOf (DestOfAndJpOrJp) = CurrentLabel ())
   END AtDestOfOrJpOrAndJp;


PROCEDURE DealWithBeingThere ();
   VAR cc : ConditionCode;
       NextInstruc : CARDINAL;
       dest : ITEM;
       HitAnotherOne : BOOLEAN;
    BEGIN
        RemoveTopItemFrom (DestOfAndJpOrJp);
        cc := Cond ();
        IF cc <> NE THEN
            label1 := NewLabel (); label2 := NewLabel ();
            GenRelBranch (cc, label1);
            GenMoveToCCR (FalseValue);
            GenRelBranch (T, label2);
            PushLabelStack (label1);
            GenMoveToCCR (TrueValue);
            PushLabelStack (label2);
        END (* if *);
        GenLabel ();
        CondCode := NE;
        HitAnotherOne := FALSE;
        NextInstruc := ValueOfNextByte ();
        WHILE (NextInstruc = NOTT (* NOT *)) AND NOT HitAnotherOne DO
            CondCode := opposite (CondCode);
            StepToNextInstruction (1);
            HitAnotherOne := AtDestOfOrJpOrAndJp ();
            NextInstruc := ValueOfNextByte ();
        END (* while *);
        IF HitAnotherOne OR IsCondJump (NextInstruc) THEN
            PreparedCondCode := TRUE;
        ELSIF IsSimpleStore (NextInstruc) THEN
            GetDestAddress (NextInstruc, dest);
            label1 := NewLabel ();
            GenRelBranch (CondCode, label1);
            GenClr (dest);
            label2 := NewLabel ();
            GenRelBranch (T, label2);
            PushLabelStack (label1);
            GenMoveImm (1, dest);
            PushLabelStack (label2);
        ELSIF NextInstruc = RTN (* RTN - return *) THEN
            label1 := NewLabel (); label2 := NewLabel ();
            GenRelBranch (CondCode, label1);
            GenClr (DataReg0);
            GenRelBranch (T, label2);
            PushLabelStack (label1);
            GenMoveQ (1, D0);
            PushLabelStack (label2);
        ELSE
            FreeReg (dest, NOT UseBitSet);
            label1 := NewLabel (); label2 := NewLabel ();
            GenRelBranch (CondCode, label1);
 	    GenClr (dest);
 	    GenRelBranch (T, label2);
            PushLabelStack (label1);
            GenMoveImm (1, dest);
            PushLabelStack (label2);
            push (dest);
        END (* if *);
    END DealWithBeingThere;

PROCEDURE IsCondJump (instr : CARDINAL) : BOOLEAN;
   BEGIN
      RETURN (instr=JPC) OR (instr=JPFC) OR (instr=JPBC)
          OR (instr=ORJP) OR (instr=ANDJP)
   END IsCondJump;


PROCEDURE IsSimpleStore (instr : CARDINAL) : BOOLEAN;
BEGIN
    RETURN (instr = SLW) OR ((SLWn+4 <= instr) AND (instr <= SLWn+15))
        OR (instr = SGW) OR ((SGWn+2 <= instr) AND (instr <= SGWn+15))
END IsSimpleStore;


PROCEDURE GetDestAddress (instr : CARDINAL; VAR dest : ITEM);
   VAR disp : INTEGER;
   BEGIN
      IF (instr=SLW) OR (instr=SGW)
      THEN ignore (next ()); disp := next ();
           StepToNextInstruction (2);
      ELSE disp := instr MOD 16;
           StepToNextInstruction (1)
      END;
      IF instr > SLWn+15 THEN 
         PackItem (dest, AddrRegDisp, G, disp, NoReg)
      ELSE
         PackItem (dest, AddrRegDisp, L, - LocalOffset (INTEGER (disp)),
						NoReg)
      END;
   END GetDestAddress;


PROCEDURE jump (instruction : CARDINAL);
   BEGIN
      CASE instruction OF

       JP
        : GenLabel (); GenLongBcc (T); |

       JPF, JPB
	: GenLabel (); GenBcc (T) |
    
       JPC, JPFC, JPBC
        : CondJump (instruction = JPC) |

       ORJP : OrJump () |

       ANDJP : AndJump ()

      END
   END jump;


BEGIN
   initialise (DestOfAndJpOrJp);
   PreparedCondCode := FALSE;
   RegJustLoadedOnStack := FALSE
END Jumps.
