IMPLEMENTATION MODULE MainPass;

FROM GenProcs IMPORT GenTableEntry, GenLabel, GenDataAndStringFrames,
                     GenEndOfCodeSequence, GenRangeCheckSequence;
FROM labels IMPORT CurrentInstrByteNo;
FROM CodeProcs IMPORT NextInstruction, 
                      StepToNextInstruction,
                      AtCaseLabelTable,
		      NotAtEndOfTable, NotAtEndOfCode;
FROM CaseFor IMPORT case, SelectCase, for, DoEXIT;
FROM LoadStore IMPORT SimpleLoads, SimpleStores;
FROM StackItems IMPORT StackLoad, StackStore;
FROM IndexItems IMPORT IndexLoad, IndexStore;
FROM Externals IMPORT ExternalLoad, ExternalStore;
FROM Jumps IMPORT jump, AtDestOfOrJpOrAndJp, 
                  DealWithBeingThere;
FROM Procs IMPORT procs;
FROM Operations IMPORT DyadicOp, MonadicOp;
FROM Misc IMPORT traps, misc;
FROM Float IMPORT float;
FROM SYS IMPORT sys;
FROM PsStk IMPORT top, base, DecTopOfPsStk, ClearPendingOp;
FROM UtilError IMPORT ReportError;
FROM MCMnemonics IMPORT
       LIn, LIB, LIW, LGA, LLW, LLD, LLWn, LGW, LGWn, SLW, SLD, SLWn, SGW,
       SGWn, LEA, LEW, LED, SEW, SED, LSA, LSWn, LSW, LSD0, TS, SSWn, SSW,
       SSD0, LXB, LXD, SXB, SXD, JPC, ANDJP, ULSS, UGEQ, UADD, SHR, EQL, GEQ,
       ORR, ANDD, INN, ADD, MODD, ABSS, NEG, COM, MSK, NOTT, BIT, FADD, FFCT,
       SYS, LODFW, STORE, COPT, LIN, MOV, FOR1, FOR2, ENTC, EXC, DECS, UCHK,
       TRAP, CHKS, TRA, PCOP, GB, CLn, NOP;


CONST OneByte = 1; TwoBytes = 2; ThreeBytes = 3;
      FourBytes = 4; FiveBytes = 5;

PROCEDURE GenHeaderTable ();
   BEGIN
      GenDataAndStringFrames ();
      GenLabel ();
      ProcedureCount := 0;
      WHILE NotAtEndOfTable DO
         INC (ProcedureCount);
         GenTableEntry ();
	 StepToNextInstruction (TwoBytes);
      END;
      GenRangeCheckSequence;
   END GenHeaderTable;


PROCEDURE GenCode ();
   BEGIN
      WHILE NotAtEndOfCode DO
         GenInstruction ()
      END;
      GenEndOfCodeSequence ();
   END GenCode;


PROCEDURE GenInstruction ();
   VAR Instruction, Length : CARDINAL;
BEGIN
    ExtraBytesProcessed := 0;
    IF AtCaseLabelTable () THEN
        SelectCase ()
    ELSIF AtDestOfOrJpOrAndJp () THEN
        DealWithBeingThere ()
    ELSE Instruction := NextInstruction (Length);
        CASE Instruction OF
          LIn+0..LIB, LIW..LGA, LLW, LLD, LLWn+4..LLWn+15, LGW..LGWn+15:
            GenLabel (); SimpleLoads (Instruction)
        | SLW, SLD, SLWn+4..SLWn+15, SGW..SGWn+15:
            SimpleStores (Instruction)
        | LEA, LEW, LED: GenLabel (); ExternalLoad (Instruction)
        | SEW, SED: ExternalStore (Instruction)
        | LSA, LSWn+0..LSWn+15, LSW..LSD0, TS:
            GenLabel (); StackLoad (Instruction)
        (* Deals with all pushes that use stack addressing (and test & set) *)
        | SSWn+0..SSWn+15, SSW..SSD0: StackStore (Instruction)
        (* Deals with all pops that use stack addressing *)
        | LXB..LXD: GenLabel (); IndexLoad (Instruction)
        (* Deals with all pushes that use indexed addressing *)
        | SXB..SXD: IndexStore (Instruction)
        (* Deals with all pops that use indexed addressing *)
        | JPC..ANDJP: jump (Instruction)
        | ULSS..UGEQ, UADD..SHR, EQL..GEQ, ORR..ANDD, INN, ADD..MODD:
            DyadicOp (Instruction)
        | ABSS, NEG, COM, MSK, NOTT, BIT: MonadicOp (Instruction)
        | FADD..FFCT: float (Instruction)
        | SYS: GenLabel (); sys ()
        | LODFW, STORE..COPT, LIN, MOV: misc (Instruction)
        | FOR1, FOR2: GenLabel (); for (Instruction)
        | ENTC, EXC: GenLabel (); case (Instruction)
        | DECS: GenLabel (); DoEXIT
        | UCHK, TRAP..CHKS: traps (Instruction)
        | TRA, PCOP, GB..CLn+15: procs (Instruction)
        | NOP: GenLabel ();
            IF top > base THEN
                DecTopOfPsStk (top);
                ClearPendingOp (TRUE); ClearPendingOp (FALSE);
            END (* if *);
        ELSE ReportError (TRUE,
            "Unimplemented MCODE instruction (%O3)", Instruction, 0)
        END; (* Of CASE *)

        StepToNextInstruction (Length + ExtraBytesProcessed);
    END;
END GenInstruction;

END MainPass.
