IMPLEMENTATION MODULE Procs;

FROM items IMPORT ITEM, REGISTER, MODE, GlobalLabel, StackPreDec, PackItem,
                  OnRealStack, InDataReg, address, BytesPerWord,
                  Operator, immediate, DataReg0, ConditionCode,
                  PackReg, PackRealStackItem;
FROM CodeProcs IMPORT next, next2, next4, ValueOfNextByte, ignore;
FROM GenProcs IMPORT GenSubaImm, GenMoveReg, GenLink, GenMoveToReg,
                     GenUnlink, GenMovea, GenMoveaReg, GenJsr, GenLabel,
		     GenBsr, GenRtn, GenMove, GenLea, GenLeaGlobalLabel,
                     GenDBra, GenComment, GenMoveQ, GenSuba, GenSubImm,
                     GenLeaExternal, GenMoveMultipleWords, GenShiftImm,
                     GenSimpleDyadicOp, GenNeg, GenRelBranch, GenAddImm,
                     GenMoveImm, GenShift, GenJsrToGlobalLabel,
                     GenJsrToExternalProc, GenAddaImm;
FROM PsStk IMPORT PopToD0, PseudoStackEmpty, DecTopOfPsStk, top, base,
                  PseudStk, DoAnyPendingOp, Pop, NotOnPsStk, pop,
                  RemoveAnyOther, Push, push, SetDyadicOp, DyadicOpPending,
                  PsStkState, StoreParams, ResetScratchHwm, StoreParamsNotTop,
		  StoreParamsNotBottom;
IMPORT PsStk;
FROM RegSlv IMPORT FreeReg, UseBitSet, slave, StopSlaving, PurgeRegSlaves,
                    MarkOnStack;
FROM AddrReg IMPORT GetExternalBaseAddr, FreeAddrReg, InAddrRegSlave,
                    ReleaseAddrReg, SlavedExternal, IncUseCount, 
                    PutIntoAddrRegSlave;
FROM MainPass IMPORT ExtraBytesProcessed;
FROM labels IMPORT PushLabelStack, NewLabel, CurrentLabel, none;
FROM UtilError IMPORT ReportError;
FROM MCMnemonics IMPORT
    LIn, LIB, LIW, LID, CX, TRA, PCOP, GB, GB1, ALOC, ENTR, RTN, CI, CF, CL, 
    CLn, SLW, SLD, SLWn, FUNC, DECS, LLD, LSA, LSW, LSWn, LSD, LSD0, SHR, SHL,
    UMUL, UADD, COPT;
FROM LoadStore IMPORT LocalOffset;
FROM Misc IMPORT ResetStaticStackHeight;

VAR StackPostInc, A0PostInc, A1PostInc, StaticLink : ITEM;


PROCEDURE procs (instruction : CARDINAL);
    VAR n : CARDINAL;
BEGIN
    IF instruction # CX THEN GenLabel () END;
    IF instruction = TRA THEN (* coroutine transfer *)
        Transfer;
    ELSIF instruction = PCOP THEN (* should be handled by 'enter' *)
        ReportError (TRUE, "Unexpected PCOP", 0, 0)
    ELSE
        CASE instruction OF
                
        GB : (* GB get base address n levels down *)
               gb () |

        GB1 : (* GB1 get base address 1 level down *)
               gb1 () |

        ALOC : (* ALLOC allocate block *)
               ReportError (FALSE, "ALLOC got away", 0, 0); HALT |

        ENTR : (* ENTR enter procedure *)
               enter () |

        RTN : (* RTN return from procedure *)
               return () | (* labelling ?*)

        CX : (* CX call external procedure *)
               n := next ();
               IF JustDoneCX AND ((CurrentLabel () # none) OR (n = 0)) THEN
                   JustDoneCX := FALSE;
                   GenMovea (StackPostInc, G);
               END (* if *);
               GenLabel ();
               IF n = 0 THEN
                   SYSTEMproc;
               ELSE
                   CallExternalProc (n); AfterProc ();
               END (* if *); |

        CI : (* CI call procedure at intermediate level *)
               ignore (next ()); CallIntermediateProc (); AfterProc () |

        CF : (* CF call formal procedure *)
               CallFormalProc (); AfterProc () |

        CL : (* CL *)
               ignore (next ()); CallLocalProc (); 
               AfterProc ()

        ELSE (* CL1-CL15 call local procedure *)
               CallLocalProc (); AfterProc ()
        END;
    END;
END procs;

PROCEDURE SYSTEMproc;
    VAR n : CARDINAL;
    BEGIN
        n := next ();
        IF n = 2 THEN
            Newprocess;
        ELSE ReportError (FALSE, "Unknown SYSTEM proc (%R0)*N", n, 0);
        END;
    END SYSTEMproc;

PROCEDURE Newprocess;
    VAR StackSize : INTEGER;
    BEGIN
        StackSize := StoreParams ();
        GenJsrToGlobalLabel (newprocess);
	IF StackSize # 0 THEN GenAddaImm (StackSize, S) END;
        PurgeRegSlaves (TRUE);
    END Newprocess;

PROCEDURE Transfer;
    VAR StackSize : INTEGER;
    BEGIN
	GenMoveReg (G, StackPreDec);
        StackSize := StoreParams ();
        GenJsrToGlobalLabel (transfer);
	IF StackSize # 0 THEN GenAddaImm (StackSize, S) END;
	GenMovea (StackPostInc, G);
        PurgeRegSlaves (TRUE);
    END Transfer;


TYPE
    PCOPTYPE = (char, shifted, multiplied, normal);
    PCOPDATA =
        RECORD
            Type:   PCOPTYPE;
            Pos:    INTEGER;
            Factor: INTEGER;
        END;


PROCEDURE DoPCOP;
    VAR label: CARDINAL;
BEGIN
    label := NewLabel ();
    PushLabelStack (label);
    GenMove (A0PostInc, A1PostInc);
    GenDBra (D0, label);
END DoPCOP;


PROCEDURE PCOPHigh (VAR pcop: PCOPDATA; VAR high: ITEM);
    VAR DataReg4, DataReg5: ITEM;
BEGIN
    PackReg (DataReg4, D4);
    PackReg (DataReg5, D5);
    WITH pcop DO
        GenMoveToReg (high, D4);
        IF (Type = normal) OR (Type = char) THEN
            IF Type = char THEN GenShiftImm (Shr, 2, DataReg4) END;
            GenMoveReg (D4, DataReg0);
            GenAddImm (1, DataReg4);
            GenShiftImm (Shl, 2, DataReg4);
        ELSE
            GenAddImm (1, DataReg4);
            IF Type = shifted THEN
                IF Factor <= 8 THEN
                    GenShiftImm (Shl, Factor, DataReg4);
                ELSE GenMoveImm (Factor, DataReg5);
                    GenShift (Shl, DataReg5, DataReg4);
                END;
            ELSE GenMoveImm (Factor, DataReg5);
                GenJsrToGlobalLabel (multiply);
            END;
            GenMoveReg (D4, DataReg0);
            GenSubImm (1, DataReg0);
            GenShiftImm (Shl, 2, DataReg4);
        END;
        GenSuba (DataReg4, S);
    END;
END PCOPHigh;


PROCEDURE VarPCOP (VAR pcop: PCOPDATA);
    VAR label: CARDINAL;
    VAR array, high: ITEM;
BEGIN
    WITH pcop DO
        PackItem (array, AddrRegDisp, L, -LocalOffset (Pos), NoReg);
        PackItem (high, AddrRegDisp, L, -LocalOffset (Pos+1), NoReg);
        GenMovea (array, A0);
        PCOPHigh (pcop, high);
        (* adjusts S to point at start of array & puts upb in words into D0 *)
        GenMoveReg (S, array);
        GenMoveaReg (S, A1);
        DoPCOP;
    END;
END VarPCOP;


PROCEDURE LoadImm (inst: INTEGER; VAR val, length: INTEGER): BOOLEAN;
BEGIN
    IF (LIn+0 <= inst) AND (inst <= LIn+15) THEN
        val := INTEGER (CARDINAL (inst) MOD 16);
        length := 1;
        RETURN TRUE;
    ELSIF inst = LIB THEN
        val := INTEGER (next ());
        length := 2;
        RETURN TRUE;
    ELSIF inst = LIW THEN
        val := INTEGER (next2 ());
        length := 3;
        RETURN TRUE;
    ELSIF inst = LID THEN
        val := INTEGER (next4 ());
        length := 5;
        RETURN TRUE;
    ELSE RETURN FALSE
    END;
END LoadImm;


PROCEDURE IsPCOP (VAR pcop: PCOPDATA; VAR length: INTEGER): BOOLEAN;

    PROCEDURE lnext (): INTEGER;
    BEGIN
        INC (length);
        RETURN INTEGER (next ());
    END lnext;

    VAR l, amount, inst: INTEGER;

BEGIN
    length := 0;
    WITH pcop DO
        Pos := lnext ();
        IF LoadImm (lnext (), amount, l) THEN
            INC (length, l-1); (* have already taken account of first byte *)
            inst := lnext ();
            IF (inst = SHR) AND (amount = 2) THEN
                Type := char;
                RETURN (lnext () = LIn+1) AND (lnext () = UADD)
                    AND (lnext () = PCOP) AND (lnext () = Pos)
            ELSIF (inst = UADD) AND (amount = 1) THEN
                inst := lnext ();
                IF inst = PCOP THEN
                    Type := normal;
                    RETURN lnext () = Pos;
                ELSIF LoadImm (inst, Factor, l) THEN
                    INC (length, l-1);
                    inst := lnext ();
                    IF inst = UMUL THEN
                        Type := multiplied;
                        RETURN (lnext () = PCOP) AND (lnext () = Pos);
                    ELSIF inst = SHL THEN
                        Type := shifted;
                        RETURN (lnext () = PCOP) AND (lnext () = Pos);
                    ELSE RETURN FALSE
                    END;
                ELSIF inst = COPT THEN
                    Type := shifted;
                    Factor := 1;
                    RETURN (lnext () = UADD)
                        AND (lnext () = PCOP) AND (lnext () = Pos)
                ELSE RETURN FALSE;
                END;
            ELSE RETURN FALSE
            END;
        ELSE RETURN FALSE;
        END;
    END;
END IsPCOP;


PROCEDURE ConstPCOP (sz: INTEGER);
    VAR pos: INTEGER;
        array: ITEM;
BEGIN
    GenSubaImm ((sz * BytesPerWord), S);
    pos := INTEGER (next ());
    INC (ExtraBytesProcessed);
    PackItem (array, AddrRegDisp, L, -LocalOffset (pos), NoReg);
    GenMovea (array, A0);
    GenMoveaReg (S, A1);
    GenMoveReg (S, array);
    IF sz > 4 THEN
        GenMoveImm (sz-1, DataReg0);
        DoPCOP;
    ELSE WHILE sz > 0 DO GenMove (A0PostInc, A1PostInc); DEC (sz) END;
    END;
END ConstPCOP;


PROCEDURE Alloc (sz: INTEGER);
    VAR inst: INTEGER;
        array: ITEM;
BEGIN
    inst := next ();
    IF inst = SLW THEN
        INC (ExtraBytesProcessed, 2);
        PackItem (array, AddrRegDisp, L, - LocalOffset (INTEGER (next ())),
					 NoReg);
    ELSIF (SLWn+4 <= inst) AND (inst <= SLWn+15) THEN
        INC (ExtraBytesProcessed);
        PackItem (array, AddrRegDisp, L,
		- LocalOffset (INTEGER (CARDINAL (inst) MOD 16)), NoReg);
    ELSE
        ReportError (TRUE, "Unexpected instruction (%O3) after ALLOC*N", inst, 0)
    END (* if *);
    GenSubaImm ((sz * BytesPerWord), S);
    GenMoveReg (S, array);
END Alloc;


PROCEDURE LocalStore (inst: INTEGER; VAR length: INTEGER): BOOLEAN;
BEGIN
    IF (inst = SLW) OR (inst = SLD) THEN
        ignore (next ());
        length := 2;
        RETURN TRUE;
    ELSIF (SLWn+4 <= inst) AND (inst <= SLWn+15) THEN
        length := 1;
        RETURN TRUE;
    ELSE RETURN FALSE;
    END;
END LocalStore;


PROCEDURE InitStackFrame ();
    VAR inst, length, val: INTEGER;
        pcop: PCOPDATA;
BEGIN
    LOOP
        inst := next ();
        IF LocalStore (inst, length) THEN INC (ExtraBytesProcessed, length);
        ELSIF LoadImm (inst, val, length) THEN
            inst := next ();
            IF inst = ALOC THEN
                INC (ExtraBytesProcessed, length + 1); (* +1 for the ALOC *)
                Alloc (val)
            ELSIF inst = PCOP THEN
                INC (ExtraBytesProcessed, length + 1); (* +1 for the PCOP *)
                ConstPCOP (val)
            ELSE EXIT
            END;
        ELSIF inst = LLD THEN
            IF IsPCOP (pcop, length) THEN
                VarPCOP (pcop);
                INC (ExtraBytesProcessed, length + 1); (* +1 for the LLD *)
            ELSE EXIT
            END;
        ELSE EXIT
        END;
    END;
END InitStackFrame;


PROCEDURE enter ();
BEGIN
    GenComment ("Procedure Entry");
    GenLeaExternal (0, G);
    top := base;
    ResetScratchHwm ();
    ResetStaticStackHeight;
    InitSize := next ();
    ParmSpace := next ();
    ExtraBytesProcessed := 0;
    GenLink (L, -(InitSize - ParmSpace) * BytesPerWord);
    InitStackFrame ();
END enter;


PROCEDURE return ();
    BEGIN
        IF NOT PseudoStackEmpty () THEN 
            PopToD0 ();
        END (* if *);
	GenUnlink (L);
        GenRtn ();
    END return;

(*
PROCEDURE StoreParams (TotalStackFrameSize : CARDINAL);
    BEGIN 
        IF DyadicOpPending THEN
            Pop (AddrRegDisp, S, -INTEGER (top - 1 + TotalStackFrameSize));
        END (* if *);
        WHILE NOT PseudoStackEmpty () DO
            Pop (AddrRegDisp, S, - INTEGER (top + TotalStackFrameSize));
        END (* while *);
    END StoreParams;
*)

PROCEDURE AfterProc ();
    VAR NextByte : CARDINAL;
    BEGIN
        PurgeRegSlaves (TRUE);
        NextByte := next ();
        ExtraBytesProcessed := 0;
        IF NextByte = FUNC (* is a function call *) THEN
            INC (ExtraBytesProcessed);
            push (DataReg0);
            NextByte := next ()
        END (* if *);
(* used to be a DECS after every CF - front end now changed
        IF IsFormal THEN
            IF NextByte <> DECS (* DECS *) THEN
                ReportError (TRUE,
                             "CF not followed by DECS (%O3)*N",
                             NextByte, 0);
            END (* if *);
            INC (ExtraBytesProcessed);
        END (* if *);
*)
    END AfterProc;


VAR JustDoneCX : BOOLEAN;
    CurrentCXDisp : CARDINAL;


PROCEDURE CallExternalProc (ExtDisp : CARDINAL);
    VAR 
        AddrReg : REGISTER;
        SameMod : BOOLEAN;
	StackSize : INTEGER;
    BEGIN
	SameMod := FALSE;
	StackSize := 0;
        IF JustDoneCX THEN
            IF CurrentCXDisp = ExtDisp THEN SameMod := TRUE END;
            JustDoneCX := FALSE;
        ELSE
            GenMoveReg (G, StackPreDec);
            StackSize := StoreParams();
        END (* if *);
        IF SameMod THEN
(*	   GenQuickJsrToExternalProc (ExtDisp, next ());  *)
GenJsrToExternalProc (ExtDisp, next ());
	ELSE (* GenSlowJsrToExternalProc (ExtDisp, next ()); *)
GenJsrToExternalProc (ExtDisp, next ());
	     IF StackSize # 0 THEN GenAddaImm (StackSize, S) END;
	END;
        IF ValueOfNextByte () = CX THEN
            JustDoneCX := TRUE;
            CurrentCXDisp := ExtDisp;
        ELSE
            GenMovea (StackPostInc, G)
        END (* if *);
   END CallExternalProc;


PROCEDURE CallIntermediateProc ();
    VAR Reg, StoreStaticLink : ITEM;
        StackSize : INTEGER;
    BEGIN
        DoAnyPendingOp ();
(*        PackRealStackItem (StoreStaticLink, 2);
        IF (top >= 3) AND OnRealStack (PseudStk [2]) THEN
            (* PseudStk item is stored where static link is going to go *)
            FreeReg (Reg, NOT UseBitSet);
            GenMove (PseudStk [2], Reg);
            PseudStk [2] := Reg;
            MarkOnStack (Reg.Register);
        END (* if *);
        pop (StoreStaticLink);  *)
	GenMoveReg (L, StackPreDec);	
        StackSize := StoreParamsNotTop ();
(*        GenLink (L, - (2 * BytesPerWord));  *)
	GenMovea (PseudStk [top], L);
	DecTopOfPsStk (1);
        GenBsr ();
        IF StackSize # 0 THEN GenAddaImm (StackSize, S) END;
	GenMovea (StackPostInc, L);
(*        GenUnlink (L);   *)
    END CallIntermediateProc;


PROCEDURE CallLocalProc ();
    VAR StoreStaticLink : ITEM;
	StackSize : INTEGER;
    BEGIN
        StackSize := StoreParams ();
        GenBsr ();
	IF StackSize # 0 THEN GenAddaImm (StackSize, S) END;
    END CallLocalProc;


PROCEDURE CallFormalProc ();
    VAR item : ITEM;
 	StackSize : INTEGER;
	AddrReg : REGISTER;
    BEGIN
        (*
        MOVEM.W (S)+,D0-D1          Proc offset to D1 & module offset to D0
        MOVE.L G,(S)-               Store Static link
        LINK L,#-8                  And dynamic link
        LEA @_GFT,A1
        MOVEA 0(A1,D0.L),G          Module data frame base to G
        MOVEA (G),A0                Module code frame base to A0
        MOVEA 0(A0,D1.L),A0         Procedure start address to A0
        JSR (A0)                    Jump to subroutine
        UNLK L                      Restore dynamic link
        MOVEA (S)+,G                Restore static link
        *)
(*        GenMovea (StackPostInc, AddrReg);  *)
(* This cod should be changed to 
Push G,
StoreparamsNotBottom
Decide what the bottom is and gen a JSR 
*)
        GenMoveReg (G, StackPreDec);
	StackSize := StoreParamsNotBottom();
	IF NOT address (PseudStk [top]) THEN
	    AddrReg := FreeAddrReg (NotOnPsStk);
            GenMovea (PseudStk [top], AddrReg);
            PackItem (item, AddrRegDisp, AddrReg, 0, NoReg);
            GenJsr (item);
	ELSE GenJsr (PseudStk [top])
	END;
	DecTopOfPsStk (1);
        GenAddaImm (StackSize, S);
        GenMovea (StackPostInc, G);
    END CallFormalProc;
    

PROCEDURE IsStackLoad (inst : CARDINAL; VAR disp : INTEGER; 
                                            VAR IsDouble : BOOLEAN) : BOOLEAN;
BEGIN
    IF (LSWn+0 <= inst) AND (inst <= LSWn+15) THEN
        ExtraBytesProcessed := 1;
        disp := -LocalOffset (INTEGER (inst MOD 16));
        IsDouble := FALSE;
        RETURN TRUE;
    ELSIF inst = LSW THEN
        ExtraBytesProcessed := 2;
        disp := - LocalOffset (INTEGER (next ()));
        IsDouble := FALSE;
        RETURN TRUE;
    ELSIF inst = LSD THEN
        ExtraBytesProcessed := 2;
        disp := -LocalOffset (INTEGER (next ()));
        IsDouble := TRUE;
        RETURN TRUE;
    ELSIF inst = LSD0 THEN
ReportError (FALSE, "LSD0 ????????????????", 0, 0);
        ExtraBytesProcessed := 1;
        disp := 0;
        IsDouble := TRUE;
        RETURN TRUE;
    ELSE RETURN FALSE;
    END;
END IsStackLoad;


PROCEDURE gb1 ();
    VAR AddrReg : REGISTER;
        AddrRegIndirect : ITEM;
        disp : INTEGER;
        nextinstruc : CARDINAL;
        IsDouble : BOOLEAN;
        SaveParmSpace : INTEGER;
    BEGIN
        DoAnyPendingOp ();
	SaveParmSpace := ParmSpace;
	ParmSpace := INTEGER (next ());
        nextinstruc := next ();
        IF nextinstruc = LSA THEN
            disp := -LocalOffset (INTEGER (next ()));
            ExtraBytesProcessed := 2;
            IF NOT InAddrRegSlave (StaticLink, AddrReg) THEN
                AddrReg := FreeAddrReg (NotOnPsStk);
                GenMovea (StaticLink, AddrReg);
                PutIntoAddrRegSlave (AddrReg, StaticLink);
            END (* if *);
            Push (ADDRaddrRegDisp, AddrReg, disp);
            IncUseCount (AddrReg, top);
        ELSIF IsStackLoad (nextinstruc, disp, IsDouble) THEN
        (* this proc adjusts ExtraBytesProcessed as necessary *)
            IF NOT InAddrRegSlave (StaticLink, AddrReg) THEN
                AddrReg := FreeAddrReg (NotOnPsStk);
                GenMovea (StaticLink, AddrReg);
                PutIntoAddrRegSlave (AddrReg, StaticLink);
            END (* if *);
            Push (AddrRegDisp, AddrReg, disp);
            IncUseCount (AddrReg, top);
            IF IsDouble THEN
                Push (AddrRegDisp, AddrReg, disp-1);
                IncUseCount (AddrReg, top);
            END (* if *);
        ELSE
            push (StaticLink);
            INCL (PsStkState [top], PsStk.GB);
	    GBParmSpace [top]:= ParmSpace;
        END (* if *);
	ParmSpace := SaveParmSpace;
    END gb1;


PROCEDURE gb ();
    VAR level, loopvar, nextinstruc : CARDINAL;
        disp : INTEGER;
        AddrReg : REGISTER;
        AddrRegIndirect : ITEM;
        IsDouble : BOOLEAN;
	SaveParmSpace : INTEGER;
    BEGIN
        level := next ();
        IF level = 1 THEN
            gb1 (); RETURN;
        END (* if *);
	SaveParmSpace := ParmSpace;
	ParmSpace := INTEGER (next());
        nextinstruc := next ();
        DoAnyPendingOp ();
        AddrReg := FreeAddrReg (NotOnPsStk);
        PackItem (AddrRegIndirect, AddrRegDisp, AddrReg, 0, NoReg);
        GenMovea (StaticLink, AddrReg);
        FOR loopvar := 1 TO level - 2 DO
            GenMovea (AddrRegIndirect, AddrReg);
        END (* for *);
        IF nextinstruc = LSA THEN
            disp := -LocalOffset (INTEGER (next ()));
            ExtraBytesProcessed := 2;
            GenMovea (AddrRegIndirect, AddrReg);
            Push (ADDRaddrRegDisp, AddrReg, disp);
        ELSIF IsStackLoad (nextinstruc, disp, IsDouble) THEN
            GenMovea (AddrRegIndirect, AddrReg);
            Push (AddrRegDisp, AddrReg, disp);
            IF IsDouble THEN
                IncUseCount (AddrReg, top);
                Push (AddrRegDisp, AddrReg, disp-1);
            END (* if *);
        ELSE
            push (AddrRegIndirect);
            INCL (PsStkState [top], PsStk.GB);
	    GBParmSpace [top] := ParmSpace;
        END (* if *);
        IncUseCount (AddrReg, top);
	ParmSpace := SaveParmSpace;
    END gb;


BEGIN
    JustDoneCX := FALSE;
    PackItem (StackPostInc, PostInc, S, 0, NoReg);
    PackItem (A0PostInc, PostInc, A0, 0, NoReg);
    PackItem (A1PostInc, PostInc, A1, 0, NoReg);
    PackItem (StaticLink, AddrRegDisp, L, 0, NoReg);
END Procs.
