IMPLEMENTATION MODULE PsStk;

FROM items IMPORT REGISTER, MODE, ITEM, Operator,
                  MonOperator, ConditionCode,
                  PackItem, StackPreDec,
		  equal, NullRegItem, DataReg0,
                  immediate, address, InDataReg, BytesPerWord,
		  UsesSlavedAddrReg;
FROM RegSlv IMPORT FreeReg, protect, slave, Slave,
                   PutIntoSlave, ResetUseBits, MarkOnStack,
		   UseBitSet, SetRegFree;
FROM AddrReg IMPORT FreeAddrReg, PutIntoAddrRegSlave,
                    InAddrRegSlave, DecUseCount, AddressMove,
		    ProtectAddress, UnprotectAddress;
FROM GenOp IMPORT generate, GenerateMonadicOp;
FROM GenProcs IMPORT GenMove, GenSubImm, GenMoveReg, GenLea,
		     SetFrameSize;
FROM Relations IMPORT compare, NotTest;
FROM UtilError IMPORT ReportError;
FROM Misc IMPORT StaticStackHeight;
FROM Procs IMPORT InitSize, ParmSpace;

CONST dyadic = TRUE;
      monadic = FALSE;

VAR DyadicOp        : Operator;
    MonadicOp       : MonOperator;

(* This procedure is used when the top item
   on the pseudo stack is to be moved to
   a destination address 'dest'. If there
   is an operation pending a procedure is
   called to generate code to work out the
   result of the operation and store it in
   'dest' in the most efficient way possible.
   Otherwise a simple MOVE instruction is
   used.
*)
PROCEDURE pop (VAR dest : ITEM);
    VAR opd1, opd2 : ITEM;
    BEGIN
        IF DyadicOpPending THEN
            Get2Operands (opd1, opd2, dest);
            generate (DyadicOp, opd1, opd2, dest);
            DEC (top, 2);
            ClearPendingOp (dyadic);
        ELSIF MonadicOpPending THEN
            Get1Operand (opd1, dest);
            GenerateMonadicOp (MonadicOp, opd1, dest);
            DEC (top);
            ClearPendingOp (monadic);
        ELSE
            MoveTopOfStackToDest (dest);
            DEC (top);
        END (* if *);
        ResetUseBits ();
    END pop;

PROCEDURE Get2Operands (VAR opd1, opd2, dest : ITEM);
    BEGIN
        GetOperand (opd1, top);
        GetOperand (opd2, top-1);
        RemoveAnyOther (dest, top - 2);
    END Get2Operands;

PROCEDURE Get1Operand (VAR opd, dest : ITEM);
    BEGIN
        GetOperand (opd, top);
        RemoveAnyOther (dest, top - 1);
    END Get1Operand;

PROCEDURE GetOperand (VAR opd : ITEM; posn : CARDINAL);
    BEGIN
        opd := PseudStk [posn];
        FreeAnyRegUsedBy (posn);
        PsStkState [posn] := STATE {};
        ProtectOpd (opd);
    END GetOperand;

PROCEDURE ProtectOpd (VAR opd : ITEM);
    VAR SlvOpd : ITEM;
    BEGIN
        IF InDataReg (opd) THEN
            protect (opd);
        ELSE
            Slave (opd, SlvOpd);
            protect (SlvOpd);
        END (* if *);
    END ProtectOpd;

PROCEDURE MoveTopOfStackToDest (VAR dest : ITEM);
    VAR source : ITEM;
        AddrReg : REGISTER;
    BEGIN
        Get1Operand (source, dest);
        IF address (source) THEN
            IF NOT InAddrRegSlave (source, AddrReg) THEN
                AddrReg := FreeAddrReg (NotOnPsStk);
                GenLea (source, AddrReg);
                PutIntoAddrRegSlave (AddrReg, source);
            END (* if *);
            GenMoveReg (AddrReg, dest);
        ELSE
            IF NOT equal (source, dest) THEN
                IF InAddrRegSlave (source, AddrReg) THEN
                    GenMoveReg (AddrReg, dest);
                ELSE
                    slave (source);
                    GenMove (source, dest);
                    PutIntoSlave (source, dest, FALSE);
                END (* if *);
            END (* if *);
        END (* if *);
    END MoveTopOfStackToDest;

(* PopToD0 is used when returning from
   a function call. The code generator
   assumes that all function results
   are returned in D0
*)
PROCEDURE PopToD0 ();
    BEGIN
        protect (DataReg0);
        pop (DataReg0);
    END PopToD0;


(* This procedure is only actually useful on rare occasions.
   It is used by 'pop' - when popping a value off the pseudo
   stack to some location. The point of 'RemoveAnyOther' is
   that this location will be altered when such a 'pop'
   occurs so if, further down the pseudo stack, this location
   is given as the location where a value x, say, can be found
   then when the location is changed the value x will be lost.
   This procedure detects such cases and saves values like x
   at offsets to the Stack Pointer register. Note, however that
   usually 'popping' a value to somewhere in store implies that
   the pseudo stack will be empty after the pop so this saving
   of values only happens very rarely.
*)
PROCEDURE RemoveAnyOther (VAR item : ITEM; limit : CARDINAL);
    VAR count : CARDINAL;
        reg   : REGISTER;
        stacked, PsStkItem : ITEM;
    BEGIN
        IF (limit <= base) OR equal (item, NullRegItem)
        THEN RETURN
        END; (* this is the usual case *)

        FOR count := base+1 TO limit DO
            PsStkItem := PseudStk [count];
            IF equal (item, PsStkItem) THEN
                MoveIntoScratchSpace (count);
            END (* if *);
        END (* for *);
    END RemoveAnyOther;


(* This procedure pushes the given
   item onto the pseudo stack. If
   an operation is pending the result
   is computed into a register before
   the push is done
*)
PROCEDURE push (VAR item : ITEM);
   BEGIN
      DoAnyPendingOp ();
      IncTop;
      PseudStk [top] := item;
      IF InDataReg (item) THEN
        MarkOnStack (item.Register);
      END (* if *);
   END push;

PROCEDURE IncTop;
    BEGIN
        (* $T-,$R- *)
        IF top = PseudStkUpb THEN
            ReportError (TRUE,
             "Expression stack overflow (too many parameters to procedure?)",
                         0, 0);
        ELSE
            INC (top);
        END (* if *);
        (* $T+,$R+ *)
    END IncTop;

PROCEDURE DoAnyPendingOp ();
   BEGIN
      IF DyadicOpPending OR MonadicOpPending
      THEN pop (NullRegItem);
           push (ResultItem)
      END
   END DoAnyPendingOp;


PROCEDURE SetDyadicOp (op : Operator);
   BEGIN
      DoAnyPendingOp ();
      IF address (PseudStk [top])
      THEN GetValueOfAddressOntoPsStk (top)
      END;
      IF address (PseudStk [top-1])
      THEN GetValueOfAddressOntoPsStk (top-1)
      END;
      DyadicOp := op;
      DyadicOpPending := TRUE
   END SetDyadicOp;

PROCEDURE SetMonadicOp (op : MonOperator);
    BEGIN
        DoAnyPendingOp ();
        IF address (PseudStk [top]) THEN
            GetValueOfAddressOntoPsStk (top)
        END (* if *);
        MonadicOp := op;
        MonadicOpPending := TRUE;
    END SetMonadicOp;

PROCEDURE ClearPendingOp (Dyadic : BOOLEAN);
   BEGIN
      IF Dyadic
      THEN DyadicOpPending := FALSE
      ELSE MonadicOpPending := FALSE
      END
   END ClearPendingOp;


PROCEDURE PseudoStackEmpty () : BOOLEAN;
   BEGIN RETURN top = base END PseudoStackEmpty;


PROCEDURE Push (mode : MODE;
                register : REGISTER;
                dispOrData : INTEGER);
   VAR item : ITEM;
   BEGIN
      PackItem (item, mode, register, dispOrData, NoReg);
      push (item);
   END Push;


PROCEDURE Pop (mode : MODE;
               register : REGISTER;
               dispOrData : INTEGER);
   VAR item : ITEM;
   BEGIN
      PackItem (item, mode, register, dispOrData, NoReg);
      pop (item);
   END Pop;


(* Given a position on the pseudo stack
   this procedure notifies the register
   slaves that the item at this position
   is no longer on the pseudo stack.
*)
PROCEDURE FreeAnyRegUsedBy (PsStkPosn : CARDINAL);
   VAR item : ITEM;
   BEGIN
      item := PseudStk [PsStkPosn];
      IF InDataReg (item)
      THEN SetRegFree (item.Register)
      END;
      IF UsesSlavedAddrReg (item)
      THEN DecUseCount (item.Register, PsStkPosn)
      END;
   END FreeAnyRegUsedBy;

PROCEDURE MoveIntoScratchSpace (posn : CARDINAL);
   VAR dest : ITEM;
   BEGIN
      PackItem (dest, AddrRegDisp, L, -(INTEGER (posn - base)
		+ InitSize - ParmSpace + StaticStackHeight), NoReg);
      GenMove (PseudStk [posn], dest);
      PseudStk [posn] := dest;
      IF posn + CARDINAL (StaticStackHeight) > ScratchHwm THEN
	ScratchHwm := posn + CARDINAL (StaticStackHeight);
	SetFrameSize (ScratchHwm);
      END;
   END MoveIntoScratchSpace;


PROCEDURE ResetScratchHwm ();
    BEGIN
        ScratchHwm := base;
    END ResetScratchHwm;

(* Used by 'FreeReg' when it has to free
   a register holding a pseudo stack item
*)
PROCEDURE PutInScratchSpace (reg : REGISTER);
   VAR Point : CARDINAL;
   BEGIN
      FOR Point := base+1 TO top DO
	 WITH PseudStk [Point] DO
	    IF (Mode = DataReg) AND (Register = reg)
	    THEN MoveIntoScratchSpace (Point);
            END
         END
      END
   END PutInScratchSpace;

PROCEDURE FindPbase (low, high : INTEGER): INTEGER;
    VAR i, res : INTEGER;
    BEGIN
        res := low;
        FOR i := low TO high DO
	    IF STOTed IN PsStkState [i] THEN res := i + 1 END
	END;
	RETURN (res);
    END FindPbase;

PROCEDURE StoreParamsNotBottom (): INTEGER;
    VAR i : CARDINAL;
        pbase : CARDINAL;
	size : INTEGER;
    BEGIN
	DoAnyPendingOp();
        pbase := FindPbase (base + 2, top);
	FOR i := pbase TO top DO PushParam (PseudStk [i]) END;
        size := top - pbase + 1; 
	top := pbase - 1;
	RETURN (size * BytesPerWord);
    END StoreParamsNotBottom;

PROCEDURE StoreParamsNotTop () : INTEGER;
    VAR size : INTEGER;
        save : ITEM;
    BEGIN
	DoAnyPendingOp();
	save := PseudStk [top];
        ProtectAddress (save);
	DecTopOfPsStk (1);
	size := StoreParams();
	top := top + 1; 
 	PseudStk [top] := save;
        UnprotectAddress (save);
        RETURN (size);
    END StoreParamsNotTop;

PROCEDURE PushParam (source : ITEM);
    VAR AddrReg : REGISTER;
    BEGIN
        IF address (source) THEN
	    IF NOT InAddrRegSlave (source, AddrReg) THEN
                AddrReg := FreeAddrReg (NotOnPsStk);
                GenLea (source, AddrReg);
                PutIntoAddrRegSlave (AddrReg, source);
            END (* if *);
            GenMoveReg (AddrReg, StackPreDec);
        ELSE
            IF InAddrRegSlave (source, AddrReg) THEN
               GenMoveReg (AddrReg, StackPreDec);
            ELSE
                slave (source);
                GenMove (source, StackPreDec);
            END (* if *);
	END
    END PushParam;

PROCEDURE StoreParams () : INTEGER;
    VAR i : CARDINAL;
	size : INTEGER;
        pbase : CARDINAL;
    BEGIN
	DoAnyPendingOp();
        pbase := FindPbase (base + 1, top);
	FOR i := pbase TO top DO
	    PushParam (PseudStk [i]);
	    PsStkState [i] := STATE {}
	    END;
        size := top - pbase + 1;
	top := pbase - 1;
	RETURN (size * BytesPerWord);
    END StoreParams;

PROCEDURE BooleanOpPending () : BOOLEAN;
   BEGIN RETURN
                (DyadicOpPending AND (((DyadicOp >= Eql)
                 AND (DyadicOp <= Geq)) OR (DyadicOp = In)))
                OR
                (MonadicOpPending AND (MonadicOp = Not))
   END BooleanOpPending;


(* Generates a compare on the two
   top items of the pseudo stack and
   returns a condition code according
   to what condition should be tested
   for
*)
PROCEDURE DoBooleanOp () : ConditionCode;
    VAR cc : ConditionCode;
    VAR opd1, opd2 : ITEM;
    BEGIN
        GetOperand (opd1, top);
        IF DyadicOpPending THEN
            GetOperand (opd2, top - 1);
            cc := compare (DyadicOp, opd1, opd2);
            DEC (top, 2);
            ClearPendingOp (dyadic);
        ELSE (* NOT *)
            cc := NotTest (opd1);
            DEC (top);
            ClearPendingOp (monadic);
        END (* if *);
        ResetUseBits ();
        RETURN cc;
    END DoBooleanOp;

PROCEDURE NOTPendingRelationalOp () : BOOLEAN;
    BEGIN
        IF DyadicOpPending THEN
            CASE DyadicOp OF

              Eql  : DyadicOp := Neq  |
              Neq  : DyadicOp := Eql  |
              Ulss : DyadicOp := Ugeq |
              Lss  : DyadicOp := Geq  |
              Uleq : DyadicOp := Ugtr |
              Leq  : DyadicOp := Gtr  |
              Ugtr : DyadicOp := Uleq |
              Gtr  : DyadicOp := Leq  |
              Ugeq : DyadicOp := Ulss |
              Geq  : DyadicOp := Lss
            ELSE RETURN FALSE

            END (* case *);
            RETURN TRUE;
        ELSIF MonadicOpPending AND (MonadicOp = Not) THEN
            MonadicOpPending := FALSE;
            RETURN TRUE;
        ELSE
            RETURN FALSE;
        END; (* if *)
    END NOTPendingRelationalOp;

PROCEDURE GetValueOfAddressOntoPsStk (posn : CARDINAL);
   VAR Reg, Address : ITEM;
       AddrReg : REGISTER;
   BEGIN
      Address := PseudStk [posn];
      IF NOT InAddrRegSlave (Address, AddrReg)
      THEN
         FreeAnyRegUsedBy (posn);
         AddrReg := FreeAddrReg (NotOnPsStk);
         GenLea (Address, AddrReg);
         PutIntoAddrRegSlave (AddrReg, Address)
      END;
      FreeReg (Reg, NOT UseBitSet);
      GenMoveReg (AddrReg, Reg);
      MarkOnStack (Reg.Register);
      PseudStk [posn] := Reg
   END GetValueOfAddressOntoPsStk;

PROCEDURE PutPsStkOnRealStack (limit : CARDINAL);
   VAR PsStkPoint, dest : ITEM;
       point : CARDINAL;
       AddrReg : REGISTER;
   BEGIN
      FOR point := base+1 TO limit DO
        PackItem (dest, AddrRegDisp, L, -(INTEGER (point - base)
		 + InitSize - ParmSpace + StaticStackHeight), NoReg);
	 PsStkPoint := PseudStk [point];
	 FreeAnyRegUsedBy (point);
         IF address (PsStkPoint) THEN
	    AddressMove (PsStkPoint, dest)
         ELSIF InAddrRegSlave (PsStkPoint, AddrReg) THEN
            GenMoveReg (AddrReg, dest);
         ELSIF NOT equal (PsStkPoint, dest) THEN
            slave (PsStkPoint);
            GenMove (PsStkPoint, dest)
         END (* if *);
	 PseudStk [point] := dest;
      END;
      IF limit + CARDINAL (StaticStackHeight) > ScratchHwm
      THEN ScratchHwm := limit + CARDINAL (StaticStackHeight);
          SetFrameSize (ScratchHwm)
      END;
   END PutPsStkOnRealStack;

PROCEDURE DecTopOfPsStk (num : CARDINAL);
   VAR count : CARDINAL;
   BEGIN
      FOR count := 1 TO num DO
         FreeAnyRegUsedBy (top);
         PsStkState [top] := STATE {};
	 DEC (top)
      END
   END DecTopOfPsStk;

PROCEDURE AssertEmptyPsStk;
BEGIN
    IF top # base THEN
        ReportError (FALSE, "Pseudo stack error", 0, 0);
        HALT;
    END;
END AssertEmptyPsStk;

BEGIN
   DyadicOpPending := FALSE;
   MonadicOpPending := FALSE;
   FOR top := base+1 TO PseudStkUpb DO
       PsStkState [top] := STATE {};
   END (* for *);
   top := base;
END PsStk.
