IMPLEMENTATION MODULE ReadSym;

FROM SYSTEM IMPORT BITSPERWORD, BYTESPERWORD, WORD;

FROM Streams IMPORT Stream, Get, Backspace, Put, ErrorCode, Delete;
FROM SystemTypes IMPORT String;
FROM UtilError IMPORT ReportError;  
FROM SysStreams IMPORT sysOut;
FROM WriteF IMPORT H;

IMPORT Strings;

FROM MCSymFileDefs IMPORT SymFileSymbols;
FROM FileStream IMPORT 
    CreateOutput, StreamErrorCode;
FROM M2Files IMPORT CreateInput;
FROM UnixErrors IMPORT SysErrorString, sysErrList;


CONST
    MaxIdentSize = 256;
    MaxStringSize = 256;
    maxsymlength = 30;
CONST bufUpb = 511;
CONST bufLength = bufUpb+1;
TYPE BUFFERLENGTH = [0 .. bufLength];
     MODE = (In, Out, InOut);
     BUFFER = ARRAY [0 .. bufLength] OF CHAR;
     BUFFERPTR = POINTER TO BUFFER;

TYPE BUFFERDATA = POINTER TO 
    RECORD
    fildes: CARDINAL;			(* Unix file descriptor	      *)
    size: BUFFERLENGTH;		        (* size of current input buffer*)
    streambufno: CARDINAL;              (* data pair specifying *)
    streambufoffset: BUFFERLENGTH;      (* position on stream *)
    filebufno: CARDINAL;                (* data pair specifying position *)
    filebufoffset: BUFFERLENGTH;        (* in file *)
    bufferismodified: BOOLEAN;
    mode: MODE;
    stream: Stream;
    eof: BOOLEAN;
    buf: BUFFER;
    END;

TYPE
    LiteralString = ARRAY [0 .. MaxStringSize - 1] OF CHAR;
    IdentifierString = ARRAY [0 .. MaxIdentSize - 1] OF CHAR;

VAR bufdata : BUFFERDATA;

VAR ModuleName : ARRAY [0..30] OF CHAR;
    symStream: Stream;
    dumcard : CARDINAL;
    underline : ARRAY [0..1] OF CHAR;
    modname : String;

PROCEDURE ModuleObjectNumber (mod : String; objnum : CARDINAL): String;
VAR numchar : ARRAY [0..3] OF CHAR;
BEGIN
    numchar [0] := CHAR (CARDINAL('0') + objnum DIV 100);
    numchar [1] := CHAR (CARDINAL('0') + (objnum MOD 100) DIV 10);
    numchar [2] := CHAR (CARDINAL('0') + objnum MOD 10);
    numchar [3] := 0C;
    RETURN Strings.ConcatSC (mod, numchar);
END ModuleObjectNumber;
    
PROCEDURE DecodeSymbolFile();

    VAR
        Sym:        SymFileSymbols;
        Ident:      IdentifierString;
        string:     LiteralString;
        
        
    PROCEDURE Error(code: CARDINAL);
	VAR i : CARDINAL;
        BEGIN
            ReportError(TRUE, "error %i in symfile for %s", code, H(ModuleName));
        END Error;

    PROCEDURE ExpectSymbol(requiredsymbol: SymFileSymbols);
        BEGIN
            GetSymbol();
            IF Sym # requiredsymbol THEN Error (ORD (requiredsymbol)) END;
        END ExpectSymbol;
        
    PROCEDURE ExpectCardinal(): CARDINAL;
        VAR
            number:     CARDINAL;
        BEGIN
            number := GetCardinal();
            IF Sym # normalconstSS THEN
                Error(1)
            END (* if *);
            RETURN number
        END ExpectCardinal;
        
        
    PROCEDURE ExpectIdentifier;
        BEGIN
            GetIdentifier();
            IF Sym # identSS THEN
                Error(2)
            END (* if *);
        END ExpectIdentifier;
        
    PROCEDURE ExpectString;
        BEGIN
            GetString();
            IF Sym # stringconstSS THEN
                Error(3)
            END (* if *);
        END ExpectString;
        
    PROCEDURE GetSymbol;
        BEGIN
            Sym := SymFileSymbols(Get(symStream));
        END GetSymbol;
        
    PROCEDURE UnGetSymbol;
        BEGIN				(*  Only works for simple symbols!*)
            Backspace(symStream);
        END UnGetSymbol;
        
    PROCEDURE GetIdentifier;
        VAR
            i:      INTEGER;
        BEGIN
            GetSymbol();
            IF Sym = identSS THEN
                FOR i := 0 TO MaxIdentSize - 1 DO
                    Ident[i] := CHAR(Get(symStream));
                    IF Ident[i] = 0C THEN
                        RETURN
                    END (* if *)
                END (* for *);
                Error(4)
            END (* if *)
        END GetIdentifier;

        PROCEDURE GetNumber(): CARDINAL;
            VAR
                number:     CARDINAL;
            BEGIN
                number := CARDINAL(Get(symStream));
                IF number < 128 THEN
                    RETURN number
                ELSE
                    RETURN (number - 128) + 128 * GetNumber()
                END (* if *)
            END GetNumber;
        
    PROCEDURE GetCardinal(): CARDINAL;
        BEGIN
            GetSymbol();
            IF Sym = normalconstSS THEN
                RETURN GetNumber()
            ELSE
                RETURN 0
            END (* if *)
        END GetCardinal;
        
    PROCEDURE GetString;
        VAR
            i:      INTEGER;
        BEGIN
            FOR i := 0 TO MaxStringSize - 1 DO
                string[i] := CHAR(Get(symStream));
                IF string[i] = 0C THEN
                    RETURN
                END (* if *);
            END (* for *);
            Error(5)
        END GetString;
        
        
    PROCEDURE DecodeModule;
    
        VAR
        				(*  "klass" of the last declaration,
                                            or endunitSS.		  *)
            PrevDeclaration:    SymFileSymbols;
            
        PROCEDURE DecodeDeclaration;

            PROCEDURE DecodeQualIdent;
                BEGIN
                    UnGetSymbol();
                    ExpectIdentifier();
                    GetSymbol();
                    IF Sym = periodSS THEN
                        ExpectIdentifier();
                        GetSymbol();
                    END (* if *)
                END DecodeQualIdent;
            
            PROCEDURE DecodeConstant;
                VAR
                    i:              CARDINAL;
                BEGIN
                    CASE Sym OF
                      stringconstSS:
                        UnGetSymbol();
                        ExpectString();
                        GetSymbol()
                    | realconstSS:
                        IF BITSPERWORD = 32 THEN
                            dumcard := GetNumber();
                            dumcard := GetNumber();
                        ELSE
                            ReportError(TRUE, "Not 32 bits per word !!!",0,0);
                        END (* if *);
                        GetSymbol()
                    | normalconstSS:
                        UnGetSymbol();
                        dumcard := ExpectCardinal();
                        GetSymbol();
                        DecodeQualIdent();
                    ELSE
                        Error(6)
                    END (* case *)
                END DecodeConstant;
                
                
            PROCEDURE DecodeType;
            
                PROCEDURE DecodeVariants;
                    VAR
                        variantval:     CARDINAL;
                    BEGIN
                        IF Sym = caseSS THEN
                            ExpectSymbol(colonSS);
                            GetSymbol();
                            DecodeType();
                            IF Sym # ofSS THEN
                                Error(7)
                            END (* if *);
                            		(*  Variant fields.		  *)
                            REPEAT
                                variantval := GetCardinal();
                                IF Sym = normalconstSS THEN
                                    LOOP
                                        variantval := GetCardinal();
                                        IF Sym # normalconstSS THEN
                                            EXIT
                                        END (* if *);
                                    END (* loop *)
                                END (* if *);
                                IF Sym # colonSS THEN
                                    Error(8)
                                END (* if *);
                                GetSymbol();
                                DecodeVariants();
                                UnGetSymbol();
                                dumcard := ExpectCardinal();
                                GetSymbol()
                            UNTIL Sym # ofSS;
                            		(*  ELSE ..			  *)
                            IF Sym = elseSS THEN
                                GetSymbol();
                                DecodeVariants();
                                UnGetSymbol();
                                dumcard := ExpectCardinal();
                                GetSymbol()
                            END (* if *);
                            IF Sym # endSS THEN
                                Error(9)
                            END (* if *);
                            GetSymbol()
                        END (* if *);
                    END DecodeVariants;

            
                PROCEDURE CheckVariantField();
                    VAR
                        i, n, m: CARDINAL;
                    BEGIN
                      IF symfilekey >= 120B THEN
                        n := ExpectCardinal();
                        IF n # 0 THEN
                            m := ExpectCardinal();
                            FOR i := 1 TO m DO 
                                dumcard := ExpectCardinal ();
                            END (* for *);
                        END (* if *);
                      END (* if *);
                    END CheckVariantField;
            
            				(*  Start of DecodeType.	  *)
                VAR
                    ord:    CARDINAL;
                BEGIN
                    CASE Sym OF
                      identSS:
   					(*  Equivalent types.		  *)
                        DecodeQualIdent()
                    | lparentSS:
                    			(*  Enumerated types.		  *)
                        ExpectIdentifier();
                        REPEAT
                            ord := ExpectCardinal();
                            GetIdentifier();
                        UNTIL Sym # identSS;
                        IF Sym # rparentSS THEN
                            Error(10)
                        END (* if *);
                        GetSymbol()
                    | lbracketSS:
                    			(*  Subrange types.		  *)
                        GetSymbol();
                        DecodeConstant();
                        IF Sym # rangeSS THEN
                            Error(11)
                        END (* if *);
                        GetSymbol();
                        DecodeConstant();
                        IF Sym # rbracketSS THEN
                            Error(12)
                        END (* if *);
                        GetSymbol()
                    | pointertypSS:
                        		(*  POINTER TO .. types		  *)
                        GetSymbol();
                        DecodeType();
                    | hiddentypSS:
                        GetSymbol()
                    | settypSS:
                        GetSymbol();
                        DecodeType();
                    | arraytypSS:
                        GetSymbol();
                        DecodeType();
                        IF Sym = ofSS THEN
                            GetSymbol();
                            DecodeType();
                        END (* if *)
                    | recordtypSS:
                    			(*  RECORD .. types.		  *)
                        GetIdentifier();
                        		(*  Fixed fields.		  *)
                        WHILE Sym = identSS DO
                            dumcard := ExpectCardinal();
                            CheckVariantField();
                            ExpectSymbol(colonSS);
                            GetSymbol();
                            DecodeType();  
                            UnGetSymbol();
                            GetIdentifier()
                        END (* while *);
                        		(*  Variant field groups.	  *)
                        DecodeVariants();
                        IF Sym # endSS THEN
                            Error(13)
                        END (* if *);
                        dumcard := ExpectCardinal();
                        GetSymbol();
                    | procSS:
                    			(*  Procedure and Function types. *)
                        ExpectSymbol(lparentSS);
                        GetSymbol();
                        IF Sym # rparentSS THEN
                            LOOP
                                IF Sym = varSS THEN
                                    GetSymbol()
                                END (* if *);
                                DecodeType();
                                IF Sym = rparentSS THEN
                                    EXIT
                                END (* if *);
                            END (* loop *);
                        END (* if *);
                        GetSymbol();
                        IF Sym = colonSS THEN
                            GetSymbol();
                            DecodeType()
                        END (* if *);
                        
                    ELSE    (* corruption *)
                        Error(14);
                    END (* case *);
                END DecodeType;
                
        
        
        				(*  Start of DecodeDeclaration.	  *)
            VAR
                declsym:        SymFileSymbols;
                newdeclclass:   BOOLEAN;
                procnum,
                vaddr:          CARDINAL;
            BEGIN
                declsym := Sym;
                newdeclclass := PrevDeclaration # Sym;
                			(*  Display each alternative
                                            declaration.		  *)
                CASE Sym OF
                  constSS:
                    ExpectIdentifier();
                    GetSymbol();
                    DecodeConstant()
                | typSS:
                    ExpectIdentifier();
                    GetSymbol();
                    DecodeType()
                | varSS:
                    ExpectIdentifier();
                    GetSymbol();
                    IF Sym = lbracketSS THEN
                        vaddr := ExpectCardinal();
                        ExpectSymbol(rbracketSS);
                    ELSE
                        UnGetSymbol();
                        vaddr := ExpectCardinal();
                    END (* if *);
                    ExpectSymbol(colonSS);
                    GetSymbol();
                    DecodeType()
                | procSS:
                    ExpectIdentifier();
                    procnum := ExpectCardinal();
                    ProcTable [procnum] := Strings.ConcatSC (modname, Ident);
    IF Strings.LengthS (ProcTable [procnum]) > maxsymlength THEN
        Strings.Dispose (ProcTable [procnum]);
        ProcTable [procnum] := ModuleObjectNumber (modname, procnum);
    END;

                    GetSymbol();
                    DecodeType();
                
                ELSE
                    Error(16);
                END (* case *);
                			(*  Remember what sort of declaration
                                            was just processed, except when
                                            this was the final one.	  *)
                IF Sym = endunitSS THEN
                    PrevDeclaration := endunitSS
                ELSE
                    PrevDeclaration := declsym
                END (* if *)
            END DecodeDeclaration;
            
 					(*  Start of DecodeModule.	  *)
            
        VAR
            i:      CARDINAL;
        BEGIN
            IF Sym # unitSS THEN
                Error(17)
            END (* if *);
        				(*  Display module name and key.  *)
            FOR i := 1 TO 3 DO
                dumcard := GetCardinal()
            END (* for *);
            ExpectIdentifier();
            GetSymbol();
            IF Sym # endunitSS THEN
                IF Sym # importSS THEN
                    Error(18)
                END (* if *);
        				(*  Display "imports" list.	  *)
                ExpectIdentifier();
                REPEAT
                    GetIdentifier()
                UNTIL Sym # identSS;
        				(*  Display the exports list.	  *)
                GetIdentifier();
                WHILE Sym = identSS DO
                    GetIdentifier()
                END (* while *);
                    			(*  Display declarations for all
                                            relevant identifiers.	  *)
                WHILE Sym # endunitSS DO
                    DecodeDeclaration()
                END (* while *);
            END (* if *);
            GetSymbol()
        END DecodeModule;
        
        				(*  Start of DecodeSymbolFile.	  *)
    VAR
        i:              CARDINAL;
        symfilekey:     CARDINAL;
    BEGIN
        symfilekey := ExpectCardinal();
        FOR i := 1 TO 3 DO
           dumcard := ExpectCardinal()
        END (* for *);
        ExpectIdentifier();
        GetSymbol();
        WHILE Sym # endfileSS DO
            DecodeModule();
        END (* while *);
    END DecodeSymbolFile;
    
PROCEDURE SetProcTable (mn : ARRAY OF CHAR);
    VAR i : CARDINAL;
        temp : String;
    BEGIN
        Strings.CopyCC (mn, ModuleName);
        symStream := CreateInput (mn, "sym");
bufdata := BUFFERDATA (symStream^.instanceData);
        FOR i := 0 TO maxprocs - 1 DO  
            IF ProcTable[i] # NIL THEN Strings.Dispose (ProcTable[i]) END
        END;
        IF symStream^.status # Success THEN RETURN END;
        temp := Strings.CopyCS (mn);
        modname := Strings.ConcatCS (underline, temp);
        Strings.Dispose (temp);
        temp := Strings.ConcatSC (modname, underline);
        ProcTable [0] := Strings.ConcatSS (temp, modname);
        Strings.Dispose (temp);
        temp := Strings.ConcatSC (modname, underline);
        Strings.Dispose (modname);
        modname := temp;
        DecodeSymbolFile();
        Strings.Dispose (modname);
        Delete (symStream);
    END SetProcTable;
    
VAR i : CARDINAL;
BEGIN
    FOR i := 0 TO maxprocs - 1 DO  
        ProcTable[i] := NIL
    END;
    underline [0] := "_";
    underline [1] := 0C;
END ReadSym.
