IMPLEMENTATION MODULE StackItems;
FROM UtilError IMPORT ReportError;
FROM items IMPORT ITEM, MODE, REGISTER, InDataReg, PackItem,
                  OnRealStack, address, PackRealStackItem,
		  UsesSlavedAddrReg, Operator, BytesPerWord;
FROM CodeProcs IMPORT next;
FROM Procs IMPORT GBParmSpace, LinkSpace;
FROM PsStk IMPORT PseudStk, top, DoAnyPendingOp, push, SetDyadicOp,
                  DyadicOpPending, pop, Pop, DecTopOfPsStk, Push, IncTop,
                  NotOnPsStk, PsStkState, STATE, STATEBIT;
FROM RegSlv IMPORT Slave, slave, FreeReg, UseBitSet,
		   PurgeAfterStackStore;
FROM GenProcs IMPORT GenSubImm, GenMove, GenMovea, GenLea,
                     GenMoveImm;
FROM AddrReg IMPORT InAddrRegSlave, IncUseCount, DecUseCount,
                    GetAddrOnPsStkIntoUseableForm, FreeAddrReg,
                    ProtectAddress, UnprotectAddress,
                    PutIntoAddrRegSlave;
FROM MCMnemonics IMPORT
    LSA, LSWn, LSW, LSD, LSD0, TS, SSWn, SSW, SSD, SSD0, SLW,
    SLWn, SEW, SGW, SGWn;


PROCEDURE StackLoad (Instruction : CARDINAL);
BEGIN
    IF Instruction = LSA THEN
        LoadStackAddress (next ())
    ELSIF Instruction = TS THEN
        TestAndSet ()
    ELSE
        CASE Instruction OF
          LSW: LoadStackWord (next ())
        | LSD: LoadStackDoubleWord (next ())
        | LSD0: LoadStackDoubleWord (0)
        ELSE (* LSWn+0..LSWn+15 *)
            LoadStackWord (Instruction MOD 16)
        END;
    END;
END StackLoad;


PROCEDURE StackStore (Instruction : CARDINAL);
BEGIN
    CASE Instruction OF
      SSW: StoreStackWord (next ())
    | SSD: StoreStackDoubleWord (next ())
    | SSD0: StoreStackDoubleWord (0)
    ELSE (* SSWn+0..SSWn+15 *)
        StoreStackWord (Instruction MOD 16)
    END
END StackStore;


PROCEDURE LoadStackAddress (disp : CARDINAL);
    VAR item : ITEM;
        nextinstruc : CARDINAL;
         AddrReg : REGISTER;
    BEGIN
        DoAnyPendingOp ();
        item := PseudStk [top];
        nextinstruc := next ();
        IF address (item) THEN
            INC (item.DispOrData, disp);
            PseudStk [top] := item
        ELSIF (* nextinstruction is a simple store and so there is
                 no point in getting an ADDRaddrRegDisp value on 
                 PseudStk (as happens in entry sequence to module *)
              (nextinstruc = SLW) OR
              (nextinstruc = SEW) OR
              ((SLWn+4 <= nextinstruc) AND (nextinstruc <= SLWn+15)) OR
              (nextinstruc = SGW) OR
              ((SGWn+2 <= nextinstruc) AND (nextinstruc <= SGWn+15))
        THEN
            Push (Immediate, NoReg, disp * BytesPerWord);
            SetDyadicOp (Uadd);
        ELSE
            DecTopOfPsStk (1);
            IF NOT InAddrRegSlave (item, AddrReg) THEN
                AddrReg := FreeAddrReg (NotOnPsStk);
                GenMovea (item, AddrReg);
                PutIntoAddrRegSlave (AddrReg, item);
            END (* if *);
            Push (ADDRaddrRegDisp, AddrReg, disp);
            IncUseCount (AddrReg, top);
        END (* if *);
    END LoadStackAddress;

PROCEDURE LoadStackWord (disp : CARDINAL);
    BEGIN 
        LoadStackItem (disp, FALSE);
    END LoadStackWord;

PROCEDURE LoadStackDoubleWord (disp : CARDINAL);
    BEGIN
        LoadStackItem (disp, TRUE);
    END LoadStackDoubleWord;

PROCEDURE LoadStackItem (disp : CARDINAL; Double : BOOLEAN);
    VAR item : ITEM;
        AddrReg : REGISTER;
    BEGIN
        DoAnyPendingOp ();
        GetAddrOnPsStkIntoUseableForm (top, FALSE);
        item := PseudStk [top];
        INC (item.DispOrData, disp);
        item.Mode := AddrRegDisp;
        PseudStk [top] := item;
        IF Double THEN
            INC (item.DispOrData); IncTop;
            IF UsesSlavedAddrReg (item) THEN
                IncUseCount (item.Register, top)
            END (* if *);
            PseudStk [top] := item
        END (* if *);
    END LoadStackItem;

PROCEDURE StoreStackWord (disp : CARDINAL);
    BEGIN 
        StoreStackItem (disp, FALSE) 
    END StoreStackWord;

PROCEDURE StoreStackDoubleWord (disp : CARDINAL);
    BEGIN 
        StoreStackItem (disp, TRUE)
    END StoreStackDoubleWord;

PROCEDURE StoreStackItem (disp : CARDINAL; Double : BOOLEAN);
    VAR point : CARDINAL;
        Address : ITEM;
        AddrReg : REGISTER;
        GBStackStore : BOOLEAN;
    BEGIN 
        IF DyadicOpPending THEN
            point := top - 2
        ELSE
            point := top - 1
        END (* if *);
        IF Double THEN
            DEC (point)
        END (* if *);
        GetAddrOnPsStkIntoUseableForm (point, FALSE);
        Address := PseudStk [point];
        ProtectAddress (Address);
        GBStackStore := GB IN PsStkState [point];
        IF GBStackStore THEN
	    IF INTEGER (disp) - 4 >= GBParmSpace [point] THEN
                DEC (Address.DispOrData, INTEGER (disp)
		   		- 3 - GBParmSpace [point])
	    ELSE INC (Address.DispOrData, - INTEGER (disp)
				 + 3 + GBParmSpace [point] + LinkSpace)
	    END;
        ELSE
            INC (Address.DispOrData, disp);
        END (* if *);
        Address.Mode := AddrRegDisp;
        IF Double THEN
            IF GBStackStore THEN
                DEC (Address.DispOrData);
                pop (Address); 
                INC (Address.DispOrData);
            ELSE
                INC (Address.DispOrData);
                pop (Address); 
                DEC (Address.DispOrData);
            END (* if *);
        END (* if *);
        pop (Address);
        UnprotectAddress (Address);
        DecTopOfPsStk (1);
    END StoreStackItem;


PROCEDURE TestAndSet ();
    VAR Address, Reg : ITEM;
    BEGIN
        DoAnyPendingOp ();
        GetAddrOnPsStkIntoUseableForm (top, FALSE);
        Address := PseudStk [top];
        Address.Mode := AddrRegDisp;
        FreeReg (Reg, NOT UseBitSet);
        GenMove (Address, Reg);
        GenMoveImm (1, Address);
        PurgeAfterStackStore (Address);
        DecTopOfPsStk (1);
        push (Reg);
    END TestAndSet;
	        

END StackItems.


