(*
    Title: 	UnixObj - Implementation
    LastEdit:	"Mon Sep 17 11:35:15 1984"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE UnixObj;

FROM SYSTEM IMPORT ADDRESS;
FROM SystemTypes IMPORT LongCARDINAL;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM Streams IMPORT Put, PutN, Seek, SeekMode, Stream, Tell;
IMPORT BinaryIO;
IMPORT HashStrings, Strings, Slist, SystemTypes;

CONST bigseg = 100000;

VAR size, relocsize: CARDINAL;
    datasize, textsize: CARDINAL;
    textrelocsize: CARDINAL;
    binStream : Stream;

PROCEDURE SetObjStream (s : Stream);
    BEGIN
        binStream := s;
    END SetObjStream;

PROCEDURE WriteCARDINAL (s : Stream; c: CARDINAL);
    BEGIN
        PutLong (c);
    END WriteCARDINAL;

PROCEDURE PutHeader(magic: CARDINAL);
    VAR
        i: CARDINAL;
    BEGIN
        WriteCARDINAL(binStream, magic);
        FOR i := 1 TO 7 DO
            WriteCARDINAL(binStream, 0);
        END (* for *);
        size := 0;
    END PutHeader;

PROCEDURE PutByte(b: BYTE);
    BEGIN
        Put(binStream, b);
        INC(size, 1);
    END PutByte;

PROCEDURE PutLong(w: CARDINAL);
(*  no guarantee of word alignment, so output byte at a time if necessary *)
    VAR
        high, low: CARDINAL;
    BEGIN
        high := w DIV 65536; low := w MOD 65536;
        Put(binStream, high DIV 256);
        Put(binStream, high MOD 256);
        Put(binStream, low DIV 256);
        Put(binStream, low MOD 256);
        INC(size, 4);
    END PutLong;

PROCEDURE PutSeg (add : ADDRESS; bytes : CARDINAL);
    VAR i : CARDINAL;
    BEGIN
        i := PutN (binStream, add, bytes);
        INC (size, bytes);
    END PutSeg;

PROCEDURE Flush(udatasize: CARDINAL);
    VAR p : LongCARDINAL;
	i : CARDINAL;
    BEGIN
        PutRelocs;
        PutSymbols;
        Tell (binStream, p);
	FOR i := 1 TO (4 - p) MOD 4 DO
	    Put (binStream,  0)
	END;
        Seek(binStream, Beginning, 4);
        WriteCARDINAL(binStream, textsize);
        WriteCARDINAL(binStream, datasize);
        WriteCARDINAL(binStream, udatasize);
        WriteCARDINAL(binStream, symbolcount*12);
        WriteCARDINAL(binStream, 0);
        WriteCARDINAL(binStream, textrelocsize);
        WriteCARDINAL(binStream, relocsize);
    END Flush;

PROCEDURE SizeNow(): CARDINAL;
    BEGIN
        RETURN size;
    END SizeNow;

PROCEDURE EndText;
    BEGIN
        textsize := size;
        size := 0;
    END EndText;

PROCEDURE EndTextReloc;
    BEGIN
        datasize := size;
        PutRelocs();
        textrelocsize := relocsize;
        relocsize := 0;
    END EndTextReloc;

TYPE RELOCPTR = POINTER TO RELOC;
    RELOC = RECORD
            link: RELOCPTR;
            address, symb: CARDINAL;
        END (* record *);

VAR RelocList: RELOCPTR;

PROCEDURE AddRelocation(address, symb: CARDINAL; extern: BOOLEAN);
    CONST
        externbit = {4};
        long = {6};
    VAR
        reloc, preloc, xreloc: RELOCPTR;
        searching: BOOLEAN;
    BEGIN
        NEW(reloc);
        symb := CARDINAL(BITSET(symb * 256) + long);
	IF extern THEN
            symb := CARDINAL(BITSET(symb) + externbit);
	    END;
        reloc^.address := address;
        reloc^.symb := symb;
        xreloc := RelocList; preloc := RelocList; searching := TRUE;
        WHILE searching AND (xreloc # NIL) DO
            IF address < xreloc^.address THEN
                searching := FALSE;
            ELSE
                preloc := xreloc;
                xreloc := xreloc^.link;
            END (* if *);
        END (* while *);
        reloc^.link := xreloc;
        IF xreloc=RelocList THEN
            RelocList := reloc
        ELSE
            preloc^.link := reloc;
        END (* if *);
        INC(relocsize, 8);
    END AddRelocation;

PROCEDURE PutRelocs;
    VAR
        xr: RELOCPTR;
    BEGIN
        WHILE RelocList # NIL DO
            WriteCARDINAL(binStream, RelocList^.address);
            WriteCARDINAL(binStream, RelocList^.symb);
            xr := RelocList;
            RelocList := RelocList^.link;
            DISPOSE(xr);
        END (* while *);
    END PutRelocs;

TYPE SYMBOLENTRYPTR = POINTER TO SYMBOLENTRY;
TYPE SYMBOLENTRY = RECORD
        link: SYMBOLENTRYPTR;
        hashid: HashStrings.HashId;
        ntype, value: CARDINAL;
        id: CARDINAL;
        namelength: CARDINAL;
    END (* record *);

VAR
    symbolcount: CARDINAL;
    symbols: HashStrings.HashTable;
    symbolentries: SYMBOLENTRYPTR;
    stringtableindex: CARDINAL;

PROCEDURE AddSymbol(VAR name: ARRAY OF CHAR; xntype, xvalue: CARDINAL): CARDINAL;
    VAR
        xhashid: HashStrings.HashId;
        symbolentry: SYMBOLENTRYPTR;
    BEGIN
        IF HashStrings.EnterChars(symbols, name, xhashid) THEN
            NEW(symbolentry);
            WITH symbolentry^ DO
                ntype := xntype;
                hashid := xhashid;
                value := xvalue;
                id := symbolcount;
                namelength := Strings.LengthC(name)+1;
            END (* with *);
            INC(symbolcount);
            Slist.AddE(symbolentries, symbolentry);
            HashStrings.Assoc(symbols, xhashid, symbolentry);
        ELSE
            symbolentry := SYMBOLENTRYPTR( HashStrings.Retrieve(symbols, xhashid) );
        END (* if *);
        RETURN symbolentry^.id;
    END AddSymbol;

PROCEDURE PutSymbols;
    VAR
        xs, s: SYMBOLENTRYPTR;
        name: SystemTypes.String;
        i: CARDINAL;
    BEGIN
        s := symbolentries;
        WHILE s # NIL DO
            WITH s^ DO
                WriteCARDINAL(binStream, stringtableindex);
                INC(stringtableindex, namelength);
                Put(binStream, ntype);
                Put(binStream, 0C);Put(binStream, 0C);Put(binStream, 0C);
                WriteCARDINAL(binStream, value);
                s := link;
            END (* with *);
        END (* while *);

        WriteCARDINAL(binStream, stringtableindex);
        s := symbolentries;
        WHILE s # NIL DO
            WITH s^ DO
                name := HashStrings.StringAt(symbols, hashid);
                i := PutN(binStream, name, namelength);
                xs := s;
                s := link;
                DISPOSE(xs);
            END (* with *);
        END (* while *);
    END PutSymbols;

BEGIN
    size := 0;
    relocsize := 0; RelocList := NIL;
    symbolcount := 0; symbolentries := NIL;
    stringtableindex := 4;
    symbols := HashStrings.New(50);
END UnixObj.
