(*
    Title: 	Linker - implementation
    LastEdit:	"Mon Sep 22 15:08:31 1986"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory

    VAX/Unix version (new code-generator)
*)

IMPLEMENTATION MODULE Linker;
FROM SYSTEM IMPORT TSIZE;
FROM Streams IMPORT Stream, Delete;
FROM FileStream IMPORT CreateOutput;
FROM LModule IMPORT
    MODULEPTR, MODULENAME, modulelist, Pname, InitTable, InitTableUpb;
FROM Options IMPORT StackSize (*,HeapSize*);
FROM UnixLd IMPORT Ld;
FROM WriteF IMPORT
    WriteF0, WriteF1, WriteF2, H, S;
FROM UtilError IMPORT ReportError;

FROM UnixMCI IMPORT GFTableSize, LoadedModule;

IMPORT Config;
IMPORT Strings;
FROM SystemTypes IMPORT String;

TYPE NTYPE = (modname, filename);
TYPE GFTCARD = [0 .. GFTableSize];

VAR mcount: GFTCARD;
    sgft: Stream;

PROCEDURE Link;
VAR l,i: CARDINAL;
    mp: MODULEPTR;
    mainmodule, gftfile: ARRAY [0 .. 39] OF CHAR;
    pname: String;
    UnixMCIpresent : BOOLEAN;
BEGIN
(*  
	This code does not put out a global frame table and other goodies
	for the debugger because there is no debugger with which to test
	When (if ?) the debugger is brought up this code should be brought into
	line with the VAX linker
*)
    UnixMCIpresent := FALSE;
    Pname(modulelist^.itself^.name, pname);
    Strings.CopySC(pname, mainmodule);
    Strings.CopyCC(mainmodule, gftfile);
    Strings.ConcatCC(gftfile, ".gfs", gftfile);
    sgft := CreateOutput(gftfile);

    DecGlobal ("spbase");
    DecGlobal ("_errno");
    DecGlobal ("_etext");
    DecGlobal ("_environ");
    DecGlobal ("_run");
    DecGlobal ("start");

    mp := modulelist;
    l := 0; mcount := 0;
    WHILE mp # NIL DO
        Pname(mp^.itself^.name, pname);
        IF Strings.EqualCS ("UnixMCI", pname) THEN UnixMCIpresent := TRUE END;
        WriteF2 (sgft, "\t.globl _%s__%s\n", S(pname), S(pname));
	OutputGdef(pname, l);
	INC(l, 4);
	INC(mcount);
	IF mcount = GFTableSize-1 THEN 
	    ReportError(TRUE, "too many modules!", 0, 0);
        END;
	mp := mp^.link;
    END;

    pname := Strings.CopyCS("SYSTEM");
    OutputGdef(pname, l); INC(mcount);

    InsS(".text", "  ");

    (*  Tidy up to get rid of absolute addresses UnixMCI *)

    IF UnixMCIpresent THEN
        WriteF0 (sgft, "\tlea\t_UnixMCI__d, a4\n");
        WriteF0 (sgft, "\tmovl\ta3, a4@(12)\n");
        WriteF0 (sgft, "\tmovl\ta3, a4@(16)\n");
        WriteF0 (sgft, "\tmovl\ta3, a4@(20)\n");
        WriteF0 (sgft, "\tlea \t_errno, a0\n");
        WriteF0 (sgft, "\tmovl\ta0, a4@(24)\n");
	WriteF0 (sgft, "\tlea \t_etext, a0\n");
        WriteF0 (sgft, "\tmovl\ta0, a4@(28)\n");
        WriteF0 (sgft, "\tmovl\tsp, a4@(32)\n");
        WriteF0 (sgft, "\tlea\t_environ, a0\n");
        WriteF0 (sgft, "\tmovl\ta0, a4@(36)\n");
        WriteF0 (sgft, "\tmovl\ta3, a4@(40)\n");
    END;
    WriteF0(sgft, "\tjmp\t_run\n");
    
    WriteF2(sgft, "start: .long _%s__%s\n", H(mainmodule), H(mainmodule));    
    (* here starts the program proper 
    WriteF0(sgft, "\tjmp\t_run\n");
 *)   WriteF0(sgft, "\t.data\nspbase:\t.long 0\n_environ:\t.long 0\n");
(*    FOR i := 0 TO InitTableUpb-1 DO
        Pname(InitTable[i]^.itself^.name, pname);
        WriteF2(sgft, "\tcalls\t$0, %S__%S\n", S(pname), S(pname));
    END (* for *);
    WriteF0(sgft, "jmp _m2exit\n");

    OutputLoadedModuleList;
*)
    Delete(sgft);
    ForceLnk;
    Ld(mainmodule);

END Link;

PROCEDURE DecGlobal (s: ARRAY OF CHAR);
    BEGIN
	WriteF1(sgft, "\t.globl\t%s\n", H(s));
    END DecGlobal;

PROCEDURE InsC(s: ARRAY OF CHAR; c: INTEGER);
    BEGIN
        WriteF2(sgft, "\t%S\t%I\n", H(s), c);
    END InsC;

PROCEDURE InsS(s: ARRAY OF CHAR; arg: ARRAY OF CHAR);
    BEGIN
        WriteF2(sgft, "\t%S\t%S\n", H(s), H(arg));
    END InsS;

PROCEDURE OutputGdef(pname: String; l: CARDINAL);
    BEGIN
        (* WriteF1(sgft, "\t.long %S__d\n", S(pname)); *)
    END OutputGdef;

PROCEDURE ForceLnk;
VAR
    mp: MODULEPTR;
BEGIN
    mp := modulelist;
    WHILE mp # NIL DO
        Lnkfname(mp^.fname);
        mp := mp^.link;
    END (* while *);
END ForceLnk;

PROCEDURE OutputLoadedModuleList;
TYPE
    CListPair = RECORD module, impl: String END;
VAR
    mp: MODULEPTR;
    cname: String;
    cListPair: ARRAY GFTCARD OF CListPair;
    cL: GFTCARD;
    i: CARDINAL;
    pname: String;

    PROCEDURE OutputAsciz(module: String; ch: CHAR; textName: String);
    BEGIN
        WriteF2(sgft, '%S_%C:', S(module), ch);
	WriteF1(sgft, '\t.asciz "%S"\n', S(textName));
    END OutputAsciz;


BEGIN
    mp := modulelist; cL := 0;
    WriteF2(sgft, "\t.globl _m2gft_%C\n_m2gft_%C:\n", 'l', 'l');
    WHILE mp # NIL DO
      WITH mp^ DO
        Pname(itself^.name, pname);
        WriteF2(sgft, "\t.long %S_%C\n", S(pname), 'm');
        IF Config.Specified(itself^.name, cListPair[cL].impl) THEN
            WriteF2(sgft, "\t.long %S_%C\n", S(pname), 'f');
            cListPair[cL].module := pname;
            INC(cL);
        ELSE
            WriteF0(sgft, "\t.long 0x80000000\n");
        END (* if *);
        WriteF2(sgft, "\t.long %S__%C\n", S(pname), 'v');
	mp := link
      END;
    END;

    WriteF0(sgft, "\n.long SYSTEM_m, 0x80000000, SYSTEM__v\n");

    WriteF1(sgft, "\t.space %U0\n", 
             (GFTableSize-mcount)*TSIZE(LoadedModule)*4);

    WriteF0(sgft, ".text\n");
    
    mp := modulelist;
    WHILE mp # NIL DO
      WITH mp^ DO
        Pname(itself^.name, pname);
        OutputAsciz(pname, 'm', pname);
        mp := link;
      END;
    END;
    WriteF0(sgft, 'SYSTEM_m:\t.asciz "SYSTEM"\n');

    IF cL > 0 THEN
        FOR i := 0 TO cL-1 DO
            OutputAsciz(cListPair[i].module, 'f', cListPair[i].impl);
        END (* for *);
    END (* if *);

END OutputLoadedModuleList;

PROCEDURE Lnkfname(VAR fname: ARRAY OF CHAR);
(*  force .lnk extension on to 'fname'.
*)
VAR l: CARDINAL;
BEGIN    
        l := Strings.LengthC(fname);  
        WHILE (l > 0) AND (fname[l] # '/') DO
            DEC(l)
        END;
        
        IF l > 3 THEN
	    DEC(l, 3);
    	    fname[l] := 'l'; 	(* change directory to 'lnk' *)
	    fname[l+1] := 'n';
	    fname[l+2] := 'k';
        ELSE
            HALT;
        END;
END Lnkfname;

END Linker .
