(*
    Title: 	Unix ld interface - Sun OS 4 implementation
    LastEdit:	"19 Feb 1990" By Ian Leslie
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory
    
*)

IMPLEMENTATION MODULE UnixLd;

FROM SYSTEM IMPORT ADR;
FROM Storage IMPORT ALLOCATE;
FROM UnixProcess IMPORT Exec, Wait, Fork;
FROM UnixIOXtra IMPORT Unlink;
FROM Streams IMPORT Stream, ErrorCode, Delete;
FROM WriteF IMPORT WriteF1, WriteF2, H;
IMPORT Strings;
IMPORT Slist;
FROM SystemTypes IMPORT String;
IMPORT M2Files;
FROM LModule IMPORT MODULEPTR, modulelist, Pname;
FROM Linker IMPORT Lnkfname;
FROM Options IMPORT Debugging, Echo, DeleteGfs, OutFile, 
    LdArgs, LDARGSPTR, LdLibArgs, LDLIBARGSPTR, NoLink;
FROM UtilError IMPORT ReportError;
FROM UnixRun IMPORT
    RUNErrorCode, RUNERROR, RUNMESSAGE, Run, RunMessage;
FROM Assembly IMPORT AssembleX;
FROM IsArchive IMPORT IsArchive, ArList, ARLIST;

CONST MaxArgs = 300;

VAR
    argv: ARRAY [0 .. MaxArgs] OF CARDINAL;
    ld, lib : ARRAY [0 .. 15] OF CHAR;
    oopt, nopt: ARRAY [0 .. 2] OF CHAR;
    aslib: ARRAY [0 .. 80] OF CHAR;

PROCEDURE Ld(mainmodule: ARRAY OF CHAR);
VAR outfile, gfofile: ARRAY [0 .. 80] OF CHAR;
    mp: MODULEPTR;
    rc: INTEGER;
    argcount: CARDINAL;
    pname: String;

    PROCEDURE NextArg(arg: CARDINAL);
        BEGIN
            IF argcount > HIGH(argv) THEN
                ReportError(TRUE, "'ld' argument array overflow", 0,0);
            END (* if *);
            argv[argcount] := arg;
            INC(argcount);
        END NextArg;
BEGIN
    Pname(modulelist^.itself^.name, pname);
    IF OutFile[0] = 0C THEN
        Strings.CopyCC(mainmodule, OutFile);
        Strings.ConcatCC(OutFile, ".out", OutFile);
    END;
    Strings.CopyCC(mainmodule, gfofile);
    Strings.ConcatCC(gfofile, ".gfo", gfofile);
    IF NOT NoLink THEN AssembleX(mainmodule, DeleteGfs, "gfs", "gfo"); END;
    argcount := 0;
    NextArg(ADR(ld));
    NextArg(ADR(oopt));
    NextArg(ADR(OutFile));
    NextArg(ADR(nopt));
    WHILE LdArgs # NIL DO
        NextArg(CARDINAL(LdArgs^.arg));
        LdArgs := LdArgs^.link;
    END (* while *);

    NextArg(ADR(gfofile));
    Lookup("SYSTEM",  aslib);
    NextArg(ADR(aslib));

    mp := modulelist;
    WHILE mp # NIL DO
        IF NOT IsArchive(mp^.fname) THEN
            NextArg(ADR(mp^.fname));
        END (* if *);
	mp := mp^.link;
    END;

    WHILE ArList # NIL DO
        NextArg(CARDINAL(ArList^.archive));
        ArList := ArList^.link;
    END (* while *);

    WHILE LdLibArgs # NIL DO
        NextArg(CARDINAL(LdLibArgs^.arg));
        LdLibArgs := LdLibArgs^.link;
    END (* while *);
    NextArg(ADR(lib));
    NextArg(0);
    ExecLd;
    rc := Unlink(gfofile);
END Ld;

PROCEDURE ExecLd;
VAR RunError: RUNERROR;
    RunM: RUNMESSAGE;
BEGIN
    IF Echo THEN
        Run("/bin/echo", argv, RunError);
    END;
    IF NoLink THEN RETURN END;
    Run("/bin/ld", argv, RunError);
    
    IF RunError.code # ok THEN
        RunMessage(RunError, RunM);
        ReportError(TRUE, "ld %S - %I", H(RunM), RunError.errno);
    END;
END ExecLd;

PROCEDURE Lookup(file: ARRAY OF CHAR; VAR filename: ARRAY OF CHAR);
    VAR
        stream: Stream;
        l: CARDINAL;
    BEGIN
        stream := M2Files.CreateInput(file, "key");
        IF stream^.status # Success THEN
            ReportError(TRUE, "failed to open '%S' module", H(file), 0);
        END;
        Delete(stream);

        Strings.CopyCC(M2Files.Filename, filename);
        Lnkfname(filename);
        l := Strings.LengthC(filename);
        IF IsArchive(filename) THEN   
            (* found in archive, wont do, cos it must load first *)
            WHILE filename[l] # '/' DO
                DEC(l);
            END (* while *);
            filename[l+1] := 0C;
            Strings.ConcatCC(filename, file, filename);
        END (* if *);
   END Lookup;

BEGIN
    ld := "/bin/ld"; nopt := "-n"; oopt := "-o"; lib := "-lc";
END UnixLd .

