IMPLEMENTATION MODULE P1IOSym;

    FROM NewStreams IMPORT
      STREAM, Connect, Reset, ReadChar, Disconnect;
    FROM WriteStrings IMPORT WriteString, WriteLn;
(*    FROM Conversions IMPORT ConvertCardinal;*)
    FROM FileLookup IMPORT LookupFile;
    FROM Storage IMPORT ALLOCATE;
    FROM MCBase IMPORT Symbol;
    FROM MCPublic IMPORT
      Compilerstatus, compstat, symfileextension;
    FROM MCSymFileDefs IMPORT symFileKey, SymFileSymbols;
    FROM P1IOOut IMPORT PutS, StopSave, RestartSave, Error;
    FROM P1IOId IMPORT InIdTab, OutIdTab, EnterId;
    FROM P1IOStr IMPORT PutStrCh, InitString, TermString;
    FROM MCP1IO IMPORT sy, val, length, spix;
    FROM MCP1IOPV IMPORT ch;

    CONST maxload = 25; (* maximal number of loaded symbolfiles *)

    TYPE Byte = [0..377B];

    VAR symstream : STREAM;
        loadsym : ARRAY [1 .. maxload] OF CARDINAL;
                  (* list of symbolfiles already loaded *)
        maxindex, index : CARDINAL;
(**)    newsymfile : BOOLEAN;

    PROCEDURE ReadSym(VAR b: Byte);
    VAR ch: CHAR;
    BEGIN
        ReadChar(symstream, ch);
	b := Byte(ch);
    END ReadSym;
    
    PROCEDURE ReadSymNumber(): CARDINAL;
    VAR ch: CHAR;
        b: CARDINAL;
    BEGIN
        ReadChar(symstream, ch);
	b := CARDINAL(ch);
        IF b < 128 THEN
            RETURN b;
        ELSE
            RETURN (ReadSymNumber() * 128) + CARDINAL(b) - 128;
        END;
    END ReadSymNumber;

    PROCEDURE SymGetSy;
      (* get symbol from symstream *)
      CONST rwordnum = 2; (* number of words for a real number *)
      VAR b : Byte;
          ix : CARDINAL;
          rconv : RECORD
                    CASE BOOLEAN OF
                      FALSE : ra : ARRAY [1..rwordnum] OF CARDINAL;
                     |TRUE : rc : REAL;
                    END;
                  END;
    BEGIN
      ReadSym(b);
      IF newsymfile THEN
        (* convert symbols *)
        CASE VAL(SymFileSymbols,b) OF
          endfileSS : sy := eop;
        | unitSS : sy := symbolsy;
        | endunitSS : sy := endblock;
        | importSS : sy := importsy;
        | exportSS : sy := qualifiedsy;
        | constSS : sy := constsy;
        | normalconstSS :
            sy := cardcon;
	    val.value := ReadSymNumber();
        | realconstSS :
            sy := realcon;
            FOR ix := 1 TO rwordnum DO
	      rconv.ra[ix] := ReadSymNumber();
            END;
            NEW(val.rvalue);
            val.rvalue^ := rconv.rc;
        | stringconstSS :
            sy := stringcon;
            InitString;
            ReadSym(b);
            WHILE b <> 0 DO PutStrCh(CHR(b)); ReadSym(b) END;
            TermString(length,val);
        | typSS : sy := typesy;
        | arraytypSS : sy := arraysy;
        | recordtypSS : sy := recordsy;
        | settypSS : sy := setsy;
        | pointertypSS : sy := pointersy;
        | hiddentypSS : sy := hidden;
        | varSS : sy := varsy;
        | procSS, funcSS : sy := proceduresy;
        | identSS :
            sy := ident;
            ReadSym(b);
            WHILE b <> 0 DO ch := CHR(b); InIdTab; ReadSym(b); END;
            EnterId;
        | periodSS : sy := period;
        | colonSS : sy := colon;
        | rangeSS : sy := range;
        | lparentSS : sy := lparent;
        | rparentSS : sy := rparent;
        | lbracketSS : sy := lbrack;
        | rbracketSS : sy := rbrack;
        | caseSS : sy := casesy;
        | ofSS : sy := ofsy;
        | elseSS : sy := elsesy;
        | endSS : sy := endsy;
        END; (* CASE *)
      ELSE
        sy := VAL(Symbol,b);
        IF sy = ident THEN
          ReadSym(b);
          WHILE b <> 0 DO ch := CHR(b); InIdTab; ReadSym(b); END;
          EnterId;
        ELSIF sy = cardcon THEN
          val.value := ReadSymNumber();
        ELSIF sy = realcon THEN (* accept rwordnum words *)
          FOR ix := 1 TO rwordnum DO
            rconv.ra[ix] := ReadSymNumber();
          END;
          NEW(val.rvalue);
          val.rvalue^ := rconv.rc;
        ELSIF sy = stringcon THEN
          InitString;
          ReadSym(b);
          WHILE b <> 0 DO PutStrCh(CHR(b)); ReadSym(b) END;
          TermString(length,val);
        END;
      END;
    END SymGetSy;
 
    PROCEDURE SymFileOK(): BOOLEAN;
      VAR b : Byte;
          key : CARDINAL;
    BEGIN
      ReadSym(b);
      IF b = ORD(cardcon) THEN
        newsymfile := FALSE;
	key := ReadSymNumber();
        RETURN key = symFileKey - 1;
      ELSIF b = ORD(normalconstSS) THEN
        newsymfile := TRUE;
	key := ReadSymNumber();
        RETURN key = symFileKey;
      ELSE
        RETURN FALSE;
      END;
    END SymFileOK;

    PROCEDURE GetSeparateModule;
      CONST strmax = 24;
      VAR name : ARRAY [0 .. strmax-1] OF CHAR;
          dummyname : ARRAY [0..1] OF CHAR;
          lastinputspix : CARDINAL;
          lastinputch : CHAR;
          lastinputsy : Symbol;
          pos : CARDINAL;
          okfile : BOOLEAN;
          exitcond : BOOLEAN;
          temFile : CARDINAL;
    BEGIN
      lastinputch := ch;
      lastinputspix := spix; 
      lastinputsy := sy; 
      (* test on already loaded symbolfile *)
      index := 1;
      WHILE (index <= maxindex) AND (loadsym[index] <> spix) DO
        INC(index);
      END;
      IF index > maxindex THEN (* new symbolfile *)
        IF maxindex < maxload THEN
          INC(maxindex);
          loadsym[maxindex] := spix;
        END;
        pos := 0;
        OutIdTab;
        WHILE (pos < strmax) AND (ch <> ' ') DO
          name[pos] := ch;
          INC(pos);
          OutIdTab;
        END;
        IF pos < strmax THEN name[pos] := 0C END;
        REPEAT
          exitcond := TRUE;
          LookupFile(name,symfileextension,temFile,
                     querys IN compstat,TRUE,FALSE,
                     dummyname,okfile);
          IF okfile THEN
            Connect(symstream,temFile,FALSE);
            Reset(symstream); 
            IF SymFileOK() THEN
              SymGetSy; SymGetSy; SymGetSy; (* module key *)
              SymGetSy;
              IF (sy = ident) AND (spix = lastinputspix) THEN (* copy file *)
                StopSave;  
                SymGetSy;
                WHILE sy <> eop DO PutS; SymGetSy END;
                RestartSave; 
              ELSE
                WriteString("  ---- incorrect module name");
                WriteLn;
                IF querys IN compstat THEN exitcond := FALSE;
                ELSE INCL(compstat,symerrs);
                END;
              END;
            ELSE
              WriteString("  ---- wrong symbol file");
              WriteLn;
              IF querys IN compstat THEN exitcond := FALSE;
              ELSE INCL(compstat,symerrs);
              END;
            END;
            Disconnect(symstream,TRUE); 
          ELSE
            INCL(compstat,symerrs); (*symbolfiles incomplete*)
          END;
        UNTIL exitcond;
      END;
      ch := lastinputch;
      sy := lastinputsy; 
      spix := lastinputspix; 
    END GetSeparateModule; 

  BEGIN (* SymFileInput *)
    maxindex := 0; (* no symbolfile loaded *)

END P1IOSym .

