IMPLEMENTATION MODULE P4Case;

FROM SYSTEM IMPORT WORD;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM Exceptions IMPORT StdException;
FROM MCBase IMPORT
    Symbol, intptr;
FROM MCP4AttributSys IMPORT
    Attribut, TestBaseType, Load;
FROM MCP4Global IMPORT
    GetSymbol, sy, Error, val,
    controlRangeCheck, loadCount, spPosition;
FROM MCP4CodeSys IMPORT
    Emit, EmitLI, Emit2, Emit2GivePC, Insert2, Fixup2, PC;
FROM MCP4ExpressionSy IMPORT
    Expression;
FROM P4Stat IMPORT StatSeq1, StatSeq3;

CONST
    ENTC=302B; EXC=303B; TRAP=304B;

  PROCEDURE CaseStatement;
    (* the case expression is assumed to have integer type *)

    TYPE
      caseptr = POINTER TO caserec;
      caserec =
        RECORD
          cval: INTEGER;
          casePC: INTEGER;
          next: caseptr
        END;

    VAR
      elsePC, tabPC: INTEGER;
      IntegerSort : BOOLEAN;
      cHeader: caseptr; (* header of cicular ordered list *)
      lat: Attribut;

    PROCEDURE GreaterThan (big, little : WORD) : BOOLEAN;
    BEGIN
        IF IntegerSort THEN
            RETURN INTEGER (big) > INTEGER (little);
        ELSE
            RETURN CARDINAL (big) > CARDINAL (little);
        END;
    END GreaterThan;

    PROCEDURE CaseLabels;
      VAR cp: caseptr; lval, i: INTEGER;

      PROCEDURE EnterCaseLabel(VAR fcp: caseptr; fval: INTEGER);
        (* build new element in linear list of fcp with value fval *)
        VAR lp: caseptr;
      BEGIN
        NEW(lp);
        WITH lp^ DO
          cval := fval;
          next := fcp; fcp := lp
        END
      END EnterCaseLabel;

      PROCEDURE MarkCase(fcp: caseptr);
        (* the linear list from fcp is sorted
           and inserted into the list of cHeader *)
        VAR ival: INTEGER;
            currcp, lp, lp2: caseptr;
      BEGIN
        WHILE fcp<>NIL DO
          WITH fcp^ DO (*take currcp out of the linear list*)
            currcp := fcp; casePC := PC; fcp := next;
          END;
          (*include currcp in the circular list*)
          ival := currcp^.cval;
          cHeader^.cval := ival; (*garantees terminating of insert loop*)
          lp2 := cHeader;        (*predecessor*)
          lp := cHeader^.next;   (*successor*)
          WHILE GreaterThan (ival, lp^.cval) DO
            lp2 := lp; lp := lp^.next
          END;
          IF (lp<>cHeader) AND (lp^.cval=ival) THEN Error(223) END;
          currcp^.next := lp;
          lp2^.next := currcp
        END
      END MarkCase;

    BEGIN (*CaseLabels*)
      cp := NIL;
      REPEAT EnterCaseLabel(cp, val); lval := val; GetSymbol;
        IF sy = range THEN GetSymbol;
          FOR i := 1 TO INTEGER(val-lval) DO
            INC(lval);
            EnterCaseLabel(cp, lval)
          END;
          GetSymbol;
        END;
      UNTIL sy = colon; GetSymbol;
      MarkCase(cp)
    END CaseLabels;

    PROCEDURE Emit4 (i : WORD);
    BEGIN
        Emit2 (CARDINAL (i) DIV 65536); Emit2 (CARDINAL (i) MOD 65536);
    END Emit4;

    PROCEDURE Insert4 (at : CARDINAL; i : WORD);
    BEGIN
        Insert2 (at, CARDINAL (i) DIV 65536);
        Insert2 (at+2, CARDINAL (i) MOD 65536);
    END Insert4;

    PROCEDURE CaseTable;
      VAR
        cp, cp2: caseptr;
        upPC: INTEGER;
        upper, i: INTEGER;
    BEGIN
      cp := cHeader^.next;
      IF cp=cHeader THEN (* empty cases *)
        Emit4(1); Emit4(0); Emit2(elsePC-PC)
      ELSE
        WITH cp^ DO
          upper := cval;
          Emit4(upper);         (*is lower bound now*)
          Emit2GivePC(upPC);    (*upper bound*)
          Emit2(0);
          Emit2(elsePC-PC);     (*else case outside range*)
          Emit2(casePC-PC);
          cp2 := cp; cp := next;
          DISPOSE(cp2)
        END;
        WHILE cp<>cHeader DO
          WITH cp^ DO
            FOR i := 2 TO ABS (INTEGER(cval)-INTEGER (upper)) DO
              (* upper+1 TO cval-1 *)
              Emit2(elsePC-PC) (*jump to else*)
            END;
            upper := cval;
            Emit2(casePC-PC); (*jump belonging to cval*)
            cp2 := cp; cp := next; DISPOSE(cp2)
          END
        END; (*while*)
        Insert4(upPC, upper)
      END (*if*)
    END CaseTable;

  BEGIN (*CaseStatement*)
    NEW(cHeader); (*header of empty cicular ordered list*)
    WITH cHeader^ DO next := cHeader; cval := 0 END;
    Expression(lat);
    IF TestBaseType (lat.typtr) = intptr THEN
      IntegerSort := TRUE;
    ELSE
      IntegerSort := FALSE;
    END;
    Load(lat); DEC(loadCount);
    Emit(ENTC);
    Emit2GivePC(tabPC); INC(spPosition);
    WHILE sy = ofsy DO GetSymbol;
      CaseLabels;
      StatSeq3(ofsy, elsesy, endsy);
      Emit(EXC)
    END;
    elsePC := PC;
    IF sy = elsesy THEN GetSymbol;
      StatSeq1(endsy); Emit(EXC)
    ELSIF controlRangeCheck THEN
      (* Emit a trap *)
      EmitLI(ORD (CaseIndexOutOfRange)); DEC(loadCount); Emit(TRAP)
    END;
    GetSymbol;
    Fixup2(tabPC);
    CaseTable; DEC(spPosition);
    DISPOSE(cHeader)
  END CaseStatement;

END P4Case .
