(*
  Title:	ArgHandler - Unix
  $Header: ArgHandler.mod,v 1.5 88/02/26 16:13:35 dm Exp $
  		Copyright (C) 1984 by Acorn Research Centre
*)

IMPLEMENTATION MODULE ArgHandler;
FROM UtilError IMPORT ReportError;
FROM WriteF IMPORT H, WriteF1;
FROM UnixArgs IMPORT argCount, GetArg;
FROM Streams IMPORT Stream, ErrorCode, Delete;
FROM SysStreams IMPORT sysOut;
FROM FileStream IMPORT CreateInput;
FROM M2Files IMPORT SetM2PathFromFile;
FROM TextIO IMPORT ReadChars, WhiteSpace, ReadCHAR;
FROM CharCodes IMPORT NewLineCh;

FROM GenMakeFile IMPORT makeFilename;
FROM GenGraph IMPORT verbose;
FROM GenDirs IMPORT AddDirectoryToParseIn;

IMPORT AppendUser, Version;

PROCEDURE Usage ();
  BEGIN
    ReportError(TRUE, 
      "usage: [-v -V -E -p fileofdirs -f makefile -a appendfile]\n", 0,0);
  END Usage;

PROCEDURE ParseArguments ();
  VAR 
    arg, fileOfDirsName: ARRAY [0..255] OF CHAR;
    i: INTEGER;
  BEGIN
    IF argCount > 1 THEN
      FOR i := 1 TO argCount - 1 DO
        GetArg (i, arg);
        IF arg [0] = "-" THEN
          CASE arg[1] OF
            'V':
            verbose := TRUE
            
          | 'E':
            SetM2PathFromFile;
            
          | 'v':
            WriteF1(sysOut, "m2make V%S\n", H(Version.idString));
            
          | 'p':
            INC(i);
            IF i = INTEGER(argCount) THEN
              Usage()
            ELSE
              GetArg(i, fileOfDirsName);
            END;
            Scan(fileOfDirsName);
            
          | 'f':
            INC(i);
            IF i = INTEGER(argCount) THEN
              Usage()
            ELSE
              GetArg(i, makeFilename);
            END;

          | 'a':
            INC(i);
            IF i = INTEGER(argCount) THEN
              Usage()
            ELSE
              GetArg(i, AppendUser.fileName);
              AppendUser.noAppend := FALSE
            END;
          ELSE
            Usage();
          END;
        ELSE Usage ()
        END;
      END
    END;
  END ParseArguments;

PROCEDURE Scan(name: ARRAY OF CHAR);
  VAR
    s: Stream;
    directory: ARRAY [0..255] OF CHAR;
    ch: CHAR;
    isTarget: BOOLEAN;
  BEGIN
    s := CreateInput(name);
    IF s^.status # Success THEN
      ReportError(TRUE, "failed to open %S", H(name), 0);
    ELSE
      LOOP
        directory[0] := 0C;
        ReadChars(s, directory, WhiteSpace);
        IF s^.status = EndOfStream THEN EXIT END;
        (* look for '*' *)
        REPEAT
          ch := ReadCHAR(s);
        UNTIL (ch = NewLineCh) OR (ch = '*') OR (s^.status # Success);
        IF ch = '*' THEN
          isTarget := TRUE;
          REPEAT
            ch := ReadCHAR(s);
          UNTIL (ch = NewLineCh) OR (s^.status # Success);
        ELSE
          isTarget := FALSE;
        END;
        AddDirectoryToParseIn(directory, isTarget);
      END;
      Delete(s);
    END;
  END Scan;

END ArgHandler.

(*
$Log:	ArgHandler.mod,v $
Revision 1.5  88/02/26  16:13:35  dm
Added the -a appendfilename flag

Revision 1.4  85/01/04  10:35:09  mjj
Add support for '*' in directory file

Revision 1.3  85/01/04  09:39:15  mjj
Added -E option to set M2PATH from file

Revision 1.2  84/11/30  16:51:30  mjj
-V for verbose, -v for version identification

Revision 1.1  84/11/30  15:53:45  mjj
Initial revision

*)

