(*
  Title:	GenDirs - Unix Implementation
  $Header: GenDirs.mod,v 1.3 85/01/04 10:36:39 mjj Exp $
  		Copyright (C) 1984 by Acorn Research Centre
*)
IMPLEMENTATION MODULE GenDirs;

FROM UnixEnv IMPORT EnvRc, GetEnvVar;
FROM Storage IMPORT ALLOCATE;
IMPORT Strings;
IMPORT Slist;
FROM UtilError IMPORT ReportError;
FROM WriteF IMPORT S;
FROM Dir IMPORT IsCurrentDirectoryName;

FROM GenGraph IMPORT directoryList, Directory;

TYPE
  PDirList = POINTER TO PDir;
  PDir = RECORD
      next: PDirList;
      name: Strings.String;
      isTarget: BOOLEAN;
    END;

VAR
  pDirList: PDirList;

PROCEDURE Build;
  VAR
    rc: EnvRc;
    m2path: ARRAY [0..255] OF CHAR;
    m2dir: ARRAY [0..79] OF CHAR;
    firstIndex, lastIndex: CARDINAL;
    endvar: BOOLEAN;
    dir: Directory;
  BEGIN
    rc := GetEnvVar("M2PATH", m2path);
    IF rc # Found THEN
      ReportError(TRUE, "no M2PATH environment variable",0,0);
    ELSE
      lastIndex := 0; firstIndex := 0;
      LOOP
        endvar := Strings.FindC(m2path, ':', lastIndex);
        IF lastIndex=firstIndex THEN EXIT END;
        (* extract first directory to search *)
        Strings.ExtractCC(m2path, firstIndex, lastIndex, m2dir);
        NEW(dir);
        dir^.scanned := FALSE;
        dir^.pathName := Strings.CopyCS(m2dir);
        dir^.uniqId := -1;
        dir^.isCurrent := IsCurrentDirectoryName(dir^.pathName);
        dir^.target := dir^.isCurrent;
        dir^.toParse := dir^.target;
        
        Slist.AddE(directoryList, dir);
        IF m2path[lastIndex] = 0C THEN
          EXIT
        ELSE
          INC(lastIndex);
          firstIndex :=  lastIndex;
        END;
      END;
      CheckDirectories;
    END;
  END Build;

PROCEDURE AddDirectoryToParseIn(cname: ARRAY OF CHAR; xisTarget: BOOLEAN);
  VAR
    p: PDirList;
  BEGIN
    NEW(p);
    WITH p^ DO
      name := Strings.CopyCS(cname);
      isTarget := xisTarget;
    END;
    Slist.Add(pDirList, p);
  END AddDirectoryToParseIn;
  
PROCEDURE CheckDirectories;
  VAR
    d: Directory;
    p: PDirList;
    found: BOOLEAN;
  BEGIN
    p := pDirList;
    WHILE p # NIL DO
      d := directoryList;
      found := FALSE;
      WHILE d # NIL DO
        IF Strings.EqualSS(p^.name, d^.pathName) THEN
          d^.toParse := TRUE;
          d^.target := p^.isTarget;
          found := TRUE;
          d := NIL;
        ELSE
          d := d^.next;
        END;
      END;
      IF NOT found THEN
        ReportError(TRUE, "directory %S not on M2PATH", S(p^.name), 0);
      END;
      p := p^.next;
    END;
  END CheckDirectories;

BEGIN
  pDirList := NIL;
END GenDirs.

(*
$Log:	GenDirs.mod,v $
Revision 1.3  85/01/04  10:36:39  mjj
Add 'isTarget' argument to 'AddDirectoryToParseIn' to allow
 directory to be marked for target generation.

Revision 1.2  85/01/04  09:45:06  mjj
Maintain separate list of directories to parse in, to allow
-E option, updating directoryList at the last moment.

Revision 1.1  84/11/30  16:03:28  mjj
Initial revision

*)

