(*
  Title:	M2Lex
  $Header: M2Lex.mod,v 1.2 84/12/04 12:38:52 mjj Exp $
  		Copyright (C) 1984 by Acorn Research Centre
*)

IMPLEMENTATION MODULE M2Lex;

IMPORT HashStrings;
FROM CharCodes IMPORT SpaceCh, NewLineCh, CReturnCh, TabCh, FormFeedCh;
FROM Streams IMPORT Delete, Stream, ErrorCode, Get, Backspace;
FROM FileStream IMPORT CreateInput;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;

TYPE LStream = POINTER TO
    RECORD
        stream: Stream;
        symbol: Symbol;     (* the current symbol *)
        lineNumber: CARDINAL;
    END;

CONST 
    EndStreamCh = 0C;			(*  returned by Rdch on EndOfStream *)

VAR symbolTable: HashStrings.HashTable;

PROCEDURE OpenStream (VAR s: LStream; name: ARRAY OF CHAR): BOOLEAN;
  VAR
    str: Stream;
BEGIN
    str := CreateInput(name);
    IF str^.status = Success THEN
        NEW (s);
        WITH s^ DO
            stream := str;
            symbol := Endfilesym;
            lineNumber := 1;
        END;
        RETURN TRUE;
    ELSE RETURN FALSE
    END;
END OpenStream;


PROCEDURE CloseStream (s: LStream);
BEGIN
    Delete (s^.stream);
    DISPOSE(s^.stream); DISPOSE (s);
END CloseStream;

PROCEDURE LineNumber(s: LStream): CARDINAL;
  BEGIN
    RETURN s^.lineNumber;
  END LineNumber;

PROCEDURE Rdch(s: LStream): CHAR;
  VAR
    ch: CHAR;
  BEGIN
    ch := CHAR(Get(s^.stream));
    IF ch = NewLineCh THEN
      INC(s^.lineNumber)
    END;
    RETURN ch;
  END Rdch;

PROCEDURE UnRdch(s: LStream);
  BEGIN
    Backspace(s^.stream);
  END UnRdch;
    
PROCEDURE Declare (word: ARRAY OF CHAR; symbol: Symbol);
VAR h: HashStrings.HashId;
BEGIN
    IF HashStrings.EnterChars (symbolTable, word, h) THEN
      HashStrings.Assoc(symbolTable, h, symbol);
    END;
END Declare;


PROCEDURE LookUp (word: ARRAY OF CHAR): Symbol;
VAR h: HashStrings.HashId;
BEGIN
    IF HashStrings.LookupChars (symbolTable, word, h) THEN
        RETURN Symbol (HashStrings.Retrieve(symbolTable, h))
    ELSE RETURN Identsym;
    END;
END LookUp;


PROCEDURE SkipComment (s: LStream);
BEGIN
    LOOP
        CASE Rdch(s) OF
          EndStreamCh:
            RETURN;
        | '*':
            IF Rdch (s) = ')' THEN RETURN ELSE UnRdch (s) END;
        | '(':
            IF Rdch (s) = '*' THEN SkipComment (s) ELSE UnRdch (s) END;
        ELSE
        END;
    END;
END SkipComment;


PROCEDURE ReadString (s: LStream; VAR a: ARRAY OF CHAR);
VAR i: CARDINAL;
BEGIN
    i := 0;
    LOOP
        a [i] := Rdch (s);
        CASE a [i] OF
          'a'..'z', 'A'..'Z', '0'..'9': 
            IF i < HIGH (a) THEN INC (i) END;
        ELSE UnRdch (s); 
            a [i] := 0C; 
            RETURN
        END
    END;
END ReadString;


PROCEDURE NextSymbol (s: LStream; VAR id: ARRAY OF CHAR): Symbol;
BEGIN
    LOOP
        CASE Rdch (s) OF
          SpaceCh, TabCh, NewLineCh, FormFeedCh:
        | EndStreamCh:
            s^.symbol := Endfilesym;
            EXIT
        | ';':
            s^.symbol := Semicolonsym;
            EXIT
        | ',':
            s^.symbol := Commasym;
            EXIT
        | '(':
            IF Rdch (s) = '*' THEN
                SkipComment (s);
            ELSE UnRdch (s);
                s^.symbol := Othersym;
                EXIT
            END;
        | 'A' .. 'Z', 'a' .. 'z':
            UnRdch (s);
            ReadString (s, id);
            s^.symbol := LookUp (id);
            EXIT
        ELSE s^.symbol := Othersym;
            EXIT
        END;
    END;
    RETURN s^.symbol
END NextSymbol;


PROCEDURE Accept (s: LStream; sym: Symbol): BOOLEAN;
VAR id: ARRAY [1..80] OF CHAR;
BEGIN
    RETURN NextSymbol (s, id) = sym
END Accept;


PROCEDURE AcceptId (s: LStream; VAR id: ARRAY OF CHAR): BOOLEAN;
BEGIN
    RETURN NextSymbol (s, id) = Identsym
END AcceptId;


PROCEDURE Skip (s: LStream; sym: Symbol);
VAR id: ARRAY [1..1] OF CHAR;
    symbol: Symbol;
BEGIN
    symbol := s^.symbol;
    WHILE (symbol # sym) AND (symbol # Endfilesym) DO
        symbol := NextSymbol (s, id)
    END;
END Skip;


BEGIN
    symbolTable := HashStrings.New (16);
    Declare ("DEFINITION", Definitionsym);
    Declare ("FROM", Fromsym);
    Declare ("IMPLEMENTATION", Implementationsym);
    Declare ("IMPORT", Importsym);
    Declare ("MODULE", Modulesym);

END M2Lex.

(*
$Log:	M2Lex.mod,v $
Revision 1.2  84/12/04  12:38:52  mjj
Added PROCEDURE LineNumber for error messages

Revision 1.1  84/11/30  16:14:46  mjj
Initial revision

*)

