(*
    Title: 	Assembly - Implementation
    LastEdit:	"Tue Sep 18 10:35:13 1984"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE Assembly;

FROM SYSTEM IMPORT ADR;
FROM UnixIOXtra IMPORT Unlink;
IMPORT Strings;
FROM UnixRun IMPORT RUNERROR, RUNErrorCode, RUNMESSAGE, RunMessage, Run;
FROM WriteF IMPORT WriteF2, H;
FROM Stop IMPORT Stop;
FROM M2Files IMPORT Extension, CreateOutput, Filename;
FROM UtilError IMPORT ReportError;

PROCEDURE Assemble(module: ARRAY OF CHAR; DeleteAsfile: BOOLEAN);
    VAR
        sext: Extension;
    BEGIN
        sext[0] := 's'; sext[1] := 0C;
        AssembleX(module, DeleteAsfile, sext, "lnk");
    END Assemble;

PROCEDURE AssembleX(module: ARRAY OF CHAR; DeleteAsfile: BOOLEAN;
                   srcx, dstx: Extension);
(* Call the assembler to process the file 'module'.s     	      *)
(* /bin/as module.s -o 'objfile'  				      *)
VAR as: ARRAY [0 .. 7] OF CHAR;
    oopt: ARRAY [0 .. 2] OF CHAR;
    dot: ARRAY [0 ..1] OF CHAR;
    argv: ARRAY [0 .. 4] OF CARDINAL;
    asfile, objfile: ARRAY [0 .. 39] OF CHAR;
    RunM: RUNMESSAGE;
    RunError: RUNERROR;
    reply: INTEGER;
    i: CARDINAL;
BEGIN
    oopt := "-o"; dot[0] := '.'; dot[1] := 0C;
    i:=0;
    LOOP				(* truncate a trailing extension *)
        IF (i > HIGH(module)) OR (module[i] = 0C) THEN EXIT
        ELSIF module[i] = '.' THEN module[i] := 0C; EXIT;
	ELSE INC(i);
        END;
    END;
    
    Strings.CopyCC(module, asfile);
    Strings.ConcatCC(asfile, dot, asfile);
    Strings.ConcatCC(asfile, srcx, asfile);
    Strings.CopyCC(module, objfile);

    IF Strings.EqualCC(dstx, "lnk") THEN
        Strings.ConcatCC(".lnk/", objfile, objfile);
    END (* if *);
    Strings.ConcatCC(objfile, dot, objfile);
    Strings.ConcatCC(objfile, dstx, objfile);
    
    argv[1] := ADR(asfile);
    argv[2] := ADR(oopt);
    argv[3] := ADR(objfile);
    argv[4]  := 0;
    Run("/bin/as", argv, RunError);
    IF RunError.code # ok THEN
        RunMessage(RunError, RunM);
        ReportError(TRUE, "%S - %I\n", H(RunM), RunError.errno);
    END;
    IF DeleteAsfile THEN reply:=Unlink(asfile) END;
END AssembleX;



BEGIN
END Assembly .
