
(*
    Title: 	Read and decode and M-code file - Implementation
    LastEdit:	"Mon Feb  7 13:30:30 1983"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory
*)


IMPLEMENTATION MODULE ReadMcd;

FROM Storage IMPORT ALLOCATE;
FROM Streams IMPORT Stream, Get, ErrorCode;

IMPORT Slist;
FROM McdFrame IMPORT
    XMODULE, XIMPORT, STRINGTEXT, CODETEXT, 
    DATATEXT, FIXUP, VERSION, ENDMOD;
FROM UtilError IMPORT ReportError;

VAR Codeframe: POINTER TO ARRAY [0 .. Maxcode - 1] OF BYTE;
    framesize: CARDINAL;
    LastCard: CARDINAL;
    UnRdCard, Debug: BOOLEAN;
    McdStream : Stream;

PROCEDURE SetMcdStream (s : Stream);
BEGIN
    McdStream := s;
END SetMcdStream;

PROCEDURE Rdch () : CHAR;
BEGIN
    RETURN CHAR (Get (McdStream));
END Rdch;

PROCEDURE BinRdch () : CHAR;
BEGIN
    RETURN CHAR (Get (McdStream));
END BinRdch;

PROCEDURE ReadMCode(debug: BOOLEAN);
VAR next, i: CARDINAL;
BEGIN
    FixupSize := 0; CodeSize := 0; DataSize := 0;
    ProcadrSize :=  0; StringSize := 0;
    Importlist := NIL;  ImportlistLength := 0;
    StringConstantBase := 0;
    UnRdCard := FALSE;
    NEW(Headerframe);
    Debug := debug;
    McdError := McdOK;
    next := Lookatword();
    WHILE (next <> ENDMOD) AND (McdError = McdOK) DO
        IF (next < 200B) OR (next > 205B) THEN
            McdError := UnknownFrametype;
        ELSE
        Module;
        next := Lookatword();
        END;
    END;
END ReadMCode;

PROCEDURE SetCodeByte (ByteNo, Op : CARDINAL);
BEGIN
    Codeframe^ [ByteNo] := BYTE (Op);
END SetCodeByte;

PROCEDURE CodeByte (ByteNo: CARDINAL) : CARDINAL;
BEGIN
    RETURN CARDINAL (Codeframe^ [ByteNo]);
END CodeByte;

PROCEDURE Module;
BEGIN
    IF Optional(VERSION) THEN VersionFrame; END;
    HeaderFrame;
    IF McdError = McdOK THEN
        IF Optional(XIMPORT) THEN ImportFrame; END;

        LOOP
            Frametype := Lookatword();
            CASE Frametype OF
                CODETEXT:
                    CodeFrame;
                    IF Optional(FIXUP) THEN FixupFrame; END;

            |   DATATEXT:
                    DataFrame;

            |   STRINGTEXT:
                    StringFrame;

            |   ENDMOD:
                    EXIT;

            ELSE    McdError := UnknownFrametype;
            END;
            IF McdError <> McdOK THEN EXIT; END;
        END;
    END;
END Module;

PROCEDURE VersionFrame;
VAR i: CARDINAL;
BEGIN
    Frametype := RdCard();
    framesize := RdCard();
    IF Debug THEN Trace("VersionFrame %N",  framesize); END;
    BytesperWord := RdCard();
    FOR i:=0 TO HIGH(Version) DO Version[i] := Rdch(); END;
END VersionFrame;

PROCEDURE HeaderFrame;
BEGIN
    IF Debug THEN Trace("HeaderFrame", 0); END;
    Frametype := RdCard();
    IF Frametype = XMODULE THEN
        framesize := RdCard();
        ModuleNameandKey(Headerframe);
        Headerframe^.datasize := RdCard();
    ELSE    
        McdError := ModuleframeMissing;
    END;
END HeaderFrame;

PROCEDURE ImportFrame;
VAR i: CARDINAL;
VAR ip: IMPORTFRAMEPTR;
BEGIN
    Frametype := RdCard();
    framesize := RdCard();
    IF Debug THEN Trace("ImportFrame %N", framesize); END;
    FOR i := 1 TO framesize BY MODULENAMESIZE + KEYSIZE DO
        NEW(ip);
        ModuleNameandKey(ip);
	INC(ImportlistLength);
        Slist.AddE(Importlist, ip);
    END;
END ImportFrame;

PROCEDURE CodeFrame;
VAR byteoffset, thisCodeSize, i: CARDINAL;
BEGIN
    Frametype := RdCard();
        framesize := RdCard();
        IF Debug THEN Trace("CodeFrame %N", framesize); END;
        byteoffset := RdCard();
        thisCodeSize := byteoffset + framesize-1;
        IF thisCodeSize > HIGH(Codeframe^)+1 THEN
            McdError := CodeframeOverflow;
        ELSE
            IF thisCodeSize > CodeSize THEN
                CodeSize := thisCodeSize;
            END;
            (*  store code in actual place *)
            FOR i := 0 TO framesize-2 DO
                Codeframe^[byteoffset+i] := RdByte();
            END;
            IF framesize=3 THEN
                IF byteoffset >= ProcadrSize THEN
                    ProcadrSize := byteoffset+2;
                END;
            END;
        END;
END CodeFrame;

PROCEDURE StringFrame;
VAR wordoffset, byteoffset, thisStringMax, i: CARDINAL;
BEGIN
    Frametype := RdCard();
        framesize := RdCard();
        IF Debug THEN Trace("StringFrame %N", framesize); END;
        wordoffset := RdCard();
        IF StringConstantBase = 0 THEN
            StringConstantBase := wordoffset;
        END;
(* The M-code compiler calculates wordoffset according to BytesperWord.
*)
        thisStringMax := (wordoffset-StringConstantBase)*BytesperWord+
                          framesize-1;
        IF thisStringMax > HIGH(Stringframe)+1 THEN
            McdError := StringframeOverflow;
        ELSE
            IF thisStringMax > StringSize THEN
                StringSize := thisStringMax;
            END;
            byteoffset := (wordoffset-StringConstantBase)*BytesperWord;
            FOR i := 0 TO framesize-2 DO
                Stringframe[byteoffset+i] := Rdch()
            END;
        END;
END StringFrame;

PROCEDURE DataFrame;
VAR wordoffset, thisDataSize, i: CARDINAL;
BEGIN
    Frametype := RdCard();
        framesize := RdCard();
        IF Debug THEN Trace("DataFrame %N", framesize); END;
        wordoffset := RdCard();
        thisDataSize := wordoffset + framesize-1;
        IF thisDataSize > HIGH(Dataframe)+1 THEN
            McdError := DataframeOverflow;
        ELSE
            IF thisDataSize > DataSize THEN
                DataSize := thisDataSize;
            END;
            (*  store data in actual position *)
            FOR i := 0 TO framesize-2 DO
                Dataframe[wordoffset+i] := RdCard();
            END;
        END;
END DataFrame;

PROCEDURE FixupFrame;
VAR i: CARDINAL;
BEGIN
    Frametype := RdCard();
    framesize := RdCard();
    IF Debug THEN Trace("FixupFrame %N", framesize); END;
    IF framesize+FixupSize > HIGH(Fixupframe)+1 THEN
        McdError := FixupframeOverflow;
    ELSE
        FOR i := 1 TO framesize DO
            Fixupframe[i-1+FixupSize] := RdCard();
        END;
        INC(FixupSize, framesize);
    END;
END FixupFrame;

PROCEDURE ModuleNameandKey(ip: IMPORTFRAMEPTR);
VAR i: CARDINAL;
BEGIN
WITH ip^ DO
    FOR i := 0 TO MODULENAMESIZE-1 DO
        name[i] := Rdch();
    END;

    FOR i := 1 TO HIGH(key) DO
        key[i] := RdCard();
    END;
END;
END ModuleNameandKey;

PROCEDURE Optional(type: CARDINAL): BOOLEAN;
BEGIN
    Frametype := Lookatword();
    IF Frametype = type THEN RETURN TRUE ELSE RETURN FALSE END;
END Optional;

PROCEDURE Lookatword(): CARDINAL;
BEGIN
    LastCard := RdCard();
    IF McdStream^.status = EndOfStream THEN 
        LastCard := ENDMOD;
    ELSE
        UnRdCard := TRUE;
    END;
    RETURN LastCard;
END Lookatword;

PROCEDURE RdByte(): BYTE;
BEGIN
    RETURN BYTE( RdCard() );
END RdByte;

PROCEDURE RdCard(): CARDINAL;
BEGIN
    IF UnRdCard THEN 
        UnRdCard := FALSE;
        RETURN LastCard;
    ELSE
        LastCard := GetBytes();
(*        IF Debug THEN Trace("RdCard=%O6", LastCard); END;  *)
        RETURN LastCard;
    END;
END RdCard;

PROCEDURE Trace(str: ARRAY OF CHAR; b: CARDINAL);
BEGIN
    ReportError (FALSE, str, b, 0);
END Trace;

PROCEDURE GetBytes(): CARDINAL;
VAR b: CARDINAL;
BEGIN
    b := CARDINAL( BinRdch() );
    IF b < 128 THEN
        RETURN b;
    ELSE
        RETURN (GetBytes() * 128) + b - 128;
    END;
END GetBytes;


(*                   ReadMcd initialisation	                      *)
VAR i: CARDINAL;
BEGIN
NEW (Codeframe);
FOR i := 0 TO Maxcode - 1 DO Codeframe^[i] := 0 END;
END ReadMcd .

