IMPLEMENTATION MODULE TTIO ;	(* Ch. Jacobi	15.12.79 *)
				(* modified	 5.03.80 *)
				(* S. McKenzie	19.05.81 for UNIX *)
				(* M. J. Jordan 5.02.82 bug in ReadInt *)
FROM SYSTEM IMPORT WORD, MAXINT, ADR;
FROM SysStreams IMPORT (* procs *) sysOut, sysIn;
FROM Streams IMPORT GetN, Put;
CONST
    INBUFLENGTH = 80;	(* length of input buffer *)
    TAB = 11C;
    NL  = 12C;
VAR
    inputstream : RECORD
		    index : CARDINAL;
		    size  : CARDINAL;
		    mustagain : BOOLEAN;
		    buf   : ARRAY [0..INBUFLENGTH-1] OF CHAR
		  END;

PROCEDURE ReadAgain;
BEGIN
    inputstream.mustagain := TRUE
END ReadAgain;

PROCEDURE Read(VAR ch : CHAR);
BEGIN
    WITH inputstream DO
	IF mustagain THEN
	    ch := buf[index - 1];	(* Assert index > 0 *)
	    mustagain := FALSE
	ELSE
	    IF index >= size THEN
		ReadLine
	    END;
	    ch := buf[index];
	    INC(index)
	END
    END
END Read;

PROCEDURE ReadInt(VAR integ : INTEGER; VAR nextch : CHAR; VAR isnum : BOOLEAN);
CONST
    SAFELIMITDIV10 = MAXINT DIV 10;
    SAFELIMITREM10 = MAXINT MOD 10;
    
VAR
    int : INTEGER;
    ch  : CHAR;
    neg : BOOLEAN;
    safedigit: [0 .. 9];
    chvalue: CARDINAL;
BEGIN
    Read(ch);
    WHILE (ch = ' ') OR (ch = TAB) DO
	Read(ch)
    END;
    IF ch = '-' THEN
	neg := TRUE;
	Read(ch)
    ELSIF ch = '+' THEN
	neg := FALSE;
	Read(ch)
    ELSE
	neg := FALSE
    END;

    safedigit := SAFELIMITREM10;
    IF neg THEN safedigit := safedigit + 1 END;
    int := 0;
    IF (ch >= '0') & (ch <= '9') THEN
	WHILE (ch >= '0') & (ch <= '9') DO
  	    chvalue := ORD(ch) - ORD('0');
	    IF (int > SAFELIMITDIV10) OR 
		((int = SAFELIMITDIV10) AND (chvalue > safedigit)) THEN
		WriteString(" Error : integer overflow", 0); WriteLn;
		HALT
	    ELSE
		int := 10*int + VAL(INTEGER, chvalue);
		Read(ch)
	    END;
	END;
	IF neg THEN
   	    integ := -int
	ELSE
	    integ := int
	END;
	isnum := TRUE
    ELSE
	isnum := FALSE
    END;
    nextch := ch
END ReadInt;

PROCEDURE ReadLine;
VAR
    result : INTEGER;
BEGIN
    WITH inputstream DO
	IF index > size THEN
	    WriteString(" Error : tried to read past eof(input)", 0);
	    HALT
	ELSE
	    result := GetN (sysIn, ADR (buf), INBUFLENGTH);
	    IF result < 0 THEN
		WriteString(" Error in standard input", 0);
		HALT
	    ELSE
		IF result = 0 THEN
		    EOF := TRUE
		END;
		size := result;
		index := 0;
	    END
	END
    END
END ReadLine;

PROCEDURE OutputNum(digits : ARRAY OF CHAR; numwidth, fieldwidth : CARDINAL);
BEGIN
    WHILE fieldwidth > numwidth DO
	Write(' ');
	DEC(fieldwidth)
    END;
    REPEAT
	DEC(numwidth);
	Write(digits[numwidth])
    UNTIL numwidth = 0
END OutputNum;

PROCEDURE Write(ch : CHAR);
BEGIN
    Put (sysOut, ch);
END Write;

PROCEDURE WriteCard(card, width : CARDINAL);
VAR
    buf : ARRAY [0..11] OF CHAR;
    i : CARDINAL;
BEGIN
    i := 0;
    REPEAT
	buf[i] := CHAR(card MOD 10 + ORD('0'));
	card := card DIV 10;
	INC(i)
    UNTIL card = 0;
    OutputNum(buf, i, width)
END WriteCard;

PROCEDURE WriteInt(int : INTEGER; width : CARDINAL);
VAR
    buf : ARRAY [0..11] OF CHAR;
    i : CARDINAL;
    neg : BOOLEAN;
BEGIN
    IF int < 0 THEN
	neg := TRUE;
	int := -int
    ELSE
	neg := FALSE
    END;
    i := 0;
    REPEAT
	buf[i] := CHAR(CARDINAL(int) MOD 10 + ORD('0'));
	int := int DIV 10;
	INC(i)
    UNTIL int = 0;
    IF neg THEN
	buf[i] := '-';
	INC(i)
    END;
    OutputNum(buf, i, width)
END WriteInt;

PROCEDURE WriteHex(word : WORD; width : CARDINAL);
VAR
    buf : ARRAY [0..7] OF CHAR;
    card, digit, i : CARDINAL;
BEGIN
    card := CARDINAL(word);
    i := 0;
    REPEAT
	digit := card MOD 16;
	IF digit < 10 THEN
	    buf[i] := CHAR(digit + ORD('0'))
	ELSE
	    buf[i] := CHAR(digit - 10 + ORD('A'))
	END;
	card := card DIV 16;
	INC(i)
    UNTIL i = 4;
    OutputNum(buf, i, width)
END WriteHex;

PROCEDURE WriteLn;
BEGIN
    Write(NL)
END WriteLn;

PROCEDURE WriteOct(word : WORD; width : CARDINAL);
VAR
    buf : ARRAY [0..11] OF CHAR;
    card, i : CARDINAL;
BEGIN
    card := CARDINAL(word);
    i := 0;
    REPEAT
	buf[i] := CHAR(card MOD 8 + ORD('0'));
	card := card DIV 8;
	INC(i)
    UNTIL i = 6;
    OutputNum(buf, i, width)
END WriteOct;

PROCEDURE WriteString(str : ARRAY OF CHAR; width : CARDINAL);
VAR
    nbytes : CARDINAL;
BEGIN
    nbytes := 0;
    WHILE (nbytes <= HIGH(str)) AND (str[nbytes] <> 0C) DO
	Write(str[nbytes]);
	INC(nbytes)
    END;
    WHILE width > nbytes DO
	Write(' ');
	DEC(width)
    END
END WriteString;


BEGIN	(* TTIO initialization *)
    WITH inputstream DO
	EOF := FALSE;
	mustagain := FALSE;
	size  := INBUFLENGTH;
	index := INBUFLENGTH
    END
END TTIO.
