(*
    Title: 	KeyFile handling - Implementation
    LastEdit:	"Sun Sep 23 16:42:31 1984"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE KeyFile;
(* $T-, $R- *)

FROM Storage IMPORT ALLOCATE;
FROM Streams IMPORT Stream, ErrorCode, Get, Delete;
FROM SysStreams IMPORT sysOut;

FROM WriteF IMPORT WriteF2, H, S;

FROM FileH IMPORT
    OpenKeyFile;

FROM KeyTypes IMPORT KeyData;
FROM KeyInput IMPORT IsKeyData, NKeys, NextKey;
FROM LModule IMPORT
    MODULENAMESIZE, KEYSIZE, KEY, MODULENAME,
    IMPORTFRAMEPTR, IMPORTFRAME, MODULEPTR, Mtable, Pname;
FROM Config IMPORT Specified;
    
FROM Options IMPORT ListFiles;
FROM UtilError IMPORT ReportError;
FROM HashStrings IMPORT Assoc, EnterChars;
FROM SystemTypes IMPORT String;

IMPORT Slist;

VAR error: BOOLEAN;
VAR skey: Stream;


PROCEDURE ReadKeyFile(mname: MODULENAME): MODULEPTR;
VAR module: MODULEPTR;
    pname: String;
    nKeys, version: CARDINAL;
BEGIN
    error := FALSE;
    NEW(module);
    OpenAKeyFile(mname, module^.fname);
    IF IsKeyData(skey, version) THEN
        nKeys := NKeys(skey);
        HeaderFrame(module);
    ELSE
        error := TRUE;
    END (* if *);
    IF (NOT error) AND (nKeys > 1) THEN
        module^.imports := ImportFrame(nKeys-1);
    ELSE
        module^.imports := NIL;
    END;
    CloseKeyFile(skey);
    IF error THEN
        Pname(mname, pname);
        ReportError(TRUE, "%S has invalid format", S(pname), 0);
    END (* if *);
    RETURN module;
END ReadKeyFile;
    
    
PROCEDURE OpenAKeyFile(mname: MODULENAME; VAR fname: ARRAY OF CHAR);
VAR 
    imname: String;
    pname: String;
BEGIN
    IF Specified(mname, imname) THEN
        skey := OpenKeyFile(imname, fname);
    ELSE
        Pname(mname, pname);
        skey := OpenKeyFile(pname, fname);
    END;
    IF ListFiles THEN
        WriteF2(sysOut, "%S: %S\n", S(pname), H(fname));
    END (* if *);
END OpenAKeyFile;

PROCEDURE CloseKeyFile(k: Stream);
BEGIN
    Delete(k);
END CloseKeyFile;


PROCEDURE HeaderFrame(module: MODULEPTR);
VAR m: IMPORTFRAMEPTR;
BEGIN
	NEW(m);
        ModuleNameandKey(m, TRUE);
	module^.itself := m;
        Assoc(Mtable, m^.name, module);    (* to help 'LModule.Newmodule' *)
END HeaderFrame;

PROCEDURE ImportFrame(n: CARDINAL): IMPORTFRAMEPTR;
VAR i: CARDINAL;
VAR ip, ilist: IMPORTFRAMEPTR;
BEGIN
    ilist :=  NIL;
    FOR i := 1 TO n DO
        NEW(ip);
        ModuleNameandKey(ip, FALSE);
        Slist.AddE(ilist, ip);
    END;
    RETURN ilist;
END ImportFrame;


PROCEDURE ModuleNameandKey(ip: IMPORTFRAMEPTR; isMain: BOOLEAN);
VAR 
    keyData: KeyData;
BEGIN
    IF NextKey(skey, keyData) THEN
        IF EnterChars(Mtable, keyData.name, ip^.name) THEN
            Assoc(Mtable, ip^.name, NIL); (*  means its a 'new' module *)
        END (* if *);

        ip^.keys := keyData.keys;
        IF isMain THEN
            ip^.status := keyData.moduleStatus;
        END (* if *);
    ELSE
        error := TRUE;
    END (* if *);
END ModuleNameandKey;

END KeyFile .
