(*
    Title: 	Module handling - Implementation
    LastEdit:	"Wed Sep 12 15:44:39 1984"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE LModule;

FROM Storage IMPORT ALLOCATE;
FROM SysStreams IMPORT sysOut;
FROM WriteF IMPORT WriteF0, WriteF1, WriteF2, WriteF3, H, S;
FROM KeyFile IMPORT ReadKeyFile;
FROM Options IMPORT VCheck, CycleCheck, Debugging;
FROM UtilError IMPORT ReportError;
FROM Slist IMPORT AddE;
FROM HashStrings IMPORT EnterChars, StringAt, Retrieve, New;
FROM SystemTypes IMPORT String;

VAR CycleTable: ARRAY [0 .. 255] OF MODULEPTR;
VAR keysOk: BOOLEAN;

PROCEDURE Add(name: MODULENAME; level: CARDINAL; VAR mp: MODULEPTR);
(*  Adds the module 'name' to the 'modulelist' and to the
    dependency graph, initialising its 'importedbylist' to 'parent'.
    Then (recursively) 'Add's all those modules which it imports.
    Returns a pointer to the module record.
*)
VAR imp: MODULEPTR;
    ip: IMPORTFRAMEPTR;
    ipby: ISIMPORTEDBYPTR;
    ExceptionsName: MODULENAME;
    Exceptions: ARRAY [0..MODULENAMESIZE-1] OF CHAR;
BEGIN
    mp := ReadKeyFile(name);
(*  WriteF1(sysOut, "Add: %S\N", H(name)); *)
    AddE(modulelist, mp);
    IF CycleCheck THEN
        CycleTable[level] := mp;
    END (* if *);
    mp^.isimportedbylist := NIL;
    mp^.visited := FALSE;
    ip := mp^.imports;

    IF level=0 THEN (* always import this *)
        Exceptions := "Exceptions";
        IF EnterChars(Mtable, Exceptions, ExceptionsName) THEN
            Add(ExceptionsName, 1, imp);
        END (* if *);
    END (* if *);
    
    WHILE ip # NIL DO
        IF newmodule(ip^.name, imp) THEN
            Add(ip^.name, level+1, imp);
        ELSE
            IF CycleCheck AND (mp # imp) AND (level>0) THEN 
                CheckForCycles(level+1, imp); 
            END (* if *);
	END (* if *);
        
        ip^.module := imp;
        (*  note that this module imports 'ip^.name' *)
        NEW(ipby);
        ipby^.module := mp;
        ipby^.import := ip;
        AddE(imp^.isimportedbylist, ipby);
(*      WriteF2(sysOut, "  %S is imported by %S\N", 
                H(ip^.name), H(mp^.itself^.name)); *)
	ip := ip^.link;
    END (* while *);

    InitTable[InitTableUpb] := mp;
    INC(InitTableUpb);
(*  WriteF1(sysOut, "End Add: %S\N", H(name)); *)
END Add;

PROCEDURE newmodule(name: MODULENAME; VAR mp: MODULEPTR): BOOLEAN;
(*  Return TRUE if module 'name' is already on the module-list.
    When a module name is first entered into the hashtable 'Mtable',
    it has an 'associated' value of NIL stored.  When the keyfile
    for that module is processed the associated value is changed
    to point at the module record.
*)
BEGIN
    mp := MODULEPTR( Retrieve(Mtable, name) );
    RETURN mp=NIL;
END newmodule;

PROCEDURE CheckForCycles(level: CARDINAL; mp: MODULEPTR);
(*  If 'mp' already appears on the module stack, then there
    is a circularity which the user may like to know about,
    since it can cause run-time errors owing to initialisation
    failures.
*)
    VAR
        i, j: CARDINAL;
        pname: String;
    BEGIN
        CycleTable[level] := mp;
        FOR i := 0 TO level-1 DO
            IF CycleTable[i] = mp THEN
                ReportError(FALSE, "IMPORT cycle", 0, 0);
                WriteF0(sysOut, "    ");
                FOR j := i TO level DO
                    Pname(CycleTable[j]^.itself^.name, pname);
                    WriteF1(sysOut, "%S", S(pname));
                    IF j # level THEN
                        WriteF0(sysOut, ", ");
                    END (* if *);
                END (* for *);
                WriteF0(sysOut, "\n");
            END (* if *);
        END (* for *);
    END CheckForCycles;

PROCEDURE CheckKeys(mp: MODULEPTR): BOOLEAN;
BEGIN
    keysOk := TRUE;
    XCheckKeys(mp);
    RETURN keysOk;
END CheckKeys;

PROCEDURE XCheckKeys(mp: MODULEPTR);
VAR
    ipby: ISIMPORTEDBYPTR;
    ip: IMPORTFRAMEPTR;
    pname: String;
BEGIN
  IF mp^.visited=FALSE THEN
    ipby := mp^.isimportedbylist;
    IF VCheck THEN
        Pname(mp^.itself^.name, pname);
        WriteF1(sysOut, "Checking %S: ", S(pname)); PrintKey(mp^.itself);
    END;
    WHILE ipby # NIL DO		(* for each module which imports this one *)
        IF VCheck THEN
          Pname(ipby^.module^.itself^.name, pname);
  	  WriteF1(sysOut, " -- against %S: ", S(pname)); 
          PrintKey(ipby^.import);
        END;
        IF WrongKey(ipby^.import^.keys, mp^.itself^.keys) THEN
            VersionConflict(mp, ipby^.module);
        END;
        ipby := ipby^.link;
    END; (* while *)

    mp^.visited := TRUE;
    ip := mp^.imports;
    WHILE ip # NIL DO
    (*  all modules import themselves, so suppress repeated check *)
        IF ip^.module # mp THEN XCheckKeys(ip^.module); END;
        ip :=  ip^.link;
    END (* while *);
  END; (* if *)
    
END XCheckKeys;

PROCEDURE WrongKey(VAR k1, k2: KEY): BOOLEAN;
VAR i: CARDINAL;
BEGIN
    RETURN (k1[0] # k2[0]) OR (k1[1] # k2[1]);
END WrongKey;

PROCEDURE VersionConflict(m1, m2: MODULEPTR);
(*
    PROCEDURE WriteKey(m: IMPORTFRAMEPTR);
    VAR i: CARDINAL;
    BEGIN
        WriteF1("Key of %S: ", H(m^.name));
        FOR i:=1 TO KEYSIZE DO 
	    WriteF1("%O6 ", m^.key[i]);
        END;
	NewLine;
    END WriteKey;
*)
VAR pname1, pname2: String;
BEGIN
    keysOk := FALSE;
    Pname(m1^.itself^.name, pname1);
    Pname(m2^.itself^.name, pname2);
    ReportError(FALSE, "version conflict between %S and %S",
                H(pname1^), H(pname2^));
    (*  WriteKey(m1^.key); WriteKey(m2^.key);		      *)
END VersionConflict;


PROCEDURE PrintKey(ip: IMPORTFRAMEPTR);
VAR k: KEY;
BEGIN
    k := ip^.keys;
    WriteF2(sysOut, "key %O %O\N", k[0], k[1]);
END PrintKey;

PROCEDURE Pname(name: MODULENAME; VAR pname: String);
(*  'name' is really a hash-value, so retrieve the string
    stored at the hash value;
*)
    BEGIN
        pname := StringAt(Mtable, name);
    END Pname;

BEGIN
    modulelist := NIL;
    InitTableUpb := 0;
    Mtable := New(512);
END  LModule.

