(*
    Title: 	Run a Unix program - Implementation
    LastEdit:	"Tue May  1 10:45:43 1984"
    Author: 	Mick Jordan
		Cambridge University Computer Laboratory
*)

IMPLEMENTATION MODULE UnixRun;

FROM SYSTEM IMPORT ADR, UNIXCALL;
FROM UnixProcess IMPORT Exec, Wait, Fork;
FROM UnixIO IMPORT Close;
FROM UnixIOXtra IMPORT Dup;
IMPORT UnixCalls;

PROCEDURE Run(program: ARRAY OF CHAR; argv: ARRAY OF CARDINAL;
              VAR RunError: RUNERROR);
VAR
    reply, pid: INTEGER;
BEGIN
    argv[0] := ADR(program);
    RunError.code := ok;		(* assume success	      *)
    pid := Fork();
    IF pid<0 THEN SetRunError(RunError, fork, pid);

    ELSIF pid>0 THEN			(* parent		      *)
        RWait(RunError);
    ELSE				(* child		      *)  
        reply := Exec(program, argv);
        IF reply<0 THEN SetRunError(RunError, exec,  reply) END;
    END;
END Run;

PROCEDURE RWait(VAR runError: RUNERROR);
    VAR
        reply: INTEGER;
        exitstatus, termstatus: CARDINAL;
        status: INTEGER;
    BEGIN
        reply := Wait(status);
        IF reply<0 THEN SetRunError(runError, wait, reply);
        ELSE
            exitstatus := CARDINAL(status) DIV 400B;
            termstatus := CARDINAL(status) MOD 400B;
            IF (exitstatus # 0) THEN SetRunError(runError, exit, exitstatus);
    	    ELSIF (termstatus # 0) THEN SetRunError(runError, termination, termstatus);
            END;
        END;
    END RWait;

PROCEDURE PipeOpen(program: ARRAY OF CHAR; pipemode: PIPEMODE;
                   argv: ARRAY OF CARDINAL; VAR runError: RUNERROR): CARDINAL;

    PROCEDURE OneOf(a,b: CARDINAL): CARDINAL;
        BEGIN
            IF pipemode=READ THEN
                RETURN b;
            ELSE
                RETURN a;
            END (* if *);
        END OneOf;

    VAR
        p: ARRAY PIPEMODE OF CARDINAL;
        pid, rc: INTEGER;
    BEGIN
        argv[0] := ADR(program);
        runError.code := ok;
        rc := UNIXCALL(UnixCalls.pipe, ADR(p));
        IF rc < 0 THEN
            SetRunError(runError, pipe, rc);
            RETURN 0;
        END (* if *);
        
        pid := Fork();
        IF pid = 0 THEN (* child *)
            rc := Close(OneOf(p[WRITE], p[READ]));
            rc := Close(OneOf(0, 1));
            rc := Dup(OneOf(p[READ], p[WRITE]));
            rc := Close(OneOf(p[READ], p[WRITE]));
            rc := Exec(program, argv);
            IF rc<0 THEN SetRunError(runError, exec,  rc) END;
        ELSIF pid > 0 THEN (* parent *)
            rc := Close(OneOf(p[READ], p[WRITE]));
            RETURN OneOf(p[WRITE], p[READ]);
        ELSE
            SetRunError(runError, fork, pid);
        END (* if *);
        RETURN 0;
    END PipeOpen;

PROCEDURE PipeClose(fd: CARDINAL; VAR runError: RUNERROR);
    VAR
        rc: INTEGER;
    BEGIN
        rc := Close(fd);
        runError.code := ok;
        RWait(runError);
    END PipeClose;

PROCEDURE SetRunError(VAR RunError: RUNERROR; c: RUNErrorCode; reply: INTEGER);
BEGIN
    WITH RunError DO
        code := c;
        errno := reply;
    END;
END SetRunError;

PROCEDURE RunMessage(err: RUNERROR; VAR m: RUNMESSAGE);
BEGIN
    WITH err DO
    CASE code OF
        fork: m :=  "FORK failed";
    |   exec: m :=  "EXEC failed";
    |   wait: m :=  "WAIT failed";
    |   pipe: m :=  "pipe open failed";
    |   exit: m :=  "exited with status ";
    |   termination: 
              m := "terminated abnormally"
    END;
    END;
    
END RunMessage;

END UnixRun .
